;;  the cool cipher thing
;;  made by one sleepy ta and one wacked-out grad student
;;    sds jgm 9.15.93


;; Take a list of words dictionary and a cipher and decipher it.
;; Example:   (decipher '((g o a t) (p i g) (r a t)) '((b r k l) (f o a)))
;;    returns a list of characters that is the deciphered message, or nil
;;    if there is no match:    '((h i) (m o m))

(defun decipher (dictionary cipher)
  (uncipher cipher (update-xlate dictionary cipher nil)))


;; Either the first dictionary word is in the cipher, or it is not.
;; Try breaking the cipher with the first dictionary word.
;; If that doesn't work, try update-xlate again with the rest of dictionary.
;; Returns the xlate table, which is nil on failure or conflict.

(defun update-xlate (dictionary cipher xlate)
  (cond ((null cipher) xlate)    ;; Updating xlate impossible with no cipher.
        ((null dictionary) nil) ;; Updating xlate impossible with no dict.
	(t (or (match-word-to-cipher dictionary cipher cipher xlate)
	       (update-xlate (rest dictionary) cipher xlate)))))


;; Either the given word (first dictionary) is in the cipher, or it is not.
;; Try breaking the cipher by matching the word to the first cipher word.
;; If that doesn't work, try match-word-to-cipher again with rest of cipher.
;; Returns the xlate table, which is nil on failure or conflict.

(defun match-word-to-cipher (dictionary full-cipher cipher xlate)
  (if (null cipher) nil
      (or (replace-word-xlate dictionary full-cipher cipher xlate)
	  (match-word-to-cipher dictionary full-cipher (rest cipher) xlate))))


;; If the first dictionary word and the first cipher word match in length,
;; compute a new xlate table. If there is no conflict in the xlate table,
;; then solve the rest of the cipher by calling update-xlate with:
;;     1. the word taken out of the dictionary
;;     2. the cipher as updated with "completed" words removed.
;;     3. the update xlate table
;;
;; Returns the xlate table, which is nil on failure or conflict.

(defun replace-word-xlate (dictionary full-cipher cipher xlate)
  (setq xlate (replace-xlate (first dictionary) (first cipher) xlate))
  (if (null xlate) nil
      (update-xlate (rest dictionary)
		    (replace-cipher (first cipher) full-cipher xlate) xlate)))


;; Remove all words from the cipher that are completely deciphered by
;; the xlate table.
;; If the first word in the cipher is completely decipherable, then
;; we want to throw it out. Otherwise we want to keep it. Either throw
;; out or keep this thing, appended with the replace-cipher for the rest
;; of the cipher.

(defun replace-cipher (remove-word cipher xlate)
  (cond ((null cipher) nil)
	((equal remove-word (first cipher))
	 (replace-cipher remove-word (rest cipher) xlate))
	(t (append (list (first cipher))
		   (replace-cipher remove-word (rest cipher) xlate)))))

;; Is this word is fully decipherable by the xlate table?
;; If the first character cannot be deciphered, return nil. Otherwise
;; return the attempt on the rest of the characters.
(defun fully-decipherable (word xlate)
  (cond ((null word) t)      ;; Sure, the null word can be translated.
	((look-down (first word) xlate)
	 (fully-decipherable (rest word) xlate))
	(t nil)))  ;; Couldn't first out this character, so not decipherable.


;; Make a new xlate table given a word and a cipher, on the assumption
;; that this word matches the first word in the cipher. If this is untrue,
;; return nil.

(defun replace-xlate (word cipher-word xlate)
   (cond ((not (equal (length word) (length cipher-word))) nil)
         ((null word) xlate)    ;; No more updates to make to xlate table.
         ((and
	   (null (look-up (first word) xlate))
	   (null (look-down (first cipher-word) xlate)))
	   (replace-xlate (rest word) (rest cipher-word)
			  (append xlate (list (list (first word)
						    (first cipher-word))))))
	 ((equal (look-up (first word) xlate) (first cipher-word))
	  (replace-xlate (rest word) (rest cipher-word) xlate))
	 (t nil)))  ;; Must be a conflict. Look-up finds some wrong character.


;; Returns the translated character if it is inside the xlate table already

(defun look-up (character xlate)
   (cond ((null xlate) nil)
         ((equal character (first (first xlate))) (first (rest (first xlate))))
         (t (look-up character (rest xlate)))))

;; Returns the unciphered character if it is in the xlate table already

(defun look-down (character xlate)
   (cond ((null xlate) nil)
         ((equal character (first (rest (first xlate))))
	  (first (first xlate))) 
         (t (look-down character (rest xlate)))))

;; Returns a deciphered word

(defun word-look-down (word xlate)
   (if (null word) nil
       (append (list (look-down (first word) xlate))
             (word-look-down (rest word) xlate))))

;; Take a cipher and a translation table and decipher it

(defun uncipher (cipher xlate)
   (if (null cipher) 
       nil
     (append (list (word-look-down (first cipher) xlate))
             (uncipher (rest cipher) xlate))))

