(in-package 'user)

;; 1. Complete Exercise 10 at the end of Chapter 3 using the version
;;    of THEOREM listed in Figure 3.5, UNIFY listed in Figure 3.12,
;;    and the data abstractions for rules and logic variables listed
;;    in Chapter 3.  All of these functions are listed below.
;;    Pay special attention to your documentation and implementation
;;    of (VARSUBST PATTERN BINDINGS) which creates a new pattern 
;;    from PATTERN using the substitutions indicated in BINDINGS,
;;    and (RENAME PATTERN) which creates a new pattern from PATTERN
;;    having created brand new names for all of its variables.
;;    Call the modified version of THEOREM by the name PROLOG.
;; 2. Modify your solution to (1) to return the initial query with 
;;    variable bindings substituted where appropriate, e.g.,
;;    (theorem '((pleasing (? x))) rules1) => ((PLEASING BLOCK17))
;;    where the list of rules in rules1 is listed below.
;; 3. Implement negation as failure; a conjunct of the form 
;;    (not p) should succeed just in case p fails, e.g., 
;;    (theorem '((not (equal a b))) rules1) => ((NOT (EQUAL A B))).
;; 4. Extra Credit: In most implementations of Prolog, rules are
;;    indexed by their primary functor. The primary functor of a 
;;    rule is the predicate symbol in the consequent of the rule. 
;;    Modify your implementation of PROLOG so that the list of 
;;    rules used in attempting to reduce a given conjunct include
;;    exactly those whose primary functor is the same as that of the
;;    conjunct considered as degenerate rule.  (2 extra points)
;;    Hint: Preprocess rules to expedite indexing as in theorem.lisp.
;; 5. Contest: Come up with an interesting set of logic programming 
;;    rules (e.g., write a simple expert system for selecting courses
;;    for a computer science concentration).  First Prize: "Prolog 
;;    Programming for Artificial Intelligence" by Ivan Bratko, 
;;    Second Prize: "Introduction to Expert Systems" by Peter Jackson.
;;    
;;    Note:  Compiled versions of this file and a solution to the 
;;    assignment reside in the files logic.sbin and prolog.sbin.
;;    To load both of these compiled files type (load "prolog"). 
;;    To load the source for this file type (load "logic.lisp").

;; (PROLOG (B1 B2 ... Bn) RULES)  (nondeterministic version)
;;   nondeterministically choose R = (C if A1 A2 ... Am) from RULES
;;   rename all variables in R (RENAME R) = (C' if A1' A2' ... Am') 
;;   try to unify B1 and C'
;;   if B1 and C' successfully unify,
;;      let BDGS be the resulting most general substitution
;;        recurse on (VARSUBST (B2 ... Bn A1' ... Am') BDGS)

;; Note that the order of conjuncts involving negation as failure is
;; important. A query of the form ((not (green (? x))) (red (? x))) should
;; not be interpreted as (exists x, not(green(x)) and red(x)) but rather
;; (exists x, not(green(x))) and (exists x, red(x))).  Keep in mind that
;; negation and `negation as failure' are two very different things.
;; 
;; > (prolog '((not (green (? x))) (red (? x)))
;;           '(((green a)) ((red a)) ((red b))))
;; NIL
;; > (prolog '((red (? x)) (not (green (? x))))
;;           '(((green a)) ((red a)) ((red b))))
;; ((RED B) (NOT (GREEN B)))

;; The following makes the results corresponding to list structures
;; printed by the interpreter look a little more comprehensible. 
(setq *print-pretty* t)

;; Data abstraction for rules. 
(defun rule-consequent (rule) (first rule))
(defun rule-antecedents (rule) (cddr rule))

;; Goal-reduction type, propositional theorem prover.
(defun theorem (conjuncts rules)
  (aux-theorem conjuncts rules rules))
(defun aux-theorem (conjuncts untried rules)
  (cond ((null conjuncts) t)
        ((null untried) nil)
        (t (or (and (eq (first conjuncts) 
                        (rule-consequent (first untried)))
                    (aux-theorem 
                     (append (rule-antecedents (first untried))
                             (rest conjuncts)) 
                     rules rules))
               (aux-theorem conjuncts (rest untried) rules)))))

;; Data abstraction for logic variables.
(defun make-var (var) (list '? var))
(defun is-var (x) (and x (listp x) (eq (first x) '?)))

;; Simple and inefficient implementation of unification.
(defun unify (p q)
  (aux-unify p q '((unify t))))
(defun aux-unify (p q bdgs)
  (let ((d (disagreement p q)))
    (cond ((null d) bdgs)
	  ((occursp (first d) (second d)) nil)
          ((not (is-var (first d))) nil)
          (t (aux-unify 
              (subst (second d) (first d) p :test #'equal)
              (subst (second d) (first d) q :test #'equal)
              (cons d bdgs))))))
(defun disagreement (p q)
  (cond ((is-var p)
	 (if (equal p q) nil (list p q)))
	((is-var q)
	 (if (equal p q) nil (list q p)))
	((or (not (consp p)) (not (consp q)))
	 (if (eq p q) nil (list p q)))
	(t (or (disagreement (first p) (first q))
	       (disagreement (rest p) (rest q))))))

;; In Prolog, '(foo (? x) (baz (? x))) and '(foo (? y) (? y))) will
;; unify. This is because Prolog does not check to see if the expression
;; that a variable is bound to contains that variable.  In prolog, this
;; is considered a feature not a bug.  In our implementation, however,
;; we have added an `occurs check' to prevent such unifications. 
(defun occursp (x y)
  (cond ((or (null y) (not (consp y))) nil)
	((is-var y) (equal x y))
	(t (or (occursp x (first y)) (occursp x (rest y))))))

;; Rules for the example in the book.
(setq rules1 `(((red block41))
	       ((red block17))
	       ((sphere block17))
	       ((cube block41))
	       ((pleasing (? x1)) if (red (? x1)) (smooth (? x1)))
	       ((smooth (? x2)) if (sphere (? x2)))))

;; > (prolog '((pleasing (? x))) rules1)
;; ((PLEASING BLOCK17))

;; See rules.lisp for additional examples of rule sets.

