;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information
(in-package 'user)

(load "logic")

(defun varsubst (p bdgs)
  (let ((bdg (first bdgs)))
    (if bdgs 
        (varsubst (subst (second bdg) 
                         (first bdg) p :test #'equal) 
                  (rest bdgs))
      (lisp p))))

(defun rename (p) 
  (mapc #'(lambda (v)
            (setq p (subst (make-var (gensym)) v p
                           :test #'equal)))
        (vars p))
    p)

(defun vars (p)
  (cond ((not (consp p)) nil)
        ((is-var p) (list p))
        (t (append (vars (first p))
                   (vars (rest p))))))

(defun lisp (p) 
  (cond ((not (consp p)) p)
        ((eq (first p) 'lisp) 
         (if (null (vars p)) (eval (second p)) p))
        (t (cons (lisp (first p)) (lisp (rest p))))))

(defun convert (rules)
  (let ((preds ()))
    (mapc #'(lambda (pred)
              (if (not (member pred preds))
                  (setq preds (cons pred preds))))
          (mapcar #'(lambda (rule) (first (first rule))) rules))
    (mapcar #'(lambda (pred)
                (cons pred
                      (findall rules
                               #'(lambda (rule)
                                   (eq pred (first (first rule)))))))
            preds)))

(defun findall (list test)
  (do ((items list (rest items)) (results nil))
      ((null items) results)
      (if (funcall test (first items)) 
          (setq results (adjoin (first items) results
                                :test #'equal)))))

(defun prolog (query rules)
  (setq rules (convert rules))
  (aux-prolog (lisp query) 
              (rest (assoc (first (first query)) rules))
              rules query))
(defun aux-prolog (conjuncts untried rules query)
  (cond ((null conjuncts) query)
        ((eq (first (first conjuncts)) 'not)
         (and (let ((query (list (second (first conjuncts)))))
                (not (aux-prolog 
                      query 
                      (rest (assoc (first (first query)) rules))
                      rules query)))
              (aux-prolog (rest conjuncts) 
                          (rest (assoc (first (first (rest conjuncts))) rules))
                          rules query)))
        ((null untried) nil)
        (t (let ((rule (rename (first untried))) bdgs reduced)
             (setq bdgs (unify (first conjuncts) 
                               (rule-consequent rule)))
             (setq reduced (varsubst (append (rule-antecedents rule)
                                             (rest conjuncts))
                                     bdgs))
             (or (and bdgs 
                      (aux-prolog 
                       reduced
                       (rest (assoc (first (first reduced)) rules))
                       rules 
                       (varsubst (varsubst query bdgs) bdgs)))
                 (aux-prolog conjuncts (rest untried) rules query))))))


;; Copyright 1994, Brown University, Providence, RI
;; 
;; Permission to use and modify this software and its documentation for any
;; purpose other than its incorporation into a commercial product is hereby
;; granted without fee.  Permission to copy and distribute this software and its
;; documentation only for non-commercial use is also granted without fee,
;; provided, however, that the above copyright notice appear in all copies, that
;; both that copyright notice and this permission notice appear in supporting
;; documentation, that the name of Brown University not be used in advertising or
;; publicity pertaining to distribution of the software without specific, written
;; prior permission, and that the person doing the distribution notify Brown
;; University of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this software for
;; any purpose.  It is provided "as is" without express or implied warranty.
;; Brown University requests notification of any modifications to this software
;; or its documentation.
;; 
;; Send the following redistribution information:
;; 
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;; 
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;; 
;; 		or
;; 
;; 	brusd@cs.brown.edu
;; 
;; We will acknowledge all electronic notifications.
