;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information
(in-package 'user)

;; A set of rules for handling membership queries. 
(setq rules2 '(((member (? x) (cons (? x) (? y))))
	       ((member (? x) (cons (? z) (? y))) if (member (? x) (? y)))))

;; > (prolog '((member a (cons b (cons a ())))) rules2)
;; ((MEMBER A (CONS B (CONS A NIL))))
;; > (prolog '((member a (cons b (cons (? x) ())))) rules2)
;; ((MEMBER A (CONS B (CONS A NIL))))

;; Why won't the following implementation of equivalence work?
(setq rules3 '(((equal a b))
	       ((equal b c))
	       ;; Reflexive
       	       ((equal (? x) (? x)))
	       ;; Symmetric
	       ((equal (? x) (? y)) if (equal (? y) (? x)))
	       ;; Transitive
	       ((equal (? x) (? y)) if (equal (? x) (? z)) 
		(equal (? z) (? y)))))

;; Here is one attempt at getting around the problem with RULES3.
;; This still isn't quite right. Is there an easy fix to get it right?
(setq rules4 '(((= a b)) ((= b c)) ((= c d)) ((= e d))
	       ((equal (? x) (? y)) if (reflexive (? x) (? y)))
	       ((equal (? x) (? y)) if (symmetric (? x) (? y)))
	       ((equal (? x) (? y)) if (transitive (? x) (? y) ()))
       	       ((reflexive (? x) (? x)))
	       ((symmetric (? x) (? y)) if (= (? y) (? x)))
	       ((transitive (? x) (? x) (? w)))
	       ((transitive (? x) (? y) (? w)) if 
		(= (? x) (? z)) (not (member (? z) (? w)))
		(transitive (? z) (? y) (cons (? z) (? w))))
	       ((member (? x) (cons (? x) (? y))))
	       ((member (? x) (cons (? z) (? y))) if (member (? x) (? y)))))
		
;; > (prolog '((equal a d)) rules4)
;; ((EQUAL A D))
;; > (prolog '((equal a h)) rules4)
;; NIL

;; Here's a set of rules for equivalence that works.
(setq rules5 '(((= a b)) ((= b c)) ((= c d)) ((= e d))
               ((equal (? x) (? y)) if (reflexive (? x) (? y)))
               ((equal (? x) (? y)) if (symmetric (? x) (? y)))
               ((equal (? x) (? y)) if (transitive (? x) (? y) ()))
               ((reflexive (? x) (? x)))
               ((symmetric (? x) (? y)) if (= (? y) (? x)))
               ((either (? x) (? y)) if (= (? x) (? y)))
               ((either (? x) (? y)) if (= (? y) (? x)))
               ((transitive (? x) (? x) (? w)))
               ((transitive (? x) (? y) (? w)) if 
                (either (? x) (? z)) (not (member (? z) (? w)))
                (transitive (? z) (? y) (cons (? z) (? w))))
               ((member (? x) (cons (? x) (? y))))
               ((member (? x) (cons (? z) (? y))) if (member (? x) (? y)))))

;; Here are the rules for implementing LAST.
(setq rules6 '(((last (? x) (cons (? x) nil)))
               ((last (? x) (cons (? z) (? y))) if (last (? x) (? y)))))

;; > (prolog '((last (? x) 
;;                   (cons a (cons b nil)))) 
;;           rules6)
;; ((LAST B (CONS A (CONS B NIL))))

;; Here are the rules for implementing MERGE.
(setq rules7 '(((merge nil (? x) (? x)))
               ((merge (? x) nil (? x)))
               ((merge (cons (? x1) (? y1)) 
                       (cons (? x2) (? y2)) 
                       (cons (? x1) (? y3)))
                if (< (? x1) (? x2))
                   (merge (? y1)
                          (cons (? x2) (? y2)) 
                          (? y3)))
               ((merge (cons (? x1) (? y1)) 
                       (cons (? x2) (? y2)) 
                       (cons (? x2) (? y3))) 
                if (not (< (? x1) (? x2)))
                   (merge (cons (? x1) (? y1)) 
                          (? y2)
                          (? y3)))
               ((< 1 2))
               ((< 2 3))
               ((< 3 4))))

;; > (prolog '((merge (cons 1 (cons 3 nil)) 
;;                    (cons 2 (cons 4 nil)) 
;;                    (? x)))
;;           rules7)
;; ((MERGE (CONS 1 (CONS 3 NIL))
;;         (CONS 2 (CONS 4 NIL))
;;         (CONS 1
;;               (CONS 2
;;                     (CONS 3 (CONS 4 NIL))))))

;; Here's an example showing how to invoke Lisp inside of a rule
;; to handle < more generally than in the above rule set. 
(setq rules8 '(((merge nil (? x) (? x)))
               ((merge (? x) nil (? x)))
               ((merge (cons (? x1) (? y1)) 
                       (cons (? x2) (? y2)) 
                       (cons (? x1) (? y3)))
                if (< (? x1) (? x2))
                   (merge (? y1)
                          (cons (? x2) (? y2)) 
                          (? y3)))
               ((merge (cons (? x1) (? y1)) 
                       (cons (? x2) (? y2)) 
                       (cons (? x2) (? y3))) 
                if (not (< (? x1) (? x2)))
                   (merge (cons (? x1) (? y1)) 
                          (? y2)
                          (? y3)))
               ((< (? x) (? y)) if (lisp (less (? x) (? y))))
               ((true))))

(defun less (a b)
    (if (< a b) '(true) '(false)))

;; > (prolog '((merge (cons 1 (cons 3 (cons 11 nil)))
;;                    (cons 2 (cons 4 (cons 9 nil)))
;;                    (? x)))
;;         rules8)
;; ((MERGE (CONS 1 (CONS 3 (CONS 11 NIL)))
;;         (CONS 2 (CONS 4 (CONS 9 NIL)))
;;         (CONS 1
;;               (CONS 2
;;                     (CONS 3
;;                           (CONS 4
;;                                 (CONS 9 (CONS 11 NIL))))))))


;; Copyright 1994, Brown University, Providence, RI
;; 
;; Permission to use and modify this software and its documentation for any
;; purpose other than its incorporation into a commercial product is hereby
;; granted without fee.  Permission to copy and distribute this software and its
;; documentation only for non-commercial use is also granted without fee,
;; provided, however, that the above copyright notice appear in all copies, that
;; both that copyright notice and this permission notice appear in supporting
;; documentation, that the name of Brown University not be used in advertising or
;; publicity pertaining to distribution of the software without specific, written
;; prior permission, and that the person doing the distribution notify Brown
;; University of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this software for
;; any purpose.  It is provided "as is" without express or implied warranty.
;; Brown University requests notification of any modifications to this software
;; or its documentation.
;; 
;; Send the following redistribution information:
;; 
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;; 
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;; 
;; 		or
;; 
;; 	brusd@cs.brown.edu
;; 
;; We will acknowledge all electronic notifications.
