;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

;; Matching and unification from the logic chapter.

(define (make-var var) (list '? var))

(define (is-var x) (and (pair? x) (eq? (first x) '?)))


;; Determine if a constant and a pattern match.
(define (match const pat) (aux-match const pat '((match #t))))


;; Keep track of the bindings established so far.
(define (aux-match const pat bdgs)
  (cond ((not bdgs) #f)
        ((atom pat) (if (eq? pat const) bdgs #f))
        ((is-var pat)
         (let ((bdg (assoc2 pat bdgs equal? (lambda (x) x))))
           (cond (bdg (aux-match const (second bdg) bdgs))
                 (else (cons (list pat const) bdgs)))))
        ((or (atom const) (null? const)) #f)
        (else
         (aux-match
           (first const)
           (first pat)
           (aux-match (rest const) (rest pat) bdgs)))))


(define (unify p q) (aux-unify p q '((unify #t))))


;; Keep track of the bindings established so far.
(define (aux-unify p q bdgs)
  (let ((d (disagreement p q)))
    (cond ((not d) bdgs)
          ((occursp (first d) (second d)) #f)
          ((not (is-var (first d))) #f)
          (else
           (aux-unify
             (subst2 (second d) (first d) p equal? (lambda (x) x))
             (subst2 (second d) (first d) q equal? (lambda (x) x))
             (cons d bdgs))))))


;; Return subexpressions where P and Q first disagree.
(define (disagreement p q)
  (cond ((is-var p) (if (equal? p q) #f (list p q)))
        ((is-var q) (if (equal? p q) #f (list q p)))
        ((or (not (pair? p)) (not (pair? q))) (if (eq? p q) #f (list p q)))
        (else
         (or (disagreement (first p) (first q))
             (disagreement (rest p) (rest q))))))


;;  Check to see of if X appears anywhere in Y.
(define (occursp x y)
  (cond ((or (null? y) (not (pair? y))) #f)
        ((is-var y) (equal? x y))
        (else (or (occursp x (first y)) (occursp x (rest y))))))


;; Test MATCH and UNIFY.
(define (test)
  (and (match '(loves (dog fred) fred) '(loves (? x) (? y)))
       (match '(loves (dog fred) mary) '(loves (dog (? x)) (? y)))
       (not (match '(loves (dog fred) fred) '(loves (? x) (? x))))
       (not (match '(loves (dog fred) mary) '(loves (dog (? x)) (? x))))
       (unify '(loves (dog (? x)) (dog fred)) '(loves (? z) (? z)))))



;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;;
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;;
;; 		or
;;
;; 	brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.

