;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

;; Simple theorem prover from the logic chapter.

;; Abstract data types for rules.
(define (consequent rule) (first rule))

(define (antecedents rule) 
  (if (null? (rest rule)) (list) (rest (rest rule))))


;; Try to prove a conjunction using a list of rules.
(define (theorem conjuncts rules) (aux-theorem conjuncts rules rules))


;; Keep track of the conjuncts that you have yet to prove,
;; the rules that you have yet to apply to the first conjunct, and
;; the set of all rules for proving the rest of the conjuncts.
(define (aux-theorem conjuncts untried rules)
  (cond ((null? conjuncts) #t)
        ((null? untried) #f)
        ((and (eq? (first conjuncts) (consequent (first untried)))
              (aux-theorem
                (append (antecedents (first untried)) (rest conjuncts))
                rules
                rules)))
        (else (aux-theorem conjuncts (rest untried) rules))))


;; Test the code.
(let ((rules '((r if p q) (q if s #t) (p) (q) (s) (#t))))
  (set! test (lambda () (theorem '(r) rules))))



;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;;
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;;
;; 		or
;;
;; 	brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.

