;; CS 141 Assignment 6
;; This file contains all the data abstraction methods for making and
;; accessing a grid.

(in-package 'user)
(load "/course/cs141/lib/asgn6/search.sbin")
;; We begin with some useful data abstractions.

;; A NODE in the search space is a LOCation and a 
;; path represented as a list of LOCations. 
(defun make-NODE (loc path) (list loc path))
(defun NODE-loc (node) (first node))
(defun NODE-path (node) (second node))

;; Data abstraction for grid LOCations.
(defun make-LOC (i j) (list (list i j) nil nil))
(defun LOC-coord (loc) (first loc))
;; The OCC slot is used to indicate occupied LOCations.
(defun LOC-occ (loc) (second loc))
(defun set-LOC-occ (loc val) (setf (second loc) val))
;; The MARK slot is used in searching graphs.
(defun LOC-mark (loc) (third loc))
(defun set-LOC-mark (loc mark) (setf (third loc) mark))

;; Data abstraction for coordinate pairs.
(defun make-COORD (i j) (list i j))
;; Returns the x coordinate given a coordinate pair.
(defun COORD-x (coord) (first coord))
;; Returns the y coordinate given a coordinate pair.
(defun COORD-y (coord) (second coord))

;; Data abstraction for grids. 
;; Note that GRID LOCations are zero based.
(defun make-GRID (n m)
  (do ((i 0 (+ i 1)) (rows ()))
      ((= i n) (reverse rows))
    (setq rows
          (cons (do ((j 0 (+ j 1)) (cols ()))
                    ((= j m) (reverse cols))
                  (setq cols 
                        (cons (make-LOC i j) cols)))
                rows))))

;; Return LOCation (i j) in a GRID.
(defun GRID-ref (grid i j)
  (let ((num-rows (length grid))
        (num-cols (length (first grid))))
    ;; Make sure the indices are within bounds.
    (if (or (>= i num-rows) (< i 0)
            (>= j num-cols) (< j 0))
        nil
      (nth j (nth i grid)))))

;; Reset all the LOCation marks to nil.
(defun GRID-clear (grid)
  (dotimes (i (length grid))
    (dotimes (j (length (first grid)))
      (set-LOC-mark (GRID-ref grid i j) nil))))

;; For debugging purposes
;; Displays the grid
(defun GRID-display (grid)
  (format t "~%       ")
  (dotimes (i (length (first grid))) (format t "~D   " i))
  (dotimes (i (length grid))
    (format t "~%    ~D  " i)
    (dotimes (j (length (first grid)))
      (if (LOC-occ (GRID-ref grid i j))
        (format t "X   ")
        (format t "+   ")))))

;; Show the marking done in the grid
(defun GRID-marks (grid)
  (format t "~%       ")
  (dotimes (i (length (first grid))) (format t "~D   " i))
  (dotimes (i (length grid))
    (format t "~%    ~D  " i)
    (dotimes (j (length (first grid)))
      (if (LOC-occ (GRID-ref grid i j))
        (format t "X   ")
        (if (LOC-mark (GRID-ref grid i j))
          (format t "M   ")
          (format t "+   "))))))


