;;; This file contains the search routines found in the book

(in-package 'user)
;; This depth-first search procedure takes a set of nodes
;; a function that tests to see if a node is a goal node
;; and a function that takes a node and returns its children
(defun dfs (nodes goalp next)
  (cond ((null nodes) nil)
	;; Return the first node if it is a goal node
	((funcall goalp (first nodes)) (first nodes))
	;; Put the children in the front of the list
	(t (dfs (append (funcall next (first nodes))
			(rest nodes))
		goalp
		next))))

;; This breadth-first search procedure takes a set of nodes,
;; a function that tests to see if a node is a goal node
;; and a function that takes a node and returns its children
(defun bfs (nodes goalp next)
  (cond ((null nodes) nil)
	;; Return the first node if it is a goal node
	((funcall goalp (first nodes)) (first nodes))
	;; Put the children in the front of the list
	(t (bfs (append (rest nodes)
			(funcall next (first nodes)))
		goalp
		next))))

;; This procedure searches depth first to a fixed maximum depth
;; The procedure takes as arguments a single node, a function
;; that tests to see if a node is a goal node, a function that
;; takes a node and returns its children, an integer specifying
;; the current depth, and an integer specifying the maximum depth.
(defun dfs-fd (node goalp next depth max)
  (cond ((funcall goalp node) node)
	;; Quit if you are at the maximum depth
	((= depth max) nil)
	;; Increment the depth and extend the search
	(t (some #'(lambda (n)
		     (dfs-fd n goalp next (+ 1 depth) max))
		 (funcall next node)))))

(defun ids (start goalp next depth grid)
  (if (>= depth 50) nil
    (progn (GRID-clear grid)
    ;; First search to the current depth.
       (or (dfs-fd start goalp next 0 depth)
	   ;; Then look deeper if unnecessary
	   (ids start goalp next (+ 1 depth) grid)))))

;; This best-first search procedure takes a set of nodes
;; a function that tests to see if a node is a goal node,
;; a function that takes a node and returns its children
;; and a function that compares the value of two nodes.

(defun best (nodes goalp next comparep)
  (cond ((null nodes) nil)
	;; Return the first node if it is a goal node.
	((funcall goalp (first nodes)) (first nodes))
	;; Append the children to the set of old nodes
	(t (best (sort (append (funcall next (first nodes))
			       (rest nodes))
		       comparep)
		 goalp
		 next
		 comparep))))


