;; This is what the students are supposed to write.

(in-package 'user)
(load "/course/cs141/lib/asgn6/data.lisp")
;; This section contains all the generic code

;; The following routines refer to a node in the search space which is
;; not the same as a node in the grid.
;; Each node in the search space contains
;; (((coord of this node in the grid) (path taken to this node)) ........)
;;

;; Returns the coordinate of the specified node in the search space
(defun coord (node-in-search-space)
  (first node-in-search-space))

;; Returns the path taken to reach that node in the search space
(defun path (node-in-search-space)
  (second node-in-search-space))

;; This is the standard translation to go from one node to its neighbours
(setq translation '((1 0) (0 1) (-1 0) (0 -1)))

(defun x-translation (translation) (first translation))
(defun y-translation (translation) (second translation))

;; This returns a list of neighbours of this coordinate
(defun return-neighbours (translation coord)
  (cond ((null translation) nil)
	((null coord) nil)
	(t (append (list (make-coord (+ (COORD-X coord)
					(x-translation (first translation)))
				     (+ (COORD-Y coord)
					(y-translation (first translation)))))
		   (return-neighbours (rest translation) coord)))))

;; This returns the list of next states in the state space to look into
;; It calls return-neighbours in order to determine the list of neighbours
;; Then it calls return-path-aux to create the list of next states.

(defun return-next (translation grid coord path)
  (progn
    (set-LOC-mark (GRID-ref grid (COORD-x coord)  (COORD-y coord)) t)
    (let ((neighbours (return-neighbours translation coord)))
      (return-path-aux neighbours grid coord path))))

(defun return-path-aux (neighbours grid coord path)
  (if (or (null neighbours) (null coord)) nil
    (let ((node (GRID-ref grid (COORD-x (first neighbours))
			  (COORD-y (first neighbours)))))
      (if (and node (NULL (LOC-occ node)) (NULL (LOC-mark node)))
	  (progn
	    (set-LOC-mark node t)
	    (append (list (cons (first neighbours)
				(list (append path (list coord)))))
		    (return-path-aux (rest neighbours) grid
				     coord path)))
	(return-path-aux (rest neighbours) grid
			 coord
			 path)))))

;; For counting the number of nodes visited to reach the goal
(let ((n 0))
  (defun reset-counter()
    (setq n 0))
  (defun add-counter() (setq n (+ 1 n)))
  (defun show-counter() n))

;; Square function
(defun sqr (x) (* x x))

;; Returns the distance squared. Do not calculate the actual distance
;; since that will take extra time
(defun distance (c1 c2)
  (+ (sqr (- (COORD-x c1) (COORD-x c2))) (sqr (- (COORD-y c1) (COORD-y c2)))))

;; My depth first search function which takes a start coordinate and
;; an end-coordinate and a grid and finds a path using dfs.
(defun my-dfs (start goal grid)
  (progn 
    (GRID-clear grid)
    (reset-counter)
    (list (dfs (list (list start nil))
	       #'(lambda (s) (progn
			       (add-counter)
			       (equal goal (coord s))))
	       #'(lambda (s) (return-next translation grid
					  (coord s)
					  (path s))))
	  (show-counter))))

;; My Breadth First search function which takes a start coordinate and
;; an end-coordinate and a grid and finds a path using bfs.

(defun my-bfs (start goal grid)
  (progn
    (GRID-clear grid)
    (reset-counter)
    (list (bfs (list (list start nil))
	       #'(lambda (s) (progn
			       (add-counter)
			       (equal goal (coord s))))
	       #'(lambda (s) (return-next translation grid
					  (coord s)
					  (path s))))
	  (show-counter))))

;; My Iterative Deepening search routine which takes a start coordinate
;; and an end-coordinate and a grid and finds a path using ids.

(defun my-ids (start goal grid)
  (GRID-clear grid)
  (reset-counter)
  (list (ids (list start nil)
	     #'(lambda (s) (progn
			     (add-counter)
			     (equal goal (coord s))))
	     #'(lambda (s) (return-next translation grid
					(coord s)
					(path s)))
	     0
	     grid)
	(show-counter)))

;; Comparator used to determine which coordinate is nearer to the goal.
(defun Comparator (c1 c2 goal)
  (< (distance c1 goal) (distance c2 goal)))

;; My Best first search routine which takes a start coordinate and an end-
;;coordinate and a grid and finds a path using Best-first search.

(defun my-best (start goal grid)
  (progn
    (GRID-clear grid)
    (reset-counter)
    (list (best (list (list start nil))
		#'(lambda (s) (progn
				(add-counter)
				(equal goal (coord s))))
		#'(lambda (s) (return-next translation grid
					   (coord s)
					   (path s)))
		#'(lambda (s1 s2) (Comparator (coord s1) (coord s2) goal)))
	  (show-counter))))


;; Answer to Part 2 of Assignment 6

(let ((list-of-goals))
  (defun set-goals(goals)
    (setq list-of-goals goals))
  ;; Checks if the given coordinate is a goal and removes it from the list
  ;; if it is
  (defun remove-goal (goal)
    (setq list-of-goals (remove-goal-aux goal list-of-goals)))
  (defun remove-goal-aux (goal goal-list)
    (cond ((null goal-list) nil)
	  ((equal goal (first goal-list)) (rest goal-list))
	  (t (append (list (first goal-list))
		     (remove-goal-aux goal (rest goal-list))))))
  (defun goal-achieved ()
    (if (null list-of-goals) t nil))
  ;; Breadth-first search used when searching for a set of goals.
  (defun my-bfs-2 (start goals grid)
    (progn
      (GRID-clear grid)
      (set-goals goals)
      (bfs (list (list start nil))
	   #'(lambda (s) (progn
			   (remove-goal (coord s))
			   (goal-achieved)))
	   #'(lambda (s) (return-next translation grid
				      (coord s)
				      (path s)))))))


;; Answer to Part 3 of Assignment 6
;; Hill-climbing Search
(defun hcs (states goalp next comparator)
  (cond ((null states) nil)
        ((funcall goalp (car states)) (car states))
        (t (hcs (append (sort (funcall next (car states))
                              comparator)
                        (cdr states))
                goalp
                next
                comparator))))









