;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

;;; This file contains the search routines found in the book

;; This depth-first search procedure takes a set of nodes
;; a function that tests to see if a node is a goal node
;; and a function that takes a node and returns its children
(define (dfs nodes goalp next)
  (cond ((null? nodes) #f)
        ((goalp (first nodes)) (first nodes))
        (else (dfs (append (next (first nodes)) (rest nodes)) goalp next))))


;; This breadth-first search procedure takes a set of nodes,
;; a function that tests to see if a node is a goal node
;; and a function that takes a node and returns its children
(define (bfs nodes goalp next)
  (cond ((null? nodes) #f)
        ((goalp (first nodes)) (first nodes))
        (else (bfs (append (rest nodes) (next (first nodes))) goalp next))))


;; This procedure searches depth first to a fixed maximum depth
;; The procedure takes as arguments a single node, a function
;; that tests to see if a node is a goal node, a function that
;; takes a node and returns its children, an integer specifying
;; the current depth, and an integer specifying the maximum depth.
(define (dfs-fd node goalp next depth max)
  (cond ((goalp node) node)
        ((= depth max) #f)
        (else
         (some2 (lambda (n) (dfs-fd n goalp next (+ 1 depth) max))
                (next node)))))


(define (ids start goalp next depth)
  (or (dfs-fd start goalp next 0 depth) (ids start goalp next (+ 1 depth))))


;; This best-first search procedure takes a set of nodes
;; a function that tests to see if a node is a goal node,
;; a function that takes a node and returns its children
;; and a function that compares the value of two nodes.

(define (best nodes goalp next comparep)
  (cond ((null? nodes) #f)
        ((goalp (first nodes)) (first nodes))
        (else
         (best (sort2 (append (next (first nodes)) (rest nodes))
                      comparep
                      (lambda (x) x))
               goalp
               next
               comparep))))



;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;;      Name:
;;      Organization:
;;      Address (postal and/or electronic):
;;
;; To:
;;      Software Librarian
;;      Computer Science Department, Box 1910
;;      Brown University
;;      Providence, RI 02912
;;
;;              or
;;
;;      brusd@cs.brown.edu

;;
;; We will acknowledge all electronic notifications.

