;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

(load "search.scm")


;; Abstract data types for testing search routine.
(define (make-tree label value children) (list 'tree label value children))

(define (tree-label tree) (first (rest tree)))

(define (tree-value tree) (first (rest (rest tree))))

(define (tree-children tree) (first (rest (rest (rest tree)))))

(define (tree-print tree) (display (tree-label tree)))

(define (tree-comparator tree1 tree2)
  (< (tree-value tree1) (tree-value tree2)))


;; Build a tree and test each of the search routines.
;;                        (tree 0 2)
;;                          /    \
;;                         /      \
;;                        /        \
;;                 (tree 1 3)      (tree 4 1)
;;                 /     \            /     \
;;                /       \          /       \
;;               /         \        /         \
;;         (tree 2 4)  (tree 3 4) (tree 5 2)  (tree 6 0)
(let ((tree (make-tree
              0
              2
              (list (make-tree
                      1
                      3
                      (list (make-tree 2 4 '()) (make-tree 3 4 '())))
                    (make-tree
                      4
                      1
                      (list (make-tree 5 2 '()) (make-tree 6 0 '())))))))
  (set! test
    (lambda ()
      (display "depth-first search:")
      (newline)
      (dfs (list tree)
           (lambda (x) (tree-print x) (= 6 (tree-label x)))
           tree-children)
      (newline)
      (display "breadth-first search:")
      (newline)
      (bfs (list tree)
           (lambda (x) (tree-print x) (= 6 (tree-label x)))
           tree-children)
      (newline)
      (display "iterative-deepening search:")
      (newline)
      (ids tree
           (lambda (x) (tree-print x) (= 6 (tree-label x)))
           tree-children
           1)
      (newline)
      (display "best-first search:")
      (newline)
      (best (list tree)
            (lambda (x) (tree-print x) #f)
            tree-children
            tree-comparator))))



;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;;      Name:
;;      Organization:
;;      Address (postal and/or electronic):
;;
;; To:
;;      Software Librarian
;;      Computer Science Department, Box 1910
;;      Brown University
;;      Providence, RI 02912
;;
;;              or
;;
;;      brusd@cs.brown.edu

;;
;; We will acknowledge all electronic notifications.

