(in-package 'user)

;; Performance is defined as the percentage of correctly classified examples.
(defun performance (node examples)
  (let ((m 0))
    (mapc #'(lambda (example) 
              (if (correctp node example) (setq m (+ 1 m))))
          examples)
    (/ m (length examples))))

;; A decision tree is correct with respect to an example if the example
;; is stored in a terminal of the tree marked with the same class.  
(defun correctp (node example)
  (if (null (NODE-children node))
      (equal (EXAMPLE-class example) (NODE-class node))
    (correctp (car (findall (NODE-children node)
                            #'(lambda (c)
                                (member (NODE-feature c)
                                        (EXAMPLE-features example)
                                        :test #'equal))))
              example)))

;; Compute the number of nodes in a decision tree.
(defun size (node)
  (+ 1 (apply #'+ (mapcar #'size (NODE-children node)))))

;; Compute the maximum depth of a decision tree. 
(defun depth (node)
  (if (null (NODE-children node)) 0
    (+ 1 (apply #'max (mapcar #'depth (NODE-children node))))))

