;; File: pdp.lisp

(in-package 'user)

(setq or  '(((0 0) 0) ((1 0) 1) ((0 1) 1) ((1 1) 1))
      and '(((0 0) 0) ((1 0) 0) ((0 1) 0) ((1 1) 1))
      xor '(((0 0) 0) ((1 0) 1) ((0 1) 1) ((1 1) 0)))

(setq 3-or '(((0 0 0) 0) ((1 0 0) 1) ((0 1 0) 1) ((1 1 0) 1)
	     ((0 0 1) 1) ((1 0 1) 1) ((0 1 1) 1) ((1 1 1) 1))
      3-and '(((0 0 0) 0) ((1 0 0) 0) ((0 1 0) 0) ((1 1 0) 0)
	      ((0 0 1) 0) ((1 0 1) 0) ((0 1 0) 0) ((1 1 1) 1))
      3-xor '(((0 0 0) 0) ((1 0 0) 1) ((0 1 0) 1) ((1 1 0) 0)
	      ((0 0 1) 1) ((1 0 1) 0) ((0 1 1) 0) ((1 1 1) 1)))

(defun gradient (examples weights)
  (reduce #'(lambda (v w) (mapcar #'+ v w))
          (mapcar #'(lambda (pair) 
                      (delta (car pair) (cadr pair) weights))
                  examples)))

(defun trial (target)
  (aux-trial target (init) 0))

(let ((rate 2) (freq 10) (stat nil) (max 1000))
  (defun rate (r) (setq rate r))
  (defun freq (f) (setq freq f))
  (defun stat (s) (setq stat s))
  (defun aux-trial (target weights n)
    (cond ((successp target weights) 
	   (if stat n
	     (format t "~%Weights: ~6,3F ~6,3F ~6,3F ~6,3F ~6,3F ~6,3F ~6,3F"
		     (nth 0 weights) (nth 1 weights) (nth 2 weights)
		     (nth 3 weights) (nth 4 weights) (nth 5 weights)
		     (nth 6 weights))))
	  ((= n max) nil)
	  (t (if (and (= 0 (mod n freq)) (not stat))
		 (format t "~%Adjustments: ~D~
		            ~%Error: ~6,3F" n (mse target weights)))
	     (aux-trial target 
			(mapcar #'+ weights 
				(mapcar #'(lambda (w) (* rate w))
					(gradient target weights)))
			(+ n 1))))))

(defun average (target attempts)
  (stat t) 
  (let ((total 0) (successes 0) adjustments)
    (dotimes (i attempts)
	     (setq adjustments (trial target))
	     (if (numberp adjustments)
		 (progn (setq total (+ total adjustments))
			(setq successes (+ 1 successes)))))
    (format t "The average is ~4,1F for ~D out of ~D attempts."
	    (if (> successes 0) (/ (* total 1.0) successes) 0)
	    successes attempts)))
	     
(defun train (target weights)
  (if (successp target weights) (princ weights)
    (train target (mapcar #'+ weights (gradient target weights)))))

(defun successp (target weights)
  (every #'(lambda (x)
             (< (abs (- (cadr x) 
                        (net (car x) weights))) 0.4))
         target))

(defun mse (target weights)
  (apply #'+
         (mapcar #'(lambda (x)
                     (square (- (cadr x) 
                                (net (car x) weights))))
                 target)))

(defun square (x) (* x x))

