;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information
;**********************************************************************
;                                   ***********************************
;        Anthony R. Cassandra       ***********************************
;            CS141087               ***********************************
;          Homework #7              ***********************************
;       Due: November 18, 1992      ***********************************
;             CS141                 ***********************************
;         Prof. Tom Dean            ***********************************
;                                   ***********************************
;**********************************************************************
;
(in-package 'user)
;
;         Variables to turn debugging options on and off
(setq tree-debug nil)
(setq test-debug nil)
(setq stat-debug nil)
(setq perform-debug nil)
;
;    Set 'tree-alt' to 't' to use the buildtree function that uses the 
; chi-square statistic and 'nil'to use the given buildtree function.
(setq tree-alt t)
;
;**********************************************************************
; Function: MAKE-SOME-EXAMPLES
;
; Makes a list of examples of the length specified.  Calls 'cexample'
; which will generate them according to the currently set parameters
; of hypothesis space, distribution, etc.
;
(defun make-some-examples (number)
  (if (< number 1) 
      nil
    (cons (cexample) (make-some-examples (- number 1))))
)
;**********************************************************************
; Function: PRINT-TREE
;
;  Used in debugging to see the structure of the tree in a more
;  readable format with the examples stripped for all but the
;  terminal nodes.  Also, all the elements for the nodes are labeled.
;
(defun print-tree (tree)
  (if (null tree)
      nil
    (progn
      (princ "**************************************")
      (terpri)
      (princ "ATTRIBUTE: ")
      (princ (NODE-attribute tree))
      (terpri)
      (princ "FEATURE: ")
      (princ (NODE-feature tree))
      (terpri)
      (princ "CLASS: ")
      (princ (NODE-class tree))
      (terpri)
      (princ "ATTRIBUTE LIST: ")
      (princ (NODE-attributes tree))
      (terpri)
      (if (null (NODE-children tree))
	  (progn
	    (princ "EXAMPLES: ")
	    (terpri)
	    (princ (NODE-examples tree)))
	(progn
	  (princ "CHILDREN:")
	  (terpri)
	  (print-children (NODE-children tree))))
      (terpri)
      )))
;**********************************************************************
; Function: PRINT-CHILDREN
;
; Auxiliary function to 'print-tree' that runs through a
; list of child nodes.
;
(defun print-children (child-list)
  (if (or (null child-list) (null (car child-list)))
      nil
    (progn
      (print-tree (car child-list))
      (print-children (cdr child-list))
      )))
;**********************************************************************
; Function: DEPTH
;
;  Returns the depth of the tree.
;
(defun depth (tree)
  (- (aux-depth tree) 1))  ; compensate for one node tree
(defun aux-depth (tree)
  (if (null tree)
      0
    (+ 1 (list-max (mapcar #'aux-depth (NODE-children tree))))))
;**********************************************************************
;  Function: LIST-MAX
;
; Returns the maximum number in a list of numbers.
;
(defun list-max (list)
  (if (null list)
      0
    (max (car list) (list-max (cdr list)))))
;**********************************************************************
; Function: SIZE
;
; Returns the number of nodes in the tree.
;
(defun size (tree)
  (if (null tree)
      0
    (+ 1 (list-add (mapcar #'size (NODE-children tree))))))
;**********************************************************************
; Function: LIST-ADD
;
;  Adds all the numbers in a list of numbers
;
(defun list-add (list)
  (if (null list)
      0
    (+ (car list) (list-add (cdr list)))))
;**********************************************************************
; Function: TEST
;
; Program to run a test.  Inputs: a concept, a threshold and the number
; of training and evaluation examples to generate.  It returns
; the percentage of the performance examples that were correctly
; classified and the depth and size of the tree.  If the flag 'tree-alt'
; is set to 't' then the buildtree function within this module is used
; which is the one that uses the statistical measure.  If the flag is
; set to 'nil' then the regular buildtree with thresholding is used.
; When 'tree-alt' is 't', the threshold parameter is ignored.
;
(defun test (concept threshold num-examples)
  (set-concept concept)
  (set-distribution 0.5)
  (set-threshold threshold)
  (set-noise '((DEPT 0.8) (FLOOR 0.9)))
  (let ((training (make-some-examples num-examples))
        (evaluation (make-some-examples num-examples)) dtree)
    (setq dtree (make-NODE training))

    (if tree-alt              ; decide which function to call
	(buildtree-alt dtree)
      (buildtree dtree))

    (if test-debug
	(print-tree dtree))

    (list ( / (performance dtree evaluation)
	      num-examples)
          (depth dtree)
          (size dtree))
    ))
;**********************************************************************
; Function: PERFORMANCE
;
; Returns the number of correctly classified examples.
;
(defun performance (dtree examples)
  (if (null examples)
      0
    (+ (check-example dtree (car examples))
       (performance dtree (cdr examples)))))
;**********************************************************************
; Function: CHECK-EXAMPLE
;
; Takes a tree and an example and walks down the tree based on the
; attributes and features of the example until it gets to a leaf node.
; It compares the example classification against what the majority
; of the examples in that node are (the class of the node).  If the classes
; are the same then a '1' is returned, otherwise a '0' is returned.
;
(defun check-example (dtree example)
  (if (null (NODE-attribute dtree))
      ;THEN we are at a leaf node, check classification
      (progn
	(if perform-debug
	    (print-check dtree example))

	(if (equal (NODE-class dtree)
		   (EXAMPLE-class example))
	    1
	  0))

    ;ELSE go down to appropriate child
    (check-example (match-child (NODE-children dtree)
				(EXAMPLE-features example))
		   example)))
;**********************************************************************
;  Function: PRINT-CHECK
;
; Used in debugging the make sure the performance function is
; ariving at the appropriate nodes for each given example.
;
(defun print-check (dtree example)
  (princ "*** At leaf node class: ")
  (princ (NODE-class dtree))
  (terpri)
  (princ "Path to this node: ")
  (princ (NODE-attributes dtree))
  (terpri)
  (princ "Example: ")
  (princ example)
  (terpri))
;**********************************************************************
;  Function: MATCH-CHILD
;
; Returns the node from the list of nodes whose feature matches
; that of the value of the attribute in the feature list
;
(defun match-child (child-list feature-list)
  (cond
        ((null child-list) (princ "Error in 'match-child!"))

	((member (NODE-feature (car child-list))
		 feature-list
		 :test #'equal)   
	 (car child-list))

	(t
	 (match-child (cdr child-list) feature-list))))
;**********************************************************************
;
; Concepts used in experiments:
;
(setq concept1 '(and (dept cs) (floor five)))

(setq concept2 '(and (status faculty) (size large) (floor five)))

(setq concept3 '(and (dept cs) (status staff)))

(setq concept4 '(and (dept cs) (floor four) (status staff)))

;**********************************************************************
;
;                DATA
;
;  Number of examples: 50
;  Distribution:  0.5
;  Noise: DEPT  = 0.8
;         FLOOR = 0.9
;
; concept 1  = (and (dept cs) (floor five)))
; concept 2  = (and (status faculty) (size large) (floor five)))
; concept 3  = (and (dept cs) (status staff)))
; concept 4  = (and (dept cs) (floor four) (status staff)))
;
;
;
;                         % CORRECTLY           TREE              TREE
;CONCEPT #   THRESHOLD     IDENTIFIED           DEPTH             SIZE
;=======================================================================
;   1            .3            50                 0                 1  
;   1            .5            60                 0                 1  
;   1            .6            60                 0                 1  
;   1            .65           74                 4                22    
;   1            .7            68                 4                28   
;   1            .75           72                 4                17    
;   1            .8            84                 4                14   
;   1            .85           84                 4                24    
;   1            .9            64                 4                20   
;   1            .95           64                 4                27    
;
;   2            .3            96                 2                 7  
;   2            .5            98                 2                 7  
;   2            .6            98                 2                 7  
;   2            .7            98                 3                 9  
;   2            .8            98                 3                 9  
;   2            .9           100                 2                 7   
;
;   3            .3            92                 1                 4  
;   3            .4            90                 1                 4  
;   3            .5            94                 2                 6  
;   3            .6            94                 4                11   
;   3            .7            96                 3                 9  
;   3            .8            92                 4                13   
;   3            .9            88                 4                11   
;
;   4            .3            60                 0                 1     
;   4            .33           82                 2                 6      
;   4            .35           92                 1                 4      
;   4            .4            94                 2                 6     
;   4            .5            92                 3                 8     
;   4            .6            92                 3                 8     
;   4            .7            92                 4                11      
;   4            .8            90                 4                14      
;   4            .9            96                 4                13      
;========================================================================
;**********************************************************************
;
;  Analysis:
;
;   The threshold seems to have the most effect when there is a larger
; percentage of noise.  The first and fourth concepts are the most
; sensitive to the threshold value because they contain two 
; attributes that could possibly be jumbled by the noisy channel.
; The second and third concept only have one of their attributes
; that could be scrambled and are therefore rather insensitive
; to changes in the threshold.
;   It also seems that the more attributes there are in a concept
; the greater probability we have of correctly classifying them.
; The second and fourth concepts have three conjuncts and they
; give more reliable classifications than do the two conjunct
; counterparts.  (Note that for validity of comparison, I choose 
; two of each in length and that each of these in turn had one and
; two 'noisy' attributes respectively.)
;  Usually as the % of correctly classified examples goes up the
; size (and depth) of the tree also increase.  But, there seems 
; to be a saturation point at which the % of correctly classified
; examples stablizes (or declines) while the size of the tree
; continues to increase.  I attribute this to the function splitting
; nodes when it cannot really discrimminate what it is splitting on.
;**********************************************************************
;
;  Extra Credit:
;
;**********************************************************************
; Function: MATCH-FEATURE
;
;  Takes a list of examples and a feature (attribute and its value)
;  and returns a list of examples that is a subset of the original
;  examples, but only those that contain the feature.
;
(defun match-feature (example-list feature)
  (cond
       ((null example-list)
	nil)
       
	((member feature 
		 (EXAMPLE-features (car example-list))
		 :test #'equal)
	 (cons (car example-list)
	       (match-feature (cdr example-list)
			      feature)))

       (t
	(match-feature (cdr example-list)
		       feature))))
;**********************************************************************
; Function: MATCH-CLASS
;
;  Takes a list of examples and a class (i.e YES or NO)
;  and returns a list of examples that is a subset of the original
;  examples, but only those that match the class.
;
(defun match-class (example-list class)
  (cond
        ((null example-list)
	 nil)

	((equal class (EXAMPLE-class (car example-list)))
	 (cons (car example-list)
	       (match-class (cdr example-list)
			    class)))

	(t
	 (match-class (cdr example-list)
		      class))))
;**********************************************************************
; Function: CREATE-FEATURE-LIST
;
; Takes an attribute symbol and a list of values (hopefully the
; values that are pertinent to that attribute) and creates a 
; list of features.  Each feature being an attribute-value pair.
;
(defun create-feature-list (attribute value-list)
  (if (null value-list)
      nil
    (cons (cons attribute (cons (car value-list) ()))
	  (create-feature-list attribute
			       (cdr value-list)))))
;**********************************************************************
; Function: GET_STATISTIC
;
; Takes a list of examples and a dimension (attribute) and will
; return a statistic that has a chi-square distribution.  This statistic
; is a measure of how relevent that attribute is in classifying
; the examples.  The formulas and method are described in the homework
; handout and its supplement, so hopefully I need not repeat them here.
; The variables and their names were chosen to have as close a
; correspondence with the handout terminology to make life easy for
; me and possibly for you.  There is a check before the last
; division to avoid division by zero.  Although I cannot see why this
; should happen (because if either list is null we just stop the 
; recursion) this did happen on a rare occasion.  My best guess was
; that the feature-list somehow got to be a list containing the empty
; list.  I had a tough time reproducing this and wasn't able to
; track it down.
;
(defun get-statistic (example-list dimension)
  (if stat-debug
      (progn
	(princ "Dimension: ")
	(princ dimension)
	(terpri)))
  (aux-statistic example-list 
		 (create-feature-list (DIMENSION-attribute dimension)
				      (DIMENSION-values dimension))))

(defun aux-statistic (example-list feature-list)
  (let (p n Ci n-sub-i p-sub-i pi-hat ni-hat)
    (cond
          ((or (null feature-list)
	       (null example-list))
	       
	   0)

	   (t
	    (setq p (length (match-class example-list (car (classes)))))
	    (setq n (length (match-class example-list (cadr (classes)))))
	    (setq Ci (match-feature example-list (car feature-list)))
	    (setq p-sub-i (length (match-class Ci (car (classes)))))
	    (setq n-sub-i (length (match-class Ci (cadr (classes)))))
	    (setq pi-hat (* p (/ (+ p-sub-i n-sub-i) (+ p n))))
	    (setq ni-hat (* n (/ (+ p-sub-i n-sub-i) (+ p n))))
	    (if stat-debug
		(progn
		  (princ p)
		  (terpri)
		  (princ n)
		  (terpri)
		  (princ p-sub-i)
		  (terpri)
		  (princ n-sub-i)
		  (terpri)
		  (princ pi-hat)
		  (terpri)
		  (princ ni-hat)
		  (terpri)
		  (princ Ci)
		  (terpri)
		  ))
	    (if (or (= pi-hat 0)
		     (= ni-hat 0))
		0
	      (+ (+ (/ (square (- p-sub-i pi-hat)) pi-hat)
		    (/ (square (- n-sub-i ni-hat)) ni-hat))
		 (aux-statistic example-list (cdr feature-list))))))))
;**********************************************************************
; Lisp doesn't define this ??????????? 
(defun square (x)
  (* x x))
;**********************************************************************
; Function: CHI-SQUARE-LOOKUP
;
;  This function takes, as a parameter, the number of degrees of freedom
; and returns the chi square critical value such that 5 %
; of the area under the distribution graph lies to the right.  It
; only goes up to 8 degrees of freedom. If a number > 8 or a number
; < 1 is input then the function returns 0.
;
(let ((table '(3.843 5.992 7.815 9.488 11.070 12.592 14.067 15.507)))
  (defun chi-square-lookup (deg-freedom)
    (aux-chi-lookup table deg-freedom)))
(defun aux-chi-lookup (list num)
  (cond
      ((or (null list) 
	   (< num 1))
       0) 
     
      ((> num 1)
       (aux-chi-lookup (cdr list) (- num 1)))

      (t
       (car list))))
;**********************************************************************
; Function: BUILD-TREE-ALT
;
;    See the in line comment for changes.  I assume that the statistic
;  can only be used to accept or not accept the hypothesis of
;  'this attribute is irrelevent'.  Therefore
;  the original 'evaluate' function is still used to decide which
;  (if any) attribute the node should be split on.  Once the 'evaluate'
;  function has decided this, the statistic will actually decide
;  whether the splitting will or will not be beneficial.

(let ((verbose nil))
  (defun set-verbose-alt (value) (setq verbose value))
  (defun buildtree-alt (node)
    (if (>= 1 (length (findall (classify (NODE-examples node))
			       #'(lambda (s) (not (null s))))))
	(set-NODE-class node (majority (NODE-examples node)))
      (let ((dim (splitter node)) examples nodes)
	(if verbose (format t "Score: ~4,2F~%" (evaluate node dim)))

;  This first part of the 'OR' statement is the only part that
;  is changed from the original.  
;
;
	(if tree-debug
	    (print-stat node dim))
        (if (or (< (get-statistic (NODE-examples node) dim)
		   (chi-square-lookup (- (length (DIMENSION-values dim))

					 1)))
		(member (DIMENSION-attribute dim) (NODE-attributes node))
		(eq (DIMENSION-attribute dim) (NODE-attribute node)))
	    (set-NODE-class node (majority (NODE-examples node)))
          (progn (setq examples (NODE-examples node)
                       nodes (mapcar #'make-NODE
                                     (partition examples dim)))
                 (set-NODE-attribute node (DIMENSION-attribute dim))
                 (set-NODE-children node nodes)
                 (mapc #'(lambda (new value)
                           (set-NODE-attributes 
                            new (cons (DIMENSION-attribute dim)
                                      (NODE-attributes node)))
                           (set-NODE-feature 
                            new (list (DIMENSION-attribute dim) value)))
                       nodes
                       (DIMENSION-values dim))
                 (mapc #'buildtree-alt nodes)))))))
;**********************************************************************
; Function: PRINT-STAT
;
;  Used in debugging to give status and values about statistics
;
(defun print-stat (node dim)
  (princ "Considering spilt on node with feature: ")
  (princ (NODE-feature node))
  (terpri)
  (princ "Considering dimension: ")
  (princ dim)
  (terpri)
  (princ "Statistic: ")
  (princ (get-statistic (NODE-examples node) dim))
  (terpri)
  (princ "Chi-square value: ")
  (princ (chi-square-lookup (- (length (DIMENSION-values dim)) 1)))
  (terpri))
;**********************************************************************
;
;   Best thresholds from regular buildtree function for each concept
;
;                         % CORRECTLY           TREE              TREE
;CONCEPT #   THRESHOLD     IDENTIFIED           DEPTH             SIZE
;=======================================================================
;   1            .8            84                 4                14   
;   2            .9           100                 2                 7   
;   3            .7            96                 3                 9  
;   4            .4            94                 2                 6     
;=======================================================================
;
;
;                         % CORRECTLY           TREE              TREE
;CONCEPT #   THRESHOLD     IDENTIFIED           DEPTH             SIZE
;=======================================================================
;   1            -             84                 2                 5 
;   2            -             98                 2                 7 
;   3            -             98                 1                 4 
;   4            -             94                 2                 6    
;=======================================================================
;
;  Analysis:
;
;  Using the statistic is clearly superior to the threshold business.
;  This is especially evident when the level of noise is high.  The 
;  first and fourth concepts have the highest percentage of noisy
;  attributes and as such were very sensitive to the value of the
;  threshold.  Without having to adjust any threshold the statistic
;  gives just as good performance and usually makes a smaller tree.
;  It is obvious that the first concept with thresholding is splitting
;  far too many nodes on irrelevent attributes, whereas the statistic
;  is able to recognize this and behave more reasonably.  Even where
;  the noise level isn't that high the statistical method yields
;  excellent results.
;
;  Output to show correct execution of 'get-statistic'
;				 
;**********************************************************************
;
;  NOTE:  I originally had a vast amount of output here to show to
;         proper execution of all the functions.  I decided to omit
;         it because of its length.  I believe you will test the
;         program anyway so the inclusion of the output could just
;         be a waste.  Please let me know if you really do desire to
;         see this intermediate output so I can adjust future 
;         assignments.  Thanks.
;
;**********************************************************************


;; Copyright 1994, Brown University, Providence, RI
;; 
;; Permission to use and modify this software and its documentation for any
;; purpose other than its incorporation into a commercial product is hereby
;; granted without fee.  Permission to copy and distribute this software and its
;; documentation only for non-commercial use is also granted without fee,
;; provided, however, that the above copyright notice appear in all copies, that
;; both that copyright notice and this permission notice appear in supporting
;; documentation, that the name of Brown University not be used in advertising or
;; publicity pertaining to distribution of the software without specific, written
;; prior permission, and that the person doing the distribution notify Brown
;; University of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this software for
;; any purpose.  It is provided "as is" without express or implied warranty.
;; Brown University requests notification of any modifications to this software
;; or its documentation.
;; 
;; Send the following redistribution information:
;; 
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;; 
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;; 
;; 		or
;; 
;; 	brusd@cs.brown.edu
;; 
;; We will acknowledge all electronic notifications.
