;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information
(in-package 'user)

;;
;;                      w_0
;;		   x_0 ----- hout1
;;		       \   /   \
;;		    w_1 \ / w_4 \
;;		         X       o ---- out
;;		    w_2 / \ w_5 /
;;		       /   \   /
;;		   x_1 ----- hout2
;;		        w_3
;;

(load "pdp")

(defun sigmoid (x) (/ 1 (+ 1 (exp (- 1 x)))))

;; The choice of initial weights is important in learning in multilayer
;; networks. Generally it is a bad idea to have all the weights initially
;; the same. We use a vector of real-valued random weights in the range
;; from zero to two.
(defun init ()
  (mapcar #'random '(2.0 2.0 2.0 2.0 2.0 2.0)))

;; Compute the output of the network shown above for a specified
;; input and setting of the five adjustable weights.
(defun net (input weights)
  (sigmoid (+ (* (nth 4 weights)
                 (sigmoid (+ (* (nth 0 weights) (nth 0 input))
                             (* (nth 2 weights) (nth 1 input)))))
              (* (nth 5 weights)
                 (sigmoid (+ (* (nth 1 weights) (nth 0 input))
                             (* (nth 3 weights) (nth 1 input))))))))

;; Compute the contribution to the network for a single training example
;; for the network shown above.
(defun delta (input output weights)
  (let ((error (- output (net input weights)))
        (out (net input weights)) 
        (hout1 (sigmoid (+ (* (nth 0 weights) (nth 0 input))
                           (* (nth 2 weights) (nth 1 input)))))
        (hout2 (sigmoid (+ (* (nth 1 weights) (nth 0 input))
                           (* (nth 3 weights) (nth 1 input))))))
    (list (* (nth 0 input) hout1 (- 1 hout1)
             (nth 4 weights) out (- 1 out) error)
          (* (nth 0 input) hout2 (- 1 hout2)
             (nth 5 weights) out (- 1 out) error)
          (* (nth 1 input) hout1 (- 1 hout1)
             (nth 4 weights) out (- 1 out) error)
          (* (nth 1 input) hout2 (- 1 hout2)
             (nth 5 weights) out (- 1 out) error)
          (* hout1 out (- 1 out) error)
          (* hout2 out (- 1 out) error))))

;; The following assumes a learning rate of 1.0.
;; > (average or 10) 
;; The average is 296.7 for 3 out of 10 attempts.
;; > (average and 10)
;; The average is 301.7 for 3 out of 10 attempts.
;; > (average xor 10)
;; The average is 394.7 for 3 out of 10 attempts.

;; The following assumes a learning rate of 2.0.
;; > (average or 10) 
;; The average is 345.8 for 5 out of 10 attempts.
;; > (average and 10)
;; The average is 152.5 for 6 out of 10 attempts.
;; > (average xor 10)
;; The average is 199.0 for 6 out of 10 attempts.


;; Copyright 1994, Brown University, Providence, RI
;; 
;; Permission to use and modify this software and its documentation for any
;; purpose other than its incorporation into a commercial product is hereby
;; granted without fee.  Permission to copy and distribute this software and its
;; documentation only for non-commercial use is also granted without fee,
;; provided, however, that the above copyright notice appear in all copies, that
;; both that copyright notice and this permission notice appear in supporting
;; documentation, that the name of Brown University not be used in advertising or
;; publicity pertaining to distribution of the software without specific, written
;; prior permission, and that the person doing the distribution notify Brown
;; University of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this software for
;; any purpose.  It is provided "as is" without express or implied warranty.
;; Brown University requests notification of any modifications to this software
;; or its documentation.
;; 
;; Send the following redistribution information:
;; 
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;; 
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;; 
;; 		or
;; 
;; 	brusd@cs.brown.edu
;; 
;; We will acknowledge all electronic notifications.
