;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

(setq optimal '(1.20 -0.20 3.10 -0.90))

(let ((rate 0.1))
 (defun target-function (x)
   (+ 1.20 (* x -0.20) (* x x 3.10) (* x x x -0.90)))
 (defun set-rate (x) (setq rate x))
 (defun initw () '(0.0 0.0 0.0 0.0))
 ;; Compute the output of the network for a given input and weights.
 (defun net (input weights)
   (+ (* (nth 0 weights) 1)
      (* (nth 1 weights) input)
      (* (nth 2 weights) input input)
      (* (nth 3 weights) input input input)))
 ;; The gradient adjustment method 
 (defun adjust (examples weights)
  (let ((d (/ 1 (length examples))))
   (list 
    (+ (nth 0 weights)	
       (* rate 2 d (apply #'+ (mapcar #'(lambda (pair)
					  (* (- (cadr pair) 
						(net (car pair) weights))
					     (norm (car pair) 0)))
				      examples))))
    (+ (nth 1 weights)	
       (* rate 2 d (apply #'+ (mapcar #'(lambda (pair)
					  (* (- (cadr pair) 
						(net (car pair) weights))
					     (norm (car pair) 1)))
				      examples))))
    (+ (nth 2 weights)
       (* rate 2 d (apply #'+ (mapcar #'(lambda (pair)
					  (* (- (cadr pair) 
						(net (car pair) weights))
					     (norm (car pair) 2)))
				      examples))))
    (+ (nth 3 weights)
       (* rate 2 d (apply #'+ (mapcar #'(lambda (pair)
					  (* (- (cadr pair) 
						(net (car pair) weights))
					     (norm (car pair) 3)))
				      examples))))))))

(defun raise (x n)
  (if (= n 0) 1 (* x (raise x (- n 1)))))

(defun square (x) (* x x))

(defun mse (pairs weights)
  (apply #'+
	 (mapcar #'(lambda (x)
		     (square (- (cadr x) 
				(net (car x) weights))))
		 pairs)))

;; Generate n training examples in the range -1 to 1. 
(defun gen-examples (n)
  (do ((i 1 (+ i 1)) (examples (list)) (x nil))
      ((= i n) examples)
      (setq x (- (random 2.0) 1))
      (setq examples (cons (list x (target-function x))
			   examples))))


;; Here is a test version with running commentary and summary statistics.
(let ((max 40) (examples (gen-examples 40)) (online nil) (normv nil))
  (defun norm (x n)
    (if normv (/ (raise x n) 
		(let ((z (abs x)))
		  (+ 1 z (* z z) (* z z z))))
      (raise x n)))
  (defun online (v) (setq online v) (setq normv v))
  (defun sample () 
    (if online (list (nth (random (length examples)) examples))
      examples))
  (defun patience (adjustments) (setq max adjustments))
  (defun test ()
    (train (initw) 0))
  (defun train (weights adjustments)
    (if (= adjustments max) 
	(format t "~F + ~F x + ~F x^2 + ~F x^3"
		(nth 0 weights)(nth 1 weights)
		(nth 2 weights)(nth 3 weights))
      (progn (format t "~F, " (mse examples weights))
	     (train (adjust (sample) weights)
		    (+ 1 adjustments))))))


;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;;
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;;
;; 		or
;;
;; 	brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.

