;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

(set! optimal '(1.2 -.2 3.1 -.9))


(let ((rate .1))
  (set! target-function
    (lambda (x) (+ 1.2 (* x -.2) (* x x 3.1) (* x x x -.9))))
  (set! set-rate (lambda (x) (set! rate x)))
  (set! initw (lambda () '(0. 0. 0. 0.)))
  (set! net
    (lambda (input weights)
      (+ (* (nth 0 weights) 1)
         (* (nth 1 weights) input)
         (* (nth 2 weights) input input)
         (* (nth 3 weights) input input input))))
  (set! adjust
    (lambda (examples weights)
      (let ((d (/ 1 (length examples))))
        (list (+ (nth 0 weights)
                 (* rate
                    2
                    d
                    (apply +
                           (map (lambda (pair)
                                  (* (- (cadr pair) (net (car pair) weights))
                                     (norm (car pair) 0)))
                                examples))))
              (+ (nth 1 weights)
                 (* rate
                    2
                    d
                    (apply +
                           (map (lambda (pair)
                                  (* (- (cadr pair) (net (car pair) weights))
                                     (norm (car pair) 1)))
                                examples))))
              (+ (nth 2 weights)
                 (* rate
                    2
                    d
                    (apply +
                           (map (lambda (pair)
                                  (* (- (cadr pair) (net (car pair) weights))
                                     (norm (car pair) 2)))
                                examples))))
              (+ (nth 3 weights)
                 (* rate
                    2
                    d
                    (apply +
                           (map (lambda (pair)
                                  (* (- (cadr pair) (net (car pair) weights))
                                     (norm (car pair) 3)))
                                examples)))))))))


(define (raise x n) (if (= n 0) 1 (* x (raise x (- n 1)))))


(define (square x) (* x x))


(define (mse pairs weights)
  (apply +
         (map (lambda (x) (square (- (cadr x) (net (car x) weights)))) pairs)))


;; Generate n training examples in the range -1 to 1. 
(define (gen-examples n)
  (do ((i 1 (+ i 1)) (examples (list)) (x #f))
      ((= i n) examples)
    (set! x (- (random 2.) 1))
    (set! examples (cons (list x (target-function x)) examples))))



;; Here is a test version with running commentary and summary statistics.
(let ((max 40) (examples (gen-examples 40)) (online #f) (normv #f))
  (set! norm
    (lambda (x n)
      (if normv
        (/ (raise x n) (let ((z (abs x))) (+ 1 z (* z z) (* z z z))))
        (raise x n))))
  (set! online (lambda (v) (set! online v) (set! normv v)))
  (set! sample
    (lambda ()
      (if online (list (nth (random (length examples)) examples)) examples)))
  (set! patience (lambda (adjustments) (set! max adjustments)))
  (set! test (lambda () (train (initw) 0)))
  (set! train
    (lambda (weights adjustments)
      (if (= adjustments max)
        (begin
          (display (nth 0 weights))
          (display " + ")
          (display (nth 1 weights))
          (display " x + ")
          (display (nth 2 weights))
          (display " x^2 + ")
          (display (nth 3 weights))
          (display " x^3"))
        (begin
          (begin (display (mse examples weights)) (display ", "))
          (train (adjust (sample) weights) (+ 1 adjustments)))))))



;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;;      Name:
;;      Organization:
;;      Address (postal and/or electronic):
;;
;; To:
;;      Software Librarian
;;      Computer Science Department, Box 1910
;;      Brown University
;;      Providence, RI 02912
;;
;;              or
;;
;;      brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.

