;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

(begin
  (set! or-test '(((0 0) 0) ((1 0) 1) ((0 1) 1) ((1 1) 1)))
  (set! and-test '(((0 0) 0) ((1 0) 0) ((0 1) 0) ((1 1) 1)))
  (set! xor-test '(((0 0) 0) ((1 0) 1) ((0 1) 1) ((1 1) 0))))


;; Sum the contributions to the gradient for each example at a point
;; in parameter space specified by the weights.
(define (gradient examples weights)
  (reduce (lambda (v w) (map + v w))
          (map (lambda (pair) (delta (car pair) (cadr pair) weights))
               examples)))


(define (trial target) (aux-trial target (init) 0))

(let ((rate 2) (freq 10) (stat #f) (max 1000))
  (set! set-rate (lambda (r) (set! rate r)))
  (set! set-freq (lambda (f) (set! freq f)))
  (set! set-stat (lambda (s) (set! stat s)))
  (set! aux-trial
    (lambda (target weights n)
      (cond ((successp target weights)
             (if stat
               n
               (begin
                 (newline)
                 (display "Weights: ")
                 (display (nth 0 weights))
                 (display " ")
                 (display (nth 1 weights))
                 (display " ")
                 (display (nth 2 weights))
                 (display " ")
                 (display (nth 3 weights))
                 (display " ")
                 (display (nth 4 weights))
                 (display " ")
                 (display (nth 5 weights))
                 (display " ")
                 (display (nth 6 weights)))))
            ((= n max) #f)
            (else
             (if (and (= 0 (modulo n freq)) (not stat))
               (begin
                 (newline)
                 (display "Adjustments: ")
                 (display n)
                 (display (mse target weights))))
             (aux-trial
               target
               (map +
                    weights
                    (map (lambda (w) (* rate w)) 
                         (gradient target weights)))
               (+ n 1)))))))


;; Attempt to train the network several times, and average the number
;; of adjustments it took to train each successful time.
(define (average target attempts)
  (set-stat #t)
  (let ((total 0) (successes 0) (adjustments #f))
    (do ((i 1 (+ 1 i)))
        ((> i attempts))
      (set! adjustments (trial target))
      (if (number? adjustments)
        (begin
          (set! total (+ total adjustments))
          (set! successes (+ 1 successes)))))
    (begin
      (display "The average is ")
      (display (if (> successes 0) (/ (* total 1.) successes) 0))
      (display " for ")
      (display successes)
      (display " out of ")
      (display attempts)
      (display " attempts."))))

             
(define (train target weights)
  (if (successp target weights)
    (display weights)
    (train target (map + weights (gradient target weights)))))


;; For a given input, the sigmoid function corresponding to the output of
;; the network provides a value between zero and one. A given set of
;; weights is considered to have successfully implemented a target
;; function if the difference between the actual and the desired outputs
;; is less than 0.4.  All that is absolutely necessary is that the
;; difference is strictly less than 0.5, so that the rounded output is
;; equal to the desired output.  The function ABS returns the absolute 
;; value of its single argument.
(define (successp target weights)
  (every2 (lambda (x) (< (abs (- (cadr x) (net (car x) weights))) .4)) 
          target))


(define (mse target weights)
  (apply +
         (map (lambda (x) (square (- (cadr x) (net (car x) weights))))
              target)))


(define (square x) (* x x))


;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;;      Name:
;;      Organization:
;;      Address (postal and/or electronic):
;;
;; To:
;;      Software Librarian
;;      Computer Science Department, Box 1910
;;      Brown University
;;      Providence, RI 02912
;;
;;              or
;;
;;      brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.

