;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

;; 
;;             x_0 --\--------\
;;                    \        \
;;                    w_0      w_3
;;                      \  hout  \       out
;;             -1 --w_1- o --w_4-- o ---
;;                      /         / \
;;                    w_2       w_5  w_6
;;                    /         /      \
;;             x_1 --/---------/       -1
;;

(load "pdp.scm")


(define (sigmoid x) (/ 1 (+ 1 (exp (- x)))))


;; The choice of initial weights is important in learning in multilayer
;; networks. Generally it is a bad idea to have all the weights initially
;; the same. We use a vector of real-valued random weights in the range
;; from zero to two.
(define (init) (map random '(2. 2. 2. 2. 2. 2. 2.)))


;; Compute the output of the network shown in Figure 5.19 for a specified
;; input and setting of the six adjustable weights.
(define (net input weights)
  (sigmoid (+ (* (nth 3 weights) (nth 0 input))
              (* (nth 4 weights)
                 (sigmoid (+ (* (nth 0 weights) (nth 0 input))
                             (* (nth 1 weights) -1)
                             (* (nth 2 weights) (nth 1 input)))))
              (* (nth 5 weights) (nth 1 input))
              (* (nth 6 weights) -1))))


;; Compute the contribution to the network for a single training example
;; for the network shown in Figure 5.19.
(define (delta input output weights)
  (let ((error (- output (net input weights)))
        (out (net input weights))
        (hout (sigmoid (+ (* (nth 0 weights) (nth 0 input))
                          (* (nth 1 weights) -1)
                          (* (nth 2 weights) (nth 1 input))))))
    (list (* (nth 0 input) hout (- 1 hout) (nth 4 weights) out (- 1 out) error)
          (* -1 hout (- 1 hout) (nth 4 weights) out (- 1 out) error)
          (* (nth 1 input) hout (- 1 hout) (nth 4 weights) out (- 1 out) error)
          (* (nth 0 input) out (- 1 out) error)
          (* hout out (- 1 out) error)
          (* (nth 1 input) out (- 1 out) error)
          (* -1 out (- 1 out) error))))


;; The following assumes a learning rate of 1.0.
;; > (average or-test 10) 
;; The average is  5.1 for 10 out of 10 attempts.
;; > (average and-test 10)
;; The average is 12.2 for 10 out of 10 attempts.
;; > (average xor-test 10)
;; The average is 264.0 for 9 out of 10 attempts.

;; The following assumes a learning rate of 2.0.
;; > (average or-test 10) 
;; The average is 5.0 for 10 out of 10 attempts.
;; > (average and-test 10)
;; The average is 5.6 for 10 out of 10 attempts.
;; > (average xor-test 10)
;; The average is 145.3 for 10 out of 10 attempts.
                        

;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;;      Name:
;;      Organization:
;;      Address (postal and/or electronic):
;;
;; To:
;;      Software Librarian
;;      Computer Science Department, Box 1910
;;      Brown University
;;      Providence, RI 02912
;;
;;              or
;;
;;      brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.

