;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

;; Gaussian function used as a radial basis function.
(define (gaussian x c w) (exp (- (* (/ 1 (* 2 w)) (square (- x c))))))


;; Generate n training examples in the range -1 to 1. 
(define (examples n)
  (do ((i 1 (+ i 1)) (examples (list)) (x #f))
      ((= i n) examples)
    (set! x (- (random 2.) 1))
    (set! examples (cons (list x (target x)) examples))))


;; Target function for the learning example:
(define (target x) (+ 1.2 (* x -.2) (* x x 3.1) (* x x x -.9)))


(define (raise x n) (if (= n 0) 1 (* x (raise x (- n 1)))))


(define (square x) (* x x))


(define (mse pairs weights)
  (apply +
         (map (lambda (x) (square (- (cadr x) (net (car x) weights)))) 
              pairs)))


;; Here is a test version with running commentary and summary statistics.
(let ((max 100)
      (examples (examples 40))
      (online #f)
      (verbose #f)
      (ctr1 -1.)
      (ctr2 -.5)
      (ctr3 0.)
      (ctr4 .5)
      (width 1.)
      (rate .1))
  (set! wc
    (lambda (a b c d w)
      (begin
        (set! ctr1 a)
        (set! ctr2 b)
        (set! ctr3 c)
        (set! ctr4 d)
        (set! width w))))
  (set! set-verbose (lambda (v) (set! verbose v)))
  (set! set-online (lambda (v) (set! online v)))
  (set! sample
    (lambda ()
      (if online (list (nth (random (length examples)) examples)) examples)))
  (set! patience (lambda (adjustments) (set! max adjustments)))
  (set! set-rate (lambda (x) (if (number? x) (set! rate x) rate)))
  (set! initw (lambda () '(0. 0. 0. 0.)))
  (set! net
    (lambda (input weights)
      (+ (* (nth 0 weights) (gaussian input ctr1 width))
         (* (nth 1 weights) (gaussian input ctr2 width))
         (* (nth 2 weights) (gaussian input ctr3 width))
         (* (nth 3 weights) (gaussian input ctr4 width)))))
  (set! adjust
    (lambda (examples weights)
      (let ((d (/ 1 (length examples))))
        (list (+ (nth 0 weights)
                 (* rate
                    2
                    d
                    (apply +
                           (map (lambda (pair)
                                  (* (- (cadr pair) (net (car pair) weights))
                                     (gaussian (car pair) ctr1 width)))
                                examples))))
              (+ (nth 1 weights)
                 (* rate
                    2
                    d
                    (apply +
                           (map (lambda (pair)
                                  (* (- (cadr pair) (net (car pair) weights))
                                     (gaussian (car pair) ctr2 width)))
                                examples))))
              (+ (nth 2 weights)
                 (* rate
                    2
                    d
                    (apply +
                           (map (lambda (pair)
                                  (* (- (cadr pair) (net (car pair) weights))
                                     (gaussian (car pair) ctr3 width)))
                                examples))))
              (+ (nth 3 weights)
                 (* rate
                    2
                    d
                    (apply +
                           (map (lambda (pair)
                                  (* (- (cadr pair) (net (car pair) weights))
                                     (gaussian (car pair) ctr4 width)))
                                examples))))))))
  (set! test (lambda () (train (initw) 0)))
  (set! train
    (lambda (weights adjustments)
      (if (= adjustments max)
        (begin
          (newline)
          (display "Plot[")
          (display (nth 0 weights))
          (display " Gaussian[x,")
          (display ctr1)
          (display ",")
          (display width)
          (display "] + (* RATE ")
          (display rate)
          (display " *)")
          (display (nth 1 weights))
          (display " Gaussian[x,")
          (display ctr2)
          (display ",")
          (display width)
          (display "] + (* MSE ")
          (display (mse examples weights))
          (display " *)")
          (display (nth 2 weights))
          (display " Gaussian[x,")
          (display ctr3)
          (display ",")
          (display width)
          (display "] + (* STEPS ")
          (display max)
          (display " *)")
          (display (nth 3 weights))
          (display " Gaussian[x,")
          (display ctr4)
          (display ",")
          (display width)
          (display "],{x,-1,1}]")
          (newline))
        (begin
          (if verbose (begin (display (mse examples weights)) (display ", ")))
          (train (adjust (sample) weights) (+ 1 adjustments)))))))


;; Experiments for learning chapter:

(define (data)
  (patience 500)
  (newline)
  (wc -.75 -.25 .25 .75 1.)
  (test)
  (newline)
  (wc -.75 -.25 .25 .75 .5)
  (test)
  (newline)
  (wc -.75 -.375 0. .5 1.)
  (test)
  (newline)
  (wc -.75 -.375 0. .5 .5)
  (test)
  (newline)
  (wc -1. -.5 0. 1. 1.)
  (test)
  (newline)
  (wc -1. -.5 0. 1. .5)
  (test))



;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;;      Name:
;;      Organization:
;;      Address (postal and/or electronic):
;;
;; To:
;;      Software Librarian
;;      Computer Science Department, Box 1910
;;      Brown University
;;      Providence, RI 02912
;;
;;              or
;;
;;      brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.

