;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information
;; File: circuits.lisp
;; Date: 10/21/92
;; Description: Temporal reasoning exercises modeling digitial circuits.

(in-package 'user)

;; 1. Complete Exercise 1 at the end of Chapter 6.
;; 2. Complete Exercise 2 at the end of Chapter 6. 
;; 3. Complete Exercise 3.b at the end of Chapter 6. You don't have
;;    to do 3.a to complete this assignment.  Instead of writing a theory
;;    in the interval logic, write your theory directly in the CRULE
;;    notation. As an alternative notation, you can use the PCAUSE/ECAUSE
;;    notation described in class.  Your theory should work for any set of
;;    initial conditions and events, but provide a script showing your
;;    theory working via the function DISPLAY on the initial conditions:
;;    W1 and W2 high, W3 low, and W4 and W5 determined by the appropriate
;;    boolean functions, both high in this case.  If you use the CRULE 
;;    notation, project using (project initial-conditions initial-events
;;    causal-theory).  If you use the PCAUSE/ECAUSE notation, project
;;    using (tm initial-conditions initial-events causal-theory).
;;    Hints: Use propositions like (HIGH W) where W is a wire and events
;;           like (GOES-LOW W) and (CHANGE G) where G is a gate.
;;           You don't have to understand the code below in order to 
;;           complete the assignment; you do have to understand what
;;           it is that PROJECT computes as explained in class. 


;; Answers:
;;
;; 1.
;;
;; We wish to prove 
;; 
;; occurs(result(s1, pulse(trigger(latch1))), transition(output(and1), 1 0))
;; 
;; Working backward, we use universal instantiation, the fact that
;; output(latch1) = input1(and1), and the axiom
;; 
;; forall s, b1, b2, b3,
;;  (holds(s, status(input1(and1), b1)) and
;;   holds(s, status(input2(and1), b2)) and
;;   holds(s, status(output(and1), b3)) and
;;   not(b3 = boolean_and(b1, b2))) implies
;;  occurs(s, transition(output(and1), b3, boolean_not(b3)))
;; 
;; to obtain
;; 
;; (holds(result(s1, pulse(trigger(latch1))), status(output(latch1), 0)) and
;;  holds(result(s1, pulse(trigger(latch1))), status(input2(and1), 1)) and
;;  holds(result(s1, pulse(trigger(latch1))), status(output(and1), 1)) and
;;  not(1 = boolean_and(0,1))) implies
;; occurs(result(s1, pulse(trigger(latch1))), transition(output(and1), 1, 0))
;; 
;; We need to satisfy each antecedent of the above implication to obtain
;; the desired conclusion.  The last antecedent is immediate
;; 
;; not(1 = boolean_and(0,1)))
;; 
;; given that boolean_and(0,1) = 0
;; 
;; The first antecedent is relatively straightforward. 
;; Working forward, we use universal instantiation and the axiom
;; 
;; forall s, b,
;;  holds(s, status(input(latch1), b) implies
;;  holds(result(s, pulse(trigger(latch1))), 
;;        status(output(latch1), b))
;; 
;; to obtain 
;; 
;; holds(s, status(input(latch1), 1) implies
;; holds(result(s, pulse(trigger(latch1))), 
;;       status(output(latch1), 1))
;; 
;; which together with 
;; 
;; holds(s1, status(input1(latch1), 0))
;; 
;; allows us to conclude that 
;; 
;; holds(result(s, pulse(trigger(latch1))), status(output(latch1), 0))
;;      
;; The other two antecedents require the frame axioms, 
;; 
;; forall s, b, 
;;  holds(s, status(output(and1), b)) implies
;;  holds(result(s, pulse(trigger(latch1))), status(output(and1), b))
;; 
;; forall s, b, 
;;  holds(s, status(input2(and1), b)) implies
;;  holds(result(s, pulse(trigger(latch1))), status(input2(and1), b))
;; 
;; from which using 
;; 
;; holds(s1, status(input2(and1), 1))
;; 
;; and
;; 
;; holds(s1, status(output(and1), 1))
;; 
;; we obtain 
;; 
;; holds(result(s1, pulse(trigger(latch1))), status(input2(and1), 1)) 
;; 
;; and
;; 
;; holds(result(s1, pulse(trigger(latch1))), status(output(and1), 1)) 
;;  
;; Now we conjoin the appropriate conjuncts and apply modus ponens to
;; obtain the desired result. 
;; 
;; Note that 
;; 
;; occurs(s1, pulse(trigger(latch1)))
;; 
;; is never used. 
;; 
;; The answer is just the two frame axioms shown above. 


;; 2. Two approaches are listed; the first uses qualifications
;;    and distinct interruption and restoration events. 
;;    The second approach uses toggle(power) to represent
;;    both interruption and restoration and toggle(switch)
;;    to represent flipping the switch to its opposite position.
;;    The second approach also uses `ramification' axioms 
;;    that were not talked about in class, but that several 
;;    students have invented on their own; ramification axioms
;;    constrain fluents within a situation. 
;;
;; Axioms governing the result of the toggling the light switch.
;; 
;; forall s, (holds(s,off) and not(ab(s))) implies 
;;           holds(result(s,toggle),on) 
;; 
;; forall s, (holds(s,on) and not(ab(s))) implies 
;;           holds(result(s,toggle),off))
;; 
;; 
;; If there is no power, then toggling a light switch has no effect.
;; 
;; forall s, holds(s,nopower) implies ab(s)
;; 
;; 
;; We have to keep track of state information about the light switch.
;;           
;; forall s, holds(s,up) implies holds(result(s,toggle),down)           
;; 
;; forall s, holds(s,down) implies holds(result(s,toggle),up))           
;; 
;; 
;; Interruptions interrupt the power and turn the light off if on.
;; 
;; forall s, holds(result(s,interruption),nopower) 
;; 
;; forall s, holds(s,on) implies holds(result(s,interruption),off) 
;; 
;; 
;; Restorations restore power and turn the light on if the switch is up.
;; 
;; forall s, holds(result(s,restoration),power) 
;; 
;; forall s, holds(s,up) implies holds(result(s,restoration),on) 
;; 
;; 
;; We also need assorted frame axioms not listed here. 
;; 
;; Given 
;; 
;; holds(s0,on)
;; 
;; and the above axioms, you should be able to prove
;; 
;; holds(result(result(result(result(s0,
;;                                   toggle),
;;                            interruption),
;;                     toggle),
;;              restoration), on)
;; 
;; 
;; Here's a second, somwhat simpler approach to solving the problem
;; that makes use of ramification axioms. 
;; 
;; holds(s,on(switch)) and holds(s,on(power)) implies holds(s,on(light))
;; 
;; holds(s,off(switch)) implies holds(s,off(light))
;; 
;; holds(s,off(power)) implies holds(s,off(light))
;; 
;; holds(s,on(switch)) implies holds(result(s,toggle(switch)),off(switch))
;; 
;; holds(s,on(power)) implies holds(result(s,toggle(power)),off(power))


;; 3.b. 

;; Ontologically speaking there are wires and gates.  Wires (GO-HIGH w)
;; or (GO-LOW w) and gates (CHANGE g); these are the basic event types.
;; Wires are either (HIGH w) or (NOT (HIGH w)); thess are the basic fact
;; types.  Events concerning wires ECAUSE events concerning gates.
;; Events concerning gates ECAUSE events concerning wires.  Events
;; concerning wires PCAUSE changes in facts concerning wires.  Events
;; concerning gates don't PCAUSE anything. It is sufficient to provide
;; rules that cover the particular scenario described in the question.

(defun test-circuit ()
  (tm 
     ;; Initial conditions:
     '((HIGH W1)
       (HIGH W2)
       (NOT (HIGH W3))
       (HIGH W4)
       (HIGH W5))
     ;; Initial events:
     '(((GOES-LOW W1) 10)
       ((GOES-HIGH W1) 40))
     ;; Causal rules:
     '(
       ;; Propagating events through the AND gate:
       (ECAUSE (AND)
               (GOES-HIGH W1)
               (CHANGE AND) 7)
       (ECAUSE (AND)
               (GOES-LOW W1)
               (CHANGE AND) 7)
       (ECAUSE (AND (HIGH W1) (HIGH W2))
               (CHANGE AND)
               (GOES-HIGH W4) 1)
       (ECAUSE (NOT (HIGH W1))
               (CHANGE AND)
               (GOES-LOW W4) 1)
       ;; Propagating events through the OR gate:
       (ECAUSE (AND)
               (GOES-HIGH W4)
               (CHANGE OR) 4)
       (ECAUSE (AND)
               (GOES-LOW W4)
               (CHANGE OR) 4)
       (ECAUSE (AND (NOT (HIGH W3)) (NOT (HIGH W4)))
               (CHANGE OR)
               (GOES-LOW W5) 1)
       (ECAUSE (HIGH W4)
               (CHANGE OR)
               (GOES-HIGH W5) 1)
       ;; The rules governing W1:
       (PCAUSE (HIGH W1)
               (GOES-LOW W1)
               (NOT (HIGH W1)))
       (PCAUSE (NOT (HIGH W1))
               (GOES-HIGH W1)
               (HIGH W1))
       ;; The rules governing W4:
       (PCAUSE (HIGH W4)
               (GOES-LOW W4)
               (NOT (HIGH W4)))
       (PCAUSE (NOT (HIGH W4))
               (GOES-HIGH W4)
               (HIGH W4))
       ;; The rules governing W5:
       (PCAUSE (HIGH W5)
               (GOES-LOW W5)
               (NOT (HIGH W5)))
       (PCAUSE (NOT (HIGH W5))
               (GOES-HIGH W5)
               (HIGH W5)) ))
  (display))

;; Copyright 1994, Brown University, Providence, RI
;; 
;; Permission to use and modify this software and its documentation for any
;; purpose other than its incorporation into a commercial product is hereby
;; granted without fee.  Permission to copy and distribute this software and its
;; documentation only for non-commercial use is also granted without fee,
;; provided, however, that the above copyright notice appear in all copies, that
;; both that copyright notice and this permission notice appear in supporting
;; documentation, that the name of Brown University not be used in advertising or
;; publicity pertaining to distribution of the software without specific, written
;; prior permission, and that the person doing the distribution notify Brown
;; University of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this software for
;; any purpose.  It is provided "as is" without express or implied warranty.
;; Brown University requests notification of any modifications to this software
;; or its documentation.
;; 
;; Send the following redistribution information:
;; 
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;; 
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;; 
;; 		or
;; 
;; 	brusd@cs.brown.edu
;; 
;; We will acknowledge all electronic notifications.
