;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information
;; File: dungeon.lisp
;; Date: 11/2/93
;; Description: Solution to the dungeons-and-dragons exercise.

(in-package 'user)

(load "tm")

;;-------------
;; Causal rules 
;;-------------

(setq causal-rules '(

;; a. movement
  ;; from dungeon to chamber
  (ecause (and (alive knight)
	       (not (frozen knight))
	       (location knight dungeon))
	  (move knight dungeon chamber)
	  (enter knight chamber) 2)
  
  ;; from chamber to dungeon 
  (ecause (and (alive knight)
	       (not (frozen knight))
	       (location knight chamber))
	  (move knight chamber dungeon)
	  (enter knight dungeon) 2)

  ;; from chamber to lair
  (ecause (and (alive knight)
	       (not (frozen knight))
	       (location knight chamber))
	  (move knight chamber lair)
	  (enter knight lair) 2)

  ;; from lair to chamber
  (ecause (and (alive knight)
	       (not (frozen knight))
	       (location knight lair))
	  (move knight lair chamber)
	  (enter knight chamber) 2)

  ;; entering dungeon
  (pcause (and)
	  (enter knight dungeon)
	  (location knight dungeon) 0)

  
  ;; entering chamber
  (pcause (and)
	  (enter knight chamber)
	  (location knight chamber) 0)

  ;; entering lair
  (pcause (and)
	  (enter knight lair)
          (location knight lair) 0)

;; b. knight disrupts dragon; dragon toasts knight
  (ecause (not (enchanted dragon))
	  (enter knight lair)
	  (toast dragon knight) 1)
  
  (pcause (location knight lair)
	  (toast dragon knight)
	  (not (alive knight)) 0)


;; c. knight thaws wizard; wizard freezes knight
  (pcause (and (location knight chamber)
	       (frozen wizard)
	       (possess knight torch))
	  (thaw knight wizard)
	  (not (frozen wizard)) 0)
  
  (ecause (and)
	  (thaw knight wizard)
	  (freeze wizard knight) 6)

  (pcause (and (not (possess wizard jewel))
	       (location knight chamber))
	  (freeze wizard knight)
	  (frozen knight) 0)


;; d. picking up
  (pcause (location knight dungeon)
	  (pickup knight torch)
	  (possess knight torch) 0)

  (pcause (location knight dungeon)
	  (pickup knight jewel)
	  (possess knight jewel) 0)

  (pcause (location knight lair)
	  (pickup knight gold)
	  (possess knight gold) 0)


;; e. giving the jewel
  (pcause (and (location knight chamber)
	       (possess knight jewel)
	       (not (frozen wizard)))
	  (give knight wizard jewel)
	  (and (not (possess knight jewel)) 0
	       (possess wizard jewel) 0
	       (enchanted dragon) 2))))

;;-------------------
;; Initial conditions
;;-------------------

(setq initial-conditions '(
  (location knight dungeon)
  (location dragon lair)
  (location wizard chamber)
  (location jewel dungeon)
  (location torch dungeon)
  (location gold lair)
  (not (possess knight torch))
  (not (possess knight gold))
  (not (possess knight jewel))
  (not (possess wizard jewel))
  (not (enchanted dragon))
  (not (frozen knight))
  (frozen wizard)
  (alive knight)))

;;---------------
;; Initial events
;;---------------

(setq initial-events1 '(
  ((pickup knight torch) 10)
  ((pickup knight jewel) 20)
  ((move knight dungeon chamber) 30)
  ((thaw knight wizard) 40)
  ((give knight wizard jewel) 45)
  ((move knight chamber lair) 50)
  ((pickup knight gold) 60)))

(setq initial-events2 '(
  ((pickup knight torch) 10)
  ((move knight dungeon chamber) 20)
  ((thaw knight wizard) 30)
  ((move knight chamber lair) 40)))

(setq initial-events3 '(
  ((move knight dungeon chamber) 10)
  ((move knight chamber lair) 20)))

;;-------------------
;; Projection routine
;;-------------------

(defun tls (initial-conditions 
	    initial-events 
	    causal-rules)
  (reset-events) (reset-facts)
  (tm initial-conditions 
      initial-events 
      causal-rules)
  (display))

;;--------------
;; Test routines
;;--------------

(defun test1 () (tls initial-conditions initial-events1 causal-rules))
(defun test2 () (tls initial-conditions initial-events2 causal-rules))
(defun test3 () (tls initial-conditions initial-events3 causal-rules))


;; Copyright 1994, Brown University, Providence, RI
;; 
;; Permission to use and modify this software and its documentation for any
;; purpose other than its incorporation into a commercial product is hereby
;; granted without fee.  Permission to copy and distribute this software and its
;; documentation only for non-commercial use is also granted without fee,
;; provided, however, that the above copyright notice appear in all copies, that
;; both that copyright notice and this permission notice appear in supporting
;; documentation, that the name of Brown University not be used in advertising or
;; publicity pertaining to distribution of the software without specific, written
;; prior permission, and that the person doing the distribution notify Brown
;; University of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this software for
;; any purpose.  It is provided "as is" without express or implied warranty.
;; Brown University requests notification of any modifications to this software
;; or its documentation.
;; 
;; Send the following redistribution information:
;; 
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;; 
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;; 
;; 		or
;; 
;; 	brusd@cs.brown.edu
;; 
;; We will acknowledge all electronic notifications.
