;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information
(in-package 'user)

;; An easier initial case
(setq easy '((on a 1) (on b 2) (on c 3) 
	     (clear a) (clear b) (clear c)
	     (smaller a b) (smaller b c) (smaller a c)
	     (smaller a 1) (smaller a 2) (smaller a 3)
	     (smaller b 1) (smaller b 2) (smaller b 3)
	     (smaller c 1) (smaller c 2) (smaller c 3)))

;; Initially all the disks are stacked on the first peg
(setq hard '((on a b) (on b c) (on c 1) 
	     (clear a) (clear 2) (clear 3)
	     (smaller a b) (smaller b c) (smaller a c)
	     (smaller a 1) (smaller a 2) (smaller a 3)
	     (smaller b 1) (smaller b 2) (smaller b 3)
	     (smaller c 1) (smaller c 2) (smaller c 3)))

;; The goal is to have the disks stacked on the third peg
(setq goal '((on a b) (on b c) (on c 3)))

;; One operator schema suffices for this version of the problem.
(setq operators '((((on (? x) (? y)) (clear (? x)) (clear (? z))
                    (smaller (? x) (? z)))
                   ((clear (? y)) (on (? x) (? z)))
                   ((on (? x) (? y)) (clear (? z))))))

(defun test (state) 
  (gps-bfs-pp (list state) goal operators))

(let ((n 0))
  (defun gps-bfs-pp (states goal operators)
    (setq n 0)
    (bfs states
	 #'(lambda (s) (setq n (+ n 1))
	     (if (= 0 (mod n 100)) (display s))
	     (subsetp goal s :test #'equal))
	 #'(lambda (s) (apply-operators operators s)))))

(defun findit (pat consts bdgs)
  (cond ((null consts) ())
	((match (car consts) pat bdgs))
	(t (findit pat (cdr consts) bdgs))))

(defun display (state)
  (mapcar #'(lambda (n)
	      (let (bdgs (bottom (list 'on '(? x) n)))
		(setq bdgs (findit bottom state '((match t))))
		(if bdgs
		    (progn 
		      (princ (varsubst bottom bdgs))
		      (setq bdgs (findit `(on (? y) (? x)) state bdgs))
		      (if bdgs
			  (progn 
			    (princ (varsubst '(on (? y) (? x)) bdgs))
			    (setq bdgs (findit `(on (? z) (? y)) state bdgs))
			    (if bdgs
				(princ (varsubst '(on (? z) (? y)) bdgs))
			      (terpri)))
			(terpri)))
		  (terpri))))
	  `(1 2 3))
  nil)


;; Copyright 1994, Brown University, Providence, RI
;; 
;; Permission to use and modify this software and its documentation for any
;; purpose other than its incorporation into a commercial product is hereby
;; granted without fee.  Permission to copy and distribute this software and its
;; documentation only for non-commercial use is also granted without fee,
;; provided, however, that the above copyright notice appear in all copies, that
;; both that copyright notice and this permission notice appear in supporting
;; documentation, that the name of Brown University not be used in advertising or
;; publicity pertaining to distribution of the software without specific, written
;; prior permission, and that the person doing the distribution notify Brown
;; University of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this software for
;; any purpose.  It is provided "as is" without express or implied warranty.
;; Brown University requests notification of any modifications to this software
;; or its documentation.
;; 
;; Send the following redistribution information:
;; 
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;; 
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;; 
;; 		or
;; 
;; 	brusd@cs.brown.edu
;; 
;; We will acknowledge all electronic notifications.
