;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

;;**********
;; Examples:
;;**********

;;******************************************
;; Consistency check for vertex descriptors:
;;******************************************

;; Check if a list of vertex descriptors is consistent.
(define (check-diagram vertex-descriptors)
  (let ((errors 0))
    (for-each
      (lambda (v-d)
        (let ((vertex1 (first v-d)) (v-type (second v-d)))
          (if (not (= (length (v-d-neighbors v-d))
                      (cond ((member2 v-type
                                      '(arrow fork tee)
                                      eq?
                                      (lambda (x) x))
                             3)
                            ((eq? v-type 'ell) 2)
                            (else -1))))
            (begin
              (begin
                (newline)
                (display "Illegal type/neighbor combo: ")
                (display v-d))
              (set! errors (+ 1 errors))))
          (for-each
            (lambda (vertex2)
              (if (not (= 1
                          (let ((count 0))
                            (for-each
                              (lambda (v-d2)
                                (if (and (eq? (first v-d2) vertex2)
                                         (member2 vertex1
                                                  (v-d-neighbors v-d2)
                                                  eq?
                                                  (lambda (x) x)))
                                  (set! count (+ 1 count))))
                              vertex-descriptors)
                            count)))
                (begin
                  (begin
                    (newline)
                    (display "Inconsistent vertex: ")
                    (display vertex1)
                    (display "-")
                    (display vertex2))
                  (set! errors (+ 1 errors)))))
            (v-d-neighbors v-d))))
      vertex-descriptors)
    (if (> errors 0)
      (begin
        (newline)
        (display "Inconsistent diagram: ")
        (display errors)
        (display " total error(s)."))))
  vertex-descriptors)


(set! cube
  (construct-diagram
    (check-diagram
      '((a fork b c d)
        (b arrow g e a)
        (c arrow e f a)
        (d arrow f g a)
        (e ell c b)
        (f ell d c)
        (g ell b d)))))


(set! cube-on-plate
  (construct-diagram
    (check-diagram
      '((a fork b c d)
        (b arrow g e a)
        (c arrow e f a)
        (d arrow f g a)
        (e ell c b)
        (f fork d c i)
        (g fork b d h)
        (h arrow l g j)
        (i arrow f m j)
        (j fork h i k)
        (k arrow m l j)
        (l ell h k)
        (m ell k i)))))


(set! tower
  (construct-diagram
    (check-diagram
      '((a fork b c d)
        (n ell q o)
        (b arrow g e a)
        (o arrow y j n)
        (c arrow e f a)
        (p ell r i)
        (d arrow f g a)
        (q arrow n s w)
        (e ell c b)
        (r arrow s p x)
        (f fork d c i)
        (s ell r q)
        (g fork b d h)
        (t arrow w x z)
        (h arrow l g j)
        (u arrow x y z)
        (i arrow f m p)
        (v arrow y w z)
        (j fork h o k)
        (w fork t v q)
        (k arrow m l j)
        (x fork r u t)
        (l ell h k)
        (y fork v u o)
        (m ell k i)
        (z fork t u v)))))


(set! arch1
  (construct-diagram
    (check-diagram
      '((a arrow e b c)
        (p ell o q)
        (b ell d a)
        (q tee p i r)
        (c fork a d g)
        (r tee j s q)
        (d fork c b m)
        (s ell r t)
        (e ell a f)
        (t arrow v s k)
        (f tee e g n)
        (u ell t l)
        (g arrow h f c)
        (v ell 2 4)
        (h tee g i o)
        (w arrow x 1 y)
        (i tee h j q)
        (x ell w z)
        (j tee i k r)
        (y fork w 2 z)
        (k tee j l t)
        (z arrow 3 x y)
        (l tee k m v)
        (1 tee n o w)
        (m ell l d)
        (2 arrow v 3 y)
        (n ell f 1)
        (3 ell z 2)
        (o arrow p 1 h)
        (4 tee u l v)))))


(set! arch2
  (construct-diagram
    (check-diagram
      '((a arrow e b c)
        (p ell o q)
        (b ell d a)
        (q tee p i r)
        (c fork a d g)
        (r tee j s q)
        (d fork c b m)
        (s ell r t)
        (e ell a f)
        (t arrow u s k)
        (f tee e g n)
        (u ell t 4)
        (g arrow h f c)
        (v ell 2 4)
        (h tee g i o)
        (w arrow x 1 y)
        (i tee h j q)
        (x ell w z)
        (j tee i k r)
        (y fork w 2 z)
        (k tee j l t)
        (z arrow 3 x y)
        (l tee k m 4)
        (1 tee n o w)
        (m ell l d)
        (2 arrow v 3 y)
        (n ell f 1)
        (3 ell z 2)
        (o arrow p 1 h)
        (4 tee u l v)))))


(set! arch3
  (construct-diagram
    (check-diagram
      '((a arrow e b c)
        (p ell o q)
        (b ell d a)
        (q tee p i r)
        (c fork a d g)
        (r tee j s q)
        (d arrow b m c)
        (s ell r t)
        (e ell a f)
        (t arrow u s k)
        (f tee e g n)
        (u ell t 4)
        (g arrow h f c)
        (v ell 2 4)
        (h tee g i o)
        (w arrow x 1 y)
        (i tee h j q)
        (x ell w z)
        (j tee i k r)
        (y fork w 2 z)
        (k tee j l t)
        (z arrow 3 x y)
        (l tee k m 4)
        (1 tee n o w)
        (m ell l d)
        (2 arrow v 3 y)
        (n ell f 1)
        (3 ell z 2)
        (o arrow p 1 h)
        (4 tee u l v)))))


(display "Diagrams are loaded.")



;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;;
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;;
;; 		or
;;
;; 	brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.

