(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 3.0,
MathReader 3.0, or any compatible application. The data for the notebook 
starts with the line of stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[      9964,        223]*)
(*NotebookOutlinePosition[     10793,        252]*)
(*  CellTagsIndexPosition[     10749,        248]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Supplementary Notes for the Agents and Environments Lecture", "Title"],

Cell["Conundrum: You can't predict your own behavior. Why not? ", "Text"],

Cell[CellGroupData[{

Cell["Knowledge, Adaptation, and Free Lunches", "Section"],

Cell["\<\
Here are a few observations about natural intelligence that should \
provide some guidance in building artificial intelligences and determining \
what is possible and what is not.\
\>", "Text"],

Cell["\<\
Many (some would say all) of the amazing features of animal \
intelligence and human intelligence in particular were \"built\" using an \
existing set of parts - components that were originally \"designed\" for \
other purposes.\
\>", "Text"],

Cell["\<\
Just as the first computers were build from vacuum tubes designed \
for purposes other than computing boolean functions and modern computers are \
built using functional units such as half-adders and multiplexers, nature had \
to use what was available to produce the bat's wonderful echolocation \
capability or the human facility for speech.\
\>", "Text"],

Cell["\<\
Spoken language is constrained by many properties of the human \
vocal and auditory systems: systems that were not originally designed to \
produce or process speech.  The result is that all speech used by humans is \
highly constrained by the capabilities of the throat, larynx, ear, cochlea, \
etc.  It shouldn't come as a surprise that human language is also tied \
closely to other primitive capabilities that concern how we make sense of the \
world which in turn are largely determined by how we experience the world \
through our senses and physical extensions.\
\>", "Text"],

Cell[TextData[{
  "Speech recognition - which we take for granted - is hard for computers in \
part because we can't reproduce the biological means of speech production \
(encoding) and recognition (decoding).  Often, the \"quirks\" of language can \
be explained in terms of the hardware and its limitations, ",
  StyleBox["e.g.,",
    FontSlant->"Italic"],
  " redundancies in spoken language are there to offset particular \
limitations of our sensory and data processing capabilities."
}], "Text"],

Cell["\<\
If we were forced to communicate by Morse code, humans would be \
very inefficient communicators.  Machines on the other hand could be designed \
that would be very efficient communicating in Morse code - though if we were \
to design a language from the bottom up knowing what we know about coding \
theory and working with existing commmunications technology, we would \
probably develop a very different coding method.\
\>", "Text"],

Cell[TextData[{
  "People often express amazement at the ability of humans to learn language. \
 And surely it is an amazing capability.  However, the ability to learn \
language is largely built in - it is as though we were born with a special \
purpose processor that with a limited amount of data can \"recognize\" a \
language (even multiple languages) and then use that language to communicate \
effectively.  It seems to us that humans can learn any language (after all, \
some languages sound so \"foreign\" to our ear), but really we are only able \
to easily learn human languages, ",
  StyleBox["i.e.,",
    FontSlant->"Italic"],
  " languages that are designed for humans, with human ears, human vocal \
tracts, and human ways of experiencing the world."
}], "Text"],

Cell["\<\
It will become clearer when we investigate learning theory, but \
there is no free lunch - learning from nothing, tabular rasa learning is \
impossible.  Nature in the form of evolution took many generations to \
converge on a language facility that works well for human physiology. No \
doubt human ears and vocal tracts have co-evolved to some extent with the \
\"software\" for learning and using language. It wouldn't have worked to \
simply specify the language in advance even if nature could manage such a \
feat; having a more general facility so that language can adapt to \
circumstances has proved to be enormously useful.\
\>", "Text"],

Cell["\<\
We won't talk about the details of language learning in this \
course.  But we will talk about other sorts of learning and about systems \
that adapt to circumstances.  To apply the strategy that nature uses in \
learning language to, say, learning to walk or learning navigate in a city, \
we need some way to encode general knowledge about walking and navigating.  \
As engineers, our general strategy will be to encode as much knowledge as \
possible about what the artificial agent is likely to encounter in its \
environment.  For those things we can't anticipate we encode the knowledge in \
a parameterized form.  Then as the agent explores it's environment it only \
has determine the values of a few parameters in order to adapt effectively to \
what nature (the environment) hands it.\
\>", "Text"],

Cell[TextData[{
  "Note that if we are to follow the strategy outlined in the previous \
paragraph we have to be able to do the following: First, we have to be able \
to encode knowledge about the world.  Second, we have to be able to encode it \
in a parameterized form that captures the variation in the possible \
environments the agent will be dealt by chance.  Finally, we have to make it \
possible for the agent to make the appropriate observations (",
  StyleBox["i.e.,",
    FontSlant->"Italic"],
  " gather additional information) so that it can fix the parameters thereby \
taking advantage of its knowledge and adapting to its environment."
}], "Text"],

Cell[TextData[{
  "Imagine designing a car that adapts to the driver (",
  StyleBox["e.g.,",
    FontSlant->"Italic"],
  " someone accelerates and brakes hard versus someone who never goes over 20 \
MPH) and to the environment (",
  StyleBox["e.g.,",
    FontSlant->"Italic"],
  " higher elevations versus lower, high temperatures versus low)."
}], "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Pattern, Structure and Computational Leverage", "Section"],

Cell[TextData[{
  "The world appears to be full of structure and pattern.  There is some \
evidence that structure arises spontaneously in various kinds of \
combinatorial systems.  Computer science and Dynamical Systems theory (to \
name just a couple of relevant disciplines) are concerned with measuring and \
characterizing structure and related notions of complexity, randomness, \
information content, entropy, ",
  StyleBox["etc.",
    FontSlant->"Italic"],
  "  This is wonderful stuff and a lot of fun but we will take the existence \
of structure for granted; the only trick as far as we are concerned is \
finding it.  [Demonstrate Wolfram's Cellular Automata Notebook.]"
}], "Text"],

Cell[TextData[{
  "Structure and pattern make it easier to cope with the blooming, buzzing \
complex world that surrounds us.  Scientists and engineers alike talk about \
the curse of dimensionality.  If we have m sensor inputs (these could be bits \
in an encoding of an auditory signal or pixels in a binary image, then we \
have 2^m possibilities to consider.  It would be practically impossible to \
keep track of all these possibilities.  Luckily the space of these 2^m \
possibilities has structure.  To take a simple example, an animal in the wild \
may have some threats to worry about at dawn and a disjoint set of threat at \
dusk, moreover those threats may announce themselves in ways make it easy to \
allocate attention, ",
  StyleBox["e.g.,",
    FontSlant->"Italic"],
  " for a rabbit, anything that casts a fast moving shadow on the ground is a \
good reason to flee and quickly.  The rabbit doesn't have to distinguish \
between hawks and horned owls and running from an occasional canada goose \
won't be a disaster."
}], "Text"],

Cell["\<\
Knowing what to attend to when and how to limit your options to a \
relevant few is the trick.  If you go on to study operations research or \
combinatorial optimization you'll find a similar perspective is applied to \
devising algorithms for scheduling and sequencing problems.  We're interested \
in finding structure and then exploiting it to simplify the computational \
problems faced by intelligence agents that have to make decisions in a timely \
manner.\
\>", "Text"]
}, Open  ]]
},
FrontEndVersion->"X 3.0",
ScreenRectangle->{{0, 1280}, {0, 1024}},
WindowSize->{997, 727},
WindowMargins->{{85, Automatic}, {89, Automatic}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"Magnification"->1},
Magnification->2,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1709, 49, 76, 0, 194, "Title"],
Cell[1788, 51, 73, 0, 52, "Text"],

Cell[CellGroupData[{
Cell[1886, 55, 58, 0, 111, "Section"],
Cell[1947, 57, 203, 4, 84, "Text"],
Cell[2153, 63, 252, 5, 116, "Text"],
Cell[2408, 70, 367, 6, 148, "Text"],
Cell[2778, 78, 592, 9, 212, "Text"],
Cell[3373, 89, 501, 9, 55, "Text"],
Cell[3877, 100, 445, 7, 55, "Text"],
Cell[4325, 109, 777, 13, 79, "Text"],
Cell[5105, 124, 657, 10, 79, "Text"],
Cell[5765, 136, 818, 12, 91, "Text"],
Cell[6586, 150, 664, 11, 67, "Text"],
Cell[7253, 163, 356, 9, 31, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[7646, 177, 64, 0, 72, "Section"],
Cell[7713, 179, 694, 12, 67, "Text"],
Cell[8410, 193, 1048, 17, 103, "Text"],
Cell[9461, 212, 487, 8, 55, "Text"]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

