(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 3.0,
MathReader 3.0, or any compatible application. The data for the notebook 
starts with the line of stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    135937,       4234]*)
(*NotebookOutlinePosition[    136768,       4263]*)
(*  CellTagsIndexPosition[    136724,       4259]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Supplementary Notes for the Introductory Lecture", "Title"],

Cell["\<\
The quotations below were taken from the chapter headings in \"The \
Universe and the Teacup: The mathematics of truth and beauty\" by K. C. Cole, \
printed by Harcourt Brace & Co., 1998.\
\>", "Text"],

Cell[CellGroupData[{

Cell["Why study artificial intelligence?", "Section"],

Cell[CellGroupData[{

Cell["Philosophy and the history of ideas about thinking", "Subsection"],

Cell["\<\
In part, it's a class about ideas, and extraordinary ideas at that. \
 Specifically, ideas concerning whether we can construct artifacts (as \
opposed to reproduce ourselves by more or less conventional biological means) \
that behave (in terms of inferences drawn and decisions made) in a manner \
that we would expect of a human.  The underlying assumption is that humans \
constitute instances of a more general class of \"things that are \
intelligent.\"  And in addition that there exist other instances of this \
class that are (or can be) constructed from the same sorts of materials \
(steel, silicon, plastic) that we use to build other engineered artifacts \
(cars, stereos).\
\>", "Text"],

Cell["\<\
Why do we need a historical perspective?  The problem is that AI \
has at its core problems that are not well defined in a mathematical sense \
(they are somtimes said to be \"ill posed\").  [To get some insight into \
ill-posed problems consider the following: \"Find me the optimal course \
schedule for getting a degree at Brown.\"  This is not well posed because \
there are different answers based on your objective function.  If you just \
want to get through as quickly as possible and you don't care what degree you \
get or how my classes you take per semester then you'll get one answer.  If, \
on the other hand, you want an Sc.B. in computer science and you want a job \
starting at $50K per year, then you'll get another answer.]  A good example \
of an ill-posed problem in AI is \"What does it mean to be intelligent?\"  or \
\"If something can simulate intelligent behavior, it is then intelligent?\"  \
or \"Is it possible for a machine to be conscious?\"\
\>", "Text"],

Cell["\<\
If we were to agree on our terms, then these questions would be \
open to scientific inquiry.  But without consensus on the precise meaning of \
our terms, we end up talking past one another.  In the end, I don't think \
it's important that we pin down these particular terms.  The history of \
science is studded with the pitfalls of getting bogged down in words: does \
light behave like a wave or a particle (answer: one or the other or both \
depending on what measurable behavior you choose to associate with waves and \
particles).  Given the confusion and controversy over some of the most basic \
questions concerning intelligence, we'll often find it useful to turn to \
other disciplines such as philosophy, psychology, biology, and physics which \
have things to add to the discussion.\
\>", "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Engineering, control theory, and cybernetics on steroids ", "Subsection"],

Cell["\<\
To some students of the field, artificial intelligence is just \
another engineering or applied mathematics discipline.  Some parts of the \
class concerning robots (or knowbots) embedded in real environments look \
essentially like the classical automatic control agenda (this is the field of \
cybernetics repackaged for political reasons) on steroids (or perhaps happy \
dust depending on your estimate of the prospects for success).\
\>", "Text"],

Cell["\<\
Parts of this class will seem like a conventional \
tools-and-techniques oriented computer science or math class in that almost \
every lecture will include one or more well-defined problems (we may not \
agree that the problems have any relevance to the ill-posed questions raised \
above) and algorithms that claim to solve these problems.\
\>", "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Sciences of the artificial and thinking as a natural kind ", 
  "Subsection"],

Cell["\<\
\"Understanding is a lot like sex.  It's got a practical purpose, \
but that's not why people do it normally.\"  - Frank Oppenheimer \
(physicist)\
\>", "Text"],

Cell["\<\
This class is also a bit like a course in the physical sciences in \
that we have a particular physical phenomena that we wish to understand.  \
Some researchers focus on the only known instance, humans, and their work is \
closer to cognitive science.  Others focus on what might be or might have \
been.  The latter are a bit like cosmologists who wonder what other sorts of \
universes might have been (e.g., what would have happened had certain \
cosmological constants been different at the time of the big bang).\
\>", 
  "Text"],

Cell["\<\
As much as cosmology as is about big questions and big ideas, so is \
CS141.  One of the most interesting things about the universe is the fact \
that it contains a being capable of wondering about its origins.  A related \
question that arises in machine learning concerns the notion of a bias; in a \
rather deep mathematical sense, humans and all learning machines require a \
bias in order to successfully generalize from things that they have \
experienced to things that they haven't.  A question arises as to how much \
different the world would have to be before our built-in bias would fail \
catastrophically.  Imagine beings of very different sizes than humans.  For \
example, imagine being very small; the dynamics governing fleas is determined \
less by gravity than by electrostatic charges - for a flea, standing near a \
puddle of water is like a human standing near a precipice.  Lean over a \
little too far or get a little too close to the water and we have a disaster. \
 Could we cope with being much smaller or much larger tha\.04n we are; \
perhaps, but there probably is a point past which we just wouldn't be able \
comprehend our environment or adapt quickly enough to survive.\
\>", "Text"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Observations on growth functions, data interpretation and inference\
\
\>", "Section"],

Cell[CellGroupData[{

Cell["\<\
Size matters - focus on exponential growth and doubling processes\
\
\>", "Subsection"],

Cell["\<\
\"The greatest shortcoming of the human race is our inability to \
understand the exponential function.\"  - Albert Bartlett (physicist)\
\>", 
  "Text"],

Cell["\<\
A growth rate of only 7% results in a doubling of the population in \
only ten years.  Most people don't see 7% population growth as problematic.\
\
\>", "Text"],

Cell["\<\
Suppose that two bacteria settle in a coke bottle (somehow filled \
ample reserves of nutrient) at 11:00am suppose the population doubles once a \
minute and the coke bottle is full at noon.  At what point would the most \
far-sighted member of the colony suspect that overcrowding was likely to be a \
problem.  Certainly not when the bottle was half full, but in fact this was \
at 11:59, at a time when it was likely too late to do anything about it.\
\>",
   "Text"],

Cell["\<\
Big numbers - on a line representing 0 to 1 trillion mark the point \
approximately at which 1 billion would occur.  Actually our conception of \
large numbers (> 100,000) seems to correspond to log scale rather than linear \
scale, i.e., a million (10^6) is pretty close to a billion (10^9) which is \
pretty close to trillion (10^12) - in each case the pairs are only 3 units \
apart on a log scale plot.  How do you feel about getting $1 versus $100 for \
a birthday check?  Why does the congress debate a few million (or billion) in \
the budget for days or weeks, but quickly agree on changes to rates that \
affect (exponential) growth (e.g., raising the ceiling on the national debt) \
which will have consequences in the range of trillions of dollars?\
\>", 
  "Text"],

Cell[CellGroupData[{

Cell["Demonstration of exponential growth behavior", "Subsubsection"],

Cell["\<\
Here is an example of the classic exponential curve; as x gets \
larger 2^x gets very large, very quickly.\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\
Plot[2^x,{x,0,20},
     {PlotLabel -> \"2^x for x from 0 to 20\", PlotRange -> All}] ;\
\>", 
  "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0476191 0.0147145 5.61337e-07 [
[.2619 .00221 -3 -9 ]
[.2619 .00221 3 0 ]
[.5 .00221 -6 -9 ]
[.5 .00221 6 0 ]
[.7381 .00221 -6 -9 ]
[.7381 .00221 6 0 ]
[.97619 .00221 -6 -9 ]
[.97619 .00221 6 0 ]
[.01131 .12698 -36 -4.5 ]
[.01131 .12698 0 4.5 ]
[.01131 .23925 -36 -4.5 ]
[.01131 .23925 0 4.5 ]
[.01131 .35152 -36 -4.5 ]
[.01131 .35152 0 4.5 ]
[.01131 .46378 -36 -4.5 ]
[.01131 .46378 0 4.5 ]
[.01131 .57605 -28.5 -6.25 ]
[.01131 .57605 0 6.25 ]
[.5 .63053 -68 0 ]
[.5 .63053 68 14 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.2619 .01471 m
.2619 .02096 L
s
[(5)] .2619 .00221 0 1 Mshowa
.5 .01471 m
.5 .02096 L
s
[(10)] .5 .00221 0 1 Mshowa
.7381 .01471 m
.7381 .02096 L
s
[(15)] .7381 .00221 0 1 Mshowa
.97619 .01471 m
.97619 .02096 L
s
[(20)] .97619 .00221 0 1 Mshowa
.125 Mabswid
.07143 .01471 m
.07143 .01846 L
s
.11905 .01471 m
.11905 .01846 L
s
.16667 .01471 m
.16667 .01846 L
s
.21429 .01471 m
.21429 .01846 L
s
.30952 .01471 m
.30952 .01846 L
s
.35714 .01471 m
.35714 .01846 L
s
.40476 .01471 m
.40476 .01846 L
s
.45238 .01471 m
.45238 .01846 L
s
.54762 .01471 m
.54762 .01846 L
s
.59524 .01471 m
.59524 .01846 L
s
.64286 .01471 m
.64286 .01846 L
s
.69048 .01471 m
.69048 .01846 L
s
.78571 .01471 m
.78571 .01846 L
s
.83333 .01471 m
.83333 .01846 L
s
.88095 .01471 m
.88095 .01846 L
s
.92857 .01471 m
.92857 .01846 L
s
.25 Mabswid
0 .01471 m
1 .01471 L
s
.02381 .12698 m
.03006 .12698 L
s
[(200000)] .01131 .12698 1 0 Mshowa
.02381 .23925 m
.03006 .23925 L
s
[(400000)] .01131 .23925 1 0 Mshowa
.02381 .35152 m
.03006 .35152 L
s
[(600000)] .01131 .35152 1 0 Mshowa
.02381 .46378 m
.03006 .46378 L
s
[(800000)] .01131 .46378 1 0 Mshowa
.02381 .57605 m
.03006 .57605 L
s
gsave
.01131 .57605 -28.5 -4.25 Mabsadd m
1 1 Mabs scale
/Courier findfont 10 scalefont setfont
(1) show
/Math1Mono findfont 10 scalefont setfont
(\\264) show
/Courier findfont 10 scalefont setfont
(10) show
0 5.25 rmoveto
/Courier findfont 7.5 scalefont setfont
(6) show
grestore
.125 Mabswid
.02381 .04278 m
.02756 .04278 L
s
.02381 .07085 m
.02756 .07085 L
s
.02381 .09892 m
.02756 .09892 L
s
.02381 .15505 m
.02756 .15505 L
s
.02381 .18312 m
.02756 .18312 L
s
.02381 .21118 m
.02756 .21118 L
s
.02381 .26732 m
.02756 .26732 L
s
.02381 .29538 m
.02756 .29538 L
s
.02381 .32345 m
.02756 .32345 L
s
.02381 .37958 m
.02756 .37958 L
s
.02381 .40765 m
.02756 .40765 L
s
.02381 .43572 m
.02756 .43572 L
s
.02381 .49185 m
.02756 .49185 L
s
.02381 .51992 m
.02756 .51992 L
s
.02381 .54798 m
.02756 .54798 L
s
.02381 .60412 m
.02756 .60412 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
gsave
.5 .63053 -129 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 22 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(2) show
69.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(^) show
75.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(x) show
87.000000 15.000000 moveto
(for) show
111.000000 15.000000 moveto
(x) show
123.000000 15.000000 moveto
(from) show
153.000000 15.000000 moveto
(0) show
165.000000 15.000000 moveto
(to) show
183.000000 15.000000 moveto
(20) show
195.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.5 Mabswid
.02381 .01472 m
.04262 .01472 L
.06244 .01472 L
.08255 .01472 L
.10458 .01472 L
.12415 .01472 L
.14509 .01472 L
.16639 .01472 L
.18653 .01472 L
.20651 .01472 L
.22495 .01473 L
.24367 .01473 L
.2643 .01473 L
.28371 .01474 L
.30459 .01475 L
.32572 .01476 L
.34581 .01478 L
.36562 .0148 L
.384 .01482 L
.40453 .01486 L
.42313 .0149 L
.44238 .01496 L
.4632 .01505 L
.48314 .01516 L
.5042 .01533 L
.52384 .01553 L
.54217 .01578 L
.56253 .01614 L
.58108 .01659 L
.60017 .01718 L
.62093 .01806 L
.6407 .01917 L
.66171 .02077 L
.68118 .02275 L
.69946 .0252 L
.71964 .02878 L
.73815 .03312 L
.75707 .03896 L
.77777 .04749 L
.79738 .05831 L
.81833 .07386 L
.83968 .09541 L
.85983 .12292 L
.87985 .15951 L
.8983 .20413 L
.91706 .26362 L
.9377 .35085 L
.95626 .45508 L
.97619 .60332 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/@P10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdk34g_ojn_l004J2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_P00oiT0IP3o0;C^]:JV
YZf][Jo^kP20P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;0000?`00O`00_`0o000o
?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0
_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo?cooOcoo
_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no?gnoOgno
_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo
_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0_ol0oolo0?lo?oloOolo_olooomo
0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_mWEc:0Pl=WIfJj^[WMgM`0:0>o<knl1
k`1^06EUI@00I@00[C8bI@1U002]039U<VD00:d006Db<RPX:000:00X02P002PX:00X02P00<`cVL`c
c<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co
<lcoI/coVLcoc<coool0<ol0I_l0VOl0c?l0oolc0?lc<olcI_lcVOlcc?lcoomV0?mVI_mVVOmVoonI
0?nI<onII_nIVOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<ool08@000?l08@0006<00`4e00<110020C40
0`4300<1<@040@<00P480000IP000`40000c00030@00008000@10001<@000`40000500030@0002l0
00<100000P00104000470000IP000`40000c00030@00008000@10001<@000`40000500030@000300
00D100000@0200030@0000D0001S00<1=P000`40000200040@000C4000<100000P030C@000@10001
0P000`4000050000H`000`40000f00030@00008000@10001<@000`40000200030@00034000@10001
0P00104000470000H`040C@00P450081<@020@@0104a008110020@P0000[00030@000?80000[0003
0@000?80000[00030@000?80000[00030@000?80000U0?H11@0002/000<100002@000`4000080003
0@0000T000<100002@000`40000800030@0000T000<100002@000`40000800030@0000T000<10000
2@000`40000800030@0000T000<100001P0:0@@000<100002@000`40000900030@0000P000<10000
2@000`40000900030@0000P000<100002@0002/000<10000W0090Dd0000[00030@000:D01@580000
:`000`40002Z00@1A00002/000<10000[P030D40000[00030@000;400P4o0000:`020K@00`4l0000
:`000`40002f0081>P0002/000<10000^0020CP0000[00030@000;X00P4f0000:`000`40002l0003
0@0003<0000[00030@000;d00P4c0000:`020L0000<10000<00002/000<10000`0000`40000_0000
:`000`40003100030@0002h0000[00030@000<8000<10000;@0002/000<10000``000`40000/0000
:`000`40003400030@0002/0000[00030@000<D000<10000:P0002/00P7600030@0002X0000[0003
0@000<H000<10000:@0002/000<10000a`000`40000X0000:`000`40003800030@0002L0000[0003
0@000<P000<100009`0000D01043008110020@@00P44008110020@D000<10000b@000`40000V0000
1P000`40000200040@000@8000@100010P001040004200040@000@8000@1000110000`40003:0003
0@0002D0000700050@0000400P001040004200040@000@8000@100010P001040004200030@000080
0P7;00030@0002D0000800040@000@8000@100010P001040004200040@000@8000@100010P000`40
000200030@000</000<10000900000D000@100010P001040004200040@000@8000@100010P001040
004200040@000@@000<10000b`000`40000T00001P020@@00P44008110020@@00P4400811@000`40
003<00030@0002<0000[00030@000<d000<100008P0002/000<10000c@000`40000R0000:`000`40
003>00030@000240000[0081c`000`40000Q0000:`000`40003?00030@000200000[00030@000<l0
00<10000800002/000<10000d0000`40000O0000:`000`40003@00030@0001l0000[00030@000=00
00<100007`0002/000<10000d@000`40000N0000:`020M8000<100007P0002/000<10000d@000`40
000N0000:`000`40003B00030@0001d0000[00030@000=8000<100007@0002/000<10000dP000`40
000M0000:`000`40003C00030@0001`0000[00030@000=<000<10000700002/00P7E00030@0001/0
000[00030@000=@000<100006`0002/000<10000e0000`40000K0000:`000`40003E00030@0001X0
000[00030@000=D000<100006P0000H00`43008110020@@00P44008110020@D000<10000e@000`40
000J00001`001@400001008000@100010P001040004200040@000@8000@100010P000`4000020003
0@000=H000<100006@0000D0104200040@000@8000@100010P001040004200040@000@8000@10001
10020ML000<100006@0000D000<100400`001040004200040@000@8000@100010P00104000420004
0@000@@000<10000eP000`40000I00001@000`400@0300040@000@8000@100010P00104000420004
0@000@8000@1000110000`40003G00030@0001P00006008110020@@00P44008110020@@00P450003
0@000=L000<10000600002/000<10000e`000`40000H0000:`000`40003H00030@0001L0000[0003
0@000=P000<100005`0002/00P7I00030@0001L0000[00030@000=P000<100005`0002/000<10000
f@000`40000F0000:`000`40003I00030@0001H0000[00030@000=T000<100005P0002/000<10000
f@000`40000F0000:`020M/000<100005@0002/000<10000fP000`40000E0000:`000`40003J0003
0@0001D0000[00030@000=X000<100005@0002/000<10000f`000`40000D0000:`000`40003K0003
0@0001@0000[00030@000=/000<10000500002/00P7L00030@0001@0000[00030@000=/000<10000
500002/000<10000g0000`40000C0000:`000`40003L00030@0001<0000[00030@000=`000<10000
4`0000H00P44008110020@@00P44008110020@D000<10000g0000`40000C00001@00104000420004
0@000@8000@100010P001040004200040@000@8000@1000110000`40003M00030@00018000050004
0@000@8000@100010P001040004200040@000@8000@100010P00104000440081gP000`40000B0000
1@030@<000@100010P001040004200040@000@8000@100010P001040004400030@000=d000<10000
4P0000D000<100000`001040004200040@000@8000@100010P001040004200040@000@@000<10000
g@000`40000B00001P030@<00P44008110020@@00P4400811@000`40003M00030@000180000[0003
0@000=h000<100004@0002/000<10000gP000`40000A0000:`000`40003N00030@000140000[0081
g`000`40000A0000:`000`40003N00030@000140000[00030@000=h000<100004@0002/000<10000
gP000`40000A0000:`000`40003O00030@000100000[00030@000=l000<10000400002/000<10000
g`000`40000@0000:`020N0000<10000400002/000<10000g`000`40000@0000:`000`40003O0003
0@000100000[00030@000>0000<100003`0002/000<10000h0000`40000?0000:`000`40003P0003
0@0000l0000[00030@000>0000<100003`0002/00P7Q00030@0000l0000[00030@000>0000<10000
3`0002/000<10000h0000`40000?0000:`000`40003Q00030@0000h0000[00030@000>4000<10000
3P0000H00P44008110020@@00P44008110020@D000<10000h@000`40000>00001@00104000420004
0@000@8000@100010P001040004200040@000@8000@1000110000`40003Q00030@0000h000050004
0@000@8000@100010P001040004200040@000@8000@100010P00104000440081hP000`40000>0000
1P020@<000@100010P001040004200040@000@8000@100010P001040004400030@000>4000<10000
3P0000D000@100010P001040004200040@000@8000@100010P001040004200040@000@@000<10000
h@000`40000>00001P020@@00P44008110020@@00P4400811@000`40003R00030@0000d0000[0003
0@000>8000<100003@0002/000<10000hP000`40000=0000:`000`40003R00030@0000d0000[0081
h`000`40000=0000:`000`40003R00030@0000d0000[00030@000>8000<100003@0002/000<10000
h`000`40000<0000:`000`40003S00030@0000`0000[00030@000><000<10000300002/00P7T0003
0@0000`0000[00030@000><000<10000300002/000<10000h`000`40000<0000:`000`40003S0003
0@0000`0000[00030@000>@000<100002`0002/000<10000i0000`40000;0000:`000`40003T0003
0@0000/0000[0081i@000`40000;0000:`000`40003T00030@0000/0000[00030@000>@000<10000
2`0002/000<10000i0000`40000;000030030@<000D100000@0200<110020@T000<10000i@000`40
000:00003@000`40000300030@0100@000<100000P001040004800030@000>D000<100002P0000d0
00<1000010000`40000300030@00008000@1000120000`40003U00030@0000X0000=00030@0000<0
00<1004010000`40000200040@000@P00P7V00030@0000X0000=00030@00008000D100000@030003
0@00008000@1000120000`40003U00030@0000X0000<00812`020@D00P4400810`000`40003U0003
0@0000X0000U00030@0100<000<10000i@000`40000:00009@030@<000<10000iP000`4000090000
9@000`40000300030@000>H000<100002@0002H00P4300030@000>H000<100002@0002/00P7W0003
0@0000T0000[00030@000?80000[00030@000?80000[00030@000?80003o0240003o0240003o0240
003o0240003o0240003o0240001O00@11`020@03004100P0104300<10P040@L00P400`010@0800@1
0P040@<00`4200050@01004020020@X00`4300<120040@<00P500000H0000`40000800030@0100X0
00<100000P001@400001008000<1000020000`400@0:00030@0000<000<100000P002`4000010040
0@0100L000@100012@001P4000400@<000<100001P000`40000200040@000Cl0001Q00030@0000P0
00<100002@000`40000200050@0000400P000`40000900030@0000T000<100000`000`400002000;
0@0000400@0100401`001040004900030@00008000D100000@0900050@0000400P000`40000m0000
HP000`40000600030@0100X000<100000P001@40000100800P4900030@0100X000<100000`020@<0
00/100000@0100400@0700040@000@T000<100000P001@40000100X000@100010P000`40000m0000
G`0010400047008100<00@4020040@<00`42008100<00@401P020@03004100P01042008100<00@40
0P030@03004100@11`001040004800@10`030@P000@100010P001040004o0000H0020@<000<10040
40000`40000Q00030@0001/00P4:00030@0000l00P440081@00006H000<100003`030B400`4W0003
0@0005L0003o0240003o0240003o0240003o02400000\
\>"],
  ImageRangeCache->{{{151, 438}, {354.188, 177.25}} -> {-16.6602, 1.20113*^6, 
  0.0857155, 7271.38}}]
}, Open  ]],

Cell["\<\
Most other functions you're familiar with grow slowly though there \
are exceptions like factorial that embarrass even the exponential function.\
\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\
Print[MatrixForm[Prepend[Table[{n,n^2,n^3,n^4,2^n,n!},
                               {n,1,32}],
      \t                 {\"n\",\"n^2\",\"n^3\",\"n^4\",\"2^n\",\"n!\"}]]] \
;\
\>", "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", GridBox[{
            {\("n"\), \("n^2"\), \("n^3"\), \("n^4"\), \("2^n"\), \("n!"\)},
            {"1", "1", "1", "1", "2", "1"},
            {"2", "4", "8", "16", "4", "2"},
            {"3", "9", "27", "81", "8", "6"},
            {"4", "16", "64", "256", "16", "24"},
            {"5", "25", "125", "625", "32", "120"},
            {"6", "36", "216", "1296", "64", "720"},
            {"7", "49", "343", "2401", "128", "5040"},
            {"8", "64", "512", "4096", "256", "40320"},
            {"9", "81", "729", "6561", "512", "362880"},
            {"10", "100", "1000", "10000", "1024", "3628800"},
            {"11", "121", "1331", "14641", "2048", "39916800"},
            {"12", "144", "1728", "20736", "4096", "479001600"},
            {"13", "169", "2197", "28561", "8192", "6227020800"},
            {"14", "196", "2744", "38416", "16384", "87178291200"},
            {"15", "225", "3375", "50625", "32768", "1307674368000"},
            {"16", "256", "4096", "65536", "65536", "20922789888000"},
            {"17", "289", "4913", "83521", "131072", "355687428096000"},
            {"18", "324", "5832", "104976", "262144", "6402373705728000"},
            {"19", "361", "6859", "130321", "524288", "121645100408832000"},
            {"20", "400", "8000", "160000", "1048576", "2432902008176640000"},
            
           
             {"21", "441", "9261", "194481", "2097152", 
              "51090942171709440000"},
           
             {"22", "484", "10648", "234256", "4194304", 
              "1124000727777607680000"},
           
             {"23", "529", "12167", "279841", "8388608", 
              "25852016738884976640000"},
           
             {"24", "576", "13824", "331776", "16777216", 
              "620448401733239439360000"},
           
             {"25", "625", "15625", "390625", "33554432", 
              "15511210043330985984000000"},
           
             {"26", "676", "17576", "456976", "67108864", 
              "403291461126605635584000000"},
            {"27", "729", "19683", "531441", "134217728", 
              "10888869450418352160768000000"},
           
             {"28", "784", "21952", "614656", "268435456", 
              "304888344611713860501504000000"},
           
             {"29", "841", "24389", "707281", "536870912", 
              "8841761993739701954543616000000"},
           
             {"30", "900", "27000", "810000", "1073741824", 
              "265252859812191058636308480000000"},
           
             {"31", "961", "29791", "923521", "2147483648", 
              "8222838654177922817725562880000000"},
           
             {"32", "1024", "32768", "1048576", "4294967296", 
              "263130836933693530167218012160000000"}
            }], ")"}],
      (MatrixForm[ #]&)]], "Print"]
}, Open  ]],

Cell["\<\
The point of this exercise is that, if the time or space required \
to solve your problem is exponential in the size of the problem and the size \
of the problem is likely to get large, then you had better start thinking \
about reformulating your problem, perhaps by sacrificing some aspect of the \
solution requirement, e.g., give up on finding an optimal solution.  We'll \
also encounter problems in which the difference between n^2 and n^3 can make \
a huge difference in our ability to provide good solutions in real time.\
\>",
   "Text"],

Cell["\<\
The difference in the short term between 5% interest and 5% \
interest is pretty small in relative terms.\
\>", "Text"],

Cell["\<\
CompoundInterest[start_,rate_,timesPerYear_Integer,years_Integer] \
:= 
    Module[{amount = start},
\tDo[amount += amount * rate * 1/timesPerYear ,
\t   {i,1/timesPerYear,years,1/timesPerYear}] ;
\tReturn[amount]]\
\>", "Input"],

Cell[CellGroupData[{

Cell["\<\
g1 = ListPlot[Table[CompoundInterest[500000,0.05,1,i],{i,1,10}],
\t\t\t\t\tPlotJoined -> True] ;\
\>", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.0820106 0.10582 -1.0529 2.03354e-06 [
[.34127 .05305 -3 -9 ]
[.34127 .05305 3 0 ]
[.55291 .05305 -3 -9 ]
[.55291 .05305 3 0 ]
[.76455 .05305 -3 -9 ]
[.76455 .05305 3 0 ]
[.97619 .05305 -6 -9 ]
[.97619 .05305 6 0 ]
[.11713 .16723 -36 -4.5 ]
[.11713 .16723 0 4.5 ]
[.11713 .26891 -36 -4.5 ]
[.11713 .26891 0 4.5 ]
[.11713 .37059 -36 -4.5 ]
[.11713 .37059 0 4.5 ]
[.11713 .47226 -36 -4.5 ]
[.11713 .47226 0 4.5 ]
[.11713 .57394 -36 -4.5 ]
[.11713 .57394 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.34127 .06555 m
.34127 .0718 L
s
[(4)] .34127 .05305 0 1 Mshowa
.55291 .06555 m
.55291 .0718 L
s
[(6)] .55291 .05305 0 1 Mshowa
.76455 .06555 m
.76455 .0718 L
s
[(8)] .76455 .05305 0 1 Mshowa
.97619 .06555 m
.97619 .0718 L
s
[(10)] .97619 .05305 0 1 Mshowa
.125 Mabswid
.18254 .06555 m
.18254 .0693 L
s
.23545 .06555 m
.23545 .0693 L
s
.28836 .06555 m
.28836 .0693 L
s
.39418 .06555 m
.39418 .0693 L
s
.44709 .06555 m
.44709 .0693 L
s
.5 .06555 m
.5 .0693 L
s
.60582 .06555 m
.60582 .0693 L
s
.65873 .06555 m
.65873 .0693 L
s
.71164 .06555 m
.71164 .0693 L
s
.81746 .06555 m
.81746 .0693 L
s
.87037 .06555 m
.87037 .0693 L
s
.92328 .06555 m
.92328 .0693 L
s
.07672 .06555 m
.07672 .0693 L
s
.02381 .06555 m
.02381 .0693 L
s
.25 Mabswid
0 .06555 m
1 .06555 L
s
.12963 .16723 m
.13588 .16723 L
s
[(600000)] .11713 .16723 1 0 Mshowa
.12963 .26891 m
.13588 .26891 L
s
[(650000)] .11713 .26891 1 0 Mshowa
.12963 .37059 m
.13588 .37059 L
s
[(700000)] .11713 .37059 1 0 Mshowa
.12963 .47226 m
.13588 .47226 L
s
[(750000)] .11713 .47226 1 0 Mshowa
.12963 .57394 m
.13588 .57394 L
s
[(800000)] .11713 .57394 1 0 Mshowa
.125 Mabswid
.12963 .08589 m
.13338 .08589 L
s
.12963 .10622 m
.13338 .10622 L
s
.12963 .12656 m
.13338 .12656 L
s
.12963 .1469 m
.13338 .1469 L
s
.12963 .18757 m
.13338 .18757 L
s
.12963 .2079 m
.13338 .2079 L
s
.12963 .22824 m
.13338 .22824 L
s
.12963 .24857 m
.13338 .24857 L
s
.12963 .28924 m
.13338 .28924 L
s
.12963 .30958 m
.13338 .30958 L
s
.12963 .32991 m
.13338 .32991 L
s
.12963 .35025 m
.13338 .35025 L
s
.12963 .39092 m
.13338 .39092 L
s
.12963 .41126 m
.13338 .41126 L
s
.12963 .43159 m
.13338 .43159 L
s
.12963 .45193 m
.13338 .45193 L
s
.12963 .4926 m
.13338 .4926 L
s
.12963 .51293 m
.13338 .51293 L
s
.12963 .53327 m
.13338 .53327 L
s
.12963 .5536 m
.13338 .5536 L
s
.12963 .04522 m
.13338 .04522 L
s
.12963 .02488 m
.13338 .02488 L
s
.12963 .00455 m
.13338 .00455 L
s
.12963 .59428 m
.13338 .59428 L
s
.12963 .61461 m
.13338 .61461 L
s
.25 Mabswid
.12963 0 m
.12963 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .01472 m
.12963 .0681 L
.23545 .12415 L
.34127 .183 L
.44709 .24479 L
.55291 .30968 L
.65873 .3778 L
.76455 .44934 L
.87037 .52445 L
.97619 .60332 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/@P10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdk34g_ojn_l004J2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_P00oiT0IP3o0;C^]:JV
YZf][Jo^kP20P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;0000?`00O`00_`0o000o
?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0
_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo?cooOcoo
_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no?gnoOgno
_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo
_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0_ol0oolo0?lo?oloOolo_olooomo
0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_mWEc:0Pl=WIfJj^[WMgM`0:0>o<knl1
k`1^06EUI@00I@00[C8bI@1U002]039U<VD00:d006Db<RPX:000:00X02P002PX:00X02P00<`cVL`c
c<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co
<lcoI/coVLcoc<coool0<ol0I_l0VOl0c?l0oolc0?lc<olcI_lcVOlcc?lcoomV0?mVI_mVVOmVoonI
0?nI<onII_nIVOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<obL000<10000mP0002L00P7g00009`000`40
003f00009`000`40003f00002@020A`000<10000mP0000/00P4J00030@000?H0000=008160000`40
003f00003`020AH00P7g00004@000`40000C00030@0003P00`4i0081>`020CH00`44008110000180
0P4C00030@0003T000<10000=`001040004i00040@000CH000<100000P0010400043000050020A40
00<10000=`040CP000@10001>@001040004f00030@00008000@100010`0001H00P4?00030@0003L0
00<10040>@030C/00P4g00030@00008000@100010`0001P00P4=00030@0003L000<10040>@000`40
000j00040@000CH000<100000P001040004300006P020@/00P4i0081>P030CX00P4f00811@020@@0
000L00812@000`40003f00007P020@L000<10000mP00020000<1000010000`40003f00008@020@@0
00<10000mP0002<00P4200030@000?H000020?l17`4000T000<1000030000`40000<00813@000`40
000<00030@0000`000<1000030000`40000<00030@0000`000<1000030000`40000<00030@0000`0
00<1000030000`40000=00030@0000`000<1000030000`40000<00030@0000`000<1000030000`40
000500009`0010400@4h00030@0003T000<10000>P000`40000i00030@0000D0000W00040@000081
l`0002L000<100000`020O40000W00030@0000D00P7_00009`020@P00P7]00009`000`4000090081
j`0002L000<100002`020NT0000W00030@0000d00P7W00009`000`40000?0081i@0002L00P4B0081
h`0002L000<100004`020N40000W00030@0001D00P7O00009`000`40000G0081g@0002L000<10000
6@020M/0000W00030@0001/00P7I00009`020Ah00P7G00009`000`40000O00030@000=@0000W0003
0@0002000P7D00009`000`40000R0081dP0002L000<1000090020M00000W00030@0002H000<10000
c@0002L00P4X0081c@0002L000<10000:@020L/0000W00030@0002/00P7900009`000`40000]0081
a`0000800P44008110020@@00P44008110020@D000<10000;`000`400034000000D00@000@020004
0@000@8000@100010P001040004200040@000@8000@1000110000`40000`0081a000000500400040
0P001040004200040@000@8000@100010P001040004200040@000@@00`4b0081`P00000400410@<0
00@100010P001040004200040@000@8000@100010P001040004400030@0003@00P70000000<00@00
10001040004200040@000@8000@100010P001040004200040@000@@000<10000=P000`40002m0000
0P030@<00P44008110020@@00P4400811@000`40000g0081_@0002L000<10000>@020K/0000W0003
0@0003/000<10000^00002L00P4m0081^00002L000<10000?P020KH0000W00030@00040000<10000
/`0002L000<10000@@020K<0000W00030@0004<00P6a00009`020DH000<10000[P0002L000<10000
AP020Jh0000W00030@0004P00P6/00009`000`40001:00030@000:T0000W00030@0004/00P6Y0000
9`000`40001=0081Y`0002L00P5@00030@000:@0000W00030@0005000P6T00009`000`40001B0081
XP0002L000<10000E0000`40002O00009`000`40001E0081W`0002L000<10000E`020Id0000W0081
FP000`40002J00009`000`40001J0081VP0002L000<10000G0020IP0000W00030@0005h000<10000
U@0000800P4300<110020@@00P44008110020@D000<10000G`020ID000001@01000100D000@10001
0P001040004200040@000@8000@100010P000`40000200030@0006400P6C000000D00@000@050004
0@000@8000@100010P001040004200040@000@8000<100000P030F<000<10000T000000400410@<0
0`4300040@000@8000@100010P001040004200040@000@@000<10000I0020I0000000`0100040003
0@0000<000@100010P001040004200040@000@8000@1000110000`40001V0081SP0000800`4200@1
0`020@@00P44008110020@D000<10000J0000`40002;00009`000`40001Y0081R`0002L00P5/0081
R@0002L000<10000K@000`40002600009`000`40001^0081QP0002L000<10000L0020H@0000W0003
0@00078000<10000P@0002L000<10000L`020H40000W0081MP020Gl0000W00030@0007L000<10000
O00002L000<10000N0020G`0000W00030@0007X000<10000N@0002L000<10000N`020GT0000W0003
0@0007d000<10000MP0002L00P5o0081MP0002L000<10000P0000`40001c00009`000`4000210081
L`0002L000<10000P`000`40001`00009`000`4000240081L00002L000<10000QP000`40001]0000
9`020HP00P5]00009`000`40002900030@0006X0000W00030@0008X00P5Z00009`000`40002<0003
0@0006L0000200030@0000<00P44008110020@@00P4400811@000`40002=0081I`00008000<10000
0P001040004200040@000@8000@100010P001040004200040@000@@000<10000S`000`40001T0000
0`001@400001008000@100010P001040004200040@000@8000@100010P000`40000200<1T0020F@0
000300050@0000400P001040004200040@000@8000@100010P001040004200030@00008000<10000
TP000`40001Q000010001040004200040@000@8000@100010P001040004200040@000@8000<10000
0P000`40002C0081H@000003004100810`020@@00P44008110020@@00P4500030@0009D000<10000
GP0002L000<10000UP020Eh0000W0081V@000`40001K00009`000`40002I0081F`0002L000<10000
V`020ET0000W00030@0009d000<10000EP0002L000<10000WP020EH0000W00030@000:0000<10000
D`0002L00P6R0081D`0002L000<10000X`000`40001@00009`000`40002T0081D00002L000<10000
YP000`40001=00009`000`40002W0081C@0002L000<10000Z@000`40001:00009`020J/00P5:0000
9`000`40002/0081B00002L000<10000[P000`40001500009`000`40002_0081A@0002L000<10000
/@000`40001200009`000`40002b0081@P0002L00P6e00030@0003l0000W00030@000;D00P4o0000
9`000`40002g00030@0003`0000200030@0000800`44008110020@@00P4400811@000`40002h0003
0@0003/0000200030@0000D000@100010P001040004200040@000@8000@100010P000`4000020003
0@000;T00P4k00000`000`40000400040@000@8000@100010P001040004200040@000@8000<10000
0P030K/000<10000>00000<000@100000`4300040@000@8000@100010P001040004200040@000@@0
00<10000_0020CP0000400040@000@D000@100010P001040004200040@000@8000@1000110000`40
002n00030@0003D000000`010@020@801043008110020@@00P4400811@000`40002o0081=@0002L0
00<10000`@000`40000b00009`000`40003200030@000340000W0081a0020C40000W00030@000<D0
00<10000;P0002L000<10000aP020Bh0000W00030@000<P000<10000:`0002L000<10000b@020B/0
000W00030@000</000<10000:00002L00P7=00030@0002L0000W00030@000<d00P4W00009`000`40
003?00030@0002@0000W00030@000=000P4T00009`000`40003B00030@000240000W00030@000=<0
00<10000800002L00P7E0081800002L000<10000eP000`40000M00009`000`40003G00030@0001`0
000W00030@000=P000<100006`0002L000<10000f@020A/0000W00030@000=/000<10000600002L0
0P7M00030@0001L0000W00030@000=d00P4G00009`000`40003O00030@0001@00002008110020@@0
0P44008110020@@00P4500030@000>0000<100004`000005004000400P001040004200040@000@80
00@100010P001040004200040@000@@000<10000h@020A<000001@010001008000@100010P001040
004200040@000@8000@100010P001040004400<1h`000`40000@00000P020@<000@100010P001040
004200040@000@8000@100010P001040004400030@000>@000<100003`000005004000400P001040
004200040@000@8000@100010P001040004200040@000@@000<10000i@020@l00002008110020@@0
0P44008110020@@00P4500030@000>L000<10000300002L000<10000j0000`40000;00009`000`40
003Y00030@0000X0000W0081j`020@X0000W00030@000>`000<100001`0002L000<10000k@000`40
000600009`000`40003^00030@0000D0000W00030@000?H0000W00030@000?H0000W0081m`0002L0
00<10000mP000001\
\>"],
  ImageRangeCache->{{{151, 438}, {298.188, 121.25}} -> {-4.33402, 726428, 
  0.0332399, 1729.72}}]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
g2 = ListPlot[Table[CompoundInterest[500000,0.06,1,i],{i,1,10}],
\t\t\t\t\tPlotJoined -> True] ;\
\>", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.0820106 0.10582 -0.838979 1.61074e-06 [
[.34127 .03443 -3 -9 ]
[.34127 .03443 3 0 ]
[.55291 .03443 -3 -9 ]
[.55291 .03443 3 0 ]
[.76455 .03443 -3 -9 ]
[.76455 .03443 3 0 ]
[.97619 .03443 -6 -9 ]
[.97619 .03443 6 0 ]
[.11713 .12747 -36 -4.5 ]
[.11713 .12747 0 4.5 ]
[.11713 .208 -36 -4.5 ]
[.11713 .208 0 4.5 ]
[.11713 .28854 -36 -4.5 ]
[.11713 .28854 0 4.5 ]
[.11713 .36908 -36 -4.5 ]
[.11713 .36908 0 4.5 ]
[.11713 .44962 -36 -4.5 ]
[.11713 .44962 0 4.5 ]
[.11713 .53015 -36 -4.5 ]
[.11713 .53015 0 4.5 ]
[.11713 .61069 -36 -4.5 ]
[.11713 .61069 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.34127 .04693 m
.34127 .05318 L
s
[(4)] .34127 .03443 0 1 Mshowa
.55291 .04693 m
.55291 .05318 L
s
[(6)] .55291 .03443 0 1 Mshowa
.76455 .04693 m
.76455 .05318 L
s
[(8)] .76455 .03443 0 1 Mshowa
.97619 .04693 m
.97619 .05318 L
s
[(10)] .97619 .03443 0 1 Mshowa
.125 Mabswid
.18254 .04693 m
.18254 .05068 L
s
.23545 .04693 m
.23545 .05068 L
s
.28836 .04693 m
.28836 .05068 L
s
.39418 .04693 m
.39418 .05068 L
s
.44709 .04693 m
.44709 .05068 L
s
.5 .04693 m
.5 .05068 L
s
.60582 .04693 m
.60582 .05068 L
s
.65873 .04693 m
.65873 .05068 L
s
.71164 .04693 m
.71164 .05068 L
s
.81746 .04693 m
.81746 .05068 L
s
.87037 .04693 m
.87037 .05068 L
s
.92328 .04693 m
.92328 .05068 L
s
.07672 .04693 m
.07672 .05068 L
s
.02381 .04693 m
.02381 .05068 L
s
.25 Mabswid
0 .04693 m
1 .04693 L
s
.12963 .12747 m
.13588 .12747 L
s
[(600000)] .11713 .12747 1 0 Mshowa
.12963 .208 m
.13588 .208 L
s
[(650000)] .11713 .208 1 0 Mshowa
.12963 .28854 m
.13588 .28854 L
s
[(700000)] .11713 .28854 1 0 Mshowa
.12963 .36908 m
.13588 .36908 L
s
[(750000)] .11713 .36908 1 0 Mshowa
.12963 .44962 m
.13588 .44962 L
s
[(800000)] .11713 .44962 1 0 Mshowa
.12963 .53015 m
.13588 .53015 L
s
[(850000)] .11713 .53015 1 0 Mshowa
.12963 .61069 m
.13588 .61069 L
s
[(900000)] .11713 .61069 1 0 Mshowa
.125 Mabswid
.12963 .06304 m
.13338 .06304 L
s
.12963 .07914 m
.13338 .07914 L
s
.12963 .09525 m
.13338 .09525 L
s
.12963 .11136 m
.13338 .11136 L
s
.12963 .14357 m
.13338 .14357 L
s
.12963 .15968 m
.13338 .15968 L
s
.12963 .17579 m
.13338 .17579 L
s
.12963 .1919 m
.13338 .1919 L
s
.12963 .22411 m
.13338 .22411 L
s
.12963 .24022 m
.13338 .24022 L
s
.12963 .25633 m
.13338 .25633 L
s
.12963 .27243 m
.13338 .27243 L
s
.12963 .30465 m
.13338 .30465 L
s
.12963 .32076 m
.13338 .32076 L
s
.12963 .33686 m
.13338 .33686 L
s
.12963 .35297 m
.13338 .35297 L
s
.12963 .38519 m
.13338 .38519 L
s
.12963 .40129 m
.13338 .40129 L
s
.12963 .4174 m
.13338 .4174 L
s
.12963 .43351 m
.13338 .43351 L
s
.12963 .46572 m
.13338 .46572 L
s
.12963 .48183 m
.13338 .48183 L
s
.12963 .49794 m
.13338 .49794 L
s
.12963 .51405 m
.13338 .51405 L
s
.12963 .54626 m
.13338 .54626 L
s
.12963 .56237 m
.13338 .56237 L
s
.12963 .57848 m
.13338 .57848 L
s
.12963 .59458 m
.13338 .59458 L
s
.12963 .03082 m
.13338 .03082 L
s
.12963 .01472 m
.13338 .01472 L
s
.25 Mabswid
.12963 0 m
.12963 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .01472 m
.12963 .06594 L
.23545 .12023 L
.34127 .17778 L
.44709 .23879 L
.55291 .30346 L
.65873 .372 L
.76455 .44466 L
.87037 .52168 L
.97619 .60332 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/@P10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdk34g_ojn_l004J2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_P00oiT0IP3o0;C^]:JV
YZf][Jo^kP20P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;0000?`00O`00_`0o000o
?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0
_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo?cooOcoo
_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no?gnoOgno
_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo
_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0_ol0oolo0?lo?oloOolo_olooomo
0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_mWEc:0Pl=WIfJj^[WMgM`0:0>o<knl1
k`1^06EUI@00I@00[C8bI@1U002]039U<VD00:d006Db<RPX:000:00X02P002PX:00X02P00<`cVL`c
c<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co
<lcoI/coVLcoc<coool0<ol0I_l0VOl0c?l0oolc0?lc<olcI_lcVOlcc?lcoomV0?mVI_mVVOmVoonI
0?nI<onII_nIVOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<obP000<10000m@0002P000<10000m@0002P0
00<10000>0030CP00P4j0081=P030@@00P450000:0000`40000i00030@0003H000@10001>0001040
004f00030@00008000@10001100000X00P4L0081>0040CL000@10001>0001040004f00030@000080
00@10001100000`00P4J00030@0003L000<10040>0030CX00P4g00030@00008000@10001100000h0
0P4H00030@0003L000<10040>0000`40000i00040@000CH000<100000P0010400044000040020AH0
00<10000>0020CT00`4i0081=P020@D00P4500004P030A<00P7f00005@020A4000<10000m@0001L0
0P4?00030@000?D0000I00813@000`40003e00006`020@/000<10000m@0000<0o`4M0@40000:0003
0@0000`000<100000`020@L000<1000030000`40000<00030@0000`000<1000030000`40000<0003
0@0000`000<100002`000`40000<00030@0000`000<1000030000`40000<00030@0000`000<10000
30000`40000<00030@0000`000<1000030000`40000600008@030@@000<10000>@000`40000h0003
0@0003T000<10000>@000`400006000090020@8000<10000m@0002H0107f0000:0020OH0000X0004
0@010O@0000X00040@000081lP0002P000<100000`020O00000X00811P000`40003]0000:0000`40
00060081k@0002P000<1000020020N/0000X00030@0000X00P7Y0000:0020@d00P7W0000:0000`40
000>0081i@0002P000<1000040020N<0000X00030@0001800P7Q0000:0000`40000D00030@000=h0
000X00815P020Mh0000X00030@0001L00P7L0000:0000`40000I0081fP0000800P44008110020@@0
0P44008110020@H000<100006`020MP000001@010001008000@100010P001040004200040@000@80
00@100010P001040004500030@0001d00P7F000000D00@000@0200040@000@8000@100010P001040
004200040@000@8000@100011@030Al00P7D000000@00@410`001040004200040@000@8000@10001
0P001040004200040@000@D000<100008@020M8000000`01000400040@000@8000@100010P001040
004200040@000@8000@100011@000`40000S00030@000<l0000200<10`020@@00P44008110020@@0
0P4600030@0002@00P7?0000:0020BL00P7=0000:0000`40000X0081b`0002P000<10000:P020LT0
000X00030@0002`00P770000:0000`40000^0081a@0002P00P4a0081``0002P000<10000<P000`40
00300000:0000`40000c0081`00002P000<10000=@020Kh0000X0081>0020K`0000X00030@0003T0
0P6j0000:0000`40000k0081^00002P000<10000?@000`40002e0000:0000`40000n0081]@0002P0
0P510081/`0002P000<10000@P020K40000200810`030@@00P44008110020@@00P4600030@0004@0
00<10000[P000005004000401@001040004200040@000@8000@100010P001040004200030@0000<0
00<10000A@020Jh000001@01000100D000@100010P001040004200040@000@8000@100010P000`40
000300<1A`020J`0000010010@4300<10`001040004200040@000@8000@100010P00104000450003
0@0004T00P6Z000000<00@0010000`40000300040@000@8000@100010P001040004200040@000@D0
00<10000B`020JP0000200<10P040@<00P44008110020@@00P4600030@0004d000<10000Y@0002P0
00<10000CP020JD0000X0081D@020J<0000X00030@0005800P6Q0000:0000`40001D00030@0009h0
000X00030@0005D00P6N0000:0020EP00P6L0000:0000`40001I00030@0009T0000X00030@0005X0
0P6I0000:0000`40001L0081U`0002P000<10000GP000`40002D0000:0020F000P6D0000:0000`40
001Q00030@000940000X00030@0006800P6A0000:0000`40001T0081S`0002P00P5W00030@0008`0
000X00030@0006L00P6<0000:0000`40001Y0081RP00008000<100000`020@@00P44008110020@@0
0P4600030@0006/000<10000Q`00008000<100000P001040004200040@000@8000@100010P001040
004200040@000@D000<10000K0020HL0000300050@0000400P001040004200040@000@8000@10001
0P001040004200030@0000<00`5^00030@0008@0000300050@0000400P001040004200040@000@80
00@100010P001040004200030@0000<000<10000K`020H@0000400040@000@8000@100010P001040
004200040@000@8000@100010P000`40000300030@0007400P62000000<00@400P43008110020@@0
0P44008110020@H000<10000L`000`40001o0000:0000`40001d0081O`0002P00P5g00030@0007`0
000X00030@0007L00P5l0000:0000`40001i00030@0007T0000X00030@0007X00P5i0000:0020Gd0
00<10000MP0002P000<10000O@020GH0000X00030@0007l000<10000L`0002P000<10000P0020G<0
000X00030@00088000<10000L00002P00P640081L00002P000<10000Q@000`40001]0000:0000`40
00260081K@0002P000<10000R0000`40001Z0000:0020HX00P5Z0000:0000`40002;00030@0006L0
000X00030@0008`00P5W00000P000`40000200<110020@@00P44008110020@H000<10000SP000`40
001T00000P000`40000500040@000@8000@100010P001040004200040@000@8000<100000`000`40
002?0081I00000<000<1000010001040004200040@000@8000@100010P001040004200030@0000<0
0`6A00030@000640000300040@0000<10`001040004200040@000@8000@100010P00104000450003
0@0009800P5Q000010001040004500040@000@8000@100010P001040004200040@000@D000<10000
U0000`40001N000000<00@400P4200@10`020@@00P44008110020@H000<10000U@020Eh0000X0081
V0000`40001K0000:0000`40002H0081F`0002P000<10000VP000`40001H0000:0000`40002K0081
F00002P000<10000W@000`40001E0000:0020Il00P5E0000:0000`40002P00030@000580000X0003
0@000:400P5B0000:0000`40002S00030@0004l0000X0081Y@020Dl0000X00030@000:H000<10000
C00002P000<10000Y`020D`0000X00030@000:T000<10000B@0002P000<10000ZP020DT0000X0081
[@000`4000160000:0000`40002]0081AP0002P000<10000[`000`40001300000P020@@00P440081
10020@@00P4400811P000`40002`0081@`000005004000400P001040004200040@000@8000@10001
0P001040004200040@000@D000<10000/P000`400010000000D00@000@0200040@000@8000@10001
0P001040004200040@000@8000@100011@030K<000<10000?`0000800P4300040@000@8000@10001
0P001040004200040@000@8000@100011@000`40002d0081?`000005004000400P00104000420004
0@000@8000@100010P001040004200040@000@D000<10000]P000`40000l00000P020@@00P440081
10020@@00P4400811P000`40002g00030@0003/0000X0081^@020C/0000X00030@000;X000<10000
>00002P000<10000^`020CP0000X00030@000;d000<10000=@0002P000<10000_P000`40000d0000
:0020L000P4d0000:0000`40003100030@000340000X00030@000<8000<10000<00002P000<10000
``020C00000X0081aP000`40000]0000:0000`40003600030@0002`0000X00030@000<L00P4/0000
:0000`40003900030@0002T0000X00030@000<X00P4Y0000:0020Ld000<100009P0002P000<10000
c@000`40000U00000P020@<00`44008110020@@00P4400811P000`40003>00819@00000500400040
1@001040004200040@000@8000@100010P001040004200030@0000<000<10000d0000`40000R0000
00D00@000@0500040@000@8000@100010P001040004200040@000@8000<100000`030M4000<10000
8@0000800P4300<10`001040004200040@000@8000@100010P001040004500030@000=800P4Q0000
00D00@000@0200030@0000<000@100010P001040004200040@000@8000@100011@000`40003D0003
0@0001h0000200810`040@<00P44008110020@@00P4600030@000=D000<100007@0002P000<10000
eP000`40000L0000:0020MP00P4L0000:0000`40003I00030@0001T0000X00030@000=X000<10000
600002P000<10000f`020AP0000X0081gP000`40000E0000:0000`40003N00030@0001@0000X0003
0@000=l00P4D0000:0000`40003Q00030@000140000X00030@000>8000<10000400002P00P7T0081
400002P000<10000i@000`40000=0000:0000`40003V00030@0000`0000X00030@000>L000<10000
2`0002P00P7Y00812`0002P000<10000jP000`4000080000:0000`40003[00030@0000L000001001
0@44008110020@@00P44008110020@H000<10000k0000`400006000010001040004200040@000@80
00@100010P001040004200040@000@8000<100000`000`40003e00000P030@8000@100010P001040
004200040@000@8000@100010P001040004500<1m@000005004000400P001040004200040@000@80
00@100010P001040004200040@000@D000<10000m@000005004000400P001040004200040@000@80
00@100010P001040004200040@000@D000<10000m@0000800P44008110020@@00P44008110020Oh0
003o02400000\
\>"],
  ImageRangeCache->{{{151, 438}, {529.375, 352.438}} -> {-4.40521, 
  1.29577*^6, 0.0334677, 2198.71}}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Show[g1,g2] ;", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0952381 -0.81951 1.589e-06 [
[.21429 .04194 -3 -9 ]
[.21429 .04194 3 0 ]
[.40476 .04194 -3 -9 ]
[.40476 .04194 3 0 ]
[.59524 .04194 -3 -9 ]
[.59524 .04194 3 0 ]
[.78571 .04194 -3 -9 ]
[.78571 .04194 3 0 ]
[.97619 .04194 -6 -9 ]
[.97619 .04194 6 0 ]
[.01131 .13389 -36 -4.5 ]
[.01131 .13389 0 4.5 ]
[.01131 .21334 -36 -4.5 ]
[.01131 .21334 0 4.5 ]
[.01131 .29279 -36 -4.5 ]
[.01131 .29279 0 4.5 ]
[.01131 .37224 -36 -4.5 ]
[.01131 .37224 0 4.5 ]
[.01131 .45169 -36 -4.5 ]
[.01131 .45169 0 4.5 ]
[.01131 .53114 -36 -4.5 ]
[.01131 .53114 0 4.5 ]
[.01131 .61059 -36 -4.5 ]
[.01131 .61059 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.21429 .05444 m
.21429 .06069 L
s
[(2)] .21429 .04194 0 1 Mshowa
.40476 .05444 m
.40476 .06069 L
s
[(4)] .40476 .04194 0 1 Mshowa
.59524 .05444 m
.59524 .06069 L
s
[(6)] .59524 .04194 0 1 Mshowa
.78571 .05444 m
.78571 .06069 L
s
[(8)] .78571 .04194 0 1 Mshowa
.97619 .05444 m
.97619 .06069 L
s
[(10)] .97619 .04194 0 1 Mshowa
.125 Mabswid
.07143 .05444 m
.07143 .05819 L
s
.11905 .05444 m
.11905 .05819 L
s
.16667 .05444 m
.16667 .05819 L
s
.2619 .05444 m
.2619 .05819 L
s
.30952 .05444 m
.30952 .05819 L
s
.35714 .05444 m
.35714 .05819 L
s
.45238 .05444 m
.45238 .05819 L
s
.5 .05444 m
.5 .05819 L
s
.54762 .05444 m
.54762 .05819 L
s
.64286 .05444 m
.64286 .05819 L
s
.69048 .05444 m
.69048 .05819 L
s
.7381 .05444 m
.7381 .05819 L
s
.83333 .05444 m
.83333 .05819 L
s
.88095 .05444 m
.88095 .05819 L
s
.92857 .05444 m
.92857 .05819 L
s
.25 Mabswid
0 .05444 m
1 .05444 L
s
.02381 .13389 m
.03006 .13389 L
s
[(600000)] .01131 .13389 1 0 Mshowa
.02381 .21334 m
.03006 .21334 L
s
[(650000)] .01131 .21334 1 0 Mshowa
.02381 .29279 m
.03006 .29279 L
s
[(700000)] .01131 .29279 1 0 Mshowa
.02381 .37224 m
.03006 .37224 L
s
[(750000)] .01131 .37224 1 0 Mshowa
.02381 .45169 m
.03006 .45169 L
s
[(800000)] .01131 .45169 1 0 Mshowa
.02381 .53114 m
.03006 .53114 L
s
[(850000)] .01131 .53114 1 0 Mshowa
.02381 .61059 m
.03006 .61059 L
s
[(900000)] .01131 .61059 1 0 Mshowa
.125 Mabswid
.02381 .07033 m
.02756 .07033 L
s
.02381 .08622 m
.02756 .08622 L
s
.02381 .10211 m
.02756 .10211 L
s
.02381 .118 m
.02756 .118 L
s
.02381 .14978 m
.02756 .14978 L
s
.02381 .16567 m
.02756 .16567 L
s
.02381 .18156 m
.02756 .18156 L
s
.02381 .19745 m
.02756 .19745 L
s
.02381 .22923 m
.02756 .22923 L
s
.02381 .24512 m
.02756 .24512 L
s
.02381 .26101 m
.02756 .26101 L
s
.02381 .2769 m
.02756 .2769 L
s
.02381 .30868 m
.02756 .30868 L
s
.02381 .32457 m
.02756 .32457 L
s
.02381 .34046 m
.02756 .34046 L
s
.02381 .35635 m
.02756 .35635 L
s
.02381 .38813 m
.02756 .38813 L
s
.02381 .40402 m
.02756 .40402 L
s
.02381 .41991 m
.02756 .41991 L
s
.02381 .4358 m
.02756 .4358 L
s
.02381 .46758 m
.02756 .46758 L
s
.02381 .48347 m
.02756 .48347 L
s
.02381 .49936 m
.02756 .49936 L
s
.02381 .51525 m
.02756 .51525 L
s
.02381 .54703 m
.02756 .54703 L
s
.02381 .56292 m
.02756 .56292 L
s
.02381 .57881 m
.02756 .57881 L
s
.02381 .5947 m
.02756 .5947 L
s
.02381 .03855 m
.02756 .03855 L
s
.02381 .02266 m
.02756 .02266 L
s
.02381 .00677 m
.02756 .00677 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.11905 .01472 m
.21429 .05643 L
.30952 .10022 L
.40476 .14621 L
.5 .1945 L
.59524 .2452 L
.69048 .29843 L
.78571 .35433 L
.88095 .41302 L
.97619 .47465 L
s
.11905 .02266 m
.21429 .07319 L
.30952 .12675 L
.40476 .18353 L
.5 .24371 L
.59524 .3075 L
.69048 .37512 L
.78571 .4468 L
.88095 .52278 L
.97619 .60332 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/@P10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdk34g_ojn_l004J2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_P00oiT0IP3o0;C^]:JV
YZf][Jo^kP20P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;0000?`00O`00_`0o000o
?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0
_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo?cooOcoo
_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no?gnoOgno
_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo
_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0_ol0oolo0?lo?oloOolo_olooomo
0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_mWEc:0Pl=WIfJj^[WMgM`0:0>o<knl1
k`1^06EUI@00I@00[C8bI@1U002]039U<VD00:d006Db<RPX:000:00X02P002PX:00X02P00<`cVL`c
c<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co
<lcoI/coVLcoc<coool0<ol0I_l0VOl0c?l0oolc0?lc<olcI_lcVOlcc?lcoomV0?mVI_mVVOmVoonI
0?nI<onII_nIVOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<ool08@000?l08@000?l08@000?l08@000?l0
8@000?l08@000?l08@000?l08@0002L000<10000mP0002L000<10000mP0002L00P7g00009`000`40
003f00009`000`40000E008150040Bd00`4^0081;P020BX00`440081100002L000<100005`020A<0
00<10000;P000`40000/00040@000B`000@10001:P000`40000200040@000@<0000W00815P020@80
0P4B00030@0002/0104]00040@000B`000@10001:P000`40000200040@000@<0000W00030@0001L0
0P4200<140000`40000Z00030@0102h00`4^0081:`000`40000200040@000@<0000W00030@0001T0
0P4300812`001040004/00030@0102h000<10000;@001040004Z00030@00008000@100010`0002L0
00<100006`000`40000200812P020Bh00P4_00<1;@020BX00P450081100002L00P4M008110020M80
000W00030@0001h00P440081d00002L000<1000080020@@00`7=00009`000`40000R00811@020L/0
000Q0?l1000W00030@0000T000<100002@000`40000900030@0000800P4500812P000`4000090003
0@0000T000<100002@000`40000900030@0000X000<100002@000`40000900030@0000T000<10000
2@000`40000900030@0000T000<100002@000`40000900030@0000T000<100002@000`4000050000
9`000`40000X00030@00008000@10041;0000`40000^00030@0002d000<10000;@000`4000050000
9`000`40000Y00811P020L<0000W0081;0020@H00`7000009`000`40000]00811`020Kh0000W0003
0@0002l00P470081_00002L000<10000<@020@L00P6j00009`020C@000<100001P020KP0000W0003
0@0003@00P4800<1]@0002L000<10000=P020@T00P6c00009`000`40000h00812@020K40000W0081
>`020@T00P6_00009`000`40000l00812@020Jd0000W00030@0003h00P490081Z`0002L000<10000
@0000`40000800<1Z00002L00P5200812`020JH0000W00030@0004<00P4;0081Y00000800P440081
10020@@00P44008110020@D000<10000A@020@/00P6R000000D00@000@0200040@000@8000@10001
0P001040004200040@000@8000@1000110000`40001700030@0000X00P6P000000D00@000@020004
0@000@8000@100010P001040004200040@000@8000@1000110020DT00P4<00<1W@00000400410@<0
00@100010P001040004200040@000@8000@100010P001040004400030@0004X00P4=0081V`000003
004000@000@100010P001040004200040@000@8000@100010P001040004400030@0004`000<10000
30020IT0000200<10`020@@00P44008110020@@00P4500030@0004d00P4>0081U`0002L00P5@0003
0@0000d00P6E00009`000`40001@00813`020I<0000W00030@0005800P4?0081T@0002L000<10000
E0000`40000>0081S`0002L00P5F008140020Hd0000W00030@0005L000<100003`000`40002:0000
9`000`40001H008140020HX0000W00030@0005X00P4@0081R00002L00P5M00030@0000l00P660000
9`000`40001M00814@020H@0000W00030@0005l000<1000040020H80000W00030@0006000P4B0081
P00002L00P5S00814P020Gh0000W00030@0006@000<100004@020G`0000200810`030@@00P440081
10020@@00P4500030@0006D00P4C0081NP000005004000401@001040004200040@000@8000@10001
0P001040004200030@00008000<10000I`020A<000<10000M`000005004000401@00104000420004
0@000@8000@100010P001040004200030@0000800P5Z00030@0001400P5g000000@00@410`030@<0
00@100010P001040004200040@000@8000@1000110000`40001Z00814`020GD000000`0100040003
0@0000<000@100010P001040004200040@000@8000@1000110000`40001/00814`020G<0000200<1
0P040@<00P44008110020@@00P4500030@0006h000<100004P020G40000W0081L0020A@00P5_0000
9`000`40001a008150020Fd0000W00030@0007<000<100004`000`40001Z00009`000`40001d0081
50020FX0000W00030@0007H00P4D0081J00002L00P5i00030@0001<00P5V00009`000`40001i0081
5@020F@0000W00030@0007/000<1000050020F80000W00030@0007`00P4F00030@0005l0000W0081
O`000`40000D0081G`0002L000<10000O`020AH00P5M00009`000`40002100030@0001D00P5K0000
9`000`40002200815`020ET0000W0081Q@000`40000F0081E`0002L000<10000Q@020AP00P5E0000
0P000`400003008110020@@00P44008110020@D000<10000Q`000`40000G00030@00058000020003
0@00008000@100010P001040004200040@000@8000@100010P001040004400030@0008P00P4H0081
DP0000<000D100000@0200040@000@8000@100010P001040004200040@000@8000<100000P020H/0
00<100005`020E00000300050@0000400P001040004200040@000@8000@100010P00104000420003
0@00008000<10000R`020AT00P5>000010001040004200040@000@8000@100010P00104000420004
0@000@8000<100000P000`40002=00030@0001P00P5<000000<00@400P43008110020@@00P440081
10020@D000<10000SP020AX000<10000B@0002L00P6A00030@0001P00P5900009`000`40002A0003
0@0001T00P5700009`000`40002B00816`000`40001400009`000`40002D00030@0001T00P540000
9`020IH000<100006P020D80000W00030@0009H00P4L0081@00002L000<10000V0000`40000K0003
0@0003d0000W00030@0009T000<100006`020Cd0000W0081V`020Ad00P4k00009`000`40002L0003
0@0001`000<10000>00002L000<10000W@000`40000L0081>00002L000<10000WP020Ah00P4f0000
9`020J4000<100007@000`40000c00009`000`40002Q00030@0001d00P4c00000P000`40000200<1
10020@@00P44008110020@D000<10000XP020Al00P4a00000P000`40000500040@000@8000@10001
0P001040004200040@000@8000<100000P000`40002T00030@0001h000<10000;P0000<000<10000
10001040004200040@000@8000@100010P001040004200030@0000800P6V00030@0001h00P4^0000
0`00104000030@<000@100010P001040004200040@000@8000@1000110000`40002V008180000`40
000[000010001040004500040@000@8000@100010P001040004200040@000@@000<10000Z0000`40
000N0081:`000003004100810P040@<00P44008110020@@00P4500030@000:T000<100007`020BT0
000W0081Z`020B4000<100009P0002L000<10000[0000`40000O00819P0002L000<10000[@000`40
000P00030@0002<0000W00030@000:h00P4Q00818`0002L00P6a00030@0002000P4Q00009`000`40
002a00030@00024000<100007P0002L000<10000/P020B800P4N00009`000`40002d00030@000240
00<100006`0002L00P6f00030@0002400P4K00009`000`40002f00818`000`40000H00009`000`40
002h00030@0002400P4H00009`000`40002i00030@00028000<100005@0002L00P6k00818`020AD0
000W00030@000;`000<100008P000`40000B00000P020@@00P44008110020@@00P4400811@000`40
002m00030@0002800P4B000000D00@000@0200040@000@8000@100010P001040004200040@000@80
00@1000110000`40002n008190000`40000?000000D00@000@0200040@000@8000@100010P001040
004200040@000@8000@1000110020L4000<100008P020@l0000200810`001040004200040@000@80
00@100010P001040004200040@000@@000<10000`@000`40000S00030@0000`000001@0100010080
00@100010P001040004200040@000@8000@100010P001040004400030@000<8000<100008`020@`0
0002008110020@@00P44008110020@@00P4500030@000<<00P4U00030@0000T0000W0081aP000`40
000S00812@0002L000<10000aP000`40000T00030@0000H0000W00030@000<L000<1000090000`40
000500009`000`4000380081;00002L00P7;00030@0002T0000W00030@000</000<10000:00002L0
00<10000c0020BP0000W00030@000<h000<100009@0002L00P7@00030@0002@0000W00030@000=00
00<100008`0002L000<10000d@020B<0000W00030@000=<000<10000800002L00P7E00030@0001l0
000W00030@000=D000<100007P0000800P4300<110020@@00P44008110020@D000<10000eP020Ah0
00001@01000100D000@100010P001040004200040@000@8000@100010P000`40000200030@000=P0
00<100006`000005004000401@001040004200040@000@8000@100010P001040004200030@000080
0P7J00030@0001X0000200810`030@<000@100010P001040004200040@000@8000@1000110000`40
003J00030@0001T000001@010001008000<100000`001040004200040@000@8000@100010P001040
004400030@000=/000<10000600000800P4300@10`020@@00P44008110020@D000<10000g0020AP0
000W0081g`000`40000E00009`000`40003O00030@0001@0000W00030@000>0000<100004`0002L0
00<10000h@000`40000B00009`020N<00P4B00009`000`40003T00030@0000l0000W00030@000>D0
00<100003P0002L000<10000iP000`40000=00009`020NP000<10000300002L000<10000j0020@`0
000W00030@000>X000<100002@0002L000<10000j`000`40000800009`020Nd000<100001`0002L0
00<10000k@000`400006000000@00@4110020@@00P44008110020@@00P4500030@000>h000<10000
1@0000@000@100010P001040004200040@000@8000@100010P001040004200030@00008000<10000
mP0000800`4200040@000@8000@100010P001040004200040@000@8000@1000110020OL000001@01
0001008000@100010P001040004200040@000@8000@100010P001040004400030@000?H000001@01
0001008000@100010P001040004200040@000@8000@100010P001040004400030@000?H000020081
10020@@00P44008110020@@00P7n0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000
o`0Q0000o`0Q0000o`0Q0000o`0Q0000\
\>"],
  ImageRangeCache->{{{151, 438}, {292.938, 116}} -> {-7.86561, 781696, 
  0.04136, 2478.94}}]
}, Open  ]],

Cell["\<\
The doubling time in years increases significantlyly as you move \
from 5% (this is the sort of interest rate you get in a bank if your lucky) \
which results in doubling time of more than twelve years to 20% (the sort of \
return you can get from stock market, again, if you're lucky) which results \
in a doubling time of less than five years.  In looking at these numbers it's \
important to put them in human scale.  For example, if you're 25 and you put \
$10,000 in the bank at 5%, then at 50 years old you'll have around $35,000 \
with interest compounded.  If you got 20%, then you'd have close to \
$1,000,000 with interest compounded.  This is a big difference to someone \
who's thinking of retiring early.\
\>", "Text"],

Cell["\<\
DoublingTimeInYears[start_,rate_] :=
    Module[{amount = start, i = 1},
\tWhile[(amount += amount * rate) < 2 start, i += 1] ;
\tReturn[i]]\
\>", "Input"],

Cell[CellGroupData[{

Cell["\<\
ListPlot[Table[DoublingTimeInYears[100,x],
\t\t\t   {x,0.01,0.30,0.01}],
\t {PlotJoined -> True, 
\t  PlotLabel -> \"Doubling time in years for a rate of 1% to 20%.\"}] ;\
\>",
   "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.031746 0.0147151 0.0245252 [
[.18254 .00222 -3 -9 ]
[.18254 .00222 3 0 ]
[.34127 .00222 -6 -9 ]
[.34127 .00222 6 0 ]
[.5 .00222 -6 -9 ]
[.5 .00222 6 0 ]
[.65873 .00222 -6 -9 ]
[.65873 .00222 6 0 ]
[.81746 .00222 -6 -9 ]
[.81746 .00222 6 0 ]
[.97619 .00222 -6 -9 ]
[.97619 .00222 6 0 ]
[.01131 .13734 -6 -4.5 ]
[.01131 .13734 0 4.5 ]
[.01131 .25997 -12 -4.5 ]
[.01131 .25997 0 4.5 ]
[.01131 .38259 -12 -4.5 ]
[.01131 .38259 0 4.5 ]
[.01131 .50522 -12 -4.5 ]
[.01131 .50522 0 4.5 ]
[.5 .63053 -143 0 ]
[.5 .63053 143 14 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.18254 .01472 m
.18254 .02097 L
s
[(5)] .18254 .00222 0 1 Mshowa
.34127 .01472 m
.34127 .02097 L
s
[(10)] .34127 .00222 0 1 Mshowa
.5 .01472 m
.5 .02097 L
s
[(15)] .5 .00222 0 1 Mshowa
.65873 .01472 m
.65873 .02097 L
s
[(20)] .65873 .00222 0 1 Mshowa
.81746 .01472 m
.81746 .02097 L
s
[(25)] .81746 .00222 0 1 Mshowa
.97619 .01472 m
.97619 .02097 L
s
[(30)] .97619 .00222 0 1 Mshowa
.125 Mabswid
.05556 .01472 m
.05556 .01847 L
s
.0873 .01472 m
.0873 .01847 L
s
.11905 .01472 m
.11905 .01847 L
s
.15079 .01472 m
.15079 .01847 L
s
.21429 .01472 m
.21429 .01847 L
s
.24603 .01472 m
.24603 .01847 L
s
.27778 .01472 m
.27778 .01847 L
s
.30952 .01472 m
.30952 .01847 L
s
.37302 .01472 m
.37302 .01847 L
s
.40476 .01472 m
.40476 .01847 L
s
.43651 .01472 m
.43651 .01847 L
s
.46825 .01472 m
.46825 .01847 L
s
.53175 .01472 m
.53175 .01847 L
s
.56349 .01472 m
.56349 .01847 L
s
.59524 .01472 m
.59524 .01847 L
s
.62698 .01472 m
.62698 .01847 L
s
.69048 .01472 m
.69048 .01847 L
s
.72222 .01472 m
.72222 .01847 L
s
.75397 .01472 m
.75397 .01847 L
s
.78571 .01472 m
.78571 .01847 L
s
.84921 .01472 m
.84921 .01847 L
s
.88095 .01472 m
.88095 .01847 L
s
.9127 .01472 m
.9127 .01847 L
s
.94444 .01472 m
.94444 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .13734 m
.03006 .13734 L
s
[(5)] .01131 .13734 1 0 Mshowa
.02381 .25997 m
.03006 .25997 L
s
[(10)] .01131 .25997 1 0 Mshowa
.02381 .38259 m
.03006 .38259 L
s
[(15)] .01131 .38259 1 0 Mshowa
.02381 .50522 m
.03006 .50522 L
s
[(20)] .01131 .50522 1 0 Mshowa
.125 Mabswid
.02381 .03924 m
.02756 .03924 L
s
.02381 .06377 m
.02756 .06377 L
s
.02381 .08829 m
.02756 .08829 L
s
.02381 .11282 m
.02756 .11282 L
s
.02381 .16187 m
.02756 .16187 L
s
.02381 .18639 m
.02756 .18639 L
s
.02381 .21092 m
.02756 .21092 L
s
.02381 .23544 m
.02756 .23544 L
s
.02381 .28449 m
.02756 .28449 L
s
.02381 .30902 m
.02756 .30902 L
s
.02381 .33354 m
.02756 .33354 L
s
.02381 .35807 m
.02756 .35807 L
s
.02381 .40712 m
.02756 .40712 L
s
.02381 .43164 m
.02756 .43164 L
s
.02381 .45617 m
.02756 .45617 L
s
.02381 .48069 m
.02756 .48069 L
s
.02381 .52974 m
.02756 .52974 L
s
.02381 .55427 m
.02756 .55427 L
s
.02381 .57879 m
.02756 .57879 L
s
.02381 .60332 m
.02756 .60332 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
gsave
.5 .63053 -204 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 22 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Doubling) show
117.000000 15.000000 moveto
(time) show
147.000000 15.000000 moveto
(in) show
165.000000 15.000000 moveto
(years) show
201.000000 15.000000 moveto
(for) show
225.000000 15.000000 moveto
(a) show
237.000000 15.000000 moveto
(rate) show
267.000000 15.000000 moveto
(of) show
285.000000 15.000000 moveto
(1) show
291.000000 15.000000 moveto
(%) show
303.000000 15.000000 moveto
(to) show
321.000000 15.000000 moveto
(20) show
333.000000 15.000000 moveto
(%) show
339.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(.) show
345.000000 15.000000 moveto
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.11746 .61803 m
.11905 .60332 L
.15079 .45617 L
.18254 .38259 L
.21429 .30902 L
.24603 .28449 L
.27778 .25997 L
.30952 .23544 L
.34127 .21092 L
.37302 .18639 L
.40476 .18639 L
.43651 .16187 L
.46825 .16187 L
.5 .13734 L
.53175 .13734 L
.56349 .13734 L
.59524 .13734 L
.62698 .11282 L
.65873 .11282 L
.69048 .11282 L
.72222 .11282 L
.75397 .11282 L
.78571 .11282 L
.81746 .11282 L
.84921 .08829 L
.88095 .08829 L
.9127 .08829 L
.94444 .08829 L
.97619 .08829 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/@P10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdk34g_ojn_l004J2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_P00oiT0IP3o0;C^]:JV
YZf][Jo^kP20P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;0000?`00O`00_`0o000o
?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0
_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo?cooOcoo
_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no?gnoOgno
_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo
_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0_ol0oolo0?lo?oloOolo_olooomo
0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_mWEc:0Pl=WIfJj^[WMgM`0:0>o<knl1
k`1^06EUI@00I@00[C8bI@1U002]039U<VD00:d006Db<RPX:000:00X02P002PX:00X02P00<`cVL`c
c<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co
<lcoI/coVLcoc<coool0<ol0I_l0VOl0c?l0oolc0?lc<olcI_lcVOlcc?lcoomV0?mVI_mVVOmVoonI
0?nI<onII_nIVOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<ool08@000?l08@0003l00`4Q00<110020Ah0
0`4300<17P040@<00P4N00@10P030Ah00`4400816000048000<100007`000`40000200040@000Ah0
00<100001@000`40000L00030@00008000@100017P000`40000500030@0001h000@100010P000`40
000E0000@P000`40000O00030@00008000@100017P000`40000500030@0001d000D100000@020003
0@0001d000<1000010000`40000N00040@000@8000<100005@0003l00`4R00030@00008000@10001
7P000`40000200<18@001040004200030@0001h000<100000`4O00810`001040004G0000?`000`40
000R00030@00008000@100017P000`40000200030@0001h000@100010P001040004M00040@000@80
00<100008@001040004200030@0001D0000o00@180020@D00P4N008110040Ah00P4400817`020@<0
104M00<110020AP0000J00030@000?l0100001X000<10000o`0400006P000`40003o00@0000J0003
0@000?l0100001D0mP4E00006P000`40000500030@0000D000<100001@000`40000400030@0000D0
00<100001@000`40000500030@0000D000<1000010000`40000500030@0000D000<100001@000`40
000400030@0000D000<100001@000`40000500030@0000D000<1000010000`40000500030@0000D0
00<100001@000`40000400030@0000D000<100001@000`40000500030@0000D000<1000010000`40
000500030@0000D000<100001@000`40000I00006P000`40003o00@0000J00030@000?l0100001X0
00<10000o`0400006P000`40003o00@0000J0081o`0500006P000`40003o00@0000J00030@000?l0
100001X000<10000o`0400006P000`40003o00@0000J00030@000?l0100001X00P7o00D0000J0003
0@000?l0100001X000<10000o`0400006P000`40003o00@0000J00030@000?l0100001X000<10000
o`0400006P020LP08@4K00006P000`40003600030@0003X0000J00030@000<D000<10000>`0001X0
00<10000``020Ch0000J00030@000<8000<10000?P0001X000<10000`@000`40000o00006P020I40
<@5200006P000`40002?00030@000740000J00030@0008h000<10000LP0001X000<10000S0020GD0
000B00<11@000`40002;00030@0007D0000E00030@00008000<10000RP000`40001f00005@000`40
000200<1L@0I0GT0000B00<11@000`40001`00030@000900000B00030@0000D000<10000K`000`40
002A00004P040@@000<10000K@020I@0000J00030@0006`000<10000U00001X000<10000J`000`40
002E00006P020F802P6H00006P000`40001P00030@000:00000J00030@0005l000<10000X@0001X0
00<10000GP000`40002R00006P000`40001M00030@000:<0000J00030@0005`000<10000Y00001X0
0P5C00X1Y`0001X000<10000D@000`40002_00006P000`40001@00030@000;00000J00030@0004h0
0P6c00006P000`40001=00030@000;<0000J00030@0004`000<10000]00001X00P5;0081]`0001X0
00<10000B@000`40002g00006P000`40001800030@000;P0000J00030@0004H00P6k00006P000`40
001500030@000;/0000J00030@0004@000<10000_00001X00P530081_`0001X000<10000@@000`40
002o00006P000`40001000030@000<00000J00030@0003l000<10000`@0000`00`4400811@000`40
000n00030@000<80000=00030@00008000@1000110000`40000m00030@000<<0000=00030@000080
00@1000110030C/00P7600003@000`40000200040@000@@000<10000>P000`40003600003@000`40
000200040@000@@000<10000>@000`400037000030020@D00P4500030@0003L00P7:00006P000`40
000f00030@000<X0000J00030@0003D000<10000b`0001X00P4d0081cP0001X000<10000<P000`40
003>00006P000`40000a00030@000<l0000J00030@0002l00P7B00006P000`40000^00030@000=80
000J00030@0002d000<10000d`0001X00P4]00030@000=@0000J00030@0002`000<10000e00001X0
00<10000:`000`40003E00006P000`40000[00030@000=D0000J00030@0002X000<10000eP0001X0
00<10000:P000`40003F00006P020B/000<10000eP0001X000<10000:@000`40003G00006P000`40
000Y00030@000=L0000J00030@0002P000<10000f00001X000<10000:0000`40003H00006P000`40
000W00030@000=T0000J0081:0000`40003I00006P000`40000V00030@000=X0000J00030@0002H0
00<10000fP0001X000<100009P000`40003J000030030@<00`4500030@0002D000<10000f`0000d0
00<100001@000`40000200030@0002D000<10000f`0000d000<100001@000`40000200<190000`40
003L00003@000`40000200<11@000`40000T00030@000=`0000=00030@00008000<100001@000`40
000S00030@000=d0000<008110040@@000<100008`000`40003M00006P000`40000R00030@000=h0
000J00030@00028000<10000gP0001X00P4S00030@000=h0000J00030@00024000<10000g`0001X0
00<100008@000`40003O00006P000`40000P00030@000>00000J00030@00020000<10000h00001X0
00<100007`000`40003Q00006P020B0000<10000h@0001X000<100007P000`40003R00006P000`40
000N00030@000>80000J00030@0001h000<10000hP0001X000<100007@000`40003S00006P000`40
000M00030@000><0000J00817@000`40003T00006P000`40000L00030@000>@0000J00030@0001`0
00<10000i00001X000<1000070000`40003T00006P000`40000K00030@000>D0000J00030@0001/0
00<10000i@0001X00P4L00030@000>D0000J00030@0001/000<10000i@0001X000<100006`000`40
003U00006P000`40000J00030@000>H0000<00@10`020@D000<100006P000`40003V00003@000`40
000200040@000@@000<100006P000`40003V00003P001@400001008000<100000P030AX000<10000
iP0000l000@100010P000`40000200030@0001X000<10000iP0000`000@100010P00104000440003
0@0001T000<10000i`0000d00P4400811@000`40000I00030@000>L0000J00030@0001T000<10000
i`0001X000<100006@000`40003W00006P020AX000<10000i`0001X000<1000060000`40003X0000
6P000`40000H00030@000>P0000J00030@0001P000<10000j00001X000<1000060000`40003X0000
6P000`40000H00030@000>P0000J008160000`40003Y00006P000`40000G00030@000>T0000J0003
0@0001L000<10000j@0001X000<100005`000`40003Y00006P000`40000G00030@000>T0000J0003
0@0001H000<10000jP0001X00P4G00030@000>X0000J00030@0001H000<10000jP0001X000<10000
5P000`40003Z00006P000`40000F00030@000>X0000J00030@0001D000<10000j`0001X000<10000
5@000`40003[00006P020AH000<10000j`0001X000<100005@000`40003[00006P000`40000D0003
0@000>`0000J00030@0001@000<10000k0000?l08@000?l08@000?l08@000?l08@0002h00`4g00<1
]@00034000<10000=@000`40002d00000P050@<00`4200D100<00@400P4200D100<00@400`401001
0@020@80104800<10P050@0600400@010P040@L01@4010010@020@P000<100000`040@03004100<1
00<00@400P4200@120040@<00`4200@120050@L0104200D10P030@<0104800<10P040@P00`4400<1
2@030@<00`4800@10`020@@00`4300811P0000<000L100000@0100<000<100400P00104000430005
0@0000401@001@400001008000@100010`000`40000600040@000@<000/100000@0100400@0=0005
0@0000400P000`40000700030@00008000<100000`001040004300030@0000D000<100001`000`40
000200050@0000400P000`40000800040@000@T000<100000P001040004300060@000@012`001@40
0001008000<100002@000`40000300030@0100T000H10001004300030@0000H000<100000P001040
004300030@0100<00P4600000`001`40000100400`000`400@0200040@000@<000D100000@050005
0@0000400P001040004300030@0000H000<1000010002P40000100400@050@T000D100000@020003
0@0000H000<100400`050@03004100810`000`40000200@12@000`40000200050@0000400P000`40
000800@12@000`40000200@10`000`40000200D11`001@400001008000<100002@000`40000300<1
2@000`40000200050@0000402@001@400001008000@100000`4;00000`001`40000100400`000`40
0@0200040@000@<000D100000@0500050@0000400P001040004300030@0000H000<1000010002`40
000100400@0100<000<100001`001@400001008000<100001P000`400@0300050@00004010001040
00020@<000<100002P000`40000200050@0000400P020@`000<100001`020@H000D100000@040005
0@0000401`001@400001008000<100002@000`40000200<12P000`40000200050@0000402P001040
004200030@0000<1300000<000D100000@0200<100@00@400P4200@110000`40000200810P050@<0
1@4600@10`020@801P4200<12@020@801@47008100<00@400`030@<00`42008100@00@40104800@1
0`030@800P400`010@0800<120020@03004100800`4200@10`030@T00`4200@12@000`4000020003
0@0100T0104300<120001040004200040@000@8000<10040300000801@4>00030@0000D000<10000
6P000`40001A00030@0002d000<100005@000`400008008110030@X000<100003`020@@00P4300<1
300001@00P4500<11@000`40000D00030@0000@000<100005@000`40000b00<1;@000`40000E00<1
6`000`40000Y0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000\
\>"],
  ImageRangeCache->{{{151, 438}, {683.5, 506.562}} -> {-22.8643, 82.2978, 
  0.128572, 0.166426}}]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Most people aren't very good at logic or statistics", "Subsection"],

Cell["\<\
At least when it comes to dealing very large (or very small) \
numbers (equivalently in considering exponential growth processes for \
problems involving anything more than just a few stages).  From a survival \
standpoint it was never very useful to have such accuracy - for the most part \
survival depended on reacting quickly and often enough the high-risk route \
was the most advantageous to the species.\
\>", "Text"],

Cell["\<\
Suppose that smoking does not cause cancer or lung disease but that \
one cigarette in every 18,250 packs is loaded with dynamite and is \
indistinguishable from other cigarettes.  The odds of dying are the same, but \
in one case you have a painful protracted death and in the other your head \
gets blown off for a quick, relatively painless death.\
\>", "Text"],

Cell["\<\
Gamblers fallacy - \"I've been playing this one-armed bandit all \
day it just has to be due for a win.\"  or \"I've been driving along this \
road at 80 MPH for the last 500 miles and I haven't had a fatal wreck yet.\"  \
It doesn't matter if you have flipped a fair coin 100 times and seen 100 \
heads in a row (this is unlikely but not impossible), the probability of the \
next one being heads is still 1/2; the coin, the roulette wheel, or the \
one-armed bandit has no memory of past.\
\>", "Text"],

Cell["\<\
The evidence that people are not very good at solving logic and \
probability word games or that they repeatedly fall into the same traps in \
gambling situations does not imply that humans don't use logic or \
probability.  Word problems - especially in probability, logic and statistics \
- are particularly tricky and solving them quickly is not likely to have been \
a skill that would serve us particularly well during most of the time humans \
have been on earth.  There are lots of cases in which humans and other \
biological organisms perform impressive feats of calculation in which they at \
least perform in accord with the laws of logic and probability.  For example, \
there is ample evidence that infants use statistics to learn language.  It's \
as though they compile frequencies to learn the acoustic cues that signal \
word boundaries.  Certainly the machinery in cells for producing proteins is \
governed by the laws of logic (or behaves as if it is).  The fact that \
biologists use terms like repressor, promoter, negative regulation, etc.  \
rather than conjunction, implication, negation, etc.  is just a matter of \
terminolgy.  [Run the Genetics movie.]\
\>", "Text"],

Cell["\<\
Question: Anybody ever play with \"Creatures\" a game in which you \
get to \"breed\" creatures called \"Norns\" and raise them.  The game has a \
very complicated ecology and biology.  There are bacterial agents (biogens \
and antigens), norns have a complicated physiology and an immune system, \
there are genes (and even a genetic editor if you want tinker directly with \
the genetic sequence), and there are even drives and brain chemicals.  [Check \
out norn.jpg and grendel.jpg.]  Makes you wonder if silicon life is that far \
off - anybody know about John Conway's game of life and John Von Neumann's \
self replicating automata?  [Run the Cellular Automata demo.]\
\>", "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["On the problem of measuring things and interpreting data", "Subsection"],

Cell[TextData[{
  "In general people do an abysmal job in drawing conclusions from summary \
data.  A classic example concerns qualitative decisions based on averages.  \
Consider comparing norns and alts using average test scores.  Suppose that \
the means are ordered to that the mean for alts is two points less than the \
mean for norns.  (Assume that the distibution of test scores for each group \
are approximately normal with a standard deviation of five points.)  Should \
we conclude that norns are better than alts in taking this exam?  Well, \
people will conclude whatever they want but the fact is that 42% of the alts \
got better scores than 42% of the norns.  Doesn't it always seem like it's \
the norns who did poorly on the test that go around making unqualified \
statements like \"norns are better than alts\"?  ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " allows us to illustrate this simple analysis graphically."
}], "Text"],

Cell["\<\
As an example of a type of inappropriate conclusion that irks most \
students, consider assigning of letter grades based on numerical test scores. \
 All measurements have some error associated with them.  Does it make sense \
to give one person an A because she gets a 90 on a test and another person a \
B because he gets an 89 if the diagnostic error in the test is 2 or \
greater?\
\>", "Text"],

Cell["\<\
Even if we get the numbers right it's often very difficult to \
assign values to outcomes.  In most case there are many factors that can \
support or conflict with one another.  For example, lots of factors go into \
deciding where to locate an airport: cost of construction, capacity, access \
time from nearby population centers, safety, social disruption building it, \
noise polution caused by airport operations.  Do you think that the factors \
governing an organism's utility are any less complex or any easier to compare \
than those at work in the airport location problem?  You should check out the \
articles on breeding Norns on the Creatures webpages \
http://www.creatures.co.uk/.\
\>", "Text"],

Cell[CellGroupData[{

Cell["Demonstration of the problems of reasoning from averages", 
  "Subsubsection"],

Cell["<<Statistics`ContinuousDistributions`", "Input"],

Cell["\<\
We model the distribution of the norn (alt) test scores with a \
normal distribution (also called a Gaussian distribution) with a mean of 90 \
(88) and a standard deviation of 5.  This model is an approximation for a \
number of reasons but one aspect in particular may strike you as strange; a \
normal distribution has infinite tails and so in our case there is some \
finite probablity (albeit small) that a norn got a grade of 150 (or -150).  \
In the following, we use the probability density functions (PDF) for normal \
distributions to illustrate the spread of the test scores.\
\>", "Text"],

Cell["NornPDF[x_] := PDF[NormalDistribution[90,5],x]", "Input"],

Cell["AltPDF[x_] := PDF[NormalDistribution[88,5],x]", "Input"],

Cell["\<\
The PDF allows us to answer questions like \"what percentage of \
norns got a score between 90 and 95?\"  We'd obtain the answer to this \
question by simply integrating under the curve over that interval.\
\>", 
  "Text"],

Cell["x =. ; (* Make sure that x has no value. *)", "Input"],

Cell[CellGroupData[{

Cell["N[Integrate[AltPDF[x],{x,90,95}]]", "Input"],

Cell[BoxData[
    \(0.263821599155904706`\)], "Output"]
}, Open  ]],

Cell["\<\
Here we compare the density functions for the test scores and you \
see the familiar bell curves.\
\>", "Text"],

Cell[CellGroupData[{

Cell["Plot[{NornPDF[x],AltPDF[x]},{x,80,100}] ;", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-3.78571 0.0476191 0.0147151 7.37706 [
[.2619 .00222 -6 -9 ]
[.2619 .00222 6 0 ]
[.5 .00222 -6 -9 ]
[.5 .00222 6 0 ]
[.7381 .00222 -6 -9 ]
[.7381 .00222 6 0 ]
[.97619 .00222 -9 -9 ]
[.97619 .00222 9 0 ]
[.01131 .16226 -24 -4.5 ]
[.01131 .16226 0 4.5 ]
[.01131 .3098 -24 -4.5 ]
[.01131 .3098 0 4.5 ]
[.01131 .45734 -24 -4.5 ]
[.01131 .45734 0 4.5 ]
[.01131 .60488 -24 -4.5 ]
[.01131 .60488 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.2619 .01472 m
.2619 .02097 L
s
[(85)] .2619 .00222 0 1 Mshowa
.5 .01472 m
.5 .02097 L
s
[(90)] .5 .00222 0 1 Mshowa
.7381 .01472 m
.7381 .02097 L
s
[(95)] .7381 .00222 0 1 Mshowa
.97619 .01472 m
.97619 .02097 L
s
[(100)] .97619 .00222 0 1 Mshowa
.125 Mabswid
.07143 .01472 m
.07143 .01847 L
s
.11905 .01472 m
.11905 .01847 L
s
.16667 .01472 m
.16667 .01847 L
s
.21429 .01472 m
.21429 .01847 L
s
.30952 .01472 m
.30952 .01847 L
s
.35714 .01472 m
.35714 .01847 L
s
.40476 .01472 m
.40476 .01847 L
s
.45238 .01472 m
.45238 .01847 L
s
.54762 .01472 m
.54762 .01847 L
s
.59524 .01472 m
.59524 .01847 L
s
.64286 .01472 m
.64286 .01847 L
s
.69048 .01472 m
.69048 .01847 L
s
.78571 .01472 m
.78571 .01847 L
s
.83333 .01472 m
.83333 .01847 L
s
.88095 .01472 m
.88095 .01847 L
s
.92857 .01472 m
.92857 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .16226 m
.03006 .16226 L
s
[(0.02)] .01131 .16226 1 0 Mshowa
.02381 .3098 m
.03006 .3098 L
s
[(0.04)] .01131 .3098 1 0 Mshowa
.02381 .45734 m
.03006 .45734 L
s
[(0.06)] .01131 .45734 1 0 Mshowa
.02381 .60488 m
.03006 .60488 L
s
[(0.08)] .01131 .60488 1 0 Mshowa
.125 Mabswid
.02381 .0516 m
.02756 .0516 L
s
.02381 .08849 m
.02756 .08849 L
s
.02381 .12537 m
.02756 .12537 L
s
.02381 .19914 m
.02756 .19914 L
s
.02381 .23603 m
.02756 .23603 L
s
.02381 .27291 m
.02756 .27291 L
s
.02381 .34668 m
.02756 .34668 L
s
.02381 .38357 m
.02756 .38357 L
s
.02381 .42045 m
.02756 .42045 L
s
.02381 .49422 m
.02756 .49422 L
s
.02381 .53111 m
.02756 .53111 L
s
.02381 .56799 m
.02756 .56799 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .09437 m
.06244 .12347 L
.10458 .16293 L
.14415 .20737 L
.18221 .25625 L
.22272 .31347 L
.26171 .37143 L
.30316 .43293 L
.34309 .48842 L
.3815 .53475 L
.40095 .55452 L
.42237 .57285 L
.44268 .58651 L
.45178 .59137 L
.46172 .59576 L
.4671 .59773 L
.4721 .59929 L
.47727 .60064 L
.48196 .60163 L
.48658 .60238 L
.4887 .60266 L
.49093 .60289 L
.49332 .60309 L
.49438 .60316 L
.49552 .60321 L
.49675 .60326 L
.49789 .6033 L
.49859 .60331 L
.49925 .60332 L
.50049 .60332 L
.50163 .60331 L
.50286 .60328 L
.50401 .60324 L
.50508 .60319 L
.50754 .60302 L
.51014 .60279 L
.51268 .60248 L
.51504 .60215 L
.5204 .60116 L
.5293 .59888 L
.53882 .59555 L
.54906 .59096 L
.56016 .58482 L
.58032 .57077 L
.62123 .53175 L
.66064 .48351 L
.69852 .43049 L
.73886 .37057 L
.77769 .31288 L
.81897 .25466 L
Mistroke
.85873 .20391 L
.89697 .16134 L
.93767 .12338 L
.97619 .09437 L
Mfstroke
.02381 .17837 m
.06244 .22411 L
.10458 .28058 L
.14415 .33806 L
.18221 .39499 L
.22272 .45413 L
.26171 .50612 L
.30316 .55209 L
.32392 .57035 L
.34309 .5839 L
.36173 .59378 L
.37114 .59748 L
.37645 .59917 L
.3815 .60052 L
.38635 .60156 L
.38891 .60202 L
.39163 .60243 L
.39416 .60274 L
.39644 .60296 L
.39755 .60305 L
.39874 .60313 L
.39979 .60319 L
.40091 .60324 L
.40221 .60329 L
.40343 .60331 L
.40417 .60332 L
.40488 .60332 L
.4062 .60331 L
.40687 .6033 L
.40759 .60328 L
.40887 .60323 L
.41026 .60316 L
.41178 .60306 L
.41427 .60285 L
.41694 .60255 L
.42179 .60182 L
.42674 .60082 L
.43136 .59966 L
.44183 .59623 L
.45203 .59183 L
.46293 .58601 L
.48427 .5714 L
.50442 .55395 L
.54289 .51215 L
.58381 .45835 L
.62321 .40111 L
.6611 .34442 L
.70144 .28553 L
.74026 .23282 L
.78154 .183 L
Mistroke
.8213 .14213 L
.85955 .10968 L
.90024 .08224 L
.93943 .06201 L
.97619 .04776 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/@P10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdk34g_ojn_l004J2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_P00oiT0IP3o0;C^]:JV
YZf][Jo^kP20P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;0000?`00O`00_`0o000o
?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0
_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo?cooOcoo
_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no?gnoOgno
_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo
_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0_ol0oolo0?lo?oloOolo_olooomo
0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_mWEc:0Pl=WIfJj^[WMgM`0:0>o<knl1
k`1^06EUI@00I@00[C8bI@1U002]039U<VD00:d006Db<RPX:000:00X02P002PX:00X02P00<`cVL`c
c<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co
<lcoI/coVLcoc<coool0<ol0I_l0VOl0c?l0oolc0?lc<olcI_lcVOlcc?lcoomV0?mVI_mVVOmVoonI
0?nI<onII_nIVOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<ool08@000?l08@000?l08@000?l08@0005D0
0P4300<1=P030@@00P4f00<10`030C<00`44008110020@<0001D00040@000@D000<10000=P001040
004200030@0003H000<100000`000`40000a00030@00008000@100010P00104000420000E0001040
004500030@0003@00`4200040@000CH00`4500030@00034000<100000P001040004200040@000@80
001E00810`030CH000@100010P001040004e00040@000@800`4d00030@00008000@100010P001040
00420000E0001040004200030@0003H000@100010P001040004e00040@000@8000<10000=0000`40
000200040@000@8000@100010P0005D00P4300@1=P020@@00P4g00810`040C800P45008110020@<0
000K00030@000?l00`0001/000<10000o`0300006`000`40003o00<0000K00030@000?l00`0001D0
o`490@<0000K00030@0000T000<100002P000`40000900030@0000X000<100002@000`40000:0003
0@0000X000<100002@000`40000:00030@0000T000<100002P000`40000900030@0000X000<10000
2@000`40000:00030@0000T000<100002P000`40000900030@0000X000<100002P000`4000070000
6`000`40000k00030@0003`000<10000?0000`40000l00030@0000L0000K00030@000?l00`0001/0
00<10000o`0300006`000`40003o00<0000K00030@000?l00`0001/000<10000o`0300006`000`40
003o00<0000K00030@000?L00P4900006`020OD00`4;00006`000`40003b00813P0001/000<10000
k`030A00000K00030@000>d00P4C00006`000`40003[00815@0001/000<10000j@020AL0000K0003
0@000>L00P4I00006`000`40003U00816`0001/000<10000h`020Ad0000K00030@000>8000<10000
7@0001/00P7Q0081800001/00P7P00030@0001D00P4900006`000`400@3M00815`000`4000090000
6`001040007J008160000`40000:00006`001@4000010=P000<100005P020@d0000K00030@000080
0P7D008160000`40000=00006`000`40000400030@000=0000<100005P020A00000K00030@0000D0
00<10000cP000`40000F00030@000100000K00030@0000H000<10000c0000`40000F00030@000140
000K008120020L/000<100005P000`40000B00006`000`40000900030@000<L000<100005P000`40
000C00006`000`40000:00030@000<D000<100005P000`40000D00006`000`40000;00030@000<<0
00<100005P000`40000E00006`000`40000<00030@000<4000<100005P000`40000F00006`000`40
000=00030@000;l000<100005P000`40000G00006`000`40000>00030@000;d000<100005P000`40
000H00006`000`40000?00030@000;/000<100005P000`40000I00000P020@@00P4400810`040@@0
00<1000040000`40002i00030@0001H000<100006P000005004000400`020@<000@100010`000`40
000400030@00014000<10000]`000`40000F00030@0001/000001@01000100P000@1000110000`40
000300814`000`40002e00030@0001H000<10000700000050040004020001040004500030@000080
00<100004P000`40002d00030@0001L000<10000700000050040004020001040004200040@000@@0
00<100004`000`40002b00030@0001L000<100007@0000800P4:008110020@D000<1000050000`40
002`00030@0001L000<100007P0001/000<100005@000`40002^00030@0001L000<100007`0001/0
00<100005P000`40002/00030@0001L000<10000800001/00P4H00030@000:X000<100005`000`40
000Q00006`000`400@0G00030@000:X000<100005P000`40000R00006`001040004G00030@000:P0
00<100005P000`40000S00006`001@40000101L000<10000YP000`40000F00030@0002@0000K0081
0`000`40000E00030@000:@000<100005P000`40000U00006`000`40000200030@0001H000<10000
XP000`40000F00030@0002H0000K00030@0000<000<100005P000`40002P00030@0001L000<10000
9P0001/000<1000010000`40000E00030@0009l000<100005`000`40000W00006`000`4000050003
0@0001D000<10000WP000`40000F00030@0002P0000K00030@0000H000<100005@000`40002L0003
0@0001H000<10000:@0001/000<100001`000`40000E00030@0009X000<100005P000`40000Z0000
6`000`40000700030@0001D000<10000V@000`40000F00030@0002/0000K00030@0000P000<10000
5@000`40002G00030@0001H000<10000;00001/000<100002@000`40000E00030@0009D000<10000
5`000`40000/00006`020@/000<100005@000`40002D00030@0001H000<10000;@0001/000<10000
2P000`40000F00030@00098000<100005P000`40000^00006`000`40000;00030@0001D000<10000
T@000`40000F00030@0002l0000K00030@0000`000<100005@000`40002?00030@0001H000<10000
<00001/000<100003@000`40000E00030@0008h000<100005@000`40000a00006`000`40000=0003
0@0001H000<10000S0000`40000F00030@000340000K00030@0000h000<100005@000`40002;0003
0@0001H000<10000<P0001/000<100003`000`40000E00030@0008X000<100005@000`40000c0000
6`000`40000?00030@0001H000<10000R0000`40000F00030@0003<0000K00814@000`40000F0003
0@0008H000<100005P000`40000d00006`000`40000A00030@0001D000<10000Q@000`40000F0003
0@0003D0000K00030@00018000<100005@000`40002400030@0001H000<10000=@0001/000<10000
4P000`40000F00030@00088000<100005P000`40000f00006`000`40000C00030@0001H000<10000
P0000`40000F00030@0003L0000K00030@0001@000<100005@000`40002000030@0001D000<10000
>00001/000<1000050000`40000F00030@0007h000<100005P000`40000h00006`000`40000E0003
0@0001H000<10000O0000`40000F00030@0003T00002008110020@@00P4400<110000`40000F0003
0@0001D000<10000O0000`40000E00030@0003X000001@01000100<00P4300040@000@@000<10000
0`000`40000F00030@0001H000<10000NP000`40000F00030@0003X000001@01000100P000@10001
0P040@@00P4H00030@0001H000<10000N0000`40000F00030@0003/000001@01000100P000@10001
0P000`400@0500030@0001P000<100005P000`40001g00030@0001D000<10000?000000500400040
20001040004200030@0100D000<1000060000`40000F00030@0007H000<100005P000`40000l0000
0P020@X00P4400811@000`40000I00030@0001H000<10000M0000`40000F00030@0003d0000K0003
0@0001X000<100005P000`40001c00030@0001D000<10000?P0001/000<100006P000`40000F0003
0@00078000<100005P000`40000n00006`000`40000K00030@0001H000<10000L0000`40000F0003
0@0003l0000K00030@0001`000<100005P000`40001_00030@0001D000<10000@00001/000<10000
70000`40000F00030@0006h000<100005P000`40001000006`000`40000M00030@0001H000<10000
K0000`40000F00030@000440000K00817`000`40000F00030@0006/000<100005@000`4000120000
6`000`40000N00030@0001H000<10000JP000`40000F00030@000480000K00030@0001l000<10000
5P000`40001X00030@0001H000<10000@`0001/000<1000080000`40000F00030@0006L000<10000
5@000`40001400006`000`40000P00030@0001H000<10000IP000`40000F00030@0004@0000K0003
0@00024000<100005P000`40001T00030@0001H000<10000A@0001/000<100008P000`40000F0003
0@0006<000<100005@000`40001600006`000`40000R00030@0001H000<10000HP000`40000F0003
0@0004H0000K00030@0002<000<100005P000`40001P00030@0001H000<10000A`0001/00P4U0003
0@0001H000<10000G`000`40000E00030@0004P0000K00030@0002@000<100005P000`40001N0003
0@0001H000<10000B00001/000<100009@000`40000F00030@0005`000<100005P000`4000190000
6`000`40000V00030@0001H000<10000F`000`40000E00030@0004X0000K00030@0002H000<10000
5P000`40001J00030@0001H000<10000BP0001/000<100009`000`40000F00030@0005P000<10000
5P000`40001;00006`000`40000X00030@0001H000<10000E`000`40000E00030@0004`0000K0003
0@0002P000<100005`000`40001E00030@0001D000<10000C@0001/000<10000:@000`40000F0003
0@0005@000<100005P000`40001=00006`000`40000Z00030@0001H000<10000DP000`40000F0003
0@0004h0000K0081:`000`40000G00030@00054000<100005@000`40001?00006`000`40000[0003
0@0001H000<10000D0000`40000F00030@0004l0000K00030@0002`000<100005P000`40001>0003
0@0001H000<10000D00001/000<10000;@000`40000F00030@0004`000<100005P000`40001A0000
6`000`40000]00030@0001H000<10000C0000`40000F00030@000540000K00030@0002h000<10000
5P000`40001:00030@0001H000<10000DP0001/000<10000;`000`40000F00030@0004P000<10000
5P000`40001C00006`000`40000_00030@0001L000<10000A`000`40000E00030@0005@000020081
10020@@00P4400811@000`40000`00030@0001H000<10000AP000`40000F00030@0005@000001@01
000100<00P4300040@000@8000@1000110000`40000a00030@0001H000<10000A0000`40000F0003
0@0005D000001@01000100P000@100010P00104000440081<P000`40000G00030@00048000<10000
5P000`40001F000000D00@000@0800040@000@800`4500030@00038000<100005`000`4000110003
0@0001H000<10000EP0000050040004020001040004200030@0000D000<10000<`000`40000F0003
0@00040000<100005P000`40001G00000P020@X00P4400<110000`40000d00030@0001H000<10000
?P000`40000F00030@0005P0000K00030@0003@000<100005`000`40000l00030@0001H000<10000
F@0001/000<10000=@000`40000F00030@0003/000<100005`000`40001I00006`000`40000f0003
0@0001H000<10000>P000`40000F00030@0005X0000K00030@0003H000<100005`000`40000h0003
0@0001H000<10000F`0001/000<10000=`000`40000G00030@0003H000<100005`000`40001K0000
6`020CT000<100005P000`40000e00030@0001L000<10000G00001/000<10000>@000`40000F0003
0@0003@000<100005P000`40001M00006`000`40000i00030@0001L000<10000<P000`40000F0003
0@0005h0000K00030@0003X000<100005`000`40000`00030@0001H000<10000G`0001/000<10000
>`000`40000G00030@0002h000<100005P000`40001P00006`000`40000k00030@0001P000<10000
;0000`40000G00030@000600000K00030@0003`000<100005`000`40000/00030@0001H000<10000
H@0001/000<10000?@000`40000G00030@0002X000<100005P000`40001R00006`000`40000n0003
0@0001L000<10000:0000`40000F00030@0006<0000K00030@0003l000<100005`000`40000V0003
0@0001H000<10000I00001/00P5100030@0001L000<1000090000`40000F00030@0006D0000K0003
0@00044000<100005`000`40000R00030@0001H000<10000IP0001/000<10000@P000`40000G0003
0@00020000<100005P000`40001W00006`000`40001300030@0001L000<100007P000`40000F0003
0@0006P0000K00030@0004@000<100005`000`40000L00030@0001H000<10000J@0001/000<10000
A@000`40000G00030@0001X000<100005P000`40001Z00006`000`400016008160020AT000<10000
5P000`40001[00006`000`40001800030@0001L000<100005@000`40000F00030@0006`0000K0003
0@0004T000<100005`000`40000C00030@0001H000<10000K@0001/000<10000BP000`40000G0003
0@00014000<100005P000`40001^00006`020D`000<100005`000`40000?00030@0001H000<10000
K`0001/000<10000C0020AP00P4=00815`020G80000K00030@0004h000<100005`000`4000090003
0@0001H000<10000LP0001/000<10000C`020AP000<100001P020AL00P5e00006`000`40001A0081
5`000`40000400030@0001D00P5g00006`000`40001C00030@0001D00P4200815`000`40001g0000
6`000`40001D00815`020AL00P5j00006`000`40001F00814P030@800P4B00<1O00000800P440081
10020@@00P4500030@0005P0104<00811`040@`00P5o000000D00@000@0300810`00104000420004
0@000@@000<10000G00<0@d03061000000D00@000@0800040@000@8000@1000110020Ol010000005
0040004020001040004300811@000`40003o00<000001@01000100P000@100010P00104000440003
0@000?l00`0000800P4:008110020@D000<10000o`030000o`0Q0000o`0Q0000o`0Q0000\
\>"],
  ImageRangeCache->{{{151, 438}, {292.938, 116}} -> {65.7767, 0.0521716, 
  0.0797743, 0.000514945}}]
}, Open  ]],

Cell["\<\
Find the point w where the two curves intersect; according to our \
simple model, for any x > w more alts got x's than norns.\
\>", "Text"],

Cell[TextData[{
  "Off[Solve::ifun] (* Stop ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " from issuing a particular warning. *)"
}], "Input"],

Cell["w =. ; (* Make sure that w has no value. *)", "Input"],

Cell["sols = Solve[AltPDF[w] == NornPDF[w], w] ;", "Input"],

Cell["<<Graphics`FilledPlot`", "Input"],

Cell[CellGroupData[{

Cell["\<\
FilledPlot[{AltPDF[x],If[x > w,0.0,AltPDF[x]] /. \
sols[[1]]},{x,80,100}] ;\
\>", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0 1 0 0.618034 [
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% Start of sub-graphic
p
0 0 1 0.618034 MathSubStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-3.78571 0.0476191 0.0147151 7.37706 [
[.2619 .00222 -6 -9 ]
[.2619 .00222 6 0 ]
[.5 .00222 -6 -9 ]
[.5 .00222 6 0 ]
[.7381 .00222 -6 -9 ]
[.7381 .00222 6 0 ]
[.97619 .00222 -9 -9 ]
[.97619 .00222 9 0 ]
[.01131 .16226 -24 -4.5 ]
[.01131 .16226 0 4.5 ]
[.01131 .3098 -24 -4.5 ]
[.01131 .3098 0 4.5 ]
[.01131 .45734 -24 -4.5 ]
[.01131 .45734 0 4.5 ]
[.01131 .60488 -24 -4.5 ]
[.01131 .60488 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.2619 .01472 m
.2619 .02097 L
s
[(85)] .2619 .00222 0 1 Mshowa
.5 .01472 m
.5 .02097 L
s
[(90)] .5 .00222 0 1 Mshowa
.7381 .01472 m
.7381 .02097 L
s
[(95)] .7381 .00222 0 1 Mshowa
.97619 .01472 m
.97619 .02097 L
s
[(100)] .97619 .00222 0 1 Mshowa
.125 Mabswid
.07143 .01472 m
.07143 .01847 L
s
.11905 .01472 m
.11905 .01847 L
s
.16667 .01472 m
.16667 .01847 L
s
.21429 .01472 m
.21429 .01847 L
s
.30952 .01472 m
.30952 .01847 L
s
.35714 .01472 m
.35714 .01847 L
s
.40476 .01472 m
.40476 .01847 L
s
.45238 .01472 m
.45238 .01847 L
s
.54762 .01472 m
.54762 .01847 L
s
.59524 .01472 m
.59524 .01847 L
s
.64286 .01472 m
.64286 .01847 L
s
.69048 .01472 m
.69048 .01847 L
s
.78571 .01472 m
.78571 .01847 L
s
.83333 .01472 m
.83333 .01847 L
s
.88095 .01472 m
.88095 .01847 L
s
.92857 .01472 m
.92857 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .16226 m
.03006 .16226 L
s
[(0.02)] .01131 .16226 1 0 Mshowa
.02381 .3098 m
.03006 .3098 L
s
[(0.04)] .01131 .3098 1 0 Mshowa
.02381 .45734 m
.03006 .45734 L
s
[(0.06)] .01131 .45734 1 0 Mshowa
.02381 .60488 m
.03006 .60488 L
s
[(0.08)] .01131 .60488 1 0 Mshowa
.125 Mabswid
.02381 .0516 m
.02756 .0516 L
s
.02381 .08849 m
.02756 .08849 L
s
.02381 .12537 m
.02756 .12537 L
s
.02381 .19914 m
.02756 .19914 L
s
.02381 .23603 m
.02756 .23603 L
s
.02381 .27291 m
.02756 .27291 L
s
.02381 .34668 m
.02756 .34668 L
s
.02381 .38357 m
.02756 .38357 L
s
.02381 .42045 m
.02756 .42045 L
s
.02381 .49422 m
.02756 .49422 L
s
.02381 .53111 m
.02756 .53111 L
s
.02381 .56799 m
.02756 .56799 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .17837 m
.06244 .22411 L
.10458 .28058 L
.14415 .33806 L
.18221 .39499 L
.22272 .45413 L
.26171 .50612 L
.30316 .55209 L
.32392 .57035 L
.34309 .5839 L
.36173 .59378 L
.37114 .59748 L
.37645 .59917 L
.3815 .60052 L
.38635 .60156 L
.38891 .60202 L
.39163 .60243 L
.39416 .60274 L
.39644 .60296 L
.39755 .60305 L
.39874 .60313 L
.39979 .60319 L
.40091 .60324 L
.40221 .60329 L
.40343 .60331 L
.40417 .60332 L
.40488 .60332 L
.4062 .60331 L
.40687 .6033 L
.40759 .60328 L
.40887 .60323 L
.41026 .60316 L
.41178 .60306 L
.41427 .60285 L
.41694 .60255 L
.42179 .60182 L
.42674 .60082 L
.43136 .59966 L
.44183 .59623 L
.45203 .59183 L
.46293 .58601 L
.48427 .5714 L
.50442 .55395 L
.54289 .51215 L
.58381 .45835 L
.62321 .40111 L
.6611 .34442 L
.70144 .28553 L
.74026 .23282 L
.78154 .183 L
Mistroke
.8213 .14213 L
.85955 .10968 L
.90024 .08224 L
.93943 .06201 L
.97619 .04776 L
Mfstroke
.02381 .17837 m
.06244 .22411 L
.10458 .28058 L
.14415 .33806 L
.18221 .39499 L
.22272 .45413 L
.26171 .50612 L
.30316 .55209 L
.32392 .57035 L
.34309 .5839 L
.36173 .59378 L
.37114 .59748 L
.37645 .59917 L
.3815 .60052 L
.38635 .60156 L
.38891 .60202 L
.39163 .60243 L
.39416 .60274 L
.39644 .60296 L
.39755 .60305 L
.39874 .60313 L
.39979 .60319 L
.40091 .60324 L
.40221 .60329 L
.40343 .60331 L
.40417 .60332 L
.40488 .60332 L
.4062 .60331 L
.40687 .6033 L
.40759 .60328 L
.40887 .60323 L
.41026 .60316 L
.41178 .60306 L
.41427 .60285 L
.41694 .60255 L
.42179 .60182 L
.42674 .60082 L
.43136 .59966 L
.44183 .59623 L
.44707 .5941 L
.44965 .59295 L
.45078 .59243 L
.45203 .59183 L
.45272 .01472 L
.45336 .01472 L
.45477 .01472 L
.45774 .01472 L
.46293 .01472 L
.50175 .01472 L
.54303 .01472 L
Mistroke
.58279 .01472 L
.62103 .01472 L
.66173 .01472 L
.70091 .01472 L
.74255 .01472 L
.78267 .01472 L
.82127 .01472 L
.86232 .01472 L
.90186 .01472 L
.93989 .01472 L
.97619 .01472 L
Mfstroke
0 1 1 r
.02381 .17837 m
.06244 .22411 L
.10458 .28058 L
.14415 .33806 L
.18221 .39499 L
.22272 .45413 L
.26171 .50612 L
.30316 .55209 L
.32392 .57035 L
.34309 .5839 L
.36173 .59378 L
.37114 .59748 L
.37645 .59917 L
.3815 .60052 L
.38635 .60156 L
.38891 .60202 L
.39163 .60243 L
.39416 .60274 L
.39644 .60296 L
.39755 .60305 L
.39874 .60313 L
.39979 .60319 L
.40091 .60324 L
.40221 .60329 L
.40343 .60331 L
.40417 .60332 L
.40488 .60332 L
.4062 .60331 L
.40687 .6033 L
.40759 .60328 L
.40887 .60323 L
.41026 .60316 L
.41178 .60306 L
.41427 .60285 L
.41694 .60255 L
.42179 .60182 L
.42674 .60082 L
.43136 .59966 L
.44183 .59623 L
.45203 .59183 L
.46293 .58601 L
.48427 .5714 L
.50442 .55395 L
.54289 .51215 L
.58381 .45835 L
.62321 .40111 L
.6611 .34442 L
.70144 .28553 L
.74026 .23282 L
.78154 .183 L
.8213 .14213 L
.85955 .10968 L
.90024 .08224 L
.93943 .06201 L
.97619 .04776 L
.97619 .01472 L
.93989 .01472 L
.90186 .01472 L
.86232 .01472 L
.82127 .01472 L
.78267 .01472 L
.74255 .01472 L
.70091 .01472 L
.66173 .01472 L
.62103 .01472 L
.58279 .01472 L
.54303 .01472 L
.50175 .01472 L
.46293 .01472 L
.45774 .01472 L
.45477 .01472 L
.45336 .01472 L
.45272 .01472 L
.45203 .59183 L
.45078 .59243 L
.44965 .59295 L
.44707 .5941 L
.44183 .59623 L
.43136 .59966 L
.42674 .60082 L
.42179 .60182 L
.41694 .60255 L
.41427 .60285 L
.41178 .60306 L
.41026 .60316 L
.40887 .60323 L
.40759 .60328 L
.40687 .6033 L
.4062 .60331 L
.40488 .60332 L
.40417 .60332 L
.40343 .60331 L
.40221 .60329 L
.40091 .60324 L
.39979 .60319 L
.39874 .60313 L
.39755 .60305 L
.39644 .60296 L
.39416 .60274 L
.39163 .60243 L
.38891 .60202 L
.38635 .60156 L
.3815 .60052 L
.37645 .59917 L
.37114 .59748 L
.36173 .59378 L
.34309 .5839 L
.32392 .57035 L
.30316 .55209 L
.26171 .50612 L
.22272 .45413 L
.18221 .39499 L
.14415 .33806 L
.10458 .28058 L
.06244 .22411 L
.02381 .17837 L
F
0 g
.02381 .17837 m
.06244 .22411 L
.10458 .28058 L
.14415 .33806 L
.18221 .39499 L
.22272 .45413 L
.26171 .50612 L
.30316 .55209 L
.32392 .57035 L
.34309 .5839 L
.36173 .59378 L
.37114 .59748 L
.37645 .59917 L
.3815 .60052 L
.38635 .60156 L
.38891 .60202 L
.39163 .60243 L
.39416 .60274 L
.39644 .60296 L
.39755 .60305 L
.39874 .60313 L
.39979 .60319 L
.40091 .60324 L
.40221 .60329 L
.40343 .60331 L
.40417 .60332 L
.40488 .60332 L
.4062 .60331 L
.40687 .6033 L
.40759 .60328 L
.40887 .60323 L
.41026 .60316 L
.41178 .60306 L
.41427 .60285 L
.41694 .60255 L
.42179 .60182 L
.42674 .60082 L
.43136 .59966 L
.44183 .59623 L
.45203 .59183 L
.46293 .58601 L
.48427 .5714 L
.50442 .55395 L
.54289 .51215 L
.58381 .45835 L
.62321 .40111 L
.6611 .34442 L
.70144 .28553 L
.74026 .23282 L
.78154 .183 L
Mistroke
.8213 .14213 L
.85955 .10968 L
.90024 .08224 L
.93943 .06201 L
.97619 .04776 L
Mfstroke
.02381 .17837 m
.06244 .22411 L
.10458 .28058 L
.14415 .33806 L
.18221 .39499 L
.22272 .45413 L
.26171 .50612 L
.30316 .55209 L
.32392 .57035 L
.34309 .5839 L
.36173 .59378 L
.37114 .59748 L
.37645 .59917 L
.3815 .60052 L
.38635 .60156 L
.38891 .60202 L
.39163 .60243 L
.39416 .60274 L
.39644 .60296 L
.39755 .60305 L
.39874 .60313 L
.39979 .60319 L
.40091 .60324 L
.40221 .60329 L
.40343 .60331 L
.40417 .60332 L
.40488 .60332 L
.4062 .60331 L
.40687 .6033 L
.40759 .60328 L
.40887 .60323 L
.41026 .60316 L
.41178 .60306 L
.41427 .60285 L
.41694 .60255 L
.42179 .60182 L
.42674 .60082 L
.43136 .59966 L
.44183 .59623 L
.44707 .5941 L
.44965 .59295 L
.45078 .59243 L
.45203 .59183 L
.45272 .01472 L
.45336 .01472 L
.45477 .01472 L
.45774 .01472 L
.46293 .01472 L
.50175 .01472 L
.54303 .01472 L
Mistroke
.58279 .01472 L
.62103 .01472 L
.66173 .01472 L
.70091 .01472 L
.74255 .01472 L
.78267 .01472 L
.82127 .01472 L
.86232 .01472 L
.90186 .01472 L
.93989 .01472 L
.97619 .01472 L
Mfstroke
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0 0 1 0.618034 MathSubStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-3.78571 0.0476191 0.0147151 7.37706 [
[.2619 .00222 -6 -9 ]
[.2619 .00222 6 0 ]
[.5 .00222 -6 -9 ]
[.5 .00222 6 0 ]
[.7381 .00222 -6 -9 ]
[.7381 .00222 6 0 ]
[.97619 .00222 -9 -9 ]
[.97619 .00222 9 0 ]
[.01131 .16226 -24 -4.5 ]
[.01131 .16226 0 4.5 ]
[.01131 .3098 -24 -4.5 ]
[.01131 .3098 0 4.5 ]
[.01131 .45734 -24 -4.5 ]
[.01131 .45734 0 4.5 ]
[.01131 .60488 -24 -4.5 ]
[.01131 .60488 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.2619 .01472 m
.2619 .02097 L
s
[(85)] .2619 .00222 0 1 Mshowa
.5 .01472 m
.5 .02097 L
s
[(90)] .5 .00222 0 1 Mshowa
.7381 .01472 m
.7381 .02097 L
s
[(95)] .7381 .00222 0 1 Mshowa
.97619 .01472 m
.97619 .02097 L
s
[(100)] .97619 .00222 0 1 Mshowa
.125 Mabswid
.07143 .01472 m
.07143 .01847 L
s
.11905 .01472 m
.11905 .01847 L
s
.16667 .01472 m
.16667 .01847 L
s
.21429 .01472 m
.21429 .01847 L
s
.30952 .01472 m
.30952 .01847 L
s
.35714 .01472 m
.35714 .01847 L
s
.40476 .01472 m
.40476 .01847 L
s
.45238 .01472 m
.45238 .01847 L
s
.54762 .01472 m
.54762 .01847 L
s
.59524 .01472 m
.59524 .01847 L
s
.64286 .01472 m
.64286 .01847 L
s
.69048 .01472 m
.69048 .01847 L
s
.78571 .01472 m
.78571 .01847 L
s
.83333 .01472 m
.83333 .01847 L
s
.88095 .01472 m
.88095 .01847 L
s
.92857 .01472 m
.92857 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .16226 m
.03006 .16226 L
s
[(0.02)] .01131 .16226 1 0 Mshowa
.02381 .3098 m
.03006 .3098 L
s
[(0.04)] .01131 .3098 1 0 Mshowa
.02381 .45734 m
.03006 .45734 L
s
[(0.06)] .01131 .45734 1 0 Mshowa
.02381 .60488 m
.03006 .60488 L
s
[(0.08)] .01131 .60488 1 0 Mshowa
.125 Mabswid
.02381 .0516 m
.02756 .0516 L
s
.02381 .08849 m
.02756 .08849 L
s
.02381 .12537 m
.02756 .12537 L
s
.02381 .19914 m
.02756 .19914 L
s
.02381 .23603 m
.02756 .23603 L
s
.02381 .27291 m
.02756 .27291 L
s
.02381 .34668 m
.02756 .34668 L
s
.02381 .38357 m
.02756 .38357 L
s
.02381 .42045 m
.02756 .42045 L
s
.02381 .49422 m
.02756 .49422 L
s
.02381 .53111 m
.02756 .53111 L
s
.02381 .56799 m
.02756 .56799 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
MathSubEnd
P
% End of sub-graphic
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/@P10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdk34g_ojn_l004J2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_P00oiT0IP3o0;C^]:JV
YZf][Jo^kP20P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;0000?`00O`00_`0o000o
?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0
_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo?cooOcoo
_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no?gnoOgno
_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo
_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0_ol0oolo0?lo?oloOolo_olooomo
0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_mWEc:0Pl=WIfJj^[WMgM`0:0>o<knl1
k`1^06EUI@00I@00[C8bI@1U002]039U<VD00:d006Db<RPX:000:00X02P002PX:00X02P00<`cVL`c
c<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co
<lcoI/coVLcoc<coool0<ol0I_l0VOl0c?l0oolc0?lc<olcI_lcVOlcc?lcoomV0?mVI_mVVOmVoonI
0?nI<onII_nIVOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<ool08@000?l08@000?l08@000?l08@0005H0
0P4300<1=@030@@00P4f00<10`030C800`44008110020@@0001E00040@000@D000<10000=@001040
004200030@0003H000<100000`000`40000`00030@00008000@100010P00104000430000E@001040
004500030@0003<00`4200040@000CH00`4500030@00030000<100000P001040004200040@000@<0
001F00810`030CD000@100010P001040004e00040@000@800`4c00030@00008000@100010P001040
00430000E@001040004200030@0003D000@100010P001040004e00040@000@8000<10000<`000`40
000200040@000@8000@100010`0005H00P4300@1=@020@@00P4g00810`040C400P45008110020@@0
003o0240000K00030@000?l00`0001/000<10000o`0300006`000`40003o00<0000E0?l120440000
6`000`40000:00030@0000T000<100002P000`40000900030@0000X000<100002@000`40000:0003
0@0000T000<100002P000`4W9`099`030BLW00XW00<19bL02BL00`4W9`0:9`030BLW00TW00<19bL0
2RL00`4W9`099`030BLW00XW00<19bL02BL00`4W9`0:9`030BLW00TW00<10000200001/000<10000
?0000`40000_00030BLW00TW00<19bL0?2L00`4W9`0k9`030@0000P0000K00030@0006h000<19bL0
QRL;00006`000`40001^00030BLW08HW2`0001/000<10000KP000`4W9`269`/0000K00030@0006h0
00<19bL0QRL;00006`000`40001^00030BLW08HW2`0001/000<10000KP000`4W9`269`/0000K0003
0@0006h000<19bL0QBL20@X0000K0081K`000`4W9`229`<1300001/000<10000KP000`4W9`209`81
3`0001/000<10000KP000`4W9`1m9`<14@0001/000<10000KP000`4W9`1k9`81500001/000<10000
KP000`4W9`1i9`815P0001/000<10000KP000`4W9`1g9`81600001/000<10000KP000`4W9`1e9`81
6P0001/000<10000KP000`4W9`1c9`81700001/000<10000KP000`4W9`1a9`817P0001/000<10000
KP000`4W9`1`9`030@0001h0000K0081K`000`4W9`1^9`818@0001/000<10000KP000`4W9`1]9`03
0@000240000K00030@0006h000<19bL0K2L00`40000R00006`000`40001^00030BLW06XW0P4U0000
6`000`40001^00030BLW06TW00<100009@0001/000<10000KP000`4W9`1W9`81:00001/000<10000
KP000`4W9`1V9`030@0002P0000K00030@0006h000<19bL0IBL00`40000Y00006`000`40001^0003
0BLW06@W00<10000:P0001/000<10000KP000`4W9`1S9`030@0002/0000K0081K`000`4W9`1Q9`81
;P0001/000<10000KP000`4W9`1P9`030@0002h0000K00030@0006h000<19bL0GbL00`40000_0000
6`000`40001^00030BLW05hW00<10000<00001/000<10000KP000`4W9`1M9`030@000340000K0003
0@0006h000<19bL0GBL00`40000a00006`000`40001^00030BLW05`W00<10000<P0000800P440081
10020@<0104400030@0006h000<19bL0FbL00`40000c000000D00@000@0300810`00104000430003
0@0000@000<10000KP000`4W9`1J9`030@0003@000001@01000100P000@1000110000`40000300<1
KP000`4W9`1I9`030@0003D000001@01000100P000@100011@000`40000200030@0006h000<19bL0
F2L00`40000f000000D00@000@0800040@000@8000@1000110000`40001^00030BLW05LW00<10000
=`0000800P4:008110020@D000<10000KP000`4W9`1F9`030@0003P0000K00030@0006h000<19bL0
EBL00`40000i00006`000`40001^00030BLW05@W00<10000>P0001/00P5_00030BLW05<W00<10000
>`0001/000<10040KP000`4W9`1C9`030@0003/0000K00040@000Fd000<19bL0DRL00`40000l0000
6`001@40000106`000<19bL0DBL00`40000m00006`020@<000<10000J@000`4W9`1@9`030@0003h0
000K00030@00008000<10000J@000`4W9`1?9`030@0003l0000K00030@0000<000<10000J0000`4W
9`1>9`030@000400000K00030@0000@000<10000I`000`4W9`1=9`030@000440000K00030@0000D0
00<10000IP000`4W9`1=9`030@000440000K00030@0000H000<10000I@000`4W9`1<9`030@000480
000K00030@0000L000<10000I0000`4W9`1;9`030@0004<0000K00030@0000L000<10000I0000`4W
9`1:9`030@0004@0000K00030@0000P000<10000H`000`4W9`199`030@0004D0000K00030@0000T0
00<10000HP000`4W9`189`030@0004H0000K00812`000`40001Q00030BLW04PW00<10000AP0001/0
00<100002P000`40001Q00030BLW04LW00<10000A`0001/000<100002`000`40001P00030BLW04HW
00<10000B00001/000<1000030000`40001O00030BLW04DW00<10000B@0001/000<100003@000`40
001N00030BLW04DW00<10000B@0001/000<100003@000`40001N00030BLW04@W00<10000BP0001/0
00<100003P000`40001M00030BLW04<W00<10000B`0001/000<100003`000`40001L00030BLW04<W
00<10000B`0001/000<100003`000`40001L00030BLW048W00<10000C00001/00P4A00030@0005/0
00<19bL0@BL00`40001=00006`000`40000A00030@0005X000<19bL0@2L00`40001>00006`000`40
000B00030@0005T000<19bL0@2L00`40001>00006`000`40000B00030@0005T000<19bL0?bL00`40
001?00006`000`40000C00030@0005P000<19bL0?RL00`40001@00006`000`40000D00030@0005L0
00<19bL0?RL00`40001@00006`000`40000E00030@0005H000<19bL0?BL00`40001A00006`000`40
000E00030@0005H000<19bL0?2L00`40001B00000P020@@00P44008110030@@000<100005P000`40
001E00030BLW03/W00<10000D`000005004000400`020@<000@1000110000`40000300030@0001L0
00<10000E0000`4W9`0k9`030@0005<000001@01000100P000@100010P040@@00`4H00030@0005<0
00<19bL0>RL00`40001D000000D00@000@0800040@000@8000<100401@000`40000H00030@0005<0
00<19bL0>BL00`40001E000000D00@000@0800040@000@8000<100401@000`40000I00030@000580
00<19bL0>2L00`40001F00000P020@X00P4400811@000`40000J00030@00054000<19bL0>2L00`40
001F00006`000`40000J00030@00054000<19bL0=bL00`40001G00006`000`40000K00030@000500
00<19bL0=RL00`40001H00006`000`40000L00030@0004l000<19bL0=RL00`40001H00006`000`40
000M00030@0004h000<19bL0=BL00`40001I00006`000`40000M00030@0004h000<19bL0=2L00`40
001J00006`000`40000N00030@0004d000<19bL0<bL00`40001K00006`020B0000<10000C0000`4W
9`0c9`030@0005/0000K00030@0001l000<10000C0000`4W9`0b9`030@0005`0000K00030@000200
00<10000B`000`4W9`0a9`030@0005d0000K00030@00024000<10000BP000`4W9`0a9`030@0005d0
000K00030@00024000<10000BP000`4W9`0`9`030@0005h0000K00030@00028000<10000B@000`4W
9`0_9`030@0005l0000K00030@0002<000<10000B0000`4W9`0_9`030@0005l0000K00030@0002<0
00<10000B0000`4W9`0^9`030@000600000K00030@0002@000<10000A`000`4W9`0]9`030@000640
000K00819P000`40001600030BLW02dW00<10000H@0001/000<100009@000`40001600030BLW02`W
00<10000HP0001/000<100009P000`40001500030BLW02/W00<10000H`0001/000<100009`000`40
001400030BLW02/W00<10000H`0001/000<100009`000`40001400030BLW02XW00<10000I00001/0
00<10000:0000`40001300030BLW02TW00<10000I@0001/000<10000:@000`40001200030BLW02TW
00<10000I@0001/000<10000:@000`40001200030BLW02PW00<10000IP0001/000<10000:P000`40
001100030BLW02LW00<10000I`0001/000<10000:P000`40001100030BLW02LW00<10000I`0001/0
0P4/00030@00040000<19bL09RL00`40001X00006`000`40000/00030@0003l000<19bL09BL00`40
001Y00006`000`40000/00030@0003l000<19bL09BL00`40001Y00006`000`40000]00030@0003h0
00<19bL092L00`40001Z00006`000`40000^00030@0003d000<19bL08bL00`40001[00006`000`40
000^00030@0003d000<19bL08bL00`40001[00006`000`40000_00030@0003`000<19bL08RL00`40
001/00006`000`40000_00030@0003`000<19bL08BL00`40001]00000P020@@00P44008110020@D0
00<10000<0000`40000k00030BLW024W00<10000K@000005004000400`020@<000@100010P001040
004400030@00034000<10000>P000`4W9`0P9`030@0006h000001@01000100P000@100010P001040
004400<1<@000`40000j00030BLW01lW00<10000K`0000050040004020001040004200<11@000`40
000b00030@0003T000<19bL07bL00`40001_000000D00@000@0800040@000@8000<100001@000`40
000c00030@0003P000<19bL07RL00`40001`00000P020@X00P4400<110000`40000d00030@0003L0
00<19bL07BL00`40001a00006`000`40000e00030@0003H000<19bL072L00`40001b00006`000`40
000f00030@0003D000<19bL06bL00`40001c00006`000`40000g00030@0003@000<19bL06bL00`40
001c00006`000`40000g00030@0003@000<19bL06RL00`40001d00006`000`40000h00030@0003<0
00<19bL06BL00`40001e00006`020CX000<10000<P000`4W9`0H9`030@0007H0000K00030@0003X0
00<10000<@000`4W9`0H9`030@0007H0000K00030@0003/000<10000<0000`4W9`0G9`030@0007L0
000K00030@0003`000<10000;`000`4W9`0F9`030@0007P0000K00030@0003`000<10000;`000`4W
9`0E9`030@0007T0000K00030@0003d000<10000;P000`4W9`0D9`030@0007X0000K00030@0003h0
00<10000;@000`4W9`0D9`030@0007X0000K00030@0003l000<10000;0000`4W9`0C9`030@0007/0
000K00030@0003l000<10000;0000`4W9`0B9`030@0007`0000K00030@00040000<10000:`000`4W
9`0A9`030@0007d0000K0081@P000`40000Z00030BLW010W00<10000OP0001/000<10000@P000`40
000Y00030BLW00lW00<10000O`0001/000<10000@`000`40000X00030BLW00hW00<10000P00001/0
00<10000@`000`40000X00030BLW00dW00<10000P@0001/000<10000A0000`40000W00030BLW00`W
00<10000PP0001/000<10000A@000`40000V00030BLW00/W00<10000P`0001/000<10000AP020BH0
00<19bL02RL00`40002400006`000`40001800030@0002<000<19bL02BL00`40002500006`000`40
001900818`000`4W9`089`030@0008H0000K00030@0004/000<1000080000`4W9`079`030@0008L0
000K0081C@020B0000<19bL01RL00`40002800006`000`40001>00030@0001d000<19bL012L20H/0
000K00030@0004l000<1000070000`4W9`029`81S@0001/000<10000D0000`40000K00030BLW0081
S`0001/000<10000D@020A/000<19`40T@0001/000<10000D`030AP00P6B00006`000`40001F0081
5@020I<00002008110020@@00P4400811@000`40001H008140030ID000001@01000100<00P430004
0@000@8000@1000110000`40001J00812`030IP000001@01000100P000@100010P001040004400<1
G00;0I/000001@01000100P000@100010`020@D000<10000o`03000000D00@000@0800040@000@80
00@1000110000`40003o00<0000200812P020@@00P4500030@000?l00`0001/000<10000o`030000
o`0Q0000o`0Q0000o`0Q0000\
\>"],
  ImageRangeCache->{{{151, 438}, {698.875, 521.938}} -> {-0.528678, 2.94985, 
  0.00349296, 0.00565173}, {{151.312, 437.625}, {698.875, 521.938}} -> {
  65.7139, 0.26181, 0.0799908, 0.000516342}, {{151.312, 437.625}, {698.875, 
  521.938}} -> {65.7139, 0.26181, 0.0799908, 0.000516342}}]
}, Open  ]],

Cell["\<\
In fact, all of the alts whose test scores are represented in the \
graph of the alt density to the right of w did better any norn whose test \
scores are represented in the norn density to left of w.\
\>", "Text"],

Cell[CellGroupData[{

Cell["N[Integrate[AltPDF[x],{x,w,Infinity}] /. sols[[1]] ]", "Input"],

Cell[BoxData[
    \(0.420740290560896923`\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["N[Integrate[NornPDF[x],{x,-Infinity,w}] /. sols[[1]] ]", "Input"],

Cell[BoxData[
    \(0.420740290560896923`\)], "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["On distinguishing signal from noise", "Subsection"],

Cell["\<\
The term \"signal\" refers to the useful part of the data, also \
called \"information\".  The term \"noise\" refers to the useless part of the \
data.  Is it possible that one person's signal is another person's noise?  \
Sure.  This happens all the time.  You're trying to learn the lead guitar \
from listening to CD a and someone else is trying to learning bass line.  \
You're trying to measure ocean wave height and someone else is trying to \
measure the degree of foaming action caused by wind.\
\>", "Text"],

Cell["\<\
As an example of how noise (disguised as information) can interfere \
with thinking, consider the following riddle: You're driving a bus with \
twenty passengers.  At the first stop, two passengers get off and one gets \
on.  At the second stop, four passengers get off.  At the third stop, three \
passengers get on and eight get off.  Question: What is the color of the \
driver's eyes?\
\>", "Text"],

Cell["\<\
Paradoxically in order to interpret data so as to extract a \
particular sort of signal you need to have a good model for the noise (i.e., \
everything other than the signal you're looking for).  Think about this.  It \
would seem that, in order to find what you're looking for in a pile of data, \
you need to know almost everything about the process that generated the data. \
 If you know this much about the data, then it wouldn't seem that you're \
likely to learn much from it.  From a statistician's (or scientist's) point \
of you this is in fact true.\
\>", "Text"],

Cell["\<\
As we'll see time and time again, the only way you can learn \
something new (short of just being lucky) is if you already know \"most of it\
\" to begin with and even then you have to be somewhat lucky (in the sense \
that it is possible to be dealt a hand (a random sample) that is unusual (not \
representative of the distribution that generated it) that will lead you to \
jump to the wrong conclusion).  Machine learning people like to say that \
there's no such thing as a free lunch.  We're lucky however to live in a \
universe in which the same basic and relatively simple patterns occur over \
and over and over again.  It's a pretty good working hypothesis therefore \
that any new phenomena that you have to contend with is likely to behave like \
phenomena that you've seen before.  If the universe was much less regular \
then inference would be difficult indeed.\
\>", "Text"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Miscellaneous ramblings about what's possible and what's not \
possible\
\>", "Section"],

Cell[CellGroupData[{

Cell["Can we build a successful survival machine?", "Subsection"],

Cell[TextData[{
  "Let's start with a model of a robot embedded in its environment.  In the \
simplest form or such a model, we have (graphically) a block diagram \
(technically the blocks are called transfer functions) with a transition or \
next-state function ",
  StyleBox["f(x(t),u(t)) = x(t+1) ",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "where ",
  StyleBox["x(t)",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and ",
  StyleBox["x(t+1)",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " are states and ",
  StyleBox["u(t)",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is an action, and a policy or control law ",
  StyleBox["g(x(t)) = u(t)",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ".  The agent (modeled by g) and the environment (modeled by ",
  StyleBox["f",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ") play tit-for-tat in an endless cycle of interactions.  We can make the \
model more complicated to account for the fact that the agent never knows the \
complete state ",
  StyleBox["x(t)",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " of the environment."
}], "Text"],

Cell["\<\
In the following, let's focus on building something that survives \
or performs a particular set of tasks.  Question: Can we build the brain of \
Rosie the robot from the parts available at your local Radio Shack?\"\
\>", 
  "Text"],

Cell["\<\
Question: What are the components of such a survival machine?  How \
about modules that build models from data, predict what's next, determine or \
estimate the current state, evaluate policies, and execute policies?\
\>", 
  "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Suppose we're given unlimited computational and informational \
resources \
\>", "Subsection"],

Cell["\<\
In particular, you given unlimited amounts of experience (data \
about how the environment), and unlimited computing power.  Notethat with \
enough data you can simulate and so accurately predict what you WILL perceive \
next; what more could you want prediction wise?  Let's suppose that you can \
rank order (assign an appropriate figure of merit or value) to any pair of \
situations that you can distinguish perceptually (this really is prerequisite \
to being able to tell the robot what it is you want it to do).  Suppose \
further that you can enumerate all policies that you can possibly execute \
using your physical capabilities.  Then with unlimited computing cycles you \
can maximize your expected value by performing an exhaustive generate and \
test.  That is to say you can build an optimal robot!\
\>", "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Well, are there limits on experience and computation? ", "Subsection"],

Cell["\<\
In particular, experience is costly - lots of organisms have died \
for their experience, and unfortunately, it's not possible tabula rasa to \
generalize to things that you haven't seen - no free lunch theorems.  \
Knowledge is the only way to offset lack of experience and knowledge is just \
compiled experience - knowledge has a price.  Knowledge is often equated with \
bias and with good reason - in nature, bias is the only thing standing \
between the organism and extinction; being creative is really just being \
either lucky or searching deeper/further than anyone else.\
\>", "Text"],

Cell["\<\
You'll often hear in this course about the curse of dimensionality; \
it's a fact of real environments that relates to exponential growth.  The \
realy world is blooming and buzzing with facets or factors that demand our \
attention.  When you combine these factors you get an exponential explosion \
in possibilities.  Again, knowledge is the only answer - clever data \
structures are needed to expedite inference to the point where it's useful \
for an organism embedded in real time.\
\>", "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["What does digital have to say about an analog world?", "Subsection"],

Cell["\<\
At some level our experience of the world resolves into bits.  The \
problem of representation concerns how we organize the bits.  Later we'll \
address this under the rubric of the pixels-to-predicates problem.  To build \
an agent that can cope with the real blooming and buzzing world, we need to \
build internal representations (think of them as efficient data structures if \
you like) to expedite inference.  We also need to somehow add the necessary \
background knowledge: the distilled expertise so that an agent can generalize \
to examples that it hasn't seen.\
\>", "Text"],

Cell["\<\
An expert system is a set of rules and relationships that encode an \
experts knowledge of a particular (usually narrow) domain and an algorithm \
for drawing inferences from this knowledge as applied to particular problem \
(instances).  The reason that expert systems are so useful is they make it \
easy to elicit, distil and then apply expert knowledge.  [Run the Netica \
movie.]\
\>", "Text"],

Cell["\<\
The issue of analog versus digital is raised as a reason why the \
sort of computer hardware that we have on our desks can never realize \
intelligent.  However, as I mentioned at the outset of this section, at some \
level our experience of the world resolves into bits.  At some level we're \
forced to interpreting quantized data at some resolution every inference is \
approximate.  As regards discrete versus continuous, computers are very good \
at repesenting continuous processes whether it be symbolically or \
numerically.\
\>", "Text"],

Cell["\<\
Note that both human and machine vision systems have limited \
resolution and limited frequency response.  The lesson is that sooner or \
later you have to get down and grovel in the bits.\
\>", "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["What constitutes a specification for an algorithm?", "Subsection"],

Cell["\<\
How about the following as a specification for an algorithm.  \
Suppose that we represent a robot navigation problem in terms of an objective \
function (a function to minimize or maximize) and a set of (linear) \
constraints expressed as inequalities, and then we solved the problem by \
using a mathematical software package designed to solving such problems.  \
Does this seem like an algorithm to you?  Let's consider a particular problem \
that can be solved in this fashion.  To keep the discussion simple we \
consider a toy planning problem that is easy to understand.\
\>", "Text"],

Cell[TextData[{
  "This general approach of representing a problem in a general form and then \
using general methods to solve it is typical of research in a number of \
fields including operations research, economics, statistics, machine vision, \
and machine learning.  In another variant, you are given data ",
  StyleBox["d",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and a set of models ",
  StyleBox["{m}",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and asked to search for a model m that maximizes",
  StyleBox[" Pr(d|m)",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " (the probability of the data given the model).  As still another example \
which generalizes on our linear programming example, you're given an \
objective function ",
  StyleBox["f",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", a set of possible solutions ",
  StyleBox["{x}",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", and asked to find some x minimizing ",
  StyleBox["f(x)",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ".  For general ",
  StyleBox["f",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", this problem can be quite hard (we can solve the special case of linear \
programs in polynomial time) and we'll be looking at search methods such as \
simulated annealing and simulated evolution for solving these problems."
}], "Text"],

Cell["\<\
We'll be learning general methods to search for solutions in very \
large state spaces, solve constrained minimization problems, and fit \
functions to minimize error or maximize likelihood.  Many of these algorithms \
will correspond to setting up a problem so that it is amenable to a given \
search or optimization method.  The hardest part in many AI problems is \
formulating the problem so that it can be tackled.  Once formulated AI \
researchers try to avail themselves of whatever good techniques are out there \
for solving the problem so formulated.  Once a problem is well posed it \
becomes largely a problem for algorithm design, combinatorial optimization, \
applied math, or statistics.\
\>", "Text"],

Cell[CellGroupData[{

Cell["Demonstration of linear programming", "Subsubsection"],

Cell["\<\
Suppose that you've decided to live on a diet of cola and chocolate \
cake.  You're also pressed for cash, so you want to minimize the money you \
spend on groceries.  You want to find quantities of cola and chocolate cake \
to buy so that you can minimize your grocery bill while consuming enough \
calories and caffeine to survive.  Here is how you might represent the \
problem as a linear program in Mathematica.\
\>", "Text"],

Cell["\<\
We want to determine how much cola and chocolate cake to buy.  We \
represent this amount as a list (vector) of length two.\
\>", "Text"],

Cell["x = {\"x[[1]]\",\"x[[2]]\"} ;", "Input"],

Cell["\<\
x[[1]] = quantity of cola
x[[2]] = quantity of chocolate cake\
\>", "Text"],

Cell["Next we specify the cost of cola and chocolate cake. ", "Text"],

Cell["c = {1.0, 2.0} ;", "Input"],

Cell["\<\
c[[1]] = cost of cola
c[[1]] = cost of chocolate cake\
\>", "Text"],

Cell["\<\
And the minimum calories and caffeine per day we need to survive. \
\
\>", "Text"],

Cell["b = {2000, 100} ;", "Input"],

Cell["\<\
b[[1]] = minimum calories per day
b[[1]] = minimum amount of caffeine per day\
\>", "Text"],

Cell["\<\
Finally, there are constants that relate calories and caffeine \
content to the amounts of cola and chocolate cake.  These constants \
(coefficients) are represented as a two-by-two matrix.\
\>", "Text"],

Cell["m = {{100, 500},{20, 5}} ;", "Input"],

Cell["\<\
m[[1,1]] = calories in cola
m[[1,2]] = calories in chocolate cake
m[[2,1]] = caffeine in cola
m[[2,2]] = caffeine in chocolate cake\
\>", "Text"],

Cell[TextData[{
  "In the language of linear algebra, we wish to find ",
  StyleBox["x",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " minimizing ",
  StyleBox["c.x",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " (where ",
  StyleBox["c.x",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "  denotes the inner (or dot) product of the vectors ",
  StyleBox["c",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and ",
  StyleBox["x",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ") subject to the constraint that ",
  StyleBox["m.x \[GreaterEqual] b",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " (where ",
  StyleBox["m.x",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is the result of multiplying the matrix m with the vector x so that ",
  StyleBox["m.x \[GreaterEqual] b",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " actually represents a system of linear (linear in the x[[i]]) \
inequalities).  Finding such an x can be done using linear programming \
methods, ",
  StyleBox["e.g.,",
    FontSlant->"Italic"],
  " simplex or interior point methods.  Mathematica makes solving such \
problems very simple."
}], "Text"],

Cell[CellGroupData[{

Cell["x = LinearProgramming[c,m,b]", "Input"],

Cell[BoxData[
    \({4.21052631578947345`, 3.1578947368421053`}\)], "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]]
},
FrontEndVersion->"X 3.0",
ScreenRectangle->{{0, 1280}, {0, 1024}},
WindowSize->{992, 681},
WindowMargins->{{97, Automatic}, {Automatic, 91}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"Magnification"->1},
Magnification->1.5,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1709, 49, 65, 0, 144, "Title"],
Cell[1777, 51, 211, 4, 65, "Text"],

Cell[CellGroupData[{
Cell[2013, 59, 53, 0, 81, "Section"],

Cell[CellGroupData[{
Cell[2091, 63, 72, 0, 70, "Subsection"],
Cell[2166, 65, 709, 11, 165, "Text"],
Cell[2878, 78, 996, 14, 215, "Text"],
Cell[3877, 94, 820, 12, 91, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[4734, 111, 79, 0, 56, "Subsection"],
Cell[4816, 113, 460, 7, 55, "Text"],
Cell[5279, 122, 365, 6, 43, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[5681, 133, 83, 1, 56, "Subsection"],
Cell[5767, 136, 170, 4, 31, "Text"],
Cell[5940, 142, 545, 9, 67, "Text"],
Cell[6488, 153, 1228, 17, 127, "Text"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[7765, 176, 96, 3, 72, "Section"],

Cell[CellGroupData[{
Cell[7886, 183, 97, 3, 56, "Subsection"],
Cell[7986, 188, 163, 4, 31, "Text"],
Cell[8152, 194, 171, 4, 31, "Text"],
Cell[8326, 200, 480, 8, 55, "Text"],
Cell[8809, 210, 786, 12, 79, "Text"],

Cell[CellGroupData[{
Cell[9620, 226, 69, 0, 36, "Subsubsection"],
Cell[9692, 228, 130, 3, 20, "Text"],

Cell[CellGroupData[{
Cell[9847, 235, 114, 4, 45, "Input"],
Cell[9964, 241, 13710, 428, 185, 5397, 321, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[23689, 672, 170, 4, 31, "Text"],

Cell[CellGroupData[{
Cell[23884, 680, 200, 5, 53, "Input"],
Cell[24087, 687, 2840, 61, 410, "Print"]
}, Open  ]],
Cell[26942, 751, 556, 9, 67, "Text"],
Cell[27501, 762, 129, 3, 20, "Text"],
Cell[27633, 767, 239, 7, 71, "Input"],

Cell[CellGroupData[{
Cell[27897, 778, 121, 3, 45, "Input"],
Cell[28021, 783, 10560, 326, 185, 3170, 230, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],

Cell[CellGroupData[{
Cell[38618, 1114, 121, 3, 45, "Input"],
Cell[38742, 1119, 11552, 360, 185, 3594, 257, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],

Cell[CellGroupData[{
Cell[50331, 1484, 30, 0, 40, "Input"],
Cell[50364, 1486, 12537, 391, 185, 3921, 280, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[62916, 1880, 741, 11, 79, "Text"],
Cell[63660, 1893, 165, 5, 62, "Input"],

Cell[CellGroupData[{
Cell[63850, 1902, 198, 6, 62, "Input"],
Cell[64051, 1910, 13517, 438, 185, 5204, 331, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[77629, 2355, 73, 0, 56, "Subsection"],
Cell[77705, 2357, 434, 7, 55, "Text"],
Cell[78142, 2366, 374, 6, 43, "Text"],
Cell[78519, 2374, 514, 8, 55, "Text"],
Cell[79036, 2384, 1204, 17, 127, "Text"],
Cell[80243, 2403, 698, 10, 79, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[80978, 2418, 78, 0, 56, "Subsection"],
Cell[81059, 2420, 963, 15, 103, "Text"],
Cell[82025, 2437, 408, 7, 55, "Text"],
Cell[82436, 2446, 718, 11, 79, "Text"],

Cell[CellGroupData[{
Cell[83179, 2461, 84, 1, 36, "Subsubsection"],
Cell[83266, 2464, 54, 0, 40, "Input"],
Cell[83323, 2466, 609, 9, 67, "Text"],
Cell[83935, 2477, 63, 0, 40, "Input"],
Cell[84001, 2479, 62, 0, 40, "Input"],
Cell[84066, 2481, 232, 5, 31, "Text"],
Cell[84301, 2488, 60, 0, 40, "Input"],

Cell[CellGroupData[{
Cell[84386, 2492, 50, 0, 40, "Input"],
Cell[84439, 2494, 55, 1, 36, "Output"]
}, Open  ]],
Cell[84509, 2498, 121, 3, 20, "Text"],

Cell[CellGroupData[{
Cell[84655, 2505, 58, 0, 40, "Input"],
Cell[84716, 2507, 15044, 430, 185, 4274, 293, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[99775, 2940, 149, 3, 20, "Text"],
Cell[99927, 2945, 154, 5, 40, "Input"],
Cell[100084, 2952, 60, 0, 40, "Input"],
Cell[100147, 2954, 59, 0, 40, "Input"],
Cell[100209, 2956, 39, 0, 40, "Input"],

Cell[CellGroupData[{
Cell[100273, 2960, 100, 3, 40, "Input"],
Cell[100376, 2965, 20541, 857, 185, 10767, 732, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[120932, 3825, 224, 4, 31, "Text"],

Cell[CellGroupData[{
Cell[121181, 3833, 69, 0, 40, "Input"],
Cell[121253, 3835, 55, 1, 36, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[121345, 3841, 71, 0, 40, "Input"],
Cell[121419, 3843, 55, 1, 36, "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[121535, 3851, 57, 0, 56, "Subsection"],
Cell[121595, 3853, 526, 8, 55, "Text"],
Cell[122124, 3863, 412, 7, 55, "Text"],
Cell[122539, 3872, 584, 9, 67, "Text"],
Cell[123126, 3883, 901, 13, 91, "Text"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[124076, 3902, 98, 3, 72, "Section"],

Cell[CellGroupData[{
Cell[124199, 3909, 65, 0, 56, "Subsection"],
Cell[124267, 3911, 1162, 35, 79, "Text"],
Cell[125432, 3948, 242, 5, 31, "Text"],
Cell[125677, 3955, 243, 5, 31, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[125957, 3965, 104, 3, 56, "Subsection"],
Cell[126064, 3970, 837, 12, 91, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[126938, 3987, 76, 0, 56, "Subsection"],
Cell[127017, 3989, 605, 9, 67, "Text"],
Cell[127625, 4000, 511, 8, 55, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[128173, 4013, 74, 0, 56, "Subsection"],
Cell[128250, 4015, 596, 9, 67, "Text"],
Cell[128849, 4026, 408, 7, 55, "Text"],
Cell[129260, 4035, 556, 9, 67, "Text"],
Cell[129819, 4046, 212, 4, 31, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[130068, 4055, 72, 0, 56, "Subsection"],
Cell[130143, 4057, 600, 9, 67, "Text"],
Cell[130746, 4068, 1400, 37, 103, "Text"],
Cell[132149, 4107, 726, 11, 79, "Text"],

Cell[CellGroupData[{
Cell[132900, 4122, 60, 0, 36, "Subsubsection"],
Cell[132963, 4124, 440, 7, 55, "Text"],
Cell[133406, 4133, 147, 3, 20, "Text"],
Cell[133556, 4138, 46, 0, 40, "Input"],
Cell[133605, 4140, 85, 3, 37, "Text"],
Cell[133693, 4145, 69, 0, 20, "Text"],
Cell[133765, 4147, 33, 0, 40, "Input"],
Cell[133801, 4149, 77, 3, 37, "Text"],
Cell[133881, 4154, 92, 3, 20, "Text"],
Cell[133976, 4159, 34, 0, 40, "Input"],
Cell[134013, 4161, 101, 3, 37, "Text"],
Cell[134117, 4166, 213, 4, 31, "Text"],
Cell[134333, 4172, 43, 0, 40, "Input"],
Cell[134379, 4174, 155, 5, 73, "Text"],
Cell[134537, 4181, 1198, 40, 67, "Text"],

Cell[CellGroupData[{
Cell[135760, 4225, 45, 0, 40, "Input"],
Cell[135808, 4227, 77, 1, 36, "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

