(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 3.0,
MathReader 3.0, or any compatible application. The data for the notebook 
starts with the line of stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    417370,      12456]*)
(*NotebookOutlinePosition[    418202,      12485]*)
(*  CellTagsIndexPosition[    418158,      12481]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Supplementary Notes on Machine Vision", "Title"],

Cell[CellGroupData[{

Cell["Detecting Motion in Image Sequences", "Section"],

Cell["\<\
I have a Connectix Corporation Color Quickcam in my box of old \
computer hardware.  These neat little spherically shaped cameras were used \
for all sorts of things (including hood ornaments for monitors) but I think \
that one of their main intended applications was for teleconferencing over \
the internet.  In any case, the only reason that I had one is that this \
particular camera was made for the Mac and one of my friend's gave it to me \
when he abandoned the Mac for the PC. I kept thinking that some weekend (just \
for fun) I'd write some code to implement a motion detector as a part of a \
burglar alarm system for my apartment.\
\>", "Text"],

Cell["\<\
Well, now it's New Year's eve (December 31, 1999) and I'm working \
on the lecture notes for CS141 and I need a little diversion.  Part of the \
first lecture on machine vision concerns motion detection and the calculation \
of optical flow fields.  One of my students, Ted Camus, did a very nice \
dissertation on fast algorithms for detecting motion (last I heard of Ted he \
was working on retinal scanning techniques for uniquely identifying people - \
you've probably seen this sort of thing in science-fiction films but it may \
soon be used in automated teller machines \"if we can just get this last bug \
worked out\").  I thought it might be nice to implement a very simple \
correlation-based optical flow algorithm of the sort described in the lecture \
notes.  Now, from the standpoint of implementing a serious motion detector, \
this really is frivolous - interpreted Mathematica code is just too slow to \
be of much use (a burglar could rob us blind, raid the refrigerator, watch a \
couple of shows on TV, and still have time to get ya good head start on the \
cops while Mathematica was still trying to determine if anything moved in the \
room).  Of course, the important thing is that now I can justify some \
recreational hacking as working on a demonstration to supplement my class \
notes.\
\>", "Text"],

Cell[TextData[{
  "This document, like most of the ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " notebooks that accompany my lecture notes, is ",
  StyleBox["not",
    FontSlant->"Italic"],
  " self contained.  You should refer to the lecture notes and to the \
textbook for more detail about optical flow and structure from motion."
}], "Text"],

Cell[TextData[{
  "It's easy to use the Quickcam to grab pictures; the problem is getting \
them into a form that Mathematica can use.  Most real vision sysems have \
framebuffers that can store multiple images, sample high resolution images to \
produce lower resolution ones, and even perform basic operations like \
differencing or convolution.  The Quickcam probably has lots of capabilities \
that I don't know about, but I just saved a series of five shots of my hand \
moving against the background of the back of my gray office chair.  The color \
images were saved in PICT format.  I opened each image in Photoshop, \
converted the image to grayscale, and then altered the image size so that it \
was 32 pixels by 32 pixels.  (I probably should have done some smoothing and \
filtering first, as the images were pretty noisy.)  Finally, I saved each \
image using the \"File/Save as...\"  command and the \"raw\" option. The \
following invocation tells ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " where to look for images and other auxiliary files."
}], "Text"],

Cell["SetDirectory[\"Power HD:Tom:work:lectures\"] ;", "Input"],

Cell["\<\
The trick was to get the images into Mathematica.  Conveniently, \
Mathematica has a utility for loading binary images byte by byte and \
converting the bytes to integers.  Bear with me for a moment while I struggle \
with the input-output side of things.\
\>", "Text"],

Cell["\<\
Each image is just a sequence of bytes representing grayscale \
levels.  The raw image tells us nothing about the format of the images; \
luckily we know how to interpret the images as a 32 X 32 array.\
\>", "Text"],

Cell["<<Utilities`BinaryFiles`", "Input"],

Cell["myImageSize = {32,32} ;", "Input"],

Cell["Array[imageArrayOne, myImageSize] ;", "Input"],

Cell["Array[imageArrayTwo, myImageSize] ;", "Input"],

Cell["\<\
ReadBinaryImages[one_String,two_String,{n_Integer,m_Integer}] := 
    Module[{imageStreamOne = OpenReadBinary[one],
            imageStreamTwo = OpenReadBinary[two]},
        Do[Do[imageArrayOne[i,j] = ReadBinary[imageStreamOne, Byte] ;
              imageArrayTwo[i,j] = ReadBinary[imageStreamTwo, Byte],
              {i,1,n},
           {j,1,m}]] ;
        Close[imageStreamOne] ; 
        Close[imageStreamTwo]]\
\>", "Input"],

Cell["\<\
Here's the invocation used to read the images into Mathematica \
(commented out here as I'm hoping that it won't be needed when you read \
this).\
\>", "Text"],

Cell["\<\
(* ReadBinaryImages[\"imageOne.raw\",\"imageTwo.raw\",myImageSize] \
; *)\
\>", "Input"],

Cell["\<\
Since I did this on a Mac and eventually I want to demonstrate this \
on another machine, I figured it would be a good idea to write the data to a \
file in a format that any version of Mathematica would likely be able to \
read.\
\>", "Text"],

Cell["\<\
WriteImageArrays[one_String,two_String,{n_Integer,m_Integer}] := 
    Module[{arrayStreamOne = OpenWrite[one],
            arrayStreamTwo = OpenWrite[two]},
        Do[Do[Write[arrayStreamOne,imageArrayOne[i,j]] ;
              Write[arrayStreamTwo,imageArrayTwo[i,j]], 
              {i,1,n},
           {j,1,m}]] ;
        Close[arrayStreamOne] ;
        Close[arrayStreamTwo]]\
\>", "Input"],

Cell["\<\
Here's the invocation used to write the arrays to files (again, \
commented out here as I'm hoping that it won't be needed when you read this).\
\
\>", "Text"],

Cell["\<\
(* \
WriteImageArrays[\"imageOne.array\",\"imageTwo.array\",myImageSize] ; *)\
\>",
   "Input"],

Cell["\<\
Now finally we can simply read the arrays directly from these \
files.\
\>", "Text"],

Cell["\<\
ReadImageArrays[one_String,two_String,{n_Integer,m_Integer}] := 
    Module[{arrayStreamOne = OpenRead[one],
            arrayStreamTwo = OpenRead[two]},
        Do[Do[imageArrayOne[i,j] = Read[arrayStreamOne] ;
              imageArrayTwo[i,j] = Read[arrayStreamTwo], 
              {i,1,n},
           {j,1,m}]] ;
        Close[arrayStreamOne] ;
        Close[arrayStreamTwo]]\
\>", "Input"],

Cell["\<\
ReadImageArrays[\"imageOne.array\",\"imageTwo.array\",myImageSize] \
;\
\>", "Input"],

Cell["\<\
Here's what the images look like; you can compare them with the \
grainy quicktime movie in \"imageSequence.movie\" which shows a five image \
sequence from the QuickCam in which the last two images correspond to the \
ones we sampled and read into Mathematica.\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\
DensityPlot[imageArrayOne[Round[i],
                          Round[j]],
            {i,1,myImageSize[[1]]},{j,1,myImageSize[[2]]},
            {Mesh -> False, PlotPoints -> myImageSize[[2]]}] ;\
\>", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% DensityGraphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.0117866 0.0310174 -0.0117866 0.0310174 [
[.1433 -0.0125 -3 -9 ]
[.1433 -0.0125 3 0 ]
[.29839 -0.0125 -6 -9 ]
[.29839 -0.0125 6 0 ]
[.45347 -0.0125 -6 -9 ]
[.45347 -0.0125 6 0 ]
[.60856 -0.0125 -6 -9 ]
[.60856 -0.0125 6 0 ]
[.76365 -0.0125 -6 -9 ]
[.76365 -0.0125 6 0 ]
[.91873 -0.0125 -6 -9 ]
[.91873 -0.0125 6 0 ]
[ 0 0 -0.125 0 ]
[-0.0125 .1433 -6 -4.5 ]
[-0.0125 .1433 0 4.5 ]
[-0.0125 .29839 -12 -4.5 ]
[-0.0125 .29839 0 4.5 ]
[-0.0125 .45347 -12 -4.5 ]
[-0.0125 .45347 0 4.5 ]
[-0.0125 .60856 -12 -4.5 ]
[-0.0125 .60856 0 4.5 ]
[-0.0125 .76365 -12 -4.5 ]
[-0.0125 .76365 0 4.5 ]
[-0.0125 .91873 -12 -4.5 ]
[-0.0125 .91873 0 4.5 ]
[ 0 0 -0.125 0 ]
[ 0 1 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.1433 0 m
.1433 .00625 L
s
[(5)] .1433 -0.0125 0 1 Mshowa
.29839 0 m
.29839 .00625 L
s
[(10)] .29839 -0.0125 0 1 Mshowa
.45347 0 m
.45347 .00625 L
s
[(15)] .45347 -0.0125 0 1 Mshowa
.60856 0 m
.60856 .00625 L
s
[(20)] .60856 -0.0125 0 1 Mshowa
.76365 0 m
.76365 .00625 L
s
[(25)] .76365 -0.0125 0 1 Mshowa
.91873 0 m
.91873 .00625 L
s
[(30)] .91873 -0.0125 0 1 Mshowa
.125 Mabswid
.17432 0 m
.17432 .00375 L
s
.20533 0 m
.20533 .00375 L
s
.23635 0 m
.23635 .00375 L
s
.26737 0 m
.26737 .00375 L
s
.3294 0 m
.3294 .00375 L
s
.36042 0 m
.36042 .00375 L
s
.39144 0 m
.39144 .00375 L
s
.42246 0 m
.42246 .00375 L
s
.48449 0 m
.48449 .00375 L
s
.51551 0 m
.51551 .00375 L
s
.54653 0 m
.54653 .00375 L
s
.57754 0 m
.57754 .00375 L
s
.63958 0 m
.63958 .00375 L
s
.6706 0 m
.6706 .00375 L
s
.70161 0 m
.70161 .00375 L
s
.73263 0 m
.73263 .00375 L
s
.79467 0 m
.79467 .00375 L
s
.82568 0 m
.82568 .00375 L
s
.8567 0 m
.8567 .00375 L
s
.88772 0 m
.88772 .00375 L
s
.11228 0 m
.11228 .00375 L
s
.08127 0 m
.08127 .00375 L
s
.05025 0 m
.05025 .00375 L
s
.01923 0 m
.01923 .00375 L
s
.94975 0 m
.94975 .00375 L
s
.98077 0 m
.98077 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
0 .1433 m
.00625 .1433 L
s
[(5)] -0.0125 .1433 1 0 Mshowa
0 .29839 m
.00625 .29839 L
s
[(10)] -0.0125 .29839 1 0 Mshowa
0 .45347 m
.00625 .45347 L
s
[(15)] -0.0125 .45347 1 0 Mshowa
0 .60856 m
.00625 .60856 L
s
[(20)] -0.0125 .60856 1 0 Mshowa
0 .76365 m
.00625 .76365 L
s
[(25)] -0.0125 .76365 1 0 Mshowa
0 .91873 m
.00625 .91873 L
s
[(30)] -0.0125 .91873 1 0 Mshowa
.125 Mabswid
0 .17432 m
.00375 .17432 L
s
0 .20533 m
.00375 .20533 L
s
0 .23635 m
.00375 .23635 L
s
0 .26737 m
.00375 .26737 L
s
0 .3294 m
.00375 .3294 L
s
0 .36042 m
.00375 .36042 L
s
0 .39144 m
.00375 .39144 L
s
0 .42246 m
.00375 .42246 L
s
0 .48449 m
.00375 .48449 L
s
0 .51551 m
.00375 .51551 L
s
0 .54653 m
.00375 .54653 L
s
0 .57754 m
.00375 .57754 L
s
0 .63958 m
.00375 .63958 L
s
0 .6706 m
.00375 .6706 L
s
0 .70161 m
.00375 .70161 L
s
0 .73263 m
.00375 .73263 L
s
0 .79467 m
.00375 .79467 L
s
0 .82568 m
.00375 .82568 L
s
0 .8567 m
.00375 .8567 L
s
0 .88772 m
.00375 .88772 L
s
0 .11228 m
.00375 .11228 L
s
0 .08127 m
.00375 .08127 L
s
0 .05025 m
.00375 .05025 L
s
0 .01923 m
.00375 .01923 L
s
0 .94975 m
.00375 .94975 L
s
0 .98077 m
.00375 .98077 L
s
.25 Mabswid
0 0 m
0 1 L
s
.1433 .99375 m
.1433 1 L
s
.29839 .99375 m
.29839 1 L
s
.45347 .99375 m
.45347 1 L
s
.60856 .99375 m
.60856 1 L
s
.76365 .99375 m
.76365 1 L
s
.91873 .99375 m
.91873 1 L
s
.125 Mabswid
.17432 .99625 m
.17432 1 L
s
.20533 .99625 m
.20533 1 L
s
.23635 .99625 m
.23635 1 L
s
.26737 .99625 m
.26737 1 L
s
.3294 .99625 m
.3294 1 L
s
.36042 .99625 m
.36042 1 L
s
.39144 .99625 m
.39144 1 L
s
.42246 .99625 m
.42246 1 L
s
.48449 .99625 m
.48449 1 L
s
.51551 .99625 m
.51551 1 L
s
.54653 .99625 m
.54653 1 L
s
.57754 .99625 m
.57754 1 L
s
.63958 .99625 m
.63958 1 L
s
.6706 .99625 m
.6706 1 L
s
.70161 .99625 m
.70161 1 L
s
.73263 .99625 m
.73263 1 L
s
.79467 .99625 m
.79467 1 L
s
.82568 .99625 m
.82568 1 L
s
.8567 .99625 m
.8567 1 L
s
.88772 .99625 m
.88772 1 L
s
.11228 .99625 m
.11228 1 L
s
.08127 .99625 m
.08127 1 L
s
.05025 .99625 m
.05025 1 L
s
.01923 .99625 m
.01923 1 L
s
.94975 .99625 m
.94975 1 L
s
.98077 .99625 m
.98077 1 L
s
.25 Mabswid
0 1 m
1 1 L
s
.99375 .1433 m
1 .1433 L
s
.99375 .29839 m
1 .29839 L
s
.99375 .45347 m
1 .45347 L
s
.99375 .60856 m
1 .60856 L
s
.99375 .76365 m
1 .76365 L
s
.99375 .91873 m
1 .91873 L
s
.125 Mabswid
.99625 .17432 m
1 .17432 L
s
.99625 .20533 m
1 .20533 L
s
.99625 .23635 m
1 .23635 L
s
.99625 .26737 m
1 .26737 L
s
.99625 .3294 m
1 .3294 L
s
.99625 .36042 m
1 .36042 L
s
.99625 .39144 m
1 .39144 L
s
.99625 .42246 m
1 .42246 L
s
.99625 .48449 m
1 .48449 L
s
.99625 .51551 m
1 .51551 L
s
.99625 .54653 m
1 .54653 L
s
.99625 .57754 m
1 .57754 L
s
.99625 .63958 m
1 .63958 L
s
.99625 .6706 m
1 .6706 L
s
.99625 .70161 m
1 .70161 L
s
.99625 .73263 m
1 .73263 L
s
.99625 .79467 m
1 .79467 L
s
.99625 .82568 m
1 .82568 L
s
.99625 .8567 m
1 .8567 L
s
.99625 .88772 m
1 .88772 L
s
.99625 .11228 m
1 .11228 L
s
.99625 .08127 m
1 .08127 L
s
.99625 .05025 m
1 .05025 L
s
.99625 .01923 m
1 .01923 L
s
.99625 .94975 m
1 .94975 L
s
.99625 .98077 m
1 .98077 L
s
.25 Mabswid
1 0 m
1 1 L
s
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of gray image
p
.01923 .01923 translate
.96154 .96154 scale
32 string
32 32 8 [32 0 0 32 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  Mimage
1827232331303134363A5473553E41424C3A2F3B3F2F2A23
2323191A1B141010
2023272334332F37393A4D5D58404B5C5A51504C49403B34
312A17131A141417
28222A28393C2C3D403E3F555C52626D6D6C66625B54524D
423C301A14161C1C
2325242D33402E3B443A4A5E6C757C7A7B7B78746F67615C
554F463725231A16
2122252C2E403A35485B6D7984868787808585817A74726B
645E54483B1A0F11
21232221242C33476173828D9089888D8988848079757169
69685B534D2D1917
1D212326262B496A76838C929090908A8A888070676D685A
60625E5554461E1E
2228282B2F4D707C87909293929B9C8EA9A1756988775F4F
535D5D534F4B2815
1D28292648747D88939696AF98919A8DC2C17879DAA35948
435059534F4A3416
1D2323326F84899099949BDBBA87938EC4CC8278DCB25C42
5183624E4F493C19
1D222850848D959A9B9A9AD3CA8E938EC9DF9371DCBC5640
99EB80494E484826
182E3C6F909198979A9C95BFE7A1999BD3E9AB79EAD85441
BFF07E43474F4F35
112A408293949A989F9E96B1F5BD9493C4F1CB87EBE65747
DCE65C3C45504E41
1E26458A9495999C9F9D9CA9EAE7B595A9F1EB85E0ED4E6A
F8D347384450544F
1B204D8F96989D9E9E9D9CA4C6F5E3B6A0DFF1C3E6F185AE
FFB3333742545A53
1725448F93949D9F9F9EA1A6A9E1F3D7CFE5F1F4F1F1EBE9
F38E273342545F57
383A468E96969B9994969DA6A8CBF4EFDEE5F1F0F0F0F2F2
D96C26324154645C
434C578A9A99A8BAB1A6918DA5B8EEEFDCCEEEF0F0F0F0F3
DA61273343515D5E
3B515E83999DA6BFD3EEE5A788A3DBF0E1BFC3EFEFF0F0F1
CF61283446545D5E
263A5A7B96999B9698ACE7EAB19FBEDADCE8BFCFF2F1F1ED
C863283848596063
483F5C7B8F9598989793ABEDEACDB9B1A1A4AEA5CFF1F2E9
AB5A273847595E60
85817F878E939599979996D4F9E5DDDAC2A49A98A7E2F5E7
984828394C565B5E
869B9794949090939A9A94A6D4F2F6F2E6C7AEA0A3CAF6E4
8C3427384C595D5F
768F9497968D8F93938F91948EAADBF5F5EBCEC0B4C2EED2
7F2A25344552565B
6E8B9698908E8787888C8B8B8B898EB9E8EBD8DFCAC8DDB5
7D26222D404D545D
6B83969A977E6881848C898D8A89878087C6EDF0DEC7C2D2
8621212A3C4E5456
6C7C959C9D83446B7F8181848686847E779CF2F3E7D0B9D6
9E22202335474B51
7277939DA395592B6577797C8183827F818AE2F4EAD5BDE1
B4261B1F2C424849
74758D9FA69F5F1D34676A767E7D807D7A7AC1F5EEDBC0E6
C338191E263A3F43
7A77879DA5A57E7E95AB8E64656E757B77728EDEEDE5BCD7
CA5B17171E2D3635
7E7A7F99A3A7AAB0AA9A8F936D585D69716E77C4DCE5BAC4
C2791F10121E2925
747B7D94A3AAA58739151583977F7059585A5FA9CDD6BDB8
B187350C0F141715
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000180P10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdoojn_l00;34gTJ2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_[C^]:JVYZf][Jo^kP00
oiT0IP3o0020P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;0000?`00O`00_`0o000o
?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0
_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo?cooOcoo
_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no?gnoOgno
_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo
_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0_ol0oolo0?lo?oloOolo_olooomo
0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_mWEc:0Pl=WIfJj^[WMgMiW<09W<<iW<
IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`c
c<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co
<lcoI/coVLcoc<coool0<ol0I_l0VOl0c?l0oolc0?lc<olcI_lcVOlcc?lcoomV0?mVI_mVVOmVoonI
0?nI<onII_nIVOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<ool08@000?l08@000?l08@000?l08@0003@0
0`4T00<110020B400`4300<18@040@<00P4Q00@10P030B400`440081500003L000<100008P000`40
000200040@000B4000<100001@000`40000O00030@00008000@100018@000`40000500030@000240
00@100010P000`40000A0000=`000`40000R00030@00008000@100018@000`40000500030@000200
00D100000@0200030@00020000<1000010000`40000Q00040@000@8000<100004@0003@00`4U0003
0@00008000@100018@000`40000200<190001040004200030@00024000<100000`4R00810`001040
004C0000=0000`40000U00030@00008000@100018@000`40000200030@00024000@100010P001040
004P00040@000@8000<1000090001040004200030@000140000d00@18`020@D00P4Q008110040B40
0P4400818P020@<0104P00<110020A@0003o0240003o0240003o0240003o0240000?0?l14@410000
3`000`40000200030@0000H000<100001@000`40000500030@0000H000<100001@000`4000060003
0@0000D000<100001@000`40000600030@0000D000<100001P000`40000500030@0000H000<10000
1@000`40000500030@0000H000<100001@000`40000600030@0000D000<100001P000`4000050003
0@0000D000<100001P000`40000500030@0000H000<100001@000`40000500030@0000H000<10000
1@000`40000600030@0000D000<100000P010@40000?00030@0002@000<100009`000`40000W0003
0@0002L000<100009`000`40000W00030@0001<00@4100003`000`40003o00d00@4100003`000`40
003o00d00@4100003`000`40003o00d00@4100003`020Ol03@020@40000?00030@000080021G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0C9G00QfEgIGMUMfE`Uf00IGMUMfEgHZE`070EL1E`5G
0@0AE`1J0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G1@010@40000?00030@0000802@40B5L1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0@YG00MfEgIGMUMf00QG00L1E`5G0EL101]G04P1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0ELa0@D00@410000
3`000`400002000PE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4bE`08MUMfEgIGMUL9MP06
EgIGMUMf:UL01`5G0EL1E`404EL0FP5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`D00@410000
3`000`40000200T104QG0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4:E`07MUMfEgIGMP08E`070EL1E`5G0@0KE`180EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G<@4500410@0000l000<100000P0085L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
<UL027IGMUMfEgIG2GH01UMfEgIGMRYG00L1E`5G0EL1015G05X1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL500410@0000l000<100000P090@18E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL12UL01gIGMUMfEgH025L01`5G
0EL1E`406eL0B05G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1Ec411@010@40000?00030@000080021G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0C9G00QfEgIGMUMfE`Uf00IGMUMfEgHZE`070EL1E`5G0@0AE`1J0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G1@010@40000?00810`090@18E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL12UL01gIG
MUMfEgH025L01`5G0EL1E`406eL0B05G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1Ec4110020@40000?00030@000080021G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0A5G00L1E`5G0EL1029G00mfEgIGMUMfEgIGMUMfEgH0
4EL08GIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP0bE`0h0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL500410@0000l000<100000P00D@5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0@0:E`0?MUMfEgIGMUMfEgIGMUMf015G00mfEgIGMUMfEgIGMUMfEgH0
:UL0805G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G<@4500410@0000l000<100000P0085L1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL14EL01`5G0EL1E`408UL03gIGMUMfEgIGMUMfEgIG
MP0AE`0QMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf039G03P1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`D00@4100003`000`400002001A
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL100YG00mfEgIGMUMfEgIGMUMfEgH04EL03gIGMUMfEgIGMUMfEgIG
MP0ZE`0P0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0ELa0@D00@4100003`000`400002000P
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4AE`070EL1E`5G0@0RE`0?MUMfEgIGMUMfEgIG
MUMf015G025fEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH0<UL0>05G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G1@010@40000?00030@000080
0541E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`402UL03gIGMUMfEgIGMUMfEgIGMP0AE`0?MUMfEgIGMUMfEgIG
MUMf02YG0201E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1Ec411@010@40000?00030@000080
021G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0A5G00L1E`5G0EL1029G00mfEgIGMUMfEgIG
MUMfEgH04EL08GIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP0bE`0h0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL500410@0000l000<10000
0P00D@5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@0:E`0?MUMfEgIGMUMfEgIGMUMf015G00mfEgIGMUMfEgIG
MUMfEgH0:UL0805G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G<@4500410@0000l00P43000P
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4AE`070EL1E`5G0@0RE`0PMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMULIMP0XEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMR9G02P1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G10020@40000?0003
0@0000800441E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0@09E`0?0EL1E`5G0EL1E`5G0EL100YG00MfEgIGMUMf00UG03MfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf029G0101E`5G0EL1E`5G
0EL1E`5G:@4500410@0000l000<100000P0085L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
4EL01`5G0EL1E`408UL087IGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG6GH0:5MfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgHRE`0X0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`D00@4100003`000`40000200110EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`402EL03`5G0EL1E`5G0EL1E`5G
0@0:E`07MUMfEgIGMP09E`0gMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMP0RE`0@0EL1E`5G0EL1E`5G0EL1EbT11@010@40000?00030@000080021G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0A5G00L1E`5G0EL1029G021fEgIGMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEaUf02QGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
8UL0:05G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL500410@0000l000<10000
0P00@@5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL100UG00l1E`5G0EL1E`5G0EL1E`402UL01gIGMUMfEgH02EL0=gIGMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH08UL0405G0EL1E`5G0EL1E`5G
0ELY0@D00@4100003`000`400002000PE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4AE`07
0EL1E`5G0@0RE`0PMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMULIMP0XEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMR9G02P1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G1@010@40000?00030@0000800441E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@09E`0?0EL1E`5G0EL1E`5G0EL100YG
00MfEgIGMUMf00UG03MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
MUMfEgIGMUMf029G0101E`5G0EL1E`5G0EL1E`5G:@4500410@0000l00P43000PE`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`4AE`070EL1E`5G0@0RE`08MUMfEgIGMUM9MP0PEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgHJE`0P0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL40081
0@0000l000<100000P00:@5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL100UG
00l1E`5G0EL1E`5G0EL1E`402EL01`5G0EL1E`402UL05WIGMUMfEgIGMUMfEgIGMUMfEgIGMUL[MP0X
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMQQG01P1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0ELA0@D00@4100003`000`400002000PE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`4AE`070EL1E`5G0@0RE`08MUMfEgIGMUM9MP0PEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgHJE`0P0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL500410@0000l000<100000P00:@5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL100UG00l1E`5G0EL1E`5G0EL1E`40
2EL01`5G0EL1E`402UL05WIGMUMfEgIGMUMfEgIGMUMfEgIGMUL[MP0XEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMQQG01P1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0ELA0@D00@410000
3`000`400002000PE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4AE`070EL1E`5G0@0RE`08
MUMfEgIGMUM9MP0PEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgHJE`0P0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL500410@0000l000<100000P00:@5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL100UG00l1E`5G0EL1E`5G0EL1E`402EL01`5G0EL1E`402UL05WIG
MUMfEgIGMUMfEgIGMUMfEgIGMUL[MP0XEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMQQG01P1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0ELA0@D00@4100001`030@D000<100000P0085L1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL14EL01`5G0EL1E`408UL027IGMUMfEgIGBGH085Mf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf6UL0805G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G1@010@40000:00030@00008000<100000P00:@5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL100UG00l1E`5G0EL1E`5G0EL1E`402EL01`5G0EL1E`402UL05WIGMUMfEgIG
MUMfEgIGMUMfEgIGMUL[MP0XEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMQQG
01P1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0ELA0@D00@4100002P000`40000200<10P0085L1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL14EL01`5G0EL1E`408UL027IGMUMfEgIGBGH085MfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf6UL0805G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
10020@40000700<11@000`400002000Y0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`402EL03`5G0EL1E`5G0EL1E`5G0@08E`0@MUMfEgIGMUMfEgIGMUMfEe=f02QGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf45L0205G0EL1E`5G6@4500410@0000L000<10000
1@000`400002000XE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0B9G00QfEgIG
MUMfEg=f01IGMUMfEgIGMUMfEgIGMUMfEgIGMUMf4UL0205G0EL1E`5G2@401eL1E`5G0EL01@010@40
000700@110000`400002000Y0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`40
2EL03`5G0EL1E`5G0EL1E`5G0@08E`0@MUMfEgIGMUMfEgIGMUMfEe=f02QGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf45L0205G0EL1E`5G6@4500410@0000l000<100000P00:5L1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4RE`08MUMfEgIGMUMcMP0FEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMQ9G00P1E`5G0EL1E`T100MG0EL1E`5G00D00@4100003`000`400002000Y
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`402EL03`5G0EL1E`5G0EL1E`5G
0@08E`0@MUMfEgIGMUMfEgIGMUMfEe=f02QGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMf45L0205G0EL1E`5G6@4500410@0000l000<100000P00:5L1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`4RE`08MUMfEgIGMUMcMP0FEgIGMUMfEgIGMUMfEgIGMUMfEgIGMQ9G
00P1E`5G0EL1E`T100MG0EL1E`5G00D00@4100003`000`400002000Y0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`402EL03`5G0EL1E`5G0EL1E`5G0@08E`0@MUMfEgIGMUMfEgIG
MUMfEe=f02QGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf45L0205G0EL1E`5G
6@4500410@0000l000<100000P00:5L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`4RE`08MUMfEgIGMUMcMP0FEgIGMUMfEgIGMUMfEgIGMUMfEgIGMQ9G00P1E`5G0EL1E`T100MG0EL1
E`5G00D00@4100003`020@<003T1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`4025L047IGMUMfEgIGMUMfEgIGMUMKMP0`EgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf4UL01P5G0EL1Ea4110020@40000?00030@000080
031G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4BE`06MUMfEgIG
4WH0>Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;012MP0>EgIGMUMfEgIGMUMfEgH:E`0H0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G1@010@40000?0003
0@00008003T1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`4025L047IGMUMfEgIGMUMfEgIGMUMKMP0`EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMf4UL01P5G0EL1Ea411@010@40000?00030@000080031G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4BE`06MUMfEgIG4WH0>Baf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;012MP0>EgIGMUMf
EgIGMUMfEgH:E`0H0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G1@010@40000?00030@00008003T1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4025L047IG
MUMfEgIGMUMfEgIGMUMKMP0`EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMf4UL01P5G0EL1Ea411@010@40000?00030@000080031G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4BE`06MUMfEgIG4WH0>Baf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;012MP0>EgIGMUMfEgIGMUMfEgH:E`0H
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G1@010@40000?00030@00008003T1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4025L047IGMUMfEgIGMUMfEgIG
MUMKMP0`EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf4UL01P5G
0EL1Ea411@010@40000?00030@000080031G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`4BE`06MUMfEgIG4WH0>Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;012MP0>EgIGMUMfEgIGMUMfEgH:E`0H0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G1@010@40000?00810`090@0XE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0@QG011fEgIGMUMfEgIGMUMfEgIGDgH0@5MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgHIEa0110020@40000?00030@000080
031G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`49EaUf044/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;00@MP07EgIGMUMfE`0CMP0`EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMf25L0405G0EL1E`5G0EL1E`5G0EL500410@0000l000<100000P090@0XE`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@QG011fEgIGMUMfEgIGMUMfEgIGDgH0@5MfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgHIEa01
1@010@40000?00030@000080031G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`49EaUf044/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;00@MP07EgIGMUMfE`0CMP0`EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf25L0405G0EL1E`5G0EL1E`5G0EL500410@0000l000<10000
0P090@0XE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@QG011fEgIGMUMfEgIG
MUMfEgIGDgH0@5MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgHIEa011@010@40000?00030@000080031G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`49EaUf044/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00@MP07EgIGMUMfE`0CMP0`EgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf25L0405G0EL1E`5G0EL1E`5G
0EL500410@0000l000<100000P090@0XE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0@QG011fEgIGMUMfEgIGMUMfEgIGDgH0@5MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgHIEa011@010@40000?00030@000080031G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`49EaUf044/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00@MP07
EgIGMUMfE`0CMP0`EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
25L0405G0EL1E`5G0EL1E`5G0EL500410@0000l000<100000P00:@5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL100QG00QfEgIGMUMfEc9f00l/MRaf;7H/MRaf;7H/MR`02GH03baf
;7H/MRaf;7H/MRaf;002MP0?EgIGMUMfEgIGMUMfEgIG00Uf011GMUMfEgIGMUMfEgIGMUMf4EL03gIG
MUMfEgIGMUMfEgIGMP0JE`060EL1E`5G2@4500410@0000l00P43000XE`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@UG6GH0>2af;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf2B`027H/MRaf;7H/4GH01baf;7H/MR`027H025MfEgIGMUMf
2EL07gIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH04UL0405G0EL1E`5G0EL1E`5G0EL40081
0@0000l000<100000P00:@5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL100QG
00QfEgIGMUMfEc9f00l/MRaf;7H/MRaf;7H/MR`02GH03baf;7H/MRaf;7H/MRaf;002MP0?EgIGMUMf
EgIGMUMfEgIG00Uf011GMUMfEgIGMUMfEgIGMUMf4EL03gIGMUMfEgIGMUMfEgIGMP0JE`060EL1E`5G
2@4500410@0000l000<100000P00:5L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`49EaUf03P/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MPT/00Qf;7H/MRaf;15f00L/MRaf;7H/00Qf00QGMUMfEgIGMPUG01mfEgIGMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMUMf019G0101E`5G0EL1E`5G0EL1E`5G1@010@40000?00030@00008002T1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@08E`08MUMfEgIGMULbMP0?;7H/MRaf
;7H/MRaf;7H/00Uf00l/MRaf;7H/MRaf;7H/MR`00WH03eMfEgIGMUMfEgIGMUMfE`09MP0@EgIGMUMf
EgIGMUMfEgIGMQ5G00mfEgIGMUMfEgIGMUMfEgH06UL01P5G0EL1E`T11@010@40000?00030@000080
02QG0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL12ELIMP0h;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H9;008MRaf;7H/MR`AMP07
;7H/MRaf;008MP08EgIGMUMfEgH9E`0OMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP0BE`0@
0EL1E`5G0EL1E`5G0EL1E`D00@4100003`000`400002000Y0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`4025L027IGMUMfEgIG<WH03baf;7H/MRaf;7H/MRaf;009MP0?;7H/MRaf
;7H/MRaf;7H/009f00mGMUMfEgIGMUMfEgIGMUL02GH045MfEgIGMUMfEgIGMUMfEgHAE`0?MUMfEgIG
MUMfEgIGMUMf01YG00H1E`5G0EL90@D00@4100003`000`400002000XE`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@UG6GH0>2af;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf2B`027H/MRaf;7H/4GH01baf;7H/MR`027H025MfEgIGMUMf
2EL07gIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH04UL0405G0EL1E`5G0EL1E`5G0EL50041
0@0000l000<100000P090@0HE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL125L027IGMUMfEgIG:WH03baf
;7H/MRaf;7H/MRaf;009MP07;7H/MRaf;009MQ0/47H01`0/02`0;0000R`01WH/MRaf;09f00QGMUMf
EgIGMQUG00MfEgIGMUMf01YG00H1E`5G0EL90@D00@4100003`020@<0021G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0@UG4GH082af;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf2B`07gH/
MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H00R`03`0/02`0;00/02`0;00/000@MP0H;00/02`0
;01f;7H/MRaf;5MfEgIGMUMf4EL05gIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf01YG00P1E`5G0EL1E`@0
0P4100003`000`40000200T101QG0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`48E`08MUMfEgIGMULZMP0?
;7H/MRaf;7H/MRaf;7H/00Uf00L/MRaf;7H/00Uf42`@MP0702`0;00/0002;006MRaf;7H/0WH025Mf
EgIGMUMf6EL01gIGMUMfEgH06UL01P5G0EL1E`T11@010@40000?00030@000080021G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@UG4GH082af;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
2B`07gH/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H00R`03`0/02`0;00/02`0;00/000@MP0H
;00/02`0;01f;7H/MRaf;5MfEgIGMUMf4EL05gIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf01YG00P1E`5G
0EL1E`D00@4100003`000`40000200T101QG0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`48E`08MUMfEgIG
MULZMP0?;7H/MRaf;7H/MRaf;7H/00Uf00L/MRaf;7H/00Uf42`@MP0702`0;00/0002;006MRaf;7H/
0WH025MfEgIGMUMf6EL01gIGMUMfEgH06UL01P5G0EL1E`T11@010@40000?00030@000080021G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@UG4GH082af;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf2B`07gH/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H00R`03`0/02`0;00/02`0;00/
000@MP0H;00/02`0;01f;7H/MRaf;5MfEgIGMUMf4EL05gIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf01YG
00P1E`5G0EL1E`D00@4100003`000`40000200T101QG0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`48E`08
MUMfEgIGMULZMP0?;7H/MRaf;7H/MRaf;7H/00Uf00L/MRaf;7H/00Uf42`@MP0702`0;00/0002;006
MRaf;7H/0WH025MfEgIGMUMf6EL01gIGMUMfEgH06UL01P5G0EL1E`T11@010@40000010010@440081
1@000`400002000PE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`49Ea5f020/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MPT/01mf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf008/
00l0;00/02`0;00/02`0;00047H062`0;00/02`0MRaf;7H/MRaGMUMfEgIGMQ5G01MfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMP0JE`080EL1E`5G0EL500410@00008000<100000P001040004400030@000080
2@4085L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0GIGMUMfEgIG67H01baf;7H/MR`02gH01Raf;7H/MP8/
00L0;00/02`000P/6GH@;11f00L0;00/02`000T/00UfEgIGMUMfEgH045L8MP08EgIGMUMfEgHJE`06
0EL1E`5G2@4500410@00008000<100000P001040004400<10P0065L1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0@UG4GH0;baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H0;00/02`000T/
27H042af;7H/MRaf;7H/MRaf;7H2;00?02`0;00/02`0;00/02`0011f00P/02`0;00/00P/00QGMUMf
EgIGMPUG00QfEgIGMUMfE`Uf00IGMUMfEgHRE`080EL1E`5G0EL400810@00008000<100000P001040
004400030@0000802@4085L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0GIGMUMfEgIG67H01baf;7H/MR`0
2gH01Raf;7H/MP8/00L0;00/02`000P/6GH@;11f00L0;00/02`000T/00UfEgIGMUMfEgH045L8MP08
EgIGMUMfEgHJE`060EL1E`5G2@4500410@00008000<100000P001040004400030@00008001QG0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`49Ea5f02l/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf02`0;00/0009;0Qf010/MRaf;7H/MRaf;7H/MRaf0R`03`0/02`0;00/02`0;00/000@MP08
;00/02`0;008;008EgIGMUMfEgH9E`08MUMfEgIGMUL9MP06EgIGMUMf8UL0205G0EL1E`5G1@010@40
00000`010@0500811@000`40000200T1021G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5fEgIGMUMfEaQf
00L/MRaf;7H/00]f00H/MRaf;7H2;00702`0;00/0008;1Uf42`@MP0702`0;00/0009;009MUMfEgIG
MUMf011G27H025MfEgIGMUMf6UL01P5G0EL1E`T11@010@40000?00030@00008001QG0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`49Ea5f02l/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
02`0;00/0009;0Qf010/MRaf;7H/MRaf;7H/MRaf0R`03`0/02`0;00/02`0;00/000@MP08;00/02`0
;008;008EgIGMUMfEgH9E`08MUMfEgIGMUL9MP06EgIGMUMf8UL0205G0EL1E`5G1@010@40000?0003
0@0000802@4085L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0GIGMUMfEgIG67H01baf;7H/MR`02gH01Raf
;7H/MP8/00L0;00/02`000P/6GH@;11f00L0;00/02`000T/00UfEgIGMUMfEgH045L8MP08EgIGMUMf
EgHJE`060EL1E`5G2@4500410@0000l000<100000P0065L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@UG
4GH0;baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H0;00/02`000T/27H042af
;7H/MRaf;7H/MRaf;7H2;00?02`0;00/02`0;00/02`0011f00P/02`0;00/00P/00QGMUMfEgIGMPUG
00QfEgIGMUMfE`Uf00IGMUMfEgHRE`080EL1E`5G0EL500410@0000l000<100000P090@0PE`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1MUMfEgIGMULHMP07;7H/MRaf;00;MP06;7H/MRaf0R`01`0/02`0;000
22`IMQ0/47H01`0/02`0;0002B`02GIGMUMfEgIGMP0@E`Qf00QGMUMfEgIGMQYG00H1E`5G0EL90@D0
0@4100003`000`400002000PE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1MUMfEgIGMUL:MP0o;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf02`0;00/02`0;00/02`0008/
01Mf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MP02;00G02`0;00/02`0;00/02`0MRaf;7H/MR`027H022`0
;00/02`022`025MfEgIGMUMf2EL022af;7H/MRaf2008MR5G00P1E`5G0EL1E`D00@4100003`020@<0
2@403UL1E`5G0EL1E`5G0EL12ULHMP0P;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H2;007
02`0;00/0008;1Uf22`0200/02`0;00/2GH03UMfEgIGMUL0;00/02`02B`AE`07MRaf;7H/MP02;007
02`0;00/0008MR9G00L1E`5G0EL100@00P4100003`000`400002000PE`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1MUMfEgIGMUL:MP0o;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf02`0;00/02`0;00/02`0008/01Mf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MP02;00G02`0;00/
02`0;00/02`0MRaf;7H/MR`027H022`0;00/02`022`025MfEgIGMUMf2EL022af;7H/MRaf2008MR5G
00P1E`5G0EL1E`D00@4100003`000`40000200T100iG0EL1E`5G0EL1E`5G0@YG67H082af;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf0R`01`0/02`0;00022`IMPP/00P0;00/02`0;0Uf00iGMUMf
EgIG02`0;00/00T/4EL01gH/MRaf;7H00R`01`0/02`0;00027HRE`070EL1E`5G0@0500410@0000l0
00<100000P0085L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0GIGMUMfEgIG2WH0?baf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MP0/02`0;00/02`0;00/0002;00GMRaf;7H/
MRaf;7H/MRaf;7H/MRaf;7H00R`05`0/02`0;00/02`0;00/07H/MRaf;7H/00Qf00P/02`0;00/00P/
00QGMUMfEgIGMPUG00P/MRaf;7H/MPP027HQE`080EL1E`5G0EL500410@0000l000<100000P090@0>
E`5G0EL1E`5G0EL1E`4:EaQf020/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MP8/00L0;00/
02`000P/6GH8;00802`0;00/02`9MP0>EgIGMUMfE`0/02`0;009;15G00Mf;7H/MRaf008/00L0;00/
02`000Qf8UL01`5G0EL1E`401@010@40000?00030@000080021G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5fEgIGMUMfE`Yf03l/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H0;00/02`0;00/02`0;0000R`05gH/MRaf;7H/MRaf;7H/MRaf;7H/MRaf008/01L0;00/02`0;00/
02`0;01f;7H/MRaf;008MP08;00/02`0;008;008EgIGMUMfEgH9E`08;7H/MRaf;7H800Qf8EL0205G
0EL1E`5G1@010@40000?00030@0000802@403UL1E`5G0EL1E`5G0EL12ULHMP0P;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H2;00702`0;00/0008;1Uf22`0200/02`0;00/2GH03UMfEgIGMUL0
;00/02`02B`AE`07MRaf;7H/MP02;00702`0;00/0008MR9G00L1E`5G0EL100D00@4100003`000`40
0002000@E`5G0EL1E`5G0EL1E`5G0@QG2WH0>2af;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf22`8000H;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf0R`01P0/
02`0;0T022`8MPP0010/02`0;00/05MfEgIGMUMf2EL8;0P027HYE`D00@4100003`020@<02@403UL1
E`5G0EL1E`5G0EL10UL027IGMUMfEgIG47H01baf;7H/MR`02GH04Baf;7H/MRaf;7H/MRaf;7H/00Qf
2B`0;`0/02`0;01f;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf02`0;00/02`0;00/02`0008/00If;7H/
MR`9MP0@02`0;00/02`0;00/02`0;15G22`800Qf8UL01`5G0EL1E`4010020@40000?00030@000080
011G0EL1E`5G0EL1E`5G0EL125L:MP0h;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H8;0P001P/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H2;00602`0;00/
2@08;0Qf200042`0;00/02`0EgIGMUMfEgH9E`P/2008MRUG1@010@40000?00030@0000802@403UL1
E`5G0EL1E`5G0EL10UL027IGMUMfEgIG47H01baf;7H/MR`02GH04Baf;7H/MRaf;7H/MRaf;7H/00Qf
2B`0;`0/02`0;01f;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf02`0;00/02`0;00/02`0008/00If;7H/
MR`9MP0@02`0;00/02`0;00/02`0;15G22`800Qf8UL01`5G0EL1E`401@010@40000?00030@000080
011G0EL1E`5G0EL1E`5G0EL125L:MP0h;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H8;0P001P/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H2;00602`0;00/
2@08;0Qf200042`0;00/02`0EgIGMUMfEgH9E`P/2008MRUG1@010@40000?00030@0000802@403UL1
E`5G0EL1E`5G0EL10UL027IGMUMfEgIG47H01baf;7H/MR`02GH04Baf;7H/MRaf;7H/MRaf;7H/00Qf
2B`0;`0/02`0;01f;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf02`0;00/02`0;00/02`0008/00If;7H/
MR`9MP0@02`0;00/02`0;00/02`0;15G22`800Qf8UL01`5G0EL1E`401@010@40000?00030@000080
011G0EL1E`5G0EL1E`5G0EL125L:MP0h;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H8;0P001P/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H2;00602`0;00/
2@08;0Qf200042`0;00/02`0EgIGMUMfEgH9E`P/2008MRUG1@010@40000?00030@0000802@403UL1
E`5G0EL1E`5G0EL10UL027IGMUMfEgIG47H01baf;7H/MR`02GH04Baf;7H/MRaf;7H/MRaf;7H/00Qf
2B`0;`0/02`0;01f;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf02`0;00/02`0;00/02`0008/00If;7H/
MR`9MP0@02`0;00/02`0;00/02`0;15G22`800Qf8UL01`5G0EL1E`401@010@40000?00030@000080
011G0EL1E`5G0EL1E`5G0EL125L:MP0h;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H8;0P02B`03gH/MRaf;7H/MRaf;7H/MP02;00602`0;00/2@0022`0
;00/02`027H@0008EgIGMUMfEgH9E`0802`0;00/02`80007MUMfEgIGMP0CE`07MUMfEgIGMP0@E`D0
0@4100003`000`40000200T100IG0EL1E`49EaUf024/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MR`027H8;0P022`AMPP/2008;0Qf0100;00/02`0;00/02`0;00/4EL07b`0;00/02`0;00/02`0
;01GMUMfEgIGMP5G0EL1E`408UL500410@0000l00P43000@E`5G0EL1E`5G0EL1E`5G0@QG2WH0>2af
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf22`800T/
00mf;7H/MRaf;7H/MRaf;7H00R`01P0/02`0;0T000P/02`0;00/00Qf400025MfEgIGMUMf2EL0200/
02`0;00/20001gIGMUMfEgH04eL01gIGMUMfEgH045L400810@0000l000<100000P090@06E`5G0EL1
2ELIMP0Q;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Qf22`800P/4GH8;0P022`8MP0@
02`0;00/02`0;00/02`0;15G01l/02`0;00/02`0;00/02`0EgIGMUMfEgH1E`5G0EL1029G1@010@40
000?00030@000080011G0EL1E`5G0EL1E`5G0EL125L:MP0h;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H8;0P02B`03gH/MRaf;7H/MRaf;7H/MP02;006
02`0;00/2@0022`0;00/02`027H@0008EgIGMUMfEgH9E`0802`0;00/02`80007MUMfEgIGMP0CE`07
MUMfEgIGMP0@E`D00@4100003`000`40000200T100IG0EL1E`49EaUf024/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MR`027H8;0P022`AMPP/2008;0Qf0100;00/02`0;00/02`0;00/4EL07b`0
;00/02`0;00/02`0;01GMUMfEgIGMP5G0EL1E`408UL500410@0000l000<100000P0045L1E`5G0EL1
E`5G0EL1E`48E`Yf03P/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MPP/2009;00?MRaf;7H/MRaf;7H/MRaf008/00H0;00/02`90008;00/02`0;008MQ00
00QGMUMfEgIGMPUG00P0;00/02`0;0P000MfEgIGMUMf01=G00MfEgIGMUMf011G1@010@40000?0003
0@0000802@401UL1E`5G0@UG6GH08Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;008MPP/
2008;15f22`800P/27H0400/02`0;00/02`0;00/02`AE`0O;00/02`0;00/02`0;00/05MfEgIGMUMf
0EL1E`5G0@0RE`D00@4100003`000`400002000@E`5G0EL1E`5G0EL1E`5G0@QG048/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H8;100
2B`01gH/MRaf;7H02B`@00Qf4009E`Qf2@001b`0;00/02`06UL03gIGMUMfEgIGMUMfEgIGMP08E`D0
0@4100003`000`40000200T100IG0EL1E`49EaUf030/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H2;00?02`0;00/02`0;00/02`000P/2GH022af;7H/MRaf2@001b`0
;00/02`027H0400/02`0;00/02`0;00/02`:E`07MUMfEgIGMP080008;00/02`0;008E`070EL1E`5G
0@0RE`D00@4100003`020@<0011G0EL1E`5G0EL1E`5G0EL125L0@Raf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MPP/4009;007MRaf;7H/
MP09;10027H@00UG27H90007;00/02`0;00JE`0?MUMfEgIGMUMfEgIGMUMf00QG10020@40000?0003
0@0000802@401UL1E`5G0@UG6GH0<2af;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MP8/00l0;00/02`0;00/02`0;00022`9MP08;7H/MRaf;7H90007;00/02`0;008MP0@
02`0;00/02`0;00/02`0;0YG00MfEgIGMUMf00P000P/02`0;00/00QG00L1E`5G0EL1029G1@010@40
000?00030@000080011G0EL1E`5G0EL1E`5G0EL125L0@Raf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MPP/4009;007MRaf;7H/MP09;100
27H@00UG27H90007;00/02`0;00JE`0?MUMfEgIGMUMfEgIGMUMf00QG1@010@40000?00030@000080
2@401UL1E`5G0@UG6GH0<2af;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MP8/00l0;00/02`0;00/02`0;00022`9MP08;7H/MRaf;7H90007;00/02`0;008MP0@02`0;00/
02`0;00/02`0;0YG00MfEgIGMUMf00P000P/02`0;00/00QG00L1E`5G0EL1029G1@010@40000?0003
0@000080011G0EL1E`5G0EL1E`5G0EL125L0@Raf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MPP/4009;007MRaf;7H/MP09;10027H@00UG
27H90007;00/02`0;00JE`0?MUMfEgIGMUMfEgIGMUMf00QG1@010@40000?00030@0000802@401UL1
E`5G0@UG6GH0<2af;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MP8/
00l0;00/02`0;00/02`0;00022`9MP08;7H/MRaf;7H90007;00/02`0;008MP0@02`0;00/02`0;00/
02`0;0YG00MfEgIGMUMf00P000P/02`0;00/00QG00L1E`5G0EL1029G1@010@40000?00030@000080
011G0EL1E`5G0EL1E`5G0EL125L0DRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf02`0;00/02`@00T/00Mf;7H/MRaf008/
00H0;00/02`90007;00/02`0;00A00Uf22`800P/6UL05gIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf00D0
0@41000000@00@410`030@D000<100000P090@06E`5G0EL12ELAMP0h;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H9;0P000P/02`0;00/00X/00mf;7H/
MRaf02`0;00/02`02008;00802`0;00/02`800Yf00L/MRaf;7H/00P022`03`5G0EL1E`5G0EL1E`5G
0@0BE`07MUMfEgIGMP09E`D00@4100000P000`40000500030@00008000<100000P0045L1E`5G0EL1
E`5G0EL1E`48E`1B;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H0;00/02`0;1002B`01gH/MRaf;7H00R`01P0/02`0;0T0
00L/02`0;00/01402GH8;0P022`JE`0GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH01@010@4000020003
0@0000D000<100000P030@802@401UL1E`5G0@UG4GH0>2af;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf2B`80008;00/02`0;00:;00?MRaf;7H/MP0/02`0
;00/00P022`0200/02`0;00/200:MP07;7H/MRaf;00800P/00l1E`5G0EL1E`5G0EL1E`404UL01gIG
MUMfEgH02EL400810@00008000<100000P030@D000<100000P0045L1E`5G0EL1E`5G0EL1E`48E`1B
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H0;00/02`0;1002B`01gH/MRaf;7H00R`01P0/02`0;0T000L/02`0;00/0140
2GH8;0P022`JE`0GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH01@010@40000200030@00008000<10000
1@000`40000200T100IG0EL1E`49Ea5f03P/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MPT/200022`0;00/02`02R`03gH/MRaf;7H0;00/02`0;00800P/
00P0;00/02`0;0P02WH01baf;7H/MR`02008;00?0EL1E`5G0EL1E`5G0EL1019G00MfEgIGMUMf00UG
1@010@4000000`010@0400@110000`400002000@E`5G0EL1E`5G0EL1E`5G0@QG058/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MP0/02`0;00/4009;007MRaf;7H/MP02;00602`0;00/2@001b`0;00/02`04@09MPP/2008;1YG
01MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP0500410@0000l000<100000P090@06E`5G0EL12ELAMP0h
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H9;0P0
00P/02`0;00/00X/00mf;7H/MRaf02`0;00/02`02008;00802`0;00/02`800Yf00L/MRaf;7H/00P0
22`03`5G0EL1E`5G0EL1E`5G0@0BE`07MUMfEgIGMP09E`D00@4100003`000`400002000@E`5G0EL1
E`5G0EL1E`5G0@QG04X/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MPP/4@003b`0;00/02`0;00/02`0;012000@;7H/MRaf
;7IG0EL1E`5G0A9G01MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP0500410@0000l000<100000P090@06
E`5G0EL12ELIMP0h;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H2;00602`0;00/2@0022`0;00/02`02B`0200/02`0;00/8@003b`0;00/02`0;00/02`0
;00900Qf00l1E`5G0EL1E`5G0EL1E`404UL01gIGMUMfEgH02EL500410@0000l000<100000P0045L1
E`5G0EL1E`5G0EL1E`48E`1:;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H8;14000l/02`0;00/02`0;00/02`0@P0042af
;7H/MRafE`5G0EL1E`4BE`0GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH01@010@40000?00810`090@06
E`5G0EL12ELIMP0h;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H2;00602`0;00/2@0022`0;00/02`02B`0200/02`0;00/8@003b`0;00/02`0;00/02`0
;00900Qf00l1E`5G0EL1E`5G0EL1E`404UL01gIGMUMfEgH02EL400810@0000l000<100000P0045L1
E`5G0EL1E`5G0EL1E`48E`1:;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H8;14000l/02`0;00/02`0;00/02`0@P0042af
;7H/MRafE`5G0EL1E`4BE`0GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH01@010@40000?00030@000080
2@401UL1E`5G0@UG6GH0>2af;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf0R`01P0/02`0;0T000P/02`0;00/00T/00P0;00/02`0;24000l/02`0;00/02`0
;00/02`02@08MP0?0EL1E`5G0EL1E`5G0EL1019G00MfEgIGMUMf00UG1@010@40000?00030@000080
011G0EL1E`5G0EL1E`5G0EL125L0BRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf22`A000?;00/02`0;00/02`0;00/0480
010/MRaf;7H/MUL1E`5G0EL14UL05gIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf00D00@4100003`000`40
000200T100IG0EL1E`49EaUf03P/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MP8/00H0;00/02`90008;00/02`0;009;00802`0;00/02`Q000?;00/02`0
;00/02`0;00/00T027H03`5G0EL1E`5G0EL1E`5G0@0BE`07MUMfEgIGMP09E`D00@4100003`000`40
000201QG04X/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MPP/00P0;00/02`0;14000L/02`0;00/03/000L/02`0;00/00Qf
00QG0EL1E`5G0A9G01MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP0500410@0000l000<100000P003`5G
0EL1E`5G0EL1E`5G0@09EaUf00l/MRaf;7H/MRaf;7H/MR`04gH05Raf;7H/MRaf;7H/MRaf;7H/MRaf
;7H9;0P001T/02`0;00/02`0;00/02`0;00/02`0;00/034001l/02`0;00/05MfEgIGMUMf0EL1E`5G
0EL1E`5G0EL1019G011fEgIGMUMfEgIGMUMfEgIG1@010@40000?00030@00008065L0BRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf22`0200/02`0;00/4@001b`0;00/02`0>`001b`0;00/02`027H025L1E`5G0EL14UL05gIG
MUMfEgIGMUMfEgIGMUMfEgIGMUMf00D00@4100003`020@<000l1E`5G0EL1E`5G0EL1E`402ELIMP0?
;7H/MRaf;7H/MRaf;7H/01=f01H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf2B`8000I;00/02`0;00/02`0
;00/02`0;00/02`0;00a000O;00/02`0;01GMUMfEgIGMP5G0EL1E`5G0EL1E`5G0@0BE`0@MUMfEgIG
MUMfEgIGMUMfE`@00P4100003`000`40000201QG04X/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MPP/00P0;00/02`0;140
00L/02`0;00/03/000L/02`0;00/00Qf00QG0EL1E`5G0A9G01MfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
MP0500410@0000l000<100000P003`5G0EL1E`5G0EL1E`5G0@09EaUf00l/MRaf;7H/MRaf;7H/MR`0
4gH05Raf;7H/MRaf;7H/MRaf;7H/MRaf;7H9;0P001T/02`0;00/02`0;00/02`0;00/02`0;00/0340
01l/02`0;00/05MfEgIGMUMf0EL1E`5G0EL1E`5G0EL1019G011fEgIGMUMfEgIGMUMfEgIG1@010@40
000?00030@00008065L0BRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf22`0200/02`0;00/4@001b`0;00/02`0>`001b`0
;00/02`027H025L1E`5G0EL14UL05gIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf00D00@4100003`000`40
0002000?0EL1E`5G0EL1E`5G0EL100UG6GH03baf;7H/MRaf;7H/MRaf;00CMP0F;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MPT/20006B`0;00/02`0;00/02`0;00/02`0;00/02`0<@007b`0;00/02`0EgIGMUMf
EgH1E`5G0EL1E`5G0EL1E`404UL047IGMUMfEgIGMUMfEgIGMUL500410@0000l000<100000P0HE`1:
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H8;00802`0;00/02`A0007;00/02`0;00k0007;00/02`0;008MP08E`5G0EL1
E`4BE`0GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH01@010@40000?00030@00008065L9MP0H;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf4B`027H/MRaf;7H/4GH01Raf;7H/MPX/01L0;00/02`0;00/02`0;00/
02`0;00/0009;00802`0;00/02`Y000O;00/02`0;01GMUMfEgIGMP5G0EL1E`5G0EL1E`5G0@0BE`0@
MUMfEgIGMUMfEgIGMUMfE`D00@4100003`000`400002011G021fEgIGMUMfEbaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MQT/025f;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H022`A000?;00/02`0
;00/02`0;00/03<000h/02`0;00/MUMfEgIGMP9G00L1E`5G0EL1019G01MfEgIGMUMfEgIGMUMfEgIG
MUMfEgIGMP0500410@0000l000<100000P0HE`Uf01P/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7HA;008
MRaf;7H/MR`AMP06;7H/MRaf2R`05`0/02`0;00/02`0;00/02`0;00/02`000T/00P0;00/02`0;2T0
01l/02`0;00/05MfEgIGMUMf0EL1E`5G0EL1E`5G0EL1019G011fEgIGMUMfEgIGMUMfEgIG1@010@40
000?00810`0@E`0PMUMfEgIGMUL/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7HI;00QMRaf;7H/MRaf;7H/
MRaf;7H/MRaf;7H/MRaf;7H/MRaf00P/4@003b`0;00/02`0;00/02`0;00c000>;00/02`0;7IGMUMf
EgH2E`070EL1E`5G0@0BE`0GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH010020@40000?00030@000080
65L9MP0H;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf4B`027H/MRaf;7H/4GH01Raf;7H/MPX/01L0;00/
02`0;00/02`0;00/02`0;00/0009;00802`0;00/02`Y000O;00/02`0;01GMUMfEgIGMP5G0EL1E`5G
0EL1E`5G0@0BE`0@MUMfEgIGMUMfEgIGMUMfE`D00@4100003`000`400002011G021fEgIGMUMfEbaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MQT/025f;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H0
22`A000?;00/02`0;00/02`0;00/03<000h/02`0;00/MUMfEgIGMP9G00L1E`5G0EL1019G01MfEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMP0500410@0000l000<100000P0HE`Uf01P/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7HA;008MRaf;7H/MR`AMP06;7H/MRaf2R`05`0/02`0;00/02`0;00/02`0;00/02`000T/
00P0;00/02`0;2T001l/02`0;00/05MfEgIGMUMf0EL1E`5G0EL1E`5G0EL1019G011fEgIGMUMfEgIG
MUMfEgIG1@010@40000?00030@00008045L087IGMUMfEgIG;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
6B`08GH/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MP08;14000l/02`0;00/02`0;00/02`0
<`003R`0;00/02afEgIGMUMf0UL01`5G0EL1E`404UL05gIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf00D0
0@4100003`000`40000200070EL1E`5G0@0:E`06MUMfEgIG2WH062af;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MPP/0240;00/02`0;00/02`0;00/02`0;00/02`0MRaf;7H/MR`02GH01Raf;7H/MP8/00H0;00/
02`;0006;00/02`04B`0400/02`0;00/02`0;00/02`I00P/01MGMUMfEgIGMP5G0EL1E`5G0EL1E`5G
0@0BE`0@MUMfEgIGMUMfEgIGMUMfE`D00@4100003`000`40000200QG011fEgIGMUMfEgIGMUMfEgIG
2WH05Raf;7H/MRaf;7H/MRaf;7H/MRaf;7H:;00602`0;00/4P0082af;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MP0/02`0;00/4@09;00602`0;00/:`003R`0;00/02afEgIGMUMf0UL01`5G0EL1E`404UL05gIG
MUMfEgIGMUMfEgIGMUMfEgIGMUMf00D00@4100003`000`40000200070EL1E`5G0@0:E`06MUMfEgIG
2WH062af;7H/MRaf;7H/MRaf;7H/MRaf;7H/MPP/0240;00/02`0;00/02`0;00/02`0;00/02`0MRaf
;7H/MR`02GH01Raf;7H/MP8/00H0;00/02`;0006;00/02`04B`0400/02`0;00/02`0;00/02`I00P/
01MGMUMfEgIGMP5G0EL1E`5G0EL1E`5G0@0BE`0@MUMfEgIGMUMfEgIGMUMfE`D00@4100003`020@<0
25L047IGMUMfEgIGMUMfEgIGMUL:MP0F;7H/MRaf;7H/MRaf;7H/MRaf;7H/MPX/00H0;00/02`B000P
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf02`0;00/02`A00T/00H0;00/02`[000>;00/02`0;7IGMUMf
EgH2E`070EL1E`5G0@0BE`0GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH010020@40000?00030@000080
00L1E`5G0EL100YG00IfEgIGMUL:MP0H;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf22`08@0/02`0;00/
02`0;00/02`0;00/02`0;01f;7H/MRaf;009MP06;7H/MRaf0R`01P0/02`0;0/000H/02`0;00A;00@
02`0;00/02`0;00/02`0;1T022`05eMfEgIGMUMf0EL1E`5G0EL1E`5G0EL1019G011fEgIGMUMfEgIG
MUMfEgIG1@010@40000?00030@00008025L047IGMUMfEgIGMUMfEgIGMUL:MP0F;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MPX/00H0;00/02`B000P;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf02`0;00/02`A00T/
00H0;00/02`[000>;00/02`0;7IGMUMfEgH2E`070EL1E`5G0@0BE`0GMUMfEgIGMUMfEgIGMUMfEgIG
MUMfEgH01@010@40000?00030@00008000L1E`5G0EL100YG00IfEgIGMUL:MP0H;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf22`08@0/02`0;00/02`0;00/02`0;00/02`0;01f;7H/MRaf;009MP06;7H/MRaf
0R`01P0/02`0;0/000H/02`0;00A;00@02`0;00/02`0;00/02`0;1T022`05eMfEgIGMUMf0EL1E`5G
0EL1E`5G0EL1019G011fEgIGMUMfEgIGMUMfEgIG1@010@40000?00030@00008025L047IGMUMfEgIG
MUMfEgIGMUL:MP0F;7H/MRaf;7H/MRaf;7H/MRaf;7H/MPX/00H0;00/02`B000P;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf02`0;00/02`A00T/00H0;00/02`[000>;00/02`0;7IGMUMfEgH2E`070EL1E`5G
0@0BE`0GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH01@010@40000?00030@00008000l1E`5G0EL1E`5G
0EL1E`400UL01WIGMUMfEa9f00l/MRaf;7H/MRaf;7H/MR`02GH042af;7H/MRaf;7H/MRaf;7H2;00?
02`0;00/02`0;00/02`000P/00Mf;7H/MRaf00X/01L0;00/02`0;00/02`0;00/02`0;00/000A;1T0
00P/02`0;00/00P/01MGMUMfEgIGMP5G0EL1E`5G0EL1E`5G0@0:E`0HMUMfEgIGMUMfEgIGMUMfEgIG
MUMfEgIG1@010@40000?00030@00008000QG0EL1E`5G0@QG00QfEgIGMUMfE`Yf02H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MPX/4009;007MRaf;7H/MP08;00@02`0;00/02`0;00/
02`0;0T02B`01P0/02`0;2<000h/02`0;00/MUMfEgIGMP9G00L1E`5G0EL1019G01MfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMP0500410@000003004100810`020@D000<100000P003`5G0EL1E`5G0EL1E`5G
0@02E`06MUMfEgIG4WH03baf;7H/MRaf;7H/MRaf;009MP0@;7H/MRaf;7H/MRaf;7H/MP8/00l0;00/
02`0;00/02`0;00022`01gH/MRaf;7H02R`05`0/02`0;00/02`0;00/02`0;00/02`0014/6@0022`0
;00/02`022`05eMfEgIGMUMf0EL1E`5G0EL1E`5G0EL100YG01QfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
MUL500410@00008000<100000P001040004400030@00008000QG0EL1E`5G0@QG00QfEgIGMUMfE`Yf
02H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MPX/4009;007MRaf;7H/MP08;00@
02`0;00/02`0;00/02`0;0T02B`01P0/02`0;2<000h/02`0;00/MUMfEgIGMP9G00L1E`5G0EL1019G
01MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP0500410@0000<000D100000@0200030@0000800`42000?
0EL1E`5G0EL1E`5G0EL1009G00IfEgIGMULBMP0?;7H/MRaf;7H/MRaf;7H/00Uf010/MRaf;7H/MRaf
;7H/MRaf0R`03`0/02`0;00/02`0;00/0008;007MRaf;7H/MP0:;00G02`0;00/02`0;00/02`0;00/
02`0;0004B`I0008;00/02`0;008;00GEgIGMUMfEgH1E`5G0EL1E`5G0EL1E`402UL067IGMUMfEgIG
MUMfEgIGMUMfEgIGMUMfE`@00P41000010001040004200030@00008000<100000P0025L1E`5G0EL1
25L027IGMUMfEgIG2WH09Raf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf2R`@00T/
00Mf;7H/MRaf00P/0100;00/02`0;00/02`0;00/2@09;00602`0;00/8`003R`0;00/02afEgIGMUMf
0UL01`5G0EL1E`404UL05gIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf00D00@41000000D00@000@020004
0@000@@000<100000P003`5G0EL1E`5G0EL1E`5G0@02E`06MUMfEgIG4WH03baf;7H/MRaf;7H/MRaf
;009MP0@;7H/MRaf;7H/MRaf;7H/MP8/00l0;00/02`0;00/02`0;00022`01gH/MRaf;7H02R`05`0/
02`0;00/02`0;00/02`0;00/02`0014/6@0022`0;00/02`022`05eMfEgIGMUMf0EL1E`5G0EL1E`5G
0EL100YG01QfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUL500410@0000800P4400811@000`4000020008
E`5G0EL1E`48E`08MUMfEgIGMUL:MP0V;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H:;1002B`01gH/MRaf;7H022`0400/02`0;00/02`0;00/02`900T/00H0;00/02`S000>;00/02`0
;7IGMUMfEgH2E`070EL1E`5G0@0BE`0GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH01@010@40000?0003
0@00008025L01`5G0EL1E`400UL01WIGMUMfEaYf00l/MRaf;7H/MRaf;7H/MR`04gH01Raf;7H/MP8/
00l0;00/02`0;00/02`0;0006R`07gH/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H022`A000W
;00/02`0;01f;7H/MRaf;5MfEgIGMUMf0EL1E`5G0EL1E`5G0EL100YG01QfEgIGMUMfEgIGMUMfEgIG
MUMfEgIGMUL500410@0000l000<100000P0@E`08MUMfEgIGMUL:MP0`;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf22`A0006;00/02`04R`03gH/MRaf;7H/MRaf;7H/
MP09;007MRaf;7H/MP02;00602`0;00/6P08;007MUMfEgIGMP02E`070EL1E`5G0@0BE`0GMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgH01@010@40000?00030@00008025L01`5G0EL1E`400UL01WIGMUMfEaYf
00l/MRaf;7H/MRaf;7H/MR`04gH01Raf;7H/MP8/00l0;00/02`0;00/02`0;0006R`07gH/MRaf;7H/
MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H022`A000W;00/02`0;01f;7H/MRaf;5MfEgIGMUMf0EL1E`5G
0EL1E`5G0EL100YG01QfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUL500410@0000l000<100000P0@E`08
MUMfEgIGMUL:MP0`;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
22`A0006;00/02`04R`03gH/MRaf;7H/MRaf;7H/MP09;007MRaf;7H/MP02;00602`0;00/6P08;007
MUMfEgIGMP02E`070EL1E`5G0@0BE`0GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH01@010@40000?0081
0`08E`070EL1E`5G0@02E`06MUMfEgIG6WH03baf;7H/MRaf;7H/MRaf;00CMP06;7H/MRaf0R`03`0/
02`0;00/02`0;00/000J;00OMRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MP08;14002L/02`0
;00/07H/MRaf;7H/EgIGMUMfEgH1E`5G0EL1E`5G0EL1E`402UL067IGMUMfEgIGMUMfEgIGMUMfEgIG
MUMfE`@00P4100003`000`400002011G00QfEgIGMUMfE`Yf030/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H8;14000H/02`0;00B;00?MRaf;7H/MRaf;7H/MRaf00T/
00Mf;7H/MRaf008/00H0;00/02`J00P/00MfEgIGMUMf009G00L1E`5G0EL1019G01MfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMP0500410@0000l000<100000P08E`070EL1E`5G0@02E`06MUMfEgIG6WH03baf
;7H/MRaf;7H/MRaf;00CMP06;7H/MRaf0R`03`0/02`0;00/02`0;00/000J;00OMRaf;7H/MRaf;7H/
MRaf;7H/MRaf;7H/MRaf;7H/MP08;14002L/02`0;00/07H/MRaf;7H/EgIGMUMfEgH1E`5G0EL1E`5G
0EL1E`402UL067IGMUMfEgIGMUMfEgIGMUMfEgIGMUMfE`D00@4100003`000`400002011G00QfEgIG
MUMfE`Yf030/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H8;140
00H/02`0;00B;00?MRaf;7H/MRaf;7H/MRaf00T/00Mf;7H/MRaf008/00H0;00/02`J00P/00MfEgIG
MUMf009G00L1E`5G0EL1019G01MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP0500410@0000l000<10000
0P0iMP07;7H/MRaf;009MP09;7H/MRaf;7H/00Qf00L/02`0;00/00T001P/02`0;00/02`0;00/02`0
;00/02`0;00:;00WMRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MP0/02`0;00/00T0010/02`0
;00/07H/MRaf;7H/25L03`5G0EL1E`5G0EL1E`5G0@0BE`0@MUMfEgIGMUMfEgIGMUMfE`D00@410000
3`000`400002029f040/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MP0/02`0;00/4@00>2`0;00/02`0;00/02`0;00/02`0;00/07H/MRaf;7H/MRaf;7H/
MRaf;7H/MRaf;7H/MRaf;7H/6@0022af;7H/MRaf2EL01`5G0EL1E`404UL05gIGMUMfEgIGMUMfEgIG
MUMfEgIGMUMf00D00@4100003`000`40000203Uf00L/MRaf;7H/00Uf00T/MRaf;7H/MR`027H01b`0
;00/02`02@0062`0;00/02`0;00/02`0;00/02`0;00/00X/02Mf;7H/MRaf;7H/MRaf;7H/MRaf;7H/
MRaf;7H/MRaf02`0;00/02`02@0042`0;00/02`0MRaf;7H/MR`8E`0?0EL1E`5G0EL1E`5G0EL1019G
011fEgIGMUMfEgIGMUMfEgIG1@010@40000?00030@0000808WH0@2af;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf02`0;00/02`A000h;00/02`0;00/02`0
;00/02`0;00/02`0MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`I0008;7H/MRaf;7H9E`07
0EL1E`5G0@0BE`0GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH01@010@40000?00030@000080>GH01baf
;7H/MR`02GH02Baf;7H/MRaf;008MP07;00/02`0;009000H;00/02`0;00/02`0;00/02`0;00/02`0
2R`09gH/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H0;00/02`0;009000@;00/02`0;01f;7H/
MRaf;0QG00l1E`5G0EL1E`5G0EL1E`404UL047IGMUMfEgIGMUMfEgIGMUL500410@0000l00P43029f
040/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MP0/
02`0;00/4@00>2`0;00/02`0;00/02`0;00/02`0;00/07H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/
MRaf;7H/6@0022af;7H/MRaf2EL01`5G0EL1E`404UL05gIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf00@0
0P4100003`000`40000203Uf00L/MRaf;7H/00Uf00T/MRaf;7H/MR`027H01b`0;00/02`02@0062`0
;00/02`0;00/02`0;00/02`0;00/00X/02Mf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf02`0
;00/02`02@0042`0;00/02`0MRaf;7H/MR`8E`0?0EL1E`5G0EL1E`5G0EL1019G011fEgIGMUMfEgIG
MUMfEgIG1@010@40000?00030@0000808WH0@2af;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf02`0;00/02`A000h;00/02`0;00/02`0;00/02`0;00/02`0
MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`I0008;7H/MRaf;7H9E`070EL1E`5G0@0BE`0G
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH01@010@40000?00030@0000802GH01baf;7H/MR`0<GH04Baf
;7H/MRaf;7H/MRaf;7H/00Uf00H/MRaf;7H2;00602`0;00/6`001R`0;00/00X/01Mf;7H/MRaf;7H/
MRaf;7H/MRaf;7H/MP08;0T000P/02`0;00/00Qf01L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@0:E`0H
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG1@010@40000?00030@00008027H0HRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MP0/02`0;00/8@0022`0;00/02`022`03gH/MRaf;7H/MRaf;7H/MP02;006
02`0;00/4P0022af;7H/MRaf2EL01`5G0EL1E`404UL05gIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf00D0
0@4100003`000`40000200Uf00L/MRaf;7H/035f014/MRaf;7H/MRaf;7H/MRaf;009MP06;7H/MRaf
0R`01P0/02`0;1/000H/02`0;00:;00GMRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H022`90008;00/02`0
;008MP0G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`402UL067IGMUMfEgIGMUMfEgIGMUMfEgIGMUMfE`D0
0@4100003`000`40000200Qf068/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H0;00/02`0;240
00P/02`0;00/00P/00mf;7H/MRaf;7H/MRaf;7H00R`01P0/02`0;18000P/MRaf;7H/MPUG00L1E`5G
0EL1019G01MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP0500410@0000l000<100000P09MP07;7H/MRaf
;00aMP0A;7H/MRaf;7H/MRaf;7H/MR`02GH01Raf;7H/MP8/00H0;00/02`K0006;00/02`02R`05gH/
MRaf;7H/MRaf;7H/MRaf;7H/MRaf00P/2@0022`0;00/02`027H05`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL100YG01QfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUL500410@0000l00P4300Qf068/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H0;00/02`0;24000P/02`0;00/00P/00mf;7H/MRaf;7H/MRaf;7H0
0R`01P0/02`0;18000P/MRaf;7H/MPUG00L1E`5G0EL1019G01MfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
MP0400810@0000l000<100000P09MP07;7H/MRaf;00aMP0A;7H/MRaf;7H/MRaf;7H/MR`02GH01Raf
;7H/MP8/00H0;00/02`K0006;00/02`02R`05gH/MRaf;7H/MRaf;7H/MRaf;7H/MRaf00P/2@0022`0
;00/02`027H05`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL100YG01QfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
MUL500410@0000l000<100000P08MP1R;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf02`0;00/
02`Q0008;00/02`0;008;00?MRaf;7H/MRaf;7H/MRaf008/00H0;00/02`B0008;7H/MRaf;7H9E`07
0EL1E`5G0@0BE`0GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH01@010@40000?00030@0000802GH01baf
;7H/MR`0<GH04Baf;7H/MRaf;7H/MRaf;7H/00Uf00H/MRaf;7H2;00602`0;00/6`001R`0;00/00X/
01Mf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MP08;0T000P/02`0;00/00Qf01L1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0@0:E`0HMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG1@010@40000?00030@00008027H0HRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf22`0200/02`0;00/6@0042`0;00/02`0
;00/02`0;009;00602`0;00/2`001b`0;00/02`027H045L1E`5G0EL1E`5G0EL1E`4BE`0GMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgH01@010@40000?00030@00008000MGMUMfEgIG06Af00H/MRaf;7H2;006
02`0;00/4`001R`0;00/024/0140;00/02`0;00/02`0;00/0008MP0G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`406UL027IGMUMfEgIG1@010@40000?00030@00008027H0HRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf22`0200/02`0;00/6@0042`0;00/02`0;00/02`0;009;00602`0;00/
2`001b`0;00/02`027H045L1E`5G0EL1E`5G0EL1E`4BE`0GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH0
1@010@40000?00030@00008000MGMUMfEgIG06Af00H/MRaf;7H2;00602`0;00/4`001R`0;00/024/
0140;00/02`0;00/02`0;00/0008MP0G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`406UL027IGMUMfEgIG
1@010@40000?00810`08MP1R;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H8;008
02`0;00/02`I000@;00/02`0;00/02`0;00/00T/00H0;00/02`;0007;00/02`0;008MP0@E`5G0EL1
E`5G0EL1E`5G0A9G01MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP0400810@0000l000<100000P001eMf
EgIGMUL0I7H01Raf;7H/MP8/00H0;00/02`C0006;00/02`08B`04@0/02`0;00/02`0;00/02`000Qf
01L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@0JE`08MUMfEgIGMUL500410@0000l000<100000P08MP1R
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H8;00802`0;00/02`I000@;00/02`0
;00/02`0;00/00T/00H0;00/02`;0007;00/02`0;008MP0@E`5G0EL1E`5G0EL1E`5G0A9G01MfEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMP0500410@0000l000<100000P001eMfEgIGMUL0I7H01Raf;7H/MP8/
00H0;00/02`C0006;00/02`08B`04@0/02`0;00/02`0;00/02`000Qf01L1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0@0JE`08MUMfEgIGMUL500410@0000l000<100000P08MP0Y;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`04GH0>2af;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf2B`@000X;00/02`0;00/02`0;00/02`0;00/02`0;00/02`0
;00/02`0;00/00T/27H065L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0A9G00mfEgIGMUMfEgIGMUMfEgH0
1@010@40000?00030@00008000MGMUMfEgIG019f00L/MRaf;7H/05Yf2B`07`0/02`0;00/02`0;00/
02`0;00/02`0;00/02`0;0004B`02@0/02`0;00/0008;0Qf01L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0@0JE`08MUMfEgIGMUL500410@0000l000<100000P08MP0Y;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MR`04GH0>2af;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf2B`@000X;00/02`0;00/02`0;00/02`0;00/02`0;00/02`0;00/02`0
;00/00T/27H065L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0A9G00mfEgIGMUMfEgIGMUMfEgH01@010@40
00000`010@020@800`4500030@00008000MGMUMfEgIG019f00L/MRaf;7H/05Yf2B`07`0/02`0;00/
02`0;00/02`0;00/02`0;00/02`0;0004B`02@0/02`0;00/0008;0Qf01L1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0@0JE`08MUMfEgIGMUL500410@00008000<100001@000`40000200030@00008027H0:Baf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/015f03P/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MPT/4000:2`0;00/02`0;00/02`0
;00/02`0;00/02`0;00/02`0;00/02`0;009;0Qf01QG0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4BE`0?
MUMfEgIGMUMfEgIGMUMf00D00@4100000`000`40000400030@0000800`420007EgIGMUMfE`0BMP07
;7H/MRaf;01JMPT/01l0;00/02`0;00/02`0;00/02`0;00/02`0;00/02`0014/00T0;00/02`0;000
22`8MP0G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`406UL027IGMUMfEgIG10020@40000400030@0000<1
1@000`40000200Qf02T/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00AMP0h
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H9;100
02P/02`0;00/02`0;00/02`0;00/02`0;00/02`0;00/02`0;00/02`02B`8MP0HE`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL14UL03gIGMUMfEgIGMUMfEgIGMP0500410@000005004000400P000`4000050003
0@00008000MGMUMfEgIG019f00L/MRaf;7H/05Yf2B`07`0/02`0;00/02`0;00/02`0;00/02`0;00/
02`0;0004B`02@0/02`0;00/0008;0Qf01L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@0JE`08MUMfEgIG
MUL500410@0000800P4300@110000`400002011f01T/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/025f
02L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`06WH01b`0;00/02`04@008B`0
;00/02`0;00/02`0;00/02`0;00/02`0;00/02`0;008MP0HE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
4UL03gIGMUMfEgIGMUMfEgIGMP0500410@0000l000<100000P001eMfEgIGMUL04WH01baf;7H/MR`0
4WH01eMfEgIGMUL0DWH8;00802`0;00/02`90006;00/02`04b`01`0/02`0;00027H08@5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@0HE`D00@4100003`000`400002011f01T/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/025f02L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`0
6WH01b`0;00/02`04@008B`0;00/02`0;00/02`0;00/02`0;00/02`0;00/02`0;008MP0HE`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL14UL03gIGMUMfEgIGMUMfEgIGMP0500410@0000l000<100000P001eMf
EgIGMUL04WH01baf;7H/MR`04WH01eMfEgIGMUL0DWH8;00802`0;00/02`90006;00/02`04b`01`0/
02`0;00027H08@5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@0HE`D00@4100003`000`40
0002011f01T/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/025f02L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MR`06WH01b`0;00/02`04@008B`0;00/02`0;00/02`0;00/02`0;00/02`0
;00/02`0;008MP0HE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL14UL03gIGMUMfEgIGMUMfEgIGMP050041
0@0000l00P430007EgIGMUMfE`0BMP07;7H/MRaf;00BMP07EgIGMUMfE`1BMPP/00P0;00/02`0;0T0
00H/02`0;00C;00702`0;00/0008MP0Q0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL101QG
10020@40000?00030@00008047H06Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`08GH09baf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00JMP07;00/02`0;00A000Q;00/02`0;00/02`0
;00/02`0;00/02`0;00/02`0;00/00Qf01QG0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4BE`0?MUMfEgIG
MUMfEgIGMUMf00D00@4100003`000`4000020007EgIGMUMfE`0BMP07;7H/MRaf;00BMP07EgIGMUMf
E`1BMPP/00P0;00/02`0;0T000H/02`0;00C;00702`0;00/0008MP0Q0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL101QG1@010@40000?00030@00008047H06Baf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MR`027H8Ee=f00L/MRaf;7H/01P000P/02`0;00/00T/00L0;00/02`0008/01mf;7H/MRafE`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL101YG00MfEgIGMUMf00D00@4100003`000`4000020007EgIGMUMf
E`0BMP0?;7H/MRaf;7H/MRaf;7H/00Uf25L027IGMUMfEgIG@gH03eMfEgIGMUL/MRaf;7H/MP0A000>
;00/02`0;00/02`0;00;;00`02`0;00/07H/MRaf;7H/0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL165L500410@0000l000<100000P0@MP0I;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;008MPQG
DgH01baf;7H/MR`0600022`0;00/02`02B`01`0/02`0;0000R`07gH/MRaf;7IG0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`406UL01gIGMUMfEgH01@010@40000?00030@00008000MGMUMfEgIG019f00l/MRaf
;7H/MRaf;7H/MR`02GH8E`08MUMfEgIGMUM3MP0?EgIGMUMfEbaf;7H/MRaf014000h/02`0;00/02`0
;00/00//0300;00/02`0MRaf;7H/MR`1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4HE`D0
0@4100003`000`400002011f01T/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Qf25MCMP07;7H/MRaf
;00H0008;00/02`0;009;00702`0;00/0002;00OMRaf;7H/MUL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0@0JE`07MUMfEgIGMP0500410@0000l000<100000P001eMfEgIGMUL04WH03baf;7H/MRaf;7H/MRaf
;009MPQG00QfEgIGMUMfEd=f00mGMUMfEgIG;7H/MRaf;7H04@003R`0;00/02`0;00/02`02b`0<00/
02`0;01f;7H/MRaf;05G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0AQG1@010@40000?0081
0`0@MP0I;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;008MPQGDgH01baf;7H/MR`0600022`0;00/02`0
2B`01`0/02`0;0000R`07gH/MRaf;7IG0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`406UL01gIGMUMfEgH0
10020@40000?00030@00008000MGMUMfEgIG019f00l/MRaf;7H/MRaf;7H/MR`02GH8E`08MUMfEgIG
MUM3MP0?EgIGMUMfEbaf;7H/MRaf014000h/02`0;00/02`0;00/00//0300;00/02`0MRaf;7H/MR`1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4HE`D00@4100003`000`400002011f030/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaGMUMfEgIGMP5G0EL1E`42E`06MUMfEgIGA7H01baf
;7H/MR`0600022`0;00/02`02B`800P/021G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0AUG
1@010@40000?00030@00008000mGMUMfEgIGMUMfEgIGMUL02WH03baf;7H/MRaf;7H/MRaf;00:MP06
EgIGMUMf0UL05`5G0EL1E`5fEgIGMUMfEgIGMUMfEgIG049f00P0;00/02`0;0T000h/02`0;00/02`0
;00/00//00L0;00/02`000P/00P1E`5G0EL1Ea4100QG0EL1E`5G0AQG1@010@40000?00030@000080
47H0<2af;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;5MfEgIGMUMf0EL1E`5G0@9G00IfEgIG
MUM4MP07;7H/MRaf;00H0008;00/02`0;009;0P022`085L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL16EL500410@0000l000<100000P003eMfEgIGMUMfEgIGMUMfE`0:MP0?;7H/MRaf;7H/MRaf
;7H/00Yf00IGMUMfEgH2E`0G0EL1E`5G0GIGMUMfEgIGMUMfEgIGMUL0@WH0200/02`0;00/2@003R`0
;00/02`0;00/02`02b`01`0/02`0;00022`0205G0EL1E`5G4@4025L1E`5G0EL165L500410@0000l0
00<100000P0@MP0`;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/EgIGMUMfEgH1E`5G0EL1
0UL01WIGMUMfEdAf00L/MRaf;7H/01P000P/02`0;00/00T/2008;00PE`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`4IE`D00@4100003`000`400002000?EgIGMUMfEgIGMUMfEgIG00Yf00l/MRaf
;7H/MRaf;7H/MR`02WH01UMfEgIGMP9G01L1E`5G0EL1MUMfEgIGMUMfEgIGMUMfE`12MP0802`0;00/
02`9000>;00/02`0;00/02`0;00;;00702`0;00/0008;0080EL1E`5G0ELA0@08E`5G0EL1E`4HE`D0
0@4100003`020@<047H0<2af;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;5MfEgIGMUMf0EL1
E`5G0@9G00IfEgIGMUM4MP07;7H/MRaf;00H0008;00/02`0;009;0P022`085L1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL16EL400810@0000l000<100000P003eMfEgIGMUMfEgIGMUMfE`0:MP0?
;7H/MRaf;7H/MRaf;7H/00Yf00IGMUMfEgH2E`0G0EL1E`5G0GIGMUMfEgIGMUMfEgIGMUL0@WH0200/
02`0;00/2@003R`0;00/02`0;00/02`02b`01`0/02`0;00022`0205G0EL1E`5G4@4025L1E`5G0EL1
65L500410@0000l000<100000P0@MP0`;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/EgIG
MUMfEgH1E`5G0EL12UL027IGMUMfEgIG@GH01b`0;00/02`04@004B`0;00/02`0;00/02`0;00/00T0
00L/02`0;00/00UG01L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@0IE`D00@4100003`000`400002000?
EgIGMUMfEgIGMUMfEgIG00Yf01L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;002MP07EgIGMUMfE`080@0P
E`5G0EL1E`5fEgIGMUMfEgIGMUMfEgIGMUMfEgIGMULbMPP/2@003R`0;00/02`0;00/02`02b`01`0/
02`0;00022`0205G0EL1E`5G4@4065L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@QG1@010@40000?0003
0@00008047H0<2af;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;5MfEgIGMUMf0EL1E`5G0@YG
00QfEgIGMUMfEd5f00L/02`0;00/0140014/02`0;00/02`0;00/02`0;0090007;00/02`0;009E`0G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`406EL500410@0000l000<100000P003eMfEgIGMUMfEgIGMUMf
E`0:MP0G;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`00WH01eMfEgIGMUL0204085L1E`5G0EL1MUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgIG<WH8;0T000h/02`0;00/02`0;00/00//00L0;00/02`000P/00P1E`5G
0EL1Ea4101QG0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`48E`D00@4100003`000`400002011f030/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaGMUMfEgIGMP5G0EL1E`4:E`08MUMfEgIGMUM1MP07
;00/02`0;00A000A;00/02`0;00/02`0;00/02`02@001b`0;00/02`02EL05`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL101UG1@010@40000?00030@00008000mGMUMfEgIGMUMfEgIGMUL02WH05baf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/009f00MGMUMfEgIG00P1021G0EL1E`5G0GIGMUMfEgIGMUMfEgIGMUMfEgIG
MUMfEc9f22`9000>;00/02`0;00/02`0;00;;00702`0;00/0008;0080EL1E`5G0ELA0@0HE`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL125L500410@0000l00P43011f030/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaGMUMfEgIGMP5G0EL1E`4:E`08MUMfEgIGMUM1MP07;00/02`0;00A000A;00/02`0
;00/02`0;00/02`02@001b`0;00/02`02EL05`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL101UG10020@40
000?00030@00008000mGMUMfEgIGMUMfEgIGMUL02WH05baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/009f
00MGMUMfEgIG00P1021G0EL1E`5G0GIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEc9f22`9000>;00/02`0
;00/02`0;00;;00702`0;00/0008;0080EL1E`5G0ELA0@0HE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
25L500410@0000l000<100000P0@MP0`;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/EgIG
MUMfEgH1E`5G0EL12UL027IGMUMfEgIG@GH01b`0;00/02`04@004B`0;00/02`0;00/02`0;00/00T0
00L/02`0;00/00UG01L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@0IE`D00@4100003`000`40000200Qf
00MGMUMfEgIG00Yf01L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00IMP09;7H/MRaf;7H/00Qf01mGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG00]f00mGMUMfEgIGMUMfEgIGMUL027H05`0/02`0;00/
02`0;00/02`0;00/02`000//00L0;00/02`000P/00QGMUMfEgIGMQT101QG0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`4500410@0000l000<100000P0HMP0I;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00AMP06
;7H/MRaf2b`01WH/MRaf;09f00mGMUMfEgIGMUMfEgIGMUL0:WH01Raf;7H/MP8/00H0;00/02`A00T/
01L0;00/02`0;00/02`0;00/MUMfEgIGMP02E`0O0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0@0AE`D00@4100003`000`40000200Qf00MGMUMfEgIG00Yf01L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;00IMP09;7H/MRaf;7H/00Qf01mGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG00]f00mGMUMf
EgIGMUMfEgIGMUL027H05`0/02`0;00/02`0;00/02`0;00/02`000//00L0;00/02`000P/00QGMUMf
EgIGMQT101QG0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4500410@0000l000<100000P0HMP0I;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;00AMP06;7H/MRaf2b`01WH/MRaf;09f00mGMUMfEgIGMUMfEgIGMUL0
:WH01Raf;7H/MP8/00H0;00/02`A00T/01L0;00/02`0;00/02`0;00/MUMfEgIGMP02E`0O0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@0AE`D00@41000000@00@4110020@D000<100000P08MP07
EgIGMUMfE`0:MP0G;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`06GH02Baf;7H/MRaf;008MP0OEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfE`0;MP0?EgIGMUMfEgIGMUMfEgIG00Qf01L0;00/02`0;00/
02`0;00/02`0;00/000;;00702`0;00/0008;008EgIGMUMfEgHI0@0HE`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL11@010@40000400040@000@8000<100000P000`40000201Qf01T/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/015f00H/MRaf;7H;;006MRaf;7H/0WH03eMfEgIGMUMfEgIGMUMfE`0ZMP06;7H/MRaf
0R`01P0/02`0;1402B`05`0/02`0;00/02`0;00/02afEgIGMUMf009G01l1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1015G1@010@40000400040@000@8000<100000P030@8027H01eMfEgIGMUL0
2WH05baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/01Uf00T/MRaf;7H/MR`027H07eMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUL02gH03eMfEgIGMUMfEgIGMUMfE`08MP0G02`0;00/02`0;00/02`0;00/
02`0;0002b`01`0/02`0;00022`025MfEgIGMUMf6@4065L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@@0
0P4100000P020@<000@1000110000`40000201Qf01T/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/015f
00H/MRaf;7H;;006MRaf;7H/0WH03eMfEgIGMUMfEgIGMUMfE`0ZMP06;7H/MRaf0R`01P0/02`0;140
2B`05`0/02`0;00/02`0;00/02afEgIGMUMf009G01l1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1015G1@010@40000400040@000@8000<100000P000`40000201Uf020/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MPT/011f;7H/MRaf;7H/MRaf;7H/47H0>EMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfE`08;00?02`0;00/02`0;00/02`001X/
27HQ0@0@E`5G0EL1E`5G0EL1E`5G0@D00@41000000@00@4110020@D000<100000P0HMP0H;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf6B`067H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;0Yf00mGMUMfEgIGMUMf
EgIGMUL08GH03b`0;00/02`0;00/02`0;00900T/0100;00/02`0;00/02`0;00/27H0<EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL01@010@40000?00030@000080
6GH082af;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf2B`047H/MRaf;7H/MRaf;7H/MR`@MP0i
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG00P/
00l0;00/02`0;00/02`0;0006R`8MR41011G0EL1E`5G0EL1E`5G0EL11@010@40000?00030@000080
67H062af;7H/MRaf;7H/MRaf;7H/MRaf;7H/MQT/01Qf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`:MP0?
EgIGMUMfEgIGMUMfEgIG025f00l/02`0;00/02`0;00/02`02@09;00@02`0;00/02`0;00/02`0;0Qf
035G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G00D00@410000
3`000`40000201Uf020/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MPT/011f;7H/MRaf;7H/
MRaf;7H/47H0>EMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfE`08;00?02`0;00/02`0;00/02`001X/27HQ0@0@E`5G0EL1E`5G0EL1E`5G0@D00@410000
3`000`40000201Qf01P/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7HI;00HMRaf;7H/MRaf;7H/MRaf;7H/
MRaf;7H/2WH03eMfEgIGMUMfEgIGMUMfE`0QMP0?;00/02`0;00/02`0;00/00T02B`0400/02`0;00/
02`0;00/02`8MP0aE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`0500410@0000l00P4301Uf020/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MPT/011f;7H/
MRaf;7H/MRaf;7H/47H0>EMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfE`08;00?02`0;00/02`0;00/02`001X/27HQ0@0@E`5G0EL1E`5G0EL1E`5G0@@0
0P4100003`000`40000201Qf01P/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7HI;00HMRaf;7H/MRaf;7H/
MRaf;7H/MRaf;7H/2WH03eMfEgIGMUMfEgIGMUMfE`0QMP0?;00/02`0;00/02`0;00/00T02B`0400/
02`0;00/02`0;00/02`8MP0aE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`0500410@0000l000<100000P001eMfEgIGMUL06WH05baf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/00Uf00MG0EL1E`5G018167H0<EMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIG;7H/MRaf;7H022`01`0/02`0;0006R`8MP080EL1E`5G0ELY0@D00@4100003`000`40000201Qf
010/MRaf;7H/MRaf;7H/MRaf2B`027H/MRaf;7H/27H8E`0A0EL1E`5G0EL1E`5G0EL1E`4027H01baf
;7H/MR`04WH085MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf2B`03`0/02`0;00/02`0;00/
000I;0Qf25LB0@0GE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL01@010@40000?00030@00008000MGMUMf
EgIG01Yf01L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;009MP07E`5G0EL1E`0B0AQf035GMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEbaf;7H/MRaf00P/00L0;00/02`001X/27H0205G
0EL1E`5G:@4500410@0000l000<100000P0HMP0@;7H/MRaf;7H/MRaf;7H/MPT/00Qf;7H/MRaf;0Qf
25L04@5G0EL1E`5G0EL1E`5G0EL100Qf00L/MRaf;7H/019f021GMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMPT/00l0;00/02`0;00/02`0;0006B`8MPQG4P405eL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G00D00@4100003`000`4000020007EgIGMUMfE`0JMP0G;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`0
2GH01eL1E`5G0EL04P4HMP0aEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUL/
MRaf;7H/MP08;00702`0;00/000J;0Qf00P1E`5G0EL1EbT11@010@40000?00030@00008067H042af
;7H/MRaf;7H/MRaf;7H9;008MRaf;7H/MR`8MPQG0141E`5G0EL1E`5G0EL1E`5G0@08MP07;7H/MRaf
;00BMP0PEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH9;00?02`0;00/02`0;00/02`001T/
27H8Ea8101MG0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`0500410@0000l000<100000P001eMfEgIGMUL0
6WH05baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf00MG0EL1E`5G018167H0<EMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG;7H/MRaf;7H022`01`0/02`0;0006R`8MP080EL1E`5G
0ELY0@D00@4100003`020@<067H042af;7H/MRaf;7H/MRaf;7H9;008MRaf;7H/MR`8MPQG0141E`5G
0EL1E`5G0EL1E`5G0@08MP07;7H/MRaf;00BMP0PEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgH9;00?02`0;00/02`0;00/02`001T/27H8Ea8101MG0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`040081
0@0000l000<10000o`0=00410@0000l000<10000o`0=00410@0000l000<1000090000`40000W0003
0@0002L000<100009`000`40000W00030@0002L000<100004`010@40000?00030@00008000<10000
1P000`40000500030@0000D000<100001P000`40000500030@0000H000<100001@000`4000050003
0@0000H000<100001@000`40000600030@0000D000<100001P000`40000500030@0000D000<10000
1P000`40000500030@0000H000<100001@000`40000600030@0000D000<100001@000`4000060003
0@0000D000<100001P000`40000500030@0000D000<100001P000`40000500030@0000H000<10000
1@000`40000200410@0000l0o`4A0@40003o02400000\
\>"],
  ImageRangeCache->{{{108, 395}, {509.312, 222.312}} -> {-14.274, 25.1236, 
  0.118757, 0.118757}}]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
DensityPlot[imageArrayTwo[Round[i],
                          Round[j]],
            {i,1,myImageSize[[1]]},{j,1,myImageSize[[2]]},
            {Mesh -> False, PlotPoints -> myImageSize[[2]]}] ;\
\>", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% DensityGraphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.0117866 0.0310174 -0.0117866 0.0310174 [
[.1433 -0.0125 -3 -9 ]
[.1433 -0.0125 3 0 ]
[.29839 -0.0125 -6 -9 ]
[.29839 -0.0125 6 0 ]
[.45347 -0.0125 -6 -9 ]
[.45347 -0.0125 6 0 ]
[.60856 -0.0125 -6 -9 ]
[.60856 -0.0125 6 0 ]
[.76365 -0.0125 -6 -9 ]
[.76365 -0.0125 6 0 ]
[.91873 -0.0125 -6 -9 ]
[.91873 -0.0125 6 0 ]
[ 0 0 -0.125 0 ]
[-0.0125 .1433 -6 -4.5 ]
[-0.0125 .1433 0 4.5 ]
[-0.0125 .29839 -12 -4.5 ]
[-0.0125 .29839 0 4.5 ]
[-0.0125 .45347 -12 -4.5 ]
[-0.0125 .45347 0 4.5 ]
[-0.0125 .60856 -12 -4.5 ]
[-0.0125 .60856 0 4.5 ]
[-0.0125 .76365 -12 -4.5 ]
[-0.0125 .76365 0 4.5 ]
[-0.0125 .91873 -12 -4.5 ]
[-0.0125 .91873 0 4.5 ]
[ 0 0 -0.125 0 ]
[ 0 1 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.1433 0 m
.1433 .00625 L
s
[(5)] .1433 -0.0125 0 1 Mshowa
.29839 0 m
.29839 .00625 L
s
[(10)] .29839 -0.0125 0 1 Mshowa
.45347 0 m
.45347 .00625 L
s
[(15)] .45347 -0.0125 0 1 Mshowa
.60856 0 m
.60856 .00625 L
s
[(20)] .60856 -0.0125 0 1 Mshowa
.76365 0 m
.76365 .00625 L
s
[(25)] .76365 -0.0125 0 1 Mshowa
.91873 0 m
.91873 .00625 L
s
[(30)] .91873 -0.0125 0 1 Mshowa
.125 Mabswid
.17432 0 m
.17432 .00375 L
s
.20533 0 m
.20533 .00375 L
s
.23635 0 m
.23635 .00375 L
s
.26737 0 m
.26737 .00375 L
s
.3294 0 m
.3294 .00375 L
s
.36042 0 m
.36042 .00375 L
s
.39144 0 m
.39144 .00375 L
s
.42246 0 m
.42246 .00375 L
s
.48449 0 m
.48449 .00375 L
s
.51551 0 m
.51551 .00375 L
s
.54653 0 m
.54653 .00375 L
s
.57754 0 m
.57754 .00375 L
s
.63958 0 m
.63958 .00375 L
s
.6706 0 m
.6706 .00375 L
s
.70161 0 m
.70161 .00375 L
s
.73263 0 m
.73263 .00375 L
s
.79467 0 m
.79467 .00375 L
s
.82568 0 m
.82568 .00375 L
s
.8567 0 m
.8567 .00375 L
s
.88772 0 m
.88772 .00375 L
s
.11228 0 m
.11228 .00375 L
s
.08127 0 m
.08127 .00375 L
s
.05025 0 m
.05025 .00375 L
s
.01923 0 m
.01923 .00375 L
s
.94975 0 m
.94975 .00375 L
s
.98077 0 m
.98077 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
0 .1433 m
.00625 .1433 L
s
[(5)] -0.0125 .1433 1 0 Mshowa
0 .29839 m
.00625 .29839 L
s
[(10)] -0.0125 .29839 1 0 Mshowa
0 .45347 m
.00625 .45347 L
s
[(15)] -0.0125 .45347 1 0 Mshowa
0 .60856 m
.00625 .60856 L
s
[(20)] -0.0125 .60856 1 0 Mshowa
0 .76365 m
.00625 .76365 L
s
[(25)] -0.0125 .76365 1 0 Mshowa
0 .91873 m
.00625 .91873 L
s
[(30)] -0.0125 .91873 1 0 Mshowa
.125 Mabswid
0 .17432 m
.00375 .17432 L
s
0 .20533 m
.00375 .20533 L
s
0 .23635 m
.00375 .23635 L
s
0 .26737 m
.00375 .26737 L
s
0 .3294 m
.00375 .3294 L
s
0 .36042 m
.00375 .36042 L
s
0 .39144 m
.00375 .39144 L
s
0 .42246 m
.00375 .42246 L
s
0 .48449 m
.00375 .48449 L
s
0 .51551 m
.00375 .51551 L
s
0 .54653 m
.00375 .54653 L
s
0 .57754 m
.00375 .57754 L
s
0 .63958 m
.00375 .63958 L
s
0 .6706 m
.00375 .6706 L
s
0 .70161 m
.00375 .70161 L
s
0 .73263 m
.00375 .73263 L
s
0 .79467 m
.00375 .79467 L
s
0 .82568 m
.00375 .82568 L
s
0 .8567 m
.00375 .8567 L
s
0 .88772 m
.00375 .88772 L
s
0 .11228 m
.00375 .11228 L
s
0 .08127 m
.00375 .08127 L
s
0 .05025 m
.00375 .05025 L
s
0 .01923 m
.00375 .01923 L
s
0 .94975 m
.00375 .94975 L
s
0 .98077 m
.00375 .98077 L
s
.25 Mabswid
0 0 m
0 1 L
s
.1433 .99375 m
.1433 1 L
s
.29839 .99375 m
.29839 1 L
s
.45347 .99375 m
.45347 1 L
s
.60856 .99375 m
.60856 1 L
s
.76365 .99375 m
.76365 1 L
s
.91873 .99375 m
.91873 1 L
s
.125 Mabswid
.17432 .99625 m
.17432 1 L
s
.20533 .99625 m
.20533 1 L
s
.23635 .99625 m
.23635 1 L
s
.26737 .99625 m
.26737 1 L
s
.3294 .99625 m
.3294 1 L
s
.36042 .99625 m
.36042 1 L
s
.39144 .99625 m
.39144 1 L
s
.42246 .99625 m
.42246 1 L
s
.48449 .99625 m
.48449 1 L
s
.51551 .99625 m
.51551 1 L
s
.54653 .99625 m
.54653 1 L
s
.57754 .99625 m
.57754 1 L
s
.63958 .99625 m
.63958 1 L
s
.6706 .99625 m
.6706 1 L
s
.70161 .99625 m
.70161 1 L
s
.73263 .99625 m
.73263 1 L
s
.79467 .99625 m
.79467 1 L
s
.82568 .99625 m
.82568 1 L
s
.8567 .99625 m
.8567 1 L
s
.88772 .99625 m
.88772 1 L
s
.11228 .99625 m
.11228 1 L
s
.08127 .99625 m
.08127 1 L
s
.05025 .99625 m
.05025 1 L
s
.01923 .99625 m
.01923 1 L
s
.94975 .99625 m
.94975 1 L
s
.98077 .99625 m
.98077 1 L
s
.25 Mabswid
0 1 m
1 1 L
s
.99375 .1433 m
1 .1433 L
s
.99375 .29839 m
1 .29839 L
s
.99375 .45347 m
1 .45347 L
s
.99375 .60856 m
1 .60856 L
s
.99375 .76365 m
1 .76365 L
s
.99375 .91873 m
1 .91873 L
s
.125 Mabswid
.99625 .17432 m
1 .17432 L
s
.99625 .20533 m
1 .20533 L
s
.99625 .23635 m
1 .23635 L
s
.99625 .26737 m
1 .26737 L
s
.99625 .3294 m
1 .3294 L
s
.99625 .36042 m
1 .36042 L
s
.99625 .39144 m
1 .39144 L
s
.99625 .42246 m
1 .42246 L
s
.99625 .48449 m
1 .48449 L
s
.99625 .51551 m
1 .51551 L
s
.99625 .54653 m
1 .54653 L
s
.99625 .57754 m
1 .57754 L
s
.99625 .63958 m
1 .63958 L
s
.99625 .6706 m
1 .6706 L
s
.99625 .70161 m
1 .70161 L
s
.99625 .73263 m
1 .73263 L
s
.99625 .79467 m
1 .79467 L
s
.99625 .82568 m
1 .82568 L
s
.99625 .8567 m
1 .8567 L
s
.99625 .88772 m
1 .88772 L
s
.99625 .11228 m
1 .11228 L
s
.99625 .08127 m
1 .08127 L
s
.99625 .05025 m
1 .05025 L
s
.99625 .01923 m
1 .01923 L
s
.99625 .94975 m
1 .94975 L
s
.99625 .98077 m
1 .98077 L
s
.25 Mabswid
1 0 m
1 1 L
s
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of gray image
p
.01923 .01923 translate
.96154 .96154 scale
32 string
32 32 8 [32 0 0 32 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  Mimage
17272423302F3134363A5572553D40434D3B2F3B402E2923
232318191B14100F
2022272234342F3739394E5D59404C5D5B52504C48403A34
322A161219141417
28222927383A2B3D413D3E555D52636E6D6D67625D56534F
433B2F1914161B1C
2225232C33402E3B45394A5E6C767D7B7B7A77746F67625D
564F473626231915
2122242C2E413B34495D6F7A85878787808685837A75736C
655F55493A190E11
22232220232D3449627484909188888F8C8780847E757672
6D695C544F2E1817
1C202125252A4B6C78858E94928E8B8B8E82727475686A6E
6D62625B57461E1E
2229282B2E4E727E879195969596A49383787C6E66585967
6A5E5E5D574D2716
1E2729264A757E8A939A91969897C9B67988D7995E504753
615A5A5B59533616
1C23243270868B90A0CDA38C9E99CBCC8387E5B462554F75
6359575759544318
1E202751868E979A9FE1C18A9897CADA927BDDC45D4D93EA
88504F565C574F24
182D3B729193999896CCD895979ED0EFB178E0E66145B5FC
934A4B505A615335
112A428594959B999BB8F4B4939FC7F0CE86E1F57438C3F9
7B3D434A5B615545
1E25468C95979B9E9FA7F1DDA49AADF2F395CEFF8145E5F2
6837424B5D625C56
1D204E9299999FA09F9AC8F9DDAF95D5FCBDCDFE9C77FAE3
4A323E4B5C66625A
1824469195959EA1A09EA2E1F7D9C1D7F5F3F3F4EBDEF9D0
382F37485B65665D
3839478F9898A0A2A1A49FC0F7F0DEE7F3F4F4F3F5F8ECA8
342F344658646A64
444D578C9D98918D8696A6ADEAF6EEDFF0F5F2F3F3F4F2A4
322D3546585E6565
3B5260859CC0CCCCB58D8CA5D4F8E7C8C6EFF3F3F3F5EDA1
322B34485B5F6465
26395B7B97A5B9D5F4DA9693B2EAEFECCAC8F5F5F3F4E7A9
392B384C5B636669
48405C7D91949593B4F2DDB9B5C4BBC2CEB5CEF1F4F6D899
3B2B384C59626564
878281898F95969A91C0F8E8CBC7B9A49D9F9ED1F5F7D282
3729374C595D6162
899C9A969591919499A3E6F8F0F0E9CCAFA39FB2ECF8D273
2F27334C595F6462
77909699998F9094948F9AC2EAF9F7EED6B9B0AFD5F7C769
2826304651575C5E
6F8D9898928F8A88898E8A8796C7F1FAE9D6D0BDD5E2AD6E
23242B3D4C535B60
6B85979B977E6883868C8A8E88849DC4DFE3EBCFC7C9CA76
202226374955595A
6D7C969D9F84466C8183848688878177B9F9F6E6CBBEDF94
1F21232E434D5155
7378939EA3965B2C67787A7D8184847B95EFF6E9D1BCE0B1
241D202639494C4D
75778FA0A7A15F1C35676A76807E807C7BD1F9F0DBC2EAC7
39181E2231434447
7B77889FA5A7807E98AD8E63656E767A7496E7F0E4BEDDD6
5F17181A27353939
7E7B819AA5A8ADB2AB9C90946E595D6A7173C1E2ECC0C7CF
8628111318252A26
767B7D94A5ABA88938141683988171595859A2D2D9C3BABE
93460D0F10151916
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000180P10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdoojn_l00;34gTJ2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_[C^]:JVYZf][Jo^kP00
oiT0IP3o0020P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;0000?`00O`00_`0o000o
?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0
_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo?cooOcoo
_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no?gnoOgno
_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo
_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0_ol0oolo0?lo?oloOolo_olooomo
0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_mWEc:0Pl=WIfJj^[WMgMiW<09W<<iW<
IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`c
c<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co
<lcoI/coVLcoc<coool0<ol0I_l0VOl0c?l0oolc0?lc<olcI_lcVOlcc?lcoomV0?mVI_mVVOmVoonI
0?nI<onII_nIVOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<ool08@000?l08@000?l08@000?l08@0003@0
0`4T00<110020B400`4300<18@040@<00P4Q00@10P030B400`440081500003L000<100008P000`40
000200040@000B4000<100001@000`40000O00030@00008000@100018@000`40000500030@000240
00@100010P000`40000A0000=`000`40000R00030@00008000@100018@000`40000500030@000200
00D100000@0200030@00020000<1000010000`40000Q00040@000@8000<100004@0003@00`4U0003
0@00008000@100018@000`40000200<190001040004200030@00024000<100000`4R00810`001040
004C0000=0000`40000U00030@00008000@100018@000`40000200030@00024000@100010P001040
004P00040@000@8000<1000090001040004200030@000140000d00@18`020@D00P4Q008110040B40
0P4400818P020@<0104P00<110020A@0003o0240003o0240003o0240003o0240000?0?l14@410000
3`000`40000200030@0000H000<100001@000`40000500030@0000H000<100001@000`4000060003
0@0000D000<100001@000`40000600030@0000D000<100001P000`40000500030@0000H000<10000
1@000`40000500030@0000H000<100001@000`40000600030@0000D000<100001P000`4000050003
0@0000D000<100001P000`40000500030@0000H000<100001@000`40000500030@0000H000<10000
1@000`40000600030@0000D000<100000P010@40000?00030@0002@000<100009`000`40000W0003
0@0002L000<100009`000`40000W00030@0001<00@4100003`000`40003o00d00@4100003`000`40
003o00d00@4100003`000`40003o00d00@4100003`020Ol03@020@40000?00030@000080021G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@UG00L1E`5G0EL1029G00QfEgIGMUMfE`Uf00IGMUMf
EgHZE`070EL1E`5G0@0AE`1B0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`P11@010@40000?00030@000080
2@40B5L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0@YG00MfEgIGMUMf00QG00L1E`5G0EL101]G01L1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0@09E`0X0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1Ec411@010@40
000?00030@000080021G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@UG00L1E`5G0EL1029G
00QfEgIGMUMfE`Uf00IGMUMfEgHZE`070EL1E`5G0@0AE`1B0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`P1
1@010@40000?00030@0000802@40B5L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@YG00MfEgIGMUMf00QG00L1E`5G0EL101]G
01L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@09E`0X0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1Ec411@010@40000?00030@000080021G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0@UG00L1E`5G0EL1029G00QfEgIGMUMfE`Uf00IGMUMfEgHZE`070EL1E`5G0@0AE`1B0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`P11@010@40000?00030@0000802@40B5L1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@YG00MfEgIG
MUMf00QG00L1E`5G0EL101]G01L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@09E`0X0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1Ec411@010@40000?00030@000080021G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@UG00L1E`5G0EL1029G00QfEgIGMUMfE`Uf00IGMUMfEgHZE`07
0EL1E`5G0@0AE`1B0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`P11@010@40000?00810`090@18E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL12UL01gIGMUMfEgH025L01`5G0EL1E`406eL05`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL100UG
02P1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G<@4400810@0000l000<10000
0P0085L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL14EL01`5G0EL1E`408UL03gIGMUMfEgIG
MUMfEgIGMP0AE`0QMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf039G03P1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`D00@4100003`000`40
0002001A0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL100YG00mfEgIGMUMfEgIGMUMfEgH04EL03gIGMUMfEgIG
MUMfEgIGMP0ZE`0P0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0ELa0@D00@4100003`000`40
0002000PE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4AE`070EL1E`5G0@0RE`0?MUMfEgIG
MUMfEgIGMUMf015G025fEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH0<UL0>05G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G1@010@40000?0003
0@0000800541E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`402UL03gIGMUMfEgIGMUMfEgIGMP0AE`0?MUMfEgIG
MUMfEgIGMUMf02YG0201E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1Ec411@010@40000?0003
0@000080021G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0A5G00L1E`5G0EL1029G00mfEgIG
MUMfEgIGMUMfEgH04EL08GIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP0bE`0h0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL500410@0000l0
00<100000P00D@5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@0:E`0?MUMfEgIGMUMfEgIGMUMf015G00mfEgIG
MUMfEgIGMUMfEgH0:UL0805G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G<@4500410@0000l0
00<100000P0085L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL14EL01`5G0EL1E`408UL03gIG
MUMfEgIGMUMfEgIGMP0AE`0QMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf039G03P1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`D00@410000
3`000`400002001A0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL100YG00mfEgIGMUMfEgIGMUMfEgH04EL03gIG
MUMfEgIGMUMfEgIGMP0ZE`0P0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0ELa0@D00@410000
3`020@<0021G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0A5G00L1E`5G0EL1029G021fEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEaUf02QGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMf6UL0<05G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`@00P4100003`000`40000200110EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`402EL03`5G0EL1E`5G0EL1E`5G0@0:E`07MUMfEgIG
MP09E`0gMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
MP0RE`0@0EL1E`5G0EL1E`5G0EL1EbT11@010@40000?00030@000080021G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0A5G00L1E`5G0EL1029G021fEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
MUMfEaUf02QGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf6UL0<05G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`D00@4100003`000`4000020011
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`402EL03`5G0EL1E`5G0EL1E`5G0@0:E`07MUMfEgIGMP09E`0gMUMfEgIGMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP0RE`0@0EL1E`5G0EL1E`5G0EL1EbT1
1@010@40000?00030@000080021G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0A5G00L1E`5G
0EL1029G021fEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEaUf02QGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf6UL0<05G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`D00@4100003`000`40000200110EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`402EL03`5G0EL1E`5G0EL1E`5G
0@0:E`07MUMfEgIGMP09E`0gMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMP0RE`0@0EL1E`5G0EL1E`5G0EL1EbT11@010@40000?00030@000080021G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0A5G00L1E`5G0EL1029G021fEgIGMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEaUf02QGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
6UL0<05G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`D00@410000
3`000`40000200110EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`402EL03`5G0EL1E`5G0EL1E`5G0@0:E`07MUMfEgIGMP09E`0gMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP0RE`0@0EL1E`5G
0EL1E`5G0EL1EbT11@010@40000?00810`0085L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
4EL01`5G0EL1E`408UL027IGMUMfEgIGBGH085MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
6UL0805G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G10020@40000?00030@00008002T1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@09E`0?0EL1E`5G0EL1E`5G0EL100UG
00L1E`5G0EL100YG01IfEgIGMUMfEgIGMUMfEgIGMUMfEgIG8gH0<5MfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMQQG01P1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0ELA0@D0
0@4100003`000`400002000PE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4AE`070EL1E`5G
0@0RE`08MUMfEgIGMUM9MP0PEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgHJE`0P0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL500410@0000l000<100000P00:@5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL100UG00l1E`5G0EL1E`5G0EL1E`402EL01`5G0EL1E`40
2UL05WIGMUMfEgIGMUMfEgIGMUMfEgIGMULSMP0`EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMf65L0605G0EL1E`5G0EL1E`5G0EL1E`5G0EL1Ea411@010@40000?0003
0@000080021G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0A5G00L1E`5G0EL1029G00QfEgIG
MUMfEdUf021GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMQYG0201E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`D00@4100003`000`400002000Y0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`402EL03`5G0EL1E`5G0EL1E`5G0@09E`070EL1E`5G0@0:E`0FMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEb=f031GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgHHE`0H0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G4@4500410@0000L00`4500030@000080
021G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0A5G00L1E`5G0EL1029G00QfEgIGMUMfEdUf
021GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMQYG0201E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`D00@4100002P000`40000200030@00008002T1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0@09E`0?0EL1E`5G0EL1E`5G0EL100UG00L1E`5G0EL100YG01IfEgIG
MUMfEgIGMUMfEgIGMUMfEgIG8gH0<5MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMQQG01P1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0ELA0@D00@4100002P000`40000200<1
0P0085L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL14EL01`5G0EL1E`408UL027IGMUMfEgIG
BGH085MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf6UL0805G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G10020@40000700<11@000`400002000Y0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`402EL03`5G0EL1E`5G0EL1E`5G0@08E`0@MUMfEgIGMUMfEgIGMUMfEe=f
02QGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf45L0205G0EL1E`5G6@450041
0@0000L000<100001@000`400002000XE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0B9G00QfEgIGMUMfEg=f01IGMUMfEgIGMUMfEgIGMUMfEgIGMUMf4UL0205G0EL1E`5G2@401eL1
E`5G0EL01@010@40000700@110000`400002000Y0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`402EL03`5G0EL1E`5G0EL1E`5G0@08E`0@MUMfEgIGMUMfEgIGMUMfEe=f02QGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf45L0205G0EL1E`5G6@4500410@0000l0
00<100000P00:5L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4RE`08MUMfEgIG
MUMcMP0FEgIGMUMfEgIGMUMfEgIGMUMfEgIGMQ9G00P1E`5G0EL1E`T100MG0EL1E`5G00D00@410000
3`000`400002000Y0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`402EL03`5G
0EL1E`5G0EL1E`5G0@08E`0@MUMfEgIGMUMfEgIGMUMfEe=f02QGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMf45L0205G0EL1E`5G6@4500410@0000l000<100000P00:5L1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4RE`08MUMfEgIGMUMcMP0FEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMQ9G00P1E`5G0EL1E`T100MG0EL1E`5G00D00@4100003`000`400002000Y0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`402EL03`5G0EL1E`5G0EL1E`5G0@08E`0@
MUMfEgIGMUMfEgIGMUMfEe=f02QGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
45L0205G0EL1E`5G6@4500410@0000l000<100000P00:5L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`4RE`08MUMfEgIGMUMcMP0FEgIGMUMfEgIGMUMfEgIGMUMfEgIGMQ9G00P1E`5G
0EL1E`T100MG0EL1E`5G00D00@4100003`020@<003T1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4025L047IGMUMfEgIGMUMfEgIGMUMKMP0`EgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf4UL01P5G0EL1Ea4110020@40
000?00030@000080031G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`4BE`06MUMfEgIG4WH0<Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MR`0BWH03UMfEgIGMUMfEgIGMUMf2UL0605G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`D00@410000
3`000`400002000i0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL100QG011fEgIGMUMfEgIGMUMfEgIGFgH0<5MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMQ9G00H1E`5G0ELA0@D00@4100003`000`400002000`E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL14UL01WIGMUMfEa9f034/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/04Yf00iGMUMfEgIGMUMf
EgIGMPYG01P1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL500410@0000l000<100000P00>@5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@08E`0@MUMfEgIG
MUMfEgIGMUMfEe]f031GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgHBE`060EL1E`5G4@4500410@0000l000<100000P00<5L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0A9G00IfEgIGMULBMP0a;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;01:MP0>EgIGMUMfEgIGMUMfEgH:E`0H0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G1@010@40000?00030@00008003T1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4025L047IGMUMfEgIGMUMfEgIGMUMKMP0`EgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf4UL01P5G0EL1Ea411@010@40
000?00030@000080031G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`4BE`06MUMfEgIG4WH0<Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MR`0BWH03UMfEgIGMUMfEgIGMUMf2UL0605G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`D00@410000
3`020@<02@40:5L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`48E`08MUMfEgIG
MUMCMP1@EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgHAEa0110020@40000?00030@000080031G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`49EaUf03T/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`0@WH085MfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMf25L0405G0EL1E`5G0EL1E`5G0EL500410@0000l000<100000P090@0X
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@QG00QfEgIGMUMfEe=f051GMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMQ5G404500410@0000l000<100000P00<5L1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@UG6GH0>Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;012MP0PEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgH8E`0@0EL1E`5G0EL1E`5G0EL1E`D00@4100003`000`40000200T102QG0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL125L027IGMUMfEgIGDgH0D5MfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMf4EL@0@D00@4100003`000`400002000`E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL12ELIMP0i;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/049f021GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMPQG
0101E`5G0EL1E`5G0EL1E`5G1@010@40000?00030@0000802@40:5L1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`48E`08MUMfEgIGMUMCMP1@EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgHAEa01
1@010@40000?00030@000080031G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`49EaUf03T/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MR`0@WH085MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf25L0405G0EL1E`5G
0EL1E`5G0EL500410@0000l000<100000P00:@5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL100QG00QfEgIGMUMfEcYf00L/MRaf;7H/029f04QGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgHBE`060EL1E`5G
2@4500410@0000l00P43000XE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@UG
6GH0>Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;00PMP0QEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG00Uf021GMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMPQG0101E`5G0EL1E`5G0EL1E`5G10020@40000?00030@00008002T1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@08E`08MUMfEgIGMULjMP07;7H/MRaf
;00RMP18EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMf4UL01P5G0EL1E`T11@010@40000?00030@00008002QG0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL12ELIMP0i;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/021f025GMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUL02GH085MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf25L0405G0EL1E`5G
0EL1E`5G0EL500410@0000l000<100000P00:@5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL100QG00QfEgIGMUMfEcYf00L/MRaf;7H/029f04QGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgHBE`060EL1E`5G
2@4500410@0000l000<100000P00:5L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`49EaUf03T/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MR`087H08EMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfE`09MP0PEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgH8E`0@0EL1E`5G0EL1E`5G0EL1E`D00@4100003`000`400002000Y
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4025L027IGMUMfEgIG>WH01baf
;7H/MR`08WH0B5MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMQ9G00H1E`5G0EL90@D00@4100003`000`400002000XE`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@UG6GH0>Baf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00PMP0QEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIG00Uf021GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMPQG0101E`5G
0EL1E`5G0EL1E`5G1@010@40000?00030@0000802@4065L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@QG
00QfEgIGMUMfEaQf00T/MRaf;7H/MR`04GH01baf;7H/MR`027HA;11f00L0;00/02`0008/00If;7H/
MR`2MP06EgIGMUMf6eL09gIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP0:E`06
0EL1E`5G2@4500410@0000l00P43000PE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`49Ea5f
03l/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MP0/
02`0;0002R`9MP06;7H/MRaf0R`07`0/02`0;01f;7H/MRaf;5MfEgIGMUMfEgIGMUMfEgH02EL0>GIG
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP08E`08
0EL1E`5G0EL400810@0000l000<100000P090@0HE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL125L027IG
MUMfEgIG67H02Baf;7H/MRaf;00AMP07;7H/MRaf;008MQ4/47H01`0/02`0;0000R`01WH/MRaf;09f
00IGMUMfEgHKE`0WMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf00YG00H1E`5G
0EL90@D00@4100003`000`400002000PE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`49Ea5f
03l/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MP0/
02`0;0002R`9MP06;7H/MRaf0R`07`0/02`0;01f;7H/MRaf;5MfEgIGMUMfEgIGMUMfEgH02EL0>GIG
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP08E`08
0EL1E`5G0EL500410@0000l000<100000P090@0HE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL125L027IG
MUMfEgIG67H02Baf;7H/MRaf;00AMP07;7H/MRaf;008MQ4/47H01`0/02`0;0000R`01WH/MRaf;09f
00IGMUMfEgHKE`0WMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf00YG00H1E`5G
0EL90@D00@4100003`000`400002000PE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`49Ea5f
03l/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MP0/
02`0;0002R`9MP06;7H/MRaf0R`07`0/02`0;01f;7H/MRaf;5MfEgIGMUMfEgIGMUMfEgH02EL0>GIG
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP08E`08
0EL1E`5G0EL500410@0000l000<100000P090@0HE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL125L027IG
MUMfEgIG67H02Baf;7H/MRaf;00AMP07;7H/MRaf;008MQ4/47H01`0/02`0;0000R`01WH/MRaf;09f
00IGMUMfEgHKE`0WMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf00YG00H1E`5G
0EL90@D00@41000000@00@4110020@D000<100000P0085L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL12ELAMP0o;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H0;00/02`000X/2GH01Raf;7H/MP8/01l0;00/02`0MRaf;7H/MRaGMUMfEgIGMUMfEgIG
MUMf00UG03UfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
MUMfEgH025L0205G0EL1E`5G1@010@40000200030@00008000@1000110000`40000200T1021G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5fEgIGMUMfEaQf00P/MRaf;7H/MPT/00Qf;7H/MRaf;0Uf00h/MRaf
;7H/MRaf;7H/MQ8/47H01`0/02`0;0002B`01gIGMUMfEgH04eL05gIGMUMfEgIGMUMfEgIGMUMfEgIG
MUMf015G00MfEgIGMUMf015G204500410@00008000<100000P001040004400<10P0065L1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0@UG4GH05Raf;7H/MRaf;7H/MRaf;7H/MRaf;7H2;00702`0;00/0002;00`
MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MP0/02`0;00/02`0;00/02`047H800P/011GMUMf
EgIGMUMfEgIGMUMf2EL9MP0`EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMf25L0205G0EL1E`5G10020@40000200030@00008000@1000110000`40000200T1021G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5fEgIGMUMfEaQf00P/MRaf;7H/MPT/00Qf;7H/MRaf;0Uf00h/MRaf
;7H/MRaf;7H/MQ8/47H01`0/02`0;0002B`01gIGMUMfEgH04eL05gIGMUMfEgIGMUMfEgIGMUMfEgIG
MUMf015G00MfEgIGMUMf015G204500410@00008000<100000P001040004400030@00008001QG0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`49Ea5f01H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf0R`01`0/02`0;000
0R`0<7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H0;00/02`0;00/02`0;00/011f2008;00@
EgIGMUMfEgIGMUMfEgIGMPUG2GH0<5MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMPQG00P1E`5G0EL1E`D00@41000000<00@401@020@D000<100000P090@0PE`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1MUMfEgIGMULHMP08;7H/MRaf;7H9;008MRaf;7H/MR`9MP0>;7H/MRaf
;7H/MRaf;7HB;11f00L0;00/02`000T/00MfEgIGMUMf01=G01MfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
MP0AE`07MUMfEgIGMP0AE`P11@010@40000?00030@00008001QG0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`49Ea5f01H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf0R`01`0/02`0;0000R`0<7H/MRaf;7H/MRaf;7H/
MRaf;7H/MRaf;7H/MRaf;7H0;00/02`0;00/02`0;00/011f2008;00@EgIGMUMfEgIGMUMfEgIGMPUG
2GH0<5MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMPQG00P1E`5G
0EL1E`D00@4100003`000`40000200T1021G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5fEgIGMUMfEaQf
00P/MRaf;7H/MPT/00Qf;7H/MRaf;0Uf00h/MRaf;7H/MRaf;7H/MQ8/47H01`0/02`0;0002B`01gIG
MUMfEgH04eL05gIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf015G00MfEgIGMUMf015G204500410@0000l0
00<100000P0065L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@UG4GH05Raf;7H/MRaf;7H/MRaf;7H/MRaf
;7H2;00702`0;00/0002;00`MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MP0/02`0;00/02`0
;00/02`047H800P/011GMUMfEgIGMUMfEgIGMUMf2EL9MP0`EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf25L0205G0EL1E`5G1@010@40000?00030@0000802@4085L1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0GIGMUMfEgIG67H022af;7H/MRaf2B`027H/MRaf;7H/2GH03Raf
;7H/MRaf;7H/MRaf4R`@MP0702`0;00/0009;007MUMfEgIGMP0CE`0GMUMfEgIGMUMfEgIGMUMfEgIG
MUMfEgH04EL01gIGMUMfEgH04EL80@D00@4100003`000`400002000PE`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1MUMfEgIGMUL:MP0O;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00:0006;00/02`0
0R`0<7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf02`0;00/02`0;00/02`0;01f;7H/MRaf;0Qf01P/02`0
;00/02`0;00/02`0EgIGMUMfEgH8E`09MRaf;7H/MRaf00P000L/MRaf;7H/009f00IGMUMfEgH9E`0I
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP08E`080EL1E`5G0EL500410@0000l00P4300T100iG0EL1
E`5G0EL1E`5G0@YG67H06Baf;7H/MRaf;7H/MRaf;7H0;00/02`0;00022`9MP07;7H/MRaf;008MPT/
00P0;00/02`0;11f00L0;00/02`000T/00MfEgIGMUMf00UG2GH022`0;00/02`027HIE`07MUMfEgIG
MP0BE`070EL1E`5G0@0400810@0000l000<100000P0085L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0GIG
MUMfEgIG2WH07baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02P001R`0;00/008/031f;7H/
MRaf;7H/MRaf;7H/MRaf;7H/MP0/02`0;00/02`0;00/02`0MRaf;7H/MR`8MP0H;00/02`0;00/02`0
;00/05MfEgIGMUMf25L02GH/MRaf;7H/MP080007;7H/MRaf;002MP06EgIGMUMf2EL06GIGMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgH025L0205G0EL1E`5G1@010@40000?00030@0000802@403UL1E`5G0EL1
E`5G0EL12ULHMP0I;7H/MRaf;7H/MRaf;7H/MP0/02`0;00/0008;0Uf00L/MRaf;7H/00Qf2B`0200/
02`0;00/47H01`0/02`0;0002B`01gIGMUMfEgH02EL9MP08;00/02`0;008MQUG00MfEgIGMUMf019G
00L1E`5G0EL100D00@4100003`000`400002000PE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1MUMfEgIG
MUL:MP0O;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00:0006;00/02`00R`0<7H/MRaf;7H/
MRaf;7H/MRaf;7H/MRaf02`0;00/02`0;00/02`0;01f;7H/MRaf;0Qf01P/02`0;00/02`0;00/02`0
EgIGMUMfEgH8E`09MRaf;7H/MRaf00P000L/MRaf;7H/009f00IGMUMfEgH9E`0IMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMP08E`080EL1E`5G0EL500410@0000l000<100000P090@0>E`5G0EL1E`5G0EL1
E`4:EaQf01T/MRaf;7H/MRaf;7H/MRaf02`0;00/02`000P/2GH01baf;7H/MR`027H9;00802`0;00/
02`@MP0702`0;00/0009;007MUMfEgIGMP09E`Uf00P/02`0;00/00Qf6EL01gIGMUMfEgH04UL01`5G
0EL1E`401@010@40000?00030@000080021G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5fEgIGMUMfE`Yf
01l/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00X000H/02`0;002;00`MRaf;7H/MRaf;7H/
MRaf;7H/MRaf;7H0;00/02`0;00/02`0;00/07H/MRaf;7H/27H062`0;00/02`0;00/02`0;01GMUMf
EgIGMPQG00Uf;7H/MRaf;7H020001baf;7H/MR`00WH01UMfEgIGMPUG01UfEgIGMUMfEgIGMUMfEgIG
MUMfEgIGMUMf00QG00P1E`5G0EL1E`D00@4100003`000`40000200T100iG0EL1E`5G0EL1E`5G0@YG
67H06Baf;7H/MRaf;7H/MRaf;7H0;00/02`0;00022`9MP07;7H/MRaf;008MPT/00P0;00/02`0;11f
00L0;00/02`000T/00MfEgIGMUMf00UG2GH022`0;00/02`027HIE`07MUMfEgIGMP0BE`070EL1E`5G
0@0500410@0000l000<100000P0045L1E`5G0EL1E`5G0EL1E`48E`Yf02H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MP8/00l0;00/02`0;00/02`0;0000R`07gH/MRaf;7H/MRaf;7H/
MRaf;7H/MRaf02`0;00/02`02@08;0Qf400025MfEgIGMUMf25L9;0P000P/MRaf;7H/MQ1G025fEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH025L500410@0000l00P4300T100iG0EL1E`5G0EL1
E`5G0@9G00QfEgIGMUMfEa1f00l/MRaf;7H/MRaf;7H/MR`02GH:;00702`0;00/000AMP06;7H/MRaf
0R`03`0/02`0;00/02`0;00/0009;0Qf01L0;00/02`0;00/02`0;00/MUMfEgIGMP09E`T/2008MQUG
00mfEgIGMUMfEgIGMUMfEgH02UL01`5G0EL1E`4010020@40000?00030@000080011G0EL1E`5G0EL1
E`5G0EL125L:MP0V;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H2;00?02`0;00/
02`0;00/02`0008/01mf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MP0/02`0;00/00T022`8MQ0000QGMUMf
EgIGMPQG2B`80008;7H/MRaf;7H@E`0QMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf00QG
1@010@40000?00030@0000802@403UL1E`5G0EL1E`5G0EL10UL027IGMUMfEgIG47H03baf;7H/MRaf
;7H/MRaf;009MPX/00L0;00/02`0015f00H/MRaf;7H2;00?02`0;00/02`0;00/02`000T/27H05`0/
02`0;00/02`0;00/02afEgIGMUMf00UG2B`800Qf6EL03gIGMUMfEgIGMUMfEgIGMP0:E`070EL1E`5G
0@0500410@0000l000<100000P0045L1E`5G0EL1E`5G0EL1E`48E`Yf02H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MP8/00l0;00/02`0;00/02`0;0000R`07gH/MRaf;7H/MRaf;7H/
MRaf;7H/MRaf02`0;00/02`02@08;0Qf400025MfEgIGMUMf25L9;0P000P/MRaf;7H/MQ1G025fEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH025L500410@0000l000<100000P090@0>E`5G0EL1
E`5G0EL1E`42E`08MUMfEgIGMUL@MP0?;7H/MRaf;7H/MRaf;7H/00Uf2R`01`0/02`0;0004GH01Raf
;7H/MP8/00l0;00/02`0;00/02`0;0002B`8MP0G02`0;00/02`0;00/02`0;7IGMUMfEgH02EL9;0P0
27HIE`0?MUMfEgIGMUMfEgIGMUMf00YG00L1E`5G0EL100D00@4100003`000`400002000@E`5G0EL1
E`5G0EL1E`5G0@QG2WH09Raf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf0R`03`0/
02`0;00/02`0;00/0002;00OMRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H0;00/02`0;00900P/27H@0008
EgIGMUMfEgH8E`T/200022af;7H/MRaf45L08GIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
MP08E`D00@4100003`000`40000200T100iG0EL1E`5G0EL1E`5G0@9G00QfEgIGMUMfEa1f00l/MRaf
;7H/MRaf;7H/MR`02GH:;00702`0;00/000AMP06;7H/MRaf0R`03`0/02`0;00/02`0;00/0009;0Qf
01L0;00/02`0;00/02`0;00/MUMfEgIGMP09E`T/2008MQUG00mfEgIGMUMfEgIGMUMfEgH02UL01`5G
0EL1E`401@010@40000?00030@000080011G0EL1E`5G0EL1E`5G0EL125L:MP0V;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H:;0P02B`05gH/MRaf;7H/MRaf;7H/MP0/02`0;00/00T0
00P/02`0;00/00Qf4008MPQG00T/02`0;00/02`02008MQQG01UfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
MUMf00QG1@010@40000?00030@0000802@401UL1E`5G0@UG6GH062af;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MPT/2008;0Uf00H/MRaf;7H9;0T022`8MP0802`0;00/02`80007MUMfEgIGMP02E`070EL1E`5G
0@09;0P027H01`5G0EL1E`404UL03gIGMUMfEgIGMUMfEgIGMP0AE`D00@4100003`020@<0011G0EL1
E`5G0EL1E`5G0EL125L:MP0V;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H:;0P0
2B`05gH/MRaf;7H/MRaf;7H/MP0/02`0;00/00T000P/02`0;00/00Qf4008MPQG00T/02`0;00/02`0
2008MQQG01UfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf00QG10020@40000?00030@0000802@401UL1
E`5G0@UG6GH062af;7H/MRaf;7H/MRaf;7H/MRaf;7H/MPT/2008;0Uf00H/MRaf;7H9;0T022`8MP08
02`0;00/02`80007MUMfEgIGMP02E`070EL1E`5G0@09;0P027H01`5G0EL1E`404UL03gIGMUMfEgIG
MUMfEgIGMP0AE`D00@4100003`000`400002000@E`5G0EL1E`5G0EL1E`5G0@QG2WH09Raf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf2R`800T/01Mf;7H/MRaf;7H/MRaf;7H0;00/02`0
;0090008;00/02`0;008MQ0027H8E`09;00/02`0;00/00P027HHE`0IMUMfEgIGMUMfEgIGMUMfEgIG
MUMfEgIGMP08E`D00@4100003`000`40000200T100IG0EL1E`49EaUf01P/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H9;0P022`9MP06;7H/MRaf2B`900P/27H0200/02`0;00/20001gIGMUMfEgH00UL01`5G
0EL1E`402B`800Qf00L1E`5G0EL1019G00mfEgIGMUMfEgIGMUMfEgH04EL500410@0000l000<10000
0P0045L1E`5G0EL1E`5G0EL1E`48E`Yf02H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MPX/2009;00GMRaf;7H/MRaf;7H/MRaf02`0;00/02`02@0022`0;00/02`027H@00Qf25L02B`0
;00/02`0;00800Qf65L06GIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH025L500410@0000l000<10000
0P090@06E`5G0EL12ELIMP0H;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf2B`800P/2GH01Raf;7H/MPT/
2@08;0Qf00P0;00/02`0;0P000MfEgIGMUMf009G00L1E`5G0EL100T/2008MP070EL1E`5G0@0BE`0?
MUMfEgIGMUMfEgIGMUMf015G1@010@40000?00030@000080011G0EL1E`5G0EL1E`5G0EL125L0>Raf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H90006
;00/02`00R`03gH/MRaf;7H/MRaf;7H/MP08;14000Mf;7H/MRaf008/00H0;00/02`900Qf25LA00Qf
65L08GIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP0500410@0000l000<100000P090@06
E`5G0EL12ELIMP0Q;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00P0020/02`0;00/07H/
MRaf;7H/MRaf;7H/MRaf;7H/MRaf;14027H8;0P027H8E`0802`0;00/02`9000?EgIGMUMfEgH1E`5G
0EL1019G01MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP09E`D00@4100003`020@<0011G0EL1E`5G0EL1
E`5G0EL125L0>Raf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H90006;00/02`00R`03gH/MRaf;7H/MRaf;7H/MP08;14000Mf;7H/MRaf008/00H0;00/
02`900Qf25LA00Qf65L08GIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP0400810@0000l0
00<100000P090@06E`5G0EL12ELIMP0Q;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00P0
020/02`0;00/07H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;14027H8;0P027H8E`0802`0;00/02`9000?
EgIGMUMfEgH1E`5G0EL1019G01MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP09E`D00@4100003`000`40
0002000@E`5G0EL1E`5G0EL1E`5G0@QG03X/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf2@001R`0;00/008/00mf;7H/MRaf;7H/MRaf;7H022`A0007
MRaf;7H/MP02;00602`0;00/2@08MPQG4@08MQQG025fEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
MUMfEgH01@010@40000?00030@0000802@401UL1E`5G0@UG6GH08Baf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;008000P;00/02`0;01f;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`A00Qf22`800Qf
25L0200/02`0;00/2@003eMfEgIGMUMf0EL1E`5G0@0BE`0GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH0
2EL500410@0000l000<100000P0045L1E`5G0EL1E`5G0EL1E`48E`0j;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MPT000H/02`0;002;00?MRaf;7H/
MRaf;7H/MRaf00P/4@001gH/MRaf;7H00R`01P0/02`0;0T027H8Ea4027HHE`0QMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMUMf00D00@4100003`000`40000200T100IG0EL1E`49EaUf024/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`0200082`0;00/02`0MRaf;7H/MRaf;7H/MRaf;7H/
MRaf;7H/4@08MPP/2008MPQG00P0;00/02`0;0T000mGMUMfEgIGMP5G0EL1E`404UL05gIGMUMfEgIG
MUMfEgIGMUMfEgIGMUMf00UG1@010@40000?00030@000080011G0EL1E`5G0EL1E`5G0EL125L0@Raf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H0;00/
02`0;0T000H/02`0;00:;00?MRaf;7H/MP0/02`0;00/00T02B`01P0/02`0;0T000Qf;7H/MRaf;0Qf
4@0PE`0QMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf00D00@41000000@00@410`030@D0
00<100000P090@06E`5G0EL12EL9MP0`;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf2B`8000@;00/02`0;01f;7H/MRaf;0Qf00T/02`0;00/02`0200@;0P000L/MRaf
;7H/009f00MGMUMfEgIG00T000P/02`0;00/00QG00l1E`5G0EL1E`5G0EL1E`402UL087IGMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG1@010@40000200030@0000D000<100000P000`400002000@
E`5G0EL1E`5G0EL1E`5G0@QG048/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf02`0;00/02`90006;00/02`02R`03gH/MRaf;7H0;00/02`0;00900T/
00H0;00/02`90008MRaf;7H/MR`8MQ4085L08GIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
MP0500410@00008000<100001@000`40000200<10P090@06E`5G0EL12EL9MP0`;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf2B`8000@;00/02`0;01f;7H/MRaf;0Qf
00T/02`0;00/02`0200@;0P000L/MRaf;7H/009f00MGMUMfEgIG00T000P/02`0;00/00QG00l1E`5G
0EL1E`5G0EL1E`402UL087IGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG10020@4000020003
0@0000800`4500030@000080011G0EL1E`5G0EL1E`5G0EL125L0@Raf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H0;00/02`0;0T000H/02`0;00:;00?
MRaf;7H/MP0/02`0;00/00T02B`01P0/02`0;0T000Qf;7H/MRaf;0Qf4@0PE`0QMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMUMf00D00@4100000P000`40000200030@0000D000<100000P090@06
E`5G0EL12EL9MP0`;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
2B`8000@;00/02`0;01f;7H/MRaf;0Qf00T/02`0;00/02`0200@;0P000L/MRaf;7H/009f00MGMUMf
EgIG00T000P/02`0;00/00QG00l1E`5G0EL1E`5G0EL1E`402UL087IGMUMfEgIGMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgIG1@010@4000000`010@0400@110000`400002000@E`5G0EL1E`5G0EL1E`5G0@QG
048/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
02`0;00/02`90006;00/02`02R`03gH/MRaf;7H0;00/02`0;00900T/00H0;00/02`90008MRaf;7H/
MR`8MQ4085L08GIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP0500410@0000l000<10000
0P090@06E`5G0EL12EL9MP0`;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf2B`8000@;00/02`0;01f;7H/MRaf;0Qf00T/02`0;00/02`0200@;0P000L/MRaf;7H/009f
00MGMUMfEgIG00T000P/02`0;00/00QG00l1E`5G0EL1E`5G0EL1E`402UL087IGMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgIG1@010@40000?00030@000080011G0EL1E`5G0EL1E`5G0EL125L0@Raf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MQ4001L/02`0;00/02`0;00/02`0;00/02`0;00Y0007;00/02`0;00;0007;00/02`0;009E`07
0EL1E`5G0@0@E`0QMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf00D00@4100003`000`40
000200T100IG0EL1E`49EaUf02P/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
0R`01P0/02`0;0T000P/02`0;00/00T/00P0;00/02`0;24000l/02`0;00/02`0;00/02`02@007b`0
;00/02`00EL1E`5G0EL1E`5G0EL1E`5G0EL1E`402UL087IGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
MUMfEgIG1@010@40000?00030@000080011G0EL1E`5G0EL1E`5G0EL125L0@Raf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MQ4001L/02`0
;00/02`0;00/02`0;00/02`0;00Y0007;00/02`0;00;0007;00/02`0;009E`070EL1E`5G0@0@E`0Q
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf00D00@4100003`020@<02@401UL1E`5G0@UG
6GH0:2af;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H2;00602`0;00/2@0022`0
;00/02`02B`0200/02`0;00/8@003b`0;00/02`0;00/02`0;009000O;00/02`0;001E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0@0:E`0PMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUL400810@0000l0
00<100000P0045L1E`5G0EL1E`5G0EL1E`48E`12;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf4@005b`0;00/02`0;00/02`0;00/02`0
;00/02T000L/02`0;00/00/000L/02`0;00/00UG00L1E`5G0EL1011G025fEgIGMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgH01@010@40000?00030@0000802@401UL1E`5G0@UG6GH0:2af;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H2;00602`0;00/2@0022`0;00/02`02B`0200/
02`0;00/8@003b`0;00/02`0;00/02`0;009000O;00/02`0;001E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0@0:E`0PMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUL500410@0000l000<100000P0045L1
E`5G0EL1E`5G0EL1E`48E`12;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf4@005b`0;00/02`0;00/02`0;00/02`0;00/02T000L/02`0
;00/00/000L/02`0;00/00UG00L1E`5G0EL1011G025fEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
MUMfEgH01@010@40000?00030@0000802@401UL1E`5G0@UG6GH0:2af;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H2;00602`0;00/2@0022`0;00/02`02B`0200/02`0;00/8@003b`0
;00/02`0;00/02`0;009000O;00/02`0;001E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@0:E`0PMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUL500410@0000l000<100000P0HE`1:;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf02`0;00/
02`A0007;00/02`0;01200P/2EL01`5G0EL1E`4045L047IGMUMfEgIGMUMfEgIGMUL9MP08EgIGMUMf
EgH500410@0000l000<100000P003`5G0EL1E`5G0EL1E`5G0@09E`Uf03P/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MPT/40004B`0;00/02`0;00/02`0
;00/034002L/02`0;00/07H/MRaf;7H/0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`402UL087IGMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG1@010@40000?00030@00008065L0BRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MP0/02`0;00/
4@001b`0;00/02`0@P08;0UG00L1E`5G0EL1011G011fEgIGMUMfEgIGMUMfEgIG2GH025MfEgIGMUMf
1@010@40000?00810`003`5G0EL1E`5G0EL1E`5G0@09E`Uf03P/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MPT/40004B`0;00/02`0;00/02`0;00/0340
02L/02`0;00/07H/MRaf;7H/0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`402UL087IGMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgIG10020@40000?00030@00008065L0BRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MP0/02`0;00/4@001b`0
;00/02`0@P08;0UG00L1E`5G0EL1011G011fEgIGMUMfEgIGMUMfEgIG2GH025MfEgIGMUMf1@010@40
000?00030@00008000l1E`5G0EL1E`5G0EL1E`402EL9MP0h;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H9;100014/02`0;00/02`0;00/02`0;00a000W
;00/02`0;01f;7H/MRaf;05G0EL1E`5G0EL1E`5G0EL1E`5G0EL100YG021fEgIGMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMUMfE`D00@4100003`000`40000201QG04X/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H0;00/02`0;14000L/02`0
;00/048022`9E`070EL1E`5G0@0@E`0@MUMfEgIGMUMfEgIGMUMfE`Uf00QGMUMfEgIGMPD00@410000
3`000`400002000?0EL1E`5G0EL1E`5G0EL100UG2GH0>2af;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf2B`@000A;00/02`0;00/02`0;00/02`0<@009b`0
;00/02`0MRaf;7H/MR`1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@0:E`0PMUMfEgIGMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMUL500410@0000l000<100000P0HE`1:;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf02`0;00/02`A0007;00/02`0
;01200P/2EL01`5G0EL1E`4045L047IGMUMfEgIGMUMfEgIGMUL9MP08EgIGMUMfEgH500410@0000l0
00<100000P0HE`Uf00l/MRaf;7H/MRaf;7H/MR`08gH03Raf;7H/MRaf;7H/MRaf0R`01P0/02`0;0T0
014/02`0;00/02`0;00/02`0;00i000OMRaf;7H/MR`1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@0:E`0P
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUL500410@0000l000<100000P0@E`0aMUMfEgIG
MUL/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;009MP0@;7H/MRaf;7H/MRaf
;7H/MPP/6@001b`0;00/02`0>P0022af;7H/MRaf2EL01`5G0EL1E`4045L08GIGMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMP0500410@0000l000<100000P0HE`Uf00l/MRaf;7H/MRaf;7H/MR`0
8gH03Raf;7H/MRaf;7H/MRaf0R`01P0/02`0;0T0014/02`0;00/02`0;00/02`0;00i000OMRaf;7H/
MR`1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@0:E`0PMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
MUL500410@0000l00P43011G035fEgIGMUMfEbaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/00Uf010/MRaf;7H/MRaf;7H/MRaf22`I0007;00/02`0;00j0008;7H/MRaf;7H9E`07
0EL1E`5G0@0@E`0QMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf00@00P4100003`000`40
000201QG2GH03baf;7H/MRaf;7H/MRaf;00SMP0>;7H/MRaf;7H/MRaf;7H2;00602`0;00/2@004B`0
;00/02`0;00/02`0;00/03T001mf;7H/MRaf;05G0EL1E`5G0EL1E`5G0EL1E`5G0EL100YG021fEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfE`D00@4100003`000`400002011G035fEgIGMUMfEbaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf010/MRaf;7H/MRaf;7H/MRaf
22`I0007;00/02`0;00j0008;7H/MRaf;7H9E`070EL1E`5G0@0@E`0QMUMfEgIGMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMUMf00D00@4100003`000`40000201QG2GH03baf;7H/MRaf;7H/MRaf;00SMP0>
;7H/MRaf;7H/MRaf;7H2;00602`0;00/2@004B`0;00/02`0;00/02`0;00/03T001mf;7H/MRaf;05G
0EL1E`5G0EL1E`5G0EL1E`5G0EL100YG021fEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfE`D0
0@4100003`000`400002011G035fEgIGMUMfEbaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/00Uf010/MRaf;7H/MRaf;7H/MRaf22`I0007;00/02`0;00j0008;7H/MRaf;7H9E`07
0EL1E`5G0@0@E`0QMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf00D00@4100003`000`40
000200070EL1E`5G0@0:E`06MUMfEgIG2WH022af;7H/MRaf82`BMP06;7H/MRaf0R`01P0/02`0;0T0
00P/02`0;00/014/00P0;00/02`0;24002L/02`0;00/07H/MRaf;7H/0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`402UL087IGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG1@010@40000?00030@000080
25L047IGMUMfEgIGMUMfEgIGMUL:MP06;7H/MRaf0R`05`0/02`0;00/02`0;00/02`0;00/02`000P/
025f;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf02`0;00/02`04@003b`0;00/02`0;00/02`0;00b0008
;7H/MRaf;7H9E`070EL1E`5G0@0@E`0QMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf00D0
0@4100003`000`40000200070EL1E`5G0@0:E`06MUMfEgIG2WH022af;7H/MRaf82`BMP06;7H/MRaf
0R`01P0/02`0;0T000P/02`0;00/014/00P0;00/02`0;24002L/02`0;00/07H/MRaf;7H/0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`402UL087IGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG1@010@40
000?00810`08E`0@MUMfEgIGMUMfEgIGMUMfE`Yf00H/MRaf;7H2;00G02`0;00/02`0;00/02`0;00/
02`0;00022`08GH/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H0;00/02`0;00A000?;00/02`0;00/02`0
;00/038000P/MRaf;7H/MPUG00L1E`5G0EL1011G025fEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
MUMfEgH010020@40000?00030@00008000L1E`5G0EL100YG00IfEgIGMUL:MP08;7H/MRaf;7HP;19f
00H/MRaf;7H2;00602`0;00/2@0022`0;00/02`04B`0200/02`0;00/8@009b`0;00/02`0MRaf;7H/
MR`1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@0:E`0PMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
MUL500410@0000l000<100000P08E`0@MUMfEgIGMUMfEgIGMUMfE`Yf00H/MRaf;7H2;00G02`0;00/
02`0;00/02`0;00/02`0;00022`08GH/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H0;00/02`0;00A000?
;00/02`0;00/02`0;00/038000P/MRaf;7H/MPUG00L1E`5G0EL1011G025fEgIGMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgH01@010@40000?00030@00008000L1E`5G0EL100YG00IfEgIGMUL:MP08
;7H/MRaf;7HP;19f00H/MRaf;7H2;00602`0;00/2@0022`0;00/02`04B`0200/02`0;00/8@009b`0
;00/02`0MRaf;7H/MR`1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@0:E`0PMUMfEgIGMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMUL500410@0000l000<100000P08E`0@MUMfEgIGMUMfEgIGMUMfE`Yf00H/MRaf
;7H2;00G02`0;00/02`0;00/02`0;00/02`0;00022`08GH/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H0
;00/02`0;00A000?;00/02`0;00/02`0;00/038000P/MRaf;7H/MPUG00L1E`5G0EL1011G025fEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH01@010@40000?00030@00008000l1E`5G0EL1E`5G
0EL1E`400UL01WIGMUMfEa9f00P/MRaf;7H/MPP/00P0;00/02`0;0T000P/02`0;00/011f2B`05`0/
02`0;00/02`0;00/02`0;00/02`0014/8@009b`0;00/02`0MRaf;7H/MR`1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0@0:E`0PMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUL500410@0000l000<10000
0P0025L1E`5G0EL125L027IGMUMfEgIG2WH03Raf;7H/MRaf;7H/MRaf2R`01P0/02`0;0T000P/02`0
;00/008/00mf;7H/MRaf;7H/MRaf;7H022`I000?;00/02`0;00/02`0;00/02X022`9E`070EL1E`5G
0@0@E`0HMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG2GH500410@000003004100810`020@D000<10000
0P003`5G0EL1E`5G0EL1E`5G0@02E`06MUMfEgIG4WH022af;7H/MRaf22`0200/02`0;00/2@0022`0
;00/02`047H9;00G02`0;00/02`0;00/02`0;00/02`0;0004B`Q000W;00/02`0;01f;7H/MRaf;05G
0EL1E`5G0EL1E`5G0EL1E`5G0EL100YG021fEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfE`D0
0@4100000P000`40000200040@000@@000<100000P0025L1E`5G0EL125L027IGMUMfEgIG2WH03Raf
;7H/MRaf;7H/MRaf2R`01P0/02`0;0T000P/02`0;00/008/00mf;7H/MRaf;7H/MRaf;7H022`I000?
;00/02`0;00/02`0;00/02X022`9E`070EL1E`5G0@0@E`0HMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
2GH500410@0000<000D100000@0200030@0000800`42000?0EL1E`5G0EL1E`5G0EL1009G00IfEgIG
MULBMP08;7H/MRaf;7H8;00802`0;00/02`90008;00/02`0;00@MPT/01L0;00/02`0;00/02`0;00/
02`0;00/000A;24002L/02`0;00/07H/MRaf;7H/0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`402UL087IG
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG10020@40000400040@000@8000<100000P000`40
00020008E`5G0EL1E`48E`08MUMfEgIGMUL:MP0>;7H/MRaf;7H/MRaf;7H:;00602`0;00/2@0022`0
;00/02`00R`03gH/MRaf;7H/MRaf;7H/MP08;1T000l/02`0;00/02`0;00/02`0:P08;0UG00L1E`5G
0EL1011G01QfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUL9MPD00@41000000D00@000@0200040@000@@0
00<100000P003`5G0EL1E`5G0EL1E`5G0@02E`06MUMfEgIG4WH022af;7H/MRaf22`0200/02`0;00/
2@0022`0;00/02`047H9;00G02`0;00/02`0;00/02`0;00/02`0;0004B`Q000W;00/02`0;01f;7H/
MRaf;05G0EL1E`5G0EL1E`5G0EL1E`5G0EL100YG021fEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
MUMfE`D00@4100000P020@@00P4500030@00008000QG0EL1E`5G0@QG00QfEgIGMUMfE`Yf00h/MRaf
;7H/MRaf;7H/MPX/00H0;00/02`90008;00/02`0;002;00?MRaf;7H/MRaf;7H/MRaf00P/6@003b`0
;00/02`0;00/02`0;00Z00P/2EL01`5G0EL1E`4045L067IGMUMfEgIGMUMfEgIGMUMfEgIGMUMfE`Uf
1@010@40000?00030@0000804EL01WIGMUMfEbYf22`90008;00/02`0;011;1T002L/02`0;00/07H/
MRaf;7H/0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`402UL087IGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
MUMfEgIG1@010@40000?00030@00008045L027IGMUMfEgIG2WH07Raf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MPT/2P001R`0;00/014/00L0;00/02`000T/0140;00/02`0;00/02`0;00/0009;006
02`0;00/6@0022`0;00/02`00R`01gH/MRaf;7H02EL01`5G0EL1E`4045L08GIGMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMP0500410@0000l000<100000P0AE`06MUMfEgIG:WH8;0T000P/02`0
;00/044/6@009b`0;00/02`0MRaf;7H/MR`1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@0:E`0PMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUL500410@0000l000<100000P0@E`08MUMfEgIGMUL:MP0N
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf2B`:0006;00/02`04B`01`0/02`0;0002B`04@0/
02`0;00/02`0;00/02`000T/00H0;00/02`I0008;00/02`0;002;007MRaf;7H/MP09E`070EL1E`5G
0@0@E`0QMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf00D00@4100003`020@<04EL01WIG
MUMfEbYf22`90008;00/02`0;011;1T002L/02`0;00/07H/MRaf;7H/0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`402UL087IGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG10020@40000?00030@000080
45L027IGMUMfEgIG2WH07Raf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MPT/2P001R`0;00/014/
00L0;00/02`000T/0140;00/02`0;00/02`0;00/0009;00602`0;00/6@0022`0;00/02`00R`01gH/
MRaf;7H02EL01`5G0EL1E`4045L08GIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP050041
0@0000l000<100000P0AE`06MUMfEgIG:WH8;0T000P/02`0;00/044/6@009b`0;00/02`0MRaf;7H/
MR`1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@0:E`0PMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
MUL500410@0000l000<100000P0@E`08MUMfEgIGMUL:MP0N;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf2B`:0006;00/02`04B`01`0/02`0;0002B`04@0/02`0;00/02`0;00/02`000T/00H0;00/
02`I0008;00/02`0;002;007MRaf;7H/MP09E`070EL1E`5G0@0@E`0QMUMfEgIGMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMUMf00D00@4100003`000`40000203Uf00L/MRaf;7H/00Uf2B`80008;00/02`0
;00H;00YMRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf02`0;00/02`04@0022`0;00/02`0
27H05`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL100YG021fEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
MUMfE`D00@4100003`000`40000201Qf030/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H2;00802`0;00/02`A000>;00/02`0;00/02`0;00:;00OMRaf;7H/MRaf;7H/
MRaf;7H/MRaf;7H/MRaf;7H/MP02;00602`0;00/4@002B`0;00/02`0;008MPUG00L1E`5G0EL1011G
025fEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH01@010@40000?00030@000080>GH01baf
;7H/MR`02GH9;0P000P/02`0;00/01P/02Uf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H0
;00/02`0;00A0008;00/02`0;008MP0G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`402UL087IGMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG1@010@40000?00030@00008067H0<2af;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MP8/00P0;00/02`0;14000h/02`0;00/02`0
;00/00X/01mf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf008/00H0;00/02`A0009;00/02`0
;00/00Qf2EL01`5G0EL1E`4045L08GIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP050041
0@0000l000<100000P0iMP07;7H/MRaf;009MPT/200022`0;00/02`062`0:GH/MRaf;7H/MRaf;7H/
MRaf;7H/MRaf;7H/MRaf;7H/MP0/02`0;00/014000P/02`0;00/00Qf01L1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0@0:E`0PMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUL500410@0000l00P4301Qf
030/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H2;00802`0;00/
02`A000>;00/02`0;00/02`0;00:;00OMRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MP02;006
02`0;00/4@002B`0;00/02`0;008MPUG00L1E`5G0EL1011G025fEgIGMUMfEgIGMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgH010020@40000?00030@000080>GH01baf;7H/MR`02GH9;0P000P/02`0;00/01P/
02Uf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H0;00/02`0;00A0008;00/02`0;008MP0G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`402UL087IGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
1@010@40000?00030@00008067H0<2af;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MP8/00P0;00/02`0;14000h/02`0;00/02`0;00/00X/01mf;7H/MRaf;7H/MRaf;7H/
MRaf;7H/MRaf;7H/MRaf008/00H0;00/02`A0009;00/02`0;00/00Qf2EL01`5G0EL1E`4045L08GIG
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP0500410@0000l000<100000P09MP0?;7H/MRaf
;7H/MRaf;7H/02Uf010/MRaf;7H/MRaf;7H/MRaf0R`01P0/02`0;1T000P/02`0;00/00X/01Mf;7H/
MRaf;7H/MRaf;7H/MRaf;7H/MP08;00802`0;00/02`9000W;00/02`0;01GMUMfEgIGMP5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL100YG021fEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfE`D00@410000
3`000`400002001B;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRT000P/02`0;00/00P/00mf;7H/MRaf;7H/
MRaf;7H02B`@0009;00/02`0;00/00Qf011G0EL1E`5G0EL1E`5G0EL145L08GIGMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMP0500410@0000l000<100000P09MP0?;7H/MRaf;7H/MRaf;7H/02Uf
010/MRaf;7H/MRaf;7H/MRaf0R`01P0/02`0;1T000P/02`0;00/00X/01Mf;7H/MRaf;7H/MRaf;7H/
MRaf;7H/MP08;00802`0;00/02`9000W;00/02`0;01GMUMfEgIGMP5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL100YG021fEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfE`D00@4100003`000`400002001B
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRT000P/02`0;00/00P/00mf;7H/MRaf;7H/MRaf;7H02B`@0009
;00/02`0;00/00Qf011G0EL1E`5G0EL1E`5G0EL145L08GIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
MUMfEgIGMP0500410@0000l000<100000P09MP0?;7H/MRaf;7H/MRaf;7H/02Uf010/MRaf;7H/MRaf
;7H/MRaf0R`01P0/02`0;1T000P/02`0;00/00X/01Mf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MP08;008
02`0;00/02`9000W;00/02`0;01GMUMfEgIGMP5G0EL1E`5G0EL1E`5G0EL1E`5G0EL100YG021fEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfE`D00@4100003`020@<0058/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf:@0022`0;00/02`022`03gH/MRaf;7H/MRaf;7H/MP09;10000T/02`0;00/02`027H045L1
E`5G0EL1E`5G0EL1E`4@E`0QMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf00@00P410000
3`000`40000200Uf00l/MRaf;7H/MRaf;7H/MR`0:GH042af;7H/MRaf;7H/MRaf;7H2;00602`0;00/
6@0022`0;00/02`02R`05gH/MRaf;7H/MRaf;7H/MRaf;7H/MRaf00P/00P0;00/02`0;0T002L/02`0
;00/05MfEgIGMUMf0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`402UL087IGMUMfEgIGMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgIG1@010@40000?00030@000080058/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf:@0022`0
;00/02`022`03gH/MRaf;7H/MRaf;7H/MP09;10000T/02`0;00/02`027H045L1E`5G0EL1E`5G0EL1
E`4@E`0QMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf00D00@4100003`000`40000200Uf
00l/MRaf;7H/MRaf;7H/MR`0:GH042af;7H/MRaf;7H/MRaf;7H2;00602`0;00/6@0022`0;00/02`0
2R`05gH/MRaf;7H/MRaf;7H/MRaf;7H/MRaf00P/00P0;00/02`0;0T002L/02`0;00/05MfEgIGMUMf
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`402UL087IGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
1@010@40000?00030@00008027H0FRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H0;00/02`0;24000P/02`0
;00/01T/00H0;00/02`90009;00/02`0;00/00Qf011G0EL1E`5G0EL1E`5G0EL145L08GIGMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP0500410@0000l000<100000P001eMfEgIGMUL04WH03baf
;7H/MRaf;7H/MRaf;00[MP06;7H/MRaf2R`01P0/02`0;140010/02`0;00/02`0;00/02`04R`03gH/
MRaf;7H0;00/02`0;00800T/01mGMUMfEgIGMP5G0EL1E`5G0EL1E`5G0EL1E`5G0EL101YG011fEgIG
MUMfEgIGMUMfEgIG1@010@40000?00030@00008027H0FRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H0;00/
02`0;24000P/02`0;00/01T/00H0;00/02`90009;00/02`0;00/00Qf011G0EL1E`5G0EL1E`5G0EL1
45L08GIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP0500410@0000l000<100000P001eMf
EgIGMUL04WH03baf;7H/MRaf;7H/MRaf;00[MP06;7H/MRaf2R`01P0/02`0;140010/02`0;00/02`0
;00/02`04R`03gH/MRaf;7H0;00/02`0;00800T/01mGMUMfEgIGMP5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL101YG011fEgIGMUMfEgIGMUMfEgIG1@010@40000?00810`08MP1J;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MP0/02`0;00/8@0022`0;00/02`06B`01P0/02`0;0T000T/02`0;00/02`027H045L1E`5G0EL1
E`5G0EL1E`4@E`0QMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf00@00P4100003`000`40
00020007EgIGMUMfE`0BMP0?;7H/MRaf;7H/MRaf;7H/02]f00H/MRaf;7H:;00602`0;00/4@0042`0
;00/02`0;00/02`0;00B;00?MRaf;7H/MP0/02`0;00/00P02B`07eMfEgIGMUMf0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`406UL047IGMUMfEgIGMUMfEgIGMUL500410@0000l000<100000P08MP1J;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MP0/02`0;00/8@0022`0;00/02`06B`01P0/02`0;0T000T/02`0;00/02`0
27H045L1E`5G0EL1E`5G0EL1E`4@E`0QMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf00D0
0@4100003`000`4000020007EgIGMUMfE`0BMP0?;7H/MRaf;7H/MRaf;7H/02]f00H/MRaf;7H:;006
02`0;00/4@0042`0;00/02`0;00/02`0;00B;00?MRaf;7H/MP0/02`0;00/00P02B`07eMfEgIGMUMf
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`406UL047IGMUMfEgIGMUMfEgIGMUL500410@0000l000<10000
0P08MP1A;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf010/MRaf;7H/MP0/02`0;00/6@0042`0;00/02`0
;00/02`0;009;00602`0;00/2@09;0Qf01QG0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4BE`0GMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgH01@010@40000?00030@00008000MGMUMfEgIG00Yf00l/MRaf;7H/MRaf
;7H/MR`0BWH8;18001H/02`0;00/02`0;00/02`0;00/02`02B`03`0/02`0;00/02`0;00/0002;00_
MRaf;7H/MRaGMUMfEgIGMP5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`404UL047IGMUMfEgIG
MUMfEgIGMUL500410@0000l000<100000P08MP1A;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf010/MRaf
;7H/MP0/02`0;00/6@0042`0;00/02`0;00/02`0;009;00602`0;00/2@09;0Qf01QG0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`4BE`0GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH01@010@4000000`010@020@80
0`4500030@00008000MGMUMfEgIG00Yf00l/MRaf;7H/MRaf;7H/MR`0BWH8;18001H/02`0;00/02`0
;00/02`0;00/02`02B`03`0/02`0;00/02`0;00/0002;00_MRaf;7H/MRaGMUMfEgIGMP5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`404UL047IGMUMfEgIGMUMfEgIGMUL500410@00008000<10000
1@000`40000200030@00008027H0DBaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;009MP0@;7H/MRaf;7H0;00/
02`0;1T0010/02`0;00/02`0;00/02`02B`01P0/02`0;0T02B`8MP0HE`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL14UL05gIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf00D00@4100000`000`40000400030@000080
0`420007EgIGMUMfE`0:MP0?;7H/MRaf;7H/MRaf;7H/04Yf22`B000F;00/02`0;00/02`0;00/02`0
;00/00T/00l0;00/02`0;00/02`0;0000R`0;gH/MRaf;7H/EgIGMUMfEgH1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1019G011fEgIGMUMfEgIGMUMfEgIG10020@40000400030@0000<11@000`40
000200Qf054/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02GH042af;7H/MRaf02`0;00/02`I000@;00/02`0
;00/02`0;00/00T/00H0;00/02`900T/27H065L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0A9G01MfEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMP0500410@000005004000400P000`40000500030@00008000MGMUMf
EgIG00Yf00l/MRaf;7H/MRaf;7H/MR`0BWH8;18001H/02`0;00/02`0;00/02`0;00/02`02B`03`0/
02`0;00/02`0;00/0002;00_MRaf;7H/MRaGMUMfEgIGMP5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`404UL047IGMUMfEgIGMUMfEgIGMUL500410@0000800P4300@110000`400002011f01T/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/025f01l/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf
01P/MRaf;7H/MP0/02`0;00/02`0;00/02`A000Q;00/02`0;00/02`0;00/02`0;00/02`0;00/02`0
;00/00Qf01QG0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4BE`0GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH0
1@010@40000?00030@00008000MGMUMfEgIG019f00L/MRaf;7H/019f00MGMUMfEgIG03Yf00H/MRaf
;7H:;00G02`0;00/02`0;00/02`0;00/02`0;0008R`09eMfEgIGMUMf0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0@0BE`0@MUMfEgIGMUMfEgIGMUMfE`D00@4100003`000`400002011f01T/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/025f01l/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf
01P/MRaf;7H/MP0/02`0;00/02`0;00/02`A000Q;00/02`0;00/02`0;00/02`0;00/02`0;00/02`0
;00/00Qf01QG0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4BE`0GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH0
1@010@40000?00030@00008000MGMUMfEgIG019f00L/MRaf;7H/019f00MGMUMfEgIG03Yf00H/MRaf
;7H:;00G02`0;00/02`0;00/02`0;00/02`0;0008R`09eMfEgIGMUMf0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0@0BE`0@MUMfEgIGMUMfEgIGMUMfE`D00@4100003`000`400002011f01T/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/025f01l/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf
01P/MRaf;7H/MP0/02`0;00/02`0;00/02`A000Q;00/02`0;00/02`0;00/02`0;00/02`0;00/02`0
;00/00Qf01QG0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4BE`0GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH0
1@010@40000?00810`001eMfEgIGMUL04WH01baf;7H/MR`04WH01eMfEgIGMUL0>WH01Raf;7H/MPX/
01L0;00/02`0;00/02`0;00/02`0;00/000R;00WEgIGMUMfEgH1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1019G011fEgIGMUMfEgIGMUMfEgIG10020@40000?00030@00008047H06Baf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MR`08GH07baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02GH062af
;7H/MRaf02`0;00/02`0;00/02`0;140024/02`0;00/02`0;00/02`0;00/02`0;00/02`0;00/02`0
27H065L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0A9G01MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP050041
0@0000l000<100000P001eMfEgIGMUL04WH01baf;7H/MR`04WH01eMfEgIGMUL0>WH01Raf;7H/MPX/
01L0;00/02`0;00/02`0;00/02`0;00/000R;00WEgIGMUMfEgH1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1019G011fEgIGMUMfEgIGMUMfEgIG1@010@40000?00030@00008047H06Baf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MR`027H8EbUf00L/MRaf;7H/01Yf22`H0008;00/02`0;009;00702`0;00/
0002;00WMRaf;7H/MUL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1019G00mfEgIGMUMfEgIG
MUMfEgH01@010@40000?00030@00008000MGMUMfEgIG019f00l/MRaf;7H/MRaf;7H/MR`02GH8E`08
MUMfEgIGMULiMP09EgIGMUMfEgIG00P/4@001R`0;00/014/00T0;00/02`0;00027H90@0FE`5G0EL1
E`5G0EL1E`5G0EL1E`5G0B9G1@010@40000?00030@00008047H06Baf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MR`027H8EbUf00L/MRaf;7H/01Yf22`H0008;00/02`0;009;00702`0;00/0002;00WMRaf;7H/
MUL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1019G00mfEgIGMUMfEgIGMUMfEgH01@010@40
000?00030@00008000MGMUMfEgIG019f00l/MRaf;7H/MRaf;7H/MR`02GH8E`08MUMfEgIGMULiMP09
EgIGMUMfEgIG00P/4@001R`0;00/014/00T0;00/02`0;00027H90@0FE`5G0EL1E`5G0EL1E`5G0EL1
E`5G0B9G1@010@40000?00030@00008047H06Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`027H8EbUf
00L/MRaf;7H/01Yf22`H0008;00/02`0;009;00702`0;00/0002;00WMRaf;7H/MUL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1019G00mfEgIGMUMfEgIGMUMfEgH01@010@40000?00030@000080
00MGMUMfEgIG019f00l/MRaf;7H/MRaf;7H/MR`02GH8E`08MUMfEgIGMULiMP09EgIGMUMfEgIG00P/
4@001R`0;00/014/00T0;00/02`0;00027H90@0FE`5G0EL1E`5G0EL1E`5G0EL1E`5G0B9G1@010@40
000?00810`0@MP0I;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;008MPQG:GH01baf;7H/MR`06WH8;1P0
00P/02`0;00/00T/00L0;00/02`0008/02Mf;7H/MRafE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`404UL03gIGMUMfEgIGMUMfEgIGMP0400810@0000l000<100000P001eMfEgIGMUL04WH03baf
;7H/MRaf;7H/MRaf;009MPQG00QfEgIGMUMfEcUf00UGMUMfEgIGMUL022`A0006;00/02`04B`02@0/
02`0;00/0008MPT101IG0EL1E`5G0EL1E`5G0EL1E`5G0EL18UL500410@0000l000<100000P0@MP0`
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/EgIGMUMfEgH1E`5G0EL10UL01WIGMUMfEcaf
00L/MRaf;7H/01P000P/02`0;00/00P/2@08;00PE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`4QE`D00@4100003`000`4000020007EgIGMUMfE`0BMP0?;7H/MRaf;7H/MRaf;7H/00Yf00IGMUMf
EgH2E`0?0EL1E`5G0GIGMUMfEgIG049f00P0;00/02`0;0T000h/02`0;00/02`0;00/00T/00T0;00/
02`0;00022`0205G0EL1E`5G2@4065L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0AQG1@010@40000?0003
0@00008047H0<2af;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;5MfEgIGMUMf0EL1E`5G0@9G
00IfEgIGMULlMP07;7H/MRaf;00H0008;00/02`0;008;0T022`085L1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL18EL500410@0000l000<100000P001eMfEgIGMUL04WH03baf;7H/MRaf;7H/MRaf
;00:MP06EgIGMUMf0UL03`5G0EL1E`5fEgIGMUMfE`12MP0802`0;00/02`9000>;00/02`0;00/02`0
;009;00902`0;00/02`000P/00P1E`5G0EL1E`T101QG0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4HE`D0
0@4100003`000`400002011f030/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaGMUMfEgIG
MP5G0EL1E`42E`06MUMfEgIG?7H01baf;7H/MR`0600022`0;00/02`022`900P/021G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0B5G1@010@40000?00030@00008000MGMUMfEgIG019f00l/MRaf
;7H/MRaf;7H/MR`02WH01UMfEgIGMP9G00l1E`5G0EL1MUMfEgIGMUL0@WH0200/02`0;00/2@003R`0
;00/02`0;00/02`02B`02@0/02`0;00/0008;0080EL1E`5G0EL90@0HE`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL165L500410@0000l00P43011f030/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaG
MUMfEgIGMP5G0EL1E`42E`06MUMfEgIG?7H01baf;7H/MR`0600022`0;00/02`022`900P/021G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0B5G10020@40000?00030@00008000MGMUMfEgIG019f
00l/MRaf;7H/MRaf;7H/MR`02WH01UMfEgIGMP9G00l1E`5G0EL1MUMfEgIGMUL0@WH0200/02`0;00/
2@003R`0;00/02`0;00/02`02B`02@0/02`0;00/0008;0080EL1E`5G0EL90@0HE`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL165L500410@0000l000<100000P0@MP0`;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/EgIGMUMfEgH1E`5G0EL12UL027IGMUMfEgIG>GH01b`0;00/02`04@003b`0;00/02`0
;00/02`0;00;0007;00/02`0;009E`0G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`408EL500410@0000l0
00<100000P003eMfEgIGMUMfEgIGMUMfE`0:MP0G;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`00WH01eMf
EgIGMUL0204085L1E`5G0EL1MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG:WH0200/02`0;00/4@001R`0
;00/00T/00T0;00/02`0;00022`0205G0EL1E`5G4@4045L1E`5G0EL1E`5G0EL1E`4HE`D00@410000
3`000`400002011f030/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaGMUMfEgIGMP5G0EL1
E`4:E`08MUMfEgIGMULiMP07;00/02`0;00A000?;00/02`0;00/02`0;00/00/000L/02`0;00/00UG
01L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@0QE`D00@4100003`000`400002000?EgIGMUMfEgIGMUMf
EgIG00Yf01L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;002MP07EgIGMUMfE`080@0PE`5G0EL1E`5fEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMULZMP0802`0;00/02`A0006;00/02`02B`02@0/02`0;00/0008;008
0EL1E`5G0ELA0@0@E`5G0EL1E`5G0EL1E`5G0AQG1@010@40000?00030@00008047H0<2af;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;5MfEgIGMUMf0EL1E`5G0@YG00QfEgIGMUMfEcUf00L/02`0
;00/014000l/02`0;00/02`0;00/02`02`001b`0;00/02`02EL05`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1025G1@010@40000?00030@00008000mGMUMfEgIGMUMfEgIGMUL02WH05baf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/009f00MGMUMfEgIG00P1021G0EL1E`5G0GIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEbYf
00P0;00/02`0;14000H/02`0;009;00902`0;00/02`000P/00P1E`5G0EL1Ea41011G0EL1E`5G0EL1
E`5G0EL165L500410@0000l00P43011f030/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaG
MUMfEgIGMP5G0EL1E`4:E`08MUMfEgIGMULiMP07;00/02`0;00A000?;00/02`0;00/02`0;00/00/0
00L/02`0;00/00UG01L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@0QE`@00P4100003`000`400002000?
EgIGMUMfEgIGMUMfEgIG00Yf01L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;002MP07EgIGMUMfE`080@0P
E`5G0EL1E`5fEgIGMUMfEgIGMUMfEgIGMUMfEgIGMULZMP0802`0;00/02`A0006;00/02`02B`02@0/
02`0;00/0008;0080EL1E`5G0ELA0@0@E`5G0EL1E`5G0EL1E`5G0AQG1@010@40000?00030@000080
47H0<2af;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;5MfEgIGMUMf0EL1E`5G0@YG00QfEgIG
MUMfEcUf00L/02`0;00/014000l/02`0;00/02`0;00/02`02`001b`0;00/02`02EL05`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1025G1@010@40000?00030@00008027H01eMfEgIGMUL02WH05baf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/015f014/MRaf;7H/MRaf;7H/MRaf;008MP0OEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfE`0;MP07EgIGMUMfE`08MP0802`0;00/02`90006;00/02`02B`06@0/02`0;00/
02`0;00/02`0EgIGMUMfEgH06@4085L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL11@010@40
000?00030@00008047H08Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00AMP06;7H/MRaf
2b`01WH/MRaf;09f00mGMUMfEgIGMUMfEgIGMUL08WH01baf;7H/MR`06009;00G02`0;00/02`0;00/
02`0;7IGMUMfEgH00UL07`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`406EL500410@0000l0
00<100000P08MP07EgIGMUMfE`0:MP0G;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`04GH04Baf;7H/MRaf
;7H/MRaf;7H/00Qf01mGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG00]f00MGMUMfEgIG00Qf
00P0;00/02`0;0T000H/02`0;009;00I02`0;00/02`0;00/02`0;01GMUMfEgIGMP0I0@0PE`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4500410@0000l000<100000P0@MP0Q;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/015f00H/MRaf;7H;;006MRaf;7H/0WH03eMfEgIGMUMfEgIGMUMf
E`0RMP07;7H/MRaf;00H00T/01L0;00/02`0;00/02`0;00/MUMfEgIGMP02E`0O0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0@0IE`D00@41000000@00@4110020@D000<100000P08MP07EgIGMUMf
E`0:MP0G;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`04GH04Baf;7H/MRaf;7H/MRaf;7H/00Qf01mGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG00]f00MGMUMfEgIG00Qf00P0;00/02`0;0T000H/02`0
;009;00I02`0;00/02`0;00/02`0;01GMUMfEgIGMP0I0@0PE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`4500410@0000@000@100010P000`40000200030@00008047H08Baf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;00AMP06;7H/MRaf2b`01WH/MRaf;09f00mGMUMfEgIGMUMfEgIGMUL0
8WH01baf;7H/MR`06009;00G02`0;00/02`0;00/02`0;7IGMUMfEgH00UL07`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`406EL500410@0000@000@100010P000`40000200<10P08MP07EgIGMUMf
E`0:MP0G;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`04GH04Baf;7H/MRaf;7H/MRaf;7H/00Qf01mGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG00]f00MGMUMfEgIG00Qf00P0;00/02`0;0T000H/02`0
;009;00I02`0;00/02`0;00/02`0;01GMUMfEgIGMP0I0@0PE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`4400810@0000800P4300040@000@@000<100000P0@MP0Q;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/015f00H/MRaf;7H;;006MRaf;7H/0WH03eMfEgIGMUMfEgIGMUMfE`0RMP07
;7H/MRaf;00H00T/01L0;00/02`0;00/02`0;00/MUMfEgIGMP02E`0O0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0@0IE`D00@41000010001040004200030@00008000<100000P0IMP0P;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H9;00@MRaf;7H/MRaf;7H/MRaf;11f035GMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG00P/00l0;00/02`0;00/02`0;000
6R`8MP080EL1E`5G0ELI0@0HE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL11@010@40000010010@440081
1@000`40000201Qf010/MRaf;7H/MRaf;7H/MRaf8B`067H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;0Yf
00mGMUMfEgIGMUMfEgIGMUL06GH01b`0;00/02`04@006B`0;00/02`0;00/02`0;00/02`0;00/02`0
27H0>EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`0500410@0000l000<100000P0IMP0P;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H9;00@
MRaf;7H/MRaf;7H/MRaf;11f035GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIG00P/00l0;00/02`0;00/02`0;0006R`8MP080EL1E`5G0ELI0@0HE`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL11@010@40000?00030@00008067H042af;7H/MRaf;7H/MRaf;7HQ;00HMRaf;7H/
MRaf;7H/MRaf;7H/MRaf;7H/2WH03eMfEgIGMUMfEgIGMUMfE`0IMP07;00/02`0;00A000I;00/02`0
;00/02`0;00/02`0;00/02`0;008MP0iE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G00D00@4100003`000`40000201Uf020/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MPT/011f;7H/MRaf;7H/MRaf;7H/47H0<EMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUL022`03`0/02`0;00/02`0;00/000J;0Qf00P1E`5G
0EL1EaT101QG0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4500410@0000l000<100000P0HMP0@;7H/MRaf
;7H/MRaf;7H/MR4/01Qf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`:MP0?EgIGMUMfEgIGMUMfEgIG01Uf
00L/02`0;00/014001T/02`0;00/02`0;00/02`0;00/02`0;00/00Qf03UG0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL01@010@40000?00810`0IMP0P
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H9;00@MRaf;7H/MRaf;7H/MRaf;11f035GMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG00P/00l0;00/02`0;00/
02`0;0006R`8MP080EL1E`5G0ELI0@0HE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL110020@40000?0003
0@00008067H042af;7H/MRaf;7H/MRaf;7HQ;00HMRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/2WH03eMf
EgIGMUMfEgIGMUMfE`0IMP07;00/02`0;00A000I;00/02`0;00/02`0;00/02`0;00/02`0;008MP0i
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G00D0
0@4100003`000`4000020007EgIGMUMfE`0JMP0G;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02GH01eL1
E`5G0EL04P49MP07;7H/MRaf;008MP0hEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG;7H/
MRaf;7H0;00/02`0;00/02`0;00J;0Qf25La0@D00@4100003`000`40000201Qf010/MRaf;7H/MRaf
;7H/MRaf4B`027H/MRaf;7H/25L04@5G0EL1E`5G0EL1E`5G0EL100Qf00L/MRaf;7H/019f01MGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfE`02MP06;7H/MRaf0R`05`0/02`0;00/02`0;00/02`0;00/02`0018/
00Mf;7H/MRaf00QG40408EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`0500410@0000l0
00<100000P001eMfEgIGMUL06WH05baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf00MG0EL1E`5G0181
2GH01baf;7H/MR`027H0>5MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEbaf;7H/MRaf02`0
;00/02`0;00/02`06R`8MPQG<@4500410@0000l000<100000P0HMP0@;7H/MRaf;7H/MRaf;7H/MQ4/
00Qf;7H/MRaf;0QG0141E`5G0EL1E`5G0EL1E`5G0@08MP07;7H/MRaf;00BMP0GEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUL00WH01Raf;7H/MP8/01L0;00/02`0;00/02`0;00/02`0;00/000B;007MRaf;7H/
MP08Ea01025G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL01@010@40000?00030@000080
00MGMUMfEgIG01Yf01L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;009MP07E`5G0EL1E`0B0@Uf00L/MRaf
;7H/00Qf03QGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUL/MRaf;7H/MP0/02`0;00/02`0
;00/01X/27H8Ec411@010@40000?00030@00008067H042af;7H/MRaf;7H/MRaf;7HA;008MRaf;7H/
MR`8E`0A0EL1E`5G0EL1E`5G0EL1E`4027H01baf;7H/MR`04WH05eMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIG009f00H/MRaf;7H2;00G02`0;00/02`0;00/02`0;00/02`0;0004R`01gH/MRaf;7H025L@0@0Q
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G00D00@4100003`000`4000020007EgIGMUMf
E`0JMP0G;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02GH01eL1E`5G0EL04P49MP07;7H/MRaf;008MP0h
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG;7H/MRaf;7H0;00/02`0;00/02`0;00J;0Qf
25La0@D00@4100003`020@<067H042af;7H/MRaf;7H/MRaf;7HA;008MRaf;7H/MR`8E`0A0EL1E`5G
0EL1E`5G0EL1E`4027H01baf;7H/MR`04WH05eMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG009f00H/MRaf
;7H2;00G02`0;00/02`0;00/02`0;00/02`0;0004R`01gH/MRaf;7H025L@0@0QE`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G00@00P4100003`000`40003o00d00@4100003`000`40003o00d0
0@4100003`000`40000T00030@0002L000<100009`000`40000W00030@0002L000<100009`000`40
000C00410@0000l000<100000P000`40000600030@0000D000<100001@000`40000600030@0000D0
00<100001P000`40000500030@0000D000<100001P000`40000500030@0000H000<100001@000`40
000600030@0000D000<100001@000`40000600030@0000D000<100001P000`40000500030@0000H0
00<100001@000`40000500030@0000H000<100001@000`40000600030@0000D000<100001@000`40
000600030@0000D000<100001P000`40000500030@0000800@4100003`3o0A410@000?l08@000001
\
\>"],
  ImageRangeCache->{{{108, 395}, {425.5, 138.5}} -> {-14.274, 15.1703, 
  0.118757, 0.118757}}]
}, Open  ]],

Cell["\<\
Looking at the difference between two consecutive images in a \
motion sequence is often useful to indentify the pixels involved in the \
motion.\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\
DensityPlot[imageArrayOne[Round[i],
                          Round[j]] -
            imageArrayTwo[Round[i],
                          Round[j]],
            {i,1,myImageSize[[1]]},{j,1,myImageSize[[2]]},
            {Mesh -> False, PlotPoints -> myImageSize[[2]]}] ;\
\>", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% DensityGraphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.0117866 0.0310174 -0.0117866 0.0310174 [
[.1433 -0.0125 -3 -9 ]
[.1433 -0.0125 3 0 ]
[.29839 -0.0125 -6 -9 ]
[.29839 -0.0125 6 0 ]
[.45347 -0.0125 -6 -9 ]
[.45347 -0.0125 6 0 ]
[.60856 -0.0125 -6 -9 ]
[.60856 -0.0125 6 0 ]
[.76365 -0.0125 -6 -9 ]
[.76365 -0.0125 6 0 ]
[.91873 -0.0125 -6 -9 ]
[.91873 -0.0125 6 0 ]
[ 0 0 -0.125 0 ]
[-0.0125 .1433 -6 -4.5 ]
[-0.0125 .1433 0 4.5 ]
[-0.0125 .29839 -12 -4.5 ]
[-0.0125 .29839 0 4.5 ]
[-0.0125 .45347 -12 -4.5 ]
[-0.0125 .45347 0 4.5 ]
[-0.0125 .60856 -12 -4.5 ]
[-0.0125 .60856 0 4.5 ]
[-0.0125 .76365 -12 -4.5 ]
[-0.0125 .76365 0 4.5 ]
[-0.0125 .91873 -12 -4.5 ]
[-0.0125 .91873 0 4.5 ]
[ 0 0 -0.125 0 ]
[ 0 1 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.1433 0 m
.1433 .00625 L
s
[(5)] .1433 -0.0125 0 1 Mshowa
.29839 0 m
.29839 .00625 L
s
[(10)] .29839 -0.0125 0 1 Mshowa
.45347 0 m
.45347 .00625 L
s
[(15)] .45347 -0.0125 0 1 Mshowa
.60856 0 m
.60856 .00625 L
s
[(20)] .60856 -0.0125 0 1 Mshowa
.76365 0 m
.76365 .00625 L
s
[(25)] .76365 -0.0125 0 1 Mshowa
.91873 0 m
.91873 .00625 L
s
[(30)] .91873 -0.0125 0 1 Mshowa
.125 Mabswid
.17432 0 m
.17432 .00375 L
s
.20533 0 m
.20533 .00375 L
s
.23635 0 m
.23635 .00375 L
s
.26737 0 m
.26737 .00375 L
s
.3294 0 m
.3294 .00375 L
s
.36042 0 m
.36042 .00375 L
s
.39144 0 m
.39144 .00375 L
s
.42246 0 m
.42246 .00375 L
s
.48449 0 m
.48449 .00375 L
s
.51551 0 m
.51551 .00375 L
s
.54653 0 m
.54653 .00375 L
s
.57754 0 m
.57754 .00375 L
s
.63958 0 m
.63958 .00375 L
s
.6706 0 m
.6706 .00375 L
s
.70161 0 m
.70161 .00375 L
s
.73263 0 m
.73263 .00375 L
s
.79467 0 m
.79467 .00375 L
s
.82568 0 m
.82568 .00375 L
s
.8567 0 m
.8567 .00375 L
s
.88772 0 m
.88772 .00375 L
s
.11228 0 m
.11228 .00375 L
s
.08127 0 m
.08127 .00375 L
s
.05025 0 m
.05025 .00375 L
s
.01923 0 m
.01923 .00375 L
s
.94975 0 m
.94975 .00375 L
s
.98077 0 m
.98077 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
0 .1433 m
.00625 .1433 L
s
[(5)] -0.0125 .1433 1 0 Mshowa
0 .29839 m
.00625 .29839 L
s
[(10)] -0.0125 .29839 1 0 Mshowa
0 .45347 m
.00625 .45347 L
s
[(15)] -0.0125 .45347 1 0 Mshowa
0 .60856 m
.00625 .60856 L
s
[(20)] -0.0125 .60856 1 0 Mshowa
0 .76365 m
.00625 .76365 L
s
[(25)] -0.0125 .76365 1 0 Mshowa
0 .91873 m
.00625 .91873 L
s
[(30)] -0.0125 .91873 1 0 Mshowa
.125 Mabswid
0 .17432 m
.00375 .17432 L
s
0 .20533 m
.00375 .20533 L
s
0 .23635 m
.00375 .23635 L
s
0 .26737 m
.00375 .26737 L
s
0 .3294 m
.00375 .3294 L
s
0 .36042 m
.00375 .36042 L
s
0 .39144 m
.00375 .39144 L
s
0 .42246 m
.00375 .42246 L
s
0 .48449 m
.00375 .48449 L
s
0 .51551 m
.00375 .51551 L
s
0 .54653 m
.00375 .54653 L
s
0 .57754 m
.00375 .57754 L
s
0 .63958 m
.00375 .63958 L
s
0 .6706 m
.00375 .6706 L
s
0 .70161 m
.00375 .70161 L
s
0 .73263 m
.00375 .73263 L
s
0 .79467 m
.00375 .79467 L
s
0 .82568 m
.00375 .82568 L
s
0 .8567 m
.00375 .8567 L
s
0 .88772 m
.00375 .88772 L
s
0 .11228 m
.00375 .11228 L
s
0 .08127 m
.00375 .08127 L
s
0 .05025 m
.00375 .05025 L
s
0 .01923 m
.00375 .01923 L
s
0 .94975 m
.00375 .94975 L
s
0 .98077 m
.00375 .98077 L
s
.25 Mabswid
0 0 m
0 1 L
s
.1433 .99375 m
.1433 1 L
s
.29839 .99375 m
.29839 1 L
s
.45347 .99375 m
.45347 1 L
s
.60856 .99375 m
.60856 1 L
s
.76365 .99375 m
.76365 1 L
s
.91873 .99375 m
.91873 1 L
s
.125 Mabswid
.17432 .99625 m
.17432 1 L
s
.20533 .99625 m
.20533 1 L
s
.23635 .99625 m
.23635 1 L
s
.26737 .99625 m
.26737 1 L
s
.3294 .99625 m
.3294 1 L
s
.36042 .99625 m
.36042 1 L
s
.39144 .99625 m
.39144 1 L
s
.42246 .99625 m
.42246 1 L
s
.48449 .99625 m
.48449 1 L
s
.51551 .99625 m
.51551 1 L
s
.54653 .99625 m
.54653 1 L
s
.57754 .99625 m
.57754 1 L
s
.63958 .99625 m
.63958 1 L
s
.6706 .99625 m
.6706 1 L
s
.70161 .99625 m
.70161 1 L
s
.73263 .99625 m
.73263 1 L
s
.79467 .99625 m
.79467 1 L
s
.82568 .99625 m
.82568 1 L
s
.8567 .99625 m
.8567 1 L
s
.88772 .99625 m
.88772 1 L
s
.11228 .99625 m
.11228 1 L
s
.08127 .99625 m
.08127 1 L
s
.05025 .99625 m
.05025 1 L
s
.01923 .99625 m
.01923 1 L
s
.94975 .99625 m
.94975 1 L
s
.98077 .99625 m
.98077 1 L
s
.25 Mabswid
0 1 m
1 1 L
s
.99375 .1433 m
1 .1433 L
s
.99375 .29839 m
1 .29839 L
s
.99375 .45347 m
1 .45347 L
s
.99375 .60856 m
1 .60856 L
s
.99375 .76365 m
1 .76365 L
s
.99375 .91873 m
1 .91873 L
s
.125 Mabswid
.99625 .17432 m
1 .17432 L
s
.99625 .20533 m
1 .20533 L
s
.99625 .23635 m
1 .23635 L
s
.99625 .26737 m
1 .26737 L
s
.99625 .3294 m
1 .3294 L
s
.99625 .36042 m
1 .36042 L
s
.99625 .39144 m
1 .39144 L
s
.99625 .42246 m
1 .42246 L
s
.99625 .48449 m
1 .48449 L
s
.99625 .51551 m
1 .51551 L
s
.99625 .54653 m
1 .54653 L
s
.99625 .57754 m
1 .57754 L
s
.99625 .63958 m
1 .63958 L
s
.99625 .6706 m
1 .6706 L
s
.99625 .70161 m
1 .70161 L
s
.99625 .73263 m
1 .73263 L
s
.99625 .79467 m
1 .79467 L
s
.99625 .82568 m
1 .82568 L
s
.99625 .8567 m
1 .8567 L
s
.99625 .88772 m
1 .88772 L
s
.99625 .11228 m
1 .11228 L
s
.99625 .08127 m
1 .08127 L
s
.99625 .05025 m
1 .05025 L
s
.99625 .01923 m
1 .01923 L
s
.99625 .94975 m
1 .94975 L
s
.99625 .98077 m
1 .98077 L
s
.25 Mabswid
1 0 m
1 1 L
s
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of gray image
p
.01923 .01923 translate
.96154 .96154 scale
32 string
32 32 8 [32 0 0 32 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  Mimage
8B8885888B8B88888888888F8B8B8F8888858888888B8B88
88888B8B8888888B
888B888B88858888888B888B888B888888888B8B8F8B8B88
85888B8B8B888888
88888B8B8B8F8B88888B8B8B888B88888B88888B85858885
888B8B8B88888B88
8B888B8B888B8888888B8B8B8B8888888B8F8F8B8B8B8888
888B888B85888B8B
88888B888888858B888585888B8B8F8F8F8B8F888B888888
888888888B8B8B88
8588888B8B858585888888858B928F8885929C817A8B7A6D
7E88888885858B88
8B8B8F8B8B8B8585858888888895A08B81A3BA7E5C9C8548
608B7E77818B8888
888588888B8885858F8B858585A0747EFFFF747AFEF4A03A
3E88886A70858B85
858B8888858888888F81A0E38F7A0005FFFF0023FFFFC866
266A88706A6D8188
8B8885888888888F770074FFED520000FFFF0000FFFFB700
4FFFB06D6A66748B
858F8B88888B888F81000FFFFF700000FFFF0000FFFF0000
C8FFFF605C59748F
888B8B818B888B8B9C0000FFFF990000FFFF0000FFFF0000
FFFFFF604B4F7E88
888885858B8B8B8B9C370085FFF4000070FF4800FFFF0000
FFFFE05C4152747E
888B88888B8888888F6D0000FFFFAA0000FFF400FFFF0000
FFFF9C4B374F7074
81888885858B88888B99000045FFFF2A00FFFF00FFFF0000
FFFF6348344F7074
858B8588888B8B888B8F8B0000ADFF9212638B92A6D263E6
FFFF524537527477
888B888B88887E706360883A0015DCAD4B6088888177A6FF
FFFF59483E557770
88888B888592DCFFFFC54823000092C84F0F858888858BFF
FFFF594B45607074
88888588851C1266F4FFFF9500006AFFED00008585819CFF
FFFF604845667474
888B888B8B662A000000FFFF8B000055CFFE00128F88B4FF
FFFF52484B6A7777
8B888B85889299A02D0000FFFFB08859005526001581EAFF
FFFF4F484F6D747E
888B88888B888B8BA30F004FFFF7FFFFFFA08100004BFFFF
FFF0554B6074777E
858B85888B8B8B8B927000003499BEFFFFFFC1520000FFFF
FFB4604860777481
888B888885888B8B8B8F7000000034AAFAFFF4C82300FFFF
FF95634F637A7781
8888888F888B858B8B88929C6A0000008FD9ADFF6D3AFFFF
FF8F6A5263777481
8B888B8B8F8B8B88888F8B8B95A04500003099FFE08B77FF
FF85775C60747A7E
888B8B8B888B858888888588888B99A3000085BEF0CB12FF
FF8B7E635C77777E
88888F8B8F8B8585858888888F8B88994B004FB7E6E31CFF
FFA677705C747E7E
8885888B8B888B8B858B8B8B88888F8F880000A3D2E605FA
FFF4777A636D7A7E
888B8B888F88888B85888F8F8B8B888F95150055B0FF2395
FFFF857E6A6D7E7A
8B88888B888B85888B888B8B88888B888B7A002D5CFF666D
FFFFB77E74708585
858B8B8F888B85888B8B858F8B88888B8B8F00086AD2997A
F4FFFF7E85858185
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000180P10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdoojn_l00;34gTJ2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_[C^]:JVYZf][Jo^kP00
oiT0IP3o0020P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;0000?`00O`00_`0o000o
?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0
_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo?cooOcoo
_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no?gnoOgno
_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo
_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0_ol0oolo0?lo?oloOolo_olooomo
0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_mWEc:0Pl=WIfJj^[WMgMiW<09W<<iW<
IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`c
c<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co
<lcoI/coVLcoc<coool0<ol0I_l0VOl0c?l0oolc0?lc<olcI_lcVOlcc?lcoomV0?mVI_mVVOmVoonI
0?nI<onII_nIVOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<ool08@000?l08@000?l08@000?l08@0003@0
0`4T00<110020B400`4300<18@040@<00P4Q00@10P030B400`440081500003L000<100008P000`40
000200040@000B4000<100001@000`40000O00030@00008000@100018@000`40000500030@000240
00@100010P000`40000A0000=`000`40000R00030@00008000@100018@000`40000500030@000200
00D100000@0200030@00020000<1000010000`40000Q00040@000@8000<100004@0003@00`4U0003
0@00008000@100018@000`40000200<190001040004200030@00024000<100000`4R00810`001040
004C0000=0000`40000U00030@00008000@100018@000`40000200030@00024000@100010P001040
004P00040@000@8000<1000090001040004200030@000140000d00@18`020@D00P4Q008110040B40
0P4400818P020@<0104P00<110020A@0003o0240003o0240003o0240003o0240000?0?l14@410000
3`000`40000200030@0000H000<100001@000`40000500030@0000H000<100001@000`4000060003
0@0000D000<100001@000`40000600030@0000D000<100001P000`40000500030@0000H000<10000
1@000`40000500030@0000H000<100001@000`40000600030@0000D000<100001P000`4000050003
0@0000D000<100001P000`40000500030@0000H000<100001@000`40000500030@0000H000<10000
1@000`40000600030@0000D000<100000P010@40000?00030@0002@000<100009`000`40000W0003
0@0002L000<100009`000`40000W00030@0001<00@4100003`000`40003o00d00@4100003`000`40
003o00d00@4100003`000`40003o00d00@4100003`020Ol03@020@40000?00030@00008000l/MRaf
;7H/MRaf;7H/MR`02GH0Lbaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MR`02GH0LBaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/00D00@4100003`000`4000020?mf1WH500410@0000l000<100000P003baf;7H/MRaf
;7H/MRaf;009MP1c;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;009MP1a;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MR`01@010@40000?00030@000080ogH6MPD00@4100003`000`400002000?;7H/MRaf;7H/MRaf
;7H/00Uf07</MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/00Uf074/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;00500410@0000l000<100000P3oMPIf1@010@40000?00030@00008000l/MRaf;7H/MRaf;7H/MR`0
2GH0Lbaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`0
2GH0LBaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00D0
0@4100003`020@<0ogH6MP@00P4100003`000`400002000Y;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MR`02GH0TBaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02GH0>Baf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00500410@0000l0
00<100000P3oMPIf1@010@40000?00030@00008002T/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;009MP2A;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;009MP0i;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00D00@4100003`000`40
00020?mf1WH500410@0000l000<100000P00:Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/00Uf094/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf03T/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`01@010@40000?00030@000080
ogH6MPD00@4100003`000`400002000Y;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MR`02GH0TBaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02GH0>Baf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00500410@0000l000<100000P3oMPIf
1@010@40000?00810`00Xbaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`0
4GH01baf;7H/MR`02GH0@Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/00@00P4100003`000`4000020?mf1WH500410@0000l000<10000
0P00Xbaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`04GH01baf;7H/MR`0
2GH0@Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/00D00@4100003`000`4000020?mf1WH500410@0000l000<100000P00Xbaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`04GH01baf;7H/MR`02GH0@Baf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00D0
0@4100003`000`4000020?mf1WH500410@0000l000<100000P00Xbaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`04GH01baf;7H/MR`02GH0@Baf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00D00@4100003`000`40
00020?mf1WH500410@0000l00P43003S;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MR`02gH05baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00@00P4100003`000`4000020?mf
1WH500410@0000l000<100000P00hbaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/00]f01L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00500410@0000l000<100000P3oMPIf
1@010@40000?00030@0000800></MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;00;MP0G;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`01@010@40000?00030@000080ogH6MPD0
0@4100001`030@D000<100000P00hbaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/00]f01L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00500410@0000X000<100000P000`40
00020?mf1WH500410@0000X000<100000P030@800></MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;00;MP0G;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`010020@40000700<1
1@000`4000020?mf1WH500410@0000L000<100001@000`400002000a;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;009MP0?;7H/MRaf;7H/MRaf;7H/015f0://MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`01@010@40000700@1
10000`4000020?mf1WH500410@0000l000<100000P00<Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02GH03baf;7H/MRaf;7H/MRaf;00AMP2[;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00D00@4100003`000`4000020?mf
1WH500410@0000l000<100000P00<Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MR`02GH03baf;7H/MRaf;7H/MRaf;00AMP2[;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00D00@4100003`000`4000020?mf1WH500410@0000l0
00<100000P00<Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`0
2GH03baf;7H/MRaf;7H/MRaf;00AMP2[;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/00D00@4100003`020@<0TgH01baf;7H/MR`08WH01eMfEgIGMUL0@WH40081
0@0000l000<100000P08MP0Q;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/01Uf01L/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;009MP0Q;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf
00l/MRaf;7H/MRaf;7H/MR`04GH01baf;7H/MR`06GH05baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/01=f
00l/MRaf;7H/MRaf;7H/MR`01@010@40000?00030@000080TgH01baf;7H/MR`08WH01eMfEgIGMUL0
@WH500410@0000l000<100000P08MP0Q;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/01Uf
01L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;009MP0Q;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/00Uf00l/MRaf;7H/MRaf;7H/MR`04GH01baf;7H/MR`06GH05baf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/01=f00l/MRaf;7H/MRaf;7H/MR`01@010@40000?00030@000080TgH01baf;7H/MR`08WH01eMf
EgIGMUL0@WH500410@0000l000<100000P08MP0Q;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/01Uf01L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;009MP0Q;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/00Uf00l/MRaf;7H/MRaf;7H/MR`04GH01baf;7H/MR`06GH05baf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/01=f00l/MRaf;7H/MRaf;7H/MR`01@010@40000?00030@000080TgH01baf;7H/MR`0
8WH01eMfEgIGMUL0@WH500410@0000l000<100000P08MP0Q;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/01Uf01L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;009MP0Q;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/00Uf00l/MRaf;7H/MRaf;7H/MR`04GH01baf;7H/MR`06GH05baf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/01=f00l/MRaf;7H/MRaf;7H/MR`01@010@40000?00810`1cMP07;7H/MRaf
;00AMP08;7H/MRaf;7H8;0Uf00iGMUMfEgIG;7H/MRaf;0Yf2EL01WIGMUMfEa5f00MGMUMfEgIG029f
10020@40000?00030@000080034/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/01Uf03T/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MR`02GH01Raf;7H/MPT/27H01eMfEgIGMUL00WH01baf;7H/MR`02GH8E`0?MUMfEgIG
MUL/MRaf;7H/01]f01L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00500410@0000l000<100000P1cMP07
;7H/MRaf;00AMP08;7H/MRaf;7H8;0Uf00iGMUMfEgIG;7H/MRaf;0Yf2EL01WIGMUMfEa5f00MGMUMf
EgIG029f1@010@40000?00030@000080034/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/01Uf03T/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MR`02GH01Raf;7H/MPT/27H01eMfEgIGMUL00WH01baf;7H/MR`02GH8E`0?
MUMfEgIGMUL/MRaf;7H/01]f01L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00500410@0000l000<10000
0P1cMP07;7H/MRaf;00AMP08;7H/MRaf;7H8;0Uf00iGMUMfEgIG;7H/MRaf;0Yf2EL01WIGMUMfEa5f
00MGMUMfEgIG029f1@010@40000?00030@000080034/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/01Uf03T/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02GH01Raf;7H/MPT/27H01eMfEgIGMUL00WH01baf;7H/MR`0
2GH8E`0?MUMfEgIGMUL/MRaf;7H/01]f01L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00500410@0000l0
00<100000P1cMP07;7H/MRaf;00AMP08;7H/MRaf;7H8;0Uf00iGMUMfEgIG;7H/MRaf;0Yf2EL01WIG
MUMfEa5f00MGMUMfEgIG029f1@010@40000?00030@000080034/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/01Uf03T/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02GH01Raf;7H/MPT/27H01eMfEgIGMUL00WH01baf
;7H/MR`02GH8E`0?MUMfEgIGMUL/MRaf;7H/01]f01L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;0050041
0@0000l000<100000P1[MP0>;7H/MRaf;5MfEgIGMUL:MQ0000QfEgIGMUMfE`Qf40006Baf;7H/MRaf
;05G0EL1E`5G0EL1E`5G0EL047H04EMfEgIGMUMfEgIGMUMfEgIG01Qf1@010@40000?00810`001baf
;7H/MR`02GH08Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00AMP0?;7H/MRaf;7H/MRaf
;7H/01Uf00L/MRaf;7H/019f400@MQ0000Uf;7H/MRaf;7H045L03baf;7H/MRaf;7H/MRaf;00KMP07
;7H/MRaf;008MP@00P4100003`000`40000206]f00h/MRaf;7H/EgIGMUMfE`Yf400027IGMUMfEgIG
27H@000I;7H/MRaf;7H/0EL1E`5G0EL1E`5G0EL1E`0@MP0AEgIGMUMfEgIGMUMfEgIGMUL067H50041
0@0000l000<100000P001baf;7H/MR`02GH08Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;00AMP0?;7H/MRaf;7H/MRaf;7H/01Uf00L/MRaf;7H/019f400@MQ0000Uf;7H/MRaf;7H045L03baf
;7H/MRaf;7H/MRaf;00KMP07;7H/MRaf;008MPD00@4100003`000`40000206]f00h/MRaf;7H/EgIG
MUMfE`Yf400027IGMUMfEgIG27H@000I;7H/MRaf;7H/0EL1E`5G0EL1E`5G0EL1E`0@MP0AEgIGMUMf
EgIGMUMfEgIGMUL067H500410@0000l000<100000P001baf;7H/MR`02GH08Baf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;00AMP0?;7H/MRaf;7H/MRaf;7H/01Uf00L/MRaf;7H/019f400@MQ00
00Uf;7H/MRaf;7H045L03baf;7H/MRaf;7H/MRaf;00KMP07;7H/MRaf;008MPD00@4100003`000`40
000206]f00h/MRaf;7H/EgIGMUMfE`Yf400027IGMUMfEgIG27H@000I;7H/MRaf;7H/0EL1E`5G0EL1
E`5G0EL1E`0@MP0AEgIGMUMfEgIGMUMfEgIGMUL067H500410@0000l000<100000P001baf;7H/MR`0
2GH08Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00AMP0?;7H/MRaf;7H/MRaf;7H/01Uf
00L/MRaf;7H/019f400@MQ0000Uf;7H/MRaf;7H045L03baf;7H/MRaf;7H/MRaf;00KMP07;7H/MRaf
;008MPD00@4100003`000`40000205=f00H/MRaf;7H2;00702`0;00/000@MQ4140080@08E`5G0EL1
E`4@00T/00mGMUMfEgIGMP5G0EL1E`400UL01WIGMUMfE`Uf01UGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIG011f1@010@40000?00810`08MP0I;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;009MP0O;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;009MP08;7H/MRaf;7H80007;7H/MRaf;009MQ4140090@07
E`5G0EL1E`0@000@;00/02`0;00/MUMfEgIGMP9G00L1E`5G0EL100Qf00L/MRaf;7H/02=f00L/MRaf
;7H/00@00P4100003`000`40000205=f00H/MRaf;7H2;00702`0;00/000@MQ4140080@08E`5G0EL1
E`4@00T/00mGMUMfEgIGMP5G0EL1E`400UL01WIGMUMfE`Uf01UGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIG011f1@010@40000?00030@00008027H06Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02GH07baf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02GH022af;7H/MRaf20001baf;7H/MR`02GHA0A00
2@401eL1E`5G0EL0400042`0;00/02`0;7IGMUMfEgH2E`070EL1E`5G0@08MP07;7H/MRaf;00SMP07
;7H/MRaf;00500410@0000l000<100000P1CMP06;7H/MRaf0R`01`0/02`0;00047HA0A00204025L1
E`5G0EL14009;00?EgIGMUMfEgH1E`5G0EL1009G00IfEgIGMUL9MP0IEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfE`0@MPD00@4100003`000`40000200Qf01T/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf
01l/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf00P/MRaf;7H/MPP000L/MRaf;7H/00Uf
4@4@00T100MG0EL1E`5G0100010/02`0;00/02afEgIGMUMf0UL01`5G0EL1E`4027H01baf;7H/MR`0
8gH01baf;7H/MR`01@010@40000?00030@000080DgH01Raf;7H/MP8/00L0;00/02`0011f4@4@00P1
00QG0EL1E`5G0A002B`03eMfEgIGMUMf0EL1E`5G0@02E`06MUMfEgIG2GH06EMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUL047H500410@00000400410@@00P4500030@00008027H06Baf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MR`02GH07baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02GH022af;7H/MRaf
20001baf;7H/MR`02GHA0A002@401eL1E`5G0EL0400042`0;00/02`0;7IGMUMfEgH2E`070EL1E`5G
0@08MP07;7H/MRaf;00SMP07;7H/MRaf;00500410@00008000<100000P001040004400030@000080
@WH01eMfEgIGMUL02@4025MfEgIGMUMf200022`0;00/02`025LA0A00404@00T/2048E`P022`08EMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfE`08MPD00@4100000P000`40000200040@000@@0
0`42000?;7H/MRaf;7H/MRaf;7H/00Uf02T/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;008MPT127H@0008MUMfEgIGMULA0A00404@00T/2048E`P022`AMP07EgIGMUMfE`0:MP07
;7H/MRaf;00400810@00008000<100000P001040004400030@000080@WH01eMfEgIGMUL02@4025Mf
EgIGMUMf200022`0;00/02`025LA0A00404@00T/2048E`P022`08EMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfE`08MPD00@4100000P000`40000200040@000@@000<100000P003baf;7H/MRaf
;7H/MRaf;009MP0Y;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`027H90@Qf
400027IGMUMfEgIG4@4@01014009;0P125L800P/4GH01eMfEgIGMUL02WH01baf;7H/MR`01@010@40
00000`010@0500811@000`400002049f00MGMUMfEgIG00T100QGMUMfEgIGMPP000P/02`0;00/00QG
4@4@01014009;0P125L800P/025GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUL027H50041
0@0000l000<100000P003baf;7H/MRaf;7H/MRaf;009MP0Y;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MR`027H90@Qf400027IGMUMfEgIG4@4@01014009;0P125L800P/4GH01eMf
EgIGMUL02WH01baf;7H/MR`01@010@40000?00030@000080@WH01eMfEgIGMUL02@4025MfEgIGMUMf
200022`0;00/02`025LA0A00404@00T/2048E`P022`08EMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfE`08MPD00@4100003`000`400002000?;7H/MRaf;7H/MRaf;7H/00Uf02T/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;008MPT127H@0008MUMfEgIGMULA0A00404@00T/
2048E`P022`AMP07EgIGMUMfE`0:MP07;7H/MRaf;00500410@0000l000<100000P12MP07EgIGMUMf
E`090@08EgIGMUMfEgH80008;00/02`0;008Ea41400@0A002B`80@QG2008;00QEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIG00Qf1@010@40000?00030@00008027H0>Baf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;008MQ414008MQ41400@0A00
4@40200/02`0;00/400067IGMUMfEgIGMUMfEgIGMUMfEgIGMUMfE`Yf00L/MRaf;7H/00D00@410000
3`020@<0BGHA0A0000QGMUMfEgIGMQ41400@0A004@48;100025GMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUL027H400810@0000l000<100000P08MP0i;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Qf4@4@00Qf4@4@0101400A0@0802`0;00/
02`@000HMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG2WH01baf;7H/MR`01@010@40000?00030@000080
BGHA0A0000QGMUMfEgIGMQ41400@0A004@48;100025GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUL027H500410@0000l000<100000P08MP0i;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Qf4@4@00Qf4@4@0101400A0@0802`0;00/02`@000H
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG2WH01baf;7H/MR`01@010@40000?00030@000080BGHA0A00
00QGMUMfEgIGMQ41400@0A004@48;100025GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUL0
27H500410@0000l000<100000P08MP0i;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Qf4@4@00Qf4@4@0101400A0@0802`0;00/02`@000HMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgIG2WH01baf;7H/MR`01@010@40000?00030@000080BGHA0A0000QGMUMf
EgIGMQ41400@0A004@48;100025GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUL027H50041
0@0000l000<100000P005baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00]f02L/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`04@4@0008;7H/MRaf;7HA0A00404@014160001gIGMUMfEgH0
4UL9MP07;7H/MRaf;00500410@0000l00P43045f00P/MRaf;7H/MQ41400027H/MRaf;7H/4@4@0101
400A0AP000QGMUMfEgIGMQ5G47H400810@0000l000<100000P005baf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/00]f02L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`04@4@0008;7H/MRaf
;7HA0A00404@014160001gIGMUMfEgH04UL9MP07;7H/MRaf;00500410@0000l000<100000P11MP08
;7H/MRaf;7HA0A0000Qf;7H/MRaf;141400@0A004@4H0008EgIGMUMfEgHAEa1f1@010@40000?0003
0@00008001L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00;MP0W;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/0141400022af;7H/MRaf4@4@0101400A0AP000MfEgIGMUMf019G2GH01baf
;7H/MR`01@010@40000?00030@000080@GH022af;7H/MRaf4@4@0008MRaf;7H/MR`A0A00404@0141
600025MfEgIGMUMf4EL@MPD00@4100003`000`400002000G;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`0
2gH09baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00A0A0000P/MRaf;7H/MQ41
400@0A004@4H0007MUMfEgIGMP0BE`Uf00L/MRaf;7H/00D00@4100003`000`400002045f00P/MRaf
;7H/MQ41400027H/MRaf;7H/4@4@0101400A0AP000QGMUMfEgIGMQ5G47H500410@0000l000<10000
0P003baf;7H/MRaf;7H/MRaf;00CMP0W;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/00UG2048MQ004@48MPP025L80A004@4H0007MUMfEgIGMP0;E`06MUMfEgIG4GH500410@0000l0
00<100000P11MP0A;7H/MRaf;7IG0EL1E`5G0EL02048MQ004@4027IGMUMfEgIG2008E`P1400A0A00
010/02`0;00/05MfEgIGMUMf4EL027IGMUMfEgIG27H500410@0000l00P43000?;7H/MRaf;7H/MRaf
;7H/01=f02L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02EL80@Qf400A0@Qf
2008E`P1400A0AP000MfEgIGMUMf00]G00IfEgIGMULAMP@00P4100003`000`400002045f014/MRaf
;7H/MUL1E`5G0EL1E`080@Qf400A0@08MUMfEgIGMUL800QG204@0141400042`0;00/02`0EgIGMUMf
EgHAE`08MUMfEgIGMUL8MPD00@4100003`000`400002000?;7H/MRaf;7H/MRaf;7H/01=f02L/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02EL80@Qf400A0@Qf2008E`P1400A0AP0
00MfEgIGMUMf00]G00IfEgIGMULAMPD00@4100003`000`400002045f014/MRaf;7H/MUL1E`5G0EL1
E`080@Qf400A0@08MUMfEgIGMUL800QG204@0141400042`0;00/02`0EgIGMUMfEgHAE`08MUMfEgIG
MUL8MPD00@4100003`000`400002000?;7H/MRaf;7H/MRaf;7H/01=f02L/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02EL80@Qf400A0@Qf2008E`P1400A0AP000MfEgIGMUMf00]G
00IfEgIGMULAMPD00@4100003`000`400002045f014/MRaf;7H/MUL1E`5G0EL1E`080@Qf400A0@08
MUMfEgIGMUL800QG204@0141400042`0;00/02`0EgIGMUMfEgHAE`08MUMfEgIGMUL8MPD00@410000
3`000`4000020019;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;009MQ014008;14140080A004@4@0008;7H/MRaf;7HIEa1f
1@010@40000?00030@000080BWH025MfEgIGMUMf404@0008MRaf;7H/MR`A0A00204@0141400027H/
MRaf;7H/25L02@5G0EL1E`5G0@08E`0@MUMfEgIGMUMfEgIGMUMfE`D00@4100003`020@<004T/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/00Uf404@00P/4@4@00P1400A0A0000P/MRaf;7H/MQUG47H400810@0000l000<10000
0P1:MP08EgIGMUMfEgH@0A0000Qf;7H/MRaf;14140080A004@4@0008MRaf;7H/MR`8E`090EL1E`5G
0EL100QG011fEgIGMUMfEgIGMUMfEgIG1@010@40000?00030@00008004T/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf
404@00P/4@4@00P1400A0A0000P/MRaf;7H/MQUG47H500410@0000l000<100000P1:MP08EgIGMUMf
EgH@0A0000Qf;7H/MRaf;14140080A004@4@0008MRaf;7H/MR`8E`090EL1E`5G0EL100QG011fEgIG
MUMfEgIGMUMfEgIG1@010@40000?00030@00008004T/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf404@00P/4@4@00P1
400A0A0000P/MRaf;7H/MQUG47H500410@0000l000<100000P1:MP08EgIGMUMfEgH@0A0000Qf;7H/
MRaf;14140080A004@4@0008MRaf;7H/MR`8E`090EL1E`5G0EL100QG011fEgIGMUMfEgIGMUMfEgIG
1@010@40000?00030@00008027H03baf;7H/MRaf;7H/MRaf;00CMP0X;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MQ0125L@0009E`5G0EL1E`5G00P140080A004@4@0007MUMfEgIG
MP0JEa1f1@010@40000010010@4300<11@000`40000204Uf00T/MRaf;7H/MR`04048Ea0000P1E`5G
0EL1E`T140080A004@4@0008EgIGMUMfEgH8E`090EL1E`5G0EL100QG011fEgIGMUMfEgIGMUMfEgIG
1@010@40000200030@0000D000<100000P000`40000200Qf00l/MRaf;7H/MRaf;7H/MR`04gH0:2af
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H@0@QG40002EL1E`5G0EL1E`080A00
204@014140001gIGMUMfEgH06UL@MPD00@4100000P000`40000500030@0000800`4204Uf00T/MRaf
;7H/MR`04048Ea0000P1E`5G0EL1E`T140080A004@4@0008EgIGMUMfEgH8E`090EL1E`5G0EL100QG
011fEgIGMUMfEgIGMUMfEgIG10020@40000200030@0000800`4500030@00008027H03baf;7H/MRaf
;7H/MRaf;00CMP0X;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MQ0125L@0009
E`5G0EL1E`5G00P140080A004@4@0007MUMfEgIGMP0JEa1f1@010@40000200030@00008000<10000
1@000`40000204Uf00T/MRaf;7H/MR`04048Ea0000P1E`5G0EL1E`T140080A004@4@0008EgIGMUMf
EgH8E`090EL1E`5G0EL100QG011fEgIGMUMfEgIGMUMfEgIG1@010@4000000`010@0400@110000`40
000200Qf00l/MRaf;7H/MRaf;7H/MR`04gH0:2af;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H@0@QG40002EL1E`5G0EL1E`080A00204@014140001gIGMUMfEgH06UL@MPD00@410000
3`000`40000204Uf00T/MRaf;7H/MR`04048Ea0000P1E`5G0EL1E`T140080A004@4@0008EgIGMUMf
EgH8E`090EL1E`5G0EL100QG011fEgIGMUMfEgIGMUMfEgIG1@010@40000?00030@00008027H01baf
;7H/MR`02GH0@Raf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MQ0122`8000@;7H/MRaf;7H/0EL1E`5G0@9G00IfEgIGMUL2MP0F;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MP8/0100;00/02`0EgIGMUMfEgIG60001gIGMUMfEgH04eL01WIGMUMfEa5f
1@010@40000?00030@000080FWH@0@08MRaf;7H/MR`800Uf204027IGMUMfEgIG47H082af;7H/MRaf
02`0;00/02afEgIGMUMfEgH/02`0;00/4@0@E`090EL1E`5G0EL100QG011fEgIGMUMfEgIGMUMfEgIG
1@010@40000?00030@00008027H01baf;7H/MR`02GH0@Raf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MQ0122`8000@;7H/MRaf;7H/0EL1
E`5G0@9G00IfEgIGMUL2MP0F;7H/MRaf;7H/MRaf;7H/MRaf;7H/MP8/0100;00/02`0EgIGMUMfEgIG
60001gIGMUMfEgH04eL01WIGMUMfEa5f1@010@40000?00810`1JMQ0100Qf;7H/MRaf;0P02GH80@08
MUMfEgIGMUL@MP0P;7H/MRaf;7H0;00/02`0;7IGMUMfEgIGMR`0;00/02`A011G00T1E`5G0EL1E`40
25L047IGMUMfEgIGMUMfEgIGMUL400810@0000l000<100000P08MP07;7H/MRaf;009MP12;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
4048;0P0010/MRaf;7H/MR`1E`5G0EL10UL01WIGMUMfE`9f01H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
0R`0400/02`0;01GMUMfEgIGMULH0007MUMfEgIGMP0CE`06MUMfEgIG4GH500410@0000l000<10000
0P1JMQ0100Qf;7H/MRaf;0P02GH80@08MUMfEgIGMUL@MP0P;7H/MRaf;7H0;00/02`0;7IGMUMfEgIG
MR`0;00/02`A011G00T1E`5G0EL1E`4025L047IGMUMfEgIGMUMfEgIGMUL500410@0000l000<10000
0P08MP07;7H/MRaf;009MP12;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf4048;0P0010/MRaf;7H/MR`1E`5G0EL10UL01WIGMUMfE`9f
01H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf0R`0400/02`0;01GMUMfEgIGMULH0007MUMfEgIGMP0CE`06
MUMfEgIG4GH500410@0000l000<100000P1JMQ0100Qf;7H/MRaf;0P02GH80@08MUMfEgIGMUL@MP0P
;7H/MRaf;7H0;00/02`0;7IGMUMfEgIGMR`0;00/02`A011G00T1E`5G0EL1E`4025L047IGMUMfEgIG
MUMfEgIGMUL500410@0000l000<100000P00<Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MR`047H06EMfEgIGMUMfEgIGMUMfEgIG;7H/MRaf;7H025L80@0?E`5G0EL1
E`40;00/02`000X/2EL01WIGMUMfE`9f00l/MRaf;7H/MRaf;7H/MR`04GH022af;7H/MRaf60001gIG
MUMfEgH04eL01WIGMUMfEa5f1@010@40000?00030@000080>WH05eMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIG00Uf00P1E`5G0EL1Ea01014/02`0;00/07H/MRaf;7H/MP08E`08MUMfEgIGMULIMP0@EgIGMUMf
Ebaf;7H/MRaf;1P000QGMUMfEgIGMPQG00T1E`5G0EL1E`4025L047IGMUMfEgIGMUMfEgIGMUL50041
0@0000l000<100000P00<Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MR`047H06EMfEgIGMUMfEgIGMUMfEgIG;7H/MRaf;7H025L80@0?E`5G0EL1E`40;00/02`000X/
2EL01WIGMUMfE`9f00l/MRaf;7H/MRaf;7H/MR`04GH022af;7H/MRaf60001gIGMUMfEgH04eL01WIG
MUMfEa5f1@010@40000?00810`0jMP0GEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUL02GH0205G0EL1E`5G
40404B`0;00/02`0MRaf;7H/MRaf00QG00QfEgIGMUMfEaUf011GMUMfEgIG;7H/MRaf;7H/600025Mf
EgIGMUMf25L02@5G0EL1E`5G0@08E`0@MUMfEgIGMUMfEgIGMUMfE`@00P4100003`000`400002000a
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00@MP0IEgIGMUMf
EgIGMUMfEgIGMUL/MRaf;7H/MP08E`P100mG0EL1E`5G0@0/02`0;0002R`9E`06MUMfEgIG0WH03baf
;7H/MRaf;7H/MRaf;00AMP08;7H/MRaf;7HH0007MUMfEgIGMP0CE`06MUMfEgIG4GH500410@0000l0
00<100000P0jMP0GEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUL02GH0205G0EL1E`5G40404B`0;00/02`0
MRaf;7H/MRaf00QG00QfEgIGMUMfEaUf011GMUMfEgIG;7H/MRaf;7H/600025MfEgIGMUMf25L02@5G
0EL1E`5G0@08E`0@MUMfEgIGMUMfEgIGMUMfE`D00@4100003`000`400002000a;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00@MP0IEgIGMUMfEgIGMUMfEgIGMUL/
MRaf;7H/MP08E`P100mG0EL1E`5G0@0/02`0;0002R`9E`06MUMfEgIG0WH03baf;7H/MRaf;7H/MRaf
;00AMP08;7H/MRaf;7HH0007MUMfEgIGMP0CE`06MUMfEgIG4GH500410@0000l000<100000P0jMP0G
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUL02GH0205G0EL1E`5G40404B`0;00/02`0MRaf;7H/MRaf00QG
00QfEgIGMUMfEaUf011GMUMfEgIG;7H/MRaf;7H/600025MfEgIGMUMf25L02@5G0EL1E`5G0@08E`0@
MUMfEgIGMUMfEgIGMUMfE`D00@4100003`000`400002000a;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00@MP0IEgIGMUMfEgIGMUMfEgIGMUL/MRaf;7H/MP08E`P1
00mG0EL1E`5G0@0/02`0;0002R`9E`06MUMfEgIG0WH03baf;7H/MRaf;7H/MRaf;00AMP08;7H/MRaf
;7HH0007MUMfEgIGMP0CE`06MUMfEgIG4GH500410@0000l000<100000P0aMP0802`0;00/02`@00T/
25L0205G0EL1E`5G4048MPT/25L80BUf600025MfEgIGMUMf4EL067IGMUMfEgIGMUMfEgIGMUMfEgIG
MUMfE`D00@4100003`000`400002000Q;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf
00H/MRaf;7H2;00602`0;00/4@002B`0;00/02`0;009E`060EL1E`5G4@404Baf;7H/MRaf02`0;00/
02`000QG2049MP0?;7H/MRaf;7H/MRaf;7H/00Uf00P/MRaf;7H/MQP000MfEgIGMUMf01=G00IfEgIG
MULAMPD00@4100003`000`400002035f00P0;00/02`0;1002B`8E`080EL1E`5G0EL@0@Qf2B`8E`P1
:GHH0008EgIGMUMfEgHAE`0HMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG1@010@40000?00810`008Baf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;009MP06;7H/MRaf0R`01P0/02`0;14000T/02`0
;00/02`02EL01P5G0EL1Ea41014/MRaf;7H/MP0/02`0;00/0008E`P12GH03baf;7H/MRaf;7H/MRaf
;009MP08;7H/MRaf;7HH0007MUMfEgIGMP0CE`06MUMfEgIG4GH400810@0000l000<100000P0aMP08
02`0;00/02`@00T/25L0205G0EL1E`5G4048MPT/25L80BUf600025MfEgIGMUMf4EL067IGMUMfEgIG
MUMfEgIGMUMfEgIGMUMfE`D00@4100003`000`400002000Q;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/00Uf00H/MRaf;7H2;00602`0;00/4@002B`0;00/02`0;009E`060EL1E`5G4@404Baf
;7H/MRaf02`0;00/02`000QG2049MP0?;7H/MRaf;7H/MRaf;7H/00Uf00P/MRaf;7H/MQP000MfEgIG
MUMf01=G00IfEgIGMULAMPD00@4100003`000`400002035f00P0;00/02`0;1002B`8E`080EL1E`5G
0EL@0@Qf2B`8E`P1:GHH0008EgIGMUMfEgHAE`0HMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG1@010@40
000?00030@000080024/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02GH01Raf;7H/MP8/
00H0;00/02`A0009;00/02`0;00/00UG00H1E`5G0ELA0@0A;7H/MRaf;7H0;00/02`0;00025L80@Uf
00l/MRaf;7H/MRaf;7H/MR`02GH022af;7H/MRaf60001gIGMUMfEgH04eL01WIGMUMfEa5f1@010@40
000?00030@000080:GH@0@08MUMfEgIGMULI00Qf404025MfEgIGMUMf2P001b`0;00/02`0404HMP09
;7H/MRaf;7H/01P000QGMUMfEgIGMQ5G01QfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUL500410@0000l0
00<100000P003baf;7H/MRaf;7H/MRaf;009MP09;7H/MRaf;7H/00Qf00l1E`5G0EL1E`5G0EL1E`40
0UL01gIGMUMfEgH06@0022af;7H/MRaf4048MQ40404IMP08;7H/MRaf;7HH0007MUMfEgIGMP0CE`06
MUMfEgIG4GH500410@0000l000<100000P0YMQ0100QfEgIGMUMfEaT027H@0@08EgIGMUMfEgH:0007
;00/02`0;00@0AQf00T/MRaf;7H/MR`0600025MfEgIGMUMf4EL067IGMUMfEgIGMUMfEgIGMUMfEgIG
MUMfE`D00@4100003`020@<000l/MRaf;7H/MRaf;7H/MR`02GH02Baf;7H/MRaf;008MP0?0EL1E`5G
0EL1E`5G0EL1009G00MfEgIGMUMf01T000P/MRaf;7H/MQ0127HA01016GH022af;7H/MRaf60001gIG
MUMfEgH04eL01WIGMUMfEa5f10020@40000?00030@000080:GH@0@08MUMfEgIGMULI00Qf404025Mf
EgIGMUMf2P001b`0;00/02`0404HMP09;7H/MRaf;7H/01P000QGMUMfEgIGMQ5G01QfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMUL500410@0000l000<100000P003baf;7H/MRaf;7H/MRaf;009MP09;7H/MRaf
;7H/00Qf00l1E`5G0EL1E`5G0EL1E`400UL01gIGMUMfEgH06@0022af;7H/MRaf4048MQ40404IMP08
;7H/MRaf;7HH0007MUMfEgIGMP0CE`06MUMfEgIG4GH500410@0000l000<100000P0YMQ0100QfEgIG
MUMfEaT027H@0@08EgIGMUMfEgH:0007;00/02`0;00@0AQf00T/MRaf;7H/MR`0600025MfEgIGMUMf
4EL067IGMUMfEgIGMUMfEgIGMUMfEgIGMUMfE`D00@4100003`000`400002000?;7H/MRaf;7H/MRaf
;7H/00Uf00T/MRaf;7H/MR`027H03`5G0EL1E`5G0EL1E`5G0@02E`07MUMfEgIGMP0I0008;7H/MRaf
;7H@0@Qf4@0@0AUf00P/MRaf;7H/MQP000MfEgIGMUMf01=G00IfEgIGMULAMPD00@4100003`000`40
000202Yf00IGMUMfEgH2E`060EL1E`5G6P4@00Qf4049E`P/200@0A1f2B`H01UG01QfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMUL500410@0000l000<100000P00>Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/EgIGMUMfEgH1E`5G0EL1E`0I0A0000P/MRaf;7H/MQ01011fEgIGMUMfEgH/
02`0;00/2@090@0FE`5G0EL1EgH/MRaf;7H/MRaf;7H/MPX/60001gIGMUMfEgH04ULHMPD00@410000
00<00@400P4300811@000`40000202Yf00IGMUMfEgH2E`060EL1E`5G6P4@00Qf4049E`P/200@0A1f
2B`H01UG01QfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUL500410@00008000<100000P00104000440003
0@00008003T/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;5MfEgIGMUMf0EL1
E`5G0EL06@4@0008;7H/MRaf;7H@0@0@MUMfEgIGMUMf;00/02`0;0T02@405UL1E`5G0EMf;7H/MRaf
;7H/MRaf;7H:;1P000MfEgIGMUMf019G67H500410@0000<000D100000@0200030@0000800`4202Yf
00IGMUMfEgH2E`060EL1E`5G6P4@00Qf4049E`P/200@0A1f2B`H01UG01QfEgIGMUMfEgIGMUMfEgIG
MUMfEgIGMUL400810@0000@000@100010P000`40000200030@00008003T/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;5MfEgIGMUMf0EL1E`5G0EL06@4@0008;7H/MRaf;7H@0@0@
MUMfEgIGMUMf;00/02`0;0T02@405UL1E`5G0EMf;7H/MRaf;7H/MRaf;7H:;1P000MfEgIGMUMf019G
67H500410@000005004000400P001040004400030@000080:WH01UMfEgIGMP9G00H1E`5G0ELJ0A00
27H@0@UG22`8010147H9;1P06EL067IGMUMfEgIGMUMfEgIGMUMfEgIGMUMfE`D00@4100000P020@@0
0P4500030@00008003T/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;5MfEgIG
MUMf0EL1E`5G0EL06@4@0008;7H/MRaf;7H@0@0@MUMfEgIGMUMf;00/02`0;0T02@405UL1E`5G0EMf
;7H/MRaf;7H/MRaf;7H:;1P000MfEgIGMUMf019G67H500410@0000l000<100000P0aMP0G;7H/MRaf
;7H/MRaf;7H/MUL1E`5G0EL04P4@00P/27H02EMfEgIGMUMfE`080@UG00H1E`5G0ELA0@Qf00P0;00/
02`0;1T06EL047IGMUMfEgIGMUMfEgIGMUL8MPD00@4100003`000`400002000G;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MR`02gH09baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`1E`5G0EL1E`0A0A00
2B`01WH/MRaf;09f00QGMUMfEgIGMPP1011GMUMfEgIGMP5G0EL1E`5G2@401eL1E`5G0EL027HQ01UG
67H500410@0000l000<100000P0aMP0G;7H/MRaf;7H/MRaf;7H/MUL1E`5G0EL04P4@00P/27H02EMf
EgIGMUMfE`080@UG00H1E`5G0ELA0@Qf00P0;00/02`0;1T06EL047IGMUMfEgIGMUMfEgIGMUL8MPD0
0@4100003`000`400002000G;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02gH09baf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MR`1E`5G0EL1E`0A0A002B`01WH/MRaf;09f00QGMUMfEgIGMPP1011GMUMf
EgIGMP5G0EL1E`5G2@401eL1E`5G0EL027HQ01UG67H500410@0000l00P43035f01L/MRaf;7H/MRaf
;7H/MRafE`5G0EL1E`0B0A0022`8MP09EgIGMUMfEgIG00P12EL01P5G0EL1Ea4127H0200/02`0;00/
6@0IE`0@MUMfEgIGMUMfEgIGMUMfE`Qf10020@40000?00030@00008001L/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;00;MP0W;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;05G0EL1E`5G01414009;006
MRaf;7H/0WH025MfEgIGMUMf204045MfEgIGMUMf0EL1E`5G0EL90@07E`5G0EL1E`08MR406ELHMPD0
0@4100003`000`400002035f01L/MRaf;7H/MRaf;7H/MRafE`5G0EL1E`0B0A0022`8MP09EgIGMUMf
EgIG00P12EL01P5G0EL1Ea4127H0200/02`0;00/6@0IE`0@MUMfEgIGMUMfEgIGMUMfE`Qf1@010@40
000?00030@00008001L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00;MP0W;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;05G0EL1E`5G01414009;006MRaf;7H/0WH025MfEgIGMUMf204045MfEgIGMUMf
0EL1E`5G0EL90@07E`5G0EL1E`08MR406ELHMPD00@4100003`000`400002045f00P/MRaf;7H/MQ41
25LY0007;7H/MRaf;009MQ0125LQ015G01QfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUL8MPD00@410000
3`000`4000020019;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00A0@QG:@0027H/MRaf;7H/27H@0@QG8@001gIGMUMfEgH0
2EL027IGMUMfEgIG6GH500410@0000l000<100000P11MP08;7H/MRaf;7HA0@QG:@001baf;7H/MR`0
2GH@0@QG8@0AE`0HMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG27H500410@0000l000<100000P00BBaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MR`04@48EbT000Qf;7H/MRaf;0Qf4048Eb4000MfEgIGMUMf00UG00QfEgIGMUMfEaUf
1@010@40000?00030@000080@GH022af;7H/MRaf4@48EbT000L/MRaf;7H/00Uf4048Eb404EL067IG
MUMfEgIGMUMfEgIGMUMfEgIGMUMfE`Qf1@010@40000?00810`00BBaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`04@48EbT0
00Qf;7H/MRaf;0Qf4048Eb4000MfEgIGMUMf00UG00QfEgIGMUMfEaUf10020@40000?00030@000080
@GH022af;7H/MRaf4@48EbT000L/MRaf;7H/00Uf4048Eb404EL067IGMUMfEgIGMUMfEgIGMUMfEgIG
MUMfE`Qf1@010@40000?00030@00008004T/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/014125LY0008MRaf;7H/MR`8MQ01
25LQ0007MUMfEgIGMP09E`08MUMfEgIGMULIMPD00@4100003`000`40000204Yf00QGMUMfEgIGMQ41
00iG0EL1E`5GMRaf;7H/MPT/6@08;0QG404I00P/00QGMUMfEgIGMPUG01QfEgIGMUMfEgIGMUMfEgIG
MUMfEgIGMUL8MPD00@4100003`000`40000200Qf00L/MRaf;7H/00Uf034/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf4048E`08;7H/MRaf;7H8;1T0010/02`0
;00/05MfEgIGMUMf404I00P/00MfEgIGMUMf00UG00QfEgIGMUMfEaUf1@010@40000?00030@000080
BWH025MfEgIGMUMf4@403UL1E`5G0EMf;7H/MRaf2B`I00P/25L@0AT022`025MfEgIGMUMf2EL067IG
MUMfEgIGMUMfEgIGMUMfEgIGMUMfE`Qf1@010@40000?00030@00008027H01baf;7H/MR`02GH0<Baf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02GH@0@QG00P/MRaf
;7H/MPP/6@0042`0;00/02`0EgIGMUMfEgH@0AT022`01gIGMUMfEgH02EL027IGMUMfEgIG6GH50041
0@0000l000<100000P1:MP08EgIGMUMfEgHA0@0>E`5G0EL1EgH/MRaf;7H9;1T022`8Ea016@08;008
EgIGMUMfEgH9E`0HMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG27H500410@0000l00P4300Qf00L/MRaf
;7H/00Uf034/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf
4048E`08;7H/MRaf;7H8;1T0010/02`0;00/05MfEgIGMUMf404I00P/00MfEgIGMUMf00UG00QfEgIG
MUMfEaUf10020@40000?00030@000080BWH025MfEgIGMUMf4@403UL1E`5G0EMf;7H/MRaf2B`I00P/
25L@0AT022`025MfEgIGMUMf2EL067IGMUMfEgIGMUMfEgIGMUMfEgIGMUMfE`Qf1@010@40000?0003
0@00008027H01baf;7H/MR`02GH0<Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MR`02GH@0@QG00P/MRaf;7H/MPP/6@0042`0;00/02`0EgIGMUMfEgH@0AT022`01gIG
MUMfEgH02EL027IGMUMfEgIG6GH500410@0000l000<100000P1:MP08EgIGMUMfEgHA0@0>E`5G0EL1
EgH/MRaf;7H9;1T022`8Ea016@08;008EgIGMUMfEgH9E`0HMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
27H500410@0000l000<100000P008Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;009MP0W
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf6048E`T/600042`0;00/02`0
0EL1E`5G0EL80AT000L/MRaf;7H/009f00IGMUMfEgH9E`08MUMfEgIGMULIMPD00@4100003`000`40
0002059f00QGMUMfEgIGMQT1011G0EL1E`5GMRaf;7H/MRaf6008;007E`5G0EL1E`090AT027H025Mf
EgIGMUMf2EL027IGMUMfEgIG2GH01eMfEgIGMUL027H500410@0000l000<100000P008Baf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;009MP0W;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/00Uf6048E`T/600042`0;00/02`00EL1E`5G0EL80AT000L/MRaf;7H/009f00IGMUMf
EgH9E`08MUMfEgIGMULIMPD00@4100003`000`400002059f00QGMUMfEgIGMQT1011G0EL1E`5GMRaf
;7H/MRaf6008;007E`5G0EL1E`090AT027H025MfEgIGMUMf2EL027IGMUMfEgIG2GH01eMfEgIGMUL0
27H500410@0000l00P43000Q;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf02L/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02GHH0@QG2B`H000@;00/02`0;001E`5G
0EL1E`P16@001baf;7H/MR`00WH01UMfEgIGMPUG00QfEgIGMUMfEaUf10020@40000?00030@000080
DWH025MfEgIGMUMf6@4045L1E`5G0EMf;7H/MRaf;7HH00P/00MG0EL1E`5G00T16@08MP08EgIGMUMf
EgH9E`08MUMfEgIGMUL9MP07EgIGMUMfE`08MPD00@4100003`000`400002000Q;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf02L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MR`02GHH0@QG2B`H000@;00/02`0;001E`5G0EL1E`P16@001baf;7H/MR`00WH01UMfEgIGMPUG
00QfEgIGMUMfEaUf1@010@40000?00030@000080DWH025MfEgIGMUMf6@4045L1E`5G0EMf;7H/MRaf
;7HH00P/00MG0EL1E`5G00T16@08MP08EgIGMUMfEgH9E`08MUMfEgIGMUL9MP07EgIGMUMfE`08MPD0
0@4100003`000`400002000a;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;009MP0W;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf6@401gH/
MRaf;7H00R`01`0/02`0;00022`800Qf25LI0007;7H/MRaf;00:MP0?EgIGMUMfEgIGMUMfEgIG01Uf
1@010@40000?00030@000080FgH03baf;7H/MRaGMUMfEgIGMP0I0@Qf00L0;00/02`0008/00Mf;7H/
MRaf00P000MfEgIGMUMf009G00L1E`5G0EL101T027H025MfEgIGMUMf2EL067IGMUMfEgIGMUMfEgIG
MUMfEgIGMUMfE`Qf1@010@40000?00030@000080034/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf02L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MR`02GHI0@07MRaf;7H/MP02;00702`0;00/0008;0P027H8EaT000L/MRaf;7H/00Yf00mGMUMf
EgIGMUMfEgIGMUL06GH500410@000003004100810P030@D000<100000P1KMP0?;7H/MRaf;5MfEgIG
MUMf01T127H01`0/02`0;0000R`01gH/MRaf;7H020001gIGMUMfEgH00UL01`5G0EL1E`406@08MP08
EgIGMUMfEgH9E`0HMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG27H500410@00008000<100001@000`40
000200030@000080034/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/00Uf02L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02GHI0@07MRaf;7H/
MP02;00702`0;00/0008;0P027H8EaT000L/MRaf;7H/00Yf00mGMUMfEgIGMUMfEgIGMUL06GH50041
0@0000<000<1000010000`40000200<10P1KMP0?;7H/MRaf;5MfEgIGMUMf01T127H01`0/02`0;000
0R`01gH/MRaf;7H020001gIGMUMfEgH00UL01`5G0EL1E`406@08MP08EgIGMUMfEgH9E`0HMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgIG27H400810@0000@000<100000`4500030@000080034/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf02L/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02GHI0@07MRaf;7H/MP02;00702`0;00/0008;0P027H8EaT0
00L/MRaf;7H/00Yf00mGMUMfEgIGMUMfEgIGMUL06GH500410@000005004000400P000`4000050003
0@000080FgH03baf;7H/MRaGMUMfEgIGMP0I0@Qf00L0;00/02`0008/00Mf;7H/MRaf00P000MfEgIG
MUMf009G00L1E`5G0EL101T027H025MfEgIGMUMf2EL067IGMUMfEgIGMUMfEgIGMUMfEgIGMUMfE`Qf
1@010@40000200810`040@@000<100000P00LRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MPQG4@48E`08MRaf;7H/MR`@0008MRaf;7H/MR`9MQ004GH03eMf
EgIGMUMfEgIGMUMfE`0IMPD00@4100003`000`40000206]f00L/MRaf;7H/00QG4@4045L1E`5G0EL1
;7H/MRaf;7H90007;00/02`0;009MP08EgIGMUMfEgH@00Qf025GMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUL047H500410@0000l000<100000P00LRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MPQG4@48E`08MRaf;7H/MR`@0008MRaf;7H/MR`9MQ00
4GH03eMfEgIGMUMfEgIGMUMfE`0IMPD00@4100003`000`40000206]f00L/MRaf;7H/00QG4@4045L1
E`5G0EL1;7H/MRaf;7H90007;00/02`0;009MP08EgIGMUMfEgH@00Qf025GMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUL047H500410@0000l000<100000P00LRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MPQG4@48E`08MRaf;7H/MR`@0008MRaf;7H/
MR`9MQ004GH03eMfEgIGMUMfEgIGMUMfE`0IMPD00@4100003`020@<0JgH01baf;7H/MR`025LA0@0@
E`5G0EL1E`4/MRaf;7H/MPT000L/02`0;00/00Uf00QGMUMfEgIGMQ0027H08EMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfE`0@MP@00P4100003`000`400002001b;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf25LA0@QG00Qf;7H/MRaf;10000Qf;7H/
MRaf;0Uf400AMP0?EgIGMUMfEgIGMUMfEgIG01Uf1@010@40000?00030@000080JgH01baf;7H/MR`0
25LA0@0@E`5G0EL1E`4/MRaf;7H/MPT000L/02`0;00/00Uf00QGMUMfEgIGMQ0027H08EMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfE`0@MPD00@4100003`000`400002000a;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;009MP0G;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MR`02GH0:Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/010127H8;0P0
014/02`0;00/005G0EL1E`5G0@0@0007;7H/MRaf;00:MP0?EgIGMUMfEgIGMUMfEgIG01Uf1@010@40
000?00030@000080LgH042af;7H/MRaf;7H/MRaf;7H@0@Qf22`800P/2@4@011f025GMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUL027H500410@0000l000<100000P00<Baf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02GH05baf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/00Uf02T/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00@0@Qf22`8000A
;00/02`0;001E`5G0EL1E`4040001baf;7H/MR`02WH03eMfEgIGMUMfEgIGMUMfE`0IMPD00@410000
3`000`40000207=f010/MRaf;7H/MRaf;7H/MRaf4048MPP/2008;0T1400@MP0QEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIG00Qf1@010@40000?00030@000080034/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf01L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;009MP0Y;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`04048MPP/20004B`0
;00/02`00EL1E`5G0EL1010000L/MRaf;7H/00Yf00mGMUMfEgIGMUMfEgIGMUL06GH500410@0000l0
00<100000P1cMP0@;7H/MRaf;7H/MRaf;7H/MQ0127H8;0P022`90A0047H08EMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfE`08MPD00@4100003`020@<0034/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf01L/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;009MP0Y
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`04048MPP/20004B`0;00/02`0
0EL1E`5G0EL1010000L/MRaf;7H/00Yf00mGMUMfEgIGMUMfEgIGMUL06GH400810@0000l000<10000
0P1cMP0@;7H/MRaf;7H/MRaf;7H/MQ0127H8;0P022`90A0047H08EMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfE`08MPD00@4100003`000`400002000a;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00IMP0i;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00QG2048E`P/40002@5G0EL1E`5G0@0@0007;7H/MRaf
;00BMP07EgIGMUMfE`0IMPD00@4100003`000`40000207]f00P/MRaf;7H/MPQG2048E`P/0100;00/
02`0;00/02`0;00/2@4@000YMRaf;7H/MRaGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUL0
47H500410@0000l000<100000P00<Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MR`06GH0>Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;008E`P125L8;10000T1E`5G0EL1E`4040001baf;7H/MR`04WH01eMfEgIGMUL0
6GH500410@0000l000<100000P1kMP08;7H/MRaf;7H8E`P125L8;00@02`0;00/02`0;00/02`0;0T1
4000:GH/MRaf;7H/EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG011f1@010@40000?0003
0@000080034/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/01Uf
03T/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`0
25L80@QG22`@00090EL1E`5G0EL1010000L/MRaf;7H/019f00MGMUMfEgIG01Uf1@010@40000?0003
0@000080NgH022af;7H/MRaf25L80@QG22`0400/02`0;00/02`0;00/02`90A0002Uf;7H/MRaf;5Mf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfE`0@MPD00@4100003`020@<0034/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/01Uf03T/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`025L80@QG22`@00090EL1E`5G
0EL1010000L/MRaf;7H/019f00MGMUMfEgIG01Uf10020@40000?00030@000080NgH022af;7H/MRaf
25L80@QG22`0400/02`0;00/02`0;00/02`90A0002Uf;7H/MRaf;5MfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfE`0@MPD00@4100003`000`4000020007;7H/MRaf;009MP0a;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;009MP11;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`040401gH/MRaf;7H0
0R`01P0/02`0;0T02@4H015f00MGMUMfEgIG01Uf1@010@40000?00030@000080RgH@0@0H;7H/MRaf
;7H0;00/02`0;00/02`0;00/2@4H0007EgIGMUMfE`09MP0AEgIGMUMfEgIGMUMfEgIGMUL047H50041
0@0000l000<100000P001baf;7H/MR`02GH0<Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MR`02GH0@Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/010100Mf;7H/MRaf008/00H0;00/02`900T1600AMP07
EgIGMUMfE`0IMPD00@4100003`000`40000208]f404062af;7H/MRaf02`0;00/02`0;00/02`0;0T1
60001eMfEgIGMUL02GH04EMfEgIGMUMfEgIGMUMfEgIG011f1@010@40000?00030@00008000L/MRaf
;7H/00Uf034/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf
044/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;00@0@07MRaf;7H/MP02;00602`0;00/2@090AP04GH01eMfEgIGMUL06GH500410@0000l0
00<100000P2;MQ0101P/MRaf;7H/MP0/02`0;00/02`0;00/02`90AP000MGMUMfEgIG00Uf015GMUMf
EgIGMUMfEgIGMUMfE`0@MPD00@4100003`020@<000L/MRaf;7H/00Uf034/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf044/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00@0@07MRaf;7H/MP02;006
02`0;00/2@090AP04GH01eMfEgIGMUL06GH400810@0000l000<100000P2;MQ0101P/MRaf;7H/MP0/
02`0;00/02`0;00/02`90AP000MGMUMfEgIG00Uf015GMUMfEgIGMUMfEgIGMUMfE`0@MPD00@410000
3`000`4000020007;7H/MRaf;009MP0a;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;009MP11;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`040401gH/MRaf;7H00R`01P0/02`0;0T02@4H015f00MGMUMf
EgIG01Uf1@010@40000?00030@000080RgH@0@QG22`80009E`5G0EL1E`5G00Qf400@MP0AEgIGMUMf
EgIGMUMfEgIGMUL047H500410@0000l000<100000P00@Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf04T/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;05G0EL1E`5G00P1
00QGMUMfEgIGMPP/20004@5G0EL1E`5G0Baf;7H/MRaf0100<GH500410@0000l000<100000P2;MQ01
25L8;0P000UG0EL1E`5G0EL027H@011f015GMUMfEgIGMUMfEgIGMUMfE`0@MPD00@4100003`000`40
00020011;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MR`02GH0BBaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/0EL1E`5G0EL0204025MfEgIGMUMf22`8000A0EL1E`5G0EL1;7H/
MRaf;7H0400aMPD00@41000000@00@4110020@D000<100000P2;MQ0125L8;0P000UG0EL1E`5G0EL0
27H@011f015GMUMfEgIGMUMfEgIGMUMfE`0@MPD00@41000010001040004200030@00008000<10000
0P00@Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/00Uf04T/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;05G0EL1E`5G00P100QGMUMfEgIGMPP/20004@5G0EL1E`5G0Baf;7H/
MRaf0100<GH500410@0000@000@100010P000`40000200<10P2;MQ0125L8;0P000UG0EL1E`5G0EL0
27H@011f015GMUMfEgIGMUMfEgIGMUMfE`0@MP@00P4100000P020@<000@1000110000`4000020011
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MR`02GH0BBaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/0EL1E`5G0EL0204025MfEgIGMUMf22`8000A0EL1E`5G0EL1;7H/MRaf;7H0
400aMPD00@41000010001040004200030@00008000<100000P2CMPP1011G0EL1E`5G0GIGMUMfEgIG
20004GIGMUMfEgIGMUMfEgIGMUMf010022`8MP0AEgIGMUMfEgIGMUMfEgIGMUL047H500410@000004
00410@@00P4500030@000080034/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/00Uf054/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`027H90@06E`5G0EL10UL01gIGMUMfEgH0
20002EMfEgIGMUMfE`08MQ0022`YMPD00@4100003`000`40000209=f204045L1E`5G0EL1MUMfEgIG
MUL8000AMUMfEgIGMUMfEgIGMUMfEgH04008;0Qf015GMUMfEgIGMUMfEgIGMUMfE`0@MPD00@410000
3`000`400002000a;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;009MP1A;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Qf2@401UL1E`5G0@9G00MfEgIGMUMf00P000UGMUMf
EgIGMUL027H@00P/:GH500410@0000l000<100000P2CMPP1011G0EL1E`5G0GIGMUMfEgIG20004GIG
MUMfEgIGMUMfEgIGMUMf010022`8MP0AEgIGMUMfEgIGMUMfEgIGMUL047H500410@0000l000<10000
0P00<Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02GH0DBaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;008MPT100IG0EL1E`42E`07MUMfEgIGMP080009EgIGMUMfEgIG00Qf
4008;2Uf1@010@40000?00810`2CMPP1011G0EL1E`5G0GIGMUMfEgIG20004GIGMUMfEgIGMUMfEgIG
MUMf010022`8MP0AEgIGMUMfEgIGMUMfEgIGMUL047H400810@0000l000<100000P00<Baf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02GH0DBaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;008MPT100IG0EL1E`42E`07MUMfEgIGMP080009EgIGMUMfEgIG00Qf4008;2Uf1@010@40
000?00030@000080TgH@0@0?MUMfEgIGMUL0;00/02`0008/00Qf;7H/MRaf;0Qf600YMPD00@410000
3`000`40000200Qf02T/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;009MP0G
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02GH0>Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00@0@Qf010/02`0;00/07H/MRaf;7H/2GHH02Uf1@010@40
000?00030@000080TgH@0@0?MUMfEgIGMUL0;00/02`0008/00Qf;7H/MRaf;0Qf600YMPD00@410000
3`000`40000200Qf02T/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;009MP0G
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02GH0>Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00@0@Qf010/02`0;00/07H/MRaf;7H/2GHH02Uf1@010@40
000?00030@000080TgH@0@0?MUMfEgIGMUL0;00/02`0008/00Qf;7H/MRaf;0Qf600YMPD00@410000
3`000`40000200Qf02T/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;009MP0G
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MR`02GH0>Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;00@0@Qf010/02`0;00/07H/MRaf;7H/2GHH02Uf1@010@40
000?00030@000080TgH@0@0?MUMfEgIGMUL0;00/02`0008/00Qf;7H/MRaf;0Qf600YMPD00@410000
3`020@<027H0:Baf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/00Uf01L/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;009MP0i;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf
;7H/MRaf;7H/MRaf;7H/MRaf;7H/010127H042`0;00/02`0MRaf;7H/MR`9MQP0:GH400810@0000l0
00<10000o`0=00410@0000l000<10000o`0=00410@0000l000<1000090000`40000W00030@0002L0
00<100009`000`40000W00030@0002L000<100004`010@40000?00030@00008000<100001P000`40
000500030@0000D000<100001P000`40000500030@0000H000<100001@000`40000500030@0000H0
00<100001@000`40000600030@0000D000<100001P000`40000500030@0000D000<100001P000`40
000500030@0000H000<100001@000`40000600030@0000D000<100001@000`40000600030@0000D0
00<100001P000`40000500030@0000D000<100001P000`40000500030@0000H000<100001@000`40
000200410@0000l0o`4A0@40003o02400000\
\>"],
  ImageRangeCache->{{{108, 395}, {442.75, 155.75}} -> {-14.274, 17.2188, 
  0.118757, 0.118757}}]
}, Open  ]],

Cell["\<\
In the lecture, we define a correlation function for evaluating the \
possibility that a patch centered at (x,y) has moved to (x+u,y+v).  Here is \
an implementation of the correlation function for the case of a 3 X 3 patch.  \
Question: What are advantages and disadvantages of using a larger patch.\
\>",
   "Text"],

Cell["\<\
ImageCorrelation[x_Integer,y_Integer,
                 u_Integer,v_Integer,
                 {n_Integer,m_Integer}] := 
        If[And[Greater[Min[x-1,x+u-1],0],
               Less[Max[x+1,x+u+1],n+1],
               Greater[Min[y-1,y+v-1],0],
               Less[Max[y+1,y+v+1],m+1]],
          Sum[(imageArrayOne[i,j] -
               imageArrayTwo[i+u,j+v])^2,
               {i,x-1,x+1},
               {j,y-1,y+1}],
          Infinity]\
\>", "Input"],

Cell["\<\
Let's see how each pixel (patch centered at (x,y) in the first \
image) correlates with itself (patch centered at (x,y) in the second \
image).\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\
DensityPlot[ImageCorrelation[Round[i],Round[j],0,0,myImageSize],
            {i,2,myImageSize[[1]]-1},{j,2,myImageSize[[1]]-1},
            {Mesh -> False, PlotPoints -> 32}] ;\
\>", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% DensityGraphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.0470822 0.0331565 -0.0470822 0.0331565 [
[.1187 -0.0125 -3 -9 ]
[.1187 -0.0125 3 0 ]
[.28448 -0.0125 -6 -9 ]
[.28448 -0.0125 6 0 ]
[.45027 -0.0125 -6 -9 ]
[.45027 -0.0125 6 0 ]
[.61605 -0.0125 -6 -9 ]
[.61605 -0.0125 6 0 ]
[.78183 -0.0125 -6 -9 ]
[.78183 -0.0125 6 0 ]
[.94761 -0.0125 -6 -9 ]
[.94761 -0.0125 6 0 ]
[ 0 0 -0.125 0 ]
[-0.0125 .1187 -6 -4.5 ]
[-0.0125 .1187 0 4.5 ]
[-0.0125 .28448 -12 -4.5 ]
[-0.0125 .28448 0 4.5 ]
[-0.0125 .45027 -12 -4.5 ]
[-0.0125 .45027 0 4.5 ]
[-0.0125 .61605 -12 -4.5 ]
[-0.0125 .61605 0 4.5 ]
[-0.0125 .78183 -12 -4.5 ]
[-0.0125 .78183 0 4.5 ]
[-0.0125 .94761 -12 -4.5 ]
[-0.0125 .94761 0 4.5 ]
[ 0 0 -0.125 0 ]
[ 0 1 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.1187 0 m
.1187 .00625 L
s
[(5)] .1187 -0.0125 0 1 Mshowa
.28448 0 m
.28448 .00625 L
s
[(10)] .28448 -0.0125 0 1 Mshowa
.45027 0 m
.45027 .00625 L
s
[(15)] .45027 -0.0125 0 1 Mshowa
.61605 0 m
.61605 .00625 L
s
[(20)] .61605 -0.0125 0 1 Mshowa
.78183 0 m
.78183 .00625 L
s
[(25)] .78183 -0.0125 0 1 Mshowa
.94761 0 m
.94761 .00625 L
s
[(30)] .94761 -0.0125 0 1 Mshowa
.125 Mabswid
.15186 0 m
.15186 .00375 L
s
.18501 0 m
.18501 .00375 L
s
.21817 0 m
.21817 .00375 L
s
.25133 0 m
.25133 .00375 L
s
.31764 0 m
.31764 .00375 L
s
.3508 0 m
.3508 .00375 L
s
.38395 0 m
.38395 .00375 L
s
.41711 0 m
.41711 .00375 L
s
.48342 0 m
.48342 .00375 L
s
.51658 0 m
.51658 .00375 L
s
.54973 0 m
.54973 .00375 L
s
.58289 0 m
.58289 .00375 L
s
.6492 0 m
.6492 .00375 L
s
.68236 0 m
.68236 .00375 L
s
.71552 0 m
.71552 .00375 L
s
.74867 0 m
.74867 .00375 L
s
.81499 0 m
.81499 .00375 L
s
.84814 0 m
.84814 .00375 L
s
.8813 0 m
.8813 .00375 L
s
.91446 0 m
.91446 .00375 L
s
.08554 0 m
.08554 .00375 L
s
.05239 0 m
.05239 .00375 L
s
.01923 0 m
.01923 .00375 L
s
.98077 0 m
.98077 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
0 .1187 m
.00625 .1187 L
s
[(5)] -0.0125 .1187 1 0 Mshowa
0 .28448 m
.00625 .28448 L
s
[(10)] -0.0125 .28448 1 0 Mshowa
0 .45027 m
.00625 .45027 L
s
[(15)] -0.0125 .45027 1 0 Mshowa
0 .61605 m
.00625 .61605 L
s
[(20)] -0.0125 .61605 1 0 Mshowa
0 .78183 m
.00625 .78183 L
s
[(25)] -0.0125 .78183 1 0 Mshowa
0 .94761 m
.00625 .94761 L
s
[(30)] -0.0125 .94761 1 0 Mshowa
.125 Mabswid
0 .15186 m
.00375 .15186 L
s
0 .18501 m
.00375 .18501 L
s
0 .21817 m
.00375 .21817 L
s
0 .25133 m
.00375 .25133 L
s
0 .31764 m
.00375 .31764 L
s
0 .3508 m
.00375 .3508 L
s
0 .38395 m
.00375 .38395 L
s
0 .41711 m
.00375 .41711 L
s
0 .48342 m
.00375 .48342 L
s
0 .51658 m
.00375 .51658 L
s
0 .54973 m
.00375 .54973 L
s
0 .58289 m
.00375 .58289 L
s
0 .6492 m
.00375 .6492 L
s
0 .68236 m
.00375 .68236 L
s
0 .71552 m
.00375 .71552 L
s
0 .74867 m
.00375 .74867 L
s
0 .81499 m
.00375 .81499 L
s
0 .84814 m
.00375 .84814 L
s
0 .8813 m
.00375 .8813 L
s
0 .91446 m
.00375 .91446 L
s
0 .08554 m
.00375 .08554 L
s
0 .05239 m
.00375 .05239 L
s
0 .01923 m
.00375 .01923 L
s
0 .98077 m
.00375 .98077 L
s
.25 Mabswid
0 0 m
0 1 L
s
.1187 .99375 m
.1187 1 L
s
.28448 .99375 m
.28448 1 L
s
.45027 .99375 m
.45027 1 L
s
.61605 .99375 m
.61605 1 L
s
.78183 .99375 m
.78183 1 L
s
.94761 .99375 m
.94761 1 L
s
.125 Mabswid
.15186 .99625 m
.15186 1 L
s
.18501 .99625 m
.18501 1 L
s
.21817 .99625 m
.21817 1 L
s
.25133 .99625 m
.25133 1 L
s
.31764 .99625 m
.31764 1 L
s
.3508 .99625 m
.3508 1 L
s
.38395 .99625 m
.38395 1 L
s
.41711 .99625 m
.41711 1 L
s
.48342 .99625 m
.48342 1 L
s
.51658 .99625 m
.51658 1 L
s
.54973 .99625 m
.54973 1 L
s
.58289 .99625 m
.58289 1 L
s
.6492 .99625 m
.6492 1 L
s
.68236 .99625 m
.68236 1 L
s
.71552 .99625 m
.71552 1 L
s
.74867 .99625 m
.74867 1 L
s
.81499 .99625 m
.81499 1 L
s
.84814 .99625 m
.84814 1 L
s
.8813 .99625 m
.8813 1 L
s
.91446 .99625 m
.91446 1 L
s
.08554 .99625 m
.08554 1 L
s
.05239 .99625 m
.05239 1 L
s
.01923 .99625 m
.01923 1 L
s
.98077 .99625 m
.98077 1 L
s
.25 Mabswid
0 1 m
1 1 L
s
.99375 .1187 m
1 .1187 L
s
.99375 .28448 m
1 .28448 L
s
.99375 .45027 m
1 .45027 L
s
.99375 .61605 m
1 .61605 L
s
.99375 .78183 m
1 .78183 L
s
.99375 .94761 m
1 .94761 L
s
.125 Mabswid
.99625 .15186 m
1 .15186 L
s
.99625 .18501 m
1 .18501 L
s
.99625 .21817 m
1 .21817 L
s
.99625 .25133 m
1 .25133 L
s
.99625 .31764 m
1 .31764 L
s
.99625 .3508 m
1 .3508 L
s
.99625 .38395 m
1 .38395 L
s
.99625 .41711 m
1 .41711 L
s
.99625 .48342 m
1 .48342 L
s
.99625 .51658 m
1 .51658 L
s
.99625 .54973 m
1 .54973 L
s
.99625 .58289 m
1 .58289 L
s
.99625 .6492 m
1 .6492 L
s
.99625 .68236 m
1 .68236 L
s
.99625 .71552 m
1 .71552 L
s
.99625 .74867 m
1 .74867 L
s
.99625 .81499 m
1 .81499 L
s
.99625 .84814 m
1 .84814 L
s
.99625 .8813 m
1 .8813 L
s
.99625 .91446 m
1 .91446 L
s
.99625 .08554 m
1 .08554 L
s
.99625 .05239 m
1 .05239 L
s
.99625 .01923 m
1 .01923 L
s
.99625 .98077 m
1 .98077 L
s
.25 Mabswid
1 0 m
1 1 L
s
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of gray image
p
.01923 .01923 translate
.96154 .96154 scale
32 string
32 32 8 [32 0 0 32 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  Mimage
000000000000000000000000000000000000000000000000
0000000000000000
000000000000000000000000000000000000000000000000
0000000000000000
000000000000000000000000000000000000000000000000
0000000000000000
000000000000000000000000000000000000000000000000
0000000000000000
000000000000000000000000000000000001010100000101
0101000000000000
000000000000000000000000000000050B0B060507080605
0505020000000000
00000000000000000000020202060B212C432F5553511E09
0908040000010000
00000000000000090909201A1B132044557E67ACA99E4111
11150B0701010100
00000000000000090909201A1B132044557E67ACA99E4111
11150B0701010100
0000000000000018181C45393524396287AEA6FFFFFEC973
73BA615D0A030201
000000000000001E1E305C5F4F3C4C68A8ACD2FFFFFFFFFE
FEFFC6BE14040302
000000000000001717425A81595A5B67BE97E6FFFFFFFFFF
FFFFFFFF17060502
0000000000000008084553906B6D706DC88CEDFFFFFFFFFF
FFFFFFFF11080703
00000000000000020237548476757A7BA679ABC3FFFFFFFF
FFFFFFE30A0A0803
0000000000000000001B42656F665F676A566070F5FFFFFF
FFFFFFA40A0B0803
00000000000000010106274C5B593640262813214E6E79FF
FFFFFF630A0B0803
000000000209100F0F091142504C1B120C0A050001022CFF
FFFFFF360A0A0702
0000000309101732324641515A572E0E1716110600003DFF
FFFFFF2209080502
000000030C1E3E5C5C7A746C6C573F141B22211203003AFF
FFFFFF2108070401
000000030A16316B6B969A74634834181C26312A18073DF6
F6FFD42008070401
00000000020E2A4C4C7A826E49312B221C1B2E3E341E4FD6
D6E29E1908060300
000000000000022222597569351C192A2A1D2444473873CE
CECD6E1007050200
000000000000000404354F6E53402D2A30241C333E428DD0
D0C4570805040100
000000000000000000142E4953594F402F21191D283480BE
BEAC4F0405040100
000000000000000000142E4953594F402F21191D283480BE
BEAC4F0405040100
00000000000000000000061F3D585855412B1912181C67B1
B1A3540304030100
0000000000000000000000000B2A445C5C42270E111449B5
B5A96F0203030100
00000000000000000000000000010F346D60410E10123DD0
D0C9970102020200
0000000000000000000000000000000E5D675A120E192CD9
D9D0B60402020200
000000000000000000000000000000013A5B5B27101920BB
BBC2B61201010100
000000000000000000000000000000001A4B4E361217187A
7A99962701010100
000000000000000000000000000000000339403E14100F38
38666C3706000000
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000180P10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdoojn_l00;34gTJ2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_[C^]:JVYZf][Jo^kP00
oiT0IP3o0020P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;0000?`00O`00_`0o000o
?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0
_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo?cooOcoo
_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no?gnoOgno
_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo
_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0_ol0oolo0?lo?oloOolo_olooomo
0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_mWEc:0Pl=WIfJj^[WMgMiW<09W<<iW<
IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`c
c<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co
<lcoI/coVLcoc<coool0<ol0I_l0VOl0c?l0oolc0?lc<olcI_lcVOlcc?lcoomV0?mVI_mVVOmVoonI
0?nI<onII_nIVOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<ool08@000?l08@000?l08@000?l08@0002d0
0`4W00<110020B@00`4300<190040@<00P4T00@10P030B@00`4400813000030000<100009@000`40
000200040@000B@000<100001@000`40000R00030@00008000@1000190000`40000500030@0002@0
00@100010P000`4000090000<0000`40000U00030@00008000@1000190000`40000500030@0002<0
00D100000@0200030@0002<000<1000010000`40000T00040@000@8000<100002@0002d00`4X0003
0@00008000@1000190000`40000200<19`001040004200030@0002@000<100000`4U00810`001040
004;0000;@000`40000X00030@00008000@1000190000`40000200030@0002@000@100010P001040
004S00040@000@8000<100009`001040004200030@0000T0000]00@19P020@D00P4T008110040B@0
0P4400819@020@<0104S00<110020@`0003o0240003o0240003o0240003o0240000?0?l14@410000
3`000`40000200030@0000H000<100001P000`40000600030@0000H000<100001P000`4000060003
0@0000H000<100001P000`40000600030@0000H000<100001P000`40000600030@0000H000<10000
1P000`40000600030@0000H000<100001P000`40000600030@0000H000<100001P000`4000060003
0@0000H000<100001P000`40000600030@0000H000<100001P000`40000600030@0000H000<10000
1P000`40000200410@0000l000<100007@000`40000Z00030@0002X000<10000:P000`40000Z0003
0@0002X000<100002`010@40000?00030@000?l03@010@40000?00030@000?l03@010@40000?0003
0@000?l03@010@40000?0081o`0=00810@0000l000<100000P3o0@H11@010@40000?00030@000080
o`460@D00@4100003`000`4000020?l11P4500410@0000l000<100000P3o0@H11@010@40000?0003
0@000080o`460@D00@4100003`000`4000020?l11P4500410@0000l000<100000P3o0@H11@010@40
000?00030@000080o`460@D00@4100003`020@<0o`460@@00P4100003`000`4000020?l11P450041
0@0000l000<100000P3o0@H11@010@40000?00030@000080o`460@D00@4100003`000`4000020?l1
1P4500410@0000l000<100000P3o0@H11@010@40000?00030@000080o`460@D00@4100003`000`40
00020?l11P4500410@0000l000<100000P3o0@H11@010@40000?00810`3o0@H110020@40000?0003
0@000080o`460@D00@4100003`000`4000020?l11P4500410@0000l000<100000P3o0@H11@010@40
000?00030@000080o`460@D00@4100003`000`4000020?l11P4500410@0000l000<100000P3o0@H1
1@010@40000700<11@000`4000020?l11P4500410@0000X000<100000P000`4000020?l11P450041
0@0000X000<100000P030@80o`460@@00P4100001`030@D000<100000P3o0@H11@010@4000070003
0@0000D000<100000P3o0@H11@010@40000700@110000`4000020?l11P4500410@0000l000<10000
0P3o0@H11@010@40000?00030@000080o`460@D00@4100003`000`4000020?l11P4500410@0000l0
00<100000P3o0@H11@010@40000?00030@000080o`460@D00@4100003`020@<0o`460@@00P410000
3`000`4000020?l11P4500410@0000l000<100000P3o0@H11@010@40000?00030@000080o`460@D0
0@4100003`000`4000020?l11P4500410@0000l000<100000P3o0@H11@010@40000?00030@000080
o`460@D00@4100003`000`4000020?l11P4500410@0000l000<100000P3o0@H11@010@40000?0081
0`3o0@H110020@40000?00030@000080o`460@D00@4100003`000`4000020?l11P4500410@0000l0
00<100000P3o0@H11@010@40000?00030@000080o`460@D00@4100003`000`40000207/1011G0EL1
E`5G0EL1E`5G0EL12EL01`5G0EL1E`4065MB0@D00@4100003`000`40000207X1011G0EL1E`5G0EL1
E`5G0EL12EL01`5G0EL1E`400UL07gIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf0EL1E`5G0EL0BP450041
0@0000l000<100000P1k0@0@E`5G0EL1E`5G0EL1E`5G0@UG00L1E`5G0EL101QGDP4500410@0000l0
00<100000P1j0@0@E`5G0EL1E`5G0EL1E`5G0@UG00L1E`5G0EL1009G01mfEgIGMUMfEgIGMUMfEgIG
MUMfEgIGMP5G0EL1E`5G04X11@010@40000?00810`1k0@0@E`5G0EL1E`5G0EL1E`5G0@UG00L1E`5G
0EL101QGDP4400810@0000l000<100000P1j0@0@E`5G0EL1E`5G0EL1E`5G0@UG00L1E`5G0EL1009G
01mfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP5G0EL1E`5G04X11@010@40000?00030@000080N`4045L1
E`5G0EL1E`5G0EL1E`49E`070EL1E`5G0@0HEe811@010@40000?00030@000080NP4045L1E`5G0EL1
E`5G0EL1E`49E`070EL1E`5G0@02E`0OMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH1E`5G0EL1E`1:0@D0
0@4100003`000`40000205<100MG0EL1E`5G01T100IG0EL1E`4BE`Uf01mGMUMfEgIG;7H/MRaf;7H/
MRaf;7H/MRaf;7H/MRaf00UGB@4500410@0000l000<100000P1B0@0XE`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@YG00IfEgIGMUL9MP08EgIGMUMfEgH@;008MRaf;7H/MR`:E`0F
0EL1E`5G0EL1E`5G0EL1E`5G0EL1Ec811@010@40000?00030@000080D`401eL1E`5G0EL06@401UL1
E`5G0A9G2GH07eMfEgIGMUL/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H02EM90@D00@4100003`000`40
0002058102QG0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL12UL01WIGMUMfE`Uf
00QGMUMfEgIGMQ0/00Qf;7H/MRaf;0YG01H1E`5G0EL1E`5G0EL1E`5G0EL1E`5G<P4500410@0000l0
00<100000P1C0@07E`5G0EL1E`0I0@06E`5G0EL14UL9MP0OEgIGMUMfEbaf;7H/MRaf;7H/MRaf;7H/
MRaf;7H/MP09EdT11@010@40000?00810`1B0@0XE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0@YG00IfEgIGMUL9MP08EgIGMUMfEgH@;008MRaf;7H/MR`:E`0F0EL1E`5G0EL1E`5G
0EL1E`5G0EL1Ec8110020@40000?00030@000080D`401eL1E`5G0EL06@401UL1E`5G0A9G2GH07eMf
EgIGMUL/MRaf;7H/MRaf;7H/MRaf;7H/MRaf;7H02EM90@D00@4100003`000`400002058102QG0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL12UL01WIGMUMfE`Uf00QGMUMfEgIGMQ0/
00Qf;7H/MRaf;0YG01H1E`5G0EL1E`5G0EL1E`5G0EL1E`5G<P4500410@0000l000<100000P1C0@07
E`5G0EL1E`0I0@06E`5G0EL14UL9MP0OEgIGMUMfEbaf;7H/MRaf;7H/MRaf;7H/MRaf;7H/MP09EdT1
1@010@40000?00030@000080DP40:5L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`4:E`06MUMfEgIG2GH025MfEgIGMUMf42`027H/MRaf;7H/2UL05P5G0EL1E`5G0EL1E`5G0EL1E`5G
0ELb0@D00@4100003`000`40000205<100MG0EL1E`5G01T100IG0EL1E`4BE`Uf01mGMUMfEgIG;7H/
MRaf;7H/MRaf;7H/MRaf;7H/MRaf00UGB@4500410@0000l000<100000P1B0@0XE`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@YG00IfEgIGMUL9MP08EgIGMUMfEgH@;008MRaf;7H/
MR`:E`0F0EL1E`5G0EL1E`5G0EL1E`5G0EL1Ec811@010@40000010010@4400811@000`40000205<1
00MG0EL1E`5G01T100IG0EL1E`4BE`Uf01mGMUMfEgIG;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf00UG
B@4500410@00008000<100000P001040004400030@000080DP40:5L1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`4:E`06MUMfEgIG2GH025MfEgIGMUMf42`027H/MRaf;7H/2UL05P5G
0EL1E`5G0EL1E`5G0EL1E`5G0ELb0@D00@4100000P000`40000200040@000@@00`4205<100MG0EL1
E`5G01T100IG0EL1E`4BE`Uf01mGMUMfEgIG;7H/MRaf;7H/MRaf;7H/MRaf;7H/MRaf00UGB@440081
0@00008000<100000P001040004400030@000080DP40:5L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`4:E`06MUMfEgIG2GH025MfEgIGMUMf42`027H/MRaf;7H/2UL05P5G0EL1E`5G
0EL1E`5G0EL1E`5G0ELb0@D00@4100000P000`40000200040@000@@000<100000P1B0@QG01l1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1009G00QfEgIGMUMfE`Qf010/MRaf;7H/MRaf;7H/MRaf
6009;00@EgIGMUMfEgIGMUMfEgIGMPP/011GMUMfEgIGMUMfEgIGMUMf8@4500410@000003004100D0
0P4500030@000080>P4065L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0AUG00L1E`5G0EL100QG00QfEgIG
MUMfE`Uf22`027H/MRaf;7H/60002B`0;00/02`0;00@MPP/011fEgIGMUMfEgIGMUMfEgIG8@450041
0@0000l000<100000P1B0@QG01l1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1009G00QfEgIG
MUMfE`Qf010/MRaf;7H/MRaf;7H/MRaf6009;00@EgIGMUMfEgIGMUMfEgIGMPP/011GMUMfEgIGMUMf
EgIGMUMf8@4500410@0000l000<100000P0j0@0HE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL16EL01`5G
0EL1E`4025L027IGMUMfEgIG2GH8;008MRaf;7H/MR`H0009;00/02`0;00/011f22`047IGMUMfEgIG
MUMfEgIGMULQ0@D00@4100003`000`400002058125L07`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`400UL027IGMUMfEgIG27H042af;7H/MRaf;7H/MRaf;7HH00T/011GMUMfEgIGMUMfEgIGMUMf
22`045MfEgIGMUMfEgIGMUMfEgHQ0@D00@4100003`000`40000203X101QG0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`4IE`070EL1E`5G0@08E`08MUMfEgIGMUL9MPP/00Qf;7H/MRaf;1P000T/02`0;00/02`0
47H8;00@MUMfEgIGMUMfEgIGMUMfEb411@010@40000?00030@000080DP48E`0O0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0@02E`08MUMfEgIGMUL8MP0@;7H/MRaf;7H/MRaf;7H/MQP02B`045Mf
EgIGMUMfEgIGMUMfEgH8;00@EgIGMUMfEgIGMUMfEgIGMR411@010@40000?00810`0j0@0HE`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL16EL01`5G0EL1E`4025L027IGMUMfEgIG2GH8;008MRaf;7H/MR`H0009
;00/02`0;00/011f22`047IGMUMfEgIGMUMfEgIGMULQ0@@00P4100003`000`400002058125L07`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`400UL027IGMUMfEgIG27H042af;7H/MRaf;7H/MRaf
;7HH00T/011GMUMfEgIGMUMfEgIGMUMf22`045MfEgIGMUMfEgIGMUMfEgHQ0@D00@4100003`000`40
000203X100iG0EL1E`5G0EL1E`5G0@YG00mfEgIGMUMfEgIGMUMfEgH06EL9MQ4/00H0;00/02`k0006
;00/02`02B`02EL1E`5G0EL1E`0H0@D00@4100003`000`40000204T100QG0EL1E`5G0@9G00mfEgIG
MUMfEgIGMUMfEgH025L01`5G0EL1E`402UL087IGMUMfEgIG;7H/MRaf;7H/MRaf;7H/MP0/02`0;00/
>@0@;2411@010@40000?00030@000080>P403UL1E`5G0EL1E`5G0EL12UL03gIGMUMfEgIGMUMfEgIG
MP0IE`Uf4B`01P0/02`0;3/000H/02`0;009;009E`5G0EL1E`5G01P11@010@40000?00030@000080
B@4025L1E`5G0EL10UL03gIGMUMfEgIGMUMfEgIGMP08E`070EL1E`5G0@0:E`0PMUMfEgIGMUL/MRaf
;7H/MRaf;7H/MRaf02`0;00/02`i010/8@4500410@0000l000<100000P0j0@0>E`5G0EL1E`5G0EL1
E`4:E`0?MUMfEgIGMUMfEgIGMUMf01UG2GHA;00602`0;00/>`001R`0;00/00T/00UG0EL1E`5G0EL0
604500410@0000l000<100000P190@08E`5G0EL1E`42E`0?MUMfEgIGMUMfEgIGMUMf00QG00L1E`5G
0EL100YG021fEgIGMUMfEbaf;7H/MRaf;7H/MRaf;7H0;00/02`0;3T042`Q0@D00@4100003`000`40
000203X100iG0EL1E`5G0EL1E`5G0@YG00mfEgIGMUMfEgIGMUMfEgH06EL9MQ4/00H0;00/02`k0006
;00/02`02B`02EL1E`5G0EL1E`0H0@D00@4100003`020@<0B@4025L1E`5G0EL10UL03gIGMUMfEgIG
MUMfEgIGMP08E`070EL1E`5G0@0:E`0PMUMfEgIGMUL/MRaf;7H/MRaf;7H/MRaf02`0;00/02`i010/
8@4400810@0000l000<100000P0j0@0>E`5G0EL1E`5G0EL1E`4:E`08MUMfEgIGMUL9MP0PEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH8;008MRaf;7H/MRaA0009E`5G0EL1E`5G01P11@010@40
000?00030@000080B@4:E`06MUMfEgIG2GH08EMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
E`08;0Qf00P0;00/02`0;4T08@4500410@0000l000<100000P0j0@0>E`5G0EL1E`5G0EL1E`4:E`08
MUMfEgIGMUL9MP0PEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH8;008MRaf;7H/MRaA0009
E`5G0EL1E`5G01P11@010@40000?00030@000080B@4:E`06MUMfEgIG2GH08EMfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfE`08;0Qf00P0;00/02`0;4T08@4500410@0000l000<100000P0j0@0>
E`5G0EL1E`5G0EL1E`4:E`08MUMfEgIGMUL9MP0PEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgH8;008MRaf;7H/MRaA0009E`5G0EL1E`5G01P11@010@40000?00030@000080B@4:E`06MUMfEgIG
2GH08EMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfE`08;0Qf00P0;00/02`0;4T08@450041
0@0000l000<100000P0j0@0>E`5G0EL1E`5G0EL1E`4:E`08MUMfEgIGMUL9MP0PEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgH8;008MRaf;7H/MRaA0009E`5G0EL1E`5G01P11@010@40000?0003
0@000080B@4:E`06MUMfEgIG2GH08EMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfE`08;0Qf
00P0;00/02`0;4T08@4500410@0000l00P4304T12EL03gIGMUMfEgIG;7H/MRaf;00RMP0@;00/02`0
;01f;7H/MRaf;54000UG0EL1E`5G0EL0604400810@0000l000<100000P190A5G27H08EMfEgIGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfE`08;0Qf00P0;00/02`0;4T08@4500410@0000l000<10000
0P190@UG00mfEgIGMUMfEbaf;7H/MR`08WH042`0;00/02`0MRaf;7H/MRaA0009E`5G0EL1E`5G01P1
1@010@40000?00030@000080B@4AE`Qf025GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUL0
22`8MP0802`0;00/02a902411@010@40000?00030@000080B@49E`0?MUMfEgIGMUL/MRaf;7H/029f
010/02`0;00/07H/MRaf;7H/D@002EL1E`5G0EL1E`0H0@D00@4100003`000`40000204T14EL8MP0Q
EgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG00P/27H0200/02`0;00/B@0Q0@D00@410000
3`000`40000204T12EL03gIGMUMfEgIG;7H/MRaf;00RMP0@;00/02`0;01f;7H/MRaf;54000UG0EL1
E`5G0EL0604500410@0000l000<100000P190A5G27H08EMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
EgIGMUMfE`08;0Qf00P0;00/02`0;4T08@4500410@0000l000<100000P190@UG00QfEgIGMUMfEbYf
00L/MRaf;7H/00Qf2B`01P0/02`0;4808@4500410@0000l00P4304T100QG0EL1E`5G0@UG27H03eMf
EgIGMUMfEgIGMUMfE`0BMP07;7H/MRaf;009MP08;7H/MRaf;7H8;3T000P/02`0;00/024110020@40
000?00030@000080B@49E`08MUMfEgIGMULZMP07;7H/MRaf;008MPT/00H0;00/02a202411@010@40
000?00030@000080B@4025L1E`5G0EL12EL8MP0?EgIGMUMfEgIGMUMfEgIG019f00L/MRaf;7H/00Uf
00P/MRaf;7H/MPP/>@0022`0;00/02`08@4500410@0000l000<100000P190@UG00QfEgIGMUMfEbYf
00L/MRaf;7H/00Qf2B`01P0/02`0;4808@4500410@0000l000<100000P190@08E`5G0EL1E`49E`Qf
00mGMUMfEgIGMUMfEgIGMUL04WH01baf;7H/MR`02GH022af;7H/MRaf22`i0008;00/02`0;00Q0@D0
0@4100003`000`40000204T12EL027IGMUMfEgIG:WH01baf;7H/MR`027H9;00602`0;00/@P0Q0@D0
0@4100003`000`40000204T100QG0EL1E`5G0@UG27H03eMfEgIGMUMfEgIGMUMfE`0BMP07;7H/MRaf
;009MP08;7H/MRaf;7H8;3T000P/02`0;00/02411@010@40000010010@4300<11@000`40000204X1
00QG0EL1E`5G0@QG00QfEgIGMUMfE`Uf01MGMUMfEgIGMUMfEgIGMUMfEgIGMUMfE`09MP0?EgIGMUMf
EgIGMUMfEgIG00Uf>@0022af;7H/MRaf8@4500410@00008000<100001@000`40000200030@000080
DP49E`0_MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH02EL047IG
MUMfEgIGMUMfEgIGMULi0008MRaf;7H/MR`Q0@D00@4100000P000`40000500030@0000800`4204X1
00QG0EL1E`5G0@QG00QfEgIGMUMfE`Uf01MGMUMfEgIGMUMfEgIGMUMfEgIGMUMfE`09MP0?EgIGMUMf
EgIGMUMfEgIG00Uf>@0022af;7H/MRaf8@4400810@00008000<100000P030@D000<100000P1B0@UG
02mfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP09E`0@MUMfEgIG
MUMfEgIGMUMfEcT000Qf;7H/MRaf;2411@010@40000200030@00008000<100001@000`40000204X1
00QG0EL1E`5G0@QG00QfEgIGMUMfE`Uf01MGMUMfEgIGMUMfEgIGMUMfEgIGMUMfE`09MP0?EgIGMUMf
EgIGMUMfEgIG00Uf>@0022af;7H/MRaf8@4500410@000003004100@0104400030@000080DP49E`0_
MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH02EL047IGMUMfEgIG
MUMfEgIGMULi0008MRaf;7H/MR`Q0@D00@4100003`000`40000204X100QG0EL1E`5G0@QG00QfEgIG
MUMfE`Uf01MGMUMfEgIGMUMfEgIGMUMfEgIGMUMfE`09MP0?EgIGMUMfEgIGMUMfEgIG00Uf>@0022af
;7H/MRaf8@4500410@0000l000<100000P1B0@UG02mfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMP09E`0@MUMfEgIGMUMfEgIGMUMfEcT000Qf;7H/MRaf;2411@010@40
000?00030@000080DP4025L1E`5G0EL125L03gIGMUMfEgIGMUMfEgIGMP0BE`0O0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0@09Ea5f800027IGMUMfEgIG8@4500410@0000l000<100000P1C0@06
E`5G0EL12UL05WIGMUMfEgIGMUMfEgIGMP5G0EL1E`4;E`0>0EL1E`5G0EL1E`5G0EL90@08E`5G0EL1
E`48E`08MUMfEgIGMUL9MR0000QGMUMfEgIGMR411@010@40000?00030@000080DP4025L1E`5G0EL1
25L03gIGMUMfEgIGMUMfEgIGMP0BE`0O0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@09Ea5f
800027IGMUMfEgIG8@4500410@0000l00P4305<100IG0EL1E`4:E`0FMUMfEgIGMUMfEgIGMUMf0EL1
E`5G0@]G00h1E`5G0EL1E`5G0EL1E`T100QG0EL1E`5G0@QG00QfEgIGMUMfE`Uf800025MfEgIGMUMf
8@4400810@0000l000<100000P1B0@08E`5G0EL1E`48E`0?MUMfEgIGMUMfEgIGMUMf019G01l1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL100UG4GHP0008MUMfEgIGMULQ0@D00@4100003`000`40
000205<100IG0EL1E`4:E`0FMUMfEgIGMUMfEgIGMUMf0EL1E`5G0@]G00h1E`5G0EL1E`5G0EL1E`T1
00QG0EL1E`5G0@QG00QfEgIGMUMfE`Uf800025MfEgIGMUMf8@4500410@0000l000<100000P1B0@08
E`5G0EL1E`48E`0?MUMfEgIGMUMfEgIGMUMf019G01l1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL100UG4GHP0008MUMfEgIGMULQ0@D00@4100003`000`40000205<100IG0EL1E`4:E`0FMUMfEgIG
MUMfEgIGMUMf0EL1E`5G0@]G00h1E`5G0EL1E`5G0EL1E`T100QG0EL1E`5G0@QG00QfEgIGMUMfE`Uf
800025MfEgIGMUMf8@4500410@0000l000<100000P0b0@07E`5G0EL1E`0I0@08E`5G0EL1E`48E`07
MUMfEgIGMP0:E`0@0EL1E`5G0EL1E`5G0EL1Ec4100QG0EL1E`5G0B0025LQ0@D00@4100003`000`40
000205X165M10@09E`5G0EL1E`5G020000P1E`5G0EL1Eb411@010@40000?00030@000080<P401eL1
E`5G0EL06@4025L1E`5G0EL125L01gIGMUMfEgH02UL0405G0EL1E`5G0EL1E`5G0ELa0@08E`5G0EL1
E`4P00QG8@4500410@0000l000<100000P1J0AQG@@402EL1E`5G0EL1E`0P00080EL1E`5G0ELQ0@D0
0@4100003`020@<0<P401eL1E`5G0EL06@4025L1E`5G0EL125L01gIGMUMfEgH02UL0405G0EL1E`5G
0EL1E`5G0ELa0@08E`5G0EL1E`4P00QG8@4400810@0000l000<100000P1J0AQG@@402EL1E`5G0EL1
E`0P00080EL1E`5G0ELQ0@D00@4100003`000`400002038100MG0EL1E`5G01T100QG0EL1E`5G0@QG
00MfEgIGMUMf00YG0101E`5G0EL1E`5G0EL1E`5G<@4025L1E`5G0EL18008Eb411@010@40000?0003
0@000080FP4HEd4100UG0EL1E`5G0EL08000205G0EL1E`5G8@4500410@0000l000<100000P0b0@07
E`5G0EL1E`0I0@08E`5G0EL1E`48E`07MUMfEgIGMP0:E`0@0EL1E`5G0EL1E`5G0EL1Ec4100QG0EL1
E`5G0B0025LQ0@D00@4100003`000`40000203T1011G0EL1E`5G0EL1E`5G0EL16UL01gIGMUMfEgH0
25L0205G0EL1E`5G>@402EL1E`5G0EL1E`0P00080EL1E`5G0ELQ0@D00@4100003`000`40000202X1
00iG0EL1E`5G0EL1E`5G0B9G01MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP02E`060EL1E`5G2`405eL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G01P12ELP0007E`5G0EL1E`0R0@D00@4100003`000`40000203T1
011G0EL1E`5G0EL1E`5G0EL16UL01gIGMUMfEgH025L0205G0EL1E`5G>@402EL1E`5G0EL1E`0P0008
0EL1E`5G0ELQ0@D00@4100003`000`40000202X100iG0EL1E`5G0EL1E`5G0B9G01MfEgIGMUMfEgIG
MUMfEgIGMUMfEgIGMP02E`060EL1E`5G2`405eL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G01P12ELP0007
E`5G0EL1E`0R0@D00@4100003`020@<0>@4045L1E`5G0EL1E`5G0EL1E`4JE`07MUMfEgIGMP08E`08
0EL1E`5G0ELi0@09E`5G0EL1E`5G020000P1E`5G0EL1Eb4110020@40000?00030@000080:P403UL1
E`5G0EL1E`5G0EL18UL05gIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf009G00H1E`5G0EL;0@0GE`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL06049Eb0000MG0EL1E`5G02811@010@40000?00030@000080>@4045L1
E`5G0EL1E`5G0EL1E`4JE`07MUMfEgIGMP08E`080EL1E`5G0ELi0@09E`5G0EL1E`5G020000P1E`5G
0EL1Eb411@010@40000?00030@000080:P403UL1E`5G0EL1E`5G0EL18UL05gIGMUMfEgIGMUMfEgIG
MUMfEgIGMUMf009G00H1E`5G0EL;0@0GE`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL06049Eb0000MG0EL1
E`5G02811@010@40000?00030@000080<@4065L1E`5G0EL1MUMfEgIGMUMfEgIGMUMfE`Uf01QGMUMf
EgIGMUMfEgIGMUMfEgIGMUMfEgH8E`080EL1E`5G0ELA0@0?E`5G0EL1E`5G0EL1E`5G01T100UG0EL1
E`5G0EL08000205G0EL1E`5G8@4500410@0000l000<100000P0Z0@06E`5G0EL12UL03WIGMUMfEgIG
MUMfEgIG8gH01UMfEgIGMPYG02P1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
4049Eb0000MG0EL1E`5G02811@010@40000?00030@000080<@4065L1E`5G0EL1MUMfEgIGMUMfEgIG
MUMfE`Uf01QGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH8E`080EL1E`5G0ELA0@0?E`5G0EL1E`5G0EL1
E`5G01T100UG0EL1E`5G0EL08000205G0EL1E`5G8@4500410@0000l000<100000P0Z0@06E`5G0EL1
2UL03WIGMUMfEgIGMUMfEgIG8gH01UMfEgIGMPYG02P1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G4049Eb0000MG0EL1E`5G02811@010@40000?00030@000080<@4065L1E`5G0EL1
MUMfEgIGMUMfEgIGMUMfE`Uf01QGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH8E`080EL1E`5G0ELA0@0?
E`5G0EL1E`5G0EL1E`5G01T100UG0EL1E`5G0EL08000205G0EL1E`5G8@4500410@0000l00P4302X1
00IG0EL1E`4:E`0>MUMfEgIGMUMfEgIGMULSMP06EgIGMUMf2UL0:05G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL@0@UG80001eL1E`5G0EL08P4400810@0000l000<100000P0a0@0H
E`5G0EL1E`5fEgIGMUMfEgIGMUMfEgIG2GH065MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMPQG00P1E`5G
0EL1Ea4100mG0EL1E`5G0EL1E`5G0EL06@402EL1E`5G0EL1E`0P00080EL1E`5G0ELQ0@D00@410000
3`000`40000202X100IG0EL1E`4:E`0>MUMfEgIGMUMfEgIGMULSMP06EgIGMUMf2UL0:05G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL@0@UG80001eL1E`5G0EL08P4500410@0000l0
00<100000P0a0@0HE`5G0EL1E`5fEgIGMUMfEgIGMUMfEgIG2WH05baf;7H/MRaGMUMfEgIGMUMfEgIG
MUMf00QG00P1E`5G0EL1Ea4101MG0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`0A0@09E`5G0EL1E`5G01P0
010/02`0;00/005G0EL1E`5G8@4500410@0000l000<100000P0Z0@06E`5G0EL12ELAMP0?;7H/MRaf
;7H/MRaf;7H/00Yf00IGMUMfEgHBE`0G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`402EL0405G0EL1E`5G
0EL1E`5G0EL80@UG6@003R`0;00/02aG0EL1E`5G8P4500410@0000l000<100000P0a0@0HE`5G0EL1
E`5fEgIGMUMfEgIGMUMfEgIG2WH05baf;7H/MRaGMUMfEgIGMUMfEgIGMUMf00QG00P1E`5G0EL1Ea41
01MG0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`0A0@09E`5G0EL1E`5G01P0010/02`0;00/005G0EL1E`5G
8@4500410@0000l000<100000P0Z0@06E`5G0EL12ELAMP0?;7H/MRaf;7H/MRaf;7H/00Yf00IGMUMf
EgHBE`0G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`402EL0405G0EL1E`5G0EL1E`5G0EL80@UG6@003R`0
;00/02aG0EL1E`5G8P4500410@000003004100810`020@D000<100000P0a0@0HE`5G0EL1E`5fEgIG
MUMfEgIGMUMfEgIG2WH05baf;7H/MRaGMUMfEgIGMUMfEgIGMUMf00QG00P1E`5G0EL1Ea4101MG0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`0A0@09E`5G0EL1E`5G01P0010/02`0;00/005G0EL1E`5G8@450041
0@00008000<100000P001040004400030@000080:P401UL1E`5G0@UG4GH03baf;7H/MRaf;7H/MRaf
;00:MP06EgIGMUMf4UL05`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL100UG0101E`5G0EL1E`5G0EL1E`5G
2049EaT000h/02`0;00/E`5G0EL1Eb811@010@40000300050@0000400P000`40000200<10P0a0@0H
E`5G0EL1E`5fEgIGMUMfEgIGMUMfEgIG2WH05baf;7H/MRaGMUMfEgIGMUMfEgIGMUMf00QG00P1E`5G
0EL1Ea4101MG0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`0A0@09E`5G0EL1E`5G01P0010/02`0;00/005G
0EL1E`5G8@4400810@0000@000@100010P000`40000200030@000080:P401UL1E`5G0@UG4GH03baf
;7H/MRaf;7H/MRaf;00:MP06EgIGMUMf4UL05`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL100UG0101E`5G
0EL1E`5G0EL1E`5G2049EaT000h/02`0;00/E`5G0EL1Eb811@010@4000001@010001008000@10001
10000`400002034100QG0EL1E`5G0A1G4GH025MfEgIGMUMf25L0605G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1Ea4101MG0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`090@UG020/02`0;00/02`0;00/02`0;00/02`0
;01f;7H/MRaf;2T11@010@400002008110020@D000<100000P0b0@06E`5G0EL14ELIMQ5G02L1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`404EL01`5G0EL1E`402UL0400/02`0;00/
02`0;00/02`8000?;7H/MRaf;7IG0EL1E`5G02811@010@40000?00030@000080<@4025L1E`5G0EL1
45LAMP08EgIGMUMfEgH8E`0H0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G4@405eL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G00T12EL082`0;00/02`0;00/02`0;00/02`0;00/07H/MRaf;7H/:@4500410@0000l0
00<100000P0b0@06E`5G0EL14ELIMQ5G02L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`404EL01`5G0EL1E`402UL0400/02`0;00/02`0;00/02`8000?;7H/MRaf;7IG0EL1E`5G0281
1@010@40000?00030@000080<@4025L1E`5G0EL145LAMP08EgIGMUMfEgH8E`0H0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G4@405eL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G00T12EL082`0;00/02`0;00/02`0
;00/02`0;00/07H/MRaf;7H/:@4500410@0000l000<100000P0b0@06E`5G0EL14ELIMQ5G02L1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`404EL01`5G0EL1E`402UL0400/02`0;00/
02`0;00/02`8000?;7H/MRaf;7IG0EL1E`5G02811@010@40000?00030@000080<@4025L1E`5G0EL1
45LAMP08EgIGMUMfEgH8E`0H0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G4@405eL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G00T12EL082`0;00/02`0;00/02`0;00/02`0;00/07H/MRaf;7H/:@4500410@0000l0
0P43038100IG0EL1E`4AEaUf4EL09`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0@0AE`070EL1E`5G0@0:E`0@02`0;00/02`0;00/02`0;0P000l/MRaf;7H/MUL1E`5G0EL08P440081
0@0000l000<100000P0i0@0aE`5G0EL1E`5G0EL1E`5G0GIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH1
E`5G0EL1E`0A0@0?E`5G0EL1E`5G0EL1E`5G00T100QG0EL1E`5G0A5G0101E`5G0EL1MUMfEgIGMUMf
62`025MfEgIGMUMf:@4500410@0000l000<100000P0j0@0>E`5G0EL1E`5G0EL1E`42E`08MUMfEgIG
MUL@MPUG02l1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@0IE`Uf
01P0;00/02`0;00/02`0;00/02`0;00/02`8MP07E`5G0EL1E`0R0@D00@4100003`000`40000203T1
035G0EL1E`5G0EL1E`5G0EL1MUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP5G0EL1E`5G014100mG0EL1
E`5G0EL1E`5G0EL02@4025L1E`5G0EL14EL0405G0EL1E`5fEgIGMUMfEgHH;008EgIGMUMfEgHY0@D0
0@4100003`000`40000203X100iG0EL1E`5G0EL1E`5G0@9G00QfEgIGMUMfEa1f2EL0;`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL101UG2GH0600/02`0;00/02`0;00/
02`0;00/02`0;0Qf00MG0EL1E`5G02811@010@40000?00030@000080>@40<EL1E`5G0EL1E`5G0EL1
E`5fEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf0EL1E`5G0EL04@403eL1E`5G0EL1E`5G0EL1E`090@08
E`5G0EL1E`4AE`0@0EL1E`5G0GIGMUMfEgIGMQP/00QGMUMfEgIGMRT11@010@40000?00030@000080
>P403UL1E`5G0EL1E`5G0EL10UL027IGMUMfEgIG47H9E`0_0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`406EL9MP0H02`0;00/02`0;00/02`0;00/02`0;00/27H01eL1
E`5G0EL08P4500410@0000l000<100000P0i0@0aE`5G0EL1E`5G0EL1E`5G0GIGMUMfEgIGMUMfEgIG
MUMfEgIGMUMfEgH1E`5G0EL1E`0A0@0?E`5G0EL1E`5G0EL1E`5G00T100QG0EL1E`5G0A5G0101E`5G
0EL1MUMfEgIGMUMf62`025MfEgIGMUMf:@4500410@0000l000<100000P0j0@0>E`5G0EL1E`5G0EL1
E`42E`08MUMfEgIGMUL@MPUG02l1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0@0IE`Uf01P0;00/02`0;00/02`0;00/02`0;00/02`8MP07E`5G0EL1E`0R0@D00@410000
3`020@<0B@4025L1E`5G0EL12UL01gIGMUMfEgH045L0805G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G2@4045L1E`5G0EL1E`5G0EL1E`48E`Uf010/02`0;00/02`0;00/02`022`8EbT110020@40
000?00030@000080B@4AE`Uf00IGMUMfEgH:E`0?0EL1E`5G0EL1E`5G0EL100UG00l1E`5G0EL1E`5G
0EL1E`406EL0:GH/MRaf;7H/MP0/02`0;00/02`0;00/02`0;00/02`0;7IGMUMfEgIG02T11@010@40
000?00030@000080B@4025L1E`5G0EL12UL01gIGMUMfEgH045L0805G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G2@4045L1E`5G0EL1E`5G0EL1E`48E`Uf010/02`0;00/02`0;00/02`022`8EbT1
1@010@40000?00030@000080B@4AE`Uf00IGMUMfEgH:E`0?0EL1E`5G0EL1E`5G0EL100UG00l1E`5G
0EL1E`5G0EL1E`406EL0:GH/MRaf;7H/MP0/02`0;00/02`0;00/02`0;00/02`0;7IGMUMfEgIG02T1
1@010@40000?00030@000080B@4025L1E`5G0EL12UL01gIGMUMfEgH045L0805G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G2@4045L1E`5G0EL1E`5G0EL1E`48E`Uf010/02`0;00/02`0;00/02`0
22`8EbT11@010@40000?00030@000080B@4AE`Uf00IGMUMfEgH:E`0?0EL1E`5G0EL1E`5G0EL100UG
00l1E`5G0EL1E`5G0EL1E`406EL0:GH/MRaf;7H/MP0/02`0;00/02`0;00/02`0;00/02`0;7IGMUMf
EgIG02T11@010@40000?00030@000080B@4025L1E`5G0EL12UL01gIGMUMfEgH045L0805G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G2@4045L1E`5G0EL1E`5G0EL1E`48E`Uf010/02`0;00/02`0
;00/02`022`8EbT11@010@40000?00030@000080B@4AE`Uf00IGMUMfEgH:E`0?0EL1E`5G0EL1E`5G
0EL100UG00l1E`5G0EL1E`5G0EL1E`406EL0:GH/MRaf;7H/MP0/02`0;00/02`0;00/02`0;00/02`0
;7IGMUMfEgIG02T11@010@40000?00030@000080B@4025L1E`5G0EL12UL01gIGMUMfEgH045L0805G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G2@4045L1E`5G0EL1E`5G0EL1E`48E`Uf010/02`0
;00/02`0;00/02`022`8EbT11@010@40000?00810`1:0@0@E`5G0EL1E`5G0EL1E`5G0@QG00mfEgIG
MUMfEgIGMUMfEgH04UL09`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@09E`Uf
62`8EbT110020@40000?00030@000080D`401UL1E`5G0A9G00MfEgIGMUMf019G00h1E`5G0EL1E`5G
0EL1Ea41011G0EL1E`5G0EL1E`5G0EL12GH@;008MRaf;7H/MR`8EbT11@010@40000?00030@000080
BP4045L1E`5G0EL1E`5G0EL1E`48E`0?MUMfEgIGMUMfEgIGMUMf019G02L1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`402EL9MQP/25LY0@D00@4100003`000`40000205<100IG0EL1
E`4BE`07MUMfEgIGMP0BE`0>0EL1E`5G0EL1E`5G0ELA0@0@E`5G0EL1E`5G0EL1E`5G0@Uf42`027H/
MRaf;7H/25LY0@D00@4100003`000`40000204X1011G0EL1E`5G0EL1E`5G0EL125L03gIGMUMfEgIG
MUMfEgIGMP0BE`0W0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL100UG2GHH;0QG
:@4500410@0000l000<100000P1C0@06E`5G0EL14UL01gIGMUMfEgH04UL03P5G0EL1E`5G0EL1E`5G
4@4045L1E`5G0EL1E`5G0EL1E`49MQ0/00Qf;7H/MRaf;0QG:@4500410@0000l000<100000P1:0@0@
E`5G0EL1E`5G0EL1E`5G0@QG00mfEgIGMUMfEgIGMUMfEgH04UL09`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0@09E`Uf62`8EbT11@010@40000?00030@000080D`401UL1E`5G0A9G
00MfEgIGMUMf019G00h1E`5G0EL1E`5G0EL1Ea41011G0EL1E`5G0EL1E`5G0EL12GH@;008MRaf;7H/
MR`8EbT11@010@40000?00030@000080BP4045L1E`5G0EL1E`5G0EL1E`48E`0?MUMfEgIGMUMfEgIG
MUMf019G02L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`402EL9MQP/25LY0@D0
0@4100003`020@<0D`401UL1E`5G0A9G00MfEgIGMUMf019G00h1E`5G0EL1E`5G0EL1Ea41011G0EL1
E`5G0EL1E`5G0EL12GH@;008MRaf;7H/MR`8EbT110020@40000?00030@000080BP4045L1E`5G0EL1
E`5G0EL1E`48E`0?MUMfEgIGMUMfEgIGMUMf019G02L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`402EL9MQP/25LY0@D00@4100003`000`40000205<100IG0EL1E`4BE`07MUMfEgIG
MP0BE`0>0EL1E`5G0EL1E`5G0ELA0@0@E`5G0EL1E`5G0EL1E`5G0@Uf42`027H/MRaf;7H/25LY0@D0
0@4100003`000`40000204X1011G0EL1E`5G0EL1E`5G0EL125L03gIGMUMfEgIGMUMfEgIGMP0BE`0W
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL100UG2GHH;0QG:@4500410@0000l0
00<100000P1C0@06E`5G0EL14UL01gIGMUMfEgH04UL03P5G0EL1E`5G0EL1E`5G4@4045L1E`5G0EL1
E`5G0EL1E`49MQ0/00Qf;7H/MRaf;0QG:@4500410@0000l000<100000P1:0@0@E`5G0EL1E`5G0EL1
E`5G0@QG00mfEgIGMUMfEgIGMUMfEgH04UL09`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0@09E`Uf62`8EbT11@010@40000?00030@000080D`401UL1E`5G0A9G00MfEgIGMUMf019G
00h1E`5G0EL1E`5G0EL1Ea41011G0EL1E`5G0EL1E`5G0EL12GH@;008MRaf;7H/MR`8EbT11@010@40
00000`010@020@800`4500030@000080FP4025L1E`5G0EL125L06GIGMUMfEgIGMUMfEgIGMUMfEgIG
MUMfEgH025L09`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@02E`08MUMfEgIG
MULA;006MRaf;7H/0WH01eMfEgIGMUL0:@4500410@00008000<100001@000`40000200030@000080
H`401UL1E`5G0@9G00mfEgIGMUMfEgIGMUMfEgH04UL01P5G0EL1Eb4100UfEgIGMUMfEgH042`027H/
MRaf;7H/25LY0@D00@4100000`000`40000400030@0000800`4205X100QG0EL1E`5G0@QG01UfEgIG
MUMfEgIGMUMfEgIGMUMfEgIGMUMf00QG02L1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`400UL027IGMUMfEgIG4B`01WH/MRaf;09f00MGMUMfEgIG02T110020@40000400030@0000<1
1@000`40000206<100IG0EL1E`42E`0?MUMfEgIGMUMfEgIGMUMf019G00H1E`5G0ELQ0@09MUMfEgIG
MUMf010/00Qf;7H/MRaf;0QG:@4500410@000005004000400P000`40000500030@000080FP4025L1
E`5G0EL125L06GIGMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH025L09`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1E`5G0@02E`08MUMfEgIGMULA;006MRaf;7H/0WH01eMfEgIGMUL0:@450041
0@0000800P4300@110000`40000206<100IG0EL1E`42E`0?MUMfEgIGMUMfEgIGMUMf019G00H1E`5G
0ELQ0@09MUMfEgIGMUMf010/00Qf;7H/MRaf;0QG:@4500410@0000l000<100000P1J0@08E`5G0EL1
E`48E`0IMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP08E`0W0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G
0EL1E`5G0EL1E`5G0EL1009G00QfEgIGMUMfEa4/00If;7H/MR`2MP07EgIGMUMfE`0Y0@D00@410000
3`000`40000206<100IG0EL1E`42E`0?MUMfEgIGMUMfEgIGMUMf019G00H1E`5G0ELQ0@09MUMfEgIG
MUMf010/00Qf;7H/MRaf;0QG:@4500410@0000l000<100000P1Z0@08E`5G0EL1E`48E`0AMUMfEgIG
MUMfEgIGMUMfEgH025L0205G0EL1E`5G2@403UL1E`5G0EL1E`5G0EL12ULH;0Qf:@4500410@0000l0
00<100000P1[0@06E`5G0EL12UL03gIGMUMfEgIGMUMfEgIGMP0:E`060EL1E`5G6@49Ea0/011f;7H/
MRaf;5MfEgIGMUMf:@4500410@0000l000<100000P1Z0@08E`5G0EL1E`48E`0AMUMfEgIGMUMfEgIG
MUMfEgH025L0205G0EL1E`5G2@403UL1E`5G0EL1E`5G0EL12ULH;0Qf:@4500410@0000l00P4306/1
00IG0EL1E`4:E`0?MUMfEgIGMUMfEgIGMUMf00YG00H1E`5G0ELI0@UG42`047H/MRaf;7H/EgIGMUMf
EgHY0@@00P4100003`000`40000206X100QG0EL1E`5G0@QG015fEgIGMUMfEgIGMUMfEgIGMP08E`08
0EL1E`5G0EL90@0>E`5G0EL1E`5G0EL1E`4:EaP/27HY0@D00@4100003`000`40000206/100IG0EL1
E`4:E`0?MUMfEgIGMUMfEgIGMUMf00YG00H1E`5G0ELI0@UG42`047H/MRaf;7H/EgIGMUMfEgHY0@D0
0@4100003`000`40000206X100QG0EL1E`5G0@QG015fEgIGMUMfEgIGMUMfEgIGMP08E`080EL1E`5G
0EL90@0>E`5G0EL1E`5G0EL1E`4:EaP/27HY0@D00@4100003`000`40000206/100IG0EL1E`4:E`0?
MUMfEgIGMUMfEgIGMUMf00YG00H1E`5G0ELI0@UG42`047H/MRaf;7H/EgIGMUMfEgHY0@D00@410000
3`000`40000207X12EL8MP08EgIGMUMfEgH8E`T100iG0EL1E`5G0EL1E`5G0@YG01L0;00/02`0;00/
02`0;00/02`0;00/0002;007MRaf;7H/MP0Y0@D00@4100003`000`40000207/101MG0EL1E`5G0GIG
MUMfEgIGMUMfEgIGMP09EaP101UG0EL1E`5G0EL/02`0;00/02`0;00/02`000P/27HY0@D00@410000
3`000`40000207X12EL8MP08EgIGMUMfEgH8E`T100iG0EL1E`5G0EL1E`5G0@YG01L0;00/02`0;00/
02`0;00/02`0;00/0002;007MRaf;7H/MP0Y0@D00@4100003`000`40000207/101MG0EL1E`5G0GIG
MUMfEgIGMUMfEgIGMP09EaP101UG0EL1E`5G0EL/02`0;00/02`0;00/02`000P/27HY0@D00@410000
3`020@<0NP49E`Qf00QGMUMfEgIGMPQG2@403UL1E`5G0EL1E`5G0EL12UL05`0/02`0;00/02`0;00/
02`0;00/02`0008/00Mf;7H/MRaf02T110020@40000?00030@000080N`405eL1E`5G0EL1MUMfEgIG
MUMfEgIGMUMf00UG60406EL1E`5G0EL1Eb`0;00/02`0;00/02`0;00022`8MRT11@010@40000?0003
0@000080NP49E`Qf00QGMUMfEgIGMPQG2@403UL1E`5G0EL1E`5G0EL12UL05`0/02`0;00/02`0;00/
02`0;00/02`0008/00Mf;7H/MRaf02T11@010@40000?00030@000080N`405eL1E`5G0EL1MUMfEgIG
MUMfEgIGMUMf00UG60406EL1E`5G0EL1Eb`0;00/02`0;00/02`0;00022`8MRT11@010@40000?0003
0@000080P`4085MfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMP5G0EL1E`5G2@409eL1E`5G0EL1E`5G0EL1
E`40;00/02`0;00/02`0;00/02`0;00/0009;2T11@010@40000?00030@000080P`4067IGMUMfEgIG
MUMfEgIGMUMfEgIGMUMfEaP1025G0EL1E`5G0EL/02`0;00/02`0;00/02`0;00/02`0;00022`Y0@D0
0@4100003`000`40000208<1021GMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgH1E`5G0EL1E`T102MG0EL1
E`5G0EL1E`5G0EL102`0;00/02`0;00/02`0;00/02`0;0002B`Y0@D00@4100003`000`40000208<1
01QfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMULH0@0QE`5G0EL1E`5G;00/02`0;00/02`0;00/02`0;00/
02`000P/:@4500410@0000l000<100000P230@0PEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf0EL1E`5G
0EL90@0WE`5G0EL1E`5G0EL1E`5G0@0/02`0;00/02`0;00/02`0;00/02`000T/:@4500410@0000l0
0P4308<101QfEgIGMUMfEgIGMUMfEgIGMUMfEgIGMULH0@0QE`5G0EL1E`5G;00/02`0;00/02`0;00/
02`0;00/02`000P/:@4400810@0000l000<100000P230@0PEgIGMUMfEgIGMUMfEgIGMUMfEgIGMUMf
0EL1E`5G0EL90@0WE`5G0EL1E`5G0EL1E`5G0@0/02`0;00/02`0;00/02`0;00/02`000T/:@450041
0@0000l000<100000P230@0HMUMfEgIGMUMfEgIGMUMfEgIGMUMfEgIG60408EL1E`5G0EL1Eb`0;00/
02`0;00/02`0;00/02`0;00/0008;2T11@010@40000?00030@000080P`4085MfEgIGMUMfEgIGMUMf
EgIGMUMfEgIGMP5G0EL1E`5G2@409eL1E`5G0EL1E`5G0EL1E`40;00/02`0;00/02`0;00/02`0;00/
0009;2T11@010@40000?00030@000080P`405eL1E`5G0EL1MUMfEgIGMUMfEgIGMUMf009G00H1E`5G
0ELA0@09E`5G0EL1E`5G020/:@4500410@0000l000<100000P230@UG031fEgIGMUMfEgIGMUMfEgH1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`4@;00702`0;00/0009;007E`5G0EL1E`0R0@D0
0@4100003`000`40000208<101MG0EL1E`5G0GIGMUMfEgIGMUMfEgIGMP02E`060EL1E`5G4@402EL1
E`5G0EL1E`0P;2T11@010@40000?00030@000080P`49E`0`MUMfEgIGMUMfEgIGMUMf0EL1E`5G0EL1
E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL142`01`0/02`0;0002B`01eL1E`5G0EL08P4500410@0000l0
00<100000P230@0GE`5G0EL1E`5fEgIGMUMfEgIGMUMfEgH00UL01P5G0EL1Ea4100UG0EL1E`5G0EL0
82`Y0@D00@4100003`020@<0P`49E`0`MUMfEgIGMUMfEgIGMUMf0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0EL1E`5G0EL142`01`0/02`0;0002B`01eL1E`5G0EL08P4400810@0000l000<100000P230@0G
E`5G0EL1E`5fEgIGMUMfEgIGMUMfEgH00UL01P5G0EL1Ea4100UG0EL1E`5G0EL082`Y0@D00@410000
3`000`40000208<12EL0<7IGMUMfEgIGMUMfEgIGMP5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1
E`5G0A0/00L0;00/02`000T/00MG0EL1E`5G02811@010@40000?00030@000080R`4AE`060EL1E`5G
6P4AMP07;7H/MRaf;008MP080EL1E`5G0ELQ0@D00@4100003`000`40000208@100IG0EL1E`4IE`0I
0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@0@MP0G;7H/MRaf;7H/MRaf;7H/MUL1E`5G0EL08P450041
0@0000l000<100000P2;0A5G00H1E`5G0ELJ0A5f00L/MRaf;7H/00Qf00P1E`5G0EL1Eb411@010@40
000?00030@000080Q0401UL1E`5G0AUG01T1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1011f01L/MRaf
;7H/MRaf;7H/MRafE`5G0EL1E`0R0@D00@41000000@00@4110020@D000<100000P2;0A5G00H1E`5G
0ELJ0A5f00L/MRaf;7H/00Qf00P1E`5G0EL1Eb411@010@40000400040@000@8000<100000P000`40
000208@100IG0EL1E`4IE`0I0EL1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0@0@MP0G;7H/MRaf;7H/MRaf
;7H/MUL1E`5G0EL08P4500410@0000@000@100010P000`40000200<10P2;0A5G00H1E`5G0ELJ0A5f
00L/MRaf;7H/00Qf00P1E`5G0EL1Eb4110020@40000200810`001040004400030@000080Q0401UL1
E`5G0AUG01T1E`5G0EL1E`5G0EL1E`5G0EL1E`5G0EL1011f01L/MRaf;7H/MRaf;7H/MRafE`5G0EL1
E`0R0@D00@41000010001040004200030@00008000<100000P2;0@08E`5G0EL1E`49E`060EL1E`5G
6`403UL1E`5G0EL1E`5G0EL10UL05gIGMUMfEgIGMUMfEgIGMP5G0EL1E`5G02411@010@4000001001
0@4400811@000`40000208/165L0405G0EL1E`5G0EL1E`5G0EL90A1G00QfEgIGMUMfE`Qf25LQ0@D0
0@4100003`000`40000208/100QG0EL1E`5G0@UG00H1E`5G0ELK0@0>E`5G0EL1E`5G0EL1E`42E`0G
MUMfEgIGMUMfEgIGMUMf0EL1E`5G0EL08@4500410@0000l000<100000P2;0AQG0101E`5G0EL1E`5G
0EL1E`5G2@4@E`08MUMfEgIGMUL8MPQG8@4500410@0000l000<100000P2;0@08E`5G0EL1E`49E`06
0EL1E`5G6`403UL1E`5G0EL1E`5G0EL10UL05gIGMUMfEgIGMUMfEgIGMP5G0EL1E`5G02411@010@40
000?00030@000080R`4HE`0@0EL1E`5G0EL1E`5G0EL1E`T145L027IGMUMfEgIG27H8Eb411@010@40
000?00030@000080R`4025L1E`5G0EL12EL01P5G0EL1Ea/100iG0EL1E`5G0EL1E`5G0@9G01MfEgIG
MUMfEgIGMUMfEgH1E`5G0EL1E`0Q0@D00@4100003`020@<0R`4HE`0@0EL1E`5G0EL1E`5G0EL1E`T1
45L027IGMUMfEgIG27H8Eb4110020@40000?00030@000?l03@010@40000?00030@000?l03@010@40
000?00030@0001d000<10000:P000`40000Z00030@0002X000<10000:P000`40000Z00030@0000/0
0@4100003`000`40000200030@0000H000<100001P000`40000600030@0000H000<100001P000`40
000600030@0000H000<100001P000`40000600030@0000H000<100001P000`40000600030@0000H0
00<100001P000`40000600030@0000H000<100001P000`40000600030@0000H000<100001P000`40
000600030@0000H000<100001P000`40000600030@0000H000<100001P000`40000600030@0000H0
00<100001P000`40000200410@0000l0o`4A0@40003o02400000\
\>"],
  ImageRangeCache->{{{108, 395}, {416.875, 129.875}} -> {-12.2886, 14.2979, 
  0.111095, 0.111095}}]
}, Open  ]],

Cell["\<\
Here is the subroutine that searches for the best match (defined as \
the offset (u,v) from the pixel (x,y) with the lowest correlation \
value).\
\>", "Text"],

Cell["\<\
BestMatch[x_Integer,y_Integer,w_Integer,{n_Integer,m_Integer}] := 
    Module[{b = Infinity, c, p = {x,y}},
        Do[c = ImageCorrelation[x,y,u,v,{n,m}];
           If[c < b, b = c ; p = {x+u,y+v}],
           {u,-w,w},
           {v,-w,w}] ;
        Return[p]]\
\>", "Input"],

Cell["\<\
To display the optical flow field, we need the following graphics \
utility for displaying arrows (called needles or flow vectors when we are \
referring to optical flow fields).\
\>", "Text"],

Cell["<<Graphics`Arrow`", "Input"],

Cell["\<\
OpticalFlow[w_Integer,{n_Integer,m_Integer}] :=
   Table[Arrow[{i,j},BestMatch[i,j,w,{n,m}],
               HeadLength -> 0.02],
         {i,1,n},
         {j,1,m}]\
\>", "Input"],

Cell[CellGroupData[{

Cell["\<\
Show[Graphics[OpticalFlow[2,myImageSize],
              {{AspectRatio -> 1}, {Frame -> True}}]] ;\
\>", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0297619 0.0238095 0.0297619 [
[.02381 -0.0125 -3 -9 ]
[.02381 -0.0125 3 0 ]
[.17262 -0.0125 -3 -9 ]
[.17262 -0.0125 3 0 ]
[.32143 -0.0125 -6 -9 ]
[.32143 -0.0125 6 0 ]
[.47024 -0.0125 -6 -9 ]
[.47024 -0.0125 6 0 ]
[.61905 -0.0125 -6 -9 ]
[.61905 -0.0125 6 0 ]
[.76786 -0.0125 -6 -9 ]
[.76786 -0.0125 6 0 ]
[.91667 -0.0125 -6 -9 ]
[.91667 -0.0125 6 0 ]
[ 0 0 -0.125 0 ]
[-0.0125 .02381 -6 -4.5 ]
[-0.0125 .02381 0 4.5 ]
[-0.0125 .17262 -6 -4.5 ]
[-0.0125 .17262 0 4.5 ]
[-0.0125 .32143 -12 -4.5 ]
[-0.0125 .32143 0 4.5 ]
[-0.0125 .47024 -12 -4.5 ]
[-0.0125 .47024 0 4.5 ]
[-0.0125 .61905 -12 -4.5 ]
[-0.0125 .61905 0 4.5 ]
[-0.0125 .76786 -12 -4.5 ]
[-0.0125 .76786 0 4.5 ]
[-0.0125 .91667 -12 -4.5 ]
[-0.0125 .91667 0 4.5 ]
[ 0 0 -0.125 0 ]
[ 0 1 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 0 m
.02381 .00625 L
s
[(0)] .02381 -0.0125 0 1 Mshowa
.17262 0 m
.17262 .00625 L
s
[(5)] .17262 -0.0125 0 1 Mshowa
.32143 0 m
.32143 .00625 L
s
[(10)] .32143 -0.0125 0 1 Mshowa
.47024 0 m
.47024 .00625 L
s
[(15)] .47024 -0.0125 0 1 Mshowa
.61905 0 m
.61905 .00625 L
s
[(20)] .61905 -0.0125 0 1 Mshowa
.76786 0 m
.76786 .00625 L
s
[(25)] .76786 -0.0125 0 1 Mshowa
.91667 0 m
.91667 .00625 L
s
[(30)] .91667 -0.0125 0 1 Mshowa
.125 Mabswid
.05357 0 m
.05357 .00375 L
s
.08333 0 m
.08333 .00375 L
s
.1131 0 m
.1131 .00375 L
s
.14286 0 m
.14286 .00375 L
s
.20238 0 m
.20238 .00375 L
s
.23214 0 m
.23214 .00375 L
s
.2619 0 m
.2619 .00375 L
s
.29167 0 m
.29167 .00375 L
s
.35119 0 m
.35119 .00375 L
s
.38095 0 m
.38095 .00375 L
s
.41071 0 m
.41071 .00375 L
s
.44048 0 m
.44048 .00375 L
s
.5 0 m
.5 .00375 L
s
.52976 0 m
.52976 .00375 L
s
.55952 0 m
.55952 .00375 L
s
.58929 0 m
.58929 .00375 L
s
.64881 0 m
.64881 .00375 L
s
.67857 0 m
.67857 .00375 L
s
.70833 0 m
.70833 .00375 L
s
.7381 0 m
.7381 .00375 L
s
.79762 0 m
.79762 .00375 L
s
.82738 0 m
.82738 .00375 L
s
.85714 0 m
.85714 .00375 L
s
.8869 0 m
.8869 .00375 L
s
.94643 0 m
.94643 .00375 L
s
.97619 0 m
.97619 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
0 .02381 m
.00625 .02381 L
s
[(0)] -0.0125 .02381 1 0 Mshowa
0 .17262 m
.00625 .17262 L
s
[(5)] -0.0125 .17262 1 0 Mshowa
0 .32143 m
.00625 .32143 L
s
[(10)] -0.0125 .32143 1 0 Mshowa
0 .47024 m
.00625 .47024 L
s
[(15)] -0.0125 .47024 1 0 Mshowa
0 .61905 m
.00625 .61905 L
s
[(20)] -0.0125 .61905 1 0 Mshowa
0 .76786 m
.00625 .76786 L
s
[(25)] -0.0125 .76786 1 0 Mshowa
0 .91667 m
.00625 .91667 L
s
[(30)] -0.0125 .91667 1 0 Mshowa
.125 Mabswid
0 .05357 m
.00375 .05357 L
s
0 .08333 m
.00375 .08333 L
s
0 .1131 m
.00375 .1131 L
s
0 .14286 m
.00375 .14286 L
s
0 .20238 m
.00375 .20238 L
s
0 .23214 m
.00375 .23214 L
s
0 .2619 m
.00375 .2619 L
s
0 .29167 m
.00375 .29167 L
s
0 .35119 m
.00375 .35119 L
s
0 .38095 m
.00375 .38095 L
s
0 .41071 m
.00375 .41071 L
s
0 .44048 m
.00375 .44048 L
s
0 .5 m
.00375 .5 L
s
0 .52976 m
.00375 .52976 L
s
0 .55952 m
.00375 .55952 L
s
0 .58929 m
.00375 .58929 L
s
0 .64881 m
.00375 .64881 L
s
0 .67857 m
.00375 .67857 L
s
0 .70833 m
.00375 .70833 L
s
0 .7381 m
.00375 .7381 L
s
0 .79762 m
.00375 .79762 L
s
0 .82738 m
.00375 .82738 L
s
0 .85714 m
.00375 .85714 L
s
0 .8869 m
.00375 .8869 L
s
0 .94643 m
.00375 .94643 L
s
0 .97619 m
.00375 .97619 L
s
.25 Mabswid
0 0 m
0 1 L
s
.02381 .99375 m
.02381 1 L
s
.17262 .99375 m
.17262 1 L
s
.32143 .99375 m
.32143 1 L
s
.47024 .99375 m
.47024 1 L
s
.61905 .99375 m
.61905 1 L
s
.76786 .99375 m
.76786 1 L
s
.91667 .99375 m
.91667 1 L
s
.125 Mabswid
.05357 .99625 m
.05357 1 L
s
.08333 .99625 m
.08333 1 L
s
.1131 .99625 m
.1131 1 L
s
.14286 .99625 m
.14286 1 L
s
.20238 .99625 m
.20238 1 L
s
.23214 .99625 m
.23214 1 L
s
.2619 .99625 m
.2619 1 L
s
.29167 .99625 m
.29167 1 L
s
.35119 .99625 m
.35119 1 L
s
.38095 .99625 m
.38095 1 L
s
.41071 .99625 m
.41071 1 L
s
.44048 .99625 m
.44048 1 L
s
.5 .99625 m
.5 1 L
s
.52976 .99625 m
.52976 1 L
s
.55952 .99625 m
.55952 1 L
s
.58929 .99625 m
.58929 1 L
s
.64881 .99625 m
.64881 1 L
s
.67857 .99625 m
.67857 1 L
s
.70833 .99625 m
.70833 1 L
s
.7381 .99625 m
.7381 1 L
s
.79762 .99625 m
.79762 1 L
s
.82738 .99625 m
.82738 1 L
s
.85714 .99625 m
.85714 1 L
s
.8869 .99625 m
.8869 1 L
s
.94643 .99625 m
.94643 1 L
s
.97619 .99625 m
.97619 1 L
s
.25 Mabswid
0 1 m
1 1 L
s
.99375 .02381 m
1 .02381 L
s
.99375 .17262 m
1 .17262 L
s
.99375 .32143 m
1 .32143 L
s
.99375 .47024 m
1 .47024 L
s
.99375 .61905 m
1 .61905 L
s
.99375 .76786 m
1 .76786 L
s
.99375 .91667 m
1 .91667 L
s
.125 Mabswid
.99625 .05357 m
1 .05357 L
s
.99625 .08333 m
1 .08333 L
s
.99625 .1131 m
1 .1131 L
s
.99625 .14286 m
1 .14286 L
s
.99625 .20238 m
1 .20238 L
s
.99625 .23214 m
1 .23214 L
s
.99625 .2619 m
1 .2619 L
s
.99625 .29167 m
1 .29167 L
s
.99625 .35119 m
1 .35119 L
s
.99625 .38095 m
1 .38095 L
s
.99625 .41071 m
1 .41071 L
s
.99625 .44048 m
1 .44048 L
s
.99625 .5 m
1 .5 L
s
.99625 .52976 m
1 .52976 L
s
.99625 .55952 m
1 .55952 L
s
.99625 .58929 m
1 .58929 L
s
.99625 .64881 m
1 .64881 L
s
.99625 .67857 m
1 .67857 L
s
.99625 .70833 m
1 .70833 L
s
.99625 .7381 m
1 .7381 L
s
.99625 .79762 m
1 .79762 L
s
.99625 .82738 m
1 .82738 L
s
.99625 .85714 m
1 .85714 L
s
.99625 .8869 m
1 .8869 L
s
.99625 .94643 m
1 .94643 L
s
.99625 .97619 m
1 .97619 L
s
.25 Mabswid
1 0 m
1 1 L
s
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-0.02 0.005 moveto 0 0 lineto
-0.02 -0.005 lineto fill
-0.02 0.005 moveto 0 0 lineto
-0.02 -0.005 lineto
-0.02 0.005 lineto stroke
} ifelse
grestore
} def
/MAarrowzero1 {
gsave
mathtops translate
0 0 Mdot
grestore
} def
% End of user PostScript
.5 Mabswid
.05357 .05357 m
.05357 .05357 L
s
% Start of user PostScript
1. 1. MAarrowzero1
% End of user PostScript
.05357 .08333 m
.05357 .08333 L
s
% Start of user PostScript
1. 2. MAarrowzero1
% End of user PostScript
.05357 .1131 m
.05357 .1131 L
s
% Start of user PostScript
1. 3. MAarrowzero1
% End of user PostScript
.05357 .14286 m
.05357 .14286 L
s
% Start of user PostScript
1. 4. MAarrowzero1
% End of user PostScript
.05357 .17262 m
.05357 .17262 L
s
% Start of user PostScript
1. 5. MAarrowzero1
% End of user PostScript
.05357 .20238 m
.05357 .20238 L
s
% Start of user PostScript
1. 6. MAarrowzero1
% End of user PostScript
.05357 .23214 m
.05357 .23214 L
s
% Start of user PostScript
1. 7. MAarrowzero1
% End of user PostScript
.05357 .2619 m
.05357 .2619 L
s
% Start of user PostScript
1. 8. MAarrowzero1
% End of user PostScript
.05357 .29167 m
.05357 .29167 L
s
% Start of user PostScript
1. 9. MAarrowzero1
% End of user PostScript
.05357 .32143 m
.05357 .32143 L
s
% Start of user PostScript
1. 10. MAarrowzero1
% End of user PostScript
.05357 .35119 m
.05357 .35119 L
s
% Start of user PostScript
1. 11. MAarrowzero1
% End of user PostScript
.05357 .38095 m
.05357 .38095 L
s
% Start of user PostScript
1. 12. MAarrowzero1
% End of user PostScript
.05357 .41071 m
.05357 .41071 L
s
% Start of user PostScript
1. 13. MAarrowzero1
% End of user PostScript
.05357 .44048 m
.05357 .44048 L
s
% Start of user PostScript
1. 14. MAarrowzero1
% End of user PostScript
.05357 .47024 m
.05357 .47024 L
s
% Start of user PostScript
1. 15. MAarrowzero1
% End of user PostScript
.05357 .5 m
.05357 .5 L
s
% Start of user PostScript
1. 16. MAarrowzero1
% End of user PostScript
.05357 .52976 m
.05357 .52976 L
s
% Start of user PostScript
1. 17. MAarrowzero1
% End of user PostScript
.05357 .55952 m
.05357 .55952 L
s
% Start of user PostScript
1. 18. MAarrowzero1
% End of user PostScript
.05357 .58929 m
.05357 .58929 L
s
% Start of user PostScript
1. 19. MAarrowzero1
% End of user PostScript
.05357 .61905 m
.05357 .61905 L
s
% Start of user PostScript
1. 20. MAarrowzero1
% End of user PostScript
.05357 .64881 m
.05357 .64881 L
s
% Start of user PostScript
1. 21. MAarrowzero1
% End of user PostScript
.05357 .67857 m
.05357 .67857 L
s
% Start of user PostScript
1. 22. MAarrowzero1
% End of user PostScript
.05357 .70833 m
.05357 .70833 L
s
% Start of user PostScript
1. 23. MAarrowzero1
% End of user PostScript
.05357 .7381 m
.05357 .7381 L
s
% Start of user PostScript
1. 24. MAarrowzero1
% End of user PostScript
.05357 .76786 m
.05357 .76786 L
s
% Start of user PostScript
1. 25. MAarrowzero1
% End of user PostScript
.05357 .79762 m
.05357 .79762 L
s
% Start of user PostScript
1. 26. MAarrowzero1
% End of user PostScript
.05357 .82738 m
.05357 .82738 L
s
% Start of user PostScript
1. 27. MAarrowzero1
% End of user PostScript
.05357 .85714 m
.05357 .85714 L
s
% Start of user PostScript
1. 28. MAarrowzero1
% End of user PostScript
.05357 .8869 m
.05357 .8869 L
s
% Start of user PostScript
1. 29. MAarrowzero1
% End of user PostScript
.05357 .91667 m
.05357 .91667 L
s
% Start of user PostScript
1. 30. MAarrowzero1
% End of user PostScript
.05357 .94643 m
.05357 .94643 L
s
% Start of user PostScript
1. 31. MAarrowzero1
% End of user PostScript
.05357 .97619 m
.05357 .97619 L
s
% Start of user PostScript
1. 32. MAarrowzero1
% End of user PostScript
.08333 .05357 m
.08333 .05357 L
s
% Start of user PostScript
2. 1. MAarrowzero1
% End of user PostScript
.08333 .08333 m
.08333 .08333 L
s
% Start of user PostScript
2. 2. MAarrowzero1
% End of user PostScript
.08333 .1131 m
.08333 .1131 L
s
% Start of user PostScript
2. 3. MAarrowzero1
% End of user PostScript
.08333 .14286 m
.08333 .14286 L
s
% Start of user PostScript
2. 4. MAarrowzero1
% End of user PostScript
.08333 .17262 m
.08333 .17262 L
s
% Start of user PostScript
2. 5. MAarrowzero1
% End of user PostScript
.08333 .20238 m
.08333 .20238 L
s
% Start of user PostScript
2. 6. MAarrowzero1
% End of user PostScript
.08333 .23214 m
.08333 .23214 L
s
% Start of user PostScript
2. 7. MAarrowzero1
% End of user PostScript
.08333 .2619 m
.08333 .2619 L
s
% Start of user PostScript
2. 8. MAarrowzero1
% End of user PostScript
.08333 .29167 m
.08333 .29167 L
s
% Start of user PostScript
2. 9. MAarrowzero1
% End of user PostScript
.08333 .32143 m
.08333 .32143 L
s
% Start of user PostScript
2. 10. MAarrowzero1
% End of user PostScript
.08333 .35119 m
.08333 .35119 L
s
% Start of user PostScript
2. 11. MAarrowzero1
% End of user PostScript
.08333 .38095 m
.08333 .38095 L
s
% Start of user PostScript
2. 12. MAarrowzero1
% End of user PostScript
.08333 .41071 m
.08333 .41071 L
s
% Start of user PostScript
2. 13. MAarrowzero1
% End of user PostScript
.08333 .44048 m
.08333 .44048 L
s
% Start of user PostScript
2. 14. MAarrowzero1
% End of user PostScript
.08333 .47024 m
.08333 .47024 L
s
% Start of user PostScript
2. 15. MAarrowzero1
% End of user PostScript
.08333 .5 m
.08333 .5 L
s
% Start of user PostScript
2. 16. MAarrowzero1
% End of user PostScript
.08333 .52976 m
.08333 .52976 L
s
% Start of user PostScript
2. 17. MAarrowzero1
% End of user PostScript
.08333 .55952 m
.08333 .55952 L
s
% Start of user PostScript
2. 18. MAarrowzero1
% End of user PostScript
.08333 .58929 m
.08333 .58929 L
s
% Start of user PostScript
2. 19. MAarrowzero1
% End of user PostScript
.08333 .61905 m
.08333 .61905 L
s
% Start of user PostScript
2. 20. MAarrowzero1
% End of user PostScript
.08333 .64881 m
.08333 .64881 L
s
% Start of user PostScript
2. 21. MAarrowzero1
% End of user PostScript
.08333 .67857 m
.08333 .67857 L
s
% Start of user PostScript
2. 22. MAarrowzero1
% End of user PostScript
.08333 .70833 m
.08333 .70833 L
s
% Start of user PostScript
2. 23. MAarrowzero1
% End of user PostScript
.08333 .7381 m
.08333 .7381 L
s
% Start of user PostScript
2. 24. MAarrowzero1
% End of user PostScript
.08333 .76786 m
.08333 .76786 L
s
% Start of user PostScript
2. 25. MAarrowzero1
% End of user PostScript
.08333 .79762 m
.08333 .79762 L
s
% Start of user PostScript
2. 26. MAarrowzero1
% End of user PostScript
.08333 .82738 m
.08333 .82738 L
s
% Start of user PostScript
2. 27. MAarrowzero1
% End of user PostScript
.08333 .85714 m
.08333 .85714 L
s
% Start of user PostScript
2. 28. MAarrowzero1
% End of user PostScript
.08333 .8869 m
.08333 .8869 L
s
% Start of user PostScript
2. 29. MAarrowzero1
% End of user PostScript
.08333 .91667 m
.08333 .91667 L
s
% Start of user PostScript
2. 30. MAarrowzero1
% End of user PostScript
.08333 .94643 m
.08333 .94643 L
s
% Start of user PostScript
2. 31. MAarrowzero1
% End of user PostScript
.08333 .97619 m
.08333 .97619 L
s
% Start of user PostScript
2. 32. MAarrowzero1
% End of user PostScript
.1131 .05357 m
.1131 .05357 L
s
% Start of user PostScript
3. 1. MAarrowzero1
% End of user PostScript
.1131 .08333 m
.1131 .08333 L
s
% Start of user PostScript
3. 2. MAarrowzero1
% End of user PostScript
.1131 .1131 m
.1131 .1131 L
s
% Start of user PostScript
3. 3. MAarrowzero1
% End of user PostScript
.1131 .14286 m
.1131 .14286 L
s
% Start of user PostScript
3. 4. MAarrowzero1
% End of user PostScript
.1131 .17262 m
.1131 .17262 L
s
% Start of user PostScript
3. 5. MAarrowzero1
% End of user PostScript
.1131 .20238 m
.1131 .20238 L
s
% Start of user PostScript
3. 6. MAarrowzero1
% End of user PostScript
.1131 .23214 m
.1131 .23214 L
s
% Start of user PostScript
3. 7. MAarrowzero1
% End of user PostScript
.1131 .2619 m
.1131 .2619 L
s
% Start of user PostScript
3. 8. MAarrowzero1
% End of user PostScript
.1131 .29167 m
.1131 .29167 L
s
% Start of user PostScript
3. 9. MAarrowzero1
% End of user PostScript
.1131 .32143 m
.1131 .32143 L
s
% Start of user PostScript
3. 10. MAarrowzero1
% End of user PostScript
.1131 .35119 m
.1131 .35119 L
s
% Start of user PostScript
3. 11. MAarrowzero1
% End of user PostScript
.1131 .38095 m
.1131 .38095 L
s
% Start of user PostScript
3. 12. MAarrowzero1
% End of user PostScript
.1131 .41071 m
.1131 .41071 L
s
% Start of user PostScript
3. 13. MAarrowzero1
% End of user PostScript
.1131 .44048 m
.1131 .44048 L
s
% Start of user PostScript
3. 14. MAarrowzero1
% End of user PostScript
.1131 .47024 m
.1131 .47024 L
s
% Start of user PostScript
3. 15. MAarrowzero1
% End of user PostScript
.1131 .5 m
.1131 .5 L
s
% Start of user PostScript
3. 16. MAarrowzero1
% End of user PostScript
.1131 .52976 m
.1131 .52976 L
s
% Start of user PostScript
3. 17. MAarrowzero1
% End of user PostScript
.1131 .55952 m
.1131 .55952 L
s
% Start of user PostScript
3. 18. MAarrowzero1
% End of user PostScript
.1131 .58929 m
.1131 .58929 L
s
% Start of user PostScript
3. 19. MAarrowzero1
% End of user PostScript
.1131 .61905 m
.1131 .61905 L
s
% Start of user PostScript
3. 20. MAarrowzero1
% End of user PostScript
.1131 .64881 m
.1131 .64881 L
s
% Start of user PostScript
3. 21. MAarrowzero1
% End of user PostScript
.1131 .67857 m
.1131 .67857 L
s
% Start of user PostScript
3. 22. MAarrowzero1
% End of user PostScript
.1131 .70833 m
.1131 .70833 L
s
% Start of user PostScript
3. 23. MAarrowzero1
% End of user PostScript
.1131 .7381 m
.1131 .7381 L
s
% Start of user PostScript
3. 24. MAarrowzero1
% End of user PostScript
.1131 .76786 m
.1131 .76786 L
s
% Start of user PostScript
3. 25. MAarrowzero1
% End of user PostScript
.1131 .79762 m
.1131 .79762 L
s
% Start of user PostScript
3. 26. MAarrowzero1
% End of user PostScript
.1131 .82738 m
.1131 .82738 L
s
% Start of user PostScript
3. 27. MAarrowzero1
% End of user PostScript
.1131 .85714 m
.1131 .85714 L
s
% Start of user PostScript
3. 28. MAarrowzero1
% End of user PostScript
.1131 .8869 m
.1131 .8869 L
s
% Start of user PostScript
3. 29. MAarrowzero1
% End of user PostScript
.1131 .91667 m
.1131 .91667 L
s
% Start of user PostScript
3. 30. MAarrowzero1
% End of user PostScript
.1131 .94643 m
.1131 .94643 L
s
% Start of user PostScript
3. 31. MAarrowzero1
% End of user PostScript
.1131 .97619 m
.1131 .97619 L
s
% Start of user PostScript
3. 32. MAarrowzero1
% End of user PostScript
.14286 .05357 m
.14286 .05357 L
s
% Start of user PostScript
4. 1. MAarrowzero1
% End of user PostScript
.14286 .08333 m
.14286 .08333 L
s
% Start of user PostScript
4. 2. MAarrowzero1
% End of user PostScript
.14286 .1131 m
.14286 .1131 L
s
% Start of user PostScript
4. 3. MAarrowzero1
% End of user PostScript
.14286 .14286 m
.14286 .14286 L
s
% Start of user PostScript
4. 4. MAarrowzero1
% End of user PostScript
.14286 .17262 m
.14286 .17262 L
s
% Start of user PostScript
4. 5. MAarrowzero1
% End of user PostScript
.14286 .20238 m
.14286 .20238 L
s
% Start of user PostScript
4. 6. MAarrowzero1
% End of user PostScript
.14286 .23214 m
.14286 .23214 L
s
% Start of user PostScript
4. 7. MAarrowzero1
% End of user PostScript
.14286 .2619 m
.14286 .2619 L
s
% Start of user PostScript
4. 8. MAarrowzero1
% End of user PostScript
.14286 .29167 m
.14286 .29167 L
s
% Start of user PostScript
4. 9. MAarrowzero1
% End of user PostScript
.14286 .32143 m
.14286 .32143 L
s
% Start of user PostScript
4. 10. MAarrowzero1
% End of user PostScript
.14286 .35119 m
.14286 .35119 L
s
% Start of user PostScript
4. 11. MAarrowzero1
% End of user PostScript
.14286 .38095 m
.14286 .38095 L
s
% Start of user PostScript
4. 12. MAarrowzero1
% End of user PostScript
.14286 .41071 m
.14286 .41071 L
s
% Start of user PostScript
4. 13. MAarrowzero1
% End of user PostScript
.14286 .44048 m
.14286 .44048 L
s
% Start of user PostScript
4. 14. MAarrowzero1
% End of user PostScript
.14286 .47024 m
.14286 .47024 L
s
% Start of user PostScript
4. 15. MAarrowzero1
% End of user PostScript
.14286 .5 m
.14286 .5 L
s
% Start of user PostScript
4. 16. MAarrowzero1
% End of user PostScript
.14286 .52976 m
.14286 .52976 L
s
% Start of user PostScript
4. 17. MAarrowzero1
% End of user PostScript
.14286 .55952 m
.14286 .55952 L
s
% Start of user PostScript
4. 18. MAarrowzero1
% End of user PostScript
.14286 .58929 m
.14286 .58929 L
s
% Start of user PostScript
4. 19. MAarrowzero1
% End of user PostScript
.14286 .61905 m
.14286 .61905 L
s
% Start of user PostScript
4. 20. MAarrowzero1
% End of user PostScript
.14286 .64881 m
.14286 .64881 L
s
% Start of user PostScript
4. 21. MAarrowzero1
% End of user PostScript
.14286 .67857 m
.14286 .67857 L
s
% Start of user PostScript
4. 22. MAarrowzero1
% End of user PostScript
.14286 .70833 m
.14286 .70833 L
s
% Start of user PostScript
4. 23. MAarrowzero1
% End of user PostScript
.14286 .7381 m
.14286 .7381 L
s
% Start of user PostScript
4. 24. MAarrowzero1
% End of user PostScript
.14286 .76786 m
.14286 .76786 L
s
% Start of user PostScript
4. 25. MAarrowzero1
% End of user PostScript
.14286 .79762 m
.14286 .79762 L
s
% Start of user PostScript
4. 26. MAarrowzero1
% End of user PostScript
.14286 .82738 m
.14286 .82738 L
s
% Start of user PostScript
4. 27. MAarrowzero1
% End of user PostScript
.14286 .85714 m
.14286 .85714 L
s
% Start of user PostScript
4. 28. MAarrowzero1
% End of user PostScript
.14286 .8869 m
.14286 .8869 L
s
% Start of user PostScript
4. 29. MAarrowzero1
% End of user PostScript
.14286 .91667 m
.14286 .91667 L
s
% Start of user PostScript
4. 30. MAarrowzero1
% End of user PostScript
.14286 .94643 m
.14286 .94643 L
s
% Start of user PostScript
4. 31. MAarrowzero1
% End of user PostScript
.14286 .97619 m
.14286 .97619 L
s
% Start of user PostScript
4. 32. MAarrowzero1
% End of user PostScript
.17262 .05357 m
.17262 .05357 L
s
% Start of user PostScript
5. 1. MAarrowzero1
% End of user PostScript
.17262 .08333 m
.17262 .08333 L
s
% Start of user PostScript
5. 2. MAarrowzero1
% End of user PostScript
.17262 .1131 m
.17262 .1131 L
s
% Start of user PostScript
5. 3. MAarrowzero1
% End of user PostScript
.17262 .14286 m
.17262 .14286 L
s
% Start of user PostScript
5. 4. MAarrowzero1
% End of user PostScript
.17262 .17262 m
.17262 .17262 L
s
% Start of user PostScript
5. 5. MAarrowzero1
% End of user PostScript
.17262 .20238 m
.17262 .20238 L
s
% Start of user PostScript
5. 6. MAarrowzero1
% End of user PostScript
.17262 .23214 m
.17262 .23214 L
s
% Start of user PostScript
5. 7. MAarrowzero1
% End of user PostScript
.17262 .2619 m
.17262 .2619 L
s
% Start of user PostScript
5. 8. MAarrowzero1
% End of user PostScript
.17262 .29167 m
.17262 .29167 L
s
% Start of user PostScript
5. 9. MAarrowzero1
% End of user PostScript
.17262 .32143 m
.17262 .32143 L
s
% Start of user PostScript
5. 10. MAarrowzero1
% End of user PostScript
.17262 .35119 m
.17262 .35119 L
s
% Start of user PostScript
5. 11. MAarrowzero1
% End of user PostScript
.17262 .38095 m
.17262 .38095 L
s
% Start of user PostScript
5. 12. MAarrowzero1
% End of user PostScript
.17262 .41071 m
.17262 .41071 L
s
% Start of user PostScript
5. 13. MAarrowzero1
% End of user PostScript
.17262 .44048 m
.17262 .44048 L
s
% Start of user PostScript
5. 14. MAarrowzero1
% End of user PostScript
.17262 .47024 m
.17262 .47024 L
s
% Start of user PostScript
5. 15. MAarrowzero1
% End of user PostScript
.17262 .5 m
.17262 .5 L
s
% Start of user PostScript
5. 16. MAarrowzero1
% End of user PostScript
.17262 .52976 m
.17262 .52976 L
s
% Start of user PostScript
5. 17. MAarrowzero1
% End of user PostScript
.17262 .55952 m
.17262 .52976 L
s
% Start of user PostScript
5. 18. 5. 17. MAarrowhead1
% End of user PostScript
.17262 .58929 m
.17262 .52976 L
s
% Start of user PostScript
5. 19. 5. 17. MAarrowhead1
% End of user PostScript
.17262 .61905 m
.17262 .64881 L
s
% Start of user PostScript
5. 20. 5. 21. MAarrowhead1
% End of user PostScript
.17262 .64881 m
.17262 .64881 L
s
% Start of user PostScript
5. 21. MAarrowzero1
% End of user PostScript
.17262 .67857 m
.17262 .67857 L
s
% Start of user PostScript
5. 22. MAarrowzero1
% End of user PostScript
.17262 .70833 m
.17262 .70833 L
s
% Start of user PostScript
5. 23. MAarrowzero1
% End of user PostScript
.17262 .7381 m
.17262 .7381 L
s
% Start of user PostScript
5. 24. MAarrowzero1
% End of user PostScript
.17262 .76786 m
.17262 .76786 L
s
% Start of user PostScript
5. 25. MAarrowzero1
% End of user PostScript
.17262 .79762 m
.17262 .79762 L
s
% Start of user PostScript
5. 26. MAarrowzero1
% End of user PostScript
.17262 .82738 m
.17262 .82738 L
s
% Start of user PostScript
5. 27. MAarrowzero1
% End of user PostScript
.17262 .85714 m
.17262 .85714 L
s
% Start of user PostScript
5. 28. MAarrowzero1
% End of user PostScript
.17262 .8869 m
.17262 .8869 L
s
% Start of user PostScript
5. 29. MAarrowzero1
% End of user PostScript
.17262 .91667 m
.17262 .91667 L
s
% Start of user PostScript
5. 30. MAarrowzero1
% End of user PostScript
.17262 .94643 m
.17262 .94643 L
s
% Start of user PostScript
5. 31. MAarrowzero1
% End of user PostScript
.17262 .97619 m
.17262 .97619 L
s
% Start of user PostScript
5. 32. MAarrowzero1
% End of user PostScript
.20238 .05357 m
.20238 .05357 L
s
% Start of user PostScript
6. 1. MAarrowzero1
% End of user PostScript
.20238 .08333 m
.20238 .08333 L
s
% Start of user PostScript
6. 2. MAarrowzero1
% End of user PostScript
.20238 .1131 m
.20238 .1131 L
s
% Start of user PostScript
6. 3. MAarrowzero1
% End of user PostScript
.20238 .14286 m
.20238 .14286 L
s
% Start of user PostScript
6. 4. MAarrowzero1
% End of user PostScript
.20238 .17262 m
.20238 .17262 L
s
% Start of user PostScript
6. 5. MAarrowzero1
% End of user PostScript
.20238 .20238 m
.20238 .20238 L
s
% Start of user PostScript
6. 6. MAarrowzero1
% End of user PostScript
.20238 .23214 m
.20238 .23214 L
s
% Start of user PostScript
6. 7. MAarrowzero1
% End of user PostScript
.20238 .2619 m
.20238 .2619 L
s
% Start of user PostScript
6. 8. MAarrowzero1
% End of user PostScript
.20238 .29167 m
.20238 .29167 L
s
% Start of user PostScript
6. 9. MAarrowzero1
% End of user PostScript
.20238 .32143 m
.20238 .32143 L
s
% Start of user PostScript
6. 10. MAarrowzero1
% End of user PostScript
.20238 .35119 m
.20238 .35119 L
s
% Start of user PostScript
6. 11. MAarrowzero1
% End of user PostScript
.20238 .38095 m
.20238 .38095 L
s
% Start of user PostScript
6. 12. MAarrowzero1
% End of user PostScript
.20238 .41071 m
.20238 .41071 L
s
% Start of user PostScript
6. 13. MAarrowzero1
% End of user PostScript
.20238 .44048 m
.20238 .44048 L
s
% Start of user PostScript
6. 14. MAarrowzero1
% End of user PostScript
.20238 .47024 m
.20238 .47024 L
s
% Start of user PostScript
6. 15. MAarrowzero1
% End of user PostScript
.20238 .5 m
.20238 .5 L
s
% Start of user PostScript
6. 16. MAarrowzero1
% End of user PostScript
.20238 .52976 m
.20238 .5 L
s
% Start of user PostScript
6. 17. 6. 16. MAarrowhead1
% End of user PostScript
.20238 .55952 m
.23214 .5 L
s
% Start of user PostScript
6. 18. 7. 16. MAarrowhead1
% End of user PostScript
.20238 .58929 m
.20238 .52976 L
s
% Start of user PostScript
6. 19. 6. 17. MAarrowhead1
% End of user PostScript
.20238 .61905 m
.2619 .67857 L
s
% Start of user PostScript
6. 20. 8. 22. MAarrowhead1
% End of user PostScript
.20238 .64881 m
.23214 .70833 L
s
% Start of user PostScript
6. 21. 7. 23. MAarrowhead1
% End of user PostScript
.20238 .67857 m
.20238 .67857 L
s
% Start of user PostScript
6. 22. MAarrowzero1
% End of user PostScript
.20238 .70833 m
.20238 .70833 L
s
% Start of user PostScript
6. 23. MAarrowzero1
% End of user PostScript
.20238 .7381 m
.20238 .7381 L
s
% Start of user PostScript
6. 24. MAarrowzero1
% End of user PostScript
.20238 .76786 m
.20238 .76786 L
s
% Start of user PostScript
6. 25. MAarrowzero1
% End of user PostScript
.20238 .79762 m
.20238 .79762 L
s
% Start of user PostScript
6. 26. MAarrowzero1
% End of user PostScript
.20238 .82738 m
.20238 .82738 L
s
% Start of user PostScript
6. 27. MAarrowzero1
% End of user PostScript
.20238 .85714 m
.20238 .85714 L
s
% Start of user PostScript
6. 28. MAarrowzero1
% End of user PostScript
.20238 .8869 m
.20238 .8869 L
s
% Start of user PostScript
6. 29. MAarrowzero1
% End of user PostScript
.20238 .91667 m
.20238 .91667 L
s
% Start of user PostScript
6. 30. MAarrowzero1
% End of user PostScript
.20238 .94643 m
.20238 .94643 L
s
% Start of user PostScript
6. 31. MAarrowzero1
% End of user PostScript
.20238 .97619 m
.20238 .97619 L
s
% Start of user PostScript
6. 32. MAarrowzero1
% End of user PostScript
.23214 .05357 m
.23214 .05357 L
s
% Start of user PostScript
7. 1. MAarrowzero1
% End of user PostScript
.23214 .08333 m
.23214 .08333 L
s
% Start of user PostScript
7. 2. MAarrowzero1
% End of user PostScript
.23214 .1131 m
.23214 .1131 L
s
% Start of user PostScript
7. 3. MAarrowzero1
% End of user PostScript
.23214 .14286 m
.23214 .14286 L
s
% Start of user PostScript
7. 4. MAarrowzero1
% End of user PostScript
.23214 .17262 m
.23214 .17262 L
s
% Start of user PostScript
7. 5. MAarrowzero1
% End of user PostScript
.23214 .20238 m
.23214 .20238 L
s
% Start of user PostScript
7. 6. MAarrowzero1
% End of user PostScript
.23214 .23214 m
.23214 .23214 L
s
% Start of user PostScript
7. 7. MAarrowzero1
% End of user PostScript
.23214 .2619 m
.23214 .2619 L
s
% Start of user PostScript
7. 8. MAarrowzero1
% End of user PostScript
.23214 .29167 m
.23214 .29167 L
s
% Start of user PostScript
7. 9. MAarrowzero1
% End of user PostScript
.23214 .32143 m
.23214 .32143 L
s
% Start of user PostScript
7. 10. MAarrowzero1
% End of user PostScript
.23214 .35119 m
.23214 .35119 L
s
% Start of user PostScript
7. 11. MAarrowzero1
% End of user PostScript
.23214 .38095 m
.23214 .38095 L
s
% Start of user PostScript
7. 12. MAarrowzero1
% End of user PostScript
.23214 .41071 m
.23214 .41071 L
s
% Start of user PostScript
7. 13. MAarrowzero1
% End of user PostScript
.23214 .44048 m
.23214 .44048 L
s
% Start of user PostScript
7. 14. MAarrowzero1
% End of user PostScript
.23214 .47024 m
.23214 .47024 L
s
% Start of user PostScript
7. 15. MAarrowzero1
% End of user PostScript
.23214 .5 m
.23214 .5 L
s
% Start of user PostScript
7. 16. MAarrowzero1
% End of user PostScript
.23214 .52976 m
.23214 .5 L
s
% Start of user PostScript
7. 17. 7. 16. MAarrowhead1
% End of user PostScript
.23214 .55952 m
.20238 .58929 L
s
% Start of user PostScript
7. 18. 6. 19. MAarrowhead1
% End of user PostScript
.23214 .58929 m
.20238 .61905 L
s
% Start of user PostScript
7. 19. 6. 20. MAarrowhead1
% End of user PostScript
.23214 .61905 m
.20238 .64881 L
s
% Start of user PostScript
7. 20. 6. 21. MAarrowhead1
% End of user PostScript
.23214 .64881 m
.2619 .70833 L
s
% Start of user PostScript
7. 21. 8. 23. MAarrowhead1
% End of user PostScript
.23214 .67857 m
.23214 .67857 L
s
% Start of user PostScript
7. 22. MAarrowzero1
% End of user PostScript
.23214 .70833 m
.23214 .70833 L
s
% Start of user PostScript
7. 23. MAarrowzero1
% End of user PostScript
.23214 .7381 m
.23214 .7381 L
s
% Start of user PostScript
7. 24. MAarrowzero1
% End of user PostScript
.23214 .76786 m
.23214 .76786 L
s
% Start of user PostScript
7. 25. MAarrowzero1
% End of user PostScript
.23214 .79762 m
.23214 .79762 L
s
% Start of user PostScript
7. 26. MAarrowzero1
% End of user PostScript
.23214 .82738 m
.23214 .82738 L
s
% Start of user PostScript
7. 27. MAarrowzero1
% End of user PostScript
.23214 .85714 m
.23214 .85714 L
s
% Start of user PostScript
7. 28. MAarrowzero1
% End of user PostScript
.23214 .8869 m
.23214 .8869 L
s
% Start of user PostScript
7. 29. MAarrowzero1
% End of user PostScript
.23214 .91667 m
.23214 .91667 L
s
% Start of user PostScript
7. 30. MAarrowzero1
% End of user PostScript
.23214 .94643 m
.23214 .94643 L
s
% Start of user PostScript
7. 31. MAarrowzero1
% End of user PostScript
.23214 .97619 m
.23214 .97619 L
s
% Start of user PostScript
7. 32. MAarrowzero1
% End of user PostScript
.2619 .05357 m
.2619 .05357 L
s
% Start of user PostScript
8. 1. MAarrowzero1
% End of user PostScript
.2619 .08333 m
.2619 .08333 L
s
% Start of user PostScript
8. 2. MAarrowzero1
% End of user PostScript
.2619 .1131 m
.2619 .1131 L
s
% Start of user PostScript
8. 3. MAarrowzero1
% End of user PostScript
.2619 .14286 m
.2619 .14286 L
s
% Start of user PostScript
8. 4. MAarrowzero1
% End of user PostScript
.2619 .17262 m
.2619 .17262 L
s
% Start of user PostScript
8. 5. MAarrowzero1
% End of user PostScript
.2619 .20238 m
.2619 .20238 L
s
% Start of user PostScript
8. 6. MAarrowzero1
% End of user PostScript
.2619 .23214 m
.2619 .23214 L
s
% Start of user PostScript
8. 7. MAarrowzero1
% End of user PostScript
.2619 .2619 m
.2619 .2619 L
s
% Start of user PostScript
8. 8. MAarrowzero1
% End of user PostScript
.2619 .29167 m
.2619 .29167 L
s
% Start of user PostScript
8. 9. MAarrowzero1
% End of user PostScript
.2619 .32143 m
.2619 .32143 L
s
% Start of user PostScript
8. 10. MAarrowzero1
% End of user PostScript
.2619 .35119 m
.2619 .35119 L
s
% Start of user PostScript
8. 11. MAarrowzero1
% End of user PostScript
.2619 .38095 m
.23214 .44048 L
s
% Start of user PostScript
8. 12. 7. 14. MAarrowhead1
% End of user PostScript
.2619 .41071 m
.2619 .44048 L
s
% Start of user PostScript
8. 13. 8. 14. MAarrowhead1
% End of user PostScript
.2619 .44048 m
.2619 .44048 L
s
% Start of user PostScript
8. 14. MAarrowzero1
% End of user PostScript
.2619 .47024 m
.2619 .47024 L
s
% Start of user PostScript
8. 15. MAarrowzero1
% End of user PostScript
.2619 .5 m
.2619 .47024 L
s
% Start of user PostScript
8. 16. 8. 15. MAarrowhead1
% End of user PostScript
.2619 .52976 m
.20238 .55952 L
s
% Start of user PostScript
8. 17. 6. 18. MAarrowhead1
% End of user PostScript
.2619 .55952 m
.23214 .58929 L
s
% Start of user PostScript
8. 18. 7. 19. MAarrowhead1
% End of user PostScript
.2619 .58929 m
.23214 .61905 L
s
% Start of user PostScript
8. 19. 7. 20. MAarrowhead1
% End of user PostScript
.2619 .61905 m
.23214 .64881 L
s
% Start of user PostScript
8. 20. 7. 21. MAarrowhead1
% End of user PostScript
.2619 .64881 m
.23214 .67857 L
s
% Start of user PostScript
8. 21. 7. 22. MAarrowhead1
% End of user PostScript
.2619 .67857 m
.23214 .67857 L
s
% Start of user PostScript
8. 22. 7. 22. MAarrowhead1
% End of user PostScript
.2619 .70833 m
.2619 .70833 L
s
% Start of user PostScript
8. 23. MAarrowzero1
% End of user PostScript
.2619 .7381 m
.2619 .7381 L
s
% Start of user PostScript
8. 24. MAarrowzero1
% End of user PostScript
.2619 .76786 m
.2619 .76786 L
s
% Start of user PostScript
8. 25. MAarrowzero1
% End of user PostScript
.2619 .79762 m
.2619 .79762 L
s
% Start of user PostScript
8. 26. MAarrowzero1
% End of user PostScript
.2619 .82738 m
.2619 .82738 L
s
% Start of user PostScript
8. 27. MAarrowzero1
% End of user PostScript
.2619 .85714 m
.2619 .85714 L
s
% Start of user PostScript
8. 28. MAarrowzero1
% End of user PostScript
.2619 .8869 m
.2619 .8869 L
s
% Start of user PostScript
8. 29. MAarrowzero1
% End of user PostScript
.2619 .91667 m
.2619 .91667 L
s
% Start of user PostScript
8. 30. MAarrowzero1
% End of user PostScript
.2619 .94643 m
.2619 .94643 L
s
% Start of user PostScript
8. 31. MAarrowzero1
% End of user PostScript
.2619 .97619 m
.2619 .97619 L
s
% Start of user PostScript
8. 32. MAarrowzero1
% End of user PostScript
.29167 .05357 m
.29167 .05357 L
s
% Start of user PostScript
9. 1. MAarrowzero1
% End of user PostScript
.29167 .08333 m
.29167 .08333 L
s
% Start of user PostScript
9. 2. MAarrowzero1
% End of user PostScript
.29167 .1131 m
.29167 .1131 L
s
% Start of user PostScript
9. 3. MAarrowzero1
% End of user PostScript
.29167 .14286 m
.29167 .14286 L
s
% Start of user PostScript
9. 4. MAarrowzero1
% End of user PostScript
.29167 .17262 m
.29167 .17262 L
s
% Start of user PostScript
9. 5. MAarrowzero1
% End of user PostScript
.29167 .20238 m
.29167 .20238 L
s
% Start of user PostScript
9. 6. MAarrowzero1
% End of user PostScript
.29167 .23214 m
.29167 .23214 L
s
% Start of user PostScript
9. 7. MAarrowzero1
% End of user PostScript
.29167 .2619 m
.29167 .2619 L
s
% Start of user PostScript
9. 8. MAarrowzero1
% End of user PostScript
.29167 .29167 m
.23214 .35119 L
s
% Start of user PostScript
9. 9. 7. 11. MAarrowhead1
% End of user PostScript
.29167 .32143 m
.23214 .38095 L
s
% Start of user PostScript
9. 10. 7. 12. MAarrowhead1
% End of user PostScript
.29167 .35119 m
.23214 .41071 L
s
% Start of user PostScript
9. 11. 7. 13. MAarrowhead1
% End of user PostScript
.29167 .38095 m
.2619 .44048 L
s
% Start of user PostScript
9. 12. 8. 14. MAarrowhead1
% End of user PostScript
.29167 .41071 m
.2619 .44048 L
s
% Start of user PostScript
9. 13. 8. 14. MAarrowhead1
% End of user PostScript
.29167 .44048 m
.2619 .47024 L
s
% Start of user PostScript
9. 14. 8. 15. MAarrowhead1
% End of user PostScript
.29167 .47024 m
.2619 .5 L
s
% Start of user PostScript
9. 15. 8. 16. MAarrowhead1
% End of user PostScript
.29167 .5 m
.2619 .47024 L
s
% Start of user PostScript
9. 16. 8. 15. MAarrowhead1
% End of user PostScript
.29167 .52976 m
.29167 .5 L
s
% Start of user PostScript
9. 17. 9. 16. MAarrowhead1
% End of user PostScript
.29167 .55952 m
.2619 .58929 L
s
% Start of user PostScript
9. 18. 8. 19. MAarrowhead1
% End of user PostScript
.29167 .58929 m
.2619 .61905 L
s
% Start of user PostScript
9. 19. 8. 20. MAarrowhead1
% End of user PostScript
.29167 .61905 m
.2619 .64881 L
s
% Start of user PostScript
9. 20. 8. 21. MAarrowhead1
% End of user PostScript
.29167 .64881 m
.2619 .67857 L
s
% Start of user PostScript
9. 21. 8. 22. MAarrowhead1
% End of user PostScript
.29167 .67857 m
.23214 .67857 L
s
% Start of user PostScript
9. 22. 7. 22. MAarrowhead1
% End of user PostScript
.29167 .70833 m
.23214 .67857 L
s
% Start of user PostScript
9. 23. 7. 22. MAarrowhead1
% End of user PostScript
.29167 .7381 m
.29167 .7381 L
s
% Start of user PostScript
9. 24. MAarrowzero1
% End of user PostScript
.29167 .76786 m
.29167 .76786 L
s
% Start of user PostScript
9. 25. MAarrowzero1
% End of user PostScript
.29167 .79762 m
.29167 .79762 L
s
% Start of user PostScript
9. 26. MAarrowzero1
% End of user PostScript
.29167 .82738 m
.29167 .82738 L
s
% Start of user PostScript
9. 27. MAarrowzero1
% End of user PostScript
.29167 .85714 m
.29167 .85714 L
s
% Start of user PostScript
9. 28. MAarrowzero1
% End of user PostScript
.29167 .8869 m
.29167 .8869 L
s
% Start of user PostScript
9. 29. MAarrowzero1
% End of user PostScript
.29167 .91667 m
.29167 .91667 L
s
% Start of user PostScript
9. 30. MAarrowzero1
% End of user PostScript
.29167 .94643 m
.29167 .94643 L
s
% Start of user PostScript
9. 31. MAarrowzero1
% End of user PostScript
.29167 .97619 m
.29167 .97619 L
s
% Start of user PostScript
9. 32. MAarrowzero1
% End of user PostScript
.32143 .05357 m
.32143 .05357 L
s
% Start of user PostScript
10. 1. MAarrowzero1
% End of user PostScript
.32143 .08333 m
.32143 .08333 L
s
% Start of user PostScript
10. 2. MAarrowzero1
% End of user PostScript
.32143 .1131 m
.32143 .1131 L
s
% Start of user PostScript
10. 3. MAarrowzero1
% End of user PostScript
.32143 .14286 m
.32143 .14286 L
s
% Start of user PostScript
10. 4. MAarrowzero1
% End of user PostScript
.32143 .17262 m
.32143 .17262 L
s
% Start of user PostScript
10. 5. MAarrowzero1
% End of user PostScript
.32143 .20238 m
.32143 .20238 L
s
% Start of user PostScript
10. 6. MAarrowzero1
% End of user PostScript
.32143 .23214 m
.32143 .23214 L
s
% Start of user PostScript
10. 7. MAarrowzero1
% End of user PostScript
.32143 .2619 m
.32143 .2619 L
s
% Start of user PostScript
10. 8. MAarrowzero1
% End of user PostScript
.32143 .29167 m
.35119 .2619 L
s
% Start of user PostScript
10. 9. 11. 8. MAarrowhead1
% End of user PostScript
.32143 .32143 m
.2619 .38095 L
s
% Start of user PostScript
10. 10. 8. 12. MAarrowhead1
% End of user PostScript
.32143 .35119 m
.2619 .41071 L
s
% Start of user PostScript
10. 11. 8. 13. MAarrowhead1
% End of user PostScript
.32143 .38095 m
.2619 .44048 L
s
% Start of user PostScript
10. 12. 8. 14. MAarrowhead1
% End of user PostScript
.32143 .41071 m
.2619 .47024 L
s
% Start of user PostScript
10. 13. 8. 15. MAarrowhead1
% End of user PostScript
.32143 .44048 m
.29167 .5 L
s
% Start of user PostScript
10. 14. 9. 16. MAarrowhead1
% End of user PostScript
.32143 .47024 m
.29167 .5 L
s
% Start of user PostScript
10. 15. 9. 16. MAarrowhead1
% End of user PostScript
.32143 .5 m
.2619 .47024 L
s
% Start of user PostScript
10. 16. 8. 15. MAarrowhead1
% End of user PostScript
.32143 .52976 m
.29167 .47024 L
s
% Start of user PostScript
10. 17. 9. 15. MAarrowhead1
% End of user PostScript
.32143 .55952 m
.29167 .58929 L
s
% Start of user PostScript
10. 18. 9. 19. MAarrowhead1
% End of user PostScript
.32143 .58929 m
.29167 .61905 L
s
% Start of user PostScript
10. 19. 9. 20. MAarrowhead1
% End of user PostScript
.32143 .61905 m
.29167 .64881 L
s
% Start of user PostScript
10. 20. 9. 21. MAarrowhead1
% End of user PostScript
.32143 .64881 m
.29167 .67857 L
s
% Start of user PostScript
10. 21. 9. 22. MAarrowhead1
% End of user PostScript
.32143 .67857 m
.2619 .67857 L
s
% Start of user PostScript
10. 22. 8. 22. MAarrowhead1
% End of user PostScript
.32143 .70833 m
.2619 .70833 L
s
% Start of user PostScript
10. 23. 8. 23. MAarrowhead1
% End of user PostScript
.32143 .7381 m
.2619 .70833 L
s
% Start of user PostScript
10. 24. 8. 23. MAarrowhead1
% End of user PostScript
.32143 .76786 m
.32143 .76786 L
s
% Start of user PostScript
10. 25. MAarrowzero1
% End of user PostScript
.32143 .79762 m
.32143 .79762 L
s
% Start of user PostScript
10. 26. MAarrowzero1
% End of user PostScript
.32143 .82738 m
.32143 .82738 L
s
% Start of user PostScript
10. 27. MAarrowzero1
% End of user PostScript
.32143 .85714 m
.32143 .85714 L
s
% Start of user PostScript
10. 28. MAarrowzero1
% End of user PostScript
.32143 .8869 m
.32143 .8869 L
s
% Start of user PostScript
10. 29. MAarrowzero1
% End of user PostScript
.32143 .91667 m
.32143 .91667 L
s
% Start of user PostScript
10. 30. MAarrowzero1
% End of user PostScript
.32143 .94643 m
.32143 .94643 L
s
% Start of user PostScript
10. 31. MAarrowzero1
% End of user PostScript
.32143 .97619 m
.32143 .97619 L
s
% Start of user PostScript
10. 32. MAarrowzero1
% End of user PostScript
.35119 .05357 m
.35119 .05357 L
s
% Start of user PostScript
11. 1. MAarrowzero1
% End of user PostScript
.35119 .08333 m
.35119 .08333 L
s
% Start of user PostScript
11. 2. MAarrowzero1
% End of user PostScript
.35119 .1131 m
.35119 .1131 L
s
% Start of user PostScript
11. 3. MAarrowzero1
% End of user PostScript
.35119 .14286 m
.35119 .14286 L
s
% Start of user PostScript
11. 4. MAarrowzero1
% End of user PostScript
.35119 .17262 m
.35119 .17262 L
s
% Start of user PostScript
11. 5. MAarrowzero1
% End of user PostScript
.35119 .20238 m
.35119 .20238 L
s
% Start of user PostScript
11. 6. MAarrowzero1
% End of user PostScript
.35119 .23214 m
.35119 .23214 L
s
% Start of user PostScript
11. 7. MAarrowzero1
% End of user PostScript
.35119 .2619 m
.35119 .2619 L
s
% Start of user PostScript
11. 8. MAarrowzero1
% End of user PostScript
.35119 .29167 m
.29167 .32143 L
s
% Start of user PostScript
11. 9. 9. 10. MAarrowhead1
% End of user PostScript
.35119 .32143 m
.29167 .35119 L
s
% Start of user PostScript
11. 10. 9. 11. MAarrowhead1
% End of user PostScript
.35119 .35119 m
.29167 .38095 L
s
% Start of user PostScript
11. 11. 9. 12. MAarrowhead1
% End of user PostScript
.35119 .38095 m
.29167 .38095 L
s
% Start of user PostScript
11. 12. 9. 12. MAarrowhead1
% End of user PostScript
.35119 .41071 m
.29167 .41071 L
s
% Start of user PostScript
11. 13. 9. 13. MAarrowhead1
% End of user PostScript
.35119 .44048 m
.29167 .44048 L
s
% Start of user PostScript
11. 14. 9. 14. MAarrowhead1
% End of user PostScript
.35119 .47024 m
.29167 .47024 L
s
% Start of user PostScript
11. 15. 9. 15. MAarrowhead1
% End of user PostScript
.35119 .5 m
.29167 .5 L
s
% Start of user PostScript
11. 16. 9. 16. MAarrowhead1
% End of user PostScript
.35119 .52976 m
.32143 .55952 L
s
% Start of user PostScript
11. 17. 10. 18. MAarrowhead1
% End of user PostScript
.35119 .55952 m
.32143 .58929 L
s
% Start of user PostScript
11. 18. 10. 19. MAarrowhead1
% End of user PostScript
.35119 .58929 m
.32143 .61905 L
s
% Start of user PostScript
11. 19. 10. 20. MAarrowhead1
% End of user PostScript
.35119 .61905 m
.32143 .64881 L
s
% Start of user PostScript
11. 20. 10. 21. MAarrowhead1
% End of user PostScript
.35119 .64881 m
.32143 .67857 L
s
% Start of user PostScript
11. 21. 10. 22. MAarrowhead1
% End of user PostScript
.35119 .67857 m
.29167 .67857 L
s
% Start of user PostScript
11. 22. 9. 22. MAarrowhead1
% End of user PostScript
.35119 .70833 m
.29167 .70833 L
s
% Start of user PostScript
11. 23. 9. 23. MAarrowhead1
% End of user PostScript
.35119 .7381 m
.29167 .7381 L
s
% Start of user PostScript
11. 24. 9. 24. MAarrowhead1
% End of user PostScript
.35119 .76786 m
.32143 .76786 L
s
% Start of user PostScript
11. 25. 10. 25. MAarrowhead1
% End of user PostScript
.35119 .79762 m
.35119 .79762 L
s
% Start of user PostScript
11. 26. MAarrowzero1
% End of user PostScript
.35119 .82738 m
.35119 .82738 L
s
% Start of user PostScript
11. 27. MAarrowzero1
% End of user PostScript
.35119 .85714 m
.35119 .85714 L
s
% Start of user PostScript
11. 28. MAarrowzero1
% End of user PostScript
.35119 .8869 m
.35119 .8869 L
s
% Start of user PostScript
11. 29. MAarrowzero1
% End of user PostScript
.35119 .91667 m
.35119 .91667 L
s
% Start of user PostScript
11. 30. MAarrowzero1
% End of user PostScript
.35119 .94643 m
.35119 .94643 L
s
% Start of user PostScript
11. 31. MAarrowzero1
% End of user PostScript
.35119 .97619 m
.35119 .97619 L
s
% Start of user PostScript
11. 32. MAarrowzero1
% End of user PostScript
.38095 .05357 m
.38095 .05357 L
s
% Start of user PostScript
12. 1. MAarrowzero1
% End of user PostScript
.38095 .08333 m
.38095 .08333 L
s
% Start of user PostScript
12. 2. MAarrowzero1
% End of user PostScript
.38095 .1131 m
.38095 .1131 L
s
% Start of user PostScript
12. 3. MAarrowzero1
% End of user PostScript
.38095 .14286 m
.38095 .14286 L
s
% Start of user PostScript
12. 4. MAarrowzero1
% End of user PostScript
.38095 .17262 m
.38095 .17262 L
s
% Start of user PostScript
12. 5. MAarrowzero1
% End of user PostScript
.38095 .20238 m
.38095 .20238 L
s
% Start of user PostScript
12. 6. MAarrowzero1
% End of user PostScript
.38095 .23214 m
.38095 .23214 L
s
% Start of user PostScript
12. 7. MAarrowzero1
% End of user PostScript
.38095 .2619 m
.41071 .29167 L
s
% Start of user PostScript
12. 8. 13. 9. MAarrowhead1
% End of user PostScript
.38095 .29167 m
.32143 .32143 L
s
% Start of user PostScript
12. 9. 10. 10. MAarrowhead1
% End of user PostScript
.38095 .32143 m
.32143 .35119 L
s
% Start of user PostScript
12. 10. 10. 11. MAarrowhead1
% End of user PostScript
.38095 .35119 m
.32143 .38095 L
s
% Start of user PostScript
12. 11. 10. 12. MAarrowhead1
% End of user PostScript
.38095 .38095 m
.32143 .41071 L
s
% Start of user PostScript
12. 12. 10. 13. MAarrowhead1
% End of user PostScript
.38095 .41071 m
.32143 .41071 L
s
% Start of user PostScript
12. 13. 10. 13. MAarrowhead1
% End of user PostScript
.38095 .44048 m
.32143 .44048 L
s
% Start of user PostScript
12. 14. 10. 14. MAarrowhead1
% End of user PostScript
.38095 .47024 m
.32143 .47024 L
s
% Start of user PostScript
12. 15. 10. 15. MAarrowhead1
% End of user PostScript
.38095 .5 m
.32143 .5 L
s
% Start of user PostScript
12. 16. 10. 16. MAarrowhead1
% End of user PostScript
.38095 .52976 m
.32143 .52976 L
s
% Start of user PostScript
12. 17. 10. 17. MAarrowhead1
% End of user PostScript
.38095 .55952 m
.35119 .58929 L
s
% Start of user PostScript
12. 18. 11. 19. MAarrowhead1
% End of user PostScript
.38095 .58929 m
.35119 .61905 L
s
% Start of user PostScript
12. 19. 11. 20. MAarrowhead1
% End of user PostScript
.38095 .61905 m
.35119 .64881 L
s
% Start of user PostScript
12. 20. 11. 21. MAarrowhead1
% End of user PostScript
.38095 .64881 m
.35119 .67857 L
s
% Start of user PostScript
12. 21. 11. 22. MAarrowhead1
% End of user PostScript
.38095 .67857 m
.32143 .67857 L
s
% Start of user PostScript
12. 22. 10. 22. MAarrowhead1
% End of user PostScript
.38095 .70833 m
.32143 .70833 L
s
% Start of user PostScript
12. 23. 10. 23. MAarrowhead1
% End of user PostScript
.38095 .7381 m
.32143 .7381 L
s
% Start of user PostScript
12. 24. 10. 24. MAarrowhead1
% End of user PostScript
.38095 .76786 m
.32143 .76786 L
s
% Start of user PostScript
12. 25. 10. 25. MAarrowhead1
% End of user PostScript
.38095 .79762 m
.35119 .79762 L
s
% Start of user PostScript
12. 26. 11. 26. MAarrowhead1
% End of user PostScript
.38095 .82738 m
.38095 .82738 L
s
% Start of user PostScript
12. 27. MAarrowzero1
% End of user PostScript
.38095 .85714 m
.38095 .85714 L
s
% Start of user PostScript
12. 28. MAarrowzero1
% End of user PostScript
.38095 .8869 m
.38095 .8869 L
s
% Start of user PostScript
12. 29. MAarrowzero1
% End of user PostScript
.38095 .91667 m
.38095 .91667 L
s
% Start of user PostScript
12. 30. MAarrowzero1
% End of user PostScript
.38095 .94643 m
.38095 .94643 L
s
% Start of user PostScript
12. 31. MAarrowzero1
% End of user PostScript
.38095 .97619 m
.38095 .97619 L
s
% Start of user PostScript
12. 32. MAarrowzero1
% End of user PostScript
.41071 .05357 m
.41071 .05357 L
s
% Start of user PostScript
13. 1. MAarrowzero1
% End of user PostScript
.41071 .08333 m
.41071 .08333 L
s
% Start of user PostScript
13. 2. MAarrowzero1
% End of user PostScript
.41071 .1131 m
.41071 .1131 L
s
% Start of user PostScript
13. 3. MAarrowzero1
% End of user PostScript
.41071 .14286 m
.41071 .14286 L
s
% Start of user PostScript
13. 4. MAarrowzero1
% End of user PostScript
.41071 .17262 m
.41071 .17262 L
s
% Start of user PostScript
13. 5. MAarrowzero1
% End of user PostScript
.41071 .20238 m
.41071 .20238 L
s
% Start of user PostScript
13. 6. MAarrowzero1
% End of user PostScript
.41071 .23214 m
.41071 .23214 L
s
% Start of user PostScript
13. 7. MAarrowzero1
% End of user PostScript
.41071 .2619 m
.38095 .29167 L
s
% Start of user PostScript
13. 8. 12. 9. MAarrowhead1
% End of user PostScript
.41071 .29167 m
.35119 .32143 L
s
% Start of user PostScript
13. 9. 11. 10. MAarrowhead1
% End of user PostScript
.41071 .32143 m
.35119 .35119 L
s
% Start of user PostScript
13. 10. 11. 11. MAarrowhead1
% End of user PostScript
.41071 .35119 m
.35119 .38095 L
s
% Start of user PostScript
13. 11. 11. 12. MAarrowhead1
% End of user PostScript
.41071 .38095 m
.35119 .38095 L
s
% Start of user PostScript
13. 12. 11. 12. MAarrowhead1
% End of user PostScript
.41071 .41071 m
.35119 .41071 L
s
% Start of user PostScript
13. 13. 11. 13. MAarrowhead1
% End of user PostScript
.41071 .44048 m
.35119 .44048 L
s
% Start of user PostScript
13. 14. 11. 14. MAarrowhead1
% End of user PostScript
.41071 .47024 m
.35119 .47024 L
s
% Start of user PostScript
13. 15. 11. 15. MAarrowhead1
% End of user PostScript
.41071 .5 m
.35119 .5 L
s
% Start of user PostScript
13. 16. 11. 16. MAarrowhead1
% End of user PostScript
.41071 .52976 m
.35119 .52976 L
s
% Start of user PostScript
13. 17. 11. 17. MAarrowhead1
% End of user PostScript
.41071 .55952 m
.35119 .55952 L
s
% Start of user PostScript
13. 18. 11. 18. MAarrowhead1
% End of user PostScript
.41071 .58929 m
.38095 .61905 L
s
% Start of user PostScript
13. 19. 12. 20. MAarrowhead1
% End of user PostScript
.41071 .61905 m
.38095 .64881 L
s
% Start of user PostScript
13. 20. 12. 21. MAarrowhead1
% End of user PostScript
.41071 .64881 m
.38095 .67857 L
s
% Start of user PostScript
13. 21. 12. 22. MAarrowhead1
% End of user PostScript
.41071 .67857 m
.38095 .70833 L
s
% Start of user PostScript
13. 22. 12. 23. MAarrowhead1
% End of user PostScript
.41071 .70833 m
.35119 .70833 L
s
% Start of user PostScript
13. 23. 11. 23. MAarrowhead1
% End of user PostScript
.41071 .7381 m
.35119 .7381 L
s
% Start of user PostScript
13. 24. 11. 24. MAarrowhead1
% End of user PostScript
.41071 .76786 m
.35119 .76786 L
s
% Start of user PostScript
13. 25. 11. 25. MAarrowhead1
% End of user PostScript
.41071 .79762 m
.35119 .79762 L
s
% Start of user PostScript
13. 26. 11. 26. MAarrowhead1
% End of user PostScript
.41071 .82738 m
.41071 .82738 L
s
% Start of user PostScript
13. 27. MAarrowzero1
% End of user PostScript
.41071 .85714 m
.41071 .85714 L
s
% Start of user PostScript
13. 28. MAarrowzero1
% End of user PostScript
.41071 .8869 m
.41071 .8869 L
s
% Start of user PostScript
13. 29. MAarrowzero1
% End of user PostScript
.41071 .91667 m
.41071 .91667 L
s
% Start of user PostScript
13. 30. MAarrowzero1
% End of user PostScript
.41071 .94643 m
.41071 .94643 L
s
% Start of user PostScript
13. 31. MAarrowzero1
% End of user PostScript
.41071 .97619 m
.41071 .97619 L
s
% Start of user PostScript
13. 32. MAarrowzero1
% End of user PostScript
.44048 .05357 m
.44048 .05357 L
s
% Start of user PostScript
14. 1. MAarrowzero1
% End of user PostScript
.44048 .08333 m
.44048 .08333 L
s
% Start of user PostScript
14. 2. MAarrowzero1
% End of user PostScript
.44048 .1131 m
.44048 .1131 L
s
% Start of user PostScript
14. 3. MAarrowzero1
% End of user PostScript
.44048 .14286 m
.44048 .14286 L
s
% Start of user PostScript
14. 4. MAarrowzero1
% End of user PostScript
.44048 .17262 m
.44048 .17262 L
s
% Start of user PostScript
14. 5. MAarrowzero1
% End of user PostScript
.44048 .20238 m
.44048 .20238 L
s
% Start of user PostScript
14. 6. MAarrowzero1
% End of user PostScript
.44048 .23214 m
.44048 .23214 L
s
% Start of user PostScript
14. 7. MAarrowzero1
% End of user PostScript
.44048 .2619 m
.41071 .2619 L
s
% Start of user PostScript
14. 8. 13. 8. MAarrowhead1
% End of user PostScript
.44048 .29167 m
.38095 .32143 L
s
% Start of user PostScript
14. 9. 12. 10. MAarrowhead1
% End of user PostScript
.44048 .32143 m
.38095 .35119 L
s
% Start of user PostScript
14. 10. 12. 11. MAarrowhead1
% End of user PostScript
.44048 .35119 m
.38095 .38095 L
s
% Start of user PostScript
14. 11. 12. 12. MAarrowhead1
% End of user PostScript
.44048 .38095 m
.38095 .38095 L
s
% Start of user PostScript
14. 12. 12. 12. MAarrowhead1
% End of user PostScript
.44048 .41071 m
.38095 .41071 L
s
% Start of user PostScript
14. 13. 12. 13. MAarrowhead1
% End of user PostScript
.44048 .44048 m
.38095 .44048 L
s
% Start of user PostScript
14. 14. 12. 14. MAarrowhead1
% End of user PostScript
.44048 .47024 m
.38095 .47024 L
s
% Start of user PostScript
14. 15. 12. 15. MAarrowhead1
% End of user PostScript
.44048 .5 m
.41071 .55952 L
s
% Start of user PostScript
14. 16. 13. 18. MAarrowhead1
% End of user PostScript
.44048 .52976 m
.41071 .58929 L
s
% Start of user PostScript
14. 17. 13. 19. MAarrowhead1
% End of user PostScript
.44048 .55952 m
.38095 .55952 L
s
% Start of user PostScript
14. 18. 12. 18. MAarrowhead1
% End of user PostScript
.44048 .58929 m
.41071 .61905 L
s
% Start of user PostScript
14. 19. 13. 20. MAarrowhead1
% End of user PostScript
.44048 .61905 m
.41071 .64881 L
s
% Start of user PostScript
14. 20. 13. 21. MAarrowhead1
% End of user PostScript
.44048 .64881 m
.41071 .67857 L
s
% Start of user PostScript
14. 21. 13. 22. MAarrowhead1
% End of user PostScript
.44048 .67857 m
.41071 .70833 L
s
% Start of user PostScript
14. 22. 13. 23. MAarrowhead1
% End of user PostScript
.44048 .70833 m
.38095 .70833 L
s
% Start of user PostScript
14. 23. 12. 23. MAarrowhead1
% End of user PostScript
.44048 .7381 m
.38095 .7381 L
s
% Start of user PostScript
14. 24. 12. 24. MAarrowhead1
% End of user PostScript
.44048 .76786 m
.38095 .76786 L
s
% Start of user PostScript
14. 25. 12. 25. MAarrowhead1
% End of user PostScript
.44048 .79762 m
.38095 .79762 L
s
% Start of user PostScript
14. 26. 12. 26. MAarrowhead1
% End of user PostScript
.44048 .82738 m
.41071 .82738 L
s
% Start of user PostScript
14. 27. 13. 27. MAarrowhead1
% End of user PostScript
.44048 .85714 m
.44048 .85714 L
s
% Start of user PostScript
14. 28. MAarrowzero1
% End of user PostScript
.44048 .8869 m
.44048 .8869 L
s
% Start of user PostScript
14. 29. MAarrowzero1
% End of user PostScript
.44048 .91667 m
.44048 .91667 L
s
% Start of user PostScript
14. 30. MAarrowzero1
% End of user PostScript
.44048 .94643 m
.44048 .94643 L
s
% Start of user PostScript
14. 31. MAarrowzero1
% End of user PostScript
.44048 .97619 m
.44048 .97619 L
s
% Start of user PostScript
14. 32. MAarrowzero1
% End of user PostScript
.47024 .05357 m
.47024 .05357 L
s
% Start of user PostScript
15. 1. MAarrowzero1
% End of user PostScript
.47024 .08333 m
.47024 .08333 L
s
% Start of user PostScript
15. 2. MAarrowzero1
% End of user PostScript
.47024 .1131 m
.47024 .1131 L
s
% Start of user PostScript
15. 3. MAarrowzero1
% End of user PostScript
.47024 .14286 m
.47024 .14286 L
s
% Start of user PostScript
15. 4. MAarrowzero1
% End of user PostScript
.47024 .17262 m
.47024 .17262 L
s
% Start of user PostScript
15. 5. MAarrowzero1
% End of user PostScript
.47024 .20238 m
.47024 .20238 L
s
% Start of user PostScript
15. 6. MAarrowzero1
% End of user PostScript
.47024 .23214 m
.47024 .23214 L
s
% Start of user PostScript
15. 7. MAarrowzero1
% End of user PostScript
.47024 .2619 m
.41071 .2619 L
s
% Start of user PostScript
15. 8. 13. 8. MAarrowhead1
% End of user PostScript
.47024 .29167 m
.41071 .32143 L
s
% Start of user PostScript
15. 9. 13. 10. MAarrowhead1
% End of user PostScript
.47024 .32143 m
.41071 .32143 L
s
% Start of user PostScript
15. 10. 13. 10. MAarrowhead1
% End of user PostScript
.47024 .35119 m
.41071 .35119 L
s
% Start of user PostScript
15. 11. 13. 11. MAarrowhead1
% End of user PostScript
.47024 .38095 m
.41071 .38095 L
s
% Start of user PostScript
15. 12. 13. 12. MAarrowhead1
% End of user PostScript
.47024 .41071 m
.41071 .41071 L
s
% Start of user PostScript
15. 13. 13. 13. MAarrowhead1
% End of user PostScript
.47024 .44048 m
.41071 .44048 L
s
% Start of user PostScript
15. 14. 13. 14. MAarrowhead1
% End of user PostScript
.47024 .47024 m
.41071 .47024 L
s
% Start of user PostScript
15. 15. 13. 15. MAarrowhead1
% End of user PostScript
.47024 .5 m
.41071 .5 L
s
% Start of user PostScript
15. 16. 13. 16. MAarrowhead1
% End of user PostScript
.47024 .52976 m
.41071 .52976 L
s
% Start of user PostScript
15. 17. 13. 17. MAarrowhead1
% End of user PostScript
.47024 .55952 m
.41071 .55952 L
s
% Start of user PostScript
15. 18. 13. 18. MAarrowhead1
% End of user PostScript
.47024 .58929 m
.41071 .58929 L
s
% Start of user PostScript
15. 19. 13. 19. MAarrowhead1
% End of user PostScript
.47024 .61905 m
.41071 .61905 L
s
% Start of user PostScript
15. 20. 13. 20. MAarrowhead1
% End of user PostScript
.47024 .64881 m
.44048 .67857 L
s
% Start of user PostScript
15. 21. 14. 22. MAarrowhead1
% End of user PostScript
.47024 .67857 m
.44048 .70833 L
s
% Start of user PostScript
15. 22. 14. 23. MAarrowhead1
% End of user PostScript
.47024 .70833 m
.47024 .7381 L
s
% Start of user PostScript
15. 23. 15. 24. MAarrowhead1
% End of user PostScript
.47024 .7381 m
.41071 .7381 L
s
% Start of user PostScript
15. 24. 13. 24. MAarrowhead1
% End of user PostScript
.47024 .76786 m
.41071 .76786 L
s
% Start of user PostScript
15. 25. 13. 25. MAarrowhead1
% End of user PostScript
.47024 .79762 m
.41071 .79762 L
s
% Start of user PostScript
15. 26. 13. 26. MAarrowhead1
% End of user PostScript
.47024 .82738 m
.44048 .85714 L
s
% Start of user PostScript
15. 27. 14. 28. MAarrowhead1
% End of user PostScript
.47024 .85714 m
.44048 .85714 L
s
% Start of user PostScript
15. 28. 14. 28. MAarrowhead1
% End of user PostScript
.47024 .8869 m
.47024 .8869 L
s
% Start of user PostScript
15. 29. MAarrowzero1
% End of user PostScript
.47024 .91667 m
.47024 .91667 L
s
% Start of user PostScript
15. 30. MAarrowzero1
% End of user PostScript
.47024 .94643 m
.47024 .94643 L
s
% Start of user PostScript
15. 31. MAarrowzero1
% End of user PostScript
.47024 .97619 m
.47024 .97619 L
s
% Start of user PostScript
15. 32. MAarrowzero1
% End of user PostScript
.5 .05357 m
.5 .05357 L
s
% Start of user PostScript
16. 1. MAarrowzero1
% End of user PostScript
.5 .08333 m
.5 .08333 L
s
% Start of user PostScript
16. 2. MAarrowzero1
% End of user PostScript
.5 .1131 m
.5 .1131 L
s
% Start of user PostScript
16. 3. MAarrowzero1
% End of user PostScript
.5 .14286 m
.5 .14286 L
s
% Start of user PostScript
16. 4. MAarrowzero1
% End of user PostScript
.5 .17262 m
.5 .17262 L
s
% Start of user PostScript
16. 5. MAarrowzero1
% End of user PostScript
.5 .20238 m
.5 .20238 L
s
% Start of user PostScript
16. 6. MAarrowzero1
% End of user PostScript
.5 .23214 m
.44048 .23214 L
s
% Start of user PostScript
16. 7. 14. 7. MAarrowhead1
% End of user PostScript
.5 .2619 m
.44048 .2619 L
s
% Start of user PostScript
16. 8. 14. 8. MAarrowhead1
% End of user PostScript
.5 .29167 m
.44048 .29167 L
s
% Start of user PostScript
16. 9. 14. 9. MAarrowhead1
% End of user PostScript
.5 .32143 m
.44048 .32143 L
s
% Start of user PostScript
16. 10. 14. 10. MAarrowhead1
% End of user PostScript
.5 .35119 m
.44048 .35119 L
s
% Start of user PostScript
16. 11. 14. 11. MAarrowhead1
% End of user PostScript
.5 .38095 m
.44048 .38095 L
s
% Start of user PostScript
16. 12. 14. 12. MAarrowhead1
% End of user PostScript
.5 .41071 m
.44048 .41071 L
s
% Start of user PostScript
16. 13. 14. 13. MAarrowhead1
% End of user PostScript
.5 .44048 m
.44048 .44048 L
s
% Start of user PostScript
16. 14. 14. 14. MAarrowhead1
% End of user PostScript
.5 .47024 m
.44048 .47024 L
s
% Start of user PostScript
16. 15. 14. 15. MAarrowhead1
% End of user PostScript
.5 .5 m
.44048 .5 L
s
% Start of user PostScript
16. 16. 14. 16. MAarrowhead1
% End of user PostScript
.5 .52976 m
.44048 .52976 L
s
% Start of user PostScript
16. 17. 14. 17. MAarrowhead1
% End of user PostScript
.5 .55952 m
.44048 .55952 L
s
% Start of user PostScript
16. 18. 14. 18. MAarrowhead1
% End of user PostScript
.5 .58929 m
.44048 .58929 L
s
% Start of user PostScript
16. 19. 14. 19. MAarrowhead1
% End of user PostScript
.5 .61905 m
.44048 .61905 L
s
% Start of user PostScript
16. 20. 14. 20. MAarrowhead1
% End of user PostScript
.5 .64881 m
.5 .67857 L
s
% Start of user PostScript
16. 21. 16. 22. MAarrowhead1
% End of user PostScript
.5 .67857 m
.47024 .70833 L
s
% Start of user PostScript
16. 22. 15. 23. MAarrowhead1
% End of user PostScript
.5 .70833 m
.47024 .7381 L
s
% Start of user PostScript
16. 23. 15. 24. MAarrowhead1
% End of user PostScript
.5 .7381 m
.44048 .7381 L
s
% Start of user PostScript
16. 24. 14. 24. MAarrowhead1
% End of user PostScript
.5 .76786 m
.44048 .76786 L
s
% Start of user PostScript
16. 25. 14. 25. MAarrowhead1
% End of user PostScript
.5 .79762 m
.44048 .79762 L
s
% Start of user PostScript
16. 26. 14. 26. MAarrowhead1
% End of user PostScript
.5 .82738 m
.44048 .85714 L
s
% Start of user PostScript
16. 27. 14. 28. MAarrowhead1
% End of user PostScript
.5 .85714 m
.47024 .85714 L
s
% Start of user PostScript
16. 28. 15. 28. MAarrowhead1
% End of user PostScript
.5 .8869 m
.47024 .8869 L
s
% Start of user PostScript
16. 29. 15. 29. MAarrowhead1
% End of user PostScript
.5 .91667 m
.5 .91667 L
s
% Start of user PostScript
16. 30. MAarrowzero1
% End of user PostScript
.5 .94643 m
.5 .94643 L
s
% Start of user PostScript
16. 31. MAarrowzero1
% End of user PostScript
.5 .97619 m
.5 .97619 L
s
% Start of user PostScript
16. 32. MAarrowzero1
% End of user PostScript
.52976 .05357 m
.52976 .05357 L
s
% Start of user PostScript
17. 1. MAarrowzero1
% End of user PostScript
.52976 .08333 m
.52976 .08333 L
s
% Start of user PostScript
17. 2. MAarrowzero1
% End of user PostScript
.52976 .1131 m
.52976 .1131 L
s
% Start of user PostScript
17. 3. MAarrowzero1
% End of user PostScript
.52976 .14286 m
.52976 .14286 L
s
% Start of user PostScript
17. 4. MAarrowzero1
% End of user PostScript
.52976 .17262 m
.52976 .17262 L
s
% Start of user PostScript
17. 5. MAarrowzero1
% End of user PostScript
.52976 .20238 m
.52976 .20238 L
s
% Start of user PostScript
17. 6. MAarrowzero1
% End of user PostScript
.52976 .23214 m
.47024 .23214 L
s
% Start of user PostScript
17. 7. 15. 7. MAarrowhead1
% End of user PostScript
.52976 .2619 m
.47024 .2619 L
s
% Start of user PostScript
17. 8. 15. 8. MAarrowhead1
% End of user PostScript
.52976 .29167 m
.47024 .29167 L
s
% Start of user PostScript
17. 9. 15. 9. MAarrowhead1
% End of user PostScript
.52976 .32143 m
.47024 .32143 L
s
% Start of user PostScript
17. 10. 15. 10. MAarrowhead1
% End of user PostScript
.52976 .35119 m
.47024 .35119 L
s
% Start of user PostScript
17. 11. 15. 11. MAarrowhead1
% End of user PostScript
.52976 .38095 m
.47024 .38095 L
s
% Start of user PostScript
17. 12. 15. 12. MAarrowhead1
% End of user PostScript
.52976 .41071 m
.47024 .41071 L
s
% Start of user PostScript
17. 13. 15. 13. MAarrowhead1
% End of user PostScript
.52976 .44048 m
.47024 .44048 L
s
% Start of user PostScript
17. 14. 15. 14. MAarrowhead1
% End of user PostScript
.52976 .47024 m
.47024 .47024 L
s
% Start of user PostScript
17. 15. 15. 15. MAarrowhead1
% End of user PostScript
.52976 .5 m
.47024 .5 L
s
% Start of user PostScript
17. 16. 15. 16. MAarrowhead1
% End of user PostScript
.52976 .52976 m
.47024 .52976 L
s
% Start of user PostScript
17. 17. 15. 17. MAarrowhead1
% End of user PostScript
.52976 .55952 m
.47024 .55952 L
s
% Start of user PostScript
17. 18. 15. 18. MAarrowhead1
% End of user PostScript
.52976 .58929 m
.47024 .58929 L
s
% Start of user PostScript
17. 19. 15. 19. MAarrowhead1
% End of user PostScript
.52976 .61905 m
.47024 .61905 L
s
% Start of user PostScript
17. 20. 15. 20. MAarrowhead1
% End of user PostScript
.52976 .64881 m
.47024 .64881 L
s
% Start of user PostScript
17. 21. 15. 21. MAarrowhead1
% End of user PostScript
.52976 .67857 m
.5 .70833 L
s
% Start of user PostScript
17. 22. 16. 23. MAarrowhead1
% End of user PostScript
.52976 .70833 m
.5 .7381 L
s
% Start of user PostScript
17. 23. 16. 24. MAarrowhead1
% End of user PostScript
.52976 .7381 m
.47024 .7381 L
s
% Start of user PostScript
17. 24. 15. 24. MAarrowhead1
% End of user PostScript
.52976 .76786 m
.47024 .76786 L
s
% Start of user PostScript
17. 25. 15. 25. MAarrowhead1
% End of user PostScript
.52976 .79762 m
.47024 .79762 L
s
% Start of user PostScript
17. 26. 15. 26. MAarrowhead1
% End of user PostScript
.52976 .82738 m
.5 .85714 L
s
% Start of user PostScript
17. 27. 16. 28. MAarrowhead1
% End of user PostScript
.52976 .85714 m
.5 .8869 L
s
% Start of user PostScript
17. 28. 16. 29. MAarrowhead1
% End of user PostScript
.52976 .8869 m
.5 .8869 L
s
% Start of user PostScript
17. 29. 16. 29. MAarrowhead1
% End of user PostScript
.52976 .91667 m
.5 .91667 L
s
% Start of user PostScript
17. 30. 16. 30. MAarrowhead1
% End of user PostScript
.52976 .94643 m
.47024 .91667 L
s
% Start of user PostScript
17. 31. 15. 30. MAarrowhead1
% End of user PostScript
.52976 .97619 m
.52976 .97619 L
s
% Start of user PostScript
17. 32. MAarrowzero1
% End of user PostScript
.55952 .05357 m
.55952 .05357 L
s
% Start of user PostScript
18. 1. MAarrowzero1
% End of user PostScript
.55952 .08333 m
.55952 .08333 L
s
% Start of user PostScript
18. 2. MAarrowzero1
% End of user PostScript
.55952 .1131 m
.55952 .1131 L
s
% Start of user PostScript
18. 3. MAarrowzero1
% End of user PostScript
.55952 .14286 m
.55952 .14286 L
s
% Start of user PostScript
18. 4. MAarrowzero1
% End of user PostScript
.55952 .17262 m
.55952 .17262 L
s
% Start of user PostScript
18. 5. MAarrowzero1
% End of user PostScript
.55952 .20238 m
.52976 .20238 L
s
% Start of user PostScript
18. 6. 17. 6. MAarrowhead1
% End of user PostScript
.55952 .23214 m
.5 .23214 L
s
% Start of user PostScript
18. 7. 16. 7. MAarrowhead1
% End of user PostScript
.55952 .2619 m
.5 .2619 L
s
% Start of user PostScript
18. 8. 16. 8. MAarrowhead1
% End of user PostScript
.55952 .29167 m
.5 .29167 L
s
% Start of user PostScript
18. 9. 16. 9. MAarrowhead1
% End of user PostScript
.55952 .32143 m
.5 .32143 L
s
% Start of user PostScript
18. 10. 16. 10. MAarrowhead1
% End of user PostScript
.55952 .35119 m
.5 .35119 L
s
% Start of user PostScript
18. 11. 16. 11. MAarrowhead1
% End of user PostScript
.55952 .38095 m
.5 .38095 L
s
% Start of user PostScript
18. 12. 16. 12. MAarrowhead1
% End of user PostScript
.55952 .41071 m
.5 .41071 L
s
% Start of user PostScript
18. 13. 16. 13. MAarrowhead1
% End of user PostScript
.55952 .44048 m
.5 .44048 L
s
% Start of user PostScript
18. 14. 16. 14. MAarrowhead1
% End of user PostScript
.55952 .47024 m
.5 .47024 L
s
% Start of user PostScript
18. 15. 16. 15. MAarrowhead1
% End of user PostScript
.55952 .5 m
.5 .5 L
s
% Start of user PostScript
18. 16. 16. 16. MAarrowhead1
% End of user PostScript
.55952 .52976 m
.5 .52976 L
s
% Start of user PostScript
18. 17. 16. 17. MAarrowhead1
% End of user PostScript
.55952 .55952 m
.5 .55952 L
s
% Start of user PostScript
18. 18. 16. 18. MAarrowhead1
% End of user PostScript
.55952 .58929 m
.5 .58929 L
s
% Start of user PostScript
18. 19. 16. 19. MAarrowhead1
% End of user PostScript
.55952 .61905 m
.52976 .64881 L
s
% Start of user PostScript
18. 20. 17. 21. MAarrowhead1
% End of user PostScript
.55952 .64881 m
.52976 .67857 L
s
% Start of user PostScript
18. 21. 17. 22. MAarrowhead1
% End of user PostScript
.55952 .67857 m
.52976 .70833 L
s
% Start of user PostScript
18. 22. 17. 23. MAarrowhead1
% End of user PostScript
.55952 .70833 m
.5 .70833 L
s
% Start of user PostScript
18. 23. 16. 23. MAarrowhead1
% End of user PostScript
.55952 .7381 m
.5 .7381 L
s
% Start of user PostScript
18. 24. 16. 24. MAarrowhead1
% End of user PostScript
.55952 .76786 m
.5 .76786 L
s
% Start of user PostScript
18. 25. 16. 25. MAarrowhead1
% End of user PostScript
.55952 .79762 m
.52976 .82738 L
s
% Start of user PostScript
18. 26. 17. 27. MAarrowhead1
% End of user PostScript
.55952 .82738 m
.52976 .85714 L
s
% Start of user PostScript
18. 27. 17. 28. MAarrowhead1
% End of user PostScript
.55952 .85714 m
.52976 .85714 L
s
% Start of user PostScript
18. 28. 17. 28. MAarrowhead1
% End of user PostScript
.55952 .8869 m
.52976 .8869 L
s
% Start of user PostScript
18. 29. 17. 29. MAarrowhead1
% End of user PostScript
.55952 .91667 m
.52976 .91667 L
s
% Start of user PostScript
18. 30. 17. 30. MAarrowhead1
% End of user PostScript
.55952 .94643 m
.52976 .94643 L
s
% Start of user PostScript
18. 31. 17. 31. MAarrowhead1
% End of user PostScript
.55952 .97619 m
.55952 .97619 L
s
% Start of user PostScript
18. 32. MAarrowzero1
% End of user PostScript
.58929 .05357 m
.58929 .05357 L
s
% Start of user PostScript
19. 1. MAarrowzero1
% End of user PostScript
.58929 .08333 m
.58929 .08333 L
s
% Start of user PostScript
19. 2. MAarrowzero1
% End of user PostScript
.58929 .1131 m
.58929 .1131 L
s
% Start of user PostScript
19. 3. MAarrowzero1
% End of user PostScript
.58929 .14286 m
.58929 .14286 L
s
% Start of user PostScript
19. 4. MAarrowzero1
% End of user PostScript
.58929 .17262 m
.58929 .17262 L
s
% Start of user PostScript
19. 5. MAarrowzero1
% End of user PostScript
.58929 .20238 m
.55952 .20238 L
s
% Start of user PostScript
19. 6. 18. 6. MAarrowhead1
% End of user PostScript
.58929 .23214 m
.55952 .20238 L
s
% Start of user PostScript
19. 7. 18. 6. MAarrowhead1
% End of user PostScript
.58929 .2619 m
.52976 .2619 L
s
% Start of user PostScript
19. 8. 17. 8. MAarrowhead1
% End of user PostScript
.58929 .29167 m
.52976 .29167 L
s
% Start of user PostScript
19. 9. 17. 9. MAarrowhead1
% End of user PostScript
.58929 .32143 m
.52976 .32143 L
s
% Start of user PostScript
19. 10. 17. 10. MAarrowhead1
% End of user PostScript
.58929 .35119 m
.52976 .35119 L
s
% Start of user PostScript
19. 11. 17. 11. MAarrowhead1
% End of user PostScript
.58929 .38095 m
.52976 .38095 L
s
% Start of user PostScript
19. 12. 17. 12. MAarrowhead1
% End of user PostScript
.58929 .41071 m
.52976 .41071 L
s
% Start of user PostScript
19. 13. 17. 13. MAarrowhead1
% End of user PostScript
.58929 .44048 m
.52976 .44048 L
s
% Start of user PostScript
19. 14. 17. 14. MAarrowhead1
% End of user PostScript
.58929 .47024 m
.52976 .47024 L
s
% Start of user PostScript
19. 15. 17. 15. MAarrowhead1
% End of user PostScript
.58929 .5 m
.52976 .5 L
s
% Start of user PostScript
19. 16. 17. 16. MAarrowhead1
% End of user PostScript
.58929 .52976 m
.55952 .55952 L
s
% Start of user PostScript
19. 17. 18. 18. MAarrowhead1
% End of user PostScript
.58929 .55952 m
.52976 .55952 L
s
% Start of user PostScript
19. 18. 17. 18. MAarrowhead1
% End of user PostScript
.58929 .58929 m
.52976 .58929 L
s
% Start of user PostScript
19. 19. 17. 19. MAarrowhead1
% End of user PostScript
.58929 .61905 m
.55952 .64881 L
s
% Start of user PostScript
19. 20. 18. 21. MAarrowhead1
% End of user PostScript
.58929 .64881 m
.55952 .67857 L
s
% Start of user PostScript
19. 21. 18. 22. MAarrowhead1
% End of user PostScript
.58929 .67857 m
.55952 .70833 L
s
% Start of user PostScript
19. 22. 18. 23. MAarrowhead1
% End of user PostScript
.58929 .70833 m
.52976 .70833 L
s
% Start of user PostScript
19. 23. 17. 23. MAarrowhead1
% End of user PostScript
.58929 .7381 m
.52976 .7381 L
s
% Start of user PostScript
19. 24. 17. 24. MAarrowhead1
% End of user PostScript
.58929 .76786 m
.55952 .79762 L
s
% Start of user PostScript
19. 25. 18. 26. MAarrowhead1
% End of user PostScript
.58929 .79762 m
.55952 .82738 L
s
% Start of user PostScript
19. 26. 18. 27. MAarrowhead1
% End of user PostScript
.58929 .82738 m
.55952 .85714 L
s
% Start of user PostScript
19. 27. 18. 28. MAarrowhead1
% End of user PostScript
.58929 .85714 m
.55952 .85714 L
s
% Start of user PostScript
19. 28. 18. 28. MAarrowhead1
% End of user PostScript
.58929 .8869 m
.55952 .8869 L
s
% Start of user PostScript
19. 29. 18. 29. MAarrowhead1
% End of user PostScript
.58929 .91667 m
.55952 .91667 L
s
% Start of user PostScript
19. 30. 18. 30. MAarrowhead1
% End of user PostScript
.58929 .94643 m
.55952 .94643 L
s
% Start of user PostScript
19. 31. 18. 31. MAarrowhead1
% End of user PostScript
.58929 .97619 m
.58929 .97619 L
s
% Start of user PostScript
19. 32. MAarrowzero1
% End of user PostScript
.61905 .05357 m
.61905 .05357 L
s
% Start of user PostScript
20. 1. MAarrowzero1
% End of user PostScript
.61905 .08333 m
.61905 .08333 L
s
% Start of user PostScript
20. 2. MAarrowzero1
% End of user PostScript
.61905 .1131 m
.61905 .1131 L
s
% Start of user PostScript
20. 3. MAarrowzero1
% End of user PostScript
.61905 .14286 m
.61905 .14286 L
s
% Start of user PostScript
20. 4. MAarrowzero1
% End of user PostScript
.61905 .17262 m
.61905 .17262 L
s
% Start of user PostScript
20. 5. MAarrowzero1
% End of user PostScript
.61905 .20238 m
.58929 .20238 L
s
% Start of user PostScript
20. 6. 19. 6. MAarrowhead1
% End of user PostScript
.61905 .23214 m
.55952 .23214 L
s
% Start of user PostScript
20. 7. 18. 7. MAarrowhead1
% End of user PostScript
.61905 .2619 m
.55952 .2619 L
s
% Start of user PostScript
20. 8. 18. 8. MAarrowhead1
% End of user PostScript
.61905 .29167 m
.55952 .29167 L
s
% Start of user PostScript
20. 9. 18. 9. MAarrowhead1
% End of user PostScript
.61905 .32143 m
.55952 .32143 L
s
% Start of user PostScript
20. 10. 18. 10. MAarrowhead1
% End of user PostScript
.61905 .35119 m
.55952 .35119 L
s
% Start of user PostScript
20. 11. 18. 11. MAarrowhead1
% End of user PostScript
.61905 .38095 m
.55952 .38095 L
s
% Start of user PostScript
20. 12. 18. 12. MAarrowhead1
% End of user PostScript
.61905 .41071 m
.55952 .41071 L
s
% Start of user PostScript
20. 13. 18. 13. MAarrowhead1
% End of user PostScript
.61905 .44048 m
.55952 .44048 L
s
% Start of user PostScript
20. 14. 18. 14. MAarrowhead1
% End of user PostScript
.61905 .47024 m
.55952 .47024 L
s
% Start of user PostScript
20. 15. 18. 15. MAarrowhead1
% End of user PostScript
.61905 .5 m
.55952 .5 L
s
% Start of user PostScript
20. 16. 18. 16. MAarrowhead1
% End of user PostScript
.61905 .52976 m
.58929 .55952 L
s
% Start of user PostScript
20. 17. 19. 18. MAarrowhead1
% End of user PostScript
.61905 .55952 m
.55952 .55952 L
s
% Start of user PostScript
20. 18. 18. 18. MAarrowhead1
% End of user PostScript
.61905 .58929 m
.55952 .58929 L
s
% Start of user PostScript
20. 19. 18. 19. MAarrowhead1
% End of user PostScript
.61905 .61905 m
.58929 .64881 L
s
% Start of user PostScript
20. 20. 19. 21. MAarrowhead1
% End of user PostScript
.61905 .64881 m
.58929 .67857 L
s
% Start of user PostScript
20. 21. 19. 22. MAarrowhead1
% End of user PostScript
.61905 .67857 m
.58929 .70833 L
s
% Start of user PostScript
20. 22. 19. 23. MAarrowhead1
% End of user PostScript
.61905 .70833 m
.55952 .70833 L
s
% Start of user PostScript
20. 23. 18. 23. MAarrowhead1
% End of user PostScript
.61905 .7381 m
.55952 .7381 L
s
% Start of user PostScript
20. 24. 18. 24. MAarrowhead1
% End of user PostScript
.61905 .76786 m
.58929 .79762 L
s
% Start of user PostScript
20. 25. 19. 26. MAarrowhead1
% End of user PostScript
.61905 .79762 m
.58929 .82738 L
s
% Start of user PostScript
20. 26. 19. 27. MAarrowhead1
% End of user PostScript
.61905 .82738 m
.58929 .85714 L
s
% Start of user PostScript
20. 27. 19. 28. MAarrowhead1
% End of user PostScript
.61905 .85714 m
.58929 .85714 L
s
% Start of user PostScript
20. 28. 19. 28. MAarrowhead1
% End of user PostScript
.61905 .8869 m
.58929 .8869 L
s
% Start of user PostScript
20. 29. 19. 29. MAarrowhead1
% End of user PostScript
.61905 .91667 m
.58929 .91667 L
s
% Start of user PostScript
20. 30. 19. 30. MAarrowhead1
% End of user PostScript
.61905 .94643 m
.58929 .94643 L
s
% Start of user PostScript
20. 31. 19. 31. MAarrowhead1
% End of user PostScript
.61905 .97619 m
.61905 .97619 L
s
% Start of user PostScript
20. 32. MAarrowzero1
% End of user PostScript
.64881 .05357 m
.64881 .05357 L
s
% Start of user PostScript
21. 1. MAarrowzero1
% End of user PostScript
.64881 .08333 m
.64881 .08333 L
s
% Start of user PostScript
21. 2. MAarrowzero1
% End of user PostScript
.64881 .1131 m
.64881 .1131 L
s
% Start of user PostScript
21. 3. MAarrowzero1
% End of user PostScript
.64881 .14286 m
.64881 .14286 L
s
% Start of user PostScript
21. 4. MAarrowzero1
% End of user PostScript
.64881 .17262 m
.64881 .17262 L
s
% Start of user PostScript
21. 5. MAarrowzero1
% End of user PostScript
.64881 .20238 m
.67857 .20238 L
s
% Start of user PostScript
21. 6. 22. 6. MAarrowhead1
% End of user PostScript
.64881 .23214 m
.58929 .23214 L
s
% Start of user PostScript
21. 7. 19. 7. MAarrowhead1
% End of user PostScript
.64881 .2619 m
.58929 .2619 L
s
% Start of user PostScript
21. 8. 19. 8. MAarrowhead1
% End of user PostScript
.64881 .29167 m
.58929 .29167 L
s
% Start of user PostScript
21. 9. 19. 9. MAarrowhead1
% End of user PostScript
.64881 .32143 m
.58929 .32143 L
s
% Start of user PostScript
21. 10. 19. 10. MAarrowhead1
% End of user PostScript
.64881 .35119 m
.58929 .35119 L
s
% Start of user PostScript
21. 11. 19. 11. MAarrowhead1
% End of user PostScript
.64881 .38095 m
.58929 .38095 L
s
% Start of user PostScript
21. 12. 19. 12. MAarrowhead1
% End of user PostScript
.64881 .41071 m
.58929 .38095 L
s
% Start of user PostScript
21. 13. 19. 12. MAarrowhead1
% End of user PostScript
.64881 .44048 m
.58929 .44048 L
s
% Start of user PostScript
21. 14. 19. 14. MAarrowhead1
% End of user PostScript
.64881 .47024 m
.58929 .47024 L
s
% Start of user PostScript
21. 15. 19. 15. MAarrowhead1
% End of user PostScript
.64881 .5 m
.58929 .5 L
s
% Start of user PostScript
21. 16. 19. 16. MAarrowhead1
% End of user PostScript
.64881 .52976 m
.61905 .55952 L
s
% Start of user PostScript
21. 17. 20. 18. MAarrowhead1
% End of user PostScript
.64881 .55952 m
.61905 .55952 L
s
% Start of user PostScript
21. 18. 20. 18. MAarrowhead1
% End of user PostScript
.64881 .58929 m
.64881 .64881 L
s
% Start of user PostScript
21. 19. 21. 21. MAarrowhead1
% End of user PostScript
.64881 .61905 m
.61905 .64881 L
s
% Start of user PostScript
21. 20. 20. 21. MAarrowhead1
% End of user PostScript
.64881 .64881 m
.61905 .67857 L
s
% Start of user PostScript
21. 21. 20. 22. MAarrowhead1
% End of user PostScript
.64881 .67857 m
.61905 .70833 L
s
% Start of user PostScript
21. 22. 20. 23. MAarrowhead1
% End of user PostScript
.64881 .70833 m
.61905 .7381 L
s
% Start of user PostScript
21. 23. 20. 24. MAarrowhead1
% End of user PostScript
.64881 .7381 m
.61905 .76786 L
s
% Start of user PostScript
21. 24. 20. 25. MAarrowhead1
% End of user PostScript
.64881 .76786 m
.61905 .79762 L
s
% Start of user PostScript
21. 25. 20. 26. MAarrowhead1
% End of user PostScript
.64881 .79762 m
.61905 .82738 L
s
% Start of user PostScript
21. 26. 20. 27. MAarrowhead1
% End of user PostScript
.64881 .82738 m
.61905 .85714 L
s
% Start of user PostScript
21. 27. 20. 28. MAarrowhead1
% End of user PostScript
.64881 .85714 m
.61905 .85714 L
s
% Start of user PostScript
21. 28. 20. 28. MAarrowhead1
% End of user PostScript
.64881 .8869 m
.61905 .8869 L
s
% Start of user PostScript
21. 29. 20. 29. MAarrowhead1
% End of user PostScript
.64881 .91667 m
.61905 .91667 L
s
% Start of user PostScript
21. 30. 20. 30. MAarrowhead1
% End of user PostScript
.64881 .94643 m
.61905 .94643 L
s
% Start of user PostScript
21. 31. 20. 31. MAarrowhead1
% End of user PostScript
.64881 .97619 m
.64881 .97619 L
s
% Start of user PostScript
21. 32. MAarrowzero1
% End of user PostScript
.67857 .05357 m
.67857 .05357 L
s
% Start of user PostScript
22. 1. MAarrowzero1
% End of user PostScript
.67857 .08333 m
.67857 .08333 L
s
% Start of user PostScript
22. 2. MAarrowzero1
% End of user PostScript
.67857 .1131 m
.67857 .1131 L
s
% Start of user PostScript
22. 3. MAarrowzero1
% End of user PostScript
.67857 .14286 m
.67857 .14286 L
s
% Start of user PostScript
22. 4. MAarrowzero1
% End of user PostScript
.67857 .17262 m
.67857 .17262 L
s
% Start of user PostScript
22. 5. MAarrowzero1
% End of user PostScript
.67857 .20238 m
.70833 .20238 L
s
% Start of user PostScript
22. 6. 23. 6. MAarrowhead1
% End of user PostScript
.67857 .23214 m
.61905 .23214 L
s
% Start of user PostScript
22. 7. 20. 7. MAarrowhead1
% End of user PostScript
.67857 .2619 m
.61905 .2619 L
s
% Start of user PostScript
22. 8. 20. 8. MAarrowhead1
% End of user PostScript
.67857 .29167 m
.61905 .29167 L
s
% Start of user PostScript
22. 9. 20. 9. MAarrowhead1
% End of user PostScript
.67857 .32143 m
.61905 .32143 L
s
% Start of user PostScript
22. 10. 20. 10. MAarrowhead1
% End of user PostScript
.67857 .35119 m
.61905 .35119 L
s
% Start of user PostScript
22. 11. 20. 11. MAarrowhead1
% End of user PostScript
.67857 .38095 m
.61905 .35119 L
s
% Start of user PostScript
22. 12. 20. 11. MAarrowhead1
% End of user PostScript
.67857 .41071 m
.61905 .38095 L
s
% Start of user PostScript
22. 13. 20. 12. MAarrowhead1
% End of user PostScript
.67857 .44048 m
.61905 .38095 L
s
% Start of user PostScript
22. 14. 20. 12. MAarrowhead1
% End of user PostScript
.67857 .47024 m
.61905 .47024 L
s
% Start of user PostScript
22. 15. 20. 15. MAarrowhead1
% End of user PostScript
.67857 .5 m
.61905 .5 L
s
% Start of user PostScript
22. 16. 20. 16. MAarrowhead1
% End of user PostScript
.67857 .52976 m
.61905 .52976 L
s
% Start of user PostScript
22. 17. 20. 17. MAarrowhead1
% End of user PostScript
.67857 .55952 m
.61905 .55952 L
s
% Start of user PostScript
22. 18. 20. 18. MAarrowhead1
% End of user PostScript
.67857 .58929 m
.61905 .58929 L
s
% Start of user PostScript
22. 19. 20. 19. MAarrowhead1
% End of user PostScript
.67857 .61905 m
.61905 .61905 L
s
% Start of user PostScript
22. 20. 20. 20. MAarrowhead1
% End of user PostScript
.67857 .64881 m
.64881 .67857 L
s
% Start of user PostScript
22. 21. 21. 22. MAarrowhead1
% End of user PostScript
.67857 .67857 m
.64881 .70833 L
s
% Start of user PostScript
22. 22. 21. 23. MAarrowhead1
% End of user PostScript
.67857 .70833 m
.64881 .7381 L
s
% Start of user PostScript
22. 23. 21. 24. MAarrowhead1
% End of user PostScript
.67857 .7381 m
.64881 .76786 L
s
% Start of user PostScript
22. 24. 21. 25. MAarrowhead1
% End of user PostScript
.67857 .76786 m
.64881 .76786 L
s
% Start of user PostScript
22. 25. 21. 25. MAarrowhead1
% End of user PostScript
.67857 .79762 m
.64881 .82738 L
s
% Start of user PostScript
22. 26. 21. 27. MAarrowhead1
% End of user PostScript
.67857 .82738 m
.64881 .85714 L
s
% Start of user PostScript
22. 27. 21. 28. MAarrowhead1
% End of user PostScript
.67857 .85714 m
.64881 .8869 L
s
% Start of user PostScript
22. 28. 21. 29. MAarrowhead1
% End of user PostScript
.67857 .8869 m
.64881 .8869 L
s
% Start of user PostScript
22. 29. 21. 29. MAarrowhead1
% End of user PostScript
.67857 .91667 m
.64881 .91667 L
s
% Start of user PostScript
22. 30. 21. 30. MAarrowhead1
% End of user PostScript
.67857 .94643 m
.64881 .94643 L
s
% Start of user PostScript
22. 31. 21. 31. MAarrowhead1
% End of user PostScript
.67857 .97619 m
.67857 .97619 L
s
% Start of user PostScript
22. 32. MAarrowzero1
% End of user PostScript
.70833 .05357 m
.70833 .05357 L
s
% Start of user PostScript
23. 1. MAarrowzero1
% End of user PostScript
.70833 .08333 m
.70833 .08333 L
s
% Start of user PostScript
23. 2. MAarrowzero1
% End of user PostScript
.70833 .1131 m
.70833 .1131 L
s
% Start of user PostScript
23. 3. MAarrowzero1
% End of user PostScript
.70833 .14286 m
.70833 .14286 L
s
% Start of user PostScript
23. 4. MAarrowzero1
% End of user PostScript
.70833 .17262 m
.70833 .17262 L
s
% Start of user PostScript
23. 5. MAarrowzero1
% End of user PostScript
.70833 .20238 m
.76786 .23214 L
s
% Start of user PostScript
23. 6. 25. 7. MAarrowhead1
% End of user PostScript
.70833 .23214 m
.76786 .23214 L
s
% Start of user PostScript
23. 7. 25. 7. MAarrowhead1
% End of user PostScript
.70833 .2619 m
.64881 .2619 L
s
% Start of user PostScript
23. 8. 21. 8. MAarrowhead1
% End of user PostScript
.70833 .29167 m
.64881 .29167 L
s
% Start of user PostScript
23. 9. 21. 9. MAarrowhead1
% End of user PostScript
.70833 .32143 m
.64881 .32143 L
s
% Start of user PostScript
23. 10. 21. 10. MAarrowhead1
% End of user PostScript
.70833 .35119 m
.64881 .35119 L
s
% Start of user PostScript
23. 11. 21. 11. MAarrowhead1
% End of user PostScript
.70833 .38095 m
.64881 .35119 L
s
% Start of user PostScript
23. 12. 21. 11. MAarrowhead1
% End of user PostScript
.70833 .41071 m
.64881 .38095 L
s
% Start of user PostScript
23. 13. 21. 12. MAarrowhead1
% End of user PostScript
.70833 .44048 m
.64881 .44048 L
s
% Start of user PostScript
23. 14. 21. 14. MAarrowhead1
% End of user PostScript
.70833 .47024 m
.64881 .47024 L
s
% Start of user PostScript
23. 15. 21. 15. MAarrowhead1
% End of user PostScript
.70833 .5 m
.64881 .5 L
s
% Start of user PostScript
23. 16. 21. 16. MAarrowhead1
% End of user PostScript
.70833 .52976 m
.64881 .58929 L
s
% Start of user PostScript
23. 17. 21. 19. MAarrowhead1
% End of user PostScript
.70833 .55952 m
.64881 .58929 L
s
% Start of user PostScript
23. 18. 21. 19. MAarrowhead1
% End of user PostScript
.70833 .58929 m
.64881 .58929 L
s
% Start of user PostScript
23. 19. 21. 19. MAarrowhead1
% End of user PostScript
.70833 .61905 m
.67857 .58929 L
s
% Start of user PostScript
23. 20. 22. 19. MAarrowhead1
% End of user PostScript
.70833 .64881 m
.67857 .58929 L
s
% Start of user PostScript
23. 21. 22. 19. MAarrowhead1
% End of user PostScript
.70833 .67857 m
.64881 .64881 L
s
% Start of user PostScript
23. 22. 21. 21. MAarrowhead1
% End of user PostScript
.70833 .70833 m
.64881 .67857 L
s
% Start of user PostScript
23. 23. 21. 22. MAarrowhead1
% End of user PostScript
.70833 .7381 m
.67857 .7381 L
s
% Start of user PostScript
23. 24. 22. 24. MAarrowhead1
% End of user PostScript
.70833 .76786 m
.67857 .76786 L
s
% Start of user PostScript
23. 25. 22. 25. MAarrowhead1
% End of user PostScript
.70833 .79762 m
.67857 .79762 L
s
% Start of user PostScript
23. 26. 22. 26. MAarrowhead1
% End of user PostScript
.70833 .82738 m
.67857 .85714 L
s
% Start of user PostScript
23. 27. 22. 28. MAarrowhead1
% End of user PostScript
.70833 .85714 m
.67857 .85714 L
s
% Start of user PostScript
23. 28. 22. 28. MAarrowhead1
% End of user PostScript
.70833 .8869 m
.67857 .8869 L
s
% Start of user PostScript
23. 29. 22. 29. MAarrowhead1
% End of user PostScript
.70833 .91667 m
.67857 .91667 L
s
% Start of user PostScript
23. 30. 22. 30. MAarrowhead1
% End of user PostScript
.70833 .94643 m
.67857 .94643 L
s
% Start of user PostScript
23. 31. 22. 31. MAarrowhead1
% End of user PostScript
.70833 .97619 m
.70833 .97619 L
s
% Start of user PostScript
23. 32. MAarrowzero1
% End of user PostScript
.7381 .05357 m
.7381 .05357 L
s
% Start of user PostScript
24. 1. MAarrowzero1
% End of user PostScript
.7381 .08333 m
.7381 .08333 L
s
% Start of user PostScript
24. 2. MAarrowzero1
% End of user PostScript
.7381 .1131 m
.7381 .1131 L
s
% Start of user PostScript
24. 3. MAarrowzero1
% End of user PostScript
.7381 .14286 m
.7381 .14286 L
s
% Start of user PostScript
24. 4. MAarrowzero1
% End of user PostScript
.7381 .17262 m
.7381 .17262 L
s
% Start of user PostScript
24. 5. MAarrowzero1
% End of user PostScript
.7381 .20238 m
.79762 .23214 L
s
% Start of user PostScript
24. 6. 26. 7. MAarrowhead1
% End of user PostScript
.7381 .23214 m
.67857 .2619 L
s
% Start of user PostScript
24. 7. 22. 8. MAarrowhead1
% End of user PostScript
.7381 .2619 m
.67857 .29167 L
s
% Start of user PostScript
24. 8. 22. 9. MAarrowhead1
% End of user PostScript
.7381 .29167 m
.67857 .29167 L
s
% Start of user PostScript
24. 9. 22. 9. MAarrowhead1
% End of user PostScript
.7381 .32143 m
.67857 .32143 L
s
% Start of user PostScript
24. 10. 22. 10. MAarrowhead1
% End of user PostScript
.7381 .35119 m
.67857 .35119 L
s
% Start of user PostScript
24. 11. 22. 11. MAarrowhead1
% End of user PostScript
.7381 .38095 m
.67857 .38095 L
s
% Start of user PostScript
24. 12. 22. 12. MAarrowhead1
% End of user PostScript
.7381 .41071 m
.67857 .41071 L
s
% Start of user PostScript
24. 13. 22. 13. MAarrowhead1
% End of user PostScript
.7381 .44048 m
.67857 .44048 L
s
% Start of user PostScript
24. 14. 22. 14. MAarrowhead1
% End of user PostScript
.7381 .47024 m
.67857 .47024 L
s
% Start of user PostScript
24. 15. 22. 15. MAarrowhead1
% End of user PostScript
.7381 .5 m
.67857 .5 L
s
% Start of user PostScript
24. 16. 22. 16. MAarrowhead1
% End of user PostScript
.7381 .52976 m
.67857 .58929 L
s
% Start of user PostScript
24. 17. 22. 19. MAarrowhead1
% End of user PostScript
.7381 .55952 m
.67857 .61905 L
s
% Start of user PostScript
24. 18. 22. 20. MAarrowhead1
% End of user PostScript
.7381 .58929 m
.67857 .64881 L
s
% Start of user PostScript
24. 19. 22. 21. MAarrowhead1
% End of user PostScript
.7381 .61905 m
.67857 .67857 L
s
% Start of user PostScript
24. 20. 22. 22. MAarrowhead1
% End of user PostScript
.7381 .64881 m
.70833 .58929 L
s
% Start of user PostScript
24. 21. 23. 19. MAarrowhead1
% End of user PostScript
.7381 .67857 m
.70833 .61905 L
s
% Start of user PostScript
24. 22. 23. 20. MAarrowhead1
% End of user PostScript
.7381 .70833 m
.70833 .64881 L
s
% Start of user PostScript
24. 23. 23. 21. MAarrowhead1
% End of user PostScript
.7381 .7381 m
.70833 .67857 L
s
% Start of user PostScript
24. 24. 23. 22. MAarrowhead1
% End of user PostScript
.7381 .76786 m
.70833 .76786 L
s
% Start of user PostScript
24. 25. 23. 25. MAarrowhead1
% End of user PostScript
.7381 .79762 m
.70833 .79762 L
s
% Start of user PostScript
24. 26. 23. 26. MAarrowhead1
% End of user PostScript
.7381 .82738 m
.70833 .82738 L
s
% Start of user PostScript
24. 27. 23. 27. MAarrowhead1
% End of user PostScript
.7381 .85714 m
.70833 .85714 L
s
% Start of user PostScript
24. 28. 23. 28. MAarrowhead1
% End of user PostScript
.7381 .8869 m
.70833 .8869 L
s
% Start of user PostScript
24. 29. 23. 29. MAarrowhead1
% End of user PostScript
.7381 .91667 m
.70833 .91667 L
s
% Start of user PostScript
24. 30. 23. 30. MAarrowhead1
% End of user PostScript
.7381 .94643 m
.70833 .94643 L
s
% Start of user PostScript
24. 31. 23. 31. MAarrowhead1
% End of user PostScript
.7381 .97619 m
.7381 .97619 L
s
% Start of user PostScript
24. 32. MAarrowzero1
% End of user PostScript
.76786 .05357 m
.76786 .05357 L
s
% Start of user PostScript
25. 1. MAarrowzero1
% End of user PostScript
.76786 .08333 m
.76786 .08333 L
s
% Start of user PostScript
25. 2. MAarrowzero1
% End of user PostScript
.76786 .1131 m
.76786 .1131 L
s
% Start of user PostScript
25. 3. MAarrowzero1
% End of user PostScript
.76786 .14286 m
.76786 .14286 L
s
% Start of user PostScript
25. 4. MAarrowzero1
% End of user PostScript
.76786 .17262 m
.76786 .17262 L
s
% Start of user PostScript
25. 5. MAarrowzero1
% End of user PostScript
.76786 .20238 m
.76786 .2619 L
s
% Start of user PostScript
25. 6. 25. 8. MAarrowhead1
% End of user PostScript
.76786 .23214 m
.7381 .2619 L
s
% Start of user PostScript
25. 7. 24. 8. MAarrowhead1
% End of user PostScript
.76786 .2619 m
.82738 .32143 L
s
% Start of user PostScript
25. 8. 27. 10. MAarrowhead1
% End of user PostScript
.76786 .29167 m
.70833 .29167 L
s
% Start of user PostScript
25. 9. 23. 9. MAarrowhead1
% End of user PostScript
.76786 .32143 m
.70833 .32143 L
s
% Start of user PostScript
25. 10. 23. 10. MAarrowhead1
% End of user PostScript
.76786 .35119 m
.70833 .35119 L
s
% Start of user PostScript
25. 11. 23. 11. MAarrowhead1
% End of user PostScript
.76786 .38095 m
.70833 .41071 L
s
% Start of user PostScript
25. 12. 23. 13. MAarrowhead1
% End of user PostScript
.76786 .41071 m
.70833 .44048 L
s
% Start of user PostScript
25. 13. 23. 14. MAarrowhead1
% End of user PostScript
.76786 .44048 m
.70833 .47024 L
s
% Start of user PostScript
25. 14. 23. 15. MAarrowhead1
% End of user PostScript
.76786 .47024 m
.70833 .5 L
s
% Start of user PostScript
25. 15. 23. 16. MAarrowhead1
% End of user PostScript
.76786 .5 m
.7381 .44048 L
s
% Start of user PostScript
25. 16. 24. 14. MAarrowhead1
% End of user PostScript
.76786 .52976 m
.7381 .47024 L
s
% Start of user PostScript
25. 17. 24. 15. MAarrowhead1
% End of user PostScript
.76786 .55952 m
.7381 .5 L
s
% Start of user PostScript
25. 18. 24. 16. MAarrowhead1
% End of user PostScript
.76786 .58929 m
.70833 .64881 L
s
% Start of user PostScript
25. 19. 23. 21. MAarrowhead1
% End of user PostScript
.76786 .61905 m
.70833 .67857 L
s
% Start of user PostScript
25. 20. 23. 22. MAarrowhead1
% End of user PostScript
.76786 .64881 m
.7381 .61905 L
s
% Start of user PostScript
25. 21. 24. 20. MAarrowhead1
% End of user PostScript
.76786 .67857 m
.7381 .64881 L
s
% Start of user PostScript
25. 22. 24. 21. MAarrowhead1
% End of user PostScript
.76786 .70833 m
.7381 .64881 L
s
% Start of user PostScript
25. 23. 24. 21. MAarrowhead1
% End of user PostScript
.76786 .7381 m
.7381 .67857 L
s
% Start of user PostScript
25. 24. 24. 22. MAarrowhead1
% End of user PostScript
.76786 .76786 m
.7381 .76786 L
s
% Start of user PostScript
25. 25. 24. 25. MAarrowhead1
% End of user PostScript
.76786 .79762 m
.7381 .79762 L
s
% Start of user PostScript
25. 26. 24. 26. MAarrowhead1
% End of user PostScript
.76786 .82738 m
.7381 .82738 L
s
% Start of user PostScript
25. 27. 24. 27. MAarrowhead1
% End of user PostScript
.76786 .85714 m
.7381 .85714 L
s
% Start of user PostScript
25. 28. 24. 28. MAarrowhead1
% End of user PostScript
.76786 .8869 m
.7381 .8869 L
s
% Start of user PostScript
25. 29. 24. 29. MAarrowhead1
% End of user PostScript
.76786 .91667 m
.7381 .91667 L
s
% Start of user PostScript
25. 30. 24. 30. MAarrowhead1
% End of user PostScript
.76786 .94643 m
.7381 .94643 L
s
% Start of user PostScript
25. 31. 24. 31. MAarrowhead1
% End of user PostScript
.76786 .97619 m
.76786 .97619 L
s
% Start of user PostScript
25. 32. MAarrowzero1
% End of user PostScript
.79762 .05357 m
.79762 .05357 L
s
% Start of user PostScript
26. 1. MAarrowzero1
% End of user PostScript
.79762 .08333 m
.79762 .08333 L
s
% Start of user PostScript
26. 2. MAarrowzero1
% End of user PostScript
.79762 .1131 m
.79762 .1131 L
s
% Start of user PostScript
26. 3. MAarrowzero1
% End of user PostScript
.79762 .14286 m
.79762 .14286 L
s
% Start of user PostScript
26. 4. MAarrowzero1
% End of user PostScript
.79762 .17262 m
.79762 .17262 L
s
% Start of user PostScript
26. 5. MAarrowzero1
% End of user PostScript
.79762 .20238 m
.79762 .20238 L
s
% Start of user PostScript
26. 6. MAarrowzero1
% End of user PostScript
.79762 .23214 m
.76786 .2619 L
s
% Start of user PostScript
26. 7. 25. 8. MAarrowhead1
% End of user PostScript
.79762 .2619 m
.85714 .32143 L
s
% Start of user PostScript
26. 8. 28. 10. MAarrowhead1
% End of user PostScript
.79762 .29167 m
.7381 .29167 L
s
% Start of user PostScript
26. 9. 24. 9. MAarrowhead1
% End of user PostScript
.79762 .32143 m
.7381 .32143 L
s
% Start of user PostScript
26. 10. 24. 10. MAarrowhead1
% End of user PostScript
.79762 .35119 m
.7381 .35119 L
s
% Start of user PostScript
26. 11. 24. 11. MAarrowhead1
% End of user PostScript
.79762 .38095 m
.7381 .41071 L
s
% Start of user PostScript
26. 12. 24. 13. MAarrowhead1
% End of user PostScript
.79762 .41071 m
.7381 .44048 L
s
% Start of user PostScript
26. 13. 24. 14. MAarrowhead1
% End of user PostScript
.79762 .44048 m
.7381 .47024 L
s
% Start of user PostScript
26. 14. 24. 15. MAarrowhead1
% End of user PostScript
.79762 .47024 m
.7381 .52976 L
s
% Start of user PostScript
26. 15. 24. 17. MAarrowhead1
% End of user PostScript
.79762 .5 m
.7381 .55952 L
s
% Start of user PostScript
26. 16. 24. 18. MAarrowhead1
% End of user PostScript
.79762 .52976 m
.76786 .47024 L
s
% Start of user PostScript
26. 17. 25. 15. MAarrowhead1
% End of user PostScript
.79762 .55952 m
.76786 .5 L
s
% Start of user PostScript
26. 18. 25. 16. MAarrowhead1
% End of user PostScript
.79762 .58929 m
.76786 .52976 L
s
% Start of user PostScript
26. 19. 25. 17. MAarrowhead1
% End of user PostScript
.79762 .61905 m
.7381 .67857 L
s
% Start of user PostScript
26. 20. 24. 22. MAarrowhead1
% End of user PostScript
.79762 .64881 m
.76786 .64881 L
s
% Start of user PostScript
26. 21. 25. 21. MAarrowhead1
% End of user PostScript
.79762 .67857 m
.76786 .64881 L
s
% Start of user PostScript
26. 22. 25. 21. MAarrowhead1
% End of user PostScript
.79762 .70833 m
.76786 .67857 L
s
% Start of user PostScript
26. 23. 25. 22. MAarrowhead1
% End of user PostScript
.79762 .7381 m
.76786 .67857 L
s
% Start of user PostScript
26. 24. 25. 22. MAarrowhead1
% End of user PostScript
.79762 .76786 m
.76786 .79762 L
s
% Start of user PostScript
26. 25. 25. 26. MAarrowhead1
% End of user PostScript
.79762 .79762 m
.76786 .82738 L
s
% Start of user PostScript
26. 26. 25. 27. MAarrowhead1
% End of user PostScript
.79762 .82738 m
.76786 .82738 L
s
% Start of user PostScript
26. 27. 25. 27. MAarrowhead1
% End of user PostScript
.79762 .85714 m
.76786 .85714 L
s
% Start of user PostScript
26. 28. 25. 28. MAarrowhead1
% End of user PostScript
.79762 .8869 m
.76786 .8869 L
s
% Start of user PostScript
26. 29. 25. 29. MAarrowhead1
% End of user PostScript
.79762 .91667 m
.76786 .91667 L
s
% Start of user PostScript
26. 30. 25. 30. MAarrowhead1
% End of user PostScript
.79762 .94643 m
.76786 .94643 L
s
% Start of user PostScript
26. 31. 25. 31. MAarrowhead1
% End of user PostScript
.79762 .97619 m
.79762 .97619 L
s
% Start of user PostScript
26. 32. MAarrowzero1
% End of user PostScript
.82738 .05357 m
.82738 .05357 L
s
% Start of user PostScript
27. 1. MAarrowzero1
% End of user PostScript
.82738 .08333 m
.82738 .08333 L
s
% Start of user PostScript
27. 2. MAarrowzero1
% End of user PostScript
.82738 .1131 m
.82738 .1131 L
s
% Start of user PostScript
27. 3. MAarrowzero1
% End of user PostScript
.82738 .14286 m
.82738 .14286 L
s
% Start of user PostScript
27. 4. MAarrowzero1
% End of user PostScript
.82738 .17262 m
.82738 .17262 L
s
% Start of user PostScript
27. 5. MAarrowzero1
% End of user PostScript
.82738 .20238 m
.82738 .20238 L
s
% Start of user PostScript
27. 6. MAarrowzero1
% End of user PostScript
.82738 .23214 m
.82738 .23214 L
s
% Start of user PostScript
27. 7. MAarrowzero1
% End of user PostScript
.82738 .2619 m
.85714 .2619 L
s
% Start of user PostScript
27. 8. 28. 8. MAarrowhead1
% End of user PostScript
.82738 .29167 m
.76786 .29167 L
s
% Start of user PostScript
27. 9. 25. 9. MAarrowhead1
% End of user PostScript
.82738 .32143 m
.76786 .32143 L
s
% Start of user PostScript
27. 10. 25. 10. MAarrowhead1
% End of user PostScript
.82738 .35119 m
.76786 .35119 L
s
% Start of user PostScript
27. 11. 25. 11. MAarrowhead1
% End of user PostScript
.82738 .38095 m
.76786 .41071 L
s
% Start of user PostScript
27. 12. 25. 13. MAarrowhead1
% End of user PostScript
.82738 .41071 m
.76786 .44048 L
s
% Start of user PostScript
27. 13. 25. 14. MAarrowhead1
% End of user PostScript
.82738 .44048 m
.76786 .5 L
s
% Start of user PostScript
27. 14. 25. 16. MAarrowhead1
% End of user PostScript
.82738 .47024 m
.76786 .52976 L
s
% Start of user PostScript
27. 15. 25. 17. MAarrowhead1
% End of user PostScript
.82738 .5 m
.76786 .55952 L
s
% Start of user PostScript
27. 16. 25. 18. MAarrowhead1
% End of user PostScript
.82738 .52976 m
.79762 .47024 L
s
% Start of user PostScript
27. 17. 26. 15. MAarrowhead1
% End of user PostScript
.82738 .55952 m
.79762 .5 L
s
% Start of user PostScript
27. 18. 26. 16. MAarrowhead1
% End of user PostScript
.82738 .58929 m
.79762 .64881 L
s
% Start of user PostScript
27. 19. 26. 21. MAarrowhead1
% End of user PostScript
.82738 .61905 m
.79762 .64881 L
s
% Start of user PostScript
27. 20. 26. 21. MAarrowhead1
% End of user PostScript
.82738 .64881 m
.79762 .64881 L
s
% Start of user PostScript
27. 21. 26. 21. MAarrowhead1
% End of user PostScript
.82738 .67857 m
.79762 .64881 L
s
% Start of user PostScript
27. 22. 26. 21. MAarrowhead1
% End of user PostScript
.82738 .70833 m
.79762 .67857 L
s
% Start of user PostScript
27. 23. 26. 22. MAarrowhead1
% End of user PostScript
.82738 .7381 m
.79762 .70833 L
s
% Start of user PostScript
27. 24. 26. 23. MAarrowhead1
% End of user PostScript
.82738 .76786 m
.79762 .76786 L
s
% Start of user PostScript
27. 25. 26. 25. MAarrowhead1
% End of user PostScript
.82738 .79762 m
.79762 .79762 L
s
% Start of user PostScript
27. 26. 26. 26. MAarrowhead1
% End of user PostScript
.82738 .82738 m
.79762 .82738 L
s
% Start of user PostScript
27. 27. 26. 27. MAarrowhead1
% End of user PostScript
.82738 .85714 m
.79762 .85714 L
s
% Start of user PostScript
27. 28. 26. 28. MAarrowhead1
% End of user PostScript
.82738 .8869 m
.79762 .8869 L
s
% Start of user PostScript
27. 29. 26. 29. MAarrowhead1
% End of user PostScript
.82738 .91667 m
.79762 .91667 L
s
% Start of user PostScript
27. 30. 26. 30. MAarrowhead1
% End of user PostScript
.82738 .94643 m
.79762 .94643 L
s
% Start of user PostScript
27. 31. 26. 31. MAarrowhead1
% End of user PostScript
.82738 .97619 m
.82738 .97619 L
s
% Start of user PostScript
27. 32. MAarrowzero1
% End of user PostScript
.85714 .05357 m
.85714 .05357 L
s
% Start of user PostScript
28. 1. MAarrowzero1
% End of user PostScript
.85714 .08333 m
.85714 .08333 L
s
% Start of user PostScript
28. 2. MAarrowzero1
% End of user PostScript
.85714 .1131 m
.85714 .1131 L
s
% Start of user PostScript
28. 3. MAarrowzero1
% End of user PostScript
.85714 .14286 m
.85714 .14286 L
s
% Start of user PostScript
28. 4. MAarrowzero1
% End of user PostScript
.85714 .17262 m
.85714 .17262 L
s
% Start of user PostScript
28. 5. MAarrowzero1
% End of user PostScript
.85714 .20238 m
.85714 .20238 L
s
% Start of user PostScript
28. 6. MAarrowzero1
% End of user PostScript
.85714 .23214 m
.8869 .29167 L
s
% Start of user PostScript
28. 7. 29. 9. MAarrowhead1
% End of user PostScript
.85714 .2619 m
.8869 .29167 L
s
% Start of user PostScript
28. 8. 29. 9. MAarrowhead1
% End of user PostScript
.85714 .29167 m
.8869 .29167 L
s
% Start of user PostScript
28. 9. 29. 9. MAarrowhead1
% End of user PostScript
.85714 .32143 m
.79762 .32143 L
s
% Start of user PostScript
28. 10. 26. 10. MAarrowhead1
% End of user PostScript
.85714 .35119 m
.79762 .35119 L
s
% Start of user PostScript
28. 11. 26. 11. MAarrowhead1
% End of user PostScript
.85714 .38095 m
.79762 .41071 L
s
% Start of user PostScript
28. 12. 26. 13. MAarrowhead1
% End of user PostScript
.85714 .41071 m
.79762 .44048 L
s
% Start of user PostScript
28. 13. 26. 14. MAarrowhead1
% End of user PostScript
.85714 .44048 m
.79762 .47024 L
s
% Start of user PostScript
28. 14. 26. 15. MAarrowhead1
% End of user PostScript
.85714 .47024 m
.82738 .47024 L
s
% Start of user PostScript
28. 15. 27. 15. MAarrowhead1
% End of user PostScript
.85714 .5 m
.82738 .55952 L
s
% Start of user PostScript
28. 16. 27. 18. MAarrowhead1
% End of user PostScript
.85714 .52976 m
.82738 .55952 L
s
% Start of user PostScript
28. 17. 27. 18. MAarrowhead1
% End of user PostScript
.85714 .55952 m
.82738 .55952 L
s
% Start of user PostScript
28. 18. 27. 18. MAarrowhead1
% End of user PostScript
.85714 .58929 m
.82738 .58929 L
s
% Start of user PostScript
28. 19. 27. 19. MAarrowhead1
% End of user PostScript
.85714 .61905 m
.82738 .61905 L
s
% Start of user PostScript
28. 20. 27. 20. MAarrowhead1
% End of user PostScript
.85714 .64881 m
.82738 .64881 L
s
% Start of user PostScript
28. 21. 27. 21. MAarrowhead1
% End of user PostScript
.85714 .67857 m
.82738 .64881 L
s
% Start of user PostScript
28. 22. 27. 21. MAarrowhead1
% End of user PostScript
.85714 .70833 m
.82738 .70833 L
s
% Start of user PostScript
28. 23. 27. 23. MAarrowhead1
% End of user PostScript
.85714 .7381 m
.82738 .7381 L
s
% Start of user PostScript
28. 24. 27. 24. MAarrowhead1
% End of user PostScript
.85714 .76786 m
.85714 .82738 L
s
% Start of user PostScript
28. 25. 28. 27. MAarrowhead1
% End of user PostScript
.85714 .79762 m
.85714 .85714 L
s
% Start of user PostScript
28. 26. 28. 28. MAarrowhead1
% End of user PostScript
.85714 .82738 m
.82738 .79762 L
s
% Start of user PostScript
28. 27. 27. 26. MAarrowhead1
% End of user PostScript
.85714 .85714 m
.85714 .8869 L
s
% Start of user PostScript
28. 28. 28. 29. MAarrowhead1
% End of user PostScript
.85714 .8869 m
.82738 .8869 L
s
% Start of user PostScript
28. 29. 27. 29. MAarrowhead1
% End of user PostScript
.85714 .91667 m
.82738 .91667 L
s
% Start of user PostScript
28. 30. 27. 30. MAarrowhead1
% End of user PostScript
.85714 .94643 m
.82738 .94643 L
s
% Start of user PostScript
28. 31. 27. 31. MAarrowhead1
% End of user PostScript
.85714 .97619 m
.85714 .97619 L
s
% Start of user PostScript
28. 32. MAarrowzero1
% End of user PostScript
.8869 .05357 m
.8869 .05357 L
s
% Start of user PostScript
29. 1. MAarrowzero1
% End of user PostScript
.8869 .08333 m
.8869 .08333 L
s
% Start of user PostScript
29. 2. MAarrowzero1
% End of user PostScript
.8869 .1131 m
.8869 .1131 L
s
% Start of user PostScript
29. 3. MAarrowzero1
% End of user PostScript
.8869 .14286 m
.8869 .14286 L
s
% Start of user PostScript
29. 4. MAarrowzero1
% End of user PostScript
.8869 .17262 m
.8869 .17262 L
s
% Start of user PostScript
29. 5. MAarrowzero1
% End of user PostScript
.8869 .20238 m
.8869 .20238 L
s
% Start of user PostScript
29. 6. MAarrowzero1
% End of user PostScript
.8869 .23214 m
.8869 .23214 L
s
% Start of user PostScript
29. 7. MAarrowzero1
% End of user PostScript
.8869 .2619 m
.8869 .2619 L
s
% Start of user PostScript
29. 8. MAarrowzero1
% End of user PostScript
.8869 .29167 m
.82738 .35119 L
s
% Start of user PostScript
29. 9. 27. 11. MAarrowhead1
% End of user PostScript
.8869 .32143 m
.82738 .35119 L
s
% Start of user PostScript
29. 10. 27. 11. MAarrowhead1
% End of user PostScript
.8869 .35119 m
.82738 .38095 L
s
% Start of user PostScript
29. 11. 27. 12. MAarrowhead1
% End of user PostScript
.8869 .38095 m
.82738 .41071 L
s
% Start of user PostScript
29. 12. 27. 13. MAarrowhead1
% End of user PostScript
.8869 .41071 m
.82738 .41071 L
s
% Start of user PostScript
29. 13. 27. 13. MAarrowhead1
% End of user PostScript
.8869 .44048 m
.82738 .44048 L
s
% Start of user PostScript
29. 14. 27. 14. MAarrowhead1
% End of user PostScript
.8869 .47024 m
.85714 .52976 L
s
% Start of user PostScript
29. 15. 28. 17. MAarrowhead1
% End of user PostScript
.8869 .5 m
.85714 .52976 L
s
% Start of user PostScript
29. 16. 28. 17. MAarrowhead1
% End of user PostScript
.8869 .52976 m
.85714 .52976 L
s
% Start of user PostScript
29. 17. 28. 17. MAarrowhead1
% End of user PostScript
.8869 .55952 m
.85714 .55952 L
s
% Start of user PostScript
29. 18. 28. 18. MAarrowhead1
% End of user PostScript
.8869 .58929 m
.85714 .55952 L
s
% Start of user PostScript
29. 19. 28. 18. MAarrowhead1
% End of user PostScript
.8869 .61905 m
.85714 .58929 L
s
% Start of user PostScript
29. 20. 28. 19. MAarrowhead1
% End of user PostScript
.8869 .64881 m
.85714 .64881 L
s
% Start of user PostScript
29. 21. 28. 21. MAarrowhead1
% End of user PostScript
.8869 .67857 m
.85714 .64881 L
s
% Start of user PostScript
29. 22. 28. 21. MAarrowhead1
% End of user PostScript
.8869 .70833 m
.85714 .70833 L
s
% Start of user PostScript
29. 23. 28. 23. MAarrowhead1
% End of user PostScript
.8869 .7381 m
.85714 .7381 L
s
% Start of user PostScript
29. 24. 28. 24. MAarrowhead1
% End of user PostScript
.8869 .76786 m
.8869 .82738 L
s
% Start of user PostScript
29. 25. 29. 27. MAarrowhead1
% End of user PostScript
.8869 .79762 m
.85714 .76786 L
s
% Start of user PostScript
29. 26. 28. 25. MAarrowhead1
% End of user PostScript
.8869 .82738 m
.85714 .79762 L
s
% Start of user PostScript
29. 27. 28. 26. MAarrowhead1
% End of user PostScript
.8869 .85714 m
.85714 .82738 L
s
% Start of user PostScript
29. 28. 28. 27. MAarrowhead1
% End of user PostScript
.8869 .8869 m
.85714 .85714 L
s
% Start of user PostScript
29. 29. 28. 28. MAarrowhead1
% End of user PostScript
.8869 .91667 m
.85714 .91667 L
s
% Start of user PostScript
29. 30. 28. 30. MAarrowhead1
% End of user PostScript
.8869 .94643 m
.85714 .94643 L
s
% Start of user PostScript
29. 31. 28. 31. MAarrowhead1
% End of user PostScript
.8869 .97619 m
.8869 .97619 L
s
% Start of user PostScript
29. 32. MAarrowzero1
% End of user PostScript
.91667 .05357 m
.91667 .05357 L
s
% Start of user PostScript
30. 1. MAarrowzero1
% End of user PostScript
.91667 .08333 m
.91667 .08333 L
s
% Start of user PostScript
30. 2. MAarrowzero1
% End of user PostScript
.91667 .1131 m
.91667 .1131 L
s
% Start of user PostScript
30. 3. MAarrowzero1
% End of user PostScript
.91667 .14286 m
.91667 .14286 L
s
% Start of user PostScript
30. 4. MAarrowzero1
% End of user PostScript
.91667 .17262 m
.91667 .17262 L
s
% Start of user PostScript
30. 5. MAarrowzero1
% End of user PostScript
.91667 .20238 m
.91667 .20238 L
s
% Start of user PostScript
30. 6. MAarrowzero1
% End of user PostScript
.91667 .23214 m
.91667 .23214 L
s
% Start of user PostScript
30. 7. MAarrowzero1
% End of user PostScript
.91667 .2619 m
.91667 .2619 L
s
% Start of user PostScript
30. 8. MAarrowzero1
% End of user PostScript
.91667 .29167 m
.91667 .29167 L
s
% Start of user PostScript
30. 9. MAarrowzero1
% End of user PostScript
.91667 .32143 m
.91667 .29167 L
s
% Start of user PostScript
30. 10. 30. 9. MAarrowhead1
% End of user PostScript
.91667 .35119 m
.91667 .32143 L
s
% Start of user PostScript
30. 11. 30. 10. MAarrowhead1
% End of user PostScript
.91667 .38095 m
.85714 .41071 L
s
% Start of user PostScript
30. 12. 28. 13. MAarrowhead1
% End of user PostScript
.91667 .41071 m
.85714 .41071 L
s
% Start of user PostScript
30. 13. 28. 13. MAarrowhead1
% End of user PostScript
.91667 .44048 m
.85714 .41071 L
s
% Start of user PostScript
30. 14. 28. 13. MAarrowhead1
% End of user PostScript
.91667 .47024 m
.85714 .41071 L
s
% Start of user PostScript
30. 15. 28. 13. MAarrowhead1
% End of user PostScript
.91667 .5 m
.8869 .55952 L
s
% Start of user PostScript
30. 16. 29. 18. MAarrowhead1
% End of user PostScript
.91667 .52976 m
.8869 .52976 L
s
% Start of user PostScript
30. 17. 29. 17. MAarrowhead1
% End of user PostScript
.91667 .55952 m
.8869 .55952 L
s
% Start of user PostScript
30. 18. 29. 18. MAarrowhead1
% End of user PostScript
.91667 .58929 m
.8869 .55952 L
s
% Start of user PostScript
30. 19. 29. 18. MAarrowhead1
% End of user PostScript
.91667 .61905 m
.8869 .58929 L
s
% Start of user PostScript
30. 20. 29. 19. MAarrowhead1
% End of user PostScript
.91667 .64881 m
.8869 .64881 L
s
% Start of user PostScript
30. 21. 29. 21. MAarrowhead1
% End of user PostScript
.91667 .67857 m
.8869 .67857 L
s
% Start of user PostScript
30. 22. 29. 22. MAarrowhead1
% End of user PostScript
.91667 .70833 m
.8869 .70833 L
s
% Start of user PostScript
30. 23. 29. 23. MAarrowhead1
% End of user PostScript
.91667 .7381 m
.8869 .7381 L
s
% Start of user PostScript
30. 24. 29. 24. MAarrowhead1
% End of user PostScript
.91667 .76786 m
.8869 .76786 L
s
% Start of user PostScript
30. 25. 29. 25. MAarrowhead1
% End of user PostScript
.91667 .79762 m
.8869 .76786 L
s
% Start of user PostScript
30. 26. 29. 25. MAarrowhead1
% End of user PostScript
.91667 .82738 m
.8869 .79762 L
s
% Start of user PostScript
30. 27. 29. 26. MAarrowhead1
% End of user PostScript
.91667 .85714 m
.91667 .8869 L
s
% Start of user PostScript
30. 28. 30. 29. MAarrowhead1
% End of user PostScript
.91667 .8869 m
.8869 .8869 L
s
% Start of user PostScript
30. 29. 29. 29. MAarrowhead1
% End of user PostScript
.91667 .91667 m
.8869 .91667 L
s
% Start of user PostScript
30. 30. 29. 30. MAarrowhead1
% End of user PostScript
.91667 .94643 m
.8869 .94643 L
s
% Start of user PostScript
30. 31. 29. 31. MAarrowhead1
% End of user PostScript
.91667 .97619 m
.91667 .97619 L
s
% Start of user PostScript
30. 32. MAarrowzero1
% End of user PostScript
.94643 .05357 m
.94643 .05357 L
s
% Start of user PostScript
31. 1. MAarrowzero1
% End of user PostScript
.94643 .08333 m
.94643 .08333 L
s
% Start of user PostScript
31. 2. MAarrowzero1
% End of user PostScript
.94643 .1131 m
.94643 .1131 L
s
% Start of user PostScript
31. 3. MAarrowzero1
% End of user PostScript
.94643 .14286 m
.94643 .14286 L
s
% Start of user PostScript
31. 4. MAarrowzero1
% End of user PostScript
.94643 .17262 m
.94643 .17262 L
s
% Start of user PostScript
31. 5. MAarrowzero1
% End of user PostScript
.94643 .20238 m
.94643 .20238 L
s
% Start of user PostScript
31. 6. MAarrowzero1
% End of user PostScript
.94643 .23214 m
.94643 .23214 L
s
% Start of user PostScript
31. 7. MAarrowzero1
% End of user PostScript
.94643 .2619 m
.94643 .2619 L
s
% Start of user PostScript
31. 8. MAarrowzero1
% End of user PostScript
.94643 .29167 m
.94643 .29167 L
s
% Start of user PostScript
31. 9. MAarrowzero1
% End of user PostScript
.94643 .32143 m
.94643 .32143 L
s
% Start of user PostScript
31. 10. MAarrowzero1
% End of user PostScript
.94643 .35119 m
.94643 .35119 L
s
% Start of user PostScript
31. 11. MAarrowzero1
% End of user PostScript
.94643 .38095 m
.94643 .38095 L
s
% Start of user PostScript
31. 12. MAarrowzero1
% End of user PostScript
.94643 .41071 m
.91667 .35119 L
s
% Start of user PostScript
31. 13. 30. 11. MAarrowhead1
% End of user PostScript
.94643 .44048 m
.91667 .38095 L
s
% Start of user PostScript
31. 14. 30. 12. MAarrowhead1
% End of user PostScript
.94643 .47024 m
.91667 .41071 L
s
% Start of user PostScript
31. 15. 30. 13. MAarrowhead1
% End of user PostScript
.94643 .5 m
.91667 .44048 L
s
% Start of user PostScript
31. 16. 30. 14. MAarrowhead1
% End of user PostScript
.94643 .52976 m
.91667 .47024 L
s
% Start of user PostScript
31. 17. 30. 15. MAarrowhead1
% End of user PostScript
.94643 .55952 m
.91667 .55952 L
s
% Start of user PostScript
31. 18. 30. 18. MAarrowhead1
% End of user PostScript
.94643 .58929 m
.91667 .58929 L
s
% Start of user PostScript
31. 19. 30. 19. MAarrowhead1
% End of user PostScript
.94643 .61905 m
.91667 .58929 L
s
% Start of user PostScript
31. 20. 30. 19. MAarrowhead1
% End of user PostScript
.94643 .64881 m
.91667 .64881 L
s
% Start of user PostScript
31. 21. 30. 21. MAarrowhead1
% End of user PostScript
.94643 .67857 m
.91667 .67857 L
s
% Start of user PostScript
31. 22. 30. 22. MAarrowhead1
% End of user PostScript
.94643 .70833 m
.91667 .70833 L
s
% Start of user PostScript
31. 23. 30. 23. MAarrowhead1
% End of user PostScript
.94643 .7381 m
.91667 .7381 L
s
% Start of user PostScript
31. 24. 30. 24. MAarrowhead1
% End of user PostScript
.94643 .76786 m
.91667 .76786 L
s
% Start of user PostScript
31. 25. 30. 25. MAarrowhead1
% End of user PostScript
.94643 .79762 m
.91667 .79762 L
s
% Start of user PostScript
31. 26. 30. 26. MAarrowhead1
% End of user PostScript
.94643 .82738 m
.94643 .85714 L
s
% Start of user PostScript
31. 27. 31. 28. MAarrowhead1
% End of user PostScript
.94643 .85714 m
.91667 .85714 L
s
% Start of user PostScript
31. 28. 30. 28. MAarrowhead1
% End of user PostScript
.94643 .8869 m
.91667 .8869 L
s
% Start of user PostScript
31. 29. 30. 29. MAarrowhead1
% End of user PostScript
.94643 .91667 m
.91667 .91667 L
s
% Start of user PostScript
31. 30. 30. 30. MAarrowhead1
% End of user PostScript
.94643 .94643 m
.94643 .94643 L
s
% Start of user PostScript
31. 31. MAarrowzero1
% End of user PostScript
.94643 .97619 m
.94643 .97619 L
s
% Start of user PostScript
31. 32. MAarrowzero1
% End of user PostScript
.97619 .05357 m
.97619 .05357 L
s
% Start of user PostScript
32. 1. MAarrowzero1
% End of user PostScript
.97619 .08333 m
.97619 .08333 L
s
% Start of user PostScript
32. 2. MAarrowzero1
% End of user PostScript
.97619 .1131 m
.97619 .1131 L
s
% Start of user PostScript
32. 3. MAarrowzero1
% End of user PostScript
.97619 .14286 m
.97619 .14286 L
s
% Start of user PostScript
32. 4. MAarrowzero1
% End of user PostScript
.97619 .17262 m
.97619 .17262 L
s
% Start of user PostScript
32. 5. MAarrowzero1
% End of user PostScript
.97619 .20238 m
.97619 .20238 L
s
% Start of user PostScript
32. 6. MAarrowzero1
% End of user PostScript
.97619 .23214 m
.97619 .23214 L
s
% Start of user PostScript
32. 7. MAarrowzero1
% End of user PostScript
.97619 .2619 m
.97619 .2619 L
s
% Start of user PostScript
32. 8. MAarrowzero1
% End of user PostScript
.97619 .29167 m
.97619 .29167 L
s
% Start of user PostScript
32. 9. MAarrowzero1
% End of user PostScript
.97619 .32143 m
.97619 .32143 L
s
% Start of user PostScript
32. 10. MAarrowzero1
% End of user PostScript
.97619 .35119 m
.97619 .35119 L
s
% Start of user PostScript
32. 11. MAarrowzero1
% End of user PostScript
.97619 .38095 m
.97619 .38095 L
s
% Start of user PostScript
32. 12. MAarrowzero1
% End of user PostScript
.97619 .41071 m
.97619 .41071 L
s
% Start of user PostScript
32. 13. MAarrowzero1
% End of user PostScript
.97619 .44048 m
.97619 .44048 L
s
% Start of user PostScript
32. 14. MAarrowzero1
% End of user PostScript
.97619 .47024 m
.97619 .47024 L
s
% Start of user PostScript
32. 15. MAarrowzero1
% End of user PostScript
.97619 .5 m
.97619 .5 L
s
% Start of user PostScript
32. 16. MAarrowzero1
% End of user PostScript
.97619 .52976 m
.97619 .52976 L
s
% Start of user PostScript
32. 17. MAarrowzero1
% End of user PostScript
.97619 .55952 m
.97619 .55952 L
s
% Start of user PostScript
32. 18. MAarrowzero1
% End of user PostScript
.97619 .58929 m
.97619 .58929 L
s
% Start of user PostScript
32. 19. MAarrowzero1
% End of user PostScript
.97619 .61905 m
.97619 .61905 L
s
% Start of user PostScript
32. 20. MAarrowzero1
% End of user PostScript
.97619 .64881 m
.97619 .64881 L
s
% Start of user PostScript
32. 21. MAarrowzero1
% End of user PostScript
.97619 .67857 m
.97619 .67857 L
s
% Start of user PostScript
32. 22. MAarrowzero1
% End of user PostScript
.97619 .70833 m
.97619 .70833 L
s
% Start of user PostScript
32. 23. MAarrowzero1
% End of user PostScript
.97619 .7381 m
.97619 .7381 L
s
% Start of user PostScript
32. 24. MAarrowzero1
% End of user PostScript
.97619 .76786 m
.97619 .76786 L
s
% Start of user PostScript
32. 25. MAarrowzero1
% End of user PostScript
.97619 .79762 m
.97619 .79762 L
s
% Start of user PostScript
32. 26. MAarrowzero1
% End of user PostScript
.97619 .82738 m
.97619 .82738 L
s
% Start of user PostScript
32. 27. MAarrowzero1
% End of user PostScript
.97619 .85714 m
.97619 .85714 L
s
% Start of user PostScript
32. 28. MAarrowzero1
% End of user PostScript
.97619 .8869 m
.97619 .8869 L
s
% Start of user PostScript
32. 29. MAarrowzero1
% End of user PostScript
.97619 .91667 m
.97619 .91667 L
s
% Start of user PostScript
32. 30. MAarrowzero1
% End of user PostScript
.97619 .94643 m
.97619 .94643 L
s
% Start of user PostScript
32. 31. MAarrowzero1
% End of user PostScript
.97619 .97619 m
.97619 .97619 L
s
% Start of user PostScript
32. 32. MAarrowzero1
% End of user PostScript
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000180P10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdoojn_l00;34gTJ2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_[C^]:JVYZf][Jo^kP00
oiT0IP3o0020P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;0000?`00O`00_`0o000o
?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0
_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo?cooOcoo
_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no?gnoOgno
_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo
_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0_ol0oolo0?lo?oloOolo_olooomo
0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_mWEc:0Pl=WIfJj^[WMgMiW<09W<<iW<
IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`c
c<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co
<lcoI/coVLcoc<coool0<ol0I_l0VOl0c?l0oolc0?lc<olcI_lcVOlcc?lcoomV0?mVI_mVVOmVoonI
0?nI<onII_nIVOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<ool08@000?l08@000?l08@000?l08@0001@0
0P4V00<18P030@@00P4P00<10`030Al0104300817`040@800`4P00<110020A@0000C00040@000BP0
00<1000080000`40000200040@000B0000<100001@000`40000M00030@00008000@100017`000`40
000500030@00020000@100010P000`40000A00004`001040004X00030@00020000<100000P001040
004P00030@0000D000<100007P001@400001008000<100007P000`40000400030@00020000@10001
0P000`40000A00004`001040004U00<18`000`40000200040@000B0000<100000P030B8000@10001
0P000`40000O00030@0000<18@020@<000@100014`0001<000@100019@000`40000S00030@000080
00@1000180000`40000200030@0001l000@100010P001040004N00040@000@8000<100008`001040
004200030@000140000D00819P040B400P45008180020@@0104O008110020B000P4300@17`030@@0
0P4D0000o`0Q0000o`0Q0000o`0Q0000o`0Q00003`3o0A410@0000l000<100000`000`4000050003
0@0000H000<100001@000`40000500030@0000D000<100001@000`40000500030@0000D000<10000
1@000`40000500030@0000D000<100001@000`40000500030@0000D000<100001P000`4000050003
0@0000D000<100001@000`40000500030@0000D000<100001@000`40000500030@0000D000<10000
1@000`40000500030@0000D000<100001P000`40000500030@0000D000<100001@000`4000050003
0@0000D000<100000`010@40000?00030@0000<000<100009P000`40000U00030@0002H000<10000
9@000`40000U00030@0002H000<100004`010@40000?00030@000?l03@010@40000?00030@000?l0
3@010@40000800811@000`40003o00d00@4100001`001040004400030@000?l03@010@4000070004
0@000@@00`7o00`00P4100001`001040004400030@000?l03@010@40000700040@000@@000<10000
o`0=00410@0000P00P4500030@000?l03@010@40000?00030@000?l03@010@40000?00030@000?l0
3@010@40000?00030@000?l03@010@40000?00030@000?l03@010@40000?00813@000`4000050003
0@0000D000<100001@000`40000500030@0000D000<100001@000`40000500030@0000D000<10000
1@000`40000500030@0000H000<100001@000`40000500030@0000D000<100001@000`4000050003
0@0000D000<100001@000`40000500030@0000D000<100001@000`40000500030@0000H000<10000
1@000`40000500030@0000D000<100001@000`40000500030@0000D000<100001@000`4000050003
0@0000800P4100003`000`40003o00d00@4100003`000`40003o00d00@4100003`000`40003o00d0
0@4100003`000`40003o00d00@4100003`000`40003o00d00@4100003`000`40003o00d00@410000
3`000`40003o00d00@4100003`020@d000<100001@000`40000500030@0000D000<100001@000`40
000500030@0000D000<100001@000`40000500030@0000D000<100001@000`40000600030@0000D0
00<100001@000`40000500030@0000D000<100001@000`40000500030@0000D000<100001@000`40
000500030@0000D000<100001@000`40000600030@0000D000<100001@000`40000500030@0000D0
00<100001@000`40000500030@0000D000<100001@000`40000200810@0000l000<10000o`0=0041
0@0000l000<10000o`0=00410@0000l000<10000o`0=00410@0000l000<10000o`0=00410@0000l0
00<10000o`0=00410@0000l000<10000o`0=00410@0000l000<10000o`0=00410@0000l00P4=0003
0@0000D000<100001@000`40000500030@0000D000<100001@000`40000500030@0000D000<10000
1@000`40000500030@0000D000<100001P000`40000500030@0000D000<100001@000`4000050003
0@0000D000<100001@000`40000500030@0000D000<100001@000`40000500030@0000D000<10000
1P000`40000500030@0000D000<100001@000`40000500030@0000D000<100001@000`4000050003
0@0000D000<100000P020@40000?00030@000?l03@010@40000?00030@000?l03@010@40000?0003
0@000?l03@010@40000?00030@000?l03@010@40000?00030@000?l03@010@40000?00030@000?l0
3@010@40000?00030@000?l03@010@40000?00813@000`40000500030@0000D000<100001@000`40
000500030@0000D000<100001@000`40000500030@0000D000<100001@000`40000500030@0000H0
00<100001@000`40000500030@0000D000<100001@000`40000500030@0000D000<100001@000`40
000500030@0000D000<100001@000`40000500030@0000H000<100001@000`40000500030@0000D0
00<100001@000`40000500030@0000D000<100001@000`40000500030@0000800P4100003`000`40
003o00d00@4100003`000`40003o00d00@4100003`000`40003o00d00@4100003`000`40003o00d0
0@4100003`000`40003o00d00@4100001`030@D000<10000o`0=00410@0000X000<100000P000`40
003o00d00@4100002P000`40000200<130000`40000500030@0000D000<100001@000`4000050003
0@0000D000<100001@000`40000500030@0000D000<100001@000`40000500030@0000H000<10000
1@000`40000500030@0000D000<100001@000`40000500030@0000D000<100001@000`4000050003
0@0000D000<100001@000`40000500030@0000H000<100001@000`40000500030@0000D000<10000
1@000`40000500030@0000D000<100001@000`40000500030@0000800P4100001`030@D000<10000
o`0=00410@0000L000<100001@000`40003o00d00@4100001`040@@000<10000o`0=00410@0000l0
00<10000o`0=00410@0000l000<10000o`0=00410@0000l000<10000o`0=00410@0000l000<10000
T0030@D00`4500<13@030@D00`5A00410@0000l00P4=00030@0000D000<100001@000`4000050003
0@0000D000<100001@000`40000500030@0000D000<100001@000`40000500030@0000D000<10000
1P000`40000500030@0000D000<100001@000`40000500030@0000D06@4701811P020@H000<10000
1P000`40000500030@0000D000<100001@000`40000500030@0000D000<100001@000`4000020081
0@0000l000<10000S`040@801P4400@13@040@@0104300811P020@@000<10000?0010@40000?0003
0@00098000@100001@4700030@0000/000<100001@000`40000600811P020@8000<10000?0010@40
000?00030@0009L0104X00811P030Ch00@4100003`000`40002G00<1:`020@H00P4m00410@0000l0
00<10000V0000`400@0/00@10P000`400@030CT00@4100003`000`40002K00030@0002/00`400`01
000200<1>0010@40000?00030@0007L0104500<11@030@d00`400`01000200<11@030Ad01P4200D1
=`010@40000?00813@000`40000500030@0000D000<100001@000`40000500030@0000D000<10000
1@000`40000500030@0000D000<100001@000`40000500030@0000H000<100001@000`4000040481
1`0A0@D00`4800030@0000D000<100001@000`40000500030@0000D000<100001@000`4000020081
0@0000l000<10000MP050@@0104400@130040@@0104400@15`020@@010400`01000600030@0000h0
00<100008`010@40000?00030@0007T00P4700030@0000D000<100003@000`40000500030@0000D0
00<1000050000`40000400040@0000811P000`40000@00030@0002800@4100003`000`40002o0081
2P050@<000<100004@000`40000R00410@0000l000<10000_@020@T000@100401@400`01000D0003
0@0002400@4100003`000`40002i00040@010@/00P4200<100<00@405P000`40000Q00410@0000l0
00<10000^0030@`010400`010@050AH000<1000080010@40000?00030@0007400P4600811P020@H0
0P4600811P020@H00P4600811P020@@01@4:00<110040@l00P4700030@0002000@4100003`000`40
000<00030@0000D000<100001@000`40000500030@0000D000<100001@000`40000500030@0000D0
00<100001@000`40000500030@0000D000<100003`050@<01@4300D10`050@<01@4300D10`050@<0
1@4300D10`040@`00P4500<140050@D000D100000@0700030@0000D000<100001@000`4000030041
0@0000l00P5L00811`000`40000505811P020@L00P4600811`090@<000<100007P020@40000?0003
0@0005P01@4800030@0000<000<100000P030@D00`4500<11@030@D00`4500<11@030@D00`4500<1
3`020@/000<100001@000`40000600<110001040004Q00410@0000l000<10000E`050@X000D10000
0@1E00813P000`40000500030@0000d000<100000P4O00410@0000l000<10000F0030@`000@10041
D`020A4000<100001@000`40000=00@17`010@40000?00030@0005P00P4=00@1CP000`4000020A@0
00<100001@000`40000=00@17P010@40000?00030@0005L000<1000030060D`0104G00030@0000D0
00<100002`050Ah00@4100003`000`40001F00030@0000`00`4200813@020@H00P4600811P020@H0
0P4600811P020@H00P4400D11P020@H00P4600<11`000`400009008100@00@417P010@40000?0003
0@0000`000<100001@000`40000500030@0000D000<100001@000`40000500030@0000D000<10000
1@000`40000>00030@0000d00P4400030@0000P01@4300D10`050@<01@4300D10`050@<01@4300D1
0`050@@01@4300D10`050@03004000H000<1000020070@L000<100001@000`40000500030@0000<0
0@4100003`020Dd000<100001@000`40000400811P020@H00P4606`11`090@L000<100004P020@40
000?00030@0004/000<100002`020@H00P4600811P020@D0104400<11@030@D00`4500<11@030@D0
0`4500<11@030@H00`4500<11@030@<000<100001@000`40000600<10P000`40000600030@0001<0
0@4100003`000`40001:00030@0000X00P4600811P020@H00P470081GP000`40000500030@0000T0
00<100001P020AD00@4100003`000`40001900030@0000T00P4600811P020@H00P470081H@000`40
0@0500030@0100L000<100001`030A@00@4100003`000`40001800030@0000D000<100000P430003
0@0000810`000`4000020@<00P400`010@03008100@00@41I0020@H00P4600030@0000P00`4D0041
0@0000l000<10000A`000`40000500@110040@@0104400@110040FH0104400@110000`40000800@1
50010@40000010010@4400811@000`40001600030@0000D0104400@110040@@01@4300D11P020@H0
0P4600811P020@H00P4600811P020@H00P4600811`020@H00P4600811P040@D00`4200030@0000/0
00<100004`010@40000200030@00008000@1000110000`40000<00030@0000D000<100001@000`40
000500030@0000D000<100001@000`40000500030@0000D00P4700H10P060@801P4200H10P060@<0
1@4300D10`050@<01@4300D10`050@<01@4300D10`050@@01@4300D10`050@<01`46008100<00@00
3@000`40000500030@0000D000<100000`010@40000200030@00008000@1000110030D@000<10000
1@030@D00`4400@110040@@0O04600811`000`40000B00810@00008000<100000P00104000440003
0@0004<000<100001@000`40000500030@0000<00P4600811P020@D00`4500<11@030@D00`4500<1
1@030@D00`4500<11@030@D00`4600<11@030@D00`4500<12@000`40000300812@000`40000C0041
0@00008000<100000P001040004400030@00048000<100001@000`40000500030@0000800P460081
1P020@H00P5c00030@0000800P4:00815@010@4000000`010@0500811@000`40000o00030@0100L0
00<100001@00104000020@H00P4600811P020G4000@100010`020@`00`4D00410@0000l000<10000
?`020@L000<100001@000`4000020@<000<100000P4300030@0000810`020@03004107<00`420081
3P030A@00@4100003`000`40000n00@11@000`40000500@110040@@0104400@1M0060@l0104D0041
0@0000l000<10000?@030@H000<100001@040@@0104400@110050@H00P4600811P020@H00P460081
1P020@H00P4600811P020@H00P4700811P020@H00P4600810`050A8000<100004`010@40000?0003
0@0003d00P4600030@0000D01P4200H10P060@801P4300D10`050@<01@4300D10`050@<01@4300D1
0`050@<01@4300D110050@<01@4300D10`050@<01@4B00030@0001<00@4100003`020@d000<10000
1@000`40000500030@0000D000<100001@000`40000500030@0000D000<100001@000`40000500<1
1@030@@0104400@1101l0@H00P4700030@0000D000<100001@000`40000200810@0000l000<10000
@`000`40000500030@0000D000<100000`020@H00P4600811@030@D00`4500<11@030@D00`4500<1
1@030@@0104400@11@030@H00`4500<11@030@D00`4?00812@000`40000C00410@0000l000<10000
@P000`40000500030@0000D000<100000P020@H00P4600811P020CL00`4500<1=P020@/00P4D0041
0@0000l000<10000?`000`400@0500030@0100L000@100000P4600811P020@H00P4j00040@010@@0
00@10041<P020@h00P4C00410@0000l000<10000?`020@H00P4700030@0000810`000`4000020@<0
00<100000P43008100<00@40@0020@H00P4[00030@00008140030A800@4100003`000`40000n00@1
10040@D0104400@110040@@0105400811P020BP0104B00814`010@40000?00030@0003d00`4500<1
1P040@@0104400@110050DH00P4600819@040AD000<1000040010@40000?00030@0003d00P460081
1P060@801P4200H10P060@@0104400@11@030@D00`4500<11@030@D00`4@00810`050B801@4E0003
0@0001000@4100003`020@d000<100001@000`40000500030@0000D000<100001@000`4000050003
0@0000D000<100001@000`40000506@11@0A0@H00P4600811`040@@00P4600811P020@<000D10000
0@0700030@0000800P4100003`000`40001300811P020@800`400`01000;00<11@030@D00`4500<1
1@030@D00`4500<11@030@<01@4200H10`050@D00`4?00811P020@L00P4600811P020@H00P400`01
000200030@0000l00@4100003`000`40001200811P020@H000<100002P020Ch0104300D110040AD0
0P4600811`020@H00P4600811P020@<00P4300030@0000h00@4100003`000`40000o00050@010040
0`001@400@0100D000<100002@020D400P400`010@0200H10`020@03004101400P4600811`020@H0
0P4600811P020@H00P4200030@0000h00@4100003`000`40000o008100<00@000`020@03004000@0
00<100001@000`4000020D@000@100001@4300810P00104000020@X000<100000P43008100<00@40
0`020@0400410@800P4010010@4300030@0000810`000`4000020@P00`4200030@0000d00@410000
3`000`40000n00D10`050@@000<100001@040D`00P400`01000300811P020@L0104400@110040@@0
104500@110040@X00P4300030@0000d00@4100003`000`40000m00<100<00@000P030@03004000<0
00<100001@040Dl00`4500811P020@@0104400D10`050@<01@4400@110040@d000D100000@0>0041
0@0000l000<10000?@020@8000@100000P4200060@0000400`4300H110040@@0104400@110040@D0
0`4500<11@030@D00`4@00811P020@<01@400`010@040@801P4200H10P060@<01@4300D13@001@40
000100h00@4100003`020@d000<100001@000`40000500030@0000D000<100001@000`4000050003
0@0000D000D100000@0300050@0000400`1J0@L000<100001@0C0@@0104400@11@0J0@<000D10000
0@0700030@0000800P4100003`000`40001000050@0000400`001040004300<100<00@000`030@D0
0`4500<11@030@D00`4500<11@030@D00`4500<14`000`40000700<13`020@H00P47008110040@03
004100@12P001@40000101400@4100003`000`40000o00030@00008000@100010P000`4000050003
0@0005X000<100005P020@H00P4700811P020@@01@4:00810`000`40000>00410@0000l000<10000
?@030@<01@400`01000600030@0005`000<100004`020@H00P4700811P020@H010400`010@090081
0P000`40000>00410@0000l000<10000?@040@801P4700030@0005h000<100003@000`4000020@<0
0P400`010@03008100@00@410P020@0400410@P0104300811`030@8000<100003@010@40000?0003
0@0003d00`4300L11@000`40001P00030@0000/0104400@110040@@0104<00040@000@@00P450081
0`000`40000=00410@0000l000<10000?0030@D0104600030@00068000<100002@040@@01@4300D1
0`050A0000<100000`020@D000D100000@0>00410@0000l000<10000?0020@H00`4600030@010081
1@040@@0104400@110040@@0104500<11@030@D00`4500<11@030@d00`400`01000200<10`060@80
1P4200H10P060@D00`4900030@0000@00P4300050@0000403P010@40000?00813@000`4000050003
0@0000D000<100001@000`40000500030@0000D000<100001@000`40000500030@0000D0O04400@1
10040@D04P4700050@0000400`000`40000500030@0000800P4100003`000`40001;00040@0000<1
0P00104000030@D00`4500<11@030@D00`4500<11@030@D00`4500<11@030@d00`4500<12P020@<0
0P46008120000`40000200@12P000`40000400050@0000404@010@40000?00030@0004X000<10000
1P000`40001_00D11P020@T000<100000`020@d000<100000`020@<000<100003P010@40000?0003
0@0004L000<100402@000`40001_00@11@020@X000<100000P020A0000<100000P030@8000<10000
3P010@40000?00030@0004L00P4900030@0006`000<100001`400`010@0;008100@00@414`000`40
000200<10P000`40000=00410@0000l000<10000AP040@P000<10000J`040@801P4<00@15`001040
00020@<000<100003@010@40000010010@4300<11@000`40001500<12@000`40001[00@10`060@/0
1@4H00060@000@010`000`40000<00410@00008000<100001@000`40000200030@0004D00P4800<1
1@040@@0104400@110040@@0104500<11@030@D00`4500<11@030@D00`4500<11@030@<01P4200H1
2P060@D00`4A00050@0000400`000`40000<00410@00008000<100001@000`40000200<130000`40
000500030@0000D000<100001@000`40000500030@0000D000<100001@000`40000500<11@1l0@04
004000811`030@D02P4700030@0000D000D100000@0300030@0000D000<100000P020@4000020003
0@0000800`4500030@0004@01P401@010@4000<100<00@000`030@D00`4500<11@030@D00`4500<1
1@030@D00`4500<11@030@D00`4500<11@030@X00P4200<100<00@401@030@D000<100000P030@X0
00<100001@001@40000101400@4100000P000`40000200030@0000D000<10000A0070@0300410081
0P000`40001`00H10`020@<000<100000P4300030@0000l000<100001P020@<000<100003P010@40
00000`010@0400@110000`40001400030@0100T100<00@00LP040@03004000800`400`01000200<1
00<00@004@000`40000600<10P000`40000>00410@0000l000<10000@`060@<01@5c00H100@00@00
104400@14P000`40000800<10P000`40000=00410@0000l000<10000@`070@<01@5^00@10P020@<0
00@10000104400814`000`40000800810`000`40000=00410@0000l000<10000@`050@8000<10040
1@5]00@10`000`400@0300030@0100811P020A8000<100002@000`400@0300030@0000`00@410000
3`000`40001400<110090@<0104400@130040@@0104500<11@030@D00`4500<11@030@D00`4500<1
1@030@<01P4400060@0000400`4500040@000A4000<100002`001@40000100h00@4100003`020@d0
00<100001@000`40000500030@0000D000<100001@000`40000500030@0000D000<100001@000`40
000507X10`001@40000100<000D100000@0300040@0000811`001@40000100<000<100001@001@40
000100<000<100001@000`40000200810@0000l000<10000=0000`40000400813`000`400@030@@0
1@4300D12`050@03004000D110040@@0104400@110040@@0104400@110040@@0104:00810P000`40
00030@8000<100000`4300030@0100T000D100000@0200030@0000H000D100000@0A00410@0000l0
00<10000<`020@D00`4?00030@00008000<100001@020@H00P4>00<11@020@L000<100001@000`40
000500030@0000D000<100001@000`40000500030@0000D000<100001@000`40000800<10P020@80
0P4200810P020@800P4900050@0000400P000`40000600030@0000<000<100003P010@40000?0003
0@0003<00P4400@13`000`40000200030@0001l000<10000C`040@03004000800`400`01000200<1
00<00@402@000`400@020@03004000P000<100000`000`40000>00410@0000l000<10000<`030@80
1P4=00<110000`40000M00030@0005401040100100040@04004000@100<00@001`000`4000040@T0
00<100001@000`40000=00410@0000l000<10000<P040@@0104=00<110000`40000M00030@000540
1P4300D10`020@<000<100001@000`4000040@T000<100001@000`40000=00410@0000l000<10000
<P040@<000<100400P4=00<11@000`40000K00030@0005800`4300030@0000810`000`4000020@@0
00<100000`000`4000030@X000<100001`000`40000<00410@0000l000<10000=0000`4000020004
0@000@l000<100001000104000040@@0104<00@110040@D00`4500<17@030A/00`4300030@010081
0`000`400@0200050@0000401@000`4000020@0400410@D00`4900030@0000`00@4100003`020@d0
00<100001@000`40000500030@0000D000<100001@000`40000500050@0000400`000`4000040081
1`000`40000504811`000`40000501411`000`40000500050@0000400`001@40000100<000D10000
0@0400050@0000400`0A0@L000<100001@000`40000200810@0000l000<10000;0000`4000050005
0@00004020020A<01P4200D12`050@03004000D110040@@0104A00030@0000D000<1000010400`01
000>00030@0000D000<100000P000`4000020@<000@100011@000`40000500040@000@80104400@1
60010@40000?00030@0002/00P4600810P000`400005008160020@H00P4<00030@0100811@020@L0
00<100001@000`40000>00030@0000D000<1000010020@l000<100001@000`40000400810P001040
00020@H000<1000010001040004600030@0000<000<1004060010@40000?00030@0002/00P460081
0P000`40000300816P000`40000B00810`000`40000T00030@0000D000<100001@000`40000=0003
0@0000D000<100001@00104000020@8000<100001P00104000030@03004000X00`4J00410@0000l0
00<10000:`030@D0104400816P020AD00`400`01000T00811P020@H00P4?00030@0000D000<10000
10020@04004000@100<00@001P00104000040@`00`4J00410@0000l000<10000:P040@@02@4L00<1
4`030@8000<100008`020@H00`4500<13@000`40000500030@0000D0104300<100<00@001`000`40
00040@`00`4J00410@0000l000<10000:P040@@01`4M00815@020@8000<100008`040@@00P460081
3P000`40000500030@0000D00P4300030@0000@100<00@001`040@d00`4K00410@0000l000<10000
;0000`40000500H16`020@P0104400@10P000`400@030@@0104500<11@030@D00`4300D10`020@H0
1@4;00030@0000D000<100001@020@@000<100400P4300030@0000H00`4010010@4500<10P020@04
00410@D00`4@00410@0000l00P4=00030@0000D000<100001@000`40000500030@0000D000<10000
1@030@D000<100001@000`40000500030@0000D000<100001@1J0@H00P4700030@0000D000D10000
0@0300030@0000H08@4700030@0000800P4100003`000`40000/00030@0000D000<1000010000`40
000500030@0000D000<100001@000`40000500030@0000H10P050@<01@4300D110040@@0104400@1
10040@`0104900030@0000<00`4800030@0000T000<1000030040@801P4200H110040A000@410000
3`000`40000/00030@0000D000<100000`000`40000500030@0000D000<100001@000`4000050003
0@0000@00P460081100010400@4600811`000`40000500030@0000D000<100001@000`40000=0003
0@0000H000<100000P020@03004000L000<100002`000`40000>00040@0000<100<00@000P030@03
004000H000<100003P010@40000?00030@0002`000<100001@000`40000200030@0000D000<10000
1@000`40000500030@0000D000<100001@000`40000:0081@P00104000020@8000<100001P000`40
000<00030@000180104500<16@010@40000?00030@0002`000<100001@000`4000020@H00P460081
1P020@H00P4600813@030Cl00P4200810P020@P000<100003P000`40000A00@11@030AT00@410000
3`000`40000/00030@0000D000<100000P4600811P020@H00P4600811P020@`00`5000@110020@L0
00<100003`000`40000B00040@000@D000<1004060010@40000?00030@0002`000<100001@000`40
0@030@@0104400@110040@@0104400@12`020D00104400@11@000`40000A00030@0001D000<10000
1@000`40000E00410@0000l000<10000;0000`40000500@11@030@D00`4500<11@030@D00`4=0003
0@0100<110040@D00`4500<11@030@D00`4=00<10`050@<00`4600030@00018000<100002P030@T0
00<100001@00104000030A000@4100003`020@d000<100001@000`40000500030@0000D000<10000
1@000`40000500030@0000D000<100001@000`40000500030@0000D000<100001@000`4000050003
0@0000D0D`4600030@0000D000<100001@000`40000600X11P020@H02@4700030@0000800P410000
3`000`40000l00030@0000D000<100001@000`40000500030@0000D000<100001@000`4000050003
0@0100@100<00@001@4400@110040@@0104400@130040@8000<100400`4200<11@000`4000050003
0@0000D000<100003P000`400@030@800`4500<11@060A000@4100003`000`40000k00030@0000D0
00<100001@000`40000500030@0000D000<100001@000`40000500030@0000<00`4500811`000`40
000500030@0000D000<100001@000`40000=00040@000@@000<100001@4200<11@000`4000050003
0@0000h000<100000`00104000030@D00`4500D140010@40000?00030@0003X000<100001@000`40
000500030@0000D000<100001@000`40000500030@0000D000<100001@000`40000h00030@0000D0
1`4010010@4300030@0000D000<100003`000`40000700@11@030@D0104@00410@0000l000<10000
=`000`400@0500030@0100D000<100401@000`400@0500030@0100D000<100401@000`400@050003
0@0103/000<100001P050@<00`400`01000600030@0000l000<1000020030@H00`4500<14@010@40
000?00030@0003L00P4600811P020@H00P4600811P020@H00P460081?0000`40000600@110040@L0
00<1000040000`40000900030@0100H00P4600814@010@40000?00030@0003H0104400@110040@@0
104400@110040@@0104400@1>`000`40000600D10`030@L000<1000040000`40000=00030@0000D0
00<100001@000`40000>00410@000003004100810`020@D000<10000=@030@D00`4500<11@030@D0
0`4500<11@030@D00`400`010@0600811P020BH00P4200030@0000H00`4200040@0000811P000`40
000A00030@0000@00P4800030@0000D000<100001@000`40000=00410@00008000<100000P001040
004400030@0000`000<100001@000`40000500030@0000D000<100003P020@H00P4600811P020@H0
0P4600811P020@H01P4300D10`050B<01@4200030@0000H00P401`01000100400P000`4000020003
0@00014000<100000P050@T000<100001@000`40000500030@0000H000<100000`010@4000030005
0@0000400P000`40000200<1;0000`40000500030@0000D000<100001@000`40000500030@0000D0
00<100001@000`40000500030@0000D08P4700030@0000D000<100001@0A0@<000D100000@030004
0@0000811P000`40000500050@0000400`0:0@L000<100001@000`40000500030@0000X00P410000
10001040004200030@00008000<10000;0000`40000600030@0000<000<100001@000`4000050003
0@0000D000<100001@000`40000500030@0000D000<100000P040@@00`4500<14@000`4000050003
0@0000D000@100000`400`010@0;00030@0000<10P000`4000060@8000<100001@000`4000020004
0@000@@00`4X00410@000005004000400P001040004400030@0002X0104800050@0000401`000`40
000500030@0000D000<100001@000`40000500030@0000D000<100001@000`40000N00030@0000D0
00<100001@000`40000500@12`020@0400410@800P4200H11`00104000020@8000<10000;P010@40
0002008110020@D000<10000:P040@T000<100401@000`400@0500030@0100D000<100401@000`40
0@0500030@0100D000<100401@000`400@0N00030@0100D000<100401@000`400@0500040@0100<1
2@000`400@0300<100<00@000`040@L000<100000P040C400@4100003`000`40000[00<12@020@H0
0P4600811P020@H00P4600811P020@H00P4O00811P020@H00P4600810P030@T00P40100100040@03
004000<00P4700030@000080104b00410@0000l000<10000:`030@P0104400@110040@@0104400@1
10040@@0104400@17@030@D0104400@110040@0400410@P01@4200H110000`400@0500030@0000<0
1@4a00410@0000l000<10000:`020@P00`4200030@0000<11@030@D00`4500<11@030@D00`4500<1
4@020@/00P4600<11@030@D00`4400030@0000H00`4300030@0100810`001040004300050@000040
0P020@800`400`010@0600811P020@H00P46008140010@40000?00030@0000`000<100001@000`40
000500030@0000D000<100001@000`4000060081100010400@4600811P020@H00P4600811P020@H0
0P4?00D12`000`40000500811P020@H00P4500030@0000H00P4400030@0100810`000`400@050003
0@0100D100<00@401@4300D10`050@<01@4300D12P000`40000300410@0000l00P4]00030@0000D0
00<100001@000`40000500030@0000D000<100001@000`40000500030@0000D000<100001@000`40
000500030@0000D04P4700030@0000D000<100001@000`40000500<11@000`40000500050@000040
0`020@H0<P4<00810@0000l000<10000=0000`40000500811@000`40000500030@0000D000<10000
1@000`40000500030@0000D000<100001@000`40000500030@0000800`4200030@0000`000<10000
1@000`40000500030@0000D000<1000010400`01000600<10P000`4000040@@010401@010@4000L1
0P060@801P4500<11@030A000@4100003`000`40000e00030@0000D00P4300030@0000D000<10000
1@000`40000500030@0000D000<100001@000`40000500030@0000D000<100001P040@`000<10000
1@000`40000500030@0000D000<100000P040@L000<100400P42008100<00@400`40100100040@@0
1@4400<11@030B800@4100003`000`40000e00030@0000D000D100400@0500030@0100D000<10040
1@000`400@0500030@0100D000<100401@000`400@0500030@0100T0104900030@0100D000<10040
1@000`400@0500030@0100D01@4300030@0100<00`400`01000200D110040@@00`4500@110040B40
0@4100003`000`40000f00030@0000D0104500811P020@H00P4600811P020@H00P4600812`030@T0
0P4600811P020@H00P4600810`020@03004100@01040100100040@H00P4600030@0100D000<10040
1@000`400@0Q00410@0000l000<10000=P000`40000500D10`040@@0104400@110040@@0104400@1
10040@X00`4800<11@040@@0104400@110040@@0104300H10P050@L000<100001P000`4000050003
0@0000D000<100007P010@40000?00030@0003L000<100000`070@03004100<10`050@<00`400`01
0@0200<11@030@D00`4500<130020@P00P4600<11@030@D00`4500<11@050@<00`4300030@010081
0P000`40000500030@0000H000<100001@000`40000500030@0000<00P46008140010@40000?0003
0@0000`000<100001@000`40000500030@0000D000<100001@000`40000500040@000@@020400`01
0@030@<01@4300H10P020@H00P4600811P020@h000<100001P000`40000500811P020@H00P460081
1P020@<00P4010010@4300030@0100810`000`40000500030@0000H000<100001@000`4000050003
0@0100@10`050@X000<100000`010@40000?0081>@001040000Y0@L000<100001@000`4000050081
1`000`40000500030@0000D000<100001@000`40000500030@0000D00`4500030@0000D000D10000
0@0300050@0000400`020@H00P4700030@0000D000<100001@0A0@`00P4100003`000`40000h0003
0@0000<0104500<11@030@H00`4@00030@0000D000<100001@000`40000600030@0000D000<10000
1@000`40000500030@0000D000<100001@000`4000040@03004000L00P4200040@0000810P001040
00030@D00`4I00<11@030A000@4100003`000`40000i00<10`050BH000<100001@000`4000050003
0@0000H000<100001@000`40000500030@0000D000<100001@000`40000500030@000080104900<1
0P000`4000030@8000@10000104400@1<P010@40000?00030@0003T00`4400@18`000`400@050003
0@0100D000<100401@001040004400040@000@D000<100401@000`400@0500030@0100D000<10040
1@050@T00`400`01000200<100<00@000P050@@0104a00410@0000l000<10000>0040@@01P4Q0081
1P020@H00P4600<11@030@H00P4600811P020@H00P4600810`020@L01040100100040@04004000@1
1P000`400@0a00410@0000l000<10000>@030@D00`4200817P040@@0104400@110040@@0104500<1
1@040@@0104400@110040@@00P4500810`000`4000020@<000D100000@0200030@0000H000<10000
;P010@40000?00030@0003X00`4500<10`020@H00P4700811P020@800`400`010@0200<11@030@H0
0P46008100<00@400`020@03004100<01@4300<11@030@D00`4800811@001@40000100<000D10000
0@0300040@000@P000<100000`020@H00P4600811P020A000@4100003`000`40000k00811P090@03
004100<110050@<01@4200H10P020@H00P4700030@0000D01@4300D10`050@<00P4600811P020@/0
0P4300050@0000400`001@40000100<000D100000@0800030@0100@10`050@<01@4300D140010@40
000?00813@000`40000500030@0000D000<100001@000`40000500030@0000D000<100001@000`40
0005034120000`40000502411`000`40000500030@0000D000D100000@0300050@0000400`001@40
000100<00P4702411`000`40000200810@0000l000<10000AP040@D00`4600<11@030@D00`4B0003
0@0000@000@100000`400`01000200<11@030A4000<100001@000`40000:00030@0000D000<10000
1@00104000030@T00`4500<11@030@D00`4@00410@0000l000<10000A`030C40104400030@0000D0
00<100007@000`40000500030@0000`000<100001@000`40000500040@0000@1<P010@40000?0003
0@0004P000@10041;`050@8000<100000`000`400@0M00030@0100D000<100403`000`4000050003
0@0000D000<100000P040C400@4100003`000`40001<0081;P060@H00P4N00811P020A4000<10000
1@000`40000500040@0000@1<@010@40000?00030@0004h00P4/00H11@030Ad0104400@140000`40
000500030@0000D000<100000P001040004`00410@0000l000<10000D0020BX0104600817P030@D0
0`4C00030@0000D000<100001@000`40000500030@0002d00@4100003`000`40001?00D10`040@@0
104400@110040@@01042008100@00@410`000`400@020@D00`4500<12`020@H00P4800<12@000`40
000500030@0000D000<100001P00104000030@D00`4500<11@030A000@4100003`020@d000<10000
1@000`40000500030@0000D000<100001@000`40000500030@0000D000<100001@000`40000505X1
1`000`40000500T11`000`40000500030@0000D000<100001P0Q0@L000<100000P020@40000?0003
0@0004l00`4600<11@030@D00`4500<11@030@D00`4500<11@030@D00`4A00030@0000D000@10000
0`4V00<11@030@D00`4500<140010@40000?00030@000:/000<100001@000`40001F00410@0000l0
00<10000Z0000`400@0500030@0105T00@4100003`000`40002X00811P020EX00@4100003`000`40
002W00@110040ET00@41000000<00@400P4200<11@000`40002V00<11@030E/00@4100000P000`40
000500030@00008000<10000E`040@@0104400@110040@@0104500<11@030A/00P4600D11@030@D0
104400@130040AD00`4500<140010@40000300030@0000@000<100000P030@`000<100001@000`40
000500030@0000D000<100001@000`40000500030@0000D000<100001@000`40000500030@0000D0
@P4700030@0000D000<100001@0Q0@L02P4700811P0A0@L000<100000P020@40000400030@0000<1
1@000`40001H00<11@030@D00`4500<11@030@D00`4500<14@000`40000500030@0000D000@10000
0`4500<11P030@D00`4800030@0000800`4:00<11@060@D00`4@00410@000005004000400P000`40
000500030@0009/000<100001@000`40000500030@0002D000<1000040040@@0104J00410@000080
0P4300@110000`40002H00030@0100D000<100401@000`400@0U00030@0101<01@4300030@010081
6@010@40000?00030@0009P00P4600811P020BH00P4D00D10`000`400@020AT00@4100003`000`40
002G00@110040@@0104T00@14`000`400@0200040@000@8000<1004060010@40000?00030@0009H0
0`4500<11@030BD00`4E00030@0000<000<100401@000`40000E00410@0000l000<10000G`040@@0
104400@110040@D00`4C00811P020@H00P4@00<11@040@@01042008120040@X000<1000010020@H0
00@100000`4@00410@0000l00P4=00030@0000D000<100001@000`40000500030@0000D000<10000
1@000`40000500030@0000D000<100001@000`40000500030@0000D0<P4700030@0000D000<10000
1@000`40000500030@0000D06@4700/11P020@H00P4600T11`000`40000200810@0000l000<10000
H0030@D00`4500<11@030@D00`4A00030@0000D000<100001@000`40000500030@0000D000@10000
0`4600<11@030@P000<100000P030@800`4500<11@030@P00`4@00410@0000l000<10000T`000`40
000500030@0000D000<100001@000`40000500030@0001d000<100002@030@@0104400@16P010@40
000?00030@00090000<100401@000`400@0500030@0100D000<100401@000`400@0M00030@0100`0
10400`010@050@03004100D16@010@40000?00030@0009000P4600811P020@H00P4600817P020@d0
10400`010@050@03004100D16@010@40000?00030@0008l00`4500@110040@@0104400@170040@d0
00D1000100040@8000<100400P401001004H00410@0000l000<10000SP020@H00`4500<11@030@D0
0`4M00<14`030@D00`4500030@0001D00@4100003`000`40001_00@16`000`40000500811P020@H0
0P46008140040@@0104200H110040@T00P4600811P000`40000D00410@0000l00P4=00030@0000D0
00<100001@000`40000500030@0000D000<100001@000`40000500030@0000D000<100001@000`40
000500030@0000D000<100001P000`40000400T11`020@H00P4700030@0000D000<100001@000`40
000500030@0000D000<100001@000`40000502811`020@H000<100001@000`40000500030@0000D0
00<100000P020@40000?00030@0007000`4800030@0000@00P4800030@0000D000<100001@000`40
000500030@0000D000<100001@000`40000500030@0000800`4500<11@030@D00`4:00<15@000`40
000;00410@0000l000<10000NP000`40000300812@000`40000500030@0000D000<100001@000`40
000500030@0000D000<100001@000`40000X00@150000`40000;00410@0000l000<10000M`000`40
0@04008120000`400@0500030@0100D000<100401@000`400@0500030@0100D000<100401@000`40
0@0Y00L14@040@`00@4100003`000`40001g00810P030@X00P4600811P020@H00P4600811P020@H0
0P4Z00L14@040@`00@4100003`000`40001f00D130030@D00`4500@110040@@0104400@110040BX0
104200030@0000l00`4<00410@0000l000<10000M@050@`00P4600811P030@D00`4500<11@030@D0
0`4/00811@000`40000>00813@010@40000?00030@0007D01P4500<10`000`40000500030@010081
0`050@<01@4300D10`020@H01@4500@110040@@0104400@12@020@H000<100002@030@03004100d0
0@4100003`020@d000<100001@000`40000500030@0000D000<100001@000`40000500030@0000D0
00<100001@000`40000500030@0000D000<100001@000`40000600030@0000D000<10000100B0@L0
8@4702X11`020@H000<100001@090@L000<100000P020@40000?00030@0007H01@4400@12@000`40
00040@@0104400@110040@T000<10000104400D10`050@<01@4300D12P030@d000<100400`4@0041
0@0000l000<10000N@020@L000<100001P000`40000400030@0000D000<100001@000`4000050003
0@0000H000<1000010000`40000500811P020@H00P4600812P040@`000<100000P000`40000>0041
0@0000l000<10000R0000`400@0U00030@0103401`4900@150010@40000?00030@0008P00P4V0081
<P070@T0104D00410@0000l000<10000Q`030BD0104b00@10P000`40000700<150010@40000?0003
0@0008H00P4V00<1=0020@D000<100001P020AD00@4100003`000`40002000<10`000`400@020@D0
0`4500<11@030@D00`4300D11@030@D0104400@110040@@0104500<100<00@401P00104000030@03
004100800`4@00410@0000l00P4=00030@0000D000<100001@000`40000500030@0000D000<10000
1@000`40000500030@0000D000<100001@000`40000500030@0000D000<100001P000`4000050003
0@0000D000<100001@1Z0@L04@4700030@0000800P4100003`000`40001o00@110040@@0104400@1
10040@@0104400@110040@@01@4300D10`050@<01@4400@130040@@0104@00410@0000l000<10000
PP000`40000500030@0000D000<100001@000`40000500030@0000D000<100001@000`4000050003
0@0000D00P4600811P020@H00P4700030@0000d000<100001@000`40000>00410@0000l000<10000
o`0=00410@0000l000<10000o`0=00410@0000l000<10000o`0=00410@00000400410@@00P450003
0@000?l03@010@40000400040@000@8000<100000P000`40002800<11@030@D00`4500<11@030@D0
0`4500<11@040@@0104400@110040@D00`4500<11@030@D00`4@00410@0000@000@100010P000`40
000200<130000`40000500030@0000D000<100001@000`40000500030@0000D000<100001@000`40
000500030@0000D000<100001@000`40000500030@0000H000<100001@000`40000500030@0000D0
0P4607X11`000`40000200810@0000800P4300040@000@@000<10000OP040@D0104400@110040@@0
104400@110040@@0104400D10`050@<01@4300D110040@@0104400@110040A000@41000010001040
004200030@00008000<10000O`040@L000<100001@000`40000500030@0000D000<100001@000`40
000500030@0000D000<100001@020@H00P4600811P020@L000<100001@000`40000500030@0000D0
00<100003P010@40000010010@4400811@000`40002000D1Q`010@40000?00030@00084000@10000
0P6500410@0000l000<10000Q`020H<00@4100003`000`4000290081P@010@40000?00030@0008/0
0P4300<11@030@D00`4500<11@030@D00`4500@110040@@0104400@11@030@D00`4500<160010@40
000?00813@000`40000500030@0000D000<100001@000`40000500030@0000D000<100001@000`40
000500030@0000D000<100001@000`40000500030@0000H000<100001@000`40000500030@0000D0
00<100001@000`40000506X11`000`40000500030@0000800P4100003`000`40002?00@110040@@0
104400@110040@@0104400D10`050@<01@4300D110040@@0104400@160010@40000?00030@000980
00<100001@000`40000500030@0000D000<100001@000`40000500030@0000D00P4600811P020@H0
0P4700030@0000D000<100001@000`40000F00410@0000l000<10000o`0=00410@0000l000<10000
o`0=00410@0000l000<10000o`0=00410@0000l000<10000o`0=00410@0000l000<10000o`0=0041
0@0000l00P4=00030@0000D000<100001@000`40000500030@0000D000<100001@000`4000050003
0@0000D000<100001@000`40000500030@0000D000<100001P000`40000500030@0000D000<10000
1@000`40000500030@0000D000<100001@000`40000500030@0000D000<100001@000`4000050003
0@0000D000<100001P000`40000500030@0000D000<100001@000`40000500030@0000D000<10000
1@000`40000500030@0000D000<100000P020@40000?00030@000?l03@010@40000?00030@000?l0
3@010@40000?00030@000?l03@010@40000?00030@000?l03@010@40000?00030@0000<000<10000
9P000`40000U00030@0002H000<100009@000`40000U00030@0002H000<100004`010@40000?0003
0@0000<000<100001@000`40000600030@0000D000<100001@000`40000500030@0000D000<10000
1@000`40000500030@0000D000<100001@000`40000500030@0000D000<100001@000`4000050003
0@0000H000<100001@000`40000500030@0000D000<100001@000`40000500030@0000D000<10000
1@000`40000500030@0000D000<100001@000`40000500030@0000H000<100001@000`4000050003
0@0000D000<100001@000`40000500030@0000<00@4100003`3o0A410@000?l08@000001\
\>"],
  ImageRangeCache->{{{108, 395}, {408.25, 121.25}} -> {-16.0722, 12.4792, 
  0.123767, 0.123767}}]
}, Open  ]],

Cell["\<\
We can easily change the size of the search window in which we \
search for the best match.  However, the computation for finding the best \
match scales with the square of the size of the search window so we have to \
be a little careful.  Changing from 2 to 3 is enoough to bog down my Mac.  If \
you're on a faster machine running Mathematica (as opposed to just reading \
this using MathReader) then you can uncomment the following invocation and \
experiment with larger search windows.\
\>", "Text"],

Cell["\<\
(* Show[Graphics[OpticalFlow[3,myImageSize],
              {{AspectRatio -> 1}, {Frame -> True}}]] ; *)\
\>", "Input"],

Cell["\<\
Question: What are some of the advantages and disadvantages of \
increasing the size of search window.  What if you were to take more images \
per second?  How would this effect your ability to detect faster or slower \
movements?\
\>", "Text"],

Cell["\<\
If you're interested in the tradeoffs involved in changing the \
search window and varying the rate at which images are taken, you might want \
to check out Ted Camus' thesis which offers an elegant solution to making \
such tradeoffs and also provides some interesting insights into the problem \
of detecting motion both in biological and artificial organisms.\
\>", "Text"]
}, Open  ]]
},
FrontEndVersion->"X 3.0",
ScreenRectangle->{{0, 1280}, {0, 1024}},
WindowSize->{886, 647},
WindowMargins->{{93, Automatic}, {Automatic, 36}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"Magnification"->1},
Magnification->1.25,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1709, 49, 54, 0, 73, "Title"],

Cell[CellGroupData[{
Cell[1788, 53, 54, 0, 71, "Section"],
Cell[1845, 55, 668, 10, 122, "Text"],
Cell[2516, 67, 1337, 19, 232, "Text"],
Cell[3856, 88, 358, 9, 57, "Text"],
Cell[4217, 99, 1086, 17, 190, "Text"],
Cell[5306, 118, 63, 0, 40, "Input"],
Cell[5372, 120, 279, 5, 43, "Text"],
Cell[5654, 127, 225, 4, 31, "Text"],
Cell[5882, 133, 41, 0, 40, "Input"],
Cell[5926, 135, 40, 0, 40, "Input"],
Cell[5969, 137, 52, 0, 40, "Input"],
Cell[6024, 139, 52, 0, 40, "Input"],
Cell[6079, 141, 440, 10, 105, "Input"],
Cell[6522, 153, 169, 4, 31, "Text"],
Cell[6694, 159, 98, 3, 40, "Input"],
Cell[6795, 164, 253, 5, 31, "Text"],
Cell[7051, 171, 404, 10, 105, "Input"],
Cell[7458, 183, 169, 4, 31, "Text"],
Cell[7630, 189, 105, 4, 40, "Input"],
Cell[7738, 195, 94, 3, 20, "Text"],
Cell[7835, 200, 403, 10, 105, "Input"],
Cell[8241, 212, 95, 3, 40, "Input"],
Cell[8339, 217, 285, 5, 43, "Text"],

Cell[CellGroupData[{
Cell[8649, 226, 219, 5, 62, "Input"],
Cell[8871, 233, 74189, 1384, 296, 7726, 559, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],

Cell[CellGroupData[{
Cell[83097, 1622, 219, 5, 62, "Input"],
Cell[83319, 1629, 75599, 1402, 296, 7726, 559, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[158933, 3034, 169, 4, 31, "Text"],

Cell[CellGroupData[{
Cell[159127, 3042, 293, 7, 79, "Input"],
Cell[159423, 3051, 59518, 1203, 296, 7726, 559, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[218956, 4257, 327, 6, 43, "Text"],
Cell[219286, 4265, 466, 13, 131, "Input"],
Cell[219755, 4280, 167, 4, 31, "Text"],

Cell[CellGroupData[{
Cell[219947, 4288, 201, 4, 53, "Input"],
Cell[220151, 4294, 41566, 960, 296, 7494, 535, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[261732, 5257, 169, 4, 31, "Text"],
Cell[261904, 5263, 288, 8, 88, "Input"],
Cell[262195, 5273, 202, 4, 31, "Text"],
Cell[262400, 5279, 34, 0, 40, "Input"],
Cell[262437, 5281, 189, 6, 71, "Input"],

Cell[CellGroupData[{
Cell[262651, 5291, 122, 3, 45, "Input"],
Cell[262776, 5296, 153271, 7126, 296, 118533, 6693, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[416062, 12425, 515, 8, 67, "Text"],
Cell[416580, 12435, 128, 3, 45, "Input"],
Cell[416711, 12440, 254, 5, 31, "Text"],
Cell[416968, 12447, 386, 6, 43, "Text"]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

