(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 3.0,
MathReader 3.0, or any compatible application. The data for the notebook 
starts with the line of stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    321487,       9814]*)
(*NotebookOutlinePosition[    322348,       9844]*)
(*  CellTagsIndexPosition[    322304,       9840]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Supplementary Notes on Probability and Statistics", "Title"],

Cell[CellGroupData[{

Cell["Observations on Experiments, Outcomes and Events", "Section"],

Cell[TextData[{
  "Probability is all about events and outcomes.  An outcome is the result of \
an experiment, ",
  StyleBox["e.g.,",
    FontSlant->"Italic"],
  " I flipped a coin (the experiment) and I observed that it landed heads up \
(the outcome)."
}], "Text"],

Cell[TextData[{
  "An event is something that could happen, ",
  StyleBox["e.g., ",
    FontSlant->"Italic"],
  "the event of the next coin flip coming up heads (a pretty common event) or \
the event corresponding to the next sixteen coin flips all coming up heads (a \
fairly rare event)."
}], "Text"],

Cell["\<\
We assign probabilities to events.  Before we say more about \
probabilities, let's consider a mythical but commonly spoken of object?\
\>", 
  "Text"],

Cell[TextData[{
  "Question: What is a fair coin, ",
  StyleBox["i.e.,",
    FontSlant->"Italic"],
  " the sort of thing that referees use to determine who gets the ball first \
in a sporting event?"
}], "Text"],

Cell["\<\
Answer: It is a coin such that the event of flipping the coin and \
having it come up heads will occur (in the long run) equally often as the \
event of flipping the coin and having it come up tails.  Another way to \
characterize a fair coin is the following: if you were to run a large number \
of experiments in which you flipped a coin n times, on average (the average \
is over all of your experiments of flipping a coin n times) the number of \
observed heads would be 1/2.  To be mucn more precise than this we would have \
to use the mathematical notion of a limit as in, if we repeatedly flip a \
coin, the ratio heads to tails will approach one as the number of coin flips \
becomes arbitrarily large (the limit tends to infinity).\
\>", "Text"],

Cell["\<\
Experiment: Consider the ratio of observed heads to the number of \
trials as the number of trials gets large.\
\>", "Text"],

Cell[TextData[{
  "Generate a list of random bits of length",
  StyleBox[" n",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell["\<\
Experiment[n_Integer] := 
    Table[Random[Integer,1],{n}] \
\>", "Input"],

Cell[TextData[{
  StyleBox["Random[Integer,1]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is our equivalent of flipping a fair coin. If ",
  StyleBox["Random[Integer,1]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " == 1 then we observe heads, otherwise tails."
}], "Text"],

Cell["\<\
RunningFrequencies[outcomes_List] := 
    Module[{numHeads = 0, frequencyList = {}},
        Do[numHeads = numHeads + outcomes[[i]] ;
           AppendTo[frequencyList, numHeads/i],
           {i, 1, Length[outcomes]}] ;
        Return[frequencyList]]\
\>", "Input"],

Cell[CellGroupData[{

Cell["\<\
ListPlot[RunningFrequencies[Experiment[1000]], 
         {PlotJoined -> True,
          PlotLabel -> \"Number of heads / Number of coin flips\"}] ;\
\>", 
  "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.000952381 -3.55795 7.57202 [
[.21429 .21556 -9 -9 ]
[.21429 .21556 9 0 ]
[.40476 .21556 -9 -9 ]
[.40476 .21556 9 0 ]
[.59524 .21556 -9 -9 ]
[.59524 .21556 9 0 ]
[.78571 .21556 -9 -9 ]
[.78571 .21556 9 0 ]
[.97619 .21556 -12 -9 ]
[.97619 .21556 12 0 ]
[.01131 .07662 -24 -4.5 ]
[.01131 .07662 0 4.5 ]
[.01131 .3795 -24 -4.5 ]
[.01131 .3795 0 4.5 ]
[.01131 .53094 -24 -4.5 ]
[.01131 .53094 0 4.5 ]
[.5 .63053 -116 0 ]
[.5 .63053 116 10.125 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.21429 .22806 m
.21429 .23431 L
s
[(200)] .21429 .21556 0 1 Mshowa
.40476 .22806 m
.40476 .23431 L
s
[(400)] .40476 .21556 0 1 Mshowa
.59524 .22806 m
.59524 .23431 L
s
[(600)] .59524 .21556 0 1 Mshowa
.78571 .22806 m
.78571 .23431 L
s
[(800)] .78571 .21556 0 1 Mshowa
.97619 .22806 m
.97619 .23431 L
s
[(1000)] .97619 .21556 0 1 Mshowa
.125 Mabswid
.07143 .22806 m
.07143 .23181 L
s
.11905 .22806 m
.11905 .23181 L
s
.16667 .22806 m
.16667 .23181 L
s
.2619 .22806 m
.2619 .23181 L
s
.30952 .22806 m
.30952 .23181 L
s
.35714 .22806 m
.35714 .23181 L
s
.45238 .22806 m
.45238 .23181 L
s
.5 .22806 m
.5 .23181 L
s
.54762 .22806 m
.54762 .23181 L
s
.64286 .22806 m
.64286 .23181 L
s
.69048 .22806 m
.69048 .23181 L
s
.7381 .22806 m
.7381 .23181 L
s
.83333 .22806 m
.83333 .23181 L
s
.88095 .22806 m
.88095 .23181 L
s
.92857 .22806 m
.92857 .23181 L
s
.25 Mabswid
0 .22806 m
1 .22806 L
s
.02381 .07662 m
.03006 .07662 L
s
[(0.48)] .01131 .07662 1 0 Mshowa
.02381 .3795 m
.03006 .3795 L
s
[(0.52)] .01131 .3795 1 0 Mshowa
.02381 .53094 m
.03006 .53094 L
s
[(0.54)] .01131 .53094 1 0 Mshowa
.125 Mabswid
.02381 .11448 m
.02756 .11448 L
s
.02381 .15234 m
.02756 .15234 L
s
.02381 .1902 m
.02756 .1902 L
s
.02381 .26592 m
.02756 .26592 L
s
.02381 .30378 m
.02756 .30378 L
s
.02381 .34164 m
.02756 .34164 L
s
.02381 .41736 m
.02756 .41736 L
s
.02381 .45522 m
.02756 .45522 L
s
.02381 .49308 m
.02756 .49308 L
s
.02381 .03876 m
.02756 .03876 L
s
.02381 .0009 m
.02756 .0009 L
s
.02381 .5688 m
.02756 .5688 L
s
.02381 .60666 m
.02756 .60666 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
gsave
.5 .63053 -177 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.125000 moveto
 558.000000 14.125000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.125000 lineto
closepath
clip newpath
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Number) show
105.000000 11.562500 moveto
(of) show
123.000000 11.562500 moveto
(heads) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
159.000000 11.562500 moveto
(\\220) show
171.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Number) show
213.000000 11.562500 moveto
(of) show
231.000000 11.562500 moveto
(coin) show
261.000000 11.562500 moveto
(flips) show
291.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Math2Mono
%%+ Courier
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02566 0 m
.02571 .22806 L
s
.02571 .22806 m
.02601 .61803 L
s
.03074 .61803 m
.03143 .22806 L
s
.03143 .22806 m
.03231 .61803 L
s
.03245 .61803 m
.03333 .22806 L
s
.03333 .22806 m
.03396 0 L
s
.03461 0 m
.03524 .22806 L
s
.03524 .22806 m
.03598 0 L
s
.06693 0 m
.06762 .06345 L
.06857 .14751 L
.06952 .07031 L
.07048 .1508 L
.07143 .07662 L
.07238 .00535 L
.07333 .08245 L
.07429 .15663 L
.07524 .22806 L
.07619 .15922 L
.07714 .22806 L
.0781 .16164 L
.07905 .09751 L
.08 .16389 L
.08095 .22806 L
.0819 .29013 L
.08286 .22806 L
.08381 .16797 L
.08476 .10975 L
.08571 .05332 L
.08667 .11333 L
.08762 .05854 L
.08857 .00535 L
.08952 .06345 L
.09048 .01172 L
.09143 .06809 L
.09238 .12289 L
.09333 .1762 L
.09429 .12574 L
.09524 .17758 L
.09619 .22806 L
.09714 .17889 L
.0981 .22806 L
.09905 .27599 L
.1 .32271 L
.10095 .2748 L
.1019 .3204 L
.10286 .27368 L
.10381 .3182 L
.10476 .36169 L
.10571 .40415 L
.10667 .35861 L
.10762 .40015 L
.10857 .44076 L
.10952 .48046 L
.11048 .43608 L
.11143 .47498 L
.11238 .43161 L
.11333 .46972 L
Mistroke
.11429 .50703 L
.11524 .46469 L
.11619 .50128 L
.11714 .45986 L
.1181 .49576 L
.11905 .45522 L
.12 .49046 L
.12095 .525 L
.1219 .55888 L
.12286 .51929 L
.12381 .48046 L
.12476 .5138 L
.12571 .54651 L
.12667 .50851 L
.12762 .54067 L
.12857 .50341 L
.12952 .53504 L
.13048 .5661 L
.13143 .5296 L
.13238 .49375 L
.13333 .52436 L
.13429 .55444 L
.13524 .51929 L
.13619 .54891 L
.13714 .57803 L
.1381 .54356 L
.13905 .57224 L
.14 .53839 L
.14095 .56665 L
.1419 .59445 L
Mfstroke
.1419 .59445 m
.14273 .61803 L
s
.14297 .61803 m
.14381 .58863 L
.14476 .55598 L
.14571 .583 L
.14667 .6096 L
s
.14667 .6096 m
.14697 .61803 L
s
.16369 .61803 m
.16381 .61439 L
.16476 .5862 L
.16571 .55838 L
.16667 .53094 L
.16762 .55401 L
.16857 .52696 L
.16952 .50026 L
.17048 .47391 L
.17143 .49675 L
.17238 .47075 L
.17333 .44509 L
.17429 .46768 L
.17524 .48999 L
.17619 .46469 L
.17714 .4397 L
.1781 .46177 L
.17905 .4371 L
.18 .41274 L
.18095 .38868 L
.1819 .36491 L
.18286 .38676 L
.18381 .40835 L
.18476 .42968 L
.18571 .45077 L
.18667 .47161 L
.18762 .4922 L
.18857 .51256 L
.18952 .53268 L
.19048 .50931 L
.19143 .4862 L
.19238 .46335 L
.19333 .44076 L
.19429 .46072 L
.19524 .4384 L
.19619 .45815 L
.19714 .43608 L
.1981 .45564 L
.19905 .47498 L
.2 .49411 L
.20095 .47232 L
.2019 .45077 L
.20286 .42944 L
.20381 .44841 L
.20476 .42733 L
.20571 .4461 L
.20667 .46469 L
.20762 .44384 L
.20857 .42322 L
.20952 .4028 L
Mistroke
.21048 .38259 L
.21143 .36259 L
.21238 .38103 L
.21333 .36124 L
.21429 .34164 L
.21524 .35991 L
.21619 .34052 L
.21714 .35861 L
.2181 .33941 L
.21905 .35734 L
.22 .33833 L
.22095 .35609 L
.2219 .33727 L
.22286 .31864 L
.22381 .33623 L
.22476 .35366 L
.22571 .33521 L
.22667 .31693 L
.22762 .29883 L
.22857 .28089 L
.22952 .26312 L
.23048 .2804 L
.23143 .29753 L
.23238 .27992 L
.23333 .26248 L
.23429 .27946 L
.23524 .26217 L
.23619 .24504 L
.23714 .26187 L
.2381 .27854 L
.23905 .29507 L
.24 .31145 L
.24095 .29448 L
.2419 .27766 L
.24286 .26098 L
.24381 .24445 L
.24476 .22806 L
.24571 .24431 L
.24667 .26042 L
.24762 .27639 L
.24857 .26015 L
.24952 .24404 L
.25048 .22806 L
.25143 .21222 L
.25238 .19651 L
.25333 .21235 L
.25429 .19677 L
.25524 .18132 L
.25619 .19703 L
.25714 .21261 L
Mistroke
.2581 .22806 L
.25905 .24339 L
.26 .25859 L
.26095 .27368 L
.2619 .28864 L
.26286 .30348 L
.26381 .28816 L
.26476 .30288 L
.26571 .28768 L
.26667 .3023 L
.26762 .3168 L
.26857 .33118 L
.26952 .31611 L
.27048 .33039 L
.27143 .34455 L
.27238 .3296 L
.27333 .34366 L
.27429 .32883 L
.27524 .34279 L
.27619 .35664 L
.27714 .34193 L
.2781 .35568 L
.27905 .36933 L
.28 .38288 L
.28095 .39633 L
.2819 .40968 L
.28286 .42293 L
.28381 .43608 L
.28476 .44914 L
.28571 .46211 L
.28667 .47498 L
.28762 .48775 L
.28857 .4732 L
.28952 .48589 L
.29048 .49849 L
.29143 .48406 L
.29238 .49657 L
.29333 .48225 L
.29429 .49468 L
.29524 .48046 L
.29619 .46634 L
.29714 .45232 L
.2981 .46469 L
.29905 .45077 L
.3 .43694 L
.30095 .44924 L
.3019 .46145 L
.30286 .47357 L
.30381 .45986 L
.30476 .47191 L
Mistroke
.30571 .45829 L
.30667 .44477 L
.30762 .45675 L
.30857 .46864 L
.30952 .45522 L
.31048 .44189 L
.31143 .45372 L
.31238 .44048 L
.31333 .45223 L
.31429 .43908 L
.31524 .45077 L
.31619 .46237 L
.31714 .47391 L
.3181 .46086 L
.31905 .44789 L
.32 .45936 L
.32095 .44649 L
.3219 .43369 L
.32286 .42098 L
.32381 .40835 L
.32476 .41976 L
.32571 .40721 L
.32667 .39474 L
.32762 .38235 L
.32857 .37004 L
.32952 .3578 L
.33048 .34564 L
.33143 .33355 L
.33238 .32154 L
.33333 .3329 L
.33429 .32097 L
.33524 .30911 L
.33619 .29732 L
.33714 .2856 L
.3381 .27395 L
.33905 .28525 L
.34 .29648 L
.34095 .30765 L
.3419 .29607 L
.34286 .28457 L
.34381 .27313 L
.34476 .26176 L
.34571 .27287 L
.34667 .26157 L
.34762 .25033 L
.34857 .26137 L
.34952 .27234 L
.35048 .28325 L
.35143 .2941 L
.35238 .28293 L
Mistroke
.35333 .29371 L
.35429 .30444 L
.35524 .29334 L
.35619 .2823 L
.35714 .27133 L
.3581 .28199 L
.35905 .27108 L
.36 .28169 L
.36095 .29223 L
.3619 .30272 L
.36286 .29187 L
.36381 .28109 L
.36476 .29151 L
.36571 .30188 L
.36667 .31219 L
.36762 .32245 L
.36857 .31173 L
.36952 .30107 L
.37048 .31127 L
.37143 .30067 L
.37238 .31082 L
.37333 .30027 L
.37429 .28979 L
.37524 .27936 L
.37619 .26899 L
.37714 .27909 L
.3781 .26877 L
.37905 .25851 L
.38 .26855 L
.38095 .27854 L
.3819 .28848 L
.38286 .27827 L
.38381 .26812 L
.38476 .25803 L
.38571 .26791 L
.38667 .27775 L
.38762 .28753 L
.38857 .29726 L
.38952 .30694 L
.39048 .31657 L
.39143 .30653 L
.39238 .29654 L
.39333 .28661 L
.39429 .27672 L
.39524 .28631 L
.39619 .29584 L
.39714 .28601 L
.3981 .2955 L
.39905 .28572 L
.4 .29516 L
Mistroke
.40095 .28543 L
.4019 .29482 L
.40286 .28514 L
.40381 .29448 L
.40476 .30378 L
.40571 .29415 L
.40667 .28457 L
.40762 .27503 L
.40857 .28429 L
.40952 .2935 L
.41048 .30266 L
.41143 .31178 L
.41238 .32086 L
.41333 .32989 L
.41429 .33887 L
.41524 .34781 L
.41619 .35671 L
.41714 .34723 L
.4181 .3378 L
.41905 .34666 L
.42 .35548 L
.42095 .36425 L
.4219 .37298 L
.42286 .38167 L
.42381 .39032 L
.42476 .38094 L
.42571 .37161 L
.42667 .36232 L
.42762 .35307 L
.42857 .36169 L
.42952 .35248 L
.43048 .36106 L
.43143 .36959 L
.43238 .37809 L
.43333 .38655 L
.43429 .37739 L
.43524 .38581 L
.43619 .39419 L
.43714 .40253 L
.4381 .39343 L
.43905 .38436 L
.44 .39267 L
.44095 .40094 L
.4419 .39192 L
.44286 .38294 L
.44381 .37401 L
.44476 .36511 L
.44571 .37335 L
.44667 .38155 L
.44762 .3727 L
Mistroke
.44857 .36388 L
.44952 .35511 L
.45048 .36328 L
.45143 .35454 L
.45238 .34585 L
.45333 .33719 L
.45429 .32857 L
.45524 .33671 L
.45619 .32813 L
.45714 .31959 L
.4581 .31109 L
.45905 .31919 L
.46 .32726 L
.46095 .31879 L
.4619 .31037 L
.46286 .3184 L
.46381 .31001 L
.46476 .31801 L
.46571 .30966 L
.46667 .31762 L
.46762 .30931 L
.46857 .31724 L
.46952 .30896 L
.47048 .31686 L
.47143 .32473 L
.47238 .31648 L
.47333 .32432 L
.47429 .31611 L
.47524 .32391 L
.47619 .31574 L
.47714 .3076 L
.4781 .31537 L
.47905 .30727 L
.48 .31501 L
.48095 .30694 L
.4819 .2989 L
.48286 .2909 L
.48381 .28293 L
.48476 .29064 L
.48571 .29832 L
.48667 .30596 L
.48762 .29803 L
.48857 .29013 L
.48952 .29774 L
.49048 .30533 L
.49143 .31288 L
.49238 .3204 L
.49333 .31254 L
.49429 .3047 L
.49524 .31219 L
Mistroke
.49619 .30439 L
.49714 .29662 L
.4981 .28888 L
.49905 .28117 L
.5 .27349 L
.50095 .26585 L
.5019 .25823 L
.50286 .25064 L
.50381 .25811 L
.50476 .25055 L
.50571 .25799 L
.50667 .2654 L
.50762 .25787 L
.50857 .25038 L
.50952 .24291 L
.51048 .25029 L
.51143 .24285 L
.51238 .23544 L
.51333 .22806 L
.51429 .22071 L
.51524 .21339 L
.51619 .20609 L
.51714 .21344 L
.5181 .22077 L
.51905 .2135 L
.52 .20626 L
.52095 .19905 L
.5219 .20634 L
.52286 .21361 L
.52381 .22085 L
.52476 .21367 L
.52571 .22088 L
.52667 .22806 L
.52762 .23522 L
.52857 .22806 L
.52952 .23519 L
.53048 .22806 L
.53143 .23516 L
.53238 .24224 L
.53333 .23514 L
.53429 .24219 L
.53524 .23511 L
.53619 .22806 L
.53714 .22104 L
.5381 .21404 L
.53905 .22106 L
.54 .21409 L
.54095 .22109 L
.5419 .21414 L
.54286 .22111 L
Mistroke
.54381 .22806 L
.54476 .22114 L
.54571 .21424 L
.54667 .20737 L
.54762 .20053 L
.54857 .20745 L
.54952 .20063 L
.55048 .20752 L
.55143 .21439 L
.55238 .22124 L
.55333 .21444 L
.55429 .20767 L
.55524 .20092 L
.55619 .20774 L
.55714 .20102 L
.5581 .20782 L
.55905 .20111 L
.56 .20789 L
.56095 .21464 L
.5619 .20796 L
.56286 .21468 L
.56381 .20803 L
.56476 .2014 L
.56571 .19479 L
.56667 .20149 L
.56762 .19491 L
.56857 .20159 L
.56952 .20824 L
.57048 .21487 L
.57143 .20831 L
.57238 .21492 L
.57333 .2215 L
.57429 .22806 L
.57524 .2346 L
.57619 .24112 L
.57714 .24761 L
.5781 .25408 L
.57905 .24754 L
.58 .25399 L
.58095 .24748 L
.5819 .24098 L
.58286 .24741 L
.58381 .25382 L
.58476 .24734 L
.58571 .25373 L
.58667 .24728 L
.58762 .24085 L
.58857 .24721 L
.58952 .24081 L
.59048 .23442 L
Mistroke
.59143 .24077 L
.59238 .24709 L
.59333 .25339 L
.59429 .25966 L
.59524 .2533 L
.59619 .25956 L
.59714 .25322 L
.5981 .2469 L
.59905 .25313 L
.6 .24683 L
.60095 .24056 L
.6019 .2343 L
.60286 .22806 L
.60381 .22184 L
.60476 .22806 L
.60571 .22186 L
.60667 .22806 L
.60762 .22189 L
.60857 .21573 L
.60952 .20959 L
.61048 .21577 L
.61143 .22193 L
.61238 .22806 L
.61333 .22195 L
.61429 .21585 L
.61524 .22196 L
.61619 .21589 L
.61714 .22198 L
.6181 .21593 L
.61905 .222 L
.62 .22806 L
.62095 .2341 L
.6219 .24012 L
.62286 .23408 L
.62381 .22806 L
.62476 .23406 L
.62571 .24004 L
.62667 .246 L
.62762 .25195 L
.62857 .24595 L
.62952 .23997 L
.63048 .234 L
.63143 .22806 L
.63238 .23399 L
.63333 .22806 L
.63429 .23397 L
.63524 .23986 L
.63619 .24573 L
.63714 .23982 L
.6381 .24567 L
Mistroke
.63905 .2515 L
.64 .25732 L
.64095 .26312 L
.6419 .2689 L
.64286 .27466 L
.64381 .26877 L
.64476 .27452 L
.64571 .28024 L
.64667 .27437 L
.64762 .26852 L
.64857 .27423 L
.64952 .2684 L
.65048 .26258 L
.65143 .26828 L
.65238 .26248 L
.65333 .26816 L
.65429 .26238 L
.65524 .26803 L
.65619 .26227 L
.65714 .26791 L
.6581 .27354 L
.65905 .27915 L
.66 .2734 L
.66095 .27899 L
.6619 .28457 L
.66286 .29013 L
.66381 .2844 L
.66476 .27869 L
.66571 .28423 L
.66667 .27854 L
.66762 .27287 L
.66857 .27839 L
.66952 .27273 L
.67048 .27824 L
.67143 .28374 L
.67238 .2781 L
.67333 .27247 L
.67429 .27795 L
.67524 .28341 L
.67619 .28886 L
.67714 .29429 L
.6781 .2997 L
.67905 .2941 L
.68 .28851 L
.68095 .29391 L
.6819 .28833 L
.68286 .29371 L
.68381 .28816 L
.68476 .28261 L
.68571 .27709 L
Mistroke
.68667 .27158 L
.68762 .26608 L
.68857 .26061 L
.68952 .26598 L
.69048 .26051 L
.69143 .26587 L
.69238 .27121 L
.69333 .26576 L
.69429 .27108 L
.69524 .26565 L
.69619 .26024 L
.69714 .25484 L
.6981 .24945 L
.69905 .24408 L
.7 .23873 L
.70095 .24404 L
.7019 .2387 L
.70286 .24399 L
.70381 .24927 L
.70476 .24395 L
.70571 .24921 L
.70667 .25446 L
.70762 .2597 L
.70857 .26492 L
.70952 .27013 L
.71048 .26482 L
.71143 .25952 L
.71238 .26472 L
.71333 .25944 L
.71429 .26462 L
.71524 .26978 L
.71619 .26452 L
.71714 .26967 L
.7181 .2748 L
.71905 .26955 L
.72 .26432 L
.72095 .26944 L
.7219 .27455 L
.72286 .27964 L
.72381 .27442 L
.72476 .2795 L
.72571 .28457 L
.72667 .27936 L
.72762 .27417 L
.72857 .26899 L
.72952 .27405 L
.73048 .27909 L
.73143 .28411 L
.73238 .28913 L
.73333 .29413 L
Mistroke
.73429 .29911 L
.73524 .29395 L
.73619 .29892 L
.73714 .30388 L
.7381 .30883 L
.73905 .30368 L
.74 .30861 L
.74095 .31354 L
.7419 .31844 L
.74286 .32334 L
.74381 .3182 L
.74476 .32309 L
.74571 .31797 L
.74667 .32284 L
.74762 .31773 L
.74857 .31264 L
.74952 .30756 L
.75048 .31242 L
.75143 .30735 L
.75238 .3023 L
.75333 .29726 L
.75429 .3021 L
.75524 .30694 L
.75619 .30191 L
.75714 .30673 L
.7581 .30172 L
.75905 .29672 L
.76 .29173 L
.76095 .29654 L
.7619 .30134 L
.76286 .29637 L
.76381 .30115 L
.76476 .30592 L
.76571 .30096 L
.76667 .29602 L
.76762 .29108 L
.76857 .28616 L
.76952 .28125 L
.77048 .28601 L
.77143 .29076 L
.77238 .28586 L
.77333 .2906 L
.77429 .29533 L
.77524 .29044 L
.77619 .29516 L
.77714 .29986 L
.7781 .29499 L
.77905 .29968 L
.78 .29482 L
.78095 .2995 L
Mistroke
.7819 .30416 L
.78286 .30882 L
.78381 .30397 L
.78476 .30861 L
.78571 .31325 L
.78667 .30841 L
.78762 .31303 L
.78857 .30821 L
.78952 .31282 L
.79048 .30801 L
.79143 .31261 L
.79238 .3172 L
.79333 .32177 L
.79429 .31698 L
.79524 .31219 L
.79619 .30742 L
.79714 .31199 L
.7981 .30723 L
.79905 .31178 L
.8 .31632 L
.80095 .32086 L
.8019 .31611 L
.80286 .31137 L
.80381 .30665 L
.80476 .31117 L
.80571 .30646 L
.80667 .31097 L
.80762 .30627 L
.80857 .31077 L
.80952 .30608 L
.81048 .3014 L
.81143 .29673 L
.81238 .29208 L
.81333 .28743 L
.81429 .2828 L
.81524 .27818 L
.81619 .28267 L
.81714 .27806 L
.8181 .27346 L
.81905 .27794 L
.82 .28241 L
.82095 .27782 L
.8219 .27324 L
.82286 .26867 L
.82381 .26412 L
.82476 .25957 L
.82571 .25504 L
.82667 .25052 L
.82762 .25498 L
.82857 .25046 L
Mistroke
.82952 .25491 L
.83048 .25041 L
.83143 .24592 L
.83238 .25036 L
.83333 .25479 L
.83429 .25031 L
.83524 .25472 L
.83619 .25025 L
.83714 .24579 L
.8381 .2502 L
.83905 .24575 L
.84 .25015 L
.84095 .25454 L
.8419 .25891 L
.84286 .26328 L
.84381 .26764 L
.84476 .2632 L
.84571 .26754 L
.84667 .27188 L
.84762 .26745 L
.84857 .26304 L
.84952 .26736 L
.85048 .26296 L
.85143 .26727 L
.85238 .26288 L
.85333 .25849 L
.85429 .2628 L
.85524 .26709 L
.85619 .26272 L
.85714 .267 L
.8581 .27128 L
.85905 .27555 L
.86 .27118 L
.86095 .27544 L
.8619 .27969 L
.86286 .28393 L
.86381 .28816 L
.86476 .29238 L
.86571 .29659 L
.86667 .30079 L
.86762 .30498 L
.86857 .30916 L
.86952 .31333 L
.87048 .31749 L
.87143 .32165 L
.87238 .31729 L
.87333 .32144 L
.87429 .32557 L
.87524 .3297 L
.87619 .33382 L
Mistroke
.87714 .33792 L
.8781 .33358 L
.87905 .32925 L
.88 .32492 L
.88095 .32902 L
.8819 .33311 L
.88286 .3288 L
.88381 .32449 L
.88476 .32857 L
.88571 .32428 L
.88667 .32 L
.88762 .32407 L
.88857 .31979 L
.88952 .32386 L
.89048 .31959 L
.89143 .31533 L
.89238 .31109 L
.89333 .31514 L
.89429 .31091 L
.89524 .30668 L
.89619 .30246 L
.89714 .30651 L
.8981 .3023 L
.89905 .30634 L
.9 .30214 L
.90095 .30617 L
.9019 .30197 L
.90286 .306 L
.90381 .30181 L
.90476 .30583 L
.90571 .30166 L
.90667 .30566 L
.90762 .3015 L
.90857 .30549 L
.90952 .30134 L
.91048 .29719 L
.91143 .29306 L
.91238 .28893 L
.91333 .29292 L
.91429 .2969 L
.91524 .30087 L
.91619 .29675 L
.91714 .30071 L
.9181 .2966 L
.91905 .30056 L
.92 .29646 L
.92095 .30041 L
.9219 .30434 L
.92286 .30025 L
.92381 .29617 L
Mistroke
.92476 .2921 L
.92571 .29603 L
.92667 .29995 L
.92762 .29588 L
.92857 .2998 L
.92952 .3037 L
.93048 .3076 L
.93143 .31149 L
.93238 .30743 L
.93333 .30339 L
.93429 .30727 L
.93524 .30323 L
.93619 .2992 L
.93714 .29518 L
.9381 .29116 L
.93905 .28716 L
.94 .29103 L
.94095 .2949 L
.9419 .2909 L
.94286 .28691 L
.94381 .29077 L
.94476 .28679 L
.94571 .29064 L
.94667 .28667 L
.94762 .29051 L
.94857 .29435 L
.94952 .29038 L
.95048 .28643 L
.95143 .28248 L
.95238 .27854 L
.95333 .28237 L
.95429 .27844 L
.95524 .27452 L
.95619 .27834 L
.95714 .27442 L
.9581 .27823 L
.95905 .28204 L
.96 .28583 L
.96095 .28193 L
.9619 .27803 L
.96286 .27414 L
.96381 .27026 L
.96476 .26638 L
.96571 .26251 L
.96667 .25866 L
.96762 .2548 L
.96857 .25859 L
.96952 .25475 L
.97048 .25091 L
.97143 .2547 L
Mistroke
.97238 .25847 L
.97333 .25464 L
.97429 .25841 L
.97524 .26217 L
.97619 .26592 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40005X0000gPP10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdk34g_ojn_l004J2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_P00oiT0IP3o0;C^]:JV
YZf][Jo^kP20P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;002k0?mT0]d009T0V@00
gNk^k^[Zj]gMgK^k^jZZZXR8R7MgMeEEEDA4A14A4Oooc?o<VLbIIS?oVC>IIS=VIYWoo`2Ioc<0VLc<
oiUVVNoGk`00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2o
o`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no
?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo
?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo
?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0
_ol0onj2k_Z0L_lo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoO`0l0>o<knl1k`1]
0:2PX:[Xo`3XooooP;Co]4Bkooo;og?X`?nXOOnCOOno_onooooo?oooOooo_mWEc:0Pl=WIfJj^[YWo
09Wo<iWoIYWoVIWoc<`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aV
VLaVc<aVolbI0<bI<lbIVLbIc<bIolc<0<c<<lc<IP000?l0J@000?l0J@0002800P400`010@090003
0@000?l0=P00028000<100000P4900030@000?l0=P00028000<100000P4900811@000`40003o02l0
000R00030@0000812@020@D000<10000o`0_00008P000`4000020@T00P4500030@000?l0;`000280
00<100000P4900811@000`40003o02l0000R00030@0000812@020@D000<10000o`0_00008P000`40
00020@T00P4500030@000?l0;`00028000<100000P4900811@000`40003o02l0000R00030@000081
2@020@D000<10000o`0_00008P000`4000020@T00P4500030@000?l0;`00028000<100000P490081
10020Ol0<@00028000<100000P49008110020Ol0<@0002800P400`010@09008110030Ol0<0000280
00<100000P49008110030Ol0<000028000<100000P4900<10`030Ol0<000028000<100000P4900<1
0`030Ol0<000028000<100000P4900<10`030Ol0<000028000<100000P4900<10`030Ol0<0000280
00<100000P4900<10`030Ol0<000028000<100000P4900<10`030Ol0<000028000<100000P4900<1
0`030Ol0<000028000<100000P4900<10`000`400@3o0300000200<11P000`40000500@10`030@D0
00<100000P4900<10`000`400@3o0300000010010@020@l000D100000@0300030@00008000<10000
0P4900<10`000`400@3o030000000`01000200030@0000T01P4200050@00004010050@T00`430003
0@010?l0<00000030040008000<100002@001@40000100<000D100000@0400030@0000812@030@<0
00<10040o`0`000000<00@000P000`40000:00040@000@@00`4500030@0000812@030@800P400`01
003o02l000000`01000200030@0000X00P400`01000200050@00004010000`4000020@T00`420081
00<00@00o`0_000000<00@000P000`40000;00030@0100<000D100000@0400030@0000812@030@80
0P400`01003o02l0000010010@020@h00P4300050@00004010000`4000020@T00`42008100<00@00
o`0_00000P030@l00P4400<11@000`4000020@T00`42008100<00@00o`0_00008P000`4000020@T0
0`401P010@400Ol0<000028000<100000P4900<100H00@41007o0300000R00030@0000812@030@06
00410@01o`0`00008P000`4000020@T00`401P010@400Ol0<000028000<100000P4900<100H00@41
007o0300000R008100<00@402@030@0600410@01o`0`00008P001040004:00<100<00@400P000`40
003o02h0000R00040@000@X00`400`010@0200030@000?l0;P00028000@100012P030@0300410080
00<10000o`0^00008P001040004:00<100<00@400P020Ol0;`00028000@100012P030@0300410080
0P7o02l0000R00040@000@X01P420081o`0_00008P001040004:00H10P020Ol0;`00028000@10001
2P060@800P7o02l0000R00040@000@X01P420081o`0_00008P001040004:00H10P020Ol0;`000280
00@100012P060@800P7o02l0000R008100<00@002P050@800P7o02l0000R00040@000@`010420081
o`0_00008P001040004<00@10P020Ol0;`00028000@1000130040@800P7o02l0000R00040@000@`0
10420081o`0_00008P001040004<00@10P020Ol0;`00028000@1000130040@800P7o02l0000R0004
0@000@`010420081o`0_00008P001040004<00@10P020Ol0;`00028000@1000130040@<000<10000
o`0]00008P001040004<00@10`000`40000a00030@000?P0000R00040@000@`0104300030@000340
00<10000n000028000@1000130040@<000<10000<@000`40003h00008P020@03004000/010430003
0@00034000<10000n000028000@1000130040@<000<100006@050@@00`4500<110000`40000U00@1
0`030@D00`4[00@110030@D00`4[00<11@030@D00`4W00D110030@D00`4500<10P00028000@10001
30040@<000<100006@001@40000100<00P400`010@03008100<00@400P020BT000@100000P400`01
0@03008100<00@40:P020@8000<100000P400`010@03008100<00@40:@001@40000100<00P400`01
0@03008100<00@40:0000`400003008100<00@400`020@03004100<00P400`010@0100008P001040
004<00@10`000`40000J00030@0000@000D100000@0300050@0000400P020BD01P4200050@000040
0`001@40000101/000<100001P020@04004100810P001040004300050@0000400`000`40000W0005
0@0000400`001@40000100<000D100000@0X00030@0000<000D100000@0300050@0000400`001@40
00010040000R00040@000@`01043008170000`40000300050@0000400`001@40000100800P4U0005
0@0000400`001@40000100<000D100000@0K00811`020@04004100810P001040004300050@000040
0`000`40000W00050@0000400`001@40000100<000D100000@0X00030@0000<000D100000@030005
0@0000400`001@4000010040000R00040@000@`0104300817@000`40000200050@0000400`001@40
0001008000<100409@001040004300050@0000400`001@40000101X00`4700P10P00104000430005
0@0000400`000`40000X00<110001@40000100<000D100000@0X00030@0000<000D100000@030005
0@0000400`001@4000010040000R00040@000@`0104300817P001@40000100<000D100000@030006
0@000@019@020@030040008000D100000@0300050@0000406P030@L02P4300050@0000400P020@<0
00<100009`001@40000100<000D100000@0300050@000040:0000`40000300050@0000400`001@40
000100<000D100000@0100008P001040004<00@10`020AX000D100000@0300050@0000400`001@40
0001008000<100409P000`400@0300050@0000400`001@40000101T01@4500050@010@001@440005
0@0000400P020@<000<100009`001@40000100<000D100000@0300050@000040:0000`4000030005
0@0000400`001@40000100<000D100000@0100008P001040004<00@10`020AX000<100000P430081
00<00@400`020@030041008000<100409`020@<00P400`010@03008100<00@406@050@<00`401001
0@020@03004100@00P400`010@0200H1:@001@40000100<00P400`010@03008100<00@409P000`40
0@05008100<00@400`020@03004100<00P400`010@0100008P001040004<00@10`020A/00`4500<1
1@030@<000<100409`020@@00`4500<16P020@03004100<00`4200030@0000<00`4300<10`050B/0
0`4500<11@030BP00P4600<11@030@D00`4200008P001040004<00@10`020C0000@10001DP020@03
004100<00`4200030@0000<000<100001`060H@0000R00040@000@`010430081<0001040005B0005
0@0000400`000`400@0900030@0000H0104200030@000840000J0?l1A`4800008P0010400@4;0003
0@00008110000`40000400030@0000`000<100003@000`40000600030@0100800P4?00030@0000`0
00<100003@000`40000<00030@0000d000D100000@0400<10`000`40000600030@0000@000<10040
1@020@030041008000<100003@000`40000<00030@0000d000<1000030000`40000=00030@0000`0
00<100003@000`40000=00008P0010400@4>008110000`40000S00030@0000H000<100400P000`40
000]00030@00020000<100000P030@`000<100000P001@400@0100D00P400`010@0a00030@0003/0
00<100003@00028000@100413P020@@000<10000;0030@8000<10000D0000`400003008130000`40
000200050@0000401@020@0400400Gh0000R00040@010@h00P4400030@0002`00`4200030@000500
00<100000`020@`000@100400P4200030@0000@00P401001004D00030@0006L0000R00040@010@h0
0P4400030@0002`00`4200030@0004l000<1000010020@d01P400`010007008100<00@004P030FP0
000R00040@010@h00P4400030@0002X000<100400P4200030@0004l000<100004`040@0400410@P0
00@100014`000`400@0X00040@010C`0000R00040@010@h00P4400030@0002X000<100400P420003
0@0001X000<10000;`040AD0104010010@4800050@0000404P000`400@0W00D1?000028000@10041
3P020@@000<10000:P000`400@020@8000<100006P000`40000_00<15P030@800`4800050@000040
4P001040004V00D1:P000`40000?00008P0010400@4>008110000`40000Y008100@00@410P000`40
000J00030@0002l00`4F00810`020@d000<1000040001040004V00D1:@030A00000R00040@010@h0
0P4400030@0002T00P4010010@4200030@0001X000<100001`000`400@0U00<16`020@d000<10000
40000`4000030B<000<100000`001@40000102<0104@00008P0010400@4>00030@0000<000<10000
:@020@0400410@8000<100006@020@T000<100409@000`400@0Z00040@0000812P030@<00`4S0003
0@0000<000<100400P4S00040@000A00000R00@13P000`40000300030@0002P00`4010010@420003
0@0001T00P4900030@0102D000<10040:P00104000030@T00`4300D18@000`40000300H18P001@40
000100l0000R00040@010@h000<100000`000`40000X00<100@00@410`000`40000H00<11`020@03
004002@000<10000:`000`400@030@T00`4400040@010@8000<1000070000`40000300H18P000`40
000A00008P0010400@4>00030@0000<000<10000:0030@03004100@000<10000600010400@020@@0
0P400`01000T00030@0002/00`4200030@0000H000@10041100010400@4200030@0001`000<10000
1000104000020B0000<100004P00028000@100413P000`4000030081:0040@03004100@000<10000
5P001P400@01008110030@03004000@000<1000070000`40000[00<10P020@0400410@<000<10000
20020@03004001/00P4;00817P000`400@0D00008P0010400@4>00030@0000<00P4X00@100<00@40
10000`40000F00060@0100400P4400<100@00@010`000`40000L00030@0002`000@100000P401001
0@4300030@0000T0104J00<12`020Ad0104D00008P0010400@4>00030@0000<00P4X00@100<00@00
10000`40000F00810P000`400@020@<00`4010010043008170000`40000]00040@0000L10P000`40
0009008100<00@002P000`40000<00<130000`40000K00040@010A@0000R00040@010@h000<10000
0`020BP00P401001004500030@0001H00P4200030@0100810`001`400@0100400`020AL000<10000
0P000`40000b008100@00@010P000`40000:00030@0100/00P4=00030@0100d000<100006@020@03
004101@0000R00040@010@h000<100000`020BP00P401001004500030@0001H00P4200030@0100<1
0P001P400@0100H15`020@<000<10000<`001@400001008000<1000030000`40000900813@000`40
000=00030@0001D000<100400P4200030@0001<0000R00040@010@h000<100000`020BP00P420003
0@0000@00P4G00810P020@0400410@8000H100400@060AL00`4200030@0003<000@100000`4?0003
0@0000T00`4<00030@0000d000<100003@000`40000500D16000028000@1004150020BP00P420003
0@0000@00P4G00810P020@0400410@8000H100000@060AL00`4200030@0003L00`4?00030@0000H0
00<100000`4;00030@0000h000<1000030020@<000@100000`400`01000G00008P0010400@4D00<1
9`020@8000<1000010030AD00`42008100@00@410P001P40000100H15`030@8000<10000=`020A00
0P4700<10P030@T000<100003`000`40000;00L10P020@8000<100005P00028000@1004150030BL0
0P4200030@0000@00`4E00<10`000`4000020@03004000<000<100400P030AL00`4200030@0003L0
00<1000040000`40000300030@0100810`020@T000<100003`000`40000;00L100<00@0070000280
00@1004150030BL000D100000@0600<15@000`400@0300030@0000@110000`400@0300815`030@03
004004d000<100000P050@@000<100001`000`40000?00030@0000H01P4020010@400@017@000280
104D00<19`001@40000100H000<100405@000`400@0300040@0000<11@000`40000200815P001P40
0@400Dh000<100000P030@03004000<00P4700030@00010000<100001P050@@000D100400@0M0000
8P0010400@4D00<19`001@40000100P000<100004`000`400@0700<11@000`40000400030@000100
00@100410P000`400@1>00810P000`400@0800<100<00@400`4B00030@0000H00`4900817@000280
00@1004150030BL000D100000@0800030@0001<000<100001`030@D000<1000010000`40000<00D1
00<00@400`020Dl000@100012`020@03004100<14`000`40000400812`000`40000L00008P001040
0@4D00<19`000`40000:00030@0001<000<100001`020@H000<1000010000`40000<00D100<00@40
0`000`40001>00040@010@/00P4010010@4E00030@0000@00P4Z00008P0010400@4D00<19`000`40
000:008150000`40000700811P000`40000400030@0000`01`4400030@0004h00`4>00040@010AD0
0P4500030@0002T0000R00040@010A@00`4U00030@0100`00P4D00030@0000L000<1000030000`40
000<00810P030@@000<10000CP030@h000<100405P020@<00`4[00008P0010400@4D00030@0102D0
00<1004030020A<00P4900030@0000`000<1000030020@800P5G00813P000`40000F00810P030B`0
000R00040@010AH000<100008`000`400@0<00814`020AP000<100002`000`400@0200030@000800
00<100400P4]00008P0010400@4F00030@0002<00P4=00<14P000`40000G00030@0000/000<10000
QP040Bd0000R00040@010AH000<100008`020@d00`4B00030@0001L000<100002`000`40002600<1
;P00028000@100415P000`40000S00813P020A8000<1000060000`4000090081R0020Bl0000R0004
0@010AH000<100008P030@h00P4B00030@0001P00P4:00030@0008L000<10000;P000280104F0003
0@000240104>00<140000`40000I00812P000`40002h00008P0010400@4F00030@000240104>00<1
40000`40000I00812P000`40002h00008P0010400@4F00030@000240104?008140000`40000I0081
2P000`40002h00008P0010400@4F00030@000240104@00030@0000h000<100006@020@X000<10000
^000028000@100415P000`40000P00D140000`40000>00030@0001T000@100400P450081^P000280
00@100415P000`40000P00@14@000`40000>00030@0001T000@100400P450081^P00028000@10041
5P020B400P4C00030@0000h000<100006`0010400@4500030@000;T0000R00040@010AH00P4Q0003
0@0001<000<100003@000`40000K00040@010@D000<10000^@00028000@100415P020AL000<10000
1`000`40000C00030@0000d000<100006`0010400@440081^`00028000@100415P020AL000<10000
1`000`40000C00030@0000d000<100006`020@03004000<000<10000^P0000800`4600030@0000@0
104300D1100010400@4F00815`000`40000700030@0001<000<100003@000`40000K008100<00@00
0P020K`0000010010@020@/000@100000P4200050@000040100010400@4F00815`000`4000060081
5@000`40000=00030@0001/00P4200030@000081_00000030040008000<100003P001@40000100L0
104F00815`000`40000600815@000`40000<00030@0001`00P4200030@000081_000000300400080
00<100003P000`40000200030@0000@000@100415P020AL000<100001P020AD000<1000030000`40
000L00810P000`4000020K`000000`01000200030@0000X000<100000P4500030@0000<000@10041
5P020AL000<100001P000`40000D00030@0000`000<1000070020@8000@10041_@00000300400080
00<100002P040@L000<100000P0010400@4F00815`000`40000600030@0001@000<1000030000`40
000M00050@000@000P6m000000<00@000P000`40000:00030@0000@000D100000@0400040@010AH0
0P4G00030@0000H000<1000050000`40000<00030@0001d000@100000`6m000000@00@400P4<0003
0@0000@000<100000P4400040@010AH00P4G00030@0000H000<1000050000`40000<00030@000240
0`6m00000P030@d0104400<11@0010400@4F00815`000`40000600030@0001@000<1000030000`40
000Q0081_P00028000@100415P020AL00P4700030@0001@000<1000030000`40000Q0081_P000280
00@100415P020AH000<100401`000`40000E00030@0000/000<100008@000`40002m00008P001040
0@4G00030@0001@000<100401`000`40000E00030@0000/000<10000h@00028000@100415`000`40
000D00030@0100L000<100005@000`40000:0081h`00028000@100415`000`40000D00030@0100L0
00<100005@000`40000:0081h`00028000@100415`000`40000D00030@0100L000<100005@000`40
000:0081h`000280104G00030@0001@000<100401`000`40000E00030@0000X00P7S00008P001040
0@4G00030@0001@000<100401`000`40000E00030@0000X000<10000hP00028000<1004060000`40
000D00030@0100L000<100005@000`40000:00030@000>80000R00030@0101P000<1000050000`40
0@0500030@0101L000<100002P000`40003R00008P000`400@0H00815@000`400@0500030@0101L0
00<100002P000`40003R00008P000`400@0H00815@000`400@0500030@0101L000<100002P000`40
003R00008P000`400@0H00815@000`400@0300040@010AP000<100000P000`40000500030@000>80
000R00030@0101P00P4D008100@00@000P400`010@0H00030@00008000@100000P4200030@000>80
000R00030@0101P00P4D008100@00@000P400`010@0H00030@0000800P42008100<00@00i0000280
00<1004060020A<00`40100100020@03004101P000<100000P020@0600410@01i@00028000<10040
60020A<00`40100100020@03004101P000<100000P020@0600410@01i@00028000<1004060030A80
0`40100100020@03004101P000<100000P020@0600410@01i@0002800`4H00<100<00@003`030@03
004000<100<00@406@00104000060@0300400>@0000R00030@0101P00`400`01000?00<100<00@00
0`400`010@0I00040@0000D100<00@40i@00028000<1004060050A000`4200@100<00@406@000`40
00060@0300400>D0000R00030@0101P01@4@00810`020@8000<100406@000`4000050@8000<10000
i000028000<1004060050A000P4300810P000`40000I00050@000@000`4200030@000>@0000R0003
0@0101P01@4@00810`020@8000<100006@001@40004000810`000`40003T00008P000`400@0H00D1
3`030@<00P4200030@0001T00P401001004400030@000>@0000R00030@0101P000<100400P4?00<1
0`020@8000<100006@020@0300400>`0000R00030@0101P000<100400`4>00<10`020@8000<10000
6@020@0300400>`0000R00030@0101X0104>00<10`020@8000<100006@040Nd0000R00030@0101X0
104>00<10`020@8000<100006@040Nd0000R00030@0101X0104>00<10`020@8000<100006@040Nd0
000R00<16P040@h00P4400810P000`40000J00<1k@00028000<100406P040@8000<100002@020@@0
0P4200030@0001X00`7]00008P000`400@0J00@10P000`400009008110020Al000<10000k@000280
00<100406P040@8000<100002@000`4000030081o`0@00008P000`400@0J00050@010@000P4;0003
0@0000<00P7o0100000R00030@0101X000<10040104;00030@0000<00P7o0100000R00030@0101`0
1@4;00030@0000<00P7o0100000R00030@0101`01@4;00030@0000<00P7o0100000R00030@0101`0
1@4;00030@0000<00P7o0100000R00030@0101`01P4:00030@0000<00P7o0100000R00030@0101`0
1P4:00030@0000<000<10000o`0?00000P030@H000<1000010040@D0104300030@0101`01P4:0003
0@0000<000<10000o`0?000000@00@400P4;00040@0000811P000`40000200030@0101`01P4:0003
0@0000<000<10000o`0?000000<00@000P000`40000>00030@0000H10`030A`01P4900811@000`40
003o00l000000`01000200030@0000h000@100010`000`40000200030@0101`01P490081o`0G0000
00<00@000P000`40000:00030@0000810`001040004400030@0101`01`480081o`0G000000<00@00
0P000`40000:00@110020@03004000<000<1004070000`400@040@P000<10000o`0F000000<00@00
0P000`40000:00030@0000H000<1004010000`400@0L00030@0100@120000`40003o01H000001001
0@020@`000<100001`020@@000<1004070000`400@040@P000<10000o`0F00000P030@d010460081
10000`400@0L00050@000@000P4800030@000?l05P00028000<1004070001@40004000<11`000`40
003o01H0000R00030@0101`000D1000100040@H000<10000o`0F00008P000`400@0L00050@000@00
104600030@000?l05P00028000<100407`000`400@030@H000<10000o`0F00008P000`400@0O0003
0@0100<11P000`40003o01H0000R00<18@040@H000<10000o`0F00008P000`400@0Q00@11P000`40
003o01H0000R00030@01024000@100411@000`40003o01L0000R00030@01024000@100411@000`40
003o01L0000R00030@0102<00P4500030@000?l05`00028000<100408`020@D000<10000o`0G0000
8P000`400@0S00811@000`40003o01L0000R00030@0102<00P4500030@000?l05`00028000<10040
8`020@D000<10000o`0G00008P000`400@0S00<110000`40003o01L0000R00030@0102<00`440003
0@000?l05`00028000<100408`030@@000<10000o`0G00008P030B<000<1004010000`40003o01L0
000R00030@0102<000<1004010000`40003o01L0000R00030@0102<000<1004010000`40003o01L0
000R00030@0102<000<1004010000`40003o01L0000R00030@0102<000<1004010000`40003o01L0
003o06T0003o06T0003o06T0003o06T0003o06T0003o06T0003o06T0003o03l0104V0000o`100003
0@0002H0002W00030@0009D000<100009P0002h0104200030@0000<100<00@400`4010010@020@03
004000<110040@<01P4800@10`060@L00`4010010@4300@10P070@<00`4010010@050@T000<10000
2P040@8000<100000`400`010@030@040041008100<00@000`4400@10`060@P0104300H12@040@@0
104300D10P030@0400410@P01P4200D10`050@<000<100400P4300D1700002l000@100000P420003
0@0000810P001040004200030@0000810P020@800P4300030@00008000<1000020020@800P440003
0@0000T000D100000@0300810`000`400@0400040@0000810P020@800P4300030@0000H000<10000
2`00104000020@8000<100000P4200040@000@8000<100000P4200810P020@<000<100000P000`40
000800810P020@@000<100002@020@<000<100000P42008110000`40000300050@0000402`000`40
000500030@0000D000<100000`020@800P4200810`000`40000I0000;`001P4000400@8000D10000
0@0200040@000@8000@1000110001040004900030@0000P000<100000P000`40000200030@0000T0
00D100000@0300030@0000D01@4300030@00008000@1000110000`40000700030@0000X000H10001
004200050@0000400P001040004200040@000@@000@100012@000`40000800030@00008000<10000
0P000`40000900030@0000D000<100000P000`40000200030@0000<000D100000@0;00030@0000D0
00<100001@000`40000300030@00008000@1000110000`40000I0000;`001P4000400@8000D10000
0@0200040@000@8000@1000110000`4000060@@000<1000020000`40000200030@00008000<10000
2@001@40000100<01P4600050@00004010000`4000050@X000<100002P001P4000400@8000D10000
0@0200040@000@8000@1000110000`4000060@@000<1000020000`40000200030@00008000<10000
2@000`40000200040@000@@000<100000P000`40000300050@0000402`000`40000500030@0000D0
00<100000`000`40000200030@0000D1700002l000<100400P001040004300040@000@8000@10001
0P020@800P4200030@00008000<100000P020@8000<1000010020@800P4400030@0000T000D10000
0@0300030@00008000@100010`00104000020@800P4200030@00008000<1000020000`4000090003
0@01008000@100010`001040004200040@000@800P4200810P000`40000200030@0000800P420003
0@0000@00P42008110000`40000900810P020@800P42008110000`40000300050@0000402`000`40
000500030@0000D000<100000`020@800P4200030@00008000<100006@0002l000<100400P001@40
0@4000<100<00@401P4200030@01008110040@<00`4010010@4700@10`060@P0104500@110030@D0
0`400`01000200D12P000`40000900030@01008000D1004100030@03004100H10P000`400@020@@0
104300<100@00@411`040@<01P4900<100<00@000P040@<00`4400D12P060@@000<100000`030@@0
0P4010010@4400D16`0002l00P4300030@00010000<1000090000`40000900030@0001X000<10000
4@000`40000800810`000`40000@00030@0002@000<10000<@000`40000500030@0003@0000^0081
0P040A4000<1000090000`40000900030@0001X000<100004@000`40000700810P040A4000<10000
90000`40000K00030@0001<000<100001@000`40000500030@0002`00000\
\>"],
  ImageRangeCache->{{{151, 438}, {306.812, 129.875}} -> {-719.498, 0.535108, 
  4.03639, 0.000507682}}]
}, Open  ]],

Cell["\<\
If you run a small experiment, say you flip a coin ten times, then \
you shouldn't expect that there will always be five observed heads and five \
observed tails; however if you did the same experiment, ten coin flips, many \
times you would observe that this result, five heads and five tails, occurred \
more often than any other result. To illustrate the variation in the \
frequency estimates as a function of the size of the sample (the number of \
coin flips in this case), we use a package for displaying histograms. This \
package for Mathematica is available free from http://www.mathsource.com/. \
I've put a 
a copy in subdirectory of the following.  \
\>", "Text"],

Cell["PrependTo[$Path,\"/u/tld/tmp/class/Mathematica/\"];", "Input",
  InitializationCell->True],

Cell["<<Histograms`Histogram`", "Input"],

Cell[CellGroupData[{

Cell["\<\
Module[{n = 10, m = 100},
    Histogram[Table[Apply[Plus,Experiment[n]],{m}],
        Table[{i-0.5,i+0.5},{i,0,9,1}],
        Table[i,{i,1,10}],
        True]] ;\
\>", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0714286 0.0952381 0.0414698 1.98183 [
[.07143 .03077 -3 -9 ]
[.07143 .03077 3 0 ]
[.16667 .03077 -3 -9 ]
[.16667 .03077 3 0 ]
[.2619 .03077 -3 -9 ]
[.2619 .03077 3 0 ]
[.35714 .03077 -3 -9 ]
[.35714 .03077 3 0 ]
[.45238 .03077 -3 -9 ]
[.45238 .03077 3 0 ]
[.54762 .03077 -3 -9 ]
[.54762 .03077 3 0 ]
[.64286 .03077 -3 -9 ]
[.64286 .03077 3 0 ]
[.7381 .03077 -3 -9 ]
[.7381 .03077 3 0 ]
[.83333 .03077 -3 -9 ]
[.83333 .03077 3 0 ]
[.92857 .03077 -6 -9 ]
[.92857 .03077 6 0 ]
[.07143 .05217 -11 0 ]
[.07143 .05217 11 9.8125 ]
[.16667 .07199 -11 0 ]
[.16667 .07199 11 9.8125 ]
[.2619 .09181 -11 0 ]
[.2619 .09181 11 9.8125 ]
[.35714 .1909 -11 0 ]
[.35714 .1909 11 9.8125 ]
[.45238 .44854 -14 0 ]
[.45238 .44854 14 9.8125 ]
[.54762 .58727 -14 0 ]
[.54762 .58727 14 9.8125 ]
[.64286 .46836 -14 0 ]
[.64286 .46836 14 9.8125 ]
[.7381 .38908 -14 0 ]
[.7381 .38908 14 9.8125 ]
[.83333 .13144 -11 0 ]
[.83333 .13144 11 9.8125 ]
[.92857 .07199 -11 0 ]
[.92857 .07199 11 9.8125 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
.667 g
.02381 .04147 m
.02381 .04147 L
.11905 .04147 L
.11905 .04147 L
F
.11905 .04147 m
.11905 .06129 L
.21429 .06129 L
.21429 .04147 L
F
.21429 .04147 m
.21429 .08111 L
.30952 .08111 L
.30952 .04147 L
F
.30952 .04147 m
.30952 .1802 L
.40476 .1802 L
.40476 .04147 L
F
.40476 .04147 m
.40476 .43784 L
.5 .43784 L
.5 .04147 L
F
.5 .04147 m
.5 .57656 L
.59524 .57656 L
.59524 .04147 L
F
.59524 .04147 m
.59524 .45765 L
.69048 .45765 L
.69048 .04147 L
F
.69048 .04147 m
.69048 .37838 L
.78571 .37838 L
.78571 .04147 L
F
.78571 .04147 m
.78571 .12074 L
.88095 .12074 L
.88095 .04147 L
F
.88095 .04147 m
.88095 .06129 L
.97619 .06129 L
.97619 .04147 L
F
0 g
.002 w
[ ] 0 setdash
.02381 .04147 m
.02381 .04147 L
.11905 .04147 L
.11905 .04147 L
s
.11905 .04147 m
.11905 .06129 L
.21429 .06129 L
.21429 .04147 L
s
.21429 .04147 m
.21429 .08111 L
.30952 .08111 L
.30952 .04147 L
s
.30952 .04147 m
.30952 .1802 L
.40476 .1802 L
.40476 .04147 L
s
.40476 .04147 m
.40476 .43784 L
.5 .43784 L
.5 .04147 L
s
.5 .04147 m
.5 .57656 L
.59524 .57656 L
.59524 .04147 L
s
.59524 .04147 m
.59524 .45765 L
.69048 .45765 L
.69048 .04147 L
s
.69048 .04147 m
.69048 .37838 L
.78571 .37838 L
.78571 .04147 L
s
.78571 .04147 m
.78571 .12074 L
.88095 .12074 L
.88095 .04147 L
s
.88095 .04147 m
.88095 .06129 L
.97619 .06129 L
.97619 .04147 L
s
.02381 .04147 m
.97619 .04147 L
s
[(1)] .07143 .03077 0 1 Mshowa
[(2)] .16667 .03077 0 1 Mshowa
[(3)] .2619 .03077 0 1 Mshowa
[(4)] .35714 .03077 0 1 Mshowa
[(5)] .45238 .03077 0 1 Mshowa
[(6)] .54762 .03077 0 1 Mshowa
[(7)] .64286 .03077 0 1 Mshowa
[(8)] .7381 .03077 0 1 Mshowa
[(9)] .83333 .03077 0 1 Mshowa
[(10)] .92857 .03077 0 1 Mshowa
gsave
.07143 .05217 -72 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.) show
75.000000 11.250000 moveto
(%) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
gsave
.16667 .07199 -72 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1.) show
75.000000 11.250000 moveto
(%) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
gsave
.2619 .09181 -72 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(2.) show
75.000000 11.250000 moveto
(%) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
gsave
.35714 .1909 -72 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(7.) show
75.000000 11.250000 moveto
(%) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
gsave
.45238 .44854 -75 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(20.) show
81.000000 11.250000 moveto
(%) show
87.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
gsave
.54762 .58727 -75 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(27.) show
81.000000 11.250000 moveto
(%) show
87.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
gsave
.64286 .46836 -75 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(21.) show
81.000000 11.250000 moveto
(%) show
87.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
gsave
.7381 .38908 -75 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(17.) show
81.000000 11.250000 moveto
(%) show
87.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
gsave
.83333 .13144 -72 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(4.) show
75.000000 11.250000 moveto
(%) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
gsave
.92857 .07199 -72 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1.) show
75.000000 11.250000 moveto
(%) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
1 g
.5 Mabswid
.02381 .03612 m
.02381 .01472 L
s
.02381 .58192 m
.02381 .60332 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40005X0000gPP10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdk34g_ojn_l004J2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_P00oiT0IP3o0;C^]:JV
YZf][Jo^kP20P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;002k0?mT0]d009T0V@00
gNk^k^[Zj]gMgK^k^jZZZXR8R7MgMeEEEDA4A14A4Oooc?o<VLbIIS?oVC>IIS=VIYWoo`2Ioc<0VLc<
oiUVVNoGk`00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2o
o`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no
?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo
?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo
?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0
_ol0onj2k_Z0L_lo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoO`0l0>o<knl1k`1]
0:2PX:[Xo`3XooooP;Co]4Bkooo;og?X`?nXOOnCOOno_onooooo?oooOooo_mWEc:0Pl=WIfJj^[YWo
09Wo<iWoIYWoVIWoc<`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aV
VLaVc<aVolbI0<bI<lbIVLbIc<bIolc<0<c<<lc<IP000?l0J@000?l0J@000?l0J@000?l0J@000?l0
J@0001P01@4M00D17P030Al0104M00@17`030Al000<100007P030Al00`4J00D110030AL0000J0003
0@0001d000D100000@0M00030@00008180000`40000K00040@0000817P001040004N00030@0001d0
00D100000@0Q00030@0001T000<100000`020@03004101H0000J00030@0001h000<100008P000`40
000J00H18@000`40000K00050@0000407`000`40000L00050@0000408@020AX000<100000`001@40
000101H0000J00030@0001l000<100008@000`40000J00050@0000408P000`40000K00810P000`40
000M00030@0001`000D100000@0O008100<00@006@000`40000300050@0000405P0001X000<10000
80000`40000N00817P001040004N00030@0000817@020@8000<100007@000`40000M00<17`000`40
00020AX000<100000`001@40000101H0000J00030@00024000<100007`000`40000K008100<00@00
7@040Ah000@1004180000`40000L00050@0000407P000`4000020AX000<100000`001@40000101H0
000J00030@0001d000D100000@0Q00030@0001`000<100407P000`40000O00818P000`40000L0005
0@0000407P001@40000101X000<100000`001@40000101H0000H00030@0101l000<100000P4M0005
0@0000407`020Ah000<1000080000`40000M00050@0000407@001@40000101h000@100016@000`40
0@05008100<00@405P0001T00P4P00<17`030B000P4N00@180030A`01@4N00<180030AX00P4600<1
5`000?l0J@0000X0o`5D0@/0000/003o0BbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/
TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/
TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/
TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/
TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA000b0BbA;94/TBbA;94/TBbA;94/
TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TB`;0000;0000`4/;00N;0030B`/01l/00<1;2`0
7b`00`4/;00O;0030B`/01l/00<1;2`07b`00`4/;00O;0030B`/01h/00<1;2`07b`00`4000090000
;000o`4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/
TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/
TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/
TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/
T@6A;94/TBbA;94/TBbA;94/T@00<P4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/
TBbA;94/TBbA;94/2`0002`000<1;2`07R`00`4/;00O;0030B`/01l/00<1;2`07b`00`4/;00O;003
0B`/01l/00<1;2`07b`00`4/;00N;0030B`/01l/00<100002@0002`00?l1;94/TBbA;94/TBbA;94/
TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/
TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941
TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/
TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;940
0381;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;0/0000/0003
0B`/01h/00<1;2`07b`00`4/;00O;0030B`/01l/00<1;2`07b`00`4/;00O;0030B`/01l/00<1;2`0
7R`00`4/;00O;0030@0000T0000/02810>bA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941
TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/
TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/
TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/
TBbA;94/TBbA;94/TBbA;94/TBbA;2<12`0004d000<1;2`07b`00`4/;00O;0030B`/01l/00<1;2`0
7b`00`4/;00O;0030B`/01l/00<1;2`07R`00`40000[0000C@00kP6A;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;04]0000C@000`4/;00O;0030B`/01l/
00<1;2`07b`00`4/;00O;0030B`/01l/00<1;2`07b`00`4/;00N;0030@0002/0000A00<11P000`40
000400<1:@00kP6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;04]000040020@03004100`000<10040:@000`4/;00O;0030B`/01l/00<1;2`07b`00`4/
;00O;0030B`/01l/00<1;2`07b`00`4/;00N;0030@0002/0000@00050@00004030030BT00>h1TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TB`1;@000100
00D100000@0:00030@0002/08`4Q;0030B`/01l/00<1;2`07b`00`4/;00O;0030B`/01l/00<1;2`0
7R`00`40000[000040001@40000100/0105;003<0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TB`1;@000100
00D100000@0<00040@000DX000<1;2`07b`00`4/;00O;0030B`/01l/00<1;2`07b`00`4/;00O;003
0B`/01h/00<10000:`00010000D100000@0;00030@0104`00<`1TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;04]000040020@03004100/000<100403`050@D000<1000010030BT000<1;2`07b`00`4/;00O;003
0B`/01l/00<1;2`07b`00`4/;00O;0030B`/01h/00<1000010050@D000<1000010030A<0000A00<1
3@000`40000@00030@0000`000<10040:@00c06A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/0@P000<10000
30000`400@0C0000=0000`40000<00<1:@000`4/;00O;0030B`/01l/00<1;2`07b`00`4/;00O;003
0B`/01l/00<1;2`07R`00`40000600030@0000`00`4C0000=0000`40000:00030@0002/00<`1TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;04800030@0000X000<100005@0003@000<100002`040BT000<1;2`0
7b`00`4/;00O;0030B`/01l/00<1;2`07b`00`4/;00O;0030B`/01h/00<100001P000`40000;00@1
4`0003@000<1000030001040004X003<0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TB`120000`40000<0004
0@000A80000d00030@0000/000<10040:P000`4/;00O;0030B`/01l/00<1;2`07b`00`4/;00O;003
0B`/01l/00<1;2`07R`00`40000600030@0000/000<100405000038000<100403@000`400@0?00D1
1@000`40000400<11`00c06A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/0@H000<100403@000`400@0D0000
<`020@h000<100003P001@40000100`000<100401`000`4/;00O;0030B`/01l/00<1;2`07b`00`4/
;00O;0030B`/01l/00<1;2`07R`00`40000500813P000`40000C0000E@000`40000=00<11`00c06A
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/0Bd0001F00030@0000X000<100002@000`4/;00O;0030B`/01l/
00<1;2`07b`00`4/;00O;0030B`/01l/8P4]0000E`000`40000:00@11`00Z`6A;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@1>0000F0000`40000:00040@000@H0
00<1;2`07b`00`4/;00O;0030B`/01l/00<1;2`07b`00`4/;00O;0030@0004`0001D00050@000040
2`000`400@08002[0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94104h0001D00030@0000812`000`400@0800030B`/01l/00<1;2`07b`00`4/;00O;0030B`/01l/
00<1;2`07b`00`40001<0000E@030@d000<100001`00Z`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@1>0000K`000`4/;00O;0030B`/01l/00<1;2`07b`00`4/
;00O;0030B`/01l/00<10000C00006l00:/1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/T@40CP0006l000<1;2`07b`00`4/;00O;0030B`/01l/00<1;2`07b`00`4/
;00O;0030@0004`0001_002[0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94104h0001_00030B`/01l/00<1;2`07b`00`4/;00O;0030B`/01l/00<1;2`07b`00`40
001<0000K`00Z`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
0@0700@110000`40000400<1=@0006l000<1;2`07b`00`4/;00O;0030B`/01l/00<1;2`07b`00`4/
;00O;0030@0000L000<100002P000`400@0e0000K`00Z`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@0500H12`030CD0001_00030B`/01l/00<1;2`07b`00`4/
;00O;0030B`/01l/00<1;2`07b`00`40000300050@0000402P000`40000g0000K`00Z`6A;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@0600040@000@/0104e0000
K`000`4/;00O;0030B`/01l/00<1;2`07b`00`4/;00O;0030B`/01l/00<1000010020@03004000/0
00@10001=00006l00:/1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/T@401`000`400@0;00030@0103H0001_00030B`/01l/00<1;2`07b`00`4/;00O;0030B`/01l/
00<1;2`07b`00`40000600812`000`400@0f0000K`00Z`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@08008130000`40000e0000K`000`4/;00O;0030B`/01l/
00<1;2`07b`00`4/;00O;0030B`/01l/00<10000C00006l00:/1TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40CP0006l08`4Q;0030B`/01l/00<1;2`07b`00`4/
;00O;0030@0004`0002A00290I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40CP00094000<1;2`07b`00`4/;00O;003
0B`/01l/00<1;2`07b`00`40001<0000T@00R@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94104h0002A00030B`/01l/
00<1;2`07b`00`4/;00O;0030B`/01l/00<10000C000094008T1TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@1>0000
T@000`4/;00O;0030B`/01l/00<1;2`07b`00`4/;00O;0030@0004`0002A00290I4/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/T@40CP00094000<1;2`07b`00`4/;00O;0030B`/01l/00<1;2`07b`00`40001<0000T@00R@6A
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94104h0001h00030@0000D000<1000010030@L000<1;2`07b`00`4/;00O;003
0B`/01l/00<1;2`07b`00`40001<0000N0000`40000<00030@0100L008T1TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
0@1>0000N@000`40000;00<11`000`4/;00O;0030B`/01l/00<1;2`07b`00`4/;00O;0030@0004`0
001i00030@0000T000<100002@00R@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94104h0001i00030@0000X010470003
0B`/01l/00<1;2`07b`00`4/;00O;0030B`/01l/00<10000C00007T000<100002`00104000460029
0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/T@40CP0007T000<100002P000`400@0800030B`/01l/00<1;2`07b`00`4/
;00O;0030B`/01l/00<10000C00007H000D100000@0;00030@0100P008T1TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
0@1>0000MP050@`000<100001`000`4/;00O;0030B`/01l/00<1;2`07b`00`4/;00O;0030@0004`0
002A00290I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/T@40CP00094000<1;2`07b`00`4/;00O;0030B`/01l/00<1;2`0
7b`00`40001<0000T@00R@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94104h0002A00030B`/01l/00<1;2`07b`00`4/
;00O;0030B`/01l/00<10000C000094008T1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@1>0000T@000`4/;00O;003
0B`/01l/00<1;2`07b`00`4/;00O;0030@0004`0002A00290I4/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40CP000940
00<1;2`07b`00`4/;00O;0030B`/01l/00<1;2`07b`00`40001<0000T@00R@6A;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94104h0002A00030B`/01l/00<1;2`07b`00`4/;00O;0030B`/01l/00<10000C000094008T1TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA0@1>0000T@000`4/;00O;0030B`/01l/00<1;2`07b`00`4/;00O;0030@0004`0
002A00290I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/T@40CP00094000<1;2`07b`00`4/;00O;0030B`/01l/00<1;2`0
7b`00`40001<0000T@00R@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94104h0002A00030B`/01l/00<1;2`07b`00`4/
;00O;0030B`/01l/00<10000C000094008T1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@1>0000T@000`4/;00O;003
0B`/01l/00<1;2`07b`00`4/;00O;0030@0004`0002A00290I4/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40CP000940
00<1;2`07b`00`4/;00O;0030B`/01l/00<1;2`07b`00`40001<0000T@00R@6A;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94104h0002A00030B`/01l/00<1;2`07b`00`4/;00O;0030B`/01l/00<10000C000094008T1TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA0@1>0000T@000`4/;00O;0030B`/01l/00<1;2`07b`00`4/;00O;0030@0004`0
002A00290I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/T@40CP00094000<1;2`07b`00`4/;00O;0030B`/01l/00<1;2`0
7b`00`40001<0000T@00R@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94104h0002A00030B`/01l/00<1;2`07b`00`4/
;00O;0030B`/01l/00<10000C000094008T1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@1>0000T@000`4/;00O;003
0B`/01l/00<1;2`07b`00`4/;00O;0030@0004`0002A00290I4/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40CP000940
00<1;2`07b`00`4/;00O;0030B`/01l/00<1;2`07b`00`40001<0000T@00R@6A;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94104h0002A00030B`/01l/00<1;2`07b`00`4/;00O;0030B`/01l/00<10000C000094008T1TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA0@1>0000T@000`4/;00O;0030B`/01l/00<1;2`07b`00`4/;00O;0030@0004`0
002A00290I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/T@40CP00094000<1;2`07b`00`4/;00O;0030B`/01l/00<1;2`0
7b`00`40001<0000T@00R@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94104h0002A00030B`/01l/00<1;2`07b`00`4/
;00O;0030B`/01l/00<10000C000094008T1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@1>0000T@000`4/;00O;003
0B`/01l/00<1;2`07b`00`4/;00O;0030@0004`0002A00290I4/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40CP000940
00<1;2`07b`00`4/;00O;0030B`/01l/00<1;2`07b`00`40001<0000T@00R@6A;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94104h0002A00030B`/01l/00<1;2`07b`00`4/;00O;0030B`/01l/00<10000C000094008T1TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA0@1>0000T@000`4/;00O;0030B`/01l/00<1;2`07b`00`4/;00O;0030@0004`0
002A00290I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/T@40CP00094000<1;2`07b`00`4/;00O;0030B`/01l/00<1;2`0
7b`00`40001<0000T@00R@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94104h0002A00030B`/01l/00<1;2`07b`00`4/
;00O;2<1CP00094006L1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410700
002A00030B`/01l/00<1;2`07b`00`4/;00O;0030@0006h0002A001W0I4/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@1`0000T@000`4/;00O;0030B`/01l/00<1;2`07b`00`40
001^0000T@00I`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40L0000940
00<1;2`07b`00`4/;00O;0030B`/01l/00<10000KP00094006L1TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;9410700002A00030B`/01l/00<1;2`07b`00`4/;00O;0030@0006h0
002A001W0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@1`0000T@000`4/
;00O;0030B`/01l/00<1;2`07b`00`40001^0000T@00I`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/T@400P050@D000<100001@000`40000300<1D`00094000<1;2`07b`00`4/
;00O;0030B`/01l/00<100000P000`40000500030@0000/000<10040D`00094006L1TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94100@000<100001P000`40000:00<1D`000940
00<1;2`07b`00`4/;00O;0030B`/01l/00<100000P000`40000600030@0000P000<10000E@000940
06L1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94100@000<100001P000`40
000900@1D`00094000<1;2`07b`00`4/;00O;0030B`/01l/00<100000P000`40000600030@0000X0
00@10001DP00094006L1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94100@0
00<100001P000`40000900030@0105@0002A00030B`/01l/00<1;2`07b`00`4/;00O;0060@000@01
1@001@40000100X000<10040E000094006L1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94100<00P4500D12`000`40001C0000T@000`4/;00O;0030B`/01l/00<1;2`07b`00`40
001^0000T@0S0@14TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/
TBbA;94/TBbA;94/TBbA;94/T@5`0000/`000`4/;00O;0030B`/01l/00<10000KP000;<004D1TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/T@40L0000;<000<1;2`07b`00`4/;00O;0030@0006h0002c00150I4/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410700002c0003
0B`/01l/00<1;2`07b`00`40001^0000/`00A@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@1`0000/`000`4/;00O;2<1L0000;<0
02<1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@2B0000/`000`4/;00O;0030@000900
002c000S0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40TP0009@01@4400<11P000`40
000400<10`000`4/;00O;0030@000900002D00050@0000400`020@03004100`000<100400`008`6A
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410980002E00030@0000@000D100000@0<00<1
0`000`4/;00O;0030@000900002F00030@0000<000D100000@0:00030@0000D002<1TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@2B0000U`000`40000200050@0000402`040@<000<1;2`0
7b`00`40002@0000V0001@40000100<000<100002P0010400042000S0I4/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/T@40TP0009@000D100000@0300050@0000402`000`400@0400030B`/01l/
00<10000T00009@000<100000P43008100<00@402`000`400@04000S0I4/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/T@400P050@<01@4500030@0000<00`5e0000U@030@D00`4=00030@0000<0
00<1;2`07b`01040004300030@0000<000<100002`000`400@1e0000/`008`6A;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94100<000<100001P000`40000;00<1M@000;<000<1;2`07b`00`40
000200030@0000D000<100002@000`40001g0000/`008`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94100D000<1000010000`40000:00@1M@000;<000<1;2`07b`00`40000400030@0000<0
00<100002`001040005d0000/`008`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410080
00D100000@0500030@0000X000<10040MP000;<000<1;2`07b`01040004200810`000`400@0<0003
0@0107H0002c000S0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@400`030@D00P4=0003
0@0007D0002c00030B`/01l/00<10000T0000;<002<1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA0@2B0000/`000`4/;00O;0030@000900002c000S0I4/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/T@40TP000;<000<1;2`07b`00`40002@0000/`008`6A;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;9410980002c00030B`/01l/00<10000T0000;<002<1TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA0@2B0000/`000`4/;00O;0030@000900002c000S0I4/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/T@40TP000;<000<1;2`07b`00`40002@0000/`008`6A;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410980002c00030B`/01l/00<10000T0000;<002<1TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@2B0000/`000`4/;00O;0030@000900002c000S
0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40TP000;<000<1;2`07b`00`40002@0000
/`008`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410980002c00030B`/01l/00<10000
T0000;<002<1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@2B0000/`000`4/;00O;003
0@000900002c000S0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40TP000;<08`6B0000
o`1Y0000o`1Y0000o`1Y0000o`1Y0000o`1Y0000o`1Y0000o`1Y0000o`1Y0000o`1Y0000o`1Y0000
]P050@D000<100001@000`40000300<1U`000;H000D100000@0500030@0000/000<10040U`000;L0
00<100001`000`40000:00<1U`000;P000<100001P000`40000800030@0009T0002i00030@0000D0
00<100002@040IL0002j00030@0000@000<100002P001040006F0000\
\>"],
  ImageRangeCache->{{{108, 395}, {324.062, 147.125}} -> {-4.79386, 0.239887, 
  0.0369537, 0.00177583}}]
}, Open  ]],

Cell["\<\
Similarly we can look at the actual frequency of heads in somewhat \
larger experiments (32) where we consider quite a few such experiments (256).\
\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\
Module[{n = 32, m = 256},
    Histogram[Table[Apply[Plus,Experiment[n]]/n,{m}],
        Table[{i,i+.10},{i,0,.90,.10}],
        Table[i,{i,0,.90,.10}],
        True]] ;\
\>", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.0414698 0.0978458 [
[.07143 .03077 -3 -9 ]
[.07143 .03077 3 0 ]
[.16667 .03077 -9 -9 ]
[.16667 .03077 9 0 ]
[.2619 .03077 -9 -9 ]
[.2619 .03077 9 0 ]
[.35714 .03077 -9 -9 ]
[.35714 .03077 9 0 ]
[.45238 .03077 -9 -9 ]
[.45238 .03077 9 0 ]
[.54762 .03077 -9 -9 ]
[.54762 .03077 9 0 ]
[.64286 .03077 -9 -9 ]
[.64286 .03077 9 0 ]
[.7381 .03077 -9 -9 ]
[.7381 .03077 9 0 ]
[.83333 .03077 -9 -9 ]
[.83333 .03077 9 0 ]
[.92857 .03077 -9 -9 ]
[.92857 .03077 9 0 ]
[.07143 .05217 -11 0 ]
[.07143 .05217 11 9.8125 ]
[.16667 .05217 -11 0 ]
[.16667 .05217 11 9.8125 ]
[.2619 .06364 -23 0 ]
[.2619 .06364 23 9.8125 ]
[.35714 .11333 -17 0 ]
[.35714 .11333 17 9.8125 ]
[.45238 .58727 -23 0 ]
[.45238 .58727 23 9.8125 ]
[.54762 .33119 -23 0 ]
[.54762 .33119 23 9.8125 ]
[.64286 .14008 -23 0 ]
[.64286 .14008 23 9.8125 ]
[.7381 .05599 -26 0 ]
[.7381 .05599 26 9.8125 ]
[.83333 .05217 -11 0 ]
[.83333 .05217 11 9.8125 ]
[.92857 .05217 -11 0 ]
[.92857 .05217 11 9.8125 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
.667 g
.02381 .04147 m
.02381 .04147 L
.11905 .04147 L
.11905 .04147 L
F
.11905 .04147 m
.11905 .04147 L
.21429 .04147 L
.21429 .04147 L
F
.21429 .04147 m
.21429 .05294 L
.30952 .05294 L
.30952 .04147 L
F
.30952 .04147 m
.30952 .10262 L
.40476 .10262 L
.40476 .04147 L
F
.40476 .04147 m
.40476 .57656 L
.5 .57656 L
.5 .04147 L
F
.5 .04147 m
.5 .32048 L
.59524 .32048 L
.59524 .04147 L
F
.59524 .04147 m
.59524 .12938 L
.69048 .12938 L
.69048 .04147 L
F
.69048 .04147 m
.69048 .04529 L
.78571 .04529 L
.78571 .04147 L
F
.78571 .04147 m
.78571 .04147 L
.88095 .04147 L
.88095 .04147 L
F
.88095 .04147 m
.88095 .04147 L
.97619 .04147 L
.97619 .04147 L
F
0 g
.002 w
[ ] 0 setdash
.02381 .04147 m
.02381 .04147 L
.11905 .04147 L
.11905 .04147 L
s
.11905 .04147 m
.11905 .04147 L
.21429 .04147 L
.21429 .04147 L
s
.21429 .04147 m
.21429 .05294 L
.30952 .05294 L
.30952 .04147 L
s
.30952 .04147 m
.30952 .10262 L
.40476 .10262 L
.40476 .04147 L
s
.40476 .04147 m
.40476 .57656 L
.5 .57656 L
.5 .04147 L
s
.5 .04147 m
.5 .32048 L
.59524 .32048 L
.59524 .04147 L
s
.59524 .04147 m
.59524 .12938 L
.69048 .12938 L
.69048 .04147 L
s
.69048 .04147 m
.69048 .04529 L
.78571 .04529 L
.78571 .04147 L
s
.78571 .04147 m
.78571 .04147 L
.88095 .04147 L
.88095 .04147 L
s
.88095 .04147 m
.88095 .04147 L
.97619 .04147 L
.97619 .04147 L
s
.02381 .04147 m
.97619 .04147 L
s
[(0)] .07143 .03077 0 1 Mshowa
[(0.1)] .16667 .03077 0 1 Mshowa
[(0.2)] .2619 .03077 0 1 Mshowa
[(0.3)] .35714 .03077 0 1 Mshowa
[(0.4)] .45238 .03077 0 1 Mshowa
[(0.5)] .54762 .03077 0 1 Mshowa
[(0.6)] .64286 .03077 0 1 Mshowa
[(0.7)] .7381 .03077 0 1 Mshowa
[(0.8)] .83333 .03077 0 1 Mshowa
[(0.9)] .92857 .03077 0 1 Mshowa
gsave
.07143 .05217 -72 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.) show
75.000000 11.250000 moveto
(%) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
gsave
.16667 .05217 -72 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.) show
75.000000 11.250000 moveto
(%) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
gsave
.2619 .06364 -84 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1.1719) show
99.000000 11.250000 moveto
(%) show
105.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
gsave
.35714 .11333 -78 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(6.25) show
87.000000 11.250000 moveto
(%) show
93.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
gsave
.45238 .58727 -84 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(54.687) show
99.000000 11.250000 moveto
(%) show
105.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
gsave
.54762 .33119 -84 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(28.516) show
99.000000 11.250000 moveto
(%) show
105.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
gsave
.64286 .14008 -84 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(8.9844) show
99.000000 11.250000 moveto
(%) show
105.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
gsave
.7381 .05599 -87 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.39062) show
105.000000 11.250000 moveto
(%) show
111.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
gsave
.83333 .05217 -72 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.) show
75.000000 11.250000 moveto
(%) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
gsave
.92857 .05217 -72 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.) show
75.000000 11.250000 moveto
(%) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
1 g
.5 Mabswid
.02381 .03612 m
.02381 .01472 L
s
.02381 .58192 m
.02381 .60332 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40005X0000gPP10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdk34g_ojn_l004J2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_P00oiT0IP3o0;C^]:JV
YZf][Jo^kP20P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;002k0?mT0]d009T0V@00
gNk^k^[Zj]gMgK^k^jZZZXR8R7MgMeEEEDA4A14A4Oooc?o<VLbIIS?oVC>IIS=VIYWoo`2Ioc<0VLc<
oiUVVNoGk`00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2o
o`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no
?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo
?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo
?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0
_ol0onj2k_Z0L_lo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoO`0l0>o<knl1k`1]
0:2PX:[Xo`3XooooP;Co]4Bkooo;og?X`?nXOOnCOOno_onooooo?oooOooo_mWEc:0Pl=WIfJj^[YWo
09Wo<iWoIYWoVIWoc<`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aV
VLaVc<aVolbI0<bI<lbIVLbIc<bIolc<0<c<<lc<IP000?l0J@000?l0J@000?l0J@000?l0J@000?l0
J@0001T00`4G00<11P000`40000300D13P030@H000<100000`050@h00`4600030@0000@00`4?00<1
1P000`40000500@13@030@H000<1000010040@h00`4600030@0000D00`4=00<11P000`4000050003
0@0000h00`4600030@0000@00`4?00<11P000`40000400<14`0001P00P400`010@0E008100<00@40
3@000`40000=008100<00@402`001@40000100d00P400`010@0;00030@0000813@020@03004100l0
00<100002`020@03004100/000@100000P4<008100<00@403@001040004;008100<00@403@000`40
000=008100<00@402`001@40000100d00P400`010@0?00030@000100000H00050@0000405@001@40
000100d000<100003@001@40000100`000<100003P001@40000100l000<100002`001@40000100/0
1P4<00050@00004040000`40000:00050@00004030001@40000100/000D100000@0>00030@0000`0
00D100000@0;00050@0000403@001@40000100l00P4A000060001@40000101D000D100000@0=0003
0@0000d000D100000@0=00030@0000d000D100000@0?00030@0000/000D100000@0;00050@000040
3@001@400001010000<100002P001@40000100`00P4200030@0000T000D100000@0>00030@0000`0
00D100000@0;00050@0000403@001@40000100d00P400`01000@000060001@40000101D000D10000
0@0=00030@0000d000D100000@0>00030@0000`000D100000@0=00813P001@40000100`000@10001
3@001@40000100`000<100000P4<00050@00004030020@8000<100002@001@40000100h000<10000
30001@40000100`00`4>00050@00004030000`4000020A40000H00050@0000405@001@40000100d0
00<100003@001@40000100l000<100002`001@40000100l000<100002`001@40000100`00P400`01
000<00050@00004030040@d000D100000@0<00040@010@`000D100000@0>00030@0000`000D10000
0@0;00050@0000403@001@40000100`000<100000P4A000060001@40000101D000D100000@0=0003
0@0000d000D100000@0;00050@0000403@001@40000100l000<100002`001@40000100d000<10040
3@001@40000100`000<100003P001@40000100`00P4>00050@0000403P000`40000<00050@000040
2`001@40000100d000D100000@0<00050@0000404@0001P00P400`010@0E008100<00@402`000`40
0@0?008100<00@402`000`4000020@d00P400`010@0;00050@0000403@020@03004100h00P4=0081
00<00@4030000`40000>008100<00@403@000`40000<008100<00@402`001@40000100d00P400`01
0@0;00050@0000403@020@03004100`000@100014P0001T00`4G00<13@020A000`4=00<13`030@d0
0`4?00<13`020@h00`4=00@13P030@l00`4<00<130050@h00`4=00<13`030@h00`4B0000o`1Y0000
2P3o0E@12`0004d00<d1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94104h0001=00030B`/01l/00<1;2`0
7b`00`4/;00O;0030B`/01l/00<1;2`07b`S0Dh0001=002[0I4/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410700001=02<18B`00`4/;00O;0030B`/01l/00<1;2`0
7b`00`40001^0000K`00R@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410700001_00030B`/01l/00<1;2`07b`00`4/
;00O;0030B`/01l/00<10000KP0006l008T1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@1`0000K`000`4/;00O;003
0B`/01l/00<1;2`07b`00`4/;00O;0030@0006h0001_00290I4/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40L00006l0
00<1;2`07b`00`4/;00O;0030B`/01l/00<1;2`07b`00`40001^00004@030@H000<1000010030@l0
0`4600030@0000@00`4Y00290I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40:0030@H000<1000010030@l00`460003
0@0000@00`4C000040020@03004100`000<100403P020@03004100`000<10040:@000`4/;00O;003
0B`/01l/00<1;2`07b`00`4/;00D;0<11R`00`4/0@0400<11@030@D00`4600<10`090@`000<10040
3P020@03004100`000<100404`00010000D100000@0<00<13P001@40000100`00`4Y001l0I4/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/T@81
00//0@6A;94/TBbA0@0300030@0000811`00104000020@03004100D000@100010P002@4000010040
0@0<00<13P001@40000100`00`4C000040001@40000100X000<1000040001@40000100X000<10000
:`000`4/;00O;0030B`/01l/00<1;2`07b`00`4/;00C;0050B`/;0401b`00`40000500030@0000D0
0P4200050@00004010001@40000100<000D1000100030@X000<1000040001@40000100X000<10000
5@00010000D100000@0;00@13P001@40000100/0204500030@0000<01@4500030@0000<01@4400<1
00</TB`00`40PBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA
;94/TBbA0I4/T@6A;94/TBbA0@0700030@0000<00P401@01000100<000<100000P020@8000<10000
0P000`400@0300030@0000T0104>00050@0000402`040A<0000@00050@00004030001040004=0005
0@00004030000`4000020@l000<100001@000`40000500030@0000D000<1;0400R`00`4/0@0J;003
0B`/01l/00<1;2`07b`00`4/;00C;0050B`/;0401b`00`40000300811@000`4000020@8000D10000
0@0400810P000`40000300H130001040004=00050@00004030001040004B000040001@40000100/0
00<100403`001@40000100/000<100400P000`40000=00030@0000H000<1000010000`4000050005
0I410B`00`40PBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA
;94/TBbA0I4/T@6A;94/TBbA0@0700030@00008000<100000P4200050@000040100010400@470004
0@0100812P000`400@0?00050@0000402`000`400@0D000040020@03004100/000<100403`020@03
004100/000<100400P000`40000=00030@0000H000<1000010000`400005008100</0B`07B`00`4/
;00O;0030B`/01l/00<1;2`04b`01@4/;2`100L/00<100001@000`40000200050@0000400P001@40
000100@00P4500040@00008100<00@402`000`400@0?008100<00@402`000`400@0D00004@030@d0
00<100003`030@d000D100000@0?00030@0000H000<1000010000`400004008100>A0@401040M2bA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA0P402b`1
0I4/TBbA;94100<000D100000@0400040@000@<00P400`010@0500030@0000<000<100001@4=0003
0@0000l00`4=00030@0001<0001700030@0000d000<100001P000`40000400030@0000@00P401B`1
0B`1008/00<1;2`05b`00`4/;00O;0030B`/01l/00<1;2`052`30@P/00<100000P030@H00`4400<1
1`030@<00`4200030@0004@0001700030@0000d000<100001P000`40000400030@0000@00P40R94/
0@6A0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/
TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/
TBbA;94/TBbA;94/T@5`0000A@000`400@0=00030@0100D000D100000@0300030@0100H0904Q;003
0B`/01l/00<1;2`07b`00`40001^0000AP020@h00P4500D110020@L00`4200030@0001X006L1TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410700002A00030B`/01l/00<1;2`0
7b`00`4/;00O;0030@0006h0002A001W0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA0@1`0000T@000`4/;00O;0030B`/01l/00<1;2`07b`00`40001^0000T@00I`6A;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40L000094000<1;2`07b`00`4/;00O;003
0B`/01l/00<10000KP00094006L1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;9410700002A00030B`/01l/00<1;2`07b`00`4/;00O;0030@0006h0002A00140I4/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94S0G00
002A00030B`/01l/00<1;2`07b`00`40002@0000L0030@D000<100000`050@@0104300@104BA;94/
TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/
TBbA0I80001`00040@000@X000D100000@0300040@0000810P0010400@4Q;0030B`/01l/00<10000
T00006l000D100000@0;00030@0000T000<100001040A94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/
TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TP0006l00P4200030@0000X0
00<1000020020@@000<1;2`07b`00`4/;00O;0030@000900001_00810P000`40000;00030@0000<0
00<100000P400`010@030@14TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/
TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6B0000K`0010400@4?00030@000080104300040@000B4/
00<1;2`07b`00`40002@0000K`020@d000D100000@0400030@0000<004T100400I4/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410980
001`00030@0000/000<100000P4400030@0000<000D100400@0Q;0030B`/01l/00<10000T0000740
0`4;00<11@040@<004P10001TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/
TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6B0000T@000`4/;00O;0030B`/01P/0`44;0030@0100H0
0`4500<11P040@@0104010010@5Y0000T@00A@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;941TBbA0@0;00050@0000400`000`4000050003
0@0000D000H10001005Y0000T@000`4/;00O;0030B`/01L/00D1;2`/0@03;0030@0000T00P420005
0@0000400`060@801P4010010@5Y0000T@00A@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;941TBbA0@09008100D00@000@0300050@000040
0`001@40000100<00P5]0000T@000`4/;00O;0030B`/01P/0`44;0030@0000H000<100000P4300<1
1@001040004400050@000@00105Y0000T@00A@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;941TBbA0@0800030@0000810P001@40000100@0
0P400`010003008100D00@000@0200030@0006H0002A00030B`/01l/00<1;2`05b`01@4/;2`100</
00<100001P001@400001008000D100000@0500030@0100D000L100400@0106X0002A00150I4/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/T@6A
;94100P000@100010`001@40000100H00P46008100@00@01JP00094000<1;2`07b`00`4/;00H;0<1
12`00`40000700<110030@L00P4600810P000`40001Y0000T@00A@6A;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@2B0000T@000`4/
;00O;0030B`/01l/00<10000T000094004D1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40TP00094000<1;2`07b`00`4/;00O;003
0@000900002A00150I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;9410980002A00030B`/01l/00<1;2`07b`00`40002@0000T@00A@6A
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA0@2B0000T@000`4/;00O;0030B`/01l/00<10000T000094004D1TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40TP000940
00<1;2`07b`00`4/;00O;0030@000900002A00150I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410980002A00030B`/01l/00<1;2`0
7b`00`40002@0000T@00A@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@2B0000T@000`4/;00O;0030B`/01l/00<10000T0000940
04D1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/T@40TP00094000<1;2`07b`00`4/;00O;0030@000900002A00150I4/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410980
002A00030B`/01l/00<1;2`07b`00`40002@0000T@00A@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@2B0000T@000`4/;00O;003
0B`/01l/00<10000T000094004D1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40TP00094000<1;2`07b`00`4/;00O;0030@000900
002A00150I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;9410980002A00030B`/01l/00<1;2`07b`00`40002@0000T@00A@6A;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
0@2B0000T@000`4/;00O;0030B`/01l/00<10000T000094004D1TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40TP00094000<1;2`0
7b`00`4/;00O;0030@000900002A00150I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410980002A00030B`/01l/00<1;2`07b`00`40
002@0000T@00A@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA0@2B0000T@000`4/;00O;0030B`/01l/00<10000T000094004D1TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/T@40TP00094000<1;2`07b`00`4/;00O;0030@000900002A00150I4/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410980002A0003
0B`/01l/00<1;2`07b`00`40002@0000T@00A@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@2B0000T@000`4/;00O;0030B`/01l/
00<10000T000094004D1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/T@40TP00094000<1;2`07b`00`4/;00O;0030@000900002A0015
0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;9410980002A00030B`/01l/00<1;2`07b`00`40002@0000T@00A@6A;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@2B0000
T@000`4/;00O;0030B`/01l/00<10000T000094004D1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40TP00094000<1;2`07b`00`4/
;00O;0030@000900002A00150I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410980002A00030B`/01l/00<1;2`07b`00`40002@0000
T@008P6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94S0I80002A00030B`/01l/00<10000
/P00094002<1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@2d0000T@000`4/;00O;003
0@000;80002A000S0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40]000094000<1;2`0
7b`00`40002b0000T@008`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410;@0002A0003
0B`/01l/00<10000/P00094002<1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@2d0000
T@000`4/;00O;0030@000;80002A000S0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40
]000094000<1;2`05b`50@</104600030@0000@0104300D11@030@800`6;0000T@008`6A;94/TBbA
;94/TBbA;94/TBbA;94/TBbA0I4/T@6A;94100<000<100002@00104000020@@000<100001@002040
00400@01R`00094000<1;2`062`00`4/;004;0050@00004040000`40000200030@0000@000H10000
0@030H/0002A000S0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/T@400`000`40000>0003
0@00008000<1000010020@8000<10000S@00094000<1;2`06R`00`4/;002;0@13@000`4000020@@0
00<1000010020@801@6;0000T@008`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94100<0
00<100002P040@D000<10000100010400@4200040@000HX0002A00030B`/01L/00D1;2`/0@03;005
0@00004030000`40000600030@0000@00P4300030@0108`0002A000M0I4/TBbA;94/TBbA;94/TBbA
;94/TBbA;941TB`00P40194/T@4300030@0000X000<1000010000`400@0700070@0000400@2<0000
T@000`4/;00H;0<112`40@d01044008120030@03004008`0002A000S0I4/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/T@40]000094000<1;2`07b`00`40002b0000T@008`6A;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;9410;@0002A00030B`/01l/00<10000/P00094002<1TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@2d0000T@000`4/;00O;0030@000;80002A000S0I4/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40]000094000<1;2`07b`00`40002b0000T@008`6A
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410;@0002A00030B`/01l/00<10000/P000940
02<1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@2d0000T@000`4/;00O;0030@000;80
002A000S0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40]000094000<1;2`07b`00`40
002b0000T@008`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410;@0002A00030B`/01l/
00<10000/P00094002<1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@2d0000T@000`4/
;00O;0030@000;80002A000S0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40]0000940
00<1;2`07b`00`40002b0000T@008`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410;@0
002A00030B`/01l/00<10000/P00094002<1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
0@2d0000T@000`4/;00O;0030@000;80002A000S0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/T@40]000094000<1;2`07b`00`40002b0000T@008`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;9410;@0002A00030B`/01l/00<10000/P00094002<1TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA0@2d0000T@000`4/;00O;0030@000;80002A000S0I4/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/T@40]000094000<1;2`07b`00`40002b0000T@008`6A;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;9410;@0002A00030B`/01l/00<10000/P00094002<1TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@2d0000T@000`4/;00O;0030@000;80002A000S0I4/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40]000094000<1;2`07b`00`40002b0000T@008`6A
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410;@0002A00030B`/01l/00<10000/P000940
02<1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@2d0000T@000`4/;00O;0030@000;80
002A000S0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40]000094000<1;2`07b`00`40
002b0000T@008`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410;@0002A00030B`/01l/
00<10000/P00094002<1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@2d0000T@000`4/
;00O;0030@000;80002A000S0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40]0000940
00<1;2`07b`00`40002b0000T@008`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410;@0
002A00030B`/01l/00<10000/P00094002<1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
0@2d0000T@000`4/;00O;0030@000;80002A000S0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/T@40]000094000<1;2`07b`00`40002b0000T@008`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;9410;@0002A00030B`/01l/00<10000/P00094002<1TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA0@2d0000T@000`4/;00O;0030@000;80002A000S0I4/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/T@40]000094000<1;2`07b`00`40002b0000T@008`6A;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;9410;@0002A00030B`/01l/00<10000/P00094002<1TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@2d0000T@000`4/;00O;0030@000;80002A000S0I4/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40]000094000<1;2`07b`00`40002b0000T@008`6A
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410;@0002A00030B`/01l/00<10000/P000940
02<1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@2d0000T@0S0K@0003o06T0003o06T0
003o06T0003o06T0003o06T0003o06T0003o06T0003o06T0003o06T0003o06T0002:00@11@040@@0
00<100001@030@@00`4600030@0000800`6]0000R@00104000020@H000<100002`00104000420005
0@0000401@000`40000200030@010:d0002>00030@0000H12`001@400001008000D100000@060004
0@0000<1[@0008h000@100010`000`40000:00810P001040004300030@0000@000<10040/@0008X0
00<100000P4300040@000@`00P4200040@0000<11`000`4000040Jd0002:00@110020@03004000/0
00@100410`001@40000100H000D100000@0200030@000:X00000\
\>"],
  ImageRangeCache->{{{108, 395}, {613, 436.062}} -> {-0.429386, 15.2516, 
  0.00369537, 0.0359689}}]
}, Open  ]],

Cell["\<\
Now a probability is a mathematical notion that (at least in the \
case of some interpretations) has to be defined in terms of infinite limits \
and such.  For the most part we will think of probabilities in terms of what \
we expect to see on average in an experiment whether or not we could actually \
carry out the experiment.  Having dispensed with the preliminaries we can \
proceed to apply probability theory. Learning how to define and use \
probabilities is simple and intuitive!  Wrong!\
\>", "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The Law of Averages and Other Misunderstandings about Probability\
\
\>", "Section"],

Cell["\<\
Humans often get probabilities wrong in the sense that they try to \
use the notion of probability to describe a situation and then they draw the \
wrong conclusion.  What goes wrong?\
\>", "Text"],

Cell["\<\
One problem concerns the so-called \"law of averages\".  As an \
example, consider the reasoning of a gambler contemplating whether or not to \
continue playing: \"I've just had a long string of losses, but now my luck \
just has to change - it's the law of averages.\"\
\>", "Text"],

Cell["\<\
Psychologists posit that the gambler believes that the underlying \
stochastic process, whether it is determined by the roll of a die, the deal \
of hand from well-shuffled deck, or the spin of a roulette wheel, has memory \
- as though it's keeping track of wins and losses and making sure that \
everything works out as regards the long-term frequencies.\
\>", "Text"],

Cell["\<\
For the most part, stochastic processes of the sort we are \
discussing here are memoryless, and one throw of a die or flip of a coin is \
independent of any other.  How can this be and still have the long term \
frequencies tend to a limit?  Simple, suppose that you have a long (and quite \
rare - the probability of this happening is 0.5^20 or approximately 9.5 X \
10^7) run of heads, let's say 20 heads in a row.  Don't you need 20 tails to \
offset this statistical \"debt\"?  No or at least not in the sense that there \
is any particular urgency to pay off this debt.  In the long run, 20 is a \
vanishingly small number.  If, for example, the full run is 10000 coin flips \
then the 20 flips only account for a small portion 20/10000 or 0.2% of the \
overall estimate (20 + # of heads observed in the remaining 980 flips)/1000.  \
Even with a total run of only 1000 coin flips, an initial run of 20 heads \
matters very little.\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\
ListPlot[RunningFrequencies[Join[Table[1,{20}],Experiment[980]]], 
         {PlotJoined -> True,
          PlotLabel -> \"Number of heads / Number of coin flips\"}] ;\
\>", 
  "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.000952381 -2.14066 4.72386 [
[.21429 .20877 -9 -9 ]
[.21429 .20877 9 0 ]
[.40476 .20877 -9 -9 ]
[.40476 .20877 9 0 ]
[.59524 .20877 -9 -9 ]
[.59524 .20877 9 0 ]
[.78571 .20877 -9 -9 ]
[.78571 .20877 9 0 ]
[.97619 .20877 -12 -9 ]
[.97619 .20877 12 0 ]
[.01131 .03232 -24 -4.5 ]
[.01131 .03232 0 4.5 ]
[.01131 .12679 -24 -4.5 ]
[.01131 .12679 0 4.5 ]
[.01131 .31575 -24 -4.5 ]
[.01131 .31575 0 4.5 ]
[.01131 .41023 -24 -4.5 ]
[.01131 .41023 0 4.5 ]
[.01131 .5047 -24 -4.5 ]
[.01131 .5047 0 4.5 ]
[.01131 .59918 -24 -4.5 ]
[.01131 .59918 0 4.5 ]
[.5 .63053 -116 0 ]
[.5 .63053 116 10.125 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.21429 .22127 m
.21429 .22752 L
s
[(200)] .21429 .20877 0 1 Mshowa
.40476 .22127 m
.40476 .22752 L
s
[(400)] .40476 .20877 0 1 Mshowa
.59524 .22127 m
.59524 .22752 L
s
[(600)] .59524 .20877 0 1 Mshowa
.78571 .22127 m
.78571 .22752 L
s
[(800)] .78571 .20877 0 1 Mshowa
.97619 .22127 m
.97619 .22752 L
s
[(1000)] .97619 .20877 0 1 Mshowa
.125 Mabswid
.07143 .22127 m
.07143 .22502 L
s
.11905 .22127 m
.11905 .22502 L
s
.16667 .22127 m
.16667 .22502 L
s
.2619 .22127 m
.2619 .22502 L
s
.30952 .22127 m
.30952 .22502 L
s
.35714 .22127 m
.35714 .22502 L
s
.45238 .22127 m
.45238 .22502 L
s
.5 .22127 m
.5 .22502 L
s
.54762 .22127 m
.54762 .22502 L
s
.64286 .22127 m
.64286 .22502 L
s
.69048 .22127 m
.69048 .22502 L
s
.7381 .22127 m
.7381 .22502 L
s
.83333 .22127 m
.83333 .22502 L
s
.88095 .22127 m
.88095 .22502 L
s
.92857 .22127 m
.92857 .22502 L
s
.25 Mabswid
0 .22127 m
1 .22127 L
s
.02381 .03232 m
.03006 .03232 L
s
[(0.46)] .01131 .03232 1 0 Mshowa
.02381 .12679 m
.03006 .12679 L
s
[(0.48)] .01131 .12679 1 0 Mshowa
.02381 .31575 m
.03006 .31575 L
s
[(0.52)] .01131 .31575 1 0 Mshowa
.02381 .41023 m
.03006 .41023 L
s
[(0.54)] .01131 .41023 1 0 Mshowa
.02381 .5047 m
.03006 .5047 L
s
[(0.56)] .01131 .5047 1 0 Mshowa
.02381 .59918 m
.03006 .59918 L
s
[(0.58)] .01131 .59918 1 0 Mshowa
.125 Mabswid
.02381 .05594 m
.02756 .05594 L
s
.02381 .07956 m
.02756 .07956 L
s
.02381 .10318 m
.02756 .10318 L
s
.02381 .15041 m
.02756 .15041 L
s
.02381 .17403 m
.02756 .17403 L
s
.02381 .19765 m
.02756 .19765 L
s
.02381 .24489 m
.02756 .24489 L
s
.02381 .26851 m
.02756 .26851 L
s
.02381 .29213 m
.02756 .29213 L
s
.02381 .33937 m
.02756 .33937 L
s
.02381 .36299 m
.02756 .36299 L
s
.02381 .38661 m
.02756 .38661 L
s
.02381 .43385 m
.02756 .43385 L
s
.02381 .45747 m
.02756 .45747 L
s
.02381 .48108 m
.02756 .48108 L
s
.02381 .52832 m
.02756 .52832 L
s
.02381 .55194 m
.02756 .55194 L
s
.02381 .57556 m
.02756 .57556 L
s
.02381 .0087 m
.02756 .0087 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
gsave
.5 .63053 -177 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.125000 moveto
 558.000000 14.125000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.125000 lineto
closepath
clip newpath
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Number) show
105.000000 11.562500 moveto
(of) show
123.000000 11.562500 moveto
(heads) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
159.000000 11.562500 moveto
(\\220) show
171.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Number) show
213.000000 11.562500 moveto
(of) show
231.000000 11.562500 moveto
(coin) show
261.000000 11.562500 moveto
(flips) show
291.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Math2Mono
%%+ Courier
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.10861 .61803 m
.10952 .58868 L
.11048 .6106 L
s
.11048 .6106 m
.11081 .61803 L
s
.11188 .61803 m
.11238 .60223 L
.11333 .57305 L
.11429 .59421 L
.11524 .56572 L
.11619 .53782 L
.11714 .55869 L
.1181 .57914 L
.11905 .55194 L
.12 .52528 L
.12095 .49915 L
.1219 .47352 L
.12286 .44838 L
.12381 .46871 L
.12476 .48866 L
.12571 .50824 L
.12667 .48371 L
.12762 .50297 L
.12857 .52188 L
.12952 .54045 L
.13048 .55869 L
.13143 .5348 L
.13238 .55277 L
.13333 .52935 L
.13429 .50633 L
.13524 .48371 L
.13619 .5015 L
.13714 .4793 L
.1381 .49683 L
.13905 .51407 L
.14 .53103 L
.14095 .50931 L
.1419 .48794 L
.14286 .46691 L
.14381 .48371 L
.14476 .46304 L
.14571 .4427 L
.14667 .42268 L
.14762 .40296 L
.14857 .38354 L
.14952 .40021 L
.15048 .3811 L
.15143 .36228 L
.15238 .37873 L
.15333 .36021 L
.15429 .34195 L
.15524 .32396 L
.15619 .34022 L
.15714 .35624 L
.1581 .37203 L
Mistroke
.15905 .35434 L
.16 .33689 L
.16095 .35249 L
.1619 .36787 L
.16286 .35069 L
.16381 .33374 L
.16476 .31703 L
.16571 .33224 L
.16667 .31575 L
.16762 .33077 L
.16857 .34558 L
.16952 .32933 L
.17048 .34397 L
.17143 .32794 L
.17238 .3424 L
.17333 .35667 L
.17429 .34086 L
.17524 .32526 L
.17619 .30984 L
.17714 .29462 L
.1781 .30875 L
.17905 .3227 L
.18 .33649 L
.18095 .3501 L
.1819 .36356 L
.18286 .34856 L
.18381 .36186 L
.18476 .34706 L
.18571 .33242 L
.18667 .34558 L
.18762 .33113 L
.18857 .31684 L
.18952 .32987 L
.19048 .31575 L
.19143 .30179 L
.19238 .28799 L
.19333 .27435 L
.19429 .26086 L
.19524 .27376 L
.19619 .26042 L
.19714 .27318 L
.1981 .25999 L
.19905 .24695 L
.2 .25957 L
.20095 .24667 L
.2019 .2339 L
.20286 .22127 L
.20381 .20877 L
.20476 .19641 L
.20571 .18417 L
Mistroke
.20667 .17206 L
.20762 .16008 L
.20857 .14822 L
.20952 .13648 L
.21048 .12487 L
.21143 .11337 L
.21238 .10198 L
.21333 .11445 L
.21429 .10318 L
.21524 .09201 L
.21619 .08096 L
.21714 .09329 L
.2181 .08233 L
.21905 .09453 L
.22 .08368 L
.22095 .07294 L
.2219 .0623 L
.22286 .05176 L
.22381 .06381 L
.22476 .05336 L
.22571 .0653 L
.22667 .05494 L
.22762 .04468 L
.22857 .05649 L
.22952 .06818 L
.23048 .058 L
.23143 .04792 L
.23238 .0595 L
.23333 .0495 L
.23429 .06096 L
.23524 .07232 L
.23619 .0624 L
.23714 .05256 L
.2381 .06381 L
.23905 .07496 L
.24 .0652 L
.24095 .05552 L
.2419 .06656 L
.24286 .05696 L
.24381 .04745 L
.24476 .05838 L
.24571 .04894 L
.24667 .05977 L
.24762 .07051 L
.24857 .06114 L
.24952 .05185 L
.25048 .04264 L
.25143 .05327 L
.25238 .04413 L
.25333 .05466 L
Mistroke
.25429 .04559 L
.25524 .03659 L
.25619 .02767 L
.25714 .0381 L
.2581 .02924 L
.25905 .02046 L
.26 .03079 L
.26095 .04104 L
.2619 .03232 L
.26286 .04248 L
.26381 .03382 L
.26476 .02522 L
.26571 .0167 L
.26667 .02676 L
.26762 .03675 L
.26857 .04665 L
.26952 .03818 L
.27048 .02976 L
.27143 .02142 L
.27238 .03123 L
.27333 .02294 L
.27429 .01472 L
.27524 .02444 L
.27619 .0341 L
.27714 .04368 L
.2781 .0355 L
.27905 .02738 L
.28 .03688 L
.28095 .04631 L
.2819 .05568 L
.28286 .06497 L
.28381 .05689 L
.28476 .04887 L
.28571 .05808 L
.28667 .05012 L
.28762 .04221 L
.28857 .03436 L
.28952 .02656 L
.29048 .03569 L
.29143 .02795 L
.29238 .03701 L
.29333 .046 L
.29429 .03831 L
.29524 .04723 L
.29619 .03958 L
.29714 .03199 L
.2981 .04085 L
.29905 .04964 L
.3 .05838 L
.30095 .05082 L
Mistroke
.3019 .04332 L
.30286 .05199 L
.30381 .0606 L
.30476 .06915 L
.30571 .07764 L
.30667 .07017 L
.30762 .06275 L
.30857 .07118 L
.30952 .07956 L
.31048 .07218 L
.31143 .08049 L
.31238 .08875 L
.31333 .08142 L
.31429 .08962 L
.31524 .08233 L
.31619 .07509 L
.31714 .08324 L
.3181 .09133 L
.31905 .08413 L
.32 .07697 L
.32095 .08501 L
.3219 .09299 L
.32286 .10092 L
.32381 .0938 L
.32476 .08673 L
.32571 .09461 L
.32667 .10243 L
.32762 .11021 L
.32857 .11794 L
.32952 .1109 L
.33048 .10391 L
.33143 .09696 L
.33238 .10463 L
.33333 .09772 L
.33429 .10535 L
.33524 .09848 L
.33619 .09165 L
.33714 .09923 L
.3381 .10675 L
.33905 .09996 L
.34 .09322 L
.34095 .08651 L
.3419 .07984 L
.34286 .08731 L
.34381 .08068 L
.34476 .08811 L
.34571 .08151 L
.34667 .07496 L
.34762 .08233 L
.34857 .08967 L
Mistroke
.34952 .09696 L
.35048 .09044 L
.35143 .09768 L
.35238 .09119 L
.35333 .0984 L
.35429 .10556 L
.35524 .11268 L
.35619 .10622 L
.35714 .0998 L
.3581 .10688 L
.35905 .11391 L
.36 .12091 L
.36095 .11452 L
.3619 .10817 L
.36286 .11512 L
.36381 .12203 L
.36476 .11571 L
.36571 .10943 L
.36667 .1163 L
.36762 .11005 L
.36857 .10383 L
.36952 .09764 L
.37048 .0915 L
.37143 .08538 L
.37238 .0793 L
.37333 .07325 L
.37429 .06723 L
.37524 .06125 L
.37619 .0553 L
.37714 .04938 L
.3781 .05619 L
.37905 .0503 L
.38 .05707 L
.38095 .05121 L
.3819 .04538 L
.38286 .05211 L
.38381 .05881 L
.38476 .06547 L
.38571 .0721 L
.38667 .07869 L
.38762 .08524 L
.38857 .07943 L
.38952 .08595 L
.39048 .09244 L
.39143 .09889 L
.39238 .10531 L
.39333 .09952 L
.39429 .10591 L
.39524 .10015 L
.39619 .1065 L
Mistroke
.39714 .10077 L
.3981 .10708 L
.39905 .10138 L
.4 .10766 L
.40095 .11391 L
.4019 .10823 L
.40286 .11445 L
.40381 .1088 L
.40476 .10318 L
.40571 .09758 L
.40667 .10376 L
.40762 .09819 L
.40857 .09265 L
.40952 .08714 L
.41048 .08165 L
.41143 .0878 L
.41238 .08233 L
.41333 .08845 L
.41429 .09453 L
.41524 .10059 L
.41619 .10661 L
.41714 .11261 L
.4181 .10717 L
.41905 .11314 L
.42 .10772 L
.42095 .10233 L
.4219 .10826 L
.42286 .10289 L
.42381 .1088 L
.42476 .10346 L
.42571 .09814 L
.42667 .10401 L
.42762 .10986 L
.42857 .10456 L
.42952 .09929 L
.43048 .09405 L
.43143 .08883 L
.43238 .08363 L
.43333 .08944 L
.43429 .08427 L
.43524 .09005 L
.43619 .09581 L
.43714 .10154 L
.4381 .10725 L
.43905 .10209 L
.44 .10777 L
.44095 .10264 L
.4419 .09753 L
.44286 .09244 L
.44381 .09809 L
Mistroke
.44476 .09302 L
.44571 .08798 L
.44667 .0936 L
.44762 .09919 L
.44857 .09417 L
.44952 .08917 L
.45048 .0842 L
.45143 .07924 L
.45238 .0848 L
.45333 .09034 L
.45429 .09586 L
.45524 .10135 L
.45619 .10682 L
.45714 .10188 L
.4581 .09696 L
.45905 .09206 L
.46 .08719 L
.46095 .09263 L
.4619 .09804 L
.46286 .09318 L
.46381 .08835 L
.46476 .09374 L
.46571 .0991 L
.46667 .10445 L
.46762 .09963 L
.46857 .09483 L
.46952 .10015 L
.47048 .09537 L
.47143 .10066 L
.47238 .0959 L
.47333 .10117 L
.47429 .09643 L
.47524 .09171 L
.47619 .09696 L
.47714 .10218 L
.4781 .09748 L
.47905 .0928 L
.48 .098 L
.48095 .09333 L
.4819 .09851 L
.48286 .09386 L
.48381 .09902 L
.48476 .09439 L
.48571 .08978 L
.48667 .09491 L
.48762 .10002 L
.48857 .10511 L
.48952 .10052 L
.49048 .09594 L
.49143 .10101 L
Mistroke
.49238 .10606 L
.49333 .11108 L
.49429 .11608 L
.49524 .11153 L
.49619 .10698 L
.49714 .10246 L
.4981 .09796 L
.49905 .09347 L
.5 .09845 L
.50095 .10341 L
.5019 .09894 L
.50286 .10388 L
.50381 .1088 L
.50476 .1137 L
.50571 .11858 L
.50667 .12344 L
.50762 .12828 L
.50857 .13311 L
.50952 .12865 L
.51048 .12421 L
.51143 .11978 L
.51238 .11538 L
.51333 .11099 L
.51429 .11579 L
.51524 .11141 L
.51619 .10706 L
.51714 .11184 L
.5181 .1075 L
.51905 .11226 L
.52 .117 L
.52095 .12173 L
.5219 .12643 L
.52286 .12211 L
.52381 .1178 L
.52476 .12248 L
.52571 .12715 L
.52667 .12286 L
.52762 .11858 L
.52857 .12323 L
.52952 .12786 L
.53048 .1236 L
.53143 .11935 L
.53238 .11512 L
.53333 .11973 L
.53429 .11551 L
.53524 .11131 L
.53619 .11591 L
.53714 .11172 L
.5381 .10755 L
.53905 .11213 L
Mistroke
.54 .11668 L
.54095 .12123 L
.5419 .11707 L
.54286 .11293 L
.54381 .1088 L
.54476 .11332 L
.54571 .11783 L
.54667 .12232 L
.54762 .12679 L
.54857 .13125 L
.54952 .12714 L
.55048 .12304 L
.55143 .11895 L
.55238 .12339 L
.55333 .11932 L
.55429 .11526 L
.55524 .11122 L
.55619 .11564 L
.55714 .12005 L
.5581 .12444 L
.55905 .12041 L
.56 .12478 L
.56095 .12914 L
.5619 .12512 L
.56286 .12947 L
.56381 .12546 L
.56476 .12147 L
.56571 .1175 L
.56667 .12182 L
.56762 .11786 L
.56857 .11391 L
.56952 .10998 L
.57048 .10606 L
.57143 .10215 L
.57238 .10646 L
.57333 .10256 L
.57429 .10685 L
.57524 .11113 L
.57619 .10725 L
.57714 .11151 L
.5781 .10764 L
.57905 .10378 L
.58 .09994 L
.58095 .10418 L
.5819 .10035 L
.58286 .09654 L
.58381 .10077 L
.58476 .10498 L
.58571 .10117 L
.58667 .09738 L
Mistroke
.58762 .10158 L
.58857 .0978 L
.58952 .09403 L
.59048 .09027 L
.59143 .08653 L
.59238 .0828 L
.59333 .08698 L
.59429 .09115 L
.59524 .0953 L
.59619 .09944 L
.59714 .10357 L
.5981 .09985 L
.59905 .09614 L
.6 .10025 L
.60095 .09655 L
.6019 .10065 L
.60286 .09696 L
.60381 .10104 L
.60476 .10511 L
.60571 .10144 L
.60667 .10549 L
.60762 .10183 L
.60857 .09817 L
.60952 .10222 L
.61048 .09857 L
.61143 .09494 L
.61238 .09133 L
.61333 .09535 L
.61429 .09175 L
.61524 .08815 L
.61619 .09216 L
.61714 .09616 L
.6181 .09258 L
.61905 .09656 L
.62 .09299 L
.62095 .09696 L
.6219 .0934 L
.62286 .08984 L
.62381 .0938 L
.62476 .09775 L
.62571 .09421 L
.62667 .09814 L
.62762 .09461 L
.62857 .09109 L
.62952 .08758 L
.63048 .08408 L
.63143 .088 L
.63238 .0919 L
.63333 .09579 L
.63429 .09231 L
Mistroke
.63524 .08883 L
.63619 .08536 L
.63714 .0819 L
.6381 .07846 L
.63905 .07502 L
.64 .0789 L
.64095 .08276 L
.6419 .07934 L
.64286 .08319 L
.64381 .08703 L
.64476 .09086 L
.64571 .09468 L
.64667 .09126 L
.64762 .08785 L
.64857 .09165 L
.64952 .09545 L
.65048 .09205 L
.65143 .09583 L
.65238 .09244 L
.65333 .09621 L
.65429 .09996 L
.65524 .10371 L
.65619 .10033 L
.65714 .10406 L
.6581 .10069 L
.65905 .10441 L
.66 .10105 L
.66095 .10476 L
.6619 .10141 L
.66286 .09807 L
.66381 .09474 L
.66476 .09844 L
.66571 .10212 L
.66667 .0988 L
.66762 .10248 L
.66857 .09916 L
.66952 .09586 L
.67048 .09257 L
.67143 .08928 L
.67238 .09294 L
.67333 .0966 L
.67429 .10024 L
.67524 .09696 L
.67619 .09369 L
.67714 .09732 L
.6781 .10094 L
.67905 .10455 L
.68 .10815 L
.68095 .10489 L
.6819 .10164 L
Mistroke
.68286 .10522 L
.68381 .1088 L
.68476 .11236 L
.68571 .10912 L
.68667 .11268 L
.68762 .11622 L
.68857 .11976 L
.68952 .11652 L
.69048 .12005 L
.69143 .12356 L
.69238 .12033 L
.69333 .12384 L
.69429 .12733 L
.69524 .12411 L
.69619 .12091 L
.69714 .12439 L
.6981 .12786 L
.69905 .12466 L
.7 .12813 L
.70095 .12493 L
.7019 .12839 L
.70286 .13183 L
.70381 .12865 L
.70476 .13208 L
.70571 .1355 L
.70667 .13233 L
.70762 .12916 L
.70857 .12601 L
.70952 .12286 L
.71048 .11972 L
.71143 .12313 L
.71238 .12 L
.71333 .11688 L
.71429 .11376 L
.71524 .11066 L
.71619 .11406 L
.71714 .11096 L
.7181 .11435 L
.71905 .11774 L
.72 .11465 L
.72095 .11802 L
.7219 .12138 L
.72286 .12474 L
.72381 .12165 L
.72476 .125 L
.72571 .12192 L
.72667 .11886 L
.72762 .12219 L
.72857 .11913 L
.72952 .12246 L
Mistroke
.73048 .11941 L
.73143 .11637 L
.73238 .11333 L
.73333 .11665 L
.73429 .11996 L
.73524 .11693 L
.73619 .12023 L
.73714 .11721 L
.7381 .1142 L
.73905 .1112 L
.74 .1082 L
.74095 .11149 L
.7419 .11477 L
.74286 .11178 L
.74381 .11505 L
.74476 .11831 L
.74571 .11533 L
.74667 .11858 L
.74762 .12182 L
.74857 .12506 L
.74952 .12208 L
.75048 .11912 L
.75143 .11616 L
.75238 .11321 L
.75333 .11027 L
.75429 .11349 L
.75524 .11056 L
.75619 .10763 L
.75714 .10471 L
.7581 .1018 L
.75905 .09889 L
.76 .09599 L
.76095 .09921 L
.7619 .09632 L
.76286 .09344 L
.76381 .09664 L
.76476 .09984 L
.76571 .09696 L
.76667 .10015 L
.76762 .09728 L
.76857 .10046 L
.76952 .09759 L
.77048 .09474 L
.77143 .09791 L
.77238 .09506 L
.77333 .09822 L
.77429 .09538 L
.77524 .09255 L
.77619 .08972 L
.77714 .0869 L
Mistroke
.7781 .08409 L
.77905 .08724 L
.78 .09038 L
.78095 .09352 L
.7819 .09071 L
.78286 .09384 L
.78381 .09104 L
.78476 .09416 L
.78571 .09727 L
.78667 .09448 L
.78762 .09169 L
.78857 .09479 L
.78952 .09201 L
.79048 .09511 L
.79143 .09819 L
.79238 .09542 L
.79333 .09265 L
.79429 .08989 L
.79524 .09297 L
.79619 .09022 L
.79714 .08747 L
.7981 .09054 L
.79905 .0936 L
.8 .09665 L
.80095 .09391 L
.8019 .09696 L
.80286 .1 L
.80381 .10303 L
.80476 .10606 L
.80571 .10332 L
.80667 .10059 L
.80762 .09787 L
.80857 .10088 L
.80952 .10389 L
.81048 .10117 L
.81143 .09846 L
.81238 .09576 L
.81333 .09876 L
.81429 .09606 L
.81524 .09337 L
.81619 .09068 L
.81714 .08801 L
.8181 .08533 L
.81905 .08832 L
.82 .09131 L
.82095 .08864 L
.8219 .09162 L
.82286 .09459 L
.82381 .09755 L
.82476 .10051 L
Mistroke
.82571 .10346 L
.82667 .10079 L
.82762 .10373 L
.82857 .10667 L
.82952 .10401 L
.83048 .10694 L
.83143 .10986 L
.83238 .10721 L
.83333 .11012 L
.83429 .10748 L
.83524 .10484 L
.83619 .10774 L
.83714 .10511 L
.8381 .10248 L
.83905 .10538 L
.84 .10827 L
.84095 .10565 L
.8419 .10304 L
.84286 .10592 L
.84381 .10331 L
.84476 .10619 L
.84571 .10906 L
.84667 .11192 L
.84762 .10932 L
.84857 .10672 L
.84952 .10958 L
.85048 .11243 L
.85143 .11527 L
.85238 .11268 L
.85333 .11009 L
.85429 .11293 L
.85524 .11034 L
.85619 .10777 L
.85714 .1052 L
.8581 .10803 L
.85905 .11085 L
.86 .10829 L
.86095 .10573 L
.8619 .10854 L
.86286 .11135 L
.86381 .11415 L
.86476 .1116 L
.86571 .10905 L
.86667 .11185 L
.86762 .10931 L
.86857 .1121 L
.86952 .10956 L
.87048 .10703 L
.87143 .1045 L
.87238 .10198 L
Mistroke
.87333 .09947 L
.87429 .10225 L
.87524 .09974 L
.87619 .10252 L
.87714 .10528 L
.8781 .10278 L
.87905 .10028 L
.88 .10304 L
.88095 .1058 L
.8819 .10855 L
.88286 .10606 L
.88381 .10357 L
.88476 .10631 L
.88571 .10905 L
.88667 .10656 L
.88762 .10929 L
.88857 .11202 L
.88952 .11474 L
.89048 .11745 L
.89143 .11497 L
.89238 .11768 L
.89333 .12038 L
.89429 .11791 L
.89524 .1206 L
.89619 .11813 L
.89714 .11567 L
.8981 .11321 L
.89905 .1159 L
.9 .11344 L
.90095 .11613 L
.9019 .11368 L
.90286 .11635 L
.90381 .11902 L
.90476 .11658 L
.90571 .11414 L
.90667 .11681 L
.90762 .11946 L
.90857 .11703 L
.90952 .1146 L
.91048 .11218 L
.91143 .11483 L
.91238 .11242 L
.91333 .11506 L
.91429 .11265 L
.91524 .11024 L
.91619 .11288 L
.91714 .11551 L
.9181 .11814 L
.91905 .11574 L
.92 .11836 L
Mistroke
.92095 .12098 L
.9219 .12359 L
.92286 .12619 L
.92381 .12879 L
.92476 .13139 L
.92571 .12899 L
.92667 .1266 L
.92762 .12918 L
.92857 .12679 L
.92952 .12938 L
.93048 .13196 L
.93143 .13453 L
.93238 .13709 L
.93333 .13471 L
.93429 .13233 L
.93524 .12995 L
.93619 .12758 L
.93714 .12522 L
.9381 .12286 L
.93905 .12542 L
.94 .12797 L
.94095 .13052 L
.9419 .13307 L
.94286 .13071 L
.94381 .12836 L
.94476 .1309 L
.94571 .13343 L
.94667 .13596 L
.94762 .13361 L
.94857 .13614 L
.94952 .13379 L
.95048 .13631 L
.95143 .13397 L
.95238 .13648 L
.95333 .13415 L
.95429 .13182 L
.95524 .13433 L
.95619 .13201 L
.95714 .13451 L
.9581 .13219 L
.95905 .12987 L
.96 .12756 L
.96095 .13006 L
.9619 .13255 L
.96286 .13504 L
.96381 .13273 L
.96476 .13043 L
.96571 .12813 L
.96667 .13061 L
.96762 .12832 L
Mistroke
.96857 .13079 L
.96952 .12851 L
.97048 .12622 L
.97143 .12395 L
.97238 .12642 L
.97333 .12414 L
.97429 .12661 L
.97524 .12434 L
.97619 .12207 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40005X0000gPP10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdk34g_ojn_l004J2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_P00oiT0IP3o0;C^]:JV
YZf][Jo^kP20P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;002k0?mT0]d009T0V@00
gNk^k^[Zj]gMgK^k^jZZZXR8R7MgMeEEEDA4A14A4Oooc?o<VLbIIS?oVC>IIS=VIYWoo`2Ioc<0VLc<
oiUVVNoGk`00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2o
o`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no
?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo
?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo
?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0
_ol0onj2k_Z0L_lo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoO`0l0>o<knl1k`1]
0:2PX:[Xo`3XooooP;Co]4Bkooo;og?X`?nXOOnCOOno_onooooo?oooOooo_mWEc:0Pl=WIfJj^[YWo
09Wo<iWoIYWoVIWoc<`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aV
VLaVc<aVolbI0<bI<lbIVLbIc<bIolc<0<c<<lc<IP000?l0J@000?l0J@00028000<10000o`140000
8P000`40003o04@0000R00030@000?l0A00002800P7o04D0000R00030@000?l0A000028000<10000
B`001040007d00008P000`40001;00040@000O@0000R00030@0004T000<100400`7d00000P030@H0
00<100001@040@@00`4400030@0004T000<100400`7d000000@00@400P4?00030@0000<000@10001
0`000`400018008100<00@400P40100100020Nh000000`01000200030@0000T01P4300050@000040
0`030DP01P400`010@02008100<00@00j`0000030040008000<100002@001@40000100@00P420005
0@000040BP060@03004100800P400`01003[000000<00@000P000`40000:00040@000@@00P420005
0@000040BP060@0300410080107/000000<00@000P000`40000:008100<00@000`0010400@440003
0@0004P000@100400P401P010@0100<1k00000030040008000<100002`000`400@0400811P000`40
000o00030@0000@00`4010010042008100<00@00107[000000@00@400P4>00811@000`4000040003
0@0003l00P4300030@0100810`00204000010040104I00030@000<l0000200<13`020@H00`430003
0@0003l00`4200D11`030@<00P4H0081d@00028000<10000?@060@03004100812@020@<00P4H0081
d@0002800P4n00/12@020@<00P4G008100<00@00c`00028000<10000?@0:0@X00P4300030@0101H0
00@10001d000028000<10000?@020@03004100D12P000`40000400815@001040007@00008P000`40
000m008100<00@401@4:00030@0000@00P4E00040@000M00000R00030@0003d000<100001P4B0081
5@001040007@00008P000`40000m00090@0000400@0101800`4D00040@000M00000R00030@0003d0
00<100000`000`40000B00<100<00@0020000`40000500030@00008000<10000C`000`40001k0000
8P000`40000m00030@0001P000@100400P4800030@0000D000<100000P000`40001>0081O@000280
0P4l00817@040@H00`4700030@00008000<1000010000`40000;00030@0003T00`5l00008P000`40
000k00817@040@H00`4700030@0000800P4500811P000`4000020081;@000`40000900030@010081
:`000`40001>00008P000`40000k00817@040@03004000<00`4700030@0000800P4500811P000`40
000200810P000`40000W00812`001`400@010040:@000`40000300030@0000@000<10000@@000280
00<10000>`020Ad00`400`010@0400@11P000`40000300030@0000<00P4500810`030@800P460003
0@0001l00P4600080@000@01004200811P000`40000O00030@0100<00`4400<1@P00028000<10000
>`020Al000@100410P020@<000<100000`000`40000400030@0000<000<10040100010400@030@03
004100810P000`400@020@<000<100006`020@D01P400`01000200<11@030A/000H100000@070@@0
00<10040@P00028000<10000>`020B400P4010010@43008110000`40000400030@0000<000<10040
100010400@030@03004100X10P020AX000@100400`4200030@0100<100<00@000P030@<000<10040
0P4J00<100<00@400P040@03004100800P4200030@0003l0000R00030@0003/000<1000080020@04
00410@<00`4300030@0000@00P4300040@000@8000H100400@020@802P4010010@020AT00`4200<1
00<00@000`020@P000<100000P400`010@0J00030@01008110020@<000<100400P4200030@0000d0
0P4`00008P020C/00P4R00H110020@<000<100001@030@060040004010400`010@0400810P040@<0
0P400`010@0F00030@0000<10P050@h01@401@010@0101P000@100012P040@<00P4300812@030Bl0
000R00030@0003X00P4S00D110020@<000<100001@030@06004000400`42008110020@8000<10000
10020@03004101D01P4300040@000@l00`4500815`000`40000=00811@020@03004100810`020@80
00@100400P4]00008P000`40000j00818`020@8000<100000P020@0400410@P00`4300D10P020@@0
00<100002@020@8000D100000@0600030@0100L0104P00<14@000`40000200030@0001@01@401001
0@030@03004100<00`4/00008P000`40000j00030@0002800P4700030@0100812P020@@00P400`01
000C00810P000`4000030@@00P400`01000200060@0000400P4R00030@0000L000<1000010030@80
0P4G00810`050@0400410@D000<1000010030B<0000R00030@0003X000<100008`000`4000050003
0@0100812P020@@000<100005P001040004200030@0100<01@4200810`000`40000S00030@0000H0
0`4300030@0100810P000`40000M00810P000`400005008100@00@40104S00008P000`40000j0003
0@0002<000<100001`030BT000@100010P001@400@4000<100<00@400P020@800P4V00030@0000D0
00<100400`030@0400410Bd01P4300030@000200000200<11P000`40000500@10`030@D000<10000
>P000`40000]00<1;@020@801@4200030@0100@100<00@409P030@@00P4200D10`020Bh000<10000
0P4400030@0001l0000010010@020@l000D100000@0300030@00008000<10000>@000`40000_0003
0@0002`00P4200@11@050@03004002L00`4300040@0000<11@020CL000<100000P000`40000800<1
3`0000030040008000<100002@060@8000D100000@0400<1>@000`40001N00810`030@D00P420081
:P030@03004000@000<1000010000`40000f00<100<00@401P00104000020A0000000`0100020003
0@0000T000D100000@0300050@00004010000`40000i00030@0005h00P4;00810P020B`00`5500<1
00<00@000P4400@14P0000030040008000<100002P001040004400<11@000`40000i00030@0005h0
00<100002P000`40000`0081AP001`400@400@001P4D000000<00@000P000`40000:008100<00@00
0P001@40000100@000<10000>@000`40002N00030@0004L00P4300@100<00@004`00000300400080
00<100002`000`400@0300050@00004010000`40000i00030@000?l020000004004100813P020@<0
00D100000@0400030@0003T000<10000o`0800000P030@l00P4400<11@000`40000i00030@000?l0
2000028000<10000>@000`40003o00P0000R0081>P000`40003o00P0000R00030@0003T000<10000
o`0800008P000`40000h00030@000?l02@00028000<10000>0000`40003o00T0000R00030@0003P0
00<10000o`0900008P000`40000h00030@000?l02@00028000<10000>0000`40003o00T0000R0003
0@0003P000<10000o`0900008P020CT000<10000o`0900008P000`40000h00030@000?l02@000280
00<10000>0000`40003o00T0000R00030@0003P000<10000o`0900008P000`40000`00D10`040@D0
0`4/00@10`030@D00`4/00<110030@D00`4[00<11@030@D00`4W00D110030@D00`4500<10P000280
00<10000<0001@40000100<00P400`010@03008100<00@40;@00104000020@03004100<00P400`01
0@0[00040@000@800P400`010@03008100<00@40:@001@40000100<00P400`010@03008100<00@40
:0000`400003008100<00@400`020@03004100<00P400`010@0100008P000`40000a00030@0000<0
0P4300050@0000400`000`40000W00H10P001@40000100<000D100000@0Z00050@0000400P001@40
000100<000D100000@0Y00050@0000400`001@40000100<000D100000@0X00030@0000<000D10000
0@0300050@0000400`001@4000010040000R0081<`000`40000200810`001@40000100<000<10000
9`001@40000100<000D100000@0300050@000040:P020@8000@100010`001@40000100<000<10000
9`001@40000100<000D100000@0300050@000040:0000`40000300050@0000400`001@40000100<0
00D100000@0100008P000`40000c00040@0000810`001@40000100<000<10000:000104000430005
0@0000400`001@40000102X00P4200040@000@<000D100000@0300030@0002P00`4400050@000040
0`001@40000102P000<100000`001@40000100<000D100000@0300050@0000400@00028000<10000
=0000`4000020@<000D100000@0300030@0002P00P400`01000200050@0000400`001@40000102X0
00@100410`001@40000100<000D100000@0Y00050@0000400`001@40000100<000D100000@0X0003
0@0000<000D100000@0300050@0000400`001@4000010040000R00030@00030000D100000@020081
0`001@40000100<000<10000:@000`400@0300050@0000400`001@40000102X00P4500050@000040
0`001@40000102T000D100000@0300050@0000400`001@40000102P000<100000`001@40000100<0
00D100000@0300050@0000400@00028000<10000<0000`4000020@800`400`010@03008100<00@40
;0020@<00P400`010@03008100<00@40:`000`400003008100<00@400`020@03004102T000D10000
0@03008100<00@400`020@03004102H000<100401@020@03004100<00P400`010@03008100<00@40
0@00028000<10000<@030@<000<100400P4500<1;@020@@00`4500<1;@030@<00`4500<1:`030@D0
0`4500<1:0020@H00`4500<11@030@80000R00030@0003L000<10000o`0:00008P000`40000g0003
0@000?l02P0001X0o`570@P0000R00030@0000`000<100003@000`40000<00030@0000T000D10000
0@0>00030@0000d000<1000030000`40000=00030@0000`000<100003@000`40000<00030@0000d0
00<1000030000`40000=00030@0000`000<100003@000`40000<00030@0000d000<1000030000`40
000=00030@0000d0000R00030@0003L000D100000@0m00030@0003/000<10000>`000`40000k0003
0@0000d0000R00030@0003L000<10000o`0:00008P000`40000g00030@000?l02P00028000<10000
=`000`40003o00X0000R00030@0003L000<10000o`0:00008P000`40000f00030@000?l02`000280
0P4g00030@000?l02`00028000<10000=P000`40003o00/0000R00030@0003H000<10000o`0;0000
8P000`40000e0081o`0=00008P000`40000e0081o`0=00008P000`40000d0081o`0>00008P000`40
000d0081o`0>00008P020CD00P7o00h0000R00030@0003@00P7o00h0000R00030@0003@00P7o00h0
000R00030@0003@000<10000o`0=00008P000`40000d00030@000?l03@00028000<10000<`000`40
003o00h0000R00030@0003<000<10000o`0>00008P000`40000c00030@000?l03P0002800P4d0003
0@000?l03P00028000<10000;P000`40000200030@000?l03P00028000<10000;P000`4000020003
0@000?l03P00028000<10000;P020@<000<10000o`0>00008P000`40000^00810`000`40003o00h0
000R00030@0002h00P4300030@000?l03P0000800`4600030@0000@0104300D110000`40000^0081
0`000`40003o00h0000010010@020@/000@100000P4200050@00004010000`40000^00810`000`40
003o00h000000`01000200030@0000h000D100000@0700<1:P020@800P420081o`0@000000<00@00
0P000`40000>00030@00008000<1000010000`40000Z00810P020@800P7o010000000`0100020003
0@0000X000<100000P4500030@0000<000<100009`000`4000020@800P420081o`0@000000<00@00
0P000`40000:00@11`000`40000200030@0002L000<100000P4200810P000`40003o00l000000`01
000200030@0000X000<1000010001@40000100@000<100009`000`4000060@8000<10000o`0?0000
00@00@400P4<00030@0000@000<100000P4400030@0002L00P400`010@040@0300410?l04@000080
0`4=00@110030@D000<100009`020@03004100@100<00@40o`0A00008P020BP010400`010@020@03
00410?l04@00028000<100009`040@030041008100<00@40o`0A00008P000`40000W00@100D00@41
00030Ol04@00028000<100009`040@800P400`010@3o0180000R00030@0002L0104200030@000081
o`0B00008P000`40000W00@10P000`4000020Ol04P00028000<100009`040@8000<100000P7o0180
000R00030@0002L010450081o`0B00008P020BL01@450081o`0B00008P000`40000V00<100<00@00
o`0H00008P000`40000V00<100<00@00o`0H00008P000`40000V00030@010?l06`00028000<10000
9P000`40003o01/0000R00030@0002H000<10000o`0K00008P000`40000V00030@000?l06`000280
00<100009@020Ol07@0002800P4V0081o`0M00008P000`40000U0081o`0M00008P000`40000U0003
0@000?l07000028000<100009@000`40003o01`0000R00030@0002D000<10000o`0L00000P030@H0
00<1000010040@D0104300030@0002D000<10000o`0L000000@00@400P4;00040@0000811P000`40
000200030@0002D000<10000o`0L000000<00@000P000`40000>00030@0000H10`030BD000<10000
o`0L000000<00@000P000`40000>00040@000@<000<100000P000`40000U00030@000?l070000003
0040008000<100002P000`4000020@<000@1000110000`40000U00030@000?l07000000300400080
00<100002P040@@00P400`01000300030@0002D000<10000o`0L000000<00@000P000`40000:0003
0@0000H000<1004010000`40000U00030@000?l0700000040041008130000`400007008110000`40
000U00030@000?l0700000800`4=00@11P020@@000<100009@000`40003o01`0000R00030@0002D0
00<10000o`0L00008P020BD000<10000o`0M00008P000`40000T00030@000?l07@00028000<10000
90000`40003o01d0000R00030@0002@000<10000o`0M00008P000`40000T00030@000?l07@000280
00<100007@000`40000400030@000?l07@00028000<100007@000`40000400030@000?l07@000280
00<100007@000`40000400030@000?l07@0002800P4N00030@0000@000<10000o`0M00008P000`40
000M00030@0000@000<10000o`0M00008P000`40000L00811P000`40003o01d0000R00030@0001`0
0P450081o`0O00008P000`40000L00811@020Ol07`00028000<1000070020@D00P7o01l0000R0003
0@0001`00P450081o`0O00008P020Ad00P4300040@010Ol07`00028000<1000070030@8000@10041
o`0O00008P000`40000L00<10P000`400@3o0200000R00030@0001`00`4200<1o`0P00008P000`40
000L00<10P030Ol08000028000<1000070030@800`7o0200000200<11P000`40000400@11@030@@0
00<1000070000`400@0200<1o`0P000000@00@400P4;00040@00008110001040004300030@0001`0
00<100400P030Ol0800000030040008000<100003P001@40000100<000@100000`4L00030@010080
0`7o020000000`01000200030@0000h000@100000P4200050@0000407P000`400@0200<1o`0P0000
00<00@000P000`40000:00030@0000810`020@8000D100000@0N00030@0100800`7o020000000`01
000200030@0000X0104400040@010@@000<1000070001@40004000<1o`0P000000<00@000P000`40
000:00030@0000D00P4600030@0001`000D1000100030Ol0800000040041008130000`4000060003
0@0000@000<1000070000`4000020@0300410?l0800000800`4=00@11P030@<000<1000070000`40
00020@0300400?l08000028000<1000070000`4000020@0300400?l0800002800P4M00030@000081
00<00@00o`0P00008P000`40000L00030@00008100<00@00o`0P00008P000`40000L00030@000081
o`0S00008P000`40000K00810P020Ol08`00028000<100006`020@800P7o02<0000R00030@0001/0
0P420081o`0S00008P000`40000K00810P020Ol08`00028000<100006`020@800P7o02<0000R0081
6`030@800P7o02<0000R00030@0001X00`4200030@000?l08P00028000<100006P030@8000<10000
o`0R00008P000`40000J00<1o`0W00008P000`40000J00<1o`0W00008P000`40000J0081o`0X0000
8P000`40000J0081o`0X00008P020A/00P7o02P0000R00030@0001X00P7o02P0000R00030@0001X0
00<10000o`0W00008P000`40000J00030@000?l09`00028000<1000060000`400@3o02T0000R0003
0@0001P00`7o02T0000200<11P000`40000400@110030@D000<1000060030Ol0:@00000400410081
2`00104000020@8000D100000@0400030@0001P00P7o02X000000`01000200030@0000h000@10001
0`000`40000200<160020Ol0:P0000030040008000<100003P001040004300030@00008000<10000
60020Ol0:P0000030040008000<100002P000`4000020@<00`4500030@0001P00P7o02X000000`01
000200030@0000X0104300050@00004010000`40000H0081o`0Z000000<00@000P000`40000:0003
0@0000@000D100000@0400030@0001P00P7o02X0000010010@020@`000<1000010001@40000100@0
00<1000060020Ol0:P0000800`4=00@110030@D000<1000060020Ol0:P000?l0J@000?l0J@000?l0
J@000?l0J@000?l0J@000?l0J@000?l0J@000?l0?`040BH0003o040000<100009P000:L000<10000
U@000`40000V0000;P040@8000<100000`400`010@030@040041008100<00@000`4400@10`060@P0
104300H11`030@0400410@<0104200L10`030@04004100D12@000`40000:00@10P000`4000030@03
004100<100@00@400P400`0100030@@0104300H120040@<01P4900@110040@<01@4200<100@00@41
20060@801@4300D10`000`400@020@<01@4L0000;`00104000020@8000<100000P4200040@000@80
00<100000P4200810P020@<000<100000P000`40000800810P020@@000<100002@001@40000100<0
0P4300030@0100@000@100000P4200810P020@<000<100001P000`40000;00040@0000810P000`40
00020@8000@100010P000`4000020@800P4200810`000`40000200030@0000P00P42008110000`40
000900810`000`4000020@800P4400030@0000<000D100000@0;00030@0000D000<100001@000`40
000300810P020@800P4300030@0001T0000_00060@000@010P001@400001008000@100010P001040
004400040@000@T000<1000020000`40000200030@00008000<100002@001@40000100<000<10000
1@050@<000<100000P001040004400030@0000L000<100002P001P4000400@8000D100000@020004
0@000@8000@1000110001040004900030@0000P000<100000P000`40000200030@0000T000<10000
1@000`40000200030@00008000<100000`001@40000100/000<100001@000`40000500030@0000<0
00<100000P001040004400030@0001T0000_00060@000@010P001@400001008000@100010P001040
004400030@0000H110000`40000800030@00008000<100000P000`40000900050@0000400`060@H0
00D100000@0400030@0000D12P000`40000:00060@000@010P001@400001008000@100010P001040
004400030@0000H110000`40000800030@00008000<100000P000`40000900030@00008000@10001
10000`40000200030@0000<000D100000@0;00030@0000D000<100001@000`40000300030@000080
00<100001@4L0000;`000`400@0200040@000@<000@100010P001040004200810P020@8000<10000
0P000`40000200810P000`40000400810P020@@000<100002@001@40000100<000<100000P001040
004300040@0000810P020@8000<100000P000`40000800030@0000T000<100400P00104000430004
0@000@8000@100010P020@800P4200030@00008000<100000P020@8000<1000010020@800P440003
0@0000T00P4200810P020@800P4400030@0000<000D100000@0;00030@0000D000<100001@000`40
000300810P020@8000<100000P000`40000I0000;`000`400@0200050@010@000`400`010@060@80
00<100400P4400@10`030@0400410@L0104300H120040@D0104400<11@030@03004000801@4:0003
0@0000T000<100400P001@400@4000<100<00@401P4200030@01008110040@<00`4010010@4700@1
0`060@T00`400`01000200@10`030@@01@4:00H110000`40000300<110020@0400410@@01@4K0000
;`020@<000<1000040000`40000T00030@0000T000<100006P000`40000A00030@0000P00P430003
0@00010000<1000090000`40000a00030@0000D000<10000=00002h00P4200@14@000`40000T0003
0@0000T000<100006P000`40000A00030@0000L00P4200@14@000`40000T00030@0001/000<10000
4`000`40000500030@0000D000<10000;0000001\
\>"],
  ImageRangeCache->{{{108, 395}, {602.688, 425.75}} -> {-545.933, 0.79849, 
  4.03639, 0.00081378}}]
}, Open  ]],

Cell["\<\
Another problem concerns how people frame a problem concerning \
probabilities; something that we will later refer to as setting the context \
or being clear about the set of experiments we have in mind.\
\>", "Text"],

Cell["\<\
Example: I tell you that the smiths have two children and one of \
them is a girl.\
\>", "Text"],

Cell["\<\
Question: What is the probability that the other child is a girl \
too?\
\>", "Text"],

Cell[TextData[{
  "Answer: ",
  StyleBox["1/2",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " - Wrong!  The confusion comes from thinking that we're asking for the \
probability that \"the other child\", is a girl."
}], "Text"],

Cell[TextData[StyleBox["Events = {BG,GB,BB,GG} ",
  FontFamily->"Courier",
  FontWeight->"Bold"]], "Text"],

Cell["\<\
We'll assume that these events are equally likely and mutually \
exclusive so that\
\>", "Text"],

Cell[TextData[StyleBox["Pr(GB) = Pr(BG) = Pr(GG) = Pr(BB) = 1/4",
  FontFamily->"Courier",
  FontWeight->"Bold"]], "Text"],

Cell["\<\
Think of the problem in terms of experiments.  For the experiment \
corresponding to asking a particular set of parents known to have two \
children what the sexes of the chidren are the obove events are equally \
likely.  However, we've effectively restricted our attention to experiments \
in which one of the two children is known to be female.  Now we have\
\>", 
  "Text"],

Cell[TextData[{
  StyleBox["Events = {BG,GB,BB,GG}",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[" ",
    FontFamily->"Courier"],
  "and"
}], "Text"],

Cell[TextData[StyleBox["Pr(GB) = Pr(BG) = Pr(GG) = 1/3",
  FontFamily->"Courier",
  FontWeight->"Bold"]], "Text"],

Cell[TextData[{
  "Answer: ",
  StyleBox["1/3",
    FontFamily->"Courier",
    FontWeight->"Bold"]
}], "Text"],

Cell["\<\
The trick is to carefully determine the context, or set of \
experiments, such that it is appropriate to apply the laws of probability.  \
Probability and statistics are largely about careful counting - a field of \
mathematics called combinatorics.  It can be tricky as well.  Here's an \
example.\
\>", "Text"],

Cell[TextData[{
  "Question: How large a party do you have to assemble before the probability \
of their being two people with the same birthday is larger than ",
  StyleBox["1/2",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell["Answer: Pretty small - twenty two will suffice", "Text"],

Cell[TextData[{
  "The probability that there are n (for",
  StyleBox[" n < 365",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ") people that don't have the same birthday is"
}], "Text"],

Cell["ProbNoMatchAmong[n_] := Product[(365-i)/365,{i,1,n}]", "Input"],

Cell["It so happens that ", "Text"],

Cell["ProbMatchAmong[n_] := 1 - ProbNoMatchAmong[n]", "Input"],

Cell["and you can check for yourself but ", "Text"],

Cell[CellGroupData[{

Cell["And[ProbMatchAmong[21] < 1/2, ProbMatchAmong[22] > 1/2]", "Input"],

Cell[BoxData[
    \(True\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
ListPlot[Table[ProbMatchAmong[n],{n,1,30}],
        {PlotJoined -> True,
         PlotLabel -> \"Pr( birthday match among n people )\"}] ;\
\>", 
  "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.031746 0.0147151 0.805805 [
[.18254 .00222 -3 -9 ]
[.18254 .00222 3 0 ]
[.34127 .00222 -6 -9 ]
[.34127 .00222 6 0 ]
[.5 .00222 -6 -9 ]
[.5 .00222 6 0 ]
[.65873 .00222 -6 -9 ]
[.65873 .00222 6 0 ]
[.81746 .00222 -6 -9 ]
[.81746 .00222 6 0 ]
[.97619 .00222 -6 -9 ]
[.97619 .00222 6 0 ]
[.01131 .0953 -18 -4.5 ]
[.01131 .0953 0 4.5 ]
[.01131 .17588 -18 -4.5 ]
[.01131 .17588 0 4.5 ]
[.01131 .25646 -18 -4.5 ]
[.01131 .25646 0 4.5 ]
[.01131 .33704 -18 -4.5 ]
[.01131 .33704 0 4.5 ]
[.01131 .41762 -18 -4.5 ]
[.01131 .41762 0 4.5 ]
[.01131 .4982 -18 -4.5 ]
[.01131 .4982 0 4.5 ]
[.01131 .57878 -18 -4.5 ]
[.01131 .57878 0 4.5 ]
[.5 .63053 -107 0 ]
[.5 .63053 107 10.375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.18254 .01472 m
.18254 .02097 L
s
[(5)] .18254 .00222 0 1 Mshowa
.34127 .01472 m
.34127 .02097 L
s
[(10)] .34127 .00222 0 1 Mshowa
.5 .01472 m
.5 .02097 L
s
[(15)] .5 .00222 0 1 Mshowa
.65873 .01472 m
.65873 .02097 L
s
[(20)] .65873 .00222 0 1 Mshowa
.81746 .01472 m
.81746 .02097 L
s
[(25)] .81746 .00222 0 1 Mshowa
.97619 .01472 m
.97619 .02097 L
s
[(30)] .97619 .00222 0 1 Mshowa
.125 Mabswid
.05556 .01472 m
.05556 .01847 L
s
.0873 .01472 m
.0873 .01847 L
s
.11905 .01472 m
.11905 .01847 L
s
.15079 .01472 m
.15079 .01847 L
s
.21429 .01472 m
.21429 .01847 L
s
.24603 .01472 m
.24603 .01847 L
s
.27778 .01472 m
.27778 .01847 L
s
.30952 .01472 m
.30952 .01847 L
s
.37302 .01472 m
.37302 .01847 L
s
.40476 .01472 m
.40476 .01847 L
s
.43651 .01472 m
.43651 .01847 L
s
.46825 .01472 m
.46825 .01847 L
s
.53175 .01472 m
.53175 .01847 L
s
.56349 .01472 m
.56349 .01847 L
s
.59524 .01472 m
.59524 .01847 L
s
.62698 .01472 m
.62698 .01847 L
s
.69048 .01472 m
.69048 .01847 L
s
.72222 .01472 m
.72222 .01847 L
s
.75397 .01472 m
.75397 .01847 L
s
.78571 .01472 m
.78571 .01847 L
s
.84921 .01472 m
.84921 .01847 L
s
.88095 .01472 m
.88095 .01847 L
s
.9127 .01472 m
.9127 .01847 L
s
.94444 .01472 m
.94444 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .0953 m
.03006 .0953 L
s
[(0.1)] .01131 .0953 1 0 Mshowa
.02381 .17588 m
.03006 .17588 L
s
[(0.2)] .01131 .17588 1 0 Mshowa
.02381 .25646 m
.03006 .25646 L
s
[(0.3)] .01131 .25646 1 0 Mshowa
.02381 .33704 m
.03006 .33704 L
s
[(0.4)] .01131 .33704 1 0 Mshowa
.02381 .41762 m
.03006 .41762 L
s
[(0.5)] .01131 .41762 1 0 Mshowa
.02381 .4982 m
.03006 .4982 L
s
[(0.6)] .01131 .4982 1 0 Mshowa
.02381 .57878 m
.03006 .57878 L
s
[(0.7)] .01131 .57878 1 0 Mshowa
.125 Mabswid
.02381 .03083 m
.02756 .03083 L
s
.02381 .04695 m
.02756 .04695 L
s
.02381 .06306 m
.02756 .06306 L
s
.02381 .07918 m
.02756 .07918 L
s
.02381 .11141 m
.02756 .11141 L
s
.02381 .12753 m
.02756 .12753 L
s
.02381 .14364 m
.02756 .14364 L
s
.02381 .15976 m
.02756 .15976 L
s
.02381 .19199 m
.02756 .19199 L
s
.02381 .20811 m
.02756 .20811 L
s
.02381 .22422 m
.02756 .22422 L
s
.02381 .24034 m
.02756 .24034 L
s
.02381 .27257 m
.02756 .27257 L
s
.02381 .28869 m
.02756 .28869 L
s
.02381 .3048 m
.02756 .3048 L
s
.02381 .32092 m
.02756 .32092 L
s
.02381 .35315 m
.02756 .35315 L
s
.02381 .36927 m
.02756 .36927 L
s
.02381 .38539 m
.02756 .38539 L
s
.02381 .4015 m
.02756 .4015 L
s
.02381 .43373 m
.02756 .43373 L
s
.02381 .44985 m
.02756 .44985 L
s
.02381 .46597 m
.02756 .46597 L
s
.02381 .48208 m
.02756 .48208 L
s
.02381 .51431 m
.02756 .51431 L
s
.02381 .53043 m
.02756 .53043 L
s
.02381 .54655 m
.02756 .54655 L
s
.02381 .56266 m
.02756 .56266 L
s
.02381 .59489 m
.02756 .59489 L
s
.02381 .61101 m
.02756 .61101 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
gsave
.5 .63053 -168 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.375000 moveto
 558.000000 14.375000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.375000 lineto
closepath
clip newpath
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Pr) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
75.000000 11.562500 moveto
(H) show
87.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(birthday) show
141.000000 11.562500 moveto
(match) show
177.000000 11.562500 moveto
(among) show
213.000000 11.562500 moveto
(n) show
225.000000 11.562500 moveto
(people) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
267.000000 11.562500 moveto
(L) show
273.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Math2Mono
%%+ Courier
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.05556 .01692 m
.0873 .02133 L
.11905 .02789 L
.15079 .03658 L
.18254 .04732 L
.21429 .06003 L
.24603 .07461 L
.27778 .09096 L
.30952 .10895 L
.34127 .12845 L
.37302 .1493 L
.40476 .17137 L
.43651 .19449 L
.46825 .2185 L
.5 .24324 L
.53175 .26855 L
.56349 .29426 L
.59524 .32021 L
.62698 .34625 L
.65873 .37224 L
.69048 .39803 L
.72222 .4235 L
.75397 .44852 L
.78571 .47298 L
.81746 .49678 L
.84921 .51984 L
.88095 .54208 L
.9127 .56344 L
.94444 .58387 L
.97619 .60332 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40005X0000gPP10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdk34g_ojn_l004J2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_P00oiT0IP3o0;C^]:JV
YZf][Jo^kP20P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;002k0?mT0]d009T0V@00
gNk^k^[Zj]gMgK^k^jZZZXR8R7MgMeEEEDA4A14A4Oooc?o<VLbIIS?oVC>IIS=VIYWoo`2Ioc<0VLc<
oiUVVNoGk`00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2o
o`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no
?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo
?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo
?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0
_ol0onj2k_Z0L_lo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoO`0l0>o<knl1k`1]
0:2PX:[Xo`3XooooP;Co]4Bkooo;og?X`?nXOOnCOOno_onooooo?oooOooo_mWEc:0Pl=WIfJj^[YWo
09Wo<iWoIYWoVIWoc<`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aV
VLaVc<aVolbI0<bI<lbIVLbIc<bIolc<0<c<<lc<IP000?l0J@000?l0J@000?l0J@000?l0J@000?l0
J@0005L0104Y00D110030BH01@4400@19@050@@00`4V00D110040BD00`4500<14@0005H000@10000
0P4Z00030@0000<00P400`010@0W00030@0000<000@100000P4T00050@0000400`020@03004102D0
00D100000@0300040@0000818`000`4000020@<00P400`010@0@0000F`000`40000X00030@0000<0
00D100000@0W00030@0000P000<100008`000`40000400050@0000409P000`40000900030@0002D0
00D100000@0300030@0000h0001K00030@0002P000<100000`001@40000102L000<1000020000`40
000T00030@0000<000D100000@0W00030@0000P000<100009@001@40000100<000<100003P0005L0
00<100000P4Z00030@0000<000D100000@0W00030@0000@000<100000P4W00030@00008000D10000
0@0X00030@0000<000<100000P4U008110001@4000010100001G00@1:`000`40000300050@000040
9`000`40000400@1:@001@40000100<000<100009`000`40000200@1:0001@40000100<000<10000
3P0002L000<10000;@000`40000/00030@0000<000D100000@0W00030@0000@000<100009P001@40
000100<000D100000@0U00050@00004010000`40000Y00050@0000400`000`40000>00009`000`40
000]00030@0002X000<100401@020@03004102D000<100401P000`40000V00030@0000810`020@03
004102D000<100000P4400030@0002D000D100000@03008100<00@40400002L000<10000;@040BX0
0P4600<19`020@H0104V00<11@030BL00`4500@19@030@D00`4A00009`000`40003o03l0000P0?l1
>`4>00009`000`40000700H110000`40000700030@0000L000<100001`000`40000700030@0000L0
00<100001`000`40000700030@0000L000<100001`000`40000700030@0000L000<100001`000`40
000600030@0000L000<100001`000`40000700030@0000L000<100001`000`40000700030@0000L0
00<100001`000`40000700030@0000L000<100001`000`40000700030@0000L000<100001`000`40
000700030@0001<0000W00030@0000d01`4K00030@0002l000<10000;P000`40000_00030@0002l0
00<10000;`000`40000C00009`000`40000D00D1o`0V00009`000`40000I00@1o`0R00009`020Ah0
107o01h0000W00030@0002400`7o01/0000W00030@0002@00`7o01P0000W00030@0002L00P7o01H0
000W00030@0002T00`7o01<0000W0081;@020Ol04@0002L000<10000;P030Ol03P0002L000<10000
<@020Ol0300002L000<10000<`030Ol02@0002L000<10000=P020Ol01`0002L00P4i00<1o`040000
9`000`40000k0081o`0200009`000`40000m00<1oP0002L000<10000@0020O`0000W00030@000480
0`7i00009`000`4000150081m`0002L00P580081m@0002L000<10000B@020O<0000W00030@0004/0
0P7a00003`030@H000<100000`050@@000<10000C@000`40003^00003P020@03004100d000<10000
10000`40001>0081kP0000h000D100000@0=00030@0000@00`5@0081k00000h000D100000@0=0003
0@0000@000<10000DP000`40003Y00003P001@40000100d000<1000010000`40001C0081j@0000h0
00D100000@0=00030@0000@000<10000E@020NL0000>00050@0000403@000`40000400030@0005L0
00<10000i00000h00P400`010@0;00030@0100H00P5I0081i00000l00`4=00811P000`40001J0081
hP0002L000<10000G0000`40003O00009`000`40001M0081g`0002L000<10000G`020Md0000W0081
HP000`40003J00009`000`40001R0081fP0002L000<10000I0000`40003G00009`000`40001U0081
e`0002L000<10000I`000`40003D00009`020FT000<10000d`0002L000<10000J@020M<0000W0003
0@0006/000<10000d00002L000<10000K0020M00000W00030@0006h00P7>00009`020G4000<10000
b`0002L000<10000L@020L/0000W00030@0007<00P7900003`030@H000<100000`050@@000<10000
M@000`40003600003P020@03004100/000D100000@0400030@0007H000<10000a@0000h000D10000
0@0<00030@0000D00`5g0081a@0000h000D100000@0=00030@0000@000<10000N@000`4000320000
3P001@40000100h000<100000`000`40001j00030@000<40000>00050@0000403`000`4000020003
0@0007/000<10000`00000h000D100000@0;00050@00004010000`40001l0081`00000h00P400`01
0@0;00030@00008110020Gl000<10000_@0000l00`4=00<11@000`40001o00030@000;`0000W0003
0@0008000P6l00009`000`40002200030@000;T0000W00030@0008<00P6i00009`020HH000<10000
]P0002L000<10000QP000`40002e00009`000`4000270081]@0002L000<10000R@000`40002b0000
9`000`40002:00030@000;40000W0081S0000`40002`00009`000`40002<00030@000:l0000W0003
0@0008d00P6_00009`000`40002?00030@000:`0000W00030@00090000<10000Z`0002L00P6B0003
0@000:X0000W00030@00098000<10000Z@0002L000<10000T`000`40002X00003`030@H000<10000
10030@D000<10000U0020JP0000>008100<00@402`000`4000020@@000<10000UP000`40002U0000
3P001@40000100l000<100000P030IL000<10000Y00000h000D100000@0?00030@00008000<10000
V0000`40002S00003P001@40000100d00P4500030@0009T00P6S00003P001@40000100l000<10000
0P000`40002K00030@000:00000>00050@0000403`000`40000200030@0009`000<10000W`0000h0
0P400`010@0;00050@00004010020Ih000<10000WP0000l00`4=00<11@000`40002N0081WP0002L0
00<10000X0000`40002K00009`000`40002Q00030@0009X0000W00030@000:8000<10000V@0002L0
0P6T0081V@0002L000<10000Y@000`40002F00009`000`40002V00030@0009D0000W00030@000:L0
00<10000U00002L000<10000Z0020I@0000W0081Z`000`40002A00009`000`40002[00030@000900
000W00030@000:`000<10000S`0002L000<10000[@020Hl0000W00030@000:l000<10000S00002L0
0P6a00030@0008/0000W00030@000;4000<10000RP0002L000<10000/P020HX0000?00<11P000`40
000500@10`000`40002d00030@0008L0000>008100<00@403`000`40000200030@000;D000<10000
QP0000h000D100000@0;00H10`030KH000<10000Q@0000h000D100000@0;00050@00004010000`40
002g0081Q@0000h000D100000@0<00040@000@@000<10000^@000`40002200003P001@40000100`0
0P400`01000300030@000;X000<10000P@0000h000D100000@0=00030@0100@000<10000^`000`40
002000003P020@03004100h00P440081_@020H00000?00<13`020@@000<10000_P000`40001m0000
9`000`40002o00030@0007`0000W00030@000<0000<10000N`0002L000<10000`@020G/0000W0003
0@000<<000<10000N00002L00P7500030@0007L0000W00030@000<D000<10000MP0002L000<10000
aP000`40001e00009`000`40003700030@0007@0000W00030@000<P00P5d00009`020L/000<10000
L@0002L000<10000b`000`40001`00009`000`40003<00030@0006l0000W00030@000<d000<10000
KP0002L000<10000cP000`40001]00009`020M0000<10000K00002L000<10000d0020F`0000W0003
0@000=8000<10000J@0000l00`4600030@0000@0104400030@000=<000<10000J00000h00P400`01
0@0;00040@0000810`000`40003D00030@0006L0000>00050@0000404000104000030MD00P5W0000
3P001@400001010000D100000@3I00030@0006@0000>00050@00004030000`4000020@<000<10000
f0000`40001S00003P001@40000100`0104400030@000=T00P5S00003P001@40000100`000<10000
1@000`40003K00030@000600000>008100<00@4030000`4000050081g@020F00000?00<13@040@@0
00<10000gP000`40001M00009`000`40003O00030@0005`0000W00030@000>000P5L00009`000`40
003R00030@0005T0000W0081i0000`40001H00009`000`40003T0081F00002L000<10000iP000`40
001E00009`000`40003W00030@0005@0000W00030@000>P000<10000D`0002L00P7Z0081D`0002L0
00<10000j`000`40001@00009`000`40003/00030@0004l0000W00030@000>d000<10000CP0002L0
00<10000kP020Dh0000W0081l@000`40001;00009`000`40003a00030@0004X0000W00030@000?80
00<10000B@0000l00`4600030@0000D00`4400030@000?<00P5900003P020@03004100d000@10001
0`000`40003e00030@0004H0000>00050@00004030001@40000100<00`7f00030@0004D0000>0005
0@00004030020@8000D100000@3i0081A@0000h000D100000@0<00810P001@4000010?/000<10000
@P0000h000D100000@0<00040@010@@000<10000nP020D80000>00050@00004030020@H000<10000
o0000`40000o00003P020@03004100d000<1000010020Oh000<10000?P0000l00`4?00<10`000`40
003n0081?P0002L000<10000o`0100030@0003/0000W00030@000?l00P020C/0000W00030@000?l0
10000`40000h00009`020Ol01P020CP0000W00030@000?l01`000`40000e00009`000`40003o00P0
00<10000=00002L000<10000o`090081=00002L000<10000o`0;00030@000340000W0081o`0=0081
<@0002L000<10000o`0>0081;`0002L000<10000o`0@00030@0002`0000W00030@000?l04@020B`0
000W00030@000?l04`020BX0000W0081o`0F00030@0002L0000W00030@000?l05P020BL0000W0003
0@000?l060000`40000T00003`030@H000<100001@000`40000400030@000?l06@020B@0000>0081
00<00@403@000`40000400030@000?l06`000`40000Q00003P001@40000100h000<100000`030Ol0
70000`40000P00003P001@40000100h000<100000`000`40003o01d00P4P00003P001@40000100h0
00<100000`000`40003o01l000<100007@0000h000D100000@0>00030@0000<000<10000o`0P0081
7@0000h000D100000@0>00030@0000<000<10000o`0R00816`0000h00P400`010@0;00050@000040
10020Ol09@000`40000H00003`030@`01@4400030@000?l09@020AP0000W00030@000?l09`020AH0
000W00030@000?l0:@000`40000C00009`000`40003o03l0000W0081o`1000009`000`40003o03l0
000W00030@000?l0?`0002L000<10000o`0o0000o`1Y0000o`1Y0000o`1Y0000o`1Y0000o`1Y0000
o`1Y0000T0040EH0104F00@150040DP0001<00030@0004<000<10000F@000`40000D00030@0001D0
00<100006P000`40000[0000B`000`4000140081FP000`40000D00030@0001D000<100006`000`40
000Z0000>P040@@01P4300030@0000X000<100400P4300D10`060@<0104200<100@00@410`030@03
004100L11@000`40000400<100@00@402@4300@110040@800`4010010@4600X100@00@400P4200@1
0P030@0400410@<00`400`01000500<100@00@412@000`400@020@@0104400@10`000`400@020@<0
1@4400@12@000`40000Z0000>`000`40000600030@0000<000<100002`020@800P4400030@0000D0
00<1000010001@400001008000D100000@0300810P020@03004000<000<100000`020@L000@10001
0P000`400@0400030@00008000D100000@0200810`001040004300030@0000D000<100000P001040
004200040@000@800P4200810P001@40000100<00P4200811`001@40000100X00P4200810P020@<0
00<100000P4200810P020@800P4400030@0000<00P4300030@0000L000<10000:@0003/000<10000
1P000`40000300030@0000/000<100000P000`40000200030@0000D000<1000010000`4000040005
0@0000400`000`40000200030@0000D11@000`400@0600040@000@8000<100001@4400030@0000@0
00<100001@001@40000100P01@4200040@000@8000@1000110001040004300050@00004010000`40
000500050@0000402P000`40000200040@000@L000<100000P001040004400030@00008000<10000
0`000`40000<00030@0002T0000k00@11@000`40000300030@0000/000<100000P000`4000020003
0@0000D000<1000010000`40000400050@0000400`000`40000200030@0000@000<100000P001040
004600040@000@8000<1000010000`40000200030@0000@000<100000P001040004300030@0000X0
00@100010P001040004200030@00008000@100010`001@40000100@000<100001@001@40000100X0
00<100000P000`4000060@8000<100000P001040004400030@00008000<100000`060@T000<10000
:@0003/000D100000@0400810P000`400@0=00810P020@@000<100001@020@8000@100011P001@40
000100<00P4200810P001@40000100@000D100000@0500040@000@8000@100010`000`4000020003
0@0000@00P4200810P001@40000100P000D100000@0200040@000@8000<100000P4200810P001@40
000100<00P4200811`001@40000100X00P4200810P000`40000200030@0000810P020@800P420081
10000`40000300030@00008000<100001`000`40000Y0000>`001@40000100800`4010010@420003
0@0000X000<100400P4300<11@030@03004100L10`040@D00`400`01000200<110020@800P4400P1
0`030@<01P4400<100@00@00104:00<10P080@<0104200D11@030@03004100D01@4:008100@00@41
10040@@01042008100@00@411@000`40000400@12@000`40000Z0000>`001@40000100/000<10000
2P000`40000F00030@0000@000<100002P000`40000U00030@0000`000<10000J0000`40000A0003
0@0002X00000\
\>"],
  ImageRangeCache->{{{108, 395}, {306.812, 129.875}} -> {-17.5423, 0.58267, 
  0.125614, 0.00494875}}]
}, Open  ]],

Cell["\<\
It's a bit harder to have a party with enough people so that with \
better than chance there is someone at the party with have YOUR \
birthday.\
\>", "Text"],

Cell["PrNoMatchYourDay[n_] := (364/365)^n", "Input"],

Cell["PrMatchYourDay[n_] := 1 - PrNoMatchYourDay[n]", "Input"],

Cell[CellGroupData[{

Cell["\<\
ListPlot[Table[PrMatchYourDay[n],{n,1,365}],
        {PlotJoined -> True,
         PlotLabel -> \"Pr( someone with your birthday among n people )\"}] ;\
\
\>", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.00260926 0.0147151 0.930415 [
[.15427 .00222 -6 -9 ]
[.15427 .00222 6 0 ]
[.28474 .00222 -9 -9 ]
[.28474 .00222 9 0 ]
[.4152 .00222 -9 -9 ]
[.4152 .00222 9 0 ]
[.54566 .00222 -9 -9 ]
[.54566 .00222 9 0 ]
[.67613 .00222 -9 -9 ]
[.67613 .00222 9 0 ]
[.80659 .00222 -9 -9 ]
[.80659 .00222 9 0 ]
[.93705 .00222 -9 -9 ]
[.93705 .00222 9 0 ]
[.01131 .10776 -18 -4.5 ]
[.01131 .10776 0 4.5 ]
[.01131 .2008 -18 -4.5 ]
[.01131 .2008 0 4.5 ]
[.01131 .29384 -18 -4.5 ]
[.01131 .29384 0 4.5 ]
[.01131 .38688 -18 -4.5 ]
[.01131 .38688 0 4.5 ]
[.01131 .47992 -18 -4.5 ]
[.01131 .47992 0 4.5 ]
[.01131 .57296 -18 -4.5 ]
[.01131 .57296 0 4.5 ]
[.5 .63053 -143.031 0 ]
[.5 .63053 143.031 10.375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.15427 .01472 m
.15427 .02097 L
s
[(50)] .15427 .00222 0 1 Mshowa
.28474 .01472 m
.28474 .02097 L
s
[(100)] .28474 .00222 0 1 Mshowa
.4152 .01472 m
.4152 .02097 L
s
[(150)] .4152 .00222 0 1 Mshowa
.54566 .01472 m
.54566 .02097 L
s
[(200)] .54566 .00222 0 1 Mshowa
.67613 .01472 m
.67613 .02097 L
s
[(250)] .67613 .00222 0 1 Mshowa
.80659 .01472 m
.80659 .02097 L
s
[(300)] .80659 .00222 0 1 Mshowa
.93705 .01472 m
.93705 .02097 L
s
[(350)] .93705 .00222 0 1 Mshowa
.125 Mabswid
.0499 .01472 m
.0499 .01847 L
s
.07599 .01472 m
.07599 .01847 L
s
.10209 .01472 m
.10209 .01847 L
s
.12818 .01472 m
.12818 .01847 L
s
.18037 .01472 m
.18037 .01847 L
s
.20646 .01472 m
.20646 .01847 L
s
.23255 .01472 m
.23255 .01847 L
s
.25864 .01472 m
.25864 .01847 L
s
.31083 .01472 m
.31083 .01847 L
s
.33692 .01472 m
.33692 .01847 L
s
.36301 .01472 m
.36301 .01847 L
s
.38911 .01472 m
.38911 .01847 L
s
.44129 .01472 m
.44129 .01847 L
s
.46738 .01472 m
.46738 .01847 L
s
.49348 .01472 m
.49348 .01847 L
s
.51957 .01472 m
.51957 .01847 L
s
.57175 .01472 m
.57175 .01847 L
s
.59785 .01472 m
.59785 .01847 L
s
.62394 .01472 m
.62394 .01847 L
s
.65003 .01472 m
.65003 .01847 L
s
.70222 .01472 m
.70222 .01847 L
s
.72831 .01472 m
.72831 .01847 L
s
.7544 .01472 m
.7544 .01847 L
s
.7805 .01472 m
.7805 .01847 L
s
.83268 .01472 m
.83268 .01847 L
s
.85877 .01472 m
.85877 .01847 L
s
.88487 .01472 m
.88487 .01847 L
s
.91096 .01472 m
.91096 .01847 L
s
.96314 .01472 m
.96314 .01847 L
s
.98924 .01472 m
.98924 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .10776 m
.03006 .10776 L
s
[(0.1)] .01131 .10776 1 0 Mshowa
.02381 .2008 m
.03006 .2008 L
s
[(0.2)] .01131 .2008 1 0 Mshowa
.02381 .29384 m
.03006 .29384 L
s
[(0.3)] .01131 .29384 1 0 Mshowa
.02381 .38688 m
.03006 .38688 L
s
[(0.4)] .01131 .38688 1 0 Mshowa
.02381 .47992 m
.03006 .47992 L
s
[(0.5)] .01131 .47992 1 0 Mshowa
.02381 .57296 m
.03006 .57296 L
s
[(0.6)] .01131 .57296 1 0 Mshowa
.125 Mabswid
.02381 .03332 m
.02756 .03332 L
s
.02381 .05193 m
.02756 .05193 L
s
.02381 .07054 m
.02756 .07054 L
s
.02381 .08915 m
.02756 .08915 L
s
.02381 .12636 m
.02756 .12636 L
s
.02381 .14497 m
.02756 .14497 L
s
.02381 .16358 m
.02756 .16358 L
s
.02381 .18219 m
.02756 .18219 L
s
.02381 .21941 m
.02756 .21941 L
s
.02381 .23801 m
.02756 .23801 L
s
.02381 .25662 m
.02756 .25662 L
s
.02381 .27523 m
.02756 .27523 L
s
.02381 .31245 m
.02756 .31245 L
s
.02381 .33106 m
.02756 .33106 L
s
.02381 .34966 m
.02756 .34966 L
s
.02381 .36827 m
.02756 .36827 L
s
.02381 .40549 m
.02756 .40549 L
s
.02381 .4241 m
.02756 .4241 L
s
.02381 .44271 m
.02756 .44271 L
s
.02381 .46131 m
.02756 .46131 L
s
.02381 .49853 m
.02756 .49853 L
s
.02381 .51714 m
.02756 .51714 L
s
.02381 .53575 m
.02756 .53575 L
s
.02381 .55436 m
.02756 .55436 L
s
.02381 .59157 m
.02756 .59157 L
s
.02381 .61018 m
.02756 .61018 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
gsave
.5 .63053 -204.031 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.375000 moveto
 558.000000 14.375000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.375000 lineto
closepath
clip newpath
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Pr) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
75.000000 11.562500 moveto
(H) show
87.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(someone) show
135.000000 11.562500 moveto
(with) show
165.062500 11.562500 moveto
(your) show
195.062500 11.562500 moveto
(birthday) show
249.062500 11.562500 moveto
(among) show
285.062500 11.562500 moveto
(n) show
297.062500 11.562500 moveto
(people) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
339.062500 11.562500 moveto
(L) show
345.062500 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Math2Mono
%%+ Courier
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02642 .01726 m
.02903 .01981 L
.03164 .02234 L
.03425 .02487 L
.03686 .02739 L
.03947 .02991 L
.04207 .03241 L
.04468 .03491 L
.04729 .03741 L
.0499 .03989 L
.05251 .04237 L
.05512 .04485 L
.05773 .04731 L
.06034 .04977 L
.06295 .05223 L
.06556 .05467 L
.06817 .05711 L
.07078 .05955 L
.07339 .06197 L
.07599 .06439 L
.0786 .0668 L
.08121 .06921 L
.08382 .07161 L
.08643 .074 L
.08904 .07639 L
.09165 .07877 L
.09426 .08114 L
.09687 .08351 L
.09948 .08587 L
.10209 .08823 L
.1047 .09057 L
.10731 .09291 L
.10992 .09525 L
.11252 .09758 L
.11513 .0999 L
.11774 .10222 L
.12035 .10453 L
.12296 .10683 L
.12557 .10912 L
.12818 .11142 L
.13079 .1137 L
.1334 .11598 L
.13601 .11825 L
.13862 .12051 L
.14123 .12277 L
.14384 .12503 L
.14644 .12727 L
.14905 .12951 L
.15166 .13175 L
.15427 .13398 L
Mistroke
.15688 .1362 L
.15949 .13842 L
.1621 .14063 L
.16471 .14283 L
.16732 .14503 L
.16993 .14722 L
.17254 .14941 L
.17515 .15159 L
.17776 .15376 L
.18037 .15593 L
.18297 .15809 L
.18558 .16025 L
.18819 .1624 L
.1908 .16454 L
.19341 .16668 L
.19602 .16881 L
.19863 .17094 L
.20124 .17306 L
.20385 .17518 L
.20646 .17729 L
.20907 .17939 L
.21168 .18149 L
.21429 .18358 L
.21689 .18567 L
.2195 .18775 L
.22211 .18982 L
.22472 .19189 L
.22733 .19395 L
.22994 .19601 L
.23255 .19807 L
.23516 .20011 L
.23777 .20215 L
.24038 .20419 L
.24299 .20622 L
.2456 .20824 L
.24821 .21026 L
.25082 .21228 L
.25342 .21428 L
.25603 .21629 L
.25864 .21828 L
.26125 .22027 L
.26386 .22226 L
.26647 .22424 L
.26908 .22622 L
.27169 .22818 L
.2743 .23015 L
.27691 .23211 L
.27952 .23406 L
.28213 .23601 L
.28474 .23795 L
Mistroke
.28735 .23989 L
.28995 .24182 L
.29256 .24375 L
.29517 .24567 L
.29778 .24759 L
.30039 .2495 L
.303 .2514 L
.30561 .2533 L
.30822 .2552 L
.31083 .25709 L
.31344 .25897 L
.31605 .26085 L
.31866 .26273 L
.32127 .2646 L
.32387 .26646 L
.32648 .26832 L
.32909 .27018 L
.3317 .27203 L
.33431 .27387 L
.33692 .27571 L
.33953 .27754 L
.34214 .27937 L
.34475 .2812 L
.34736 .28302 L
.34997 .28483 L
.35258 .28664 L
.35519 .28844 L
.3578 .29024 L
.3604 .29204 L
.36301 .29383 L
.36562 .29561 L
.36823 .29739 L
.37084 .29916 L
.37345 .30093 L
.37606 .3027 L
.37867 .30446 L
.38128 .30621 L
.38389 .30796 L
.3865 .30971 L
.38911 .31145 L
.39172 .31319 L
.39432 .31492 L
.39693 .31665 L
.39954 .31837 L
.40215 .32008 L
.40476 .3218 L
.40737 .3235 L
.40998 .32521 L
.41259 .32691 L
.4152 .3286 L
Mistroke
.41781 .33029 L
.42042 .33197 L
.42303 .33365 L
.42564 .33533 L
.42825 .337 L
.43085 .33867 L
.43346 .34033 L
.43607 .34198 L
.43868 .34364 L
.44129 .34528 L
.4439 .34693 L
.44651 .34857 L
.44912 .3502 L
.45173 .35183 L
.45434 .35346 L
.45695 .35508 L
.45956 .35669 L
.46217 .35831 L
.46477 .35991 L
.46738 .36152 L
.46999 .36312 L
.4726 .36471 L
.47521 .3663 L
.47782 .36789 L
.48043 .36947 L
.48304 .37105 L
.48565 .37262 L
.48826 .37419 L
.49087 .37575 L
.49348 .37731 L
.49609 .37887 L
.4987 .38042 L
.5013 .38196 L
.50391 .38351 L
.50652 .38505 L
.50913 .38658 L
.51174 .38811 L
.51435 .38964 L
.51696 .39116 L
.51957 .39268 L
.52218 .39419 L
.52479 .3957 L
.5274 .39721 L
.53001 .39871 L
.53262 .4002 L
.53523 .4017 L
.53783 .40319 L
.54044 .40467 L
.54305 .40615 L
.54566 .40763 L
Mistroke
.54827 .4091 L
.55088 .41057 L
.55349 .41203 L
.5561 .41349 L
.55871 .41495 L
.56132 .4164 L
.56393 .41785 L
.56654 .4193 L
.56915 .42074 L
.57175 .42217 L
.57436 .42361 L
.57697 .42503 L
.57958 .42646 L
.58219 .42788 L
.5848 .4293 L
.58741 .43071 L
.59002 .43212 L
.59263 .43353 L
.59524 .43493 L
.59785 .43633 L
.60046 .43772 L
.60307 .43911 L
.60568 .4405 L
.60828 .44188 L
.61089 .44326 L
.6135 .44463 L
.61611 .446 L
.61872 .44737 L
.62133 .44873 L
.62394 .45009 L
.62655 .45145 L
.62916 .4528 L
.63177 .45415 L
.63438 .4555 L
.63699 .45684 L
.6396 .45818 L
.6422 .45951 L
.64481 .46084 L
.64742 .46217 L
.65003 .46349 L
.65264 .46481 L
.65525 .46613 L
.65786 .46744 L
.66047 .46875 L
.66308 .47005 L
.66569 .47135 L
.6683 .47265 L
.67091 .47395 L
.67352 .47524 L
.67613 .47653 L
Mistroke
.67873 .47781 L
.68134 .47909 L
.68395 .48037 L
.68656 .48164 L
.68917 .48291 L
.69178 .48418 L
.69439 .48544 L
.697 .4867 L
.69961 .48795 L
.70222 .48921 L
.70483 .49046 L
.70744 .4917 L
.71005 .49294 L
.71265 .49418 L
.71526 .49542 L
.71787 .49665 L
.72048 .49788 L
.72309 .4991 L
.7257 .50033 L
.72831 .50155 L
.73092 .50276 L
.73353 .50397 L
.73614 .50518 L
.73875 .50639 L
.74136 .50759 L
.74397 .50879 L
.74658 .50998 L
.74918 .51117 L
.75179 .51236 L
.7544 .51355 L
.75701 .51473 L
.75962 .51591 L
.76223 .51709 L
.76484 .51826 L
.76745 .51943 L
.77006 .5206 L
.77267 .52176 L
.77528 .52292 L
.77789 .52408 L
.7805 .52523 L
.78311 .52638 L
.78571 .52753 L
.78832 .52867 L
.79093 .52981 L
.79354 .53095 L
.79615 .53208 L
.79876 .53322 L
.80137 .53434 L
.80398 .53547 L
.80659 .53659 L
Mistroke
.8092 .53771 L
.81181 .53883 L
.81442 .53994 L
.81703 .54105 L
.81963 .54216 L
.82224 .54326 L
.82485 .54436 L
.82746 .54546 L
.83007 .54656 L
.83268 .54765 L
.83529 .54874 L
.8379 .54982 L
.84051 .55091 L
.84312 .55199 L
.84573 .55306 L
.84834 .55414 L
.85095 .55521 L
.85356 .55628 L
.85616 .55734 L
.85877 .5584 L
.86138 .55946 L
.86399 .56052 L
.8666 .56157 L
.86921 .56263 L
.87182 .56367 L
.87443 .56472 L
.87704 .56576 L
.87965 .5668 L
.88226 .56784 L
.88487 .56887 L
.88748 .5699 L
.89008 .57093 L
.89269 .57195 L
.8953 .57298 L
.89791 .574 L
.90052 .57501 L
.90313 .57603 L
.90574 .57704 L
.90835 .57805 L
.91096 .57905 L
.91357 .58006 L
.91618 .58106 L
.91879 .58205 L
.9214 .58305 L
.92401 .58404 L
.92661 .58503 L
.92922 .58602 L
.93183 .587 L
.93444 .58798 L
.93705 .58896 L
Mistroke
.93966 .58993 L
.94227 .59091 L
.94488 .59188 L
.94749 .59285 L
.9501 .59381 L
.95271 .59477 L
.95532 .59573 L
.95793 .59669 L
.96053 .59765 L
.96314 .5986 L
.96575 .59955 L
.96836 .60049 L
.97097 .60144 L
.97358 .60238 L
.97619 .60332 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40005X0000gPP10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdk34g_ojn_l004J2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_P00oiT0IP3o0;C^]:JV
YZf][Jo^kP20P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;002k0?mT0]d009T0V@00
gNk^k^[Zj]gMgK^k^jZZZXR8R7MgMeEEEDA4A14A4Oooc?o<VLbIIS?oVC>IIS=VIYWoo`2Ioc<0VLc<
oiUVVNoGk`00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2o
o`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no
?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo
?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo
?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0
_ol0onj2k_Z0L_lo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoO`0l0>o<knl1k`1]
0:2PX:[Xo`3XooooP;Co]4Bkooo;og?X`?nXOOnCOOno_onooooo?oooOooo_mWEc:0Pl=WIfJj^[YWo
09Wo<iWoIYWoVIWoc<`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aV
VLaVc<aVolbI0<bI<lbIVLbIc<bIolc<0<c<<lc<IP000?l0J@000?l0J@000?l0J@000?l0J@000?l0
J@000440104400<16@050@@00`4500<15@050@@0104400<15@050@@00`4500<15@050@@0104400<1
5P030@D00`4500<15P030@D0104400<18P00040000@100000P42008100<00@406P000`4000030081
00<00@400`020@03004101H000<100000`00104000020@800P400`010@0D00050@0000400`020@03
004100<00P400`010@0D00050@0000400`00104000020@800P400`010@0D00030@0000810`020@03
004100<00P400`010@0D00030@0000810`00104000020@800P400`010@0Q0000A@00104000430003
0@0001P000<100000`001@40000100<000D100000@0F00030@0000P000@100010`000`40000C0003
0@0000@000D100000@0300050@0000405@000`40000900040@000@<000<100005P001@40000100<0
00D100000@0300030@0001H000<100001P001040004300030@0001l0001500040@000@<000<10000
60000`40000300050@0000400`001@40000101H000<1000020001040004300030@0001@000<10000
0`001@40000100<000D100000@0F00030@0000P000@100010`000`40000F00050@0000400`001@40
000100<000<100005P000`40000600040@000@<000<100007`00044000<100000P4200050@000040
6P000`40000300050@0000400`001@40000101H000<1000010000`4000020@8000D100000@0G0003
0@00008000D100000@0300050@0000405`000`40000300030@0000810P001@40000101H00P440005
0@0000400`001@40000101H00P4500030@0000810P001@4000010240001100@10`001@40000101X0
00<100000`001@40000100<000D100000@0F00030@0000@0104300050@00004060001@40000100<0
00D100000@0300030@0001H000<100000P040@<000D100000@0H00050@0000400`001@40000100<0
00<100005P000`40000200@10`001@4000010240000N00030@00020000<1000010001@40000101X0
00<100000`001@40000100<000D100000@0F00030@0000@000<1000010001@40000101@000D10000
0@0300050@0000400`001@40000101@000D100000@0400030@0000@000D100000@0H00050@000040
0`001@40000100<000<100005P000`40000200030@0000@000D100000@0Q00007P000`40000P0003
0@0000@00P400`010@0H00030@0100D00P400`010@03008100<00@4050000`400@0600030@0000@0
0P400`010@0D00030@0000810`020@03004100<00P400`010@0D00030@00008110000`4000040081
00<00@4050001@40000100<00P400`010@03008100<00@4050001@40000100@000<1000010020@03
00410240000N00030@000200104400<16P020@H00`4500<15P020@H0104400<15P030@D00`4500<1
5P030@D0104400<15P030@D00`4500<15P030@D0104400<18P0001h000<10000o`1800005P3o0C`1
5`0001h000<100001@000`40000500030@0000D000<100001@000`40000600030@0000D000<10000
1@000`40000500030@0000D000<100001@000`40000600030@0000D000<100001@000`4000050003
0@0000D000<100001P000`40000500030@0000D000<100001@000`40000500030@0000H000<10000
1@000`40000500030@0000D000<100001@000`40000500030@0000H000<100001@000`4000050003
0@0000D000<100001@000`40000600030@0000D000<100001@000`40000500030@0000D000<10000
1P000`40000H00007P020BL000<100009@000`40000V00030@0002H000<100009P000`40000V0003
0@0002H000<10000:@0001h000@10041o`1700007P001@4000010?l0AP0001h000D100000@3o04H0
000N00810`000`40003o04<0000N00030@0000<00P7o04<0000N00030@0000D000<10000o`100000
7P000`40000600030@000?l0?`0001h000<100001P000`40003o03l0000N00030@0000L00P7o03l0
000N00812P000`40003o03`0000N00030@0000X000<10000o`0k00007P000`40000;00030@000?l0
>P0001h000<1000030000`40003o03T0000N00030@0000d000<10000o`0h00007P000`40000>0081
o`0h00007P020A0000<10000o`0f00007P000`40000@00030@000?l0=@0001h000<100004@020Ol0
=@0001h000<100004`000`40003o0380000N00030@0001@000<10000o`0a00007P000`40000D0081
o`0b00007P020AL000<10000o`0_00007P000`40000G00030@000?l0;P0001h000<1000060000`40
003o02d0000N00030@0001T000<10000o`0/00001@030@H000<100000`050@D000<100006P020Ol0
;00000@00P400`010@0=00030@0000D00`4L00030@000?l0:@0000@000D100000@0=00030@0000D0
00<100007@000`40003o02P0000400050@0000403@000`40000500030@0001d00P7o02T000040005
0@0000403@000`40000500030@0001l000<10000o`0V000010001@40000100d000<100001@000`40
000P0081o`0V000010001@40000100d000<100001@000`40000Q00030@000?l0900000@00P400`01
0@0;00030@0100L00P4S00030@000?l08`0000D00`4=00811`000`40000S0081o`0S00007P000`40
000U00030@000?l0800001h000<100009P000`40003o01l0000N00030@0002L000<10000o`0N0000
7P000`40000X0081o`0N00007P020B/000<10000o`0K00007P000`40000Z00030@000?l06`0001h0
00<10000:`020Ol06`0001h000<10000;@000`40003o01P0000N00030@0002h00P7o01P0000N0003
0@0002l000<10000o`0F00007P020C400P7o01H0000N00030@00038000<10000o`0C00007P000`40
000c00030@000?l04P0001h000<10000=0000`40003o0140000N00030@0003D00P7o0140000N0003
0@0003L000<10000o`0>00007P020CT000<10000o`0=00007P000`40000i00030@000?l0300001h0
00<10000>P020Ol0300001h000<10000?0000`40003o00T0000500<11P000`40000300D11@000`40
000l0081o`0:000010020@03004100/000D100000@0500030@0003h00P7o00P0000400050@000040
30000`40000600<1@0000`40003o00D0000400050@0000403@000`40000500030@00044000<10000
o`04000010001@40000100h000<1000010000`40001200030@000?l00`0000@000D100000@0?0003
0@0000<000<10000@`020Ol00`0000@000D100000@0;00050@0000401@000`40001500030@000?l0
0004008100<00@402`000`4000020@D000<10000A@020Ol00@0000D00`4=00<11P020DP00P7n0000
7P000`40001900030@000?/0000N00030@0004X000<10000nP0001h000<10000B`000`40003i0000
7P000`40001<0081n@0001h00P5?00030@000?H0000N00030@0004l000<10000m@0001h000<10000
D0020OD0000N00030@0005800P7c00007P000`40001C00030@000?40000N00030@0005@00P7a0000
7P020EL00P7_00007P000`40001G00030@000>d0000N00030@0005P00P7]00007P000`40001J0081
j`0001h000<10000G0000`40003X00007P000`40001M00030@000>L0000N0081G`020NL0000N0003
0@00060000<10000i00001h000<10000H@020N@0000N00030@0006<000<10000h@0000D00`460003
0@0000@00`4600030@0006@00P7Q000010020@03004100/000<100000P4500030@0006D00P7P0000
10001@40000100l000<100000`030FL00P7N000010001@40000100l000<100000`000`40001Y0003
0@000=/0000400050@0000403@020@H000<10000J@020M`0000400050@0000403`000`4000030003
0@0006/00P7J000010001@40000100l000<100000`000`40001]0081f00000@00P400`010@0;0005
0@0000401@000`40001^0081e`0000D00`4=00<11P020G400P7E00007P000`40001b00030@000=80
000N00030@0007800P7C00007P000`40001d0081d@0001h000<10000MP020Ll0000N00030@0007L0
0P7>00007P020GX00P7<00007P000`40001k00030@000<T0000N00030@0007`00P7900007P000`40
001n00030@000<H0000N00030@0007l00P7600007P020H400P7500007P000`4000220081``0001h0
00<10000Q0000`40003000007P000`4000250081`00001h000<10000Q`020Kh0000N00030@0008T0
00<10000^`0001h00P6;0081^`0001h000<10000S0020KT0000N00030@0008d00P6h00007P000`40
002?0081]P0000D00`4600030@0000D0104400030@0009400P6d000010020@03004100l000<10000
0`000`40002B0081/`0000@000D100000@0;00H110030I@00`6`000010001@40000100/000D10000
0@0500030@0009L000<10000[@0000@000D100000@0<00040@000@D000<10000V0020Jd000040005
0@00004030020@03004000@000<10000VP020J/0000400050@0000403@000`400@0500030@0009/0
0P6Z000010020@03004100h00P4500030@0009d00P6X00001@030@l00P450081X0020JH0000N0003
0@000:000P6U00007P000`40002R00<1XP0001h000<10000Y0020J40000N00030@000:H00P6O0000
7P000`40002X0081W@0001h00P6Z00<1V`0001h000<10000[0020IT0000N00030@000:d00P6H0000
7P000`40002_0081UP0001h000<10000/@020I@0000N00030@000;<00P6B00007P020KH00P6@0000
7P000`40002f00<1SP0001h000<10000^@020H`0000N00030@000;/000<10000R@0001h000<10000
_0030HP0000N0081`0000`40002500007P000`4000300081Q@0001h000<10000`P030H80000N0003
0@000<@00P6100001@030@H000<1000010040@D000<10000aP030Gh00004008100<00@402`001040
00020@@000<10000b0020Gd0000400050@00004040000`40000200<1bP030GX0000400050@000040
40000`40000200030@000<d000<10000M`0000@000D100000@0<00030@00008110000`40003>00<1
MP0000@000D100000@0<00@11@000`40003A00030@0007<0000400050@00004030000`4000060003
0@000=800`5b000010020@03004100`000<100001P000`40003E0081L00000D00`4=00@11@020ML0
0`5^00007P000`40003I0081K00001h000<10000f`020FX0000N00030@000=d00`5W00007P000`40
003O00<1I@0001h000<10000hP020F<0000N0081i0030F40000N00030@000>H00`5N00007P000`40
003X00<1G00001h000<10000j`020EX0000N00030@000>`00`5H00007P000`40003_00<1E@0001h0
0P7b00<1D`0001h000<10000m0020E40000N00030@000?H00`5>00007P000`40003i0081C00001h0
00<10000nP030DX0000N00030@000?d00P5800007P020Ol00@030DD0000N00030@000?l00`020D<0
000N00030@000?l01@030D00000N00030@000?l020020Ch0000500<11P000`40000500<11@000`40
003o00X00`4k000010020@03004100d000@1000110030Ol03@030CP0000400050@00004030001@40
000100@000<10000o`0@0081=P0000@000D100000@0<00810P000`40000200030@000?l04P030C<0
000400050@00004030020@8000<100000P000`40003o01D00P4a000010001@40000100`000@10041
1@000`40003o01L00`4^000010001@40000100`00P4700030@000?l06P030B/00004008100<00@40
3@000`4000050081o`0N00<1:00000D00`4?00<110000`40003o01l00`4V00007P000`40003o0280
0`4S00007P000`40003o02@0104P00007P000`40003o02P000<100007@0001h000<10000o`180000
7P020Ol0B@0001h000<10000o`1800007P000`40003o04P0000N00030@000?l0B0000?l0J@000?l0
J@000?l0J@000?l0J@000?l0J@000?l0J@0008<0105J00@1:0040AH0104D00@19@0001H000<10000
K0000`40001K00030@0002/000<1000050000`40000E00030@0001/000<100001`0001D000<10000
K@020E`00P4/00030@0001@000<100005@000`40000L00030@0000H0000400@110060@<000<10000
2P050@@010401@010@40008100<00@400P040@@0104200<100@00@410`040@L00P400`010@0300D1
10040@800`4010010@4:00030@0000<0104300<100<00@400P060@P000<100400P4300D10`060@<0
104200<100@00@410`030@03004100L11@000`40000400X100@00@400P4200@10P030@0400410@<0
0`400`01000500<100@00@412@000`400@020@@0104400@10`000`400@020@<01@4400@12P000`40
000600001@000`40000600030@0000<000<100002`020@<000<100000P42008100D00@000@020003
0@0000810`000`4000020@800P4200050@0000400`020@<000<1000010020@03004100D000<10000
10001@400001008000D100000@0:008110020@800P4200030@0000811@000`40000900810P020@@0
00<100001@000`40000400050@0000400P001@40000100<00P42008100<00@000`000`4000030081
1P000`40000200040@000@8000@100010P020@800P4200050@0000400`020@800P4700050@000040
2P020@800P4200810`000`4000020@800P4200810P020@@000<100000`020@<000<1000020000`40
000500001@000`40000600030@0000<000<100002`000`40000200040@000@@000<100400P001040
004200030@0000D000<100000P001040004300050@0000402`020@03004100D000<1000010000`40
000400050@0000402P000`400@0300030@00008000@100010`000`40000300030@0000T000<10000
0P000`40000200030@0000D000<1000010000`40000400050@0000400`000`40000200030@0000D1
1@000`400@0600D10P001040004200040@000@@000@100010`001@40000100@000<100001@001@40
000100X000<100000P001040004700030@00008000@1000110000`40000200030@0000<000<10000
3@000`40000500001@040@D000<100000`000`40000;00D10`000`40000200030@01008000@10001
0P060@8000<100000P001040004300040@0000H11P001@400@0100D000<1000010000`4000040005
0@0000402@001040004300030@00008000@100010`000`40000300030@0000T000<100000P000`40
000200030@0000D000<1000010000`40000400050@0000400`000`40000200030@0000@000<10000
0P001040004:00040@000@8000@100010P000`40000200040@000@<000D100000@0400030@0000D0
00D100000@0:00030@00008000<100001P4200030@00008000@1000110000`40000200030@0000<0
1P4:00030@0000D0000500050@00004010020@8000<100403@000`40000200030@0000810P020@05
004000400P001040004400030@0000810P020@8000D100000@0300030@00008000<100000`001040
004200030@00008000<1000010000`40000400050@0000402@001@40000100800P4200810P001@40
000100D00P4200030@0000D00P42008110000`40000500810P001040004600050@0000400`020@80
0P4200050@00004010001@40000100D000D100000@0200040@000@8000<100000P4200810P001@40
000100<00P4200811`001@40000100X00P4200810P000`40000200030@0000810P020@800P420081
10000`40000300030@00008000<1000020000`40000500001@001@40000100800`4010010@420003
0@0000/01@4300@100<00@401P4300@110040@801@4500@11P020@<00P4200<110060@<010490081
0P020@<01042008100@00@410`030@0400410@L000<100400P4300<11@030@03004100L10`040@D0
0`400`01000200<110020@800P4600<10P080@<0104200D11@030@03004100D01@4:008100@00@41
10040@@01042008100@00@411@000`40000400@12P000`40000600001@001@40000100/000<10000
E`000`40000400030@00030000<100005P000`40000400030@0000X000<10000L@000`40000B0003
0@0000H00000\
\>"],
  ImageRangeCache->{{{108, 395}, {667.125, 490.188}} -> {-201.867, 2.04892, 
  1.5283, 0.00428597}}]
}, Open  ]],

Cell["\<\
People are notortiously bad about dealing with probabilities and in \
some cases this can have important sociological and political consequences.\
\
\>", "Text"],

Cell["\<\
Example: there are cases of communities living near high-voltage \
power lines that have higher incidences of cancer than the general \
population.\
\>", "Text"],

Cell["\<\
Question: Does this provide evidence for the hypthesis that \
exposure to electromagnetic radiation causes cancer?\
\>", "Text"],

Cell["\<\
Answer: That was a trick question the answer is surely not, but \
what about the following.\
\>", "Text"],

Cell["\<\
Question: Does this provide evidence for the hypthesis that \
exposure to electromagnetic radiation is strongly correlated with \
cancer?\
\>", "Text"],

Cell["\<\
Answer: No.  Just as in the birthday case (is there a day and two \
persons who were born on that day), it is very likely that there is a \
location that happens to have a higher than average cancer rate and also \
happens to be near a powerline.  Note that even the notion of a \"location\" \
is complicated in the powerline / cancer problem.  Just as the way you cut up \
a state into congressional districts the way that you cut up the country into \
areas (which you relate to statistics like the rate of reported cancers) can \
have an impact on the resulting statistics.  Have you ever looked at a map of \
the U.S. color coded according to polution, income, or health risks?\
\>", 
  "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["On Game Shows and Bertrand's Box Paradox", "Section"],

Cell["\<\
To take a rather more famous example, consider the so-called Monte \
Hall Problem (identified 100 years earlier by Joseph Bertrand and known in \
the mathematical literature as Bertrand's Box Paradox).  Here's the Monte \
Hall version named after the game-show host.\
\>", "Text"],

Cell["\<\
Question: You're participating in a game show.  There are three \
doors, behind one is a car and behind the other two doors are goats.  You are \
asked to choose a door; after doing so, the game show host selects one of the \
two doors that you did not choose and opens it revealing a goat.  The host \
then asks you if you would like to switch from the door that initially you \
chose to the remaining unopened door, the one that the host did not select.  \
Should you switch doors?\
\>", "Text"],

Cell["\<\
Answer: Many people, including lots of very smart people and people \
who should know better, answer \"no, nothing has changed, so there's no \
reason to switch\"; however, the correct answer, assuming that you want to \
win the car, is \"yes, you should switch.\"\
\>", "Text"],

Cell["\<\
Again the trick to arriving at the right answer involves \
establishing the right context, thinking about the right set of \
experiments.\
\>", "Text"],

Cell[TextData[{
  "It seems that most people wish to consider the context corresponding the \
following: let's assume that you have initially chosen door number one (it \
doesn't matter for any subsequent argument), the host has revealed a goat \
behind door number two, ",
  StyleBox["CXX",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " corresponds to the event in which the car is behind door one and the \
other goat is behind three, and ",
  StyleBox["XXC",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " corresponds to the case where the car and the goat are reversed."
}], "Text"],

Cell[TextData[{
  "Events: ",
  StyleBox["{CXX, XXC}",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " where"
}], "Text"],

Cell[TextData[StyleBox["Pr(CXX) = Pr(XXC) = 1/2",
  FontFamily->"Courier",
  FontWeight->"Bold"]], "Text"],

Cell["\<\
This context corresponds to the set of experiments in which you and \
the host have made your choices as described above.  Problem is that the \
context isn't right!  The context assumes that the host's choice is \
preordained when in fact the host's choice is dictated by where the car is.\
\
\>", "Text"],

Cell[TextData[{
  "Imagine a process in which we generate one of ",
  StyleBox["{CXX, XCX, XXC}",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " with equal probability, again we'll also assume that your initial choice \
is always door one, and then the host selects one of the remaining doors that \
doesn't have a car behind it.  The context is determined by"
}], "Text"],

Cell[TextData[{
  "Events: ",
  StyleBox["{CXX, XCX, XXC}",
    FontFamily->"Courier",
    FontWeight->"Bold"]
}], "Text"],

Cell[TextData[StyleBox["Pr(CXX) = Pr(XXC) = Pr(XCX) = 1/3",
  FontFamily->"Courier",
  FontWeight->"Bold"]], "Text"],

Cell[TextData[{
  "If you switch, you win in both ",
  StyleBox["XCX",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and ",
  StyleBox["XXC",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " but lose in ",
  StyleBox["CXX",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ". If you stick with your choice, you only win in the ",
  StyleBox["CXX",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " case and lose in the other two.  Clearly switching wins twice as often as \
sticking to your initial choice.  It's as though you are given a choice you \
can either chose one door or two doors; posed in this way, it's a no-brainer \
decsion : choosing two doors is twice as good as choosing one."
}], "Text"],

Cell["\<\
Another way to see this is to actually perform a set of \
experiments.  The following procedure does exactly that for us.\
\>", "Text"],

Cell["\<\
Monte[trials_Integer] :=
    Module[{swap = 0, stick = 0, car, door, data = {}},
           Do[car = Random[Integer, 2] ;
              door = Random[Integer, 2] ;
              If[door == car, 
                  stick = stick + 1, 
                  swap = swap + 1] ;
              AppendTo[data, {swap/i, stick/i}],
              {i,1,trials}] ;
           Return[data]]\
\>", "Input"],

Cell["data = Monte[100] ;", "Input"],

Cell["\<\
swap = ListPlot[Table[data[[i]][[1]],
                      {i,1,Length[data]}], 
                {PlotJoined -> True, 
                 DisplayFunction -> Identity}] ;\
\>", "Input"],

Cell["\<\
stick = ListPlot[Table[data[[i]][[2]],
                       {i,1,Length[data]}], 
                 {PlotJoined -> True,
                  DisplayFunction -> Identity}] ;\
\>", "Input"],

Cell[CellGroupData[{

Cell["\<\
Show[swap, {PlotLabel -> \"Switch to the other door\", 
                DisplayFunction -> $DisplayFunction}] ;\
\>", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.00952381 -1.12682 2.30685 [
[.21429 .01411 -6 -9 ]
[.21429 .01411 6 0 ]
[.40476 .01411 -6 -9 ]
[.40476 .01411 6 0 ]
[.59524 .01411 -6 -9 ]
[.59524 .01411 6 0 ]
[.78571 .01411 -6 -9 ]
[.78571 .01411 6 0 ]
[.97619 .01411 -9 -9 ]
[.97619 .01411 9 0 ]
[.01131 .14195 -24 -4.5 ]
[.01131 .14195 0 4.5 ]
[.01131 .25729 -18 -4.5 ]
[.01131 .25729 0 4.5 ]
[.01131 .37263 -24 -4.5 ]
[.01131 .37263 0 4.5 ]
[.01131 .48798 -18 -4.5 ]
[.01131 .48798 0 4.5 ]
[.01131 .60332 -24 -4.5 ]
[.01131 .60332 0 4.5 ]
[.5 .63053 -74 0 ]
[.5 .63053 74 10.125 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.21429 .02661 m
.21429 .03286 L
s
[(20)] .21429 .01411 0 1 Mshowa
.40476 .02661 m
.40476 .03286 L
s
[(40)] .40476 .01411 0 1 Mshowa
.59524 .02661 m
.59524 .03286 L
s
[(60)] .59524 .01411 0 1 Mshowa
.78571 .02661 m
.78571 .03286 L
s
[(80)] .78571 .01411 0 1 Mshowa
.97619 .02661 m
.97619 .03286 L
s
[(100)] .97619 .01411 0 1 Mshowa
.125 Mabswid
.07143 .02661 m
.07143 .03036 L
s
.11905 .02661 m
.11905 .03036 L
s
.16667 .02661 m
.16667 .03036 L
s
.2619 .02661 m
.2619 .03036 L
s
.30952 .02661 m
.30952 .03036 L
s
.35714 .02661 m
.35714 .03036 L
s
.45238 .02661 m
.45238 .03036 L
s
.5 .02661 m
.5 .03036 L
s
.54762 .02661 m
.54762 .03036 L
s
.64286 .02661 m
.64286 .03036 L
s
.69048 .02661 m
.69048 .03036 L
s
.7381 .02661 m
.7381 .03036 L
s
.83333 .02661 m
.83333 .03036 L
s
.88095 .02661 m
.88095 .03036 L
s
.92857 .02661 m
.92857 .03036 L
s
.25 Mabswid
0 .02661 m
1 .02661 L
s
.02381 .14195 m
.03006 .14195 L
s
[(0.55)] .01131 .14195 1 0 Mshowa
.02381 .25729 m
.03006 .25729 L
s
[(0.6)] .01131 .25729 1 0 Mshowa
.02381 .37263 m
.03006 .37263 L
s
[(0.65)] .01131 .37263 1 0 Mshowa
.02381 .48798 m
.03006 .48798 L
s
[(0.7)] .01131 .48798 1 0 Mshowa
.02381 .60332 m
.03006 .60332 L
s
[(0.75)] .01131 .60332 1 0 Mshowa
.125 Mabswid
.02381 .04967 m
.02756 .04967 L
s
.02381 .07274 m
.02756 .07274 L
s
.02381 .09581 m
.02756 .09581 L
s
.02381 .11888 m
.02756 .11888 L
s
.02381 .16502 m
.02756 .16502 L
s
.02381 .18809 m
.02756 .18809 L
s
.02381 .21115 m
.02756 .21115 L
s
.02381 .23422 m
.02756 .23422 L
s
.02381 .28036 m
.02756 .28036 L
s
.02381 .30343 m
.02756 .30343 L
s
.02381 .3265 m
.02756 .3265 L
s
.02381 .34957 m
.02756 .34957 L
s
.02381 .3957 m
.02756 .3957 L
s
.02381 .41877 m
.02756 .41877 L
s
.02381 .44184 m
.02756 .44184 L
s
.02381 .46491 m
.02756 .46491 L
s
.02381 .51104 m
.02756 .51104 L
s
.02381 .53411 m
.02756 .53411 L
s
.02381 .55718 m
.02756 .55718 L
s
.02381 .58025 m
.02756 .58025 L
s
.02381 .00354 m
.02756 .00354 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
gsave
.5 .63053 -135 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.125000 moveto
 558.000000 14.125000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.125000 lineto
closepath
clip newpath
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Switch) show
105.000000 11.562500 moveto
(to) show
123.000000 11.562500 moveto
(the) show
147.000000 11.562500 moveto
(other) show
183.000000 11.562500 moveto
(door) show
207.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.04264 0 m
.04286 .02661 L
.05238 .41108 L
.0619 .60332 L
.07143 .25729 L
.08095 .41108 L
.09048 .52093 L
.1 .31496 L
.10952 .41108 L
.11905 .48798 L
.12857 .34118 L
.1381 .41108 L
.14762 .47023 L
.15714 .35616 L
.16667 .41108 L
.17619 .45914 L
.18571 .36585 L
.19524 .41108 L
.20476 .33014 L
.21429 .37263 L
.22381 .30123 L
.23333 .34118 L
.24286 .37765 L
.25238 .41108 L
.2619 .34957 L
.27143 .38151 L
.28095 .32564 L
.29048 .27377 L
.3 .22547 L
.30952 .25729 L
.31905 .21264 L
.32857 .17078 L
.3381 .13146 L
.34762 .1623 L
.35714 .19138 L
.36667 .21884 L
.37619 .24482 L
.38571 .26943 L
.39524 .29278 L
.40476 .31496 L
.41429 .2798 L
.42381 .30123 L
.43333 .32167 L
.44286 .28875 L
.45238 .25729 L
.4619 .2272 L
.47143 .19839 L
.48095 .21884 L
.49048 .23846 L
.5 .25729 L
Mistroke
.50952 .27538 L
.51905 .29278 L
.52857 .266 L
.5381 .28292 L
.54762 .29923 L
.55714 .31496 L
.56667 .28967 L
.57619 .30502 L
.58571 .28075 L
.59524 .29574 L
.60476 .27242 L
.61429 .28706 L
.62381 .30123 L
.63333 .31496 L
.64286 .29278 L
.65238 .30622 L
.6619 .28484 L
.67143 .26408 L
.68095 .27735 L
.69048 .29025 L
.7 .30278 L
.70952 .31496 L
.71905 .32681 L
.72857 .33834 L
.7381 .34957 L
.74762 .36049 L
.75714 .37114 L
.76667 .38151 L
.77619 .39161 L
.78571 .37263 L
.79524 .3826 L
.80476 .39233 L
.81429 .40182 L
.82381 .41108 L
.83333 .42013 L
.84286 .42896 L
.85238 .41108 L
.8619 .41982 L
.87143 .42836 L
.88095 .43671 L
.89048 .44488 L
.9 .4278 L
.90952 .43589 L
.91905 .4438 L
.92857 .42727 L
.9381 .41108 L
.94762 .41901 L
.95714 .42677 L
.96667 .43438 L
.97619 .41877 L
Mistroke
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40005X0000gPP10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdk34g_ojn_l004J2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_P00oiT0IP3o0;C^]:JV
YZf][Jo^kP20P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;002k0?mT0]d009T0V@00
gNk^k^[Zj]gMgK^k^jZZZXR8R7MgMeEEEDA4A14A4Oooc?o<VLbIIS?oVC>IIS=VIYWoo`2Ioc<0VLc<
oiUVVNoGk`00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2o
o`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no
?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo
?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo
?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0
_ol0onj2k_Z0L_lo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoO`0l0>o<knl1k`1]
0:2PX:[Xo`3XooooP;Co]4Bkooo;og?X`?nXOOnCOOno_onooooo?oooOooo_mWEc:0Pl=WIfJj^[YWo
09Wo<iWoIYWoVIWoc<`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aV
VLaVc<aVolbI0<bI<lbIVLbIc<bIolc<0<c<<lc<IP000?l0J@000?l0J@000?l0J@000?l0J@000?l0
J@0005`01@4400<1<`040@<00`4c00<110030C800`4500<1;P050@@00`4500<11`0005`000D10000
0@03008100<00@40=000104000020@030041038000@100010P020@030041030000D100000@030081
00<00@40;`000`400003008100<00@400`020@03004100H0000V00030@0000<000<10000;P000`40
000400050@000040<0060@8000D100000@0a00050@0000400P001@400001030000D100000@030005
0@000040;`000`40000300050@0000400`001@40000100H0000V008110000`40000_00030@0000<0
00D100000@0`00050@0000400`001@40000103400P4200040@000@<000<10000;P001@40000100<0
00D100000@0_00030@0000<000D100000@0300050@0000401P0002H000<100000`000`40000`0003
0@00008000D100000@0a00040@000@<000D100000@0a00810P001040004300030@0002l00`440005
0@000040;`000`40000300050@0000400`001@40000100H0000V00030@0000<000<10000<@001@40
000100<000<10000;`020@030040008000D100000@0a00040@010@<000D100000@0`00050@000040
0`001@40000102l000<100000`001@40000100<000D100000@0600009P000`40000300030@0002d0
00D100000@0300050@000040<P000`400@0300050@000040<@020@D000D100000@0`00050@000040
0`001@40000102l000<100000`001@40000100<000D100000@0600009P000`40000300030@0002d0
00<100000P43008100<00@40<`020@<00P400`010@0b00030@0000<00P400`010@0`00050@000040
0`020@03004102d000<100401@020@03004100<00P400`010@0600009P000`40000300030@0002h0
0`4500<1=0020@@00`4d00<10`030C800`4500<1;`020@H00`4500<11`0002H000<100000`000`40
003o03X0000N0?l1@`4800009P000`40000300030@0000H000<1000030000`40000=00030@0000`0
00<1000030000`40000<00030@0000d000<1000030000`40000<00030@0000`000<100003@000`40
000<00030@0000`000<1000030000`40000=00030@0000`000<1000030000`40000<00030@0000d0
00<1000030000`40000>00009P000`40000300030@0000H000<1000030000`40000=00030@0000`0
00<1000030000`40000<00030@0000d000<1000030000`40000<00030@0000`000<100003@000`40
000<00030@0000`000<1000030000`40000=00030@0000`000<1000030000`40000<00030@0000d0
00<1000030000`40000>00009P000`40000300030@000?l0>P0002H000<100000`000`40003o03X0
000V00030@0000<000<10000o`0j00009P000`40000300030@000?l0>P0002H000<100000`000`40
003o03X0000V008110000`40003o03X0000V00030@0000<000<10000o`0j00009P000`4000030003
0@000?l0>P0002H000<100000`000`40003o03X0000V00030@0000<000<10000o`0j00009P000`40
000300030@000?l0>P0002H000<100000`000`40003o03X0000V008110000`40003o03X0000V0003
0@0000<000<10000o`0j00009P000`40000300030@000?l0>P0002H000<100000`000`40003o03X0
000V00030@0000<000<10000o`0j00009P000`40000300030@000?l0>P0002H000<1000010000`40
003o03T0000V00030@0000@000<10000o`0i00009P020@D000<10000o`0i00009P000`4000040003
0@000?l0>@0002H000<1000010000`40003o03T0000V00030@0000@000<10000o`0i00009P000`40
000400030@000?l0>@0002H000<1000010000`40003o03T0000V00030@0000@000<10000o`0i0000
9P020@D000<10000o`0i00009P000`40000400030@000?l0>@0002H000<1000010000`40003o03T0
000V00030@0000@000<10000o`0i00009P000`40000400030@0005/000<10000fP0000H00`460003
0@0000@0104400@110000`40000400030@0005/000<10000fP0000D00P400`010@0;00040@000081
0P00104000020@<000<1000010000`40001J00030@010=/0000500050@00004040000`4000050004
0@0000<110000`40001J00030@010=/0000500050@00004040000`40000500050@0000401P000`40
001J00030@010=/0000500050@00004030000`4000020@<000<100000P4300030@0000@000<10000
F@001@4000010=X0000500050@00004030040@@0104400030@0000@000<10000F@001@4000010=X0
000500050@00004030000`40000500030@0000D000<1000010000`40001I00050@000040fP0000D0
0P400`010@0<00030@0000D000<100001@000`40000400030@0005T000D100000@3J00001P030@d0
104400@110000`40000400030@0005P000<100000`000`40003G00009P000`40000400030@0005P0
00<100000`000`40003G00009P020@D000<10000F0000`40000300030@000=L0000V00030@0000@0
00<10000E`000`40000500030@000=H0000V00030@0000@000<10000E`000`40000500030@000=H0
000V00030@0000@000<10000E`000`40000500030@000=H0000V00030@0000@000<10000E`000`40
000600030@000=D0000V00030@0000@000<10000EP000`40000700030@000=D0000V00030@0000@0
00<10000EP000`40000700030@000=D0000V00811@000`40001F00030@0000P000<10000e00002H0
00<1000010000`40001F00030@0000P000<10000e00002H000<1000010000`40001E00030@0000T0
00<10000e00002H000<1000010000`40001E00030@0000X000<1000080000`40002`00009P000`40
000400030@0005D000<100002P000`40000P00030@000;00000V00030@0000@000<10000E@000`40
000:00030@0001l000<10040/@0002H000<1000010000`40001D00030@0000`000<100007P000`40
0@2a00009P020@D000<10000E0000`40000<00030@0001h000@10001/00002H000<1000010000`40
001D00030@0000`000<100007P001040006`00009P000`40000400030@0005@000<100003@000`40
000L00030@00008000<10000[@0002H000<100001@000`40001C00030@0000d000<1000070000`40
000200030@000:d0000V00030@0000D000<10000DP000`40000>00030@0001`000<100000`000`40
002/00009P000`40000500030@0004d000<100000P000`40000?00030@0001X000<1000010000`40
002/00009P000`40000500030@0004d000<100000P000`40000?00030@0001X000<100001@000`40
002[00009P000`40000500030@0004d00P4300030@00010000<100006@000`40000500030@000:/0
000V00811P000`40001<00030@0100<000<1000040000`40000H00030@0000L000<10000ZP0002H0
00<100001@000`40001<00030@01008000<100004@000`40000H00030@0000L000<10000ZP0002H0
00<100001@000`40001<00060@000@0150000`40000G00030@0000P000<10000Z@0002H000<10000
1@000`40001<00060@000@0150000`40000F00030@0000T000<10000Z@0002H000<100001@000`40
001<00060@000@0150000`40000F00030@0000X000<10000Z00000d00`4600030@0000D00`450003
0@0000D000<10000C0000`4000020AH000<100005@000`40000:00030@000:P0000<008100<00@40
3@001040004400030@0000D000<10000B`000`40000200030@0001@000<1000050000`40000<0003
0@000:L0000<00050@00004030001@40000100@00`4500030@0000@000<10000A0000`4000020003
0@0001D000<100004`000`40000<00030@000:L0000<00050@00004030020@8000<100000P000`40
000500030@0000@000<10000A0000`40000J00030@0001<000<100003@000`40002V000030001@40
000100`00P4200030@00008000<100001@000`40000400030@0004@000<100006P000`40000B0003
0@0000h000<10000<`000`40001`000030001@40000100`000@100411@000`40000500030@0000@0
00<10000A0000`40000K00030@00014000<100003`000`40000500030@0002X00P5a000030001@40
000100`00P4700030@0000D000<1000010000`40001300030@0001`000<100004@000`40000?0003
0@0000D00P4Z00030@010740000<008100<00@403@000`40000500030@0000D000<1000010000`40
001300030@0001`000<100004@000`40000@00030@0000<000<100405P000`40000A00040@000G00
000=00<13`030@@000<100001@000`40000400030@0004<000<100007@000`40000?00030@000140
00<100000`001040004E00814@001@4000010700000V00030@0000D000<1000010000`4000130003
0@0001d000<100001@000`40000700030@00018000D100000@0300030@00018000<100404@000`40
000200030@0006d0000V00811P000`40000400030@0004<000<100007P000`40000400030@0000L0
00<100004P001@40000100@000<1000030000`40000200040@000@l000<1000010000`40001/0000
9P000`40000500030@0000@00P5300030@0001l000<1000010020@L000<1000050000`400@060003
0@0000/00P4300040@000@l000<100001@000`40001[00009P000`40000500030@0000@00P530003
0@0001l000<100000`000`400@0700030@0001@000<100401P000`40000:00030@01008000<10000
0P000`40000<00030@0000D000<10000J`0002H000<100001@000`4000040081@`000`40000P0003
0@00008000@100011P000`40000E00030@0000H000<1000010000`40000200060@000@011@000`40
000:00030@0000L000<10000JP0002H000<100001@000`4000040081@`000`40000P00030@000080
00@100011@000`40000F00030@0000H000<1000010020@<000<100000P4700030@0000@000<10000
0P000`40000800030@0006T0000V00030@0000D000<1000010020D<000<1000080000`4000020005
0@00004010000`40000P00030@00008000<100400P000`40000200030@0000D000<1000010020@80
00<100002P000`40001X00009P000`40000500030@0000@00P4_00030@00014000<100008@001040
004400030@00008000<1000080000`40000200060@000@013@000`40000200060@000@0130000`40
001X00009P020@H000<1000010020Bl000<1000040000`40000R00040@000@@000<100000P000`40
000Q00050@0000400P020@l000D100000@0200813P000`40001W00009P000`40000500030@0000@0
0P4_00030@00010000<100008`000`400@0500040@000B@000@1000110000`40000=00040@000@@0
00<100003@000`40001V00009P000`40000500030@0000@00P4_00814@000`40000S00030@0100D0
00@100019@000`400@0E00030@0101D000<10000I@0002H000<100001@000`4000040081;P000`40
0@0A00030@0002@000<100001@000`400@0U00815P020AH000<10000I@0002H000<100001@000`40
000300030@0100L000<1000090000`400@0A00030@0002@000<100001@000`400@0V00030@0001D0
00<100005@000`40001T00009P000`40000500030@0000<000<100401`000`40000T00030@010100
00<10000;P000`40001F00030@0006<0000V00030@0000D000<100000`000`400@0700030@0002@0
00@100013`000`40000^00030@0005L000<10000HP0002H00P4600030@0000<000<100401`000`40
000T00040@000@l000<10000R@000`40001Q00009P000`40000500030@0000<000<100401`000`40
000T00040@000@l000<10000RP000`40001P00009P000`40000500030@0000<000<100401`000`40
000O00030@00008000@100013`000`40002;00030@0005l0000V00030@0000D000<100000`001040
0046008180000`40000200050@0000403P000`40002<00030@0005h0000V00030@0000D000<10000
0`0010400046008180001@40000100@000<100002`000`40002>00030@0005d0000V00030@0000D0
00<100000`001040004600811`000`40000F00810P000`40000200030@0000/000<10000S`000`40
001L00009P000`40000500030@0000<000@100011P020@L000<100005P020@8000<100000P000`40
000;00030@00090000<10000F`0002H000<100001@000`40000300040@000@H00P4700030@0001D0
00<100400P000`40000300030@0000X000<10000T0000`40001K00009P020@L000<100000P001040
004600811`000`40000E00030@01008000<100000`000`40000500030@00008000<10000T@000`40
001J00009P000`40000600030@00008000@100011@000`400@0700815P001P4000400@D000<10000
1@000`40000200030@00098000<10000F@0002H000<100001P000`40000200040@000@D000<10040
1`020@P000<100002`001P4000400@D000<100001@020@8000<10000U0000`40001H00009P000`40
000600030@00008000@100011@000`400@07008120000`40000;00060@000@011P000`4000040081
0P000`40002E00030@0005L0000V00030@0000H000<100000P001040004500030@0100L00P480003
0@0000/000<100000P4700030@0000<000<100400P000`40002F00030@0005H0000600<11P000`40
000500<110040@@000<100001P000`40000200040@000@D000@100011@000`400@0800811`000`40
000200050@0000401`000`40000300060@000@01V@000`40001E00001@020@03004100d000@10001
0P00104000020@<000<100001P000`40000200040@000@D000@100011@000`400@0800811`000`40
000200050@0000401`000`40000300060@000@01V@000`40001E00001@001@40000100`000D10000
0@0700040@0000<11P000`40000200040@000@D000@100011@000`400@0700030@0100L000<10000
0P001@40000100P000<100000P001P4000400IX000<100001P000`40001;00001@001@40000100`0
0P4200030@0000D000D100000@0800030@00008000@100011@001040004500030@0100L000<10040
1`020@8000<100002`000`40000200030@000081W0000`4000050081C00000D000D100000@0<0081
0P001@40000100800P4300030@0000H000<100000P001040004500040@000@D000@100011P000`40
0@0700810P000`40000;00030@00008000D100000@2M00030@0000<000@10001B`0000D000D10000
0@0<00040@010@@0104400030@0000H000<100000P001040004500040@000@D000@100011P000`40
0@0700810P000`40000;00030@00008000D100000@2N00030@00008000D100000@1:00001@001@40
000100`00P4600030@0000D000<100001P000`40000200040@000@D000@100011@00104000460004
0@000@D000<100400P000`40000<00040@000J@000@100011@000`40001700001@020@03004100d0
00<1000010000`40000500030@0000H000<100000P001@40000100@000@100011@00104000460004
0@000@D000<100400P000`40000<00040@000JD000<100401P000`40001600001P030@l00`4300@1
10000`40000600030@00008000D100000@0400040@000@D000@100011P001040004500060@000@01
3P001040006V00030@0000D000<10000AP0002H000<100001P000`40000200050@00004010001040
004500040@000@H000@100011@001P4000400@h000@10001[`000`40001500009P020@L000<10000
0P001@40000100@000D100000@0400040@000@H000@100011@001P4000400@l000<10040/0000`40
001400009P000`40000600030@00008000D100000@0400050@00004010001040004500050@000040
1@000`4000020A0000<10040/@000`40001300009P000`40000600030@00008000D100000@040005
0@00004010001@40000100@000<100000P000`40000200030@00008140020K<000<10000@P0002H0
00<100001P000`40000200050@00004010001@40000100@000D100000@0400030@00008000<10000
0P001@40000101000P6d00030@000440000V00030@0000H000<100000P001@40000100@000D10000
0@0400050@00004010000`40000200030@00008000D100000@0A00030@000;<000<10000@00002H0
00<100001P000`40000200050@00004010001@40000100@000D100000@0400030@00008000<10000
0P001@400001014000<10000]0000`40000600030@0001T000<100006P0002H000<100001P000`40
000200050@00004010001@40000100<000<100000P000`40000200030@00008000D100000@3>0081
1@0010400@4H00040@010AX0000V00811`000`40000200050@00004010001@40000100<000<10000
0P000`40000200030@00008000D100000@3@00030@00008000D100000@0G00050@00004020000`40
000>00009P000`40000600030@00008000D100000@0300030@00008000D100000@0500050@000040
1@001040007A00040@000@D000<100004`000`40000300030@0000@000<100003`0002H000<10000
1P000`40000200050@0000400`000`40000300040@000@D000D100000@0500040@000M8000<10040
1P000`40000B00030@0000@000<100000`000`40000?00009P000`40000600030@00008000D10000
0@0300030@0000<000@100011@001040004600040@000M<000<100001P000`40000700030@0000H0
00<100001P001@4000010180000V00030@0000H000D100000@0500040@000@D000@100011@001040
004600040@000Md000<100001@0010400@4500030@0000P00P400`01000A00009P000`4000060005
0@0000401@001040004500040@000@D000@100011P001040007N00030@0000<000<100000P000`40
000200030@0000X000<100004@0002H000<100001P001@40000100D000@100011@00104000450004
0@000@L000<10040g`000`40000200030@0000<000@100018@0002H00P4700050@0000401@001040
004500040@000@H000<100401`000`400@3P008100<00@001P000`400@0Q00009P000`4000060005
0@0000401@001040004500040@000@H000<100401`000`400@3R00030@0000L000<10000800002H0
00<100001`001040004500040@000@D000@100011P000`400@070081o`0A00009P000`4000070004
0@000@D000@100011P000`400@0600030@0100L00P7o0140000V00030@0000L000@100011@001040
004600030@0100H000<1004020000`40003o00l0000V00030@0000L000@100011@00104000460003
0@0100H00P4900030@000?l03`0002H000<100001`001040004500040@000@H000<100401`000`40
000700030@000?l03`0002H000<100001`001040004500040@000@H000<100401`000`40003o01T0
000V008120001040004500040@000@H00P4800030@000?l06@0002H000<100001`00104000460003
0@0100H00P4800030@000?l06@0002H000<100001`001040004600030@0100H00P7o02@0000V0003
0@0000L000@100011P000`400@0700030@000?l08P0002H000<100001`001040004600030@0100L0
00<10000o`0R00003@030@H000<100001@000`40000500030@0000L000@100011P000`400@070003
0@000?l08P0000`00P400`010@0=00030@0000D000<100001`001040004600030@0100L000<10000
o`0R000030001@40000100h000<1000010030@L000@100011P000`400@0700030@000?l08P0000`0
00D100000@0>00030@0000@000<100001`00104000460081o`0]000030001@40000100h000<10000
10000`40000700040@000@H00P7o02d0000<00050@0000403P000`40000400030@0000L000@10001
1P020Ol0;@0000`000D100000@0>00030@0000@000<100001`00104000460081o`0]000030020@03
004100/000D100000@0500030@0000L000@100011P020Ol0;@0000d00`4<00D11@000`4000070004
0@000@L000<10000o`0[00009P000`40000800030@0100L000<10000o`0[00009P020@T000<10040
1`000`40003o02/0000V00030@0000P000<100401`000`40003o02/0000V00030@0000P000<10040
1`000`40003o02/0000V00030@0000P000<100401`000`40003o02/0000V00030@0000P000<10040
o`0e00009P000`40000800030@010?l0=@0002H000<1000020000`400@3o03D0000V00812@000`40
0@3o03D0000V00030@0000P000<10040o`0e00009P000`40000800030@010?l0=@0002H000<10000
20000`400@3o03D0000V00030@0000P00P7o03H0000V00030@0000P00P7o03H0000V00030@0000P0
0P7o03H0000V00812@020Ol0=P0002H000<1000020020Ol0=P0002H000<1000020020Ol0=P0002H0
00<1000020020Ol0=P0002H000<1000020020Ol0=P0002H000<100002@000`40003o03@0000V0003
0@0000T000<10000o`0d00009P000`40000900030@000?l0=00002H00P4:00030@000?l0=00002H0
00<100002@000`40003o03@0000V00030@0000T000<10000o`0d00009P000`40000900030@000?l0
=00002H000<100002@000`40003o03@0000600<11P000`40000500030@0000@0104400030@0000T0
00<10000o`0d00001@020@03004100d000<100000`00104000020@<000<100002@000`40003o03@0
000500050@0000403P000`40000700040@0000<12@000`40003o03@0000500050@0000403P000`40
000700050@000040o`1200001@001@40000100h000<100000`000`4000020@<000<10000o`100000
1@001@40000100h000<100000`040@@000<10000o`1000001@001@40000100h000<100000`000`40
000500030@000?l0@00000D00P400`010@0;00050@00004010000`40000500030@000?l0@00000H0
0`4<00D110040Ol0A`000?l0J@000?l0J@000?l0J@000?l0J@000?l0J@000?l0J@000?l0J@000?l0
J@0006D000<100400P43008100<00@400`050@@0104400@10P030@0400410@P0104400@12P040@80
0`4010010@4300@12P040@@0104200<100@00@410`040@<01P4800<100<00@400P040@@0104300H1
C`0006D00P4300030@00008100<00@401@000`40000400050@0000400P020@<000@100010`000`40
000700050@0000400P020@800P4900050@0000400P001@40000100<00P4300030@0000H00P420081
0`001@400001008000D100000@0300810`000`40000200030@0000P00P4200810P020@800P420081
0P020@@000<10000D00006D000<100000P000`4000020@03004100D000<1000010000`4000040003
0@0000D000D100000@0900030@0000@000<100000P000`40000700030@0000@000D100000@030003
0@0000/000<100000P001@40000100H000D100000@0300030@0000L000<1000020000`4000020004
0@000@@000@1000110000`40000200030@000500001W00@10P001@400@0100D000<1000010000`40
000400030@00008000@100010`000`40000700030@0000@000<100000P000`40000700030@0000@0
00D100000@0300H120000`40000200050@0000401P001@40000100<01P4400030@0000P000<10000
0P001040004400040@000@@000<100000P000`40001@0000I@030@@000@100010P000`4000020003
0@0000@000<1000010020@800P4200050@0000402@000`40000400810P020@T000<1000010001@40
000100<000<100000P000`40000600810P020@<000<1000010001@40000100<000<100000P000`40
000200810P000`40000400810P020@800P4200810P020@800P4400810P000`40001<0000I@000`40
000200040@010@<00P4200<110060@@00`40100100040@P01P4400@120060@<0104500@12P040@80
1P4300@11@040@<00`4010010@4700<100<00@000P040@@0104300<100@00@41CP0006D000@10000
0P4C00030@0000`000<100002`000`40000C00030@0000@000<1000070000`40000400030@0001l0
00<10000I@000001\
\>"],
  ImageRangeCache->{{{108, 395}, {673.188, 496.25}} -> {-56.8817, 1.31962, 
  0.410137, 0.00169325}}]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Show[stick, {PlotLabel -> \"Stick to your original door\", 
             DisplayFunction -> $DisplayFunction}] ;\
\>", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.00952381 -0.561998 2.30685 [
[.21429 .11756 -6 -9 ]
[.21429 .11756 6 0 ]
[.40476 .11756 -6 -9 ]
[.40476 .11756 6 0 ]
[.59524 .11756 -6 -9 ]
[.59524 .11756 6 0 ]
[.78571 .11756 -6 -9 ]
[.78571 .11756 6 0 ]
[.97619 .11756 -9 -9 ]
[.97619 .11756 9 0 ]
[.01131 .01472 -24 -4.5 ]
[.01131 .01472 0 4.5 ]
[.01131 .2454 -24 -4.5 ]
[.01131 .2454 0 4.5 ]
[.01131 .36074 -18 -4.5 ]
[.01131 .36074 0 4.5 ]
[.01131 .47609 -24 -4.5 ]
[.01131 .47609 0 4.5 ]
[.01131 .59143 -18 -4.5 ]
[.01131 .59143 0 4.5 ]
[.5 .63053 -83.0312 0 ]
[.5 .63053 83.0312 10.125 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.21429 .13006 m
.21429 .13631 L
s
[(20)] .21429 .11756 0 1 Mshowa
.40476 .13006 m
.40476 .13631 L
s
[(40)] .40476 .11756 0 1 Mshowa
.59524 .13006 m
.59524 .13631 L
s
[(60)] .59524 .11756 0 1 Mshowa
.78571 .13006 m
.78571 .13631 L
s
[(80)] .78571 .11756 0 1 Mshowa
.97619 .13006 m
.97619 .13631 L
s
[(100)] .97619 .11756 0 1 Mshowa
.125 Mabswid
.07143 .13006 m
.07143 .13381 L
s
.11905 .13006 m
.11905 .13381 L
s
.16667 .13006 m
.16667 .13381 L
s
.2619 .13006 m
.2619 .13381 L
s
.30952 .13006 m
.30952 .13381 L
s
.35714 .13006 m
.35714 .13381 L
s
.45238 .13006 m
.45238 .13381 L
s
.5 .13006 m
.5 .13381 L
s
.54762 .13006 m
.54762 .13381 L
s
.64286 .13006 m
.64286 .13381 L
s
.69048 .13006 m
.69048 .13381 L
s
.7381 .13006 m
.7381 .13381 L
s
.83333 .13006 m
.83333 .13381 L
s
.88095 .13006 m
.88095 .13381 L
s
.92857 .13006 m
.92857 .13381 L
s
.25 Mabswid
0 .13006 m
1 .13006 L
s
.02381 .01472 m
.03006 .01472 L
s
[(0.25)] .01131 .01472 1 0 Mshowa
.02381 .2454 m
.03006 .2454 L
s
[(0.35)] .01131 .2454 1 0 Mshowa
.02381 .36074 m
.03006 .36074 L
s
[(0.4)] .01131 .36074 1 0 Mshowa
.02381 .47609 m
.03006 .47609 L
s
[(0.45)] .01131 .47609 1 0 Mshowa
.02381 .59143 m
.03006 .59143 L
s
[(0.5)] .01131 .59143 1 0 Mshowa
.125 Mabswid
.02381 .03778 m
.02756 .03778 L
s
.02381 .06085 m
.02756 .06085 L
s
.02381 .08392 m
.02756 .08392 L
s
.02381 .10699 m
.02756 .10699 L
s
.02381 .15313 m
.02756 .15313 L
s
.02381 .17619 m
.02756 .17619 L
s
.02381 .19926 m
.02756 .19926 L
s
.02381 .22233 m
.02756 .22233 L
s
.02381 .26847 m
.02756 .26847 L
s
.02381 .29154 m
.02756 .29154 L
s
.02381 .31461 m
.02756 .31461 L
s
.02381 .33767 m
.02756 .33767 L
s
.02381 .38381 m
.02756 .38381 L
s
.02381 .40688 m
.02756 .40688 L
s
.02381 .42995 m
.02756 .42995 L
s
.02381 .45302 m
.02756 .45302 L
s
.02381 .49915 m
.02756 .49915 L
s
.02381 .52222 m
.02756 .52222 L
s
.02381 .54529 m
.02756 .54529 L
s
.02381 .56836 m
.02756 .56836 L
s
.02381 .6145 m
.02756 .6145 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
gsave
.5 .63053 -144.031 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.125000 moveto
 558.000000 14.125000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.125000 lineto
closepath
clip newpath
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Stick) show
99.000000 11.562500 moveto
(to) show
117.062500 11.562500 moveto
(your) show
147.062500 11.562500 moveto
(original) show
201.062500 11.562500 moveto
(door) show
225.062500 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.04264 .61803 m
.04286 .59143 L
.05238 .20695 L
.0619 .01472 L
.07143 .36074 L
.08095 .20695 L
.09048 .0971 L
.1 .30307 L
.10952 .20695 L
.11905 .13006 L
.12857 .27686 L
.1381 .20695 L
.14762 .1478 L
.15714 .26188 L
.16667 .20695 L
.17619 .15889 L
.18571 .25219 L
.19524 .20695 L
.20476 .28789 L
.21429 .2454 L
.22381 .3168 L
.23333 .27686 L
.24286 .24039 L
.25238 .20695 L
.2619 .26847 L
.27143 .23653 L
.28095 .29239 L
.29048 .34427 L
.3 .39256 L
.30952 .36074 L
.31905 .40539 L
.32857 .44725 L
.3381 .48657 L
.34762 .45573 L
.35714 .42665 L
.36667 .39919 L
.37619 .37321 L
.38571 .3486 L
.39524 .32525 L
.40476 .30307 L
.41429 .33824 L
.42381 .3168 L
.43333 .29637 L
.44286 .32929 L
.45238 .36074 L
.4619 .39083 L
.47143 .41964 L
.48095 .39919 L
.49048 .37957 L
.5 .36074 L
Mistroke
.50952 .34265 L
.51905 .32525 L
.52857 .35204 L
.5381 .33511 L
.54762 .3188 L
.55714 .30307 L
.56667 .32837 L
.57619 .31301 L
.58571 .33728 L
.59524 .3223 L
.60476 .34562 L
.61429 .33098 L
.62381 .3168 L
.63333 .30307 L
.64286 .32525 L
.65238 .31181 L
.6619 .3332 L
.67143 .35396 L
.68095 .34068 L
.69048 .32779 L
.7 .31526 L
.70952 .30307 L
.71905 .29122 L
.72857 .27969 L
.7381 .26847 L
.74762 .25754 L
.75714 .2469 L
.76667 .23653 L
.77619 .22642 L
.78571 .2454 L
.79524 .23543 L
.80476 .22571 L
.81429 .21622 L
.82381 .20695 L
.83333 .19791 L
.84286 .18907 L
.85238 .20695 L
.8619 .19821 L
.87143 .18967 L
.88095 .18132 L
.89048 .17315 L
.9 .19024 L
.90952 .18215 L
.91905 .17423 L
.92857 .19076 L
.9381 .20695 L
.94762 .19903 L
.95714 .19126 L
.96667 .18365 L
.97619 .19926 L
Mistroke
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40005X0000gPP10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdk34g_ojn_l004J2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_P00oiT0IP3o0;C^]:JV
YZf][Jo^kP20P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;002k0?mT0]d009T0V@00
gNk^k^[Zj]gMgK^k^jZZZXR8R7MgMeEEEDA4A14A4Oooc?o<VLbIIS?oVC>IIS=VIYWoo`2Ioc<0VLc<
oiUVVNoGk`00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2o
o`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no
?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo
?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo
?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0
_ol0onj2k_Z0L_lo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoO`0l0>o<knl1k`1]
0:2PX:[Xo`3XooooP;Co]4Bkooo;og?X`?nXOOnCOOno_onooooo?oooOooo_mWEc:0Pl=WIfJj^[YWo
09Wo<iWoIYWoVIWoc<`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aV
VLaVc<aVolbI0<bI<lbIVLbIc<bIolc<0<c<<lc<IP000?l0J@00028000<10000o`1400008P000`40
003o04@0000R00030@000?l0A00000800`4600030@0000<01@4400@110000`40003o04@000001001
0@020@/000D100000@0300040@0000810`000`40003o04@000000`01000200030@0000X000<10000
2@00104000030@T000<10000o`0h000000<00@000P000`40000;00030@0000P000D100000@0;0003
0@000?l0>00000030040008000<1000030000`40000300030@0000810`000`40000900030@000?l0
>00000030040008000<100003@000`40000200@110000`40000900030@000?l0>000000300400080
00<100002@001@40000100@000<100001@000`40000900030@000?l0>0000004004100812`000`40
00020@@000<100001@000`40000900030@000?l0>00000800`4=00<11@040@@000<100002@000`40
003o03P0000R00812P000`40003o03P0000R00030@0000T000<10000o`0h00008P000`4000090003
0@000?l0>000028000<100002@000`40003o03P0000R00030@0000P00P7o03X0000R00030@0000P0
0P7o03X0000R00030@0000P00P7o03X0000R00030@0000P00P7o03X0000R00812@020Ol0>P000280
00<1000020020Ol0>P00028000<1000020020Ol0>P00028000<1000020020Ol0>P00028000<10000
20000`400@3o03T0000R00030@0000P000<10040o`0i00008P000`40000800030@010?l0>@000280
00<1000020000`400@3o03T0000R00812@000`400@3o03T0000R00030@0000P000<10040o`0i0000
8P000`40000800030@010?l0>@00028000<1000020000`400@3o03T0000R00030@0000P000<10040
20000`40000O00D110030CD0104300<1=0030@@00`4d00<11@030Bl01@4400<11@030@<0000R0003
0@0000P000<1004020000`40000O00050@0000400`020@03004103H000@100000P400`010@0c0004
0@000@800P400`010@0b00050@0000400`020@030041030000<100000`020@03004100<00P400`01
0@0200008P000`40000800030@0100P000<1000080000`40000400050@000040<P060@8000D10000
0@0b00050@0000400P001@400001038000D100000@0300050@000040<0000`40000300050@000040
0`001@4000010080000R00812@000`400@0800030@00024000<100000`001@400001038000D10000
0@0300050@000040<P020@8000@100010`000`40000`00050@0000400`001@400001030000<10000
0`001@40000100<000D100000@0200008P000`40000800030@0100P000<100008P000`4000020005
0@000040<`001040004300050@000040<P020@8000@100010`000`40000a00<110001@4000010300
00<100000`001@40000100<000D100000@0200008P000`40000700040@000@L00P4U00050@000040
0`000`40000a008100<00@000P001@400001038000@100410`001@400001038000D100000@030005
0@000040<0000`40000300050@0000400`001@4000010080000R00030@0000L000@100011`020B40
00D100000@0300050@000040=0000`400@0300050@000040<P020@D000D100000@0b00050@000040
0`001@400001030000<100000`001@40000100<000D100000@0200008P000`40000700040@000@L0
0P4Q00030@0000810`020@03004103D00P43008100<00@40<`000`400003008100<00@40<P001@40
000100<00P400`010@0^00030@0100D00P400`010@03008100<00@400P00028000<100001`001040
004700818P030@D00`4f008110030CD00`4300<1=0030@D00`4`00811P030@D00`4300008P000`40
000700040@000@L00P7o0300000R00030@0000L000@100011`020Ol0<00001X0o`5;0@@0000R0003
0@0000L000H10001004500030@0100L000<100003@000`40000<00030@0000d000<100003@000`40
000<00030@0000d000<100003@000`40000<00030@0000d000<100003@000`40000<00030@0000d0
00<100003@000`40000=00030@0000`000<100003@000`40000=00030@0000`000<100002P000280
00<100001`001040004700030@0100L000<1000070000`40000l00030@0003`000<10000?0000`40
000k00030@0000X0000R00030@0000L000@100011P001040004700030@000?l09@00028000<10000
1`001040004600040@000@L000<10000o`0U00008P000`40000700040@000@H000@100011P020Ol0
9`00028000<100001`001040004600040@000@H00P4800030@000?l0700002800P4800040@000@H0
00@100011P020@P000<10000o`0L00008P000`40000700040@000@H000@100011P000`400@070003
0@000?l07000028000<100001`001040004600040@000@H000<100401`000`40000700030@000?l0
4P00028000<100001`001040004600040@000@H000<100401P020@T000<10000o`0B00008P000`40
000700040@000@H000@100011P000`400@0600030@0100P000<10000o`0B00008P000`4000070004
0@000@D000D100000@0600030@0100H000<100401`020Ol05000028000<100001`00104000450005
0@0000401@001040004600030@0100L00P7o01@0000R00030@0000L000@100011@001@40000100D0
00@100011P000`400@070081jP000`40000600030@0001d0000R00811`001@40000100D000D10000
0@0500040@000@H000<100401`000`400@3X00030@0100H000<100407P00028000<100001P001@40
000100D000D100000@0500040@000@D000@100011`000`400@3W00040@000@D000@100017P000280
00<100001P001@40000100D000D100000@0500040@000@D000@100011`000`400@3V00030@000080
00D100000@0400030@0001/0000R00030@0000H000D100000@0500050@0000401@00104000450004
0@000@H000@10001i@000`40000300040@000@H000<100002P000`40000=00008P000`4000060003
0@00008000<100000P001@40000100D000@100011@001040004600040@000MT000<1000020000`40
000500811`000`400008008100<00@003@00028000<100001P000`40000200030@00008000D10000
0@0500040@000@D000D100000@0500040@000MP000<1004020000`40000600030@0000L000<10000
1P001@40000100h0000R00030@0000H000<100000P001@40000100@000<100000P001@40000100D0
00D100000@0500040@000ML000@100011P020A@000<1000010000`40000300030@0000/0000R0003
0@0000H000<100000P001@40000100@000<100000P001@40000100@000<100000P000`4000030004
0@000MH000<100000P000`40000200030@0001@000<100000`000`40000400030@0000/0000R0081
1`000`40000200050@0000401@001@40000100<000<100000P000`40000200030@00008000D10000
0@3E00030@0000<000<100000P4H00050@00004020000`40000:00008P000`40000600030@000080
00D100000@0500050@00004010001@40000100@000<100000P001@4000010=@000<100001@020AX0
00@100415P00028000<100001P000`40000200050@0000401@001@40000100@000D100000@040003
0@00008000<100000P001@400001018000<10000^@000`40000600030@0001X000<100005P000280
00<100001P000`40000200050@0000401@001@40000100@000D100000@0400030@00008000<10000
0P001@400001018000<10000^0000`40000m00008P000`40000600030@00008000D100000@050005
0@00004010001@40000100@000<100000P000`40000200050@0000404@020KT000<10000?P000280
00<100001P000`40000200050@0000401@001@40000100@000D100000@0400030@00008000<10000
0P000`4000020A400P6h00030@0003l0000R00030@0000H000<100000P001@40000100D000D10000
0@0400040@000@D000D100000@0500030@0000814@000`400@2f00030@000400000R00811`000`40
000200050@0000401@001@40000100@000@100011P001040004500060@000@0140000`400@2e0003
0@000440000R00030@0000H000<100000P001@40000100D000@100011@001040004600040@000@D0
00H10001004?00040@000K@000<10000@P00028000<100001P000`40000200050@0000401@001040
004500040@000@H000@100011@001P4000400@l000@10001ZP000`40000600030@0004<0000R0003
0@0000H000<100000P001@40000100D000@100011@001040004600040@000@D000<100400P000`40
000=00040@000JT000<100401`000`40001300008P000`40000600030@00008000@100011P001040
004500040@000@H000@100011@000`400@0200030@0000d000@10001Z0001040004600030@0004@0
000R00030@0000H000<100000P001040004600040@000@D000@100011P000`400@0700810P000`40
000<00030@00008000D100000@2R00030@00008000<100000P000`40001500000P030@H000<10000
10030@D0104400030@0000H000<100000P001040004600040@000@D000@100011P000`400@070081
0P000`40000<00030@00008000D100000@2Q00030@0000@000@10001B0000004004100812`000`40
00020@<000@100000P4300030@0000H000<100000P001040004600040@000@D000<100401`000`40
0@0700810P000`40000<00030@00008000<100000P6P00030@0000D000<10040B@00000300400080
00<100003@000`40000600040@0000<11P000`40000200040@000@H000@100011@000`400@070003
0@0100L000D100000@0400030@0000L000<100000P001P4000400Ih000<100001`000`4000180000
00<00@000P000`40000=00030@0000H000D100000@0800030@00008000@100011P00104000450003
0@0100P00P4700030@00008000D100000@0800030@0000<000H10001006M00030@0005<000000`01
000200030@0000/00P4500030@0000810`000`40000600030@00008000@100011P00104000460081
20020@L000<100000P001@40000100P000<100000`001P4000400Id000<10000D`00000300400080
00<100003@000`40000200@110000`40000600030@00008000@100011P0010400046008120000`40
000;00030@00008120000`40000300030@01008000<10000VP000`40001D000000<00@000P000`40
000=00030@00008000<100001@000`40000600030@00008000@100011P000`400@07008120000`40
000;00060@000@011`000`40000400810P000`40002I00030@0005D0000010010@020@/000D10000
0@0400030@0000D000<100001P000`40000200040@000@H000<100401`020@P000<100002`001P40
00400@H000<100001@020@8000<10000V0000`40001F00000P030@d00`4500@110000`4000060003
0@00008000@100011P000`400@0700815P001P4000400@H000<100001@000`40000200030@0009H0
00<10000E`0002800P4700030@00008000@100011P000`400@0700030@0001D000H1000100460003
0@0000D000<100000P000`40002E00030@0005P0000R00030@0000H000<100000P00104000470081
1`000`40000E00030@01008000<1000010000`40000:00030@0009@000<10000F@00028000<10000
1@000`40000300040@000@L00P4700030@0001D000<100400P000`40000300030@0000/000<10000
U0000`40001I00008P000`40000500030@0000<000@100011`020@L000<100005P020@8000<10000
0`000`40000;00030@0009<000<10000FP00028000<100001@000`40000300040@000@L00P4P0081
0P000`40000300030@0000/000<10000TP000`40001K00008P000`40000500030@0000<000@10001
1`020B0000<100000P001@40000100l000<10000T0000`40001L00008P000`40000500030@0000<0
00<1004020020B0000<100000P001@40000100l000<10000T0000`40001L00008P000`4000050003
0@0000<000<1004020000`40000O00030@00008000D100000@0?00030@0008l000<10000G@000280
0P4600030@0000<000<1004020000`40000T00040@000A0000<10000SP000`40001N00008P000`40
000500030@0000<000<1004020000`40000T00040@000A0000<10000S@000`40001O00008P000`40
000500030@0000<000<1004020000`40000T00040@000A0000<10000;`000`40001J00030@000600
000R00030@0000D000<100000`000`400@0800030@0002@000@100014@000`40000^00030@0005T0
00<10000H@00028000<100001@000`400004008120000`40000T00030@01018000<100009@000`40
000500030@0102L000<100005P000`40000F00030@000680000R00030@0000D000<1000010020C00
0P4B00030@0002D000<100001@000`400@0V00815`020AL000<10000H`00028000<100001@000`40
00040081<0020A8000<1000090000`400@0500050@0000409@000`400@0F00030@0101H000<10000
H`00028000<100001@000`4000040081<0000`40000A00030@0002@000<100401@001@40000102@0
00@100011@000`40000=00040@000@D000<100003@000`40001T00008P020@H000<1000010020C00
00<100004@000`40000S00040@000@@000<100000P000`40000R00050@0000400`020@h000<10000
0P00104000020@h000<10000I@00028000<100001@000`4000040081A@000`40000R00040@000@@0
00<100000`000`40000P00030@00008000@100010P000`40000:00030@0000<000@100010P000`40
000:00030@0006H0000R00030@0000D000<1000010020DD000<100008@000`40000200040@000@H0
00<100007`000`40000300040@000@8000<100000P000`40000500030@0000@00P4300030@0000X0
00<10000IP00028000<100001@000`4000040081A@000`40000Q00030@00008000@100011`000`40
000F00030@0000D000<1000010020@@000<100000P4600030@0000D000<100000`000`4000080003
0@0006L0000R00030@0000D000<1000010020DD000<100008@000`40000200030@0100P000<10000
5P000`40000400030@0000D000<100000`001P4000400@D000<100002`000`40000700030@0006P0
000R00030@0000D000<1000010020DD000<1000080000`40000300030@0100P000<100005@000`40
0@0500030@0000/000<100400P000`40000200030@0000d000<100001@000`40001Y00008P000`40
000500030@0000@00P5500030@00020000<1000010000`40000800030@0001@000<1004010000`40
000=00810`001040004@00030@0000D000<10000J@0002800P4600030@0000@000<10000A@000`40
000O00030@0000@000<1000020000`40000C00040@000@@000<100003@000`40000200040@000A00
00<1000010000`40001Z00008P000`40000500030@0000@000<10000A@000`40000N00030@000100
00<100004`001@400001008000<100004`000`400@0B00030@00008000<10000J`00028000<10000
1@000`40000400030@0004D000<100007P000`40000@00030@00018000<100000P001040004F0081
4P001@40000106h0000R00030@0000D000<1000010000`40001500030@0001d000<100004P000`40
000@00030@0000<000<100405`000`40000B00040@000Fh0000R00030@0000D000<1000010000`40
001600030@0001`000<100004P000`40000@00030@0000@00P4/00030@0106l0000R00030@0000D0
00<1000010000`40001600030@0001`000<100004P000`40000?00030@0000D000<10000;0020Fl0
000900<11P000`40000500@110000`40000500030@0000@000<10000AP000`40000K00030@0001<0
00<100003P000`40000e00030@0006h00008008100<00@403`000`40000300030@0000D000<10000
10000`40001600030@0001/000<1000050000`40000=00030@000:H0000800050@0000402`060@@0
0`4500030@0000@000<10000A`000`40000200030@0001D000<1000050000`40000<00030@000:L0
000800050@0000402`001@40000100D000<100001@000`40001>00030@00008000<1000050000`40
000E00030@0000`000<10000Y`0000P000D100000@0<00040@000@D000<100001@000`40001>0004
0@0000815P000`40000F00030@0000X000<10000Z00000P000D100000@0<008100<00@0010000`40
000500030@0004h000L100000@0101@000<100005`000`40000:00030@000:P0000800050@000040
3@000`400@0500030@0000D000<10000C`001P4000400A@000<100005`000`40000900030@000:T0
0008008100<00@403P020@D000<100001@000`40001?00060@000@0150000`40000H00030@0000P0
00<10000Z@0000T00`4?00811@000`40000500030@0004l000<100400P000`40000A00030@0001T0
00<100001`000`40002Z00008P020@H000<10000C`000`400@0200030@00014000<100006@000`40
000700030@000:X0000R00030@0000D000<10000D0020@<000<1000040000`40000I00030@0000H0
00<10000Z`00028000<100001@000`40001@00030@00008000<100003`000`40000K00030@0000D0
00<10000Z`00028000<100001@000`40001@00030@00008000<100003`000`40000K00030@0000@0
00<10000[000028000<100001@000`40001E00030@0000l000<100006`000`40000400030@000:`0
000R00030@0000D000<10000E@000`40000>00030@0001d000<100000P000`40002]00008P000`40
000400030@0005L000<100003@000`40000M00030@00008000<10000[@00028000<1000010000`40
001G00030@0000d000<100007@000`40000200030@000:d0000R00811@000`40001G00030@0000`0
00<100007`001040006`00008P000`40000400030@0005L000<1000030000`40000O00040@000K00
000R00030@0000@000<10000E`000`40000;00030@00020000<10040/@00028000<1000010000`40
001H00030@0000X000<100008@020K40000R00030@0000@000<10000F0000`40000900030@000280
00<10000/000028000<1000010000`40001H00030@0000T000<10000e@00028000<1000010000`40
001H00030@0000P000<10000eP00028000<1000010000`40001I00030@0000L000<10000eP000280
0P4500030@0005T000<100001`000`40003F00008P000`40000400030@0005T000<100001P000`40
003G00008P000`40000400030@0005T000<100001P000`40003G00008P000`40000400030@0005X0
00<100001@000`40003G00008P000`40000400030@0005X000<1000010000`40003H00008P000`40
000400030@0005X000<1000010000`40003H00008P000`40000400030@0005X000<1000010000`40
003H00008P020@D000<10000FP000`40000300030@000=T0000R00030@0000@000<10000F`000`40
000200030@000=T0000R00030@0000@000<10000F`000`40000200030@000=T0000R00030@0000@0
00<10000F`001@4000010=`0000R00030@0000@000<10000F`001@4000010=`0000R00030@0000@0
00<10000G0001040007L00000P030@H000<100001@040@<0104400030@0000@000<10000G0001040
007L000000@00@400P4?00050@0000400`020@<000<1000010000`40001L00030@010=d000000`01
000200030@0000T01P4700040@0000<110000`40001L00030@010=d000000`01000200030@0000T0
00D100000@0800050@0000401P000`40001M0081g@0000030040008000<100002P00104000440003
0@0000810`000`40000400030@0005d000<10000g00000030040008000<100002P020@03004000<0
104400030@0000@000<10000G@000`40003L000000<00@000P000`40000;00030@0100@000<10000
1@000`40000400030@000?l0?@000004004100813P020@@000<100001@000`40000400030@000?l0
?@0000800`4?008110040@@000<1000010000`40003o03d0000R00811@000`40003o03d0000R0003
0@0000@000<10000o`0m00008P000`40000400030@000?l0?@00028000<1000010000`40003o03d0
000R00030@0000@000<10000o`0m00008P000`40000400030@000?l0?@00028000<1000010000`40
003o03d0000R00030@0000@000<10000o`0m00008P020@D000<10000o`0m00008P000`4000040003
0@000?l0?@00028000<100000`000`40003o03h0000R00030@0000<000<10000o`0n00008P000`40
000300030@000?l0?P00028000<100000`000`40003o03h0000R00030@0000<000<10000o`0n0000
8P000`40000300030@000?l0?P0002800P4400030@000?l0?P00028000<100000`000`40003o03h0
000R00030@0000<000<10000o`0n00008P000`40000300030@000?l0?P00028000<100000`000`40
003o03h0000R00030@0000<000<10000o`0n00008P000`40000300030@000?l0?P0002800P440003
0@000?l0?P00028000<100000`000`40003o03h0000R00030@0000<000<10000o`0n00008P000`40
000300030@000?l0?P00028000<100000`000`40003o03h0000R00030@0000<000<10000o`0n0000
2@030@H000<1000010040@D000<100000`000`40003o03h00008008100<00@402`00104000020@@0
00<100000`000`40003o03h0000800050@00004040000`40000200<10`000`40003o03h000080005
0@00004040000`40000200030@0000<000<10000o`0n000020001@40000100`000<100000P440003
0@0000<000<10000o`0n000020001@40000100`0104500030@0000<000<10000o`0n000020001@40
000100`000<100001P000`40000300030@000?l0?P0000P00P400`010@0<00030@0000H000<10000
0`000`40003o03h0000900<13@040@D000<100000`000`40003o03h0000R008110000`40003o03h0
000R00030@0000<000<10000o`0n00008P000`40000300030@000?l0?P000?l0J@000?l0J@000?l0
J@000?l0J@000?l0J@000?l0J@0009h0104j00@1R0000:0000<10000?@000`4000250000X0020Ch0
00<10000Q@0005X000<100400P4400@10`050@@0104300030@0000<120040@@0104=00030@0000<0
104300<100<00@400P060@P0104300H10P050@@00`40100100050@800`401@010@4000L10P050@P0
0`400`010@0200@110040@<01P530000FP020@<000D100000@0300030@00008000<100000`020@<0
00@100010P020@T000D100000@0200810P020@/00P4400810P020@8000<100000P4500030@0000P0
0P42008110000`40000500030@0000<00P42008110000`40000300050@0000400P000`4000020003
0@0000<000<100001`020@800P4200810P020@800P42008110000`4000140000FP000`4000020005
0@00004020000`40000300030@0000D000<100402`000`40000400030@00008000<100002@000`40
0@0300030@00008000@100010`000`40000300030@0000P000<100000P000`40000200030@0000D0
00<100000`000`40000200030@00008000<100000`001@40000100<01@4500030@0000L000<10000
0P001040004400040@000@@000<100000P000`4000140000G0040@<000<100001P000`4000030003
0@00008000<100000P4<00030@0000@000<100000P000`40000800040@000@<000<100000P001040
004300030@0000<000<1000020000`40000200030@00008000<100001@000`40000300030@000080
00<100000P000`40000300050@0000401`000`40000300030@0000L000<100000P00104000440004
0@000@@000<100000P000`4000140000FP030@H000<100001P000`40000300810P020@8000@10041
2P000`40000400810P020@X000D100000@0200810P020@8000D100000@0500810P000`4000040081
0P020@@00P4200050@0000401@020@800P4400030@0000<000D100000@0300050@0000401@000`40
000700810P020@800P4200810P020@800P4400810P000`4000100000FP000`40000200030@0100D1
0`030@H00`401P01000100<11`060@@0104:00810P020@<01042008100@00@410`030@0400410@L0
104300<100D00@4100030@H00`4010010@030@@01@4500<11P000`40000800<100<00@000P040@@0
104300<100@00@41@P0005X000@100000P4300030@0001@000<100002`000`40001b00030@0000`0
00<10000F@000001\
\>"],
  ImageRangeCache->{{{108, 395}, {298.188, 121.25}} -> {-53.964, 0.44209, 
  0.398872, 0.00164674}}]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Show[swap, stick, {PlotLabel -> \"Comparing the two strategies\",
                   DisplayFunction -> $DisplayFunction}] ;\
\>", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.00952381 0.0147151 0.588604 [
[.21429 .00222 -6 -9 ]
[.21429 .00222 6 0 ]
[.40476 .00222 -6 -9 ]
[.40476 .00222 6 0 ]
[.59524 .00222 -6 -9 ]
[.59524 .00222 6 0 ]
[.78571 .00222 -6 -9 ]
[.78571 .00222 6 0 ]
[.97619 .00222 -9 -9 ]
[.97619 .00222 9 0 ]
[.01131 .13244 -18 -4.5 ]
[.01131 .13244 0 4.5 ]
[.01131 .25016 -18 -4.5 ]
[.01131 .25016 0 4.5 ]
[.01131 .36788 -18 -4.5 ]
[.01131 .36788 0 4.5 ]
[.01131 .4856 -18 -4.5 ]
[.01131 .4856 0 4.5 ]
[.01131 .60332 -6 -4.5 ]
[.01131 .60332 0 4.5 ]
[.5 .63053 -86 0 ]
[.5 .63053 86 10.125 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.21429 .01472 m
.21429 .02097 L
s
[(20)] .21429 .00222 0 1 Mshowa
.40476 .01472 m
.40476 .02097 L
s
[(40)] .40476 .00222 0 1 Mshowa
.59524 .01472 m
.59524 .02097 L
s
[(60)] .59524 .00222 0 1 Mshowa
.78571 .01472 m
.78571 .02097 L
s
[(80)] .78571 .00222 0 1 Mshowa
.97619 .01472 m
.97619 .02097 L
s
[(100)] .97619 .00222 0 1 Mshowa
.125 Mabswid
.07143 .01472 m
.07143 .01847 L
s
.11905 .01472 m
.11905 .01847 L
s
.16667 .01472 m
.16667 .01847 L
s
.2619 .01472 m
.2619 .01847 L
s
.30952 .01472 m
.30952 .01847 L
s
.35714 .01472 m
.35714 .01847 L
s
.45238 .01472 m
.45238 .01847 L
s
.5 .01472 m
.5 .01847 L
s
.54762 .01472 m
.54762 .01847 L
s
.64286 .01472 m
.64286 .01847 L
s
.69048 .01472 m
.69048 .01847 L
s
.7381 .01472 m
.7381 .01847 L
s
.83333 .01472 m
.83333 .01847 L
s
.88095 .01472 m
.88095 .01847 L
s
.92857 .01472 m
.92857 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .13244 m
.03006 .13244 L
s
[(0.2)] .01131 .13244 1 0 Mshowa
.02381 .25016 m
.03006 .25016 L
s
[(0.4)] .01131 .25016 1 0 Mshowa
.02381 .36788 m
.03006 .36788 L
s
[(0.6)] .01131 .36788 1 0 Mshowa
.02381 .4856 m
.03006 .4856 L
s
[(0.8)] .01131 .4856 1 0 Mshowa
.02381 .60332 m
.03006 .60332 L
s
[(1)] .01131 .60332 1 0 Mshowa
.125 Mabswid
.02381 .04415 m
.02756 .04415 L
s
.02381 .07358 m
.02756 .07358 L
s
.02381 .10301 m
.02756 .10301 L
s
.02381 .16187 m
.02756 .16187 L
s
.02381 .1913 m
.02756 .1913 L
s
.02381 .22073 m
.02756 .22073 L
s
.02381 .27959 m
.02756 .27959 L
s
.02381 .30902 m
.02756 .30902 L
s
.02381 .33845 m
.02756 .33845 L
s
.02381 .39731 m
.02756 .39731 L
s
.02381 .42674 m
.02756 .42674 L
s
.02381 .45617 m
.02756 .45617 L
s
.02381 .51503 m
.02756 .51503 L
s
.02381 .54446 m
.02756 .54446 L
s
.02381 .57389 m
.02756 .57389 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
gsave
.5 .63053 -147 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.125000 moveto
 558.000000 14.125000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.125000 lineto
closepath
clip newpath
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Comparing) show
123.000000 11.562500 moveto
(the) show
147.000000 11.562500 moveto
(two) show
171.000000 11.562500 moveto
(strategies) show
231.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.03333 .01472 m
.04286 .30902 L
.05238 .40712 L
.0619 .45617 L
.07143 .36788 L
.08095 .40712 L
.09048 .43515 L
.1 .38259 L
.10952 .40712 L
.11905 .42674 L
.12857 .38928 L
.1381 .40712 L
.14762 .42221 L
.15714 .3931 L
.16667 .40712 L
.17619 .41938 L
.18571 .39558 L
.19524 .40712 L
.20476 .38646 L
.21429 .39731 L
.22381 .37909 L
.23333 .38928 L
.24286 .39859 L
.25238 .40712 L
.2619 .39142 L
.27143 .39957 L
.28095 .38532 L
.29048 .37208 L
.3 .35976 L
.30952 .36788 L
.31905 .35649 L
.32857 .3458 L
.3381 .33577 L
.34762 .34364 L
.35714 .35106 L
.36667 .35807 L
.37619 .3647 L
.38571 .37098 L
.39524 .37693 L
.40476 .38259 L
.41429 .37362 L
.42381 .37909 L
.43333 .3843 L
.44286 .3759 L
.45238 .36788 L
.4619 .3602 L
.47143 .35285 L
.48095 .35807 L
.49048 .36307 L
.5 .36788 L
Mistroke
.50952 .37249 L
.51905 .37693 L
.52857 .3701 L
.5381 .37442 L
.54762 .37858 L
.55714 .38259 L
.56667 .37614 L
.57619 .38006 L
.58571 .37386 L
.59524 .37769 L
.60476 .37174 L
.61429 .37547 L
.62381 .37909 L
.63333 .38259 L
.64286 .37693 L
.65238 .38036 L
.6619 .37491 L
.67143 .36961 L
.68095 .373 L
.69048 .37629 L
.7 .37948 L
.70952 .38259 L
.71905 .38562 L
.72857 .38856 L
.7381 .39142 L
.74762 .39421 L
.75714 .39693 L
.76667 .39957 L
.77619 .40215 L
.78571 .39731 L
.79524 .39985 L
.80476 .40233 L
.81429 .40475 L
.82381 .40712 L
.83333 .40943 L
.84286 .41168 L
.85238 .40712 L
.8619 .40935 L
.87143 .41153 L
.88095 .41366 L
.89048 .41574 L
.9 .41138 L
.90952 .41345 L
.91905 .41547 L
.92857 .41125 L
.9381 .40712 L
.94762 .40914 L
.95714 .41112 L
.96667 .41306 L
.97619 .40908 L
Mistroke
Mfstroke
.03333 .60332 m
.04286 .30902 L
.05238 .21092 L
.0619 .16187 L
.07143 .25016 L
.08095 .21092 L
.09048 .18289 L
.1 .23544 L
.10952 .21092 L
.11905 .1913 L
.12857 .22875 L
.1381 .21092 L
.14762 .19582 L
.15714 .22493 L
.16667 .21092 L
.17619 .19865 L
.18571 .22246 L
.19524 .21092 L
.20476 .23157 L
.21429 .22073 L
.22381 .23895 L
.23333 .22875 L
.24286 .21945 L
.25238 .21092 L
.2619 .22661 L
.27143 .21846 L
.28095 .23272 L
.29048 .24595 L
.3 .25828 L
.30952 .25016 L
.31905 .26155 L
.32857 .27223 L
.3381 .28226 L
.34762 .27439 L
.35714 .26697 L
.36667 .25997 L
.37619 .25334 L
.38571 .24706 L
.39524 .2411 L
.40476 .23544 L
.41429 .24441 L
.42381 .23895 L
.43333 .23373 L
.44286 .24213 L
.45238 .25016 L
.4619 .25783 L
.47143 .26518 L
.48095 .25997 L
.49048 .25496 L
.5 .25016 L
Mistroke
.50952 .24554 L
.51905 .2411 L
.52857 .24794 L
.5381 .24362 L
.54762 .23945 L
.55714 .23544 L
.56667 .2419 L
.57619 .23798 L
.58571 .24417 L
.59524 .24035 L
.60476 .2463 L
.61429 .24256 L
.62381 .23895 L
.63333 .23544 L
.64286 .2411 L
.65238 .23767 L
.6619 .24313 L
.67143 .24843 L
.68095 .24504 L
.69048 .24175 L
.7 .23855 L
.70952 .23544 L
.71905 .23242 L
.72857 .22948 L
.7381 .22661 L
.74762 .22382 L
.75714 .22111 L
.76667 .21846 L
.77619 .21588 L
.78571 .22073 L
.79524 .21818 L
.80476 .2157 L
.81429 .21328 L
.82381 .21092 L
.83333 .20861 L
.84286 .20635 L
.85238 .21092 L
.8619 .20869 L
.87143 .20651 L
.88095 .20438 L
.89048 .20229 L
.9 .20665 L
.90952 .20459 L
.91905 .20257 L
.92857 .20679 L
.9381 .21092 L
.94762 .20889 L
.95714 .20691 L
.96667 .20497 L
.97619 .20895 L
Mistroke
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40005X0000gPP10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdk34g_ojn_l004J2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_P00oiT0IP3o0;C^]:JV
YZf][Jo^kP20P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;002k0?mT0]d009T0V@00
gNk^k^[Zj]gMgK^k^jZZZXR8R7MgMeEEEDA4A14A4Oooc?o<VLbIIS?oVC>IIS=VIYWoo`2Ioc<0VLc<
oiUVVNoGk`00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2o
o`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no
?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo
?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo
?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0
_ol0onj2k_Z0L_lo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoO`0l0>o<knl1k`1]
0:2PX:[Xo`3XooooP;Co]4Bkooo;og?X`?nXOOnCOOno_onooooo?oooOooo_mWEc:0Pl=WIfJj^[YWo
09Wo<iWoIYWoVIWoc<`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aV
VLaVc<aVolbI0<bI<lbIVLbIc<bIolc<0<c<<lc<IP000?l0J@000?l0J@000?l0J@000?l0J@000?l0
J@0005X01@4400<1<P040@<00`4b00<110030C400`4500<1;0050@@00`4500<13P0005X000D10000
0@03008100<00@40<`00104000020@030041034000@100010P020@03004102l000D100000@030081
00<00@40;@000`400003008100<00@400`020@03004100d0001K00030@0000@000D100000@0_00H1
0P001@400001030000D100000@0200050@000040;`001@40000100<000D100000@0]00030@0000<0
00D100000@0300050@0000403@0005`000<100000`001@40000102l000D100000@0300050@000040
<0020@8000@100010`000`40000]00050@0000400`001@40000102d000<100000`001@40000100<0
00D100000@0=0000G@000`40000200050@000040<0001040004300050@000040<0020@8000@10001
0`000`40000^00<110001@40000102d000<100000`001@40000100<000D100000@0=0000GP001@40
000100<000<10000;P020@030040008000D100000@0`00040@010@<000D100000@0_00050@000040
0`001@40000102d000<100000`001@40000100<000D100000@0=00009@000`40000b00050@000040
0`001@400001034000<100400`001@40000103000P4500050@000040;`001@40000100<000D10000
0@0]00030@0000<000D100000@0300050@0000403@0002D000<10000<P000`4000020@<00P400`01
0@0b00810`020@030041034000<100000`020@03004102l000D100000@03008100<00@40:`000`40
0@05008100<00@400`020@03004100d0000U00030@0003<00`4500<1<`020@@00`4c00<10`030C40
0`4500<1;@020@H00`4500<13P0002D000<10000o`1100007@3o0Cd13`0002D000@100012`000`40
000<00030@0000`000<1000030000`40000<00030@0000`000<1000030000`40000<00030@0000`0
00<1000030000`40000<00030@0000`000<100002`000`40000<00030@0000`000<1000030000`40
000<00030@0000`000<1000030000`40000<00030@0001D0000U00040@000CP000<10000>@000`40
000i00030@0003P000<10000>@000`40000E00009@001040007o0400000U00040@000Ol0@00002D0
00@10001o`1000009@001040007o0400000U00040@000Ol0@00002D000@10001o`1000009@001040
007o0400000U008100<00@00o`0o00009@001040007o0400000U00040@000Ol0@00002D000@10001
o`1000009@001040007o0400000U00040@000Ol0@00002D000D100000@3o03l0000U00050@000040
o`0o00009@001@4000010?l0?`0002D00P4200030@000?l0?@0002D000D100000@3o03l0000U0005
0@000040o`0o00009@001@4000010?l0?`0002D000D100000@3o03l0000U00050@000040o`0o0000
9@001@4000010?l0?`0002D000D100000@3o03l0000U00050@000040o`0o00009@020@8000<10000
o`0m00009@001@4000010?l0?`0002D000D100000@3o03l0000U00050@000040o`0o00009@001@40
00010?l0?`0002D000D100000@3o03l0000U00050@000040o`0o00003@030@H000<100000`050@@0
00D100000@3o03l0000<008100<00@402`001@40000100@000D100000@3o03l0000<00050@000040
30000`40000500<100<00@00o`0n000030001@40000100d000<1000010001@4000010?l0?`0000`0
00D100000@0>00030@0000<000D100000@3o03l0000<00050@0000403`000`40000200050@000040
o`0o000030001@40000100/000D100000@0400050@000040o`0o000030020@03004100/000<10000
0P4400050@000040o`0o00003@030@d00`4500050@000040o`0o00009@001@4000010?l0?`0002D0
00D100000@3o03l0000U00050@000040o`0o00009@020@<000<1000010000`40003o03D0000U0003
0@00008000<1000010000`40003o03D0000U00030@00008000<1000010000`40003o03D0000U0003
0@00008000<100000`020Ol0=`0002D000<100000P000`4000030081o`0g00009@000`4000020003
0@0000<000<10040o`0f00009@000`40000200030@0000<000<100401`000`40003o02`0000U0003
0@00008000<100000`000`400@0700030@000?l0;00002D000<100000P000`40000200040@000@H0
0P7o02h0000U00810`000`40000200040@000@H000<100401`000`40003o02<0000U00030@000080
00<100000P001040004600030@0100L000<100001P000`40003o01X0000U00030@00008000<10000
0P001040004500040@000@H000<100401P020@P000<10000o`0A00009@000`40000200030@000080
00@100011@001040004600030@0100H000<100401P020Mh00P470081:@0002D000<100000P001@40
000100<000<100000`001040004500040@000@D000@100011P000`400@3J00<10P020@800`420081
2`030AT0000U00030@00008000D100000@0400050@0000400`000`40000300040@000@D000@10001
1@001040007;00H10`060@L00P4700810`060@<00P4G00009@000`40000200050@00004010001@40
000100@000D100000@0400050@00004010001@40000100@000D100000@0A00030@000:l00`4600<1
60030B80000U00030@00008000D100000@0400050@00004010001@40000100@000D100000@040004
0@000@D000H10001004?00030@010:X01P5900009@000`40000200050@00004010001@40000100@0
00@100011@001040004500040@000@D000H10001004>00040@000@@000<10000U`060@<00`5?0000
9@020@<000D100000@0400040@000@D000@100011@001040004600811`020@<000D100000@080003
0@00008000H10001006E00<11P030E80000U00030@00008000D100000@0400040@000@D000@10001
1P020@L000<100001P000`40000200060@000@011P000`40000300030@01008000<10000T0030Eh0
000U00030@00008000D100000@0400040@000@D000<100401`020AH00P4200030@0000<000<10000
1@000`40000200030@0008`00`5Q00009@000`40000200040@000@D000@100011P020@L000<10000
5@000`40000200050@0000403@000`40002900<1I00002D000<100000P001040004500030@0100L0
00<100008`001040004?00030@0002d000<10000E@030FL0000U00030@00008000@100011P020@L0
00<1000090020A4000<100008P020@H00P400`01000S00811@000`40000=00<10`020A000`5Z0000
9@000`40000200040@000@H00P4^00030@00014000<100007`020@800P42008110000`40000F0003
0@0000D00`42008100@00@400P4200811`020@<00`4200812`030Fd0000=00<11P000`40000500@1
0`000`40000200040@000@H00P5200030@0001d00P4600811`000`40000C008100<00@000`020@L0
00<100000`020@800P4200<130000`40000500<1L00000`00P400`010@0?00030@00008000<10000
0P001040004600030@00048000<100006`000`40000@00814@020@@000@100415@020A000P4200<1
L`0000`000D100000@0;00H10`030@8000@100011P000`40001300030@00008000<100004`020AD0
00<1000030020@L000<10000:@020GH0000<00050@0000402`001@40000100@000<100000P001040
005=00060@000@014`000`40000F00030@0000T00P6]000030001@40000100`000@1000110000`40
000200040@000Dd000<100400`000`40000>00816P000`4000060081[`0000`000D100000@0<0081
00<00@000`000`40000200040@000Dh000<100000P000`40000=00030@0001/000<100000`020K40
000<00050@0000403@000`400@0400030@0000<00P5E00030@0000/000<100007@020@0300410;<0
000<008100<00@403P020@@000<100000`020EH000<100002@000`40000P00030@000;<0000=00<1
3`020@@000<100000`020EL000<100001`000`40003G00009@000`4000030081F0000`4000040081
fP0002D000<100000`020ET000<100000P000`40003J00009@000`4000030081FP0010400@7M0000
9@020@@00P5K00030@000=d0000U00030@0000<00P7o03`0000U00030@0000<00P7o03`0000U0003
0@0000<00P7o03`0000U00030@0000<000<10000o`0k00009@000`40000300030@000?l0>`0002D0
00<100000`000`40003o03/0000U00030@0000<000<10000o`0k00009@000`40000300030@000?l0
>`0002D00P4400030@000?l0>`0002D000<100000`000`40003o03/0000U00030@0000<000<10000
o`0k00009@000`40000300030@000?l0>`0002D000<100000`000`40003o03/0000U00030@0000<0
00<10000o`0k00009@000`4000030081o`0l00009@000`4000030081o`0l00009@000`4000030081
F`000`40003M00009@020@@00P5J00030@010=h0000U00030@0000<00P5I00050@000040g@0002D0
00<100000`020EP000<100000`000`40003J00009@000`4000030081E`000`4000050081fP0002D0
00<100000`020EL000<100001`000`40003G00009@000`4000030081EP000`40000900818@020K@0
000U00030@0000<00P5E00030@0000`000<1000070020@800P6b00003@030@H000<100001@030@@0
00<100000P001040005>00030@00008000<100003P020A/000<1000010020K00000<008100<00@40
3@001040004300030@00008000@10001C@001P400@400A<000<100005P020@T00P6^000030001@40
000100`000D100000@0300<10P001040004600030@0004@000D100000@0E00815@000`40000;0081
[00000`000D100000@0<00810P001@40000100@000@100011P000`40001300030@0001X000<10000
4@000`40000>00811P020BX00P5f000030001@40000100`00P4200050@0000401000104000460003
0@00048000<1000070020@H000<100001`000`40000A00810P020@800P4>00810`030A000P4200<1
L`0000`000D100000@0<00040@010@@000<100000P00104000460081@P000`40000O00030@000080
00@100411P000`40000D00811P020@H00P4200810P030@<00`4600810`020@L00`5`000030001@40
000100`00P4600030@00008000@100011P020Bh000<100004@000`40000P008100<00@000`020@80
0P4Q00810P020@800P4=00<100<00@400P030@`00`5]000030020@03004100d000<1000010000`40
000200040@000@D000<100401`000`40000T00814@000`40000S00030@0000D00P4U00815P000`40
000D00<1JP0000d00`4?00<10`000`40000200040@000@D000<100401`000`40000N00030@000080
00@100013`000`40002500<1I`0002D000<100000P001040004500040@000@H00P4700030@0001D0
0P4300050@0000403@000`40002900<1I00002D000<100000P001@40000100@000@100011@000`40
0@0700811`000`40000;00060@000@011@000`40000500030@00008000<10000S0030F40000U0003
0@00008000D100000@0400040@000@D000@100011P020@L00P4700030@00008000H1000100460003
0@0000<000H10041006B00H1F`0002D00P4300050@00004010001040004500040@000@D000@10001
1@000`400@0700810P000`40000200030@0000H000<100000P001@40000109T00`4300H1C`0002D0
00<100000P001@40000100@000D100000@0400040@000@D000@100011@001040004500060@000@01
3P001040006Q00<11P030D`0000U00030@00008000D100000@0400050@00004010001@40000100@0
00D100000@0400040@000@D000<100000P4@00030@010:d00`5900009@000`40000200050@000040
10001@40000100@000D100000@0400050@0000400`000`40000200050@00004010000`40000?0003
0@000:l01P4300H15@060@H00P4G00009@000`40000200050@00004010001@40000100<000<10000
0`001040004500040@000@D000@10001cP030@H00`4600<11P030@H01P4I00009@000`4000020005
0@0000400`000`40000300040@000@D000@100011P000`400@060081f`060@<01P4X00009@000`40
000200030@00008000@100011@001040004600030@0100H00P4800030@000?l04@0002D000<10000
0P000`40000200040@000@H000<100401P000`400@0700030@000?l06P0002D000<100000P000`40
000200040@000@H000<100401`000`40003o02<0000U00810`000`40000200040@000@H00P480003
0@000?l08`0002D000<100000P000`40000200040@000@L000<10000o`0/00009@000`4000020003
0@0000<000<100401`000`40003o02`0000U00030@00008000<100000`000`400@3o03H0000U0003
0@00008000<100000`000`400@3o03H0000U00030@00008000<100000`020Ol0=`0002D000<10000
0P000`4000030081o`0g00009@000`40000200030@0000@000<10000o`0e00009@000`4000020003
0@0000@000<10000o`0e00009@020@<000<1000010000`40003o03D0000U00050@000040o`0o0000
9@001@4000010?l0?`0002D000D100000@3o03l0000U00050@000040o`0o00009@001@4000010?l0
?`0002D000D100000@3o03l0000=00<11P000`40000400<11@001@4000010?l0?`0000`00P400`01
0@0;00050@00004010001@4000010?l0?`0000`000D100000@0;00050@00004010030@0300400?l0
?P0000`000D100000@0;00050@00004010001@4000010?l0?`0000`000D100000@0<00<11@001@40
00010?l0?`0000`000D100000@0;00050@00004010001@4000010?l0?`0000`000D100000@0;0005
0@00004010001@4000010?l0?`0000`00P400`010@0;00050@00004010001@4000010?l0?`0000d0
0`4=00<11@001@4000010?l0?`0002D000D100000@3o03l0000U00050@000040o`0o00009@001@40
00010?l0?`0002D00P4200030@000?l0?@0002D000D100000@3o03l0000U00050@000040o`0o0000
9@001@4000010?l0?`0002D000D100000@3o03l0000U00050@000040o`0o00009@001@4000010?l0
?`0002D000D100000@3o03l0000U00050@000040o`0o00009@020@8000<10000o`0m00009@001@40
00010?l0?`0002D000D100000@3o03l0000U00040@000Ol0@00002D000@10001o`1000009@001040
007o0400000U00040@000Ol0@00002D000@10001o`1000009@001040007o0400000U008100<00@00
o`0o00009@001040007o0400000U00040@000Ol0@00002D000@10001o`1000009@001040007o0400
000U00040@000Ol0@00002D000@10001o`1000006`050@D000@10001o`1000007@000`4000050004
0@000Ol0@00001d000<100001@040Ol0@00001d000<100001@000`40003o0440000M00030@0000D0
00<10000o`1100007@000`40000500030@000?l0@@0001d000<100001@000`40003o0440000K0003
0@0100L000<10000o`11000070020Ol0B`000?l0J@000?l0J@000?l0J@000?l0J@000?l0J@0006X0
104V00@1L`040ET0001[00030@0002X000<10000M0000`40001F0000J`000`40000Z00030@0007@0
00<10000EP0005D00`4400@100D00@4100020@06004100400`4200L10P060@801@4200<100@00@41
0`030@03004000D0104200<100@00@410`040@X01043008100<00@4010040@T01@4400@10`060@03
004100D10`040@@0104400<100@00@001@4400@10`050D40001D00050@0000400P020@800P401@01
0001008000<100000P42008100<00@000`000`40000300030@0000D000<100000`001@40000100<0
0P4200811P001@400001008000D100000@0300810`000`40000700050@0000400P020@03004100<0
0P42008120020@<000D100000@0300030@00008000<100000P000`40000200030@00008000D10000
0@0200810`000`4000020@800P4400030@0000<00P4300030@0000810`000`40000n0000D`000`40
000200040@000@@000<100400P001040004200030@00008000<100001@4500030@0000D000<10000
0`001@40000100<000<100000P000`40000400030@0000@000D100000@0300030@0000`000<10000
10020@03004100<000<100000P000`40000600030@00008000D100000@0800030@0000<01@440003
0@0000@000<100001@000`40000200030@00008000<100000`000`40000500030@00008000<10000
?P0005<000<100001@000`40000200030@01008000@100010P000`40000200030@0000@000<10000
0`000`40000500030@0000<000D100000@0300030@00008000<1000010000`40000400050@000040
0`060@T000<1000010001@400@0100<000<100000P000`40000600D110000`40000600030@0000L0
00<100000P000`40000400H10P000`40000200030@00008000<100000`060@801@510000D`000`40
000500810P020@05004000400P000`4000020@800P4200050@0000401@020@8000D100000@050005
0@0000400`020@800P4600030@0000@000D100000@0300030@00008000<100001`000`4000030004
0@000@8000<100000P42008120000`40000200050@00004020020@8000<100400`000`4000020003
0@0000@000<100000P000`4000020@800P4400030@0000<000<100000P001040004400030@0003h0
001C00030@00008000@1000010400`010@060@04004100<110030@@00`401@010@4000<110050@D0
0`400`010@0300H10`040@D0104800H10P020@<00P4300@12P050@03004100@10`030@0400410@80
0`4300H110040@@00`4010010@030@H0104400D1@00005<00P420081AP000`40000400030@0001@0
00<100008`000`40000E00030@0006X00000\
\>"],
  ImageRangeCache->{{{108, 395}, {529.375, 352.438}} -> {-57.6765, 2.30183, 
  0.418048, 0.00676416}}]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Histogram[Table[Monte[30][[30]][[1]],{100}],
          Table[{i,i+.10},{i,0,.90,.10}],
          Table[i,{i,0,.90,.10}],
          True] ;\
\>", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.0414698 0.11385 [
[.07143 .03077 -3 -9 ]
[.07143 .03077 3 0 ]
[.16667 .03077 -9 -9 ]
[.16667 .03077 9 0 ]
[.2619 .03077 -9 -9 ]
[.2619 .03077 9 0 ]
[.35714 .03077 -9 -9 ]
[.35714 .03077 9 0 ]
[.45238 .03077 -9 -9 ]
[.45238 .03077 9 0 ]
[.54762 .03077 -9 -9 ]
[.54762 .03077 9 0 ]
[.64286 .03077 -9 -9 ]
[.64286 .03077 9 0 ]
[.7381 .03077 -9 -9 ]
[.7381 .03077 9 0 ]
[.83333 .03077 -9 -9 ]
[.83333 .03077 9 0 ]
[.92857 .03077 -9 -9 ]
[.92857 .03077 9 0 ]
[.07143 .05217 -11 0 ]
[.07143 .05217 11 9.8125 ]
[.16667 .05217 -11 0 ]
[.16667 .05217 11 9.8125 ]
[.2619 .05217 -11 0 ]
[.2619 .05217 11 9.8125 ]
[.35714 .05217 -11 0 ]
[.35714 .05217 11 9.8125 ]
[.45238 .09771 -11 0 ]
[.45238 .09771 11 9.8125 ]
[.54762 .30264 -14 0 ]
[.54762 .30264 14 9.8125 ]
[.64286 .58727 -14 0 ]
[.64286 .58727 14 9.8125 ]
[.7381 .31403 -14 0 ]
[.7381 .31403 14 9.8125 ]
[.83333 .09771 -11 0 ]
[.83333 .09771 11 9.8125 ]
[.92857 .05217 -11 0 ]
[.92857 .05217 11 9.8125 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
.667 g
.02381 .04147 m
.02381 .04147 L
.11905 .04147 L
.11905 .04147 L
F
.11905 .04147 m
.11905 .04147 L
.21429 .04147 L
.21429 .04147 L
F
.21429 .04147 m
.21429 .04147 L
.30952 .04147 L
.30952 .04147 L
F
.30952 .04147 m
.30952 .04147 L
.40476 .04147 L
.40476 .04147 L
F
.40476 .04147 m
.40476 .08701 L
.5 .08701 L
.5 .04147 L
F
.5 .04147 m
.5 .29194 L
.59524 .29194 L
.59524 .04147 L
F
.59524 .04147 m
.59524 .57656 L
.69048 .57656 L
.69048 .04147 L
F
.69048 .04147 m
.69048 .30332 L
.78571 .30332 L
.78571 .04147 L
F
.78571 .04147 m
.78571 .08701 L
.88095 .08701 L
.88095 .04147 L
F
.88095 .04147 m
.88095 .04147 L
.97619 .04147 L
.97619 .04147 L
F
0 g
.002 w
[ ] 0 setdash
.02381 .04147 m
.02381 .04147 L
.11905 .04147 L
.11905 .04147 L
s
.11905 .04147 m
.11905 .04147 L
.21429 .04147 L
.21429 .04147 L
s
.21429 .04147 m
.21429 .04147 L
.30952 .04147 L
.30952 .04147 L
s
.30952 .04147 m
.30952 .04147 L
.40476 .04147 L
.40476 .04147 L
s
.40476 .04147 m
.40476 .08701 L
.5 .08701 L
.5 .04147 L
s
.5 .04147 m
.5 .29194 L
.59524 .29194 L
.59524 .04147 L
s
.59524 .04147 m
.59524 .57656 L
.69048 .57656 L
.69048 .04147 L
s
.69048 .04147 m
.69048 .30332 L
.78571 .30332 L
.78571 .04147 L
s
.78571 .04147 m
.78571 .08701 L
.88095 .08701 L
.88095 .04147 L
s
.88095 .04147 m
.88095 .04147 L
.97619 .04147 L
.97619 .04147 L
s
.02381 .04147 m
.97619 .04147 L
s
[(0)] .07143 .03077 0 1 Mshowa
[(0.1)] .16667 .03077 0 1 Mshowa
[(0.2)] .2619 .03077 0 1 Mshowa
[(0.3)] .35714 .03077 0 1 Mshowa
[(0.4)] .45238 .03077 0 1 Mshowa
[(0.5)] .54762 .03077 0 1 Mshowa
[(0.6)] .64286 .03077 0 1 Mshowa
[(0.7)] .7381 .03077 0 1 Mshowa
[(0.8)] .83333 .03077 0 1 Mshowa
[(0.9)] .92857 .03077 0 1 Mshowa
gsave
.07143 .05217 -72 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.) show
75.000000 11.250000 moveto
(%) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
gsave
.16667 .05217 -72 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.) show
75.000000 11.250000 moveto
(%) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
gsave
.2619 .05217 -72 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.) show
75.000000 11.250000 moveto
(%) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
gsave
.35714 .05217 -72 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.) show
75.000000 11.250000 moveto
(%) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
gsave
.45238 .09771 -72 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(4.) show
75.000000 11.250000 moveto
(%) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
gsave
.54762 .30264 -75 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(22.) show
81.000000 11.250000 moveto
(%) show
87.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
gsave
.64286 .58727 -75 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(47.) show
81.000000 11.250000 moveto
(%) show
87.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
gsave
.7381 .31403 -75 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(23.) show
81.000000 11.250000 moveto
(%) show
87.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
gsave
.83333 .09771 -72 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(4.) show
75.000000 11.250000 moveto
(%) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
gsave
.92857 .05217 -72 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.) show
75.000000 11.250000 moveto
(%) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
1 g
.5 Mabswid
.02381 .03612 m
.02381 .01472 L
s
.02381 .58192 m
.02381 .60332 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40005X0000gPP10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdk34g_ojn_l004J2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_P00oiT0IP3o0;C^]:JV
YZf][Jo^kP20P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;002k0?mT0]d009T0V@00
gNk^k^[Zj]gMgK^k^jZZZXR8R7MgMeEEEDA4A14A4Oooc?o<VLbIIS?oVC>IIS=VIYWoo`2Ioc<0VLc<
oiUVVNoGk`00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2o
o`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no
?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo
?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo
?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0
_ol0onj2k_Z0L_lo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoO`0l0>o<knl1k`1]
0:2PX:[Xo`3XooooP;Co]4Bkooo;og?X`?nXOOnCOOno_onooooo?oooOooo_mWEc:0Pl=WIfJj^[YWo
09Wo<iWoIYWoVIWoc<`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aV
VLaVc<aVolbI0<bI<lbIVLbIc<bIolc<0<c<<lc<IP000?l0J@000?l0J@000?l0J@000?l0J@000?l0
J@0001T00`4G00<11P000`40000300D13P030@H000<100000`050@h00`4600030@0000@00`4?00<1
1P000`40000500@13@030@H000<1000010040@h00`4600030@0000D00`4=00<11P000`4000050003
0@0000h00`4600030@0000@00`4?00<11P000`40000400<14`0001P00P400`010@0E008100<00@40
3@000`40000=008100<00@402`001@40000100d00P400`010@0;00030@0000813@020@03004100l0
00<100002`020@03004100/000@100000P4<008100<00@403@001040004;008100<00@403@000`40
000=008100<00@402`001@40000100d00P400`010@0?00030@000100000H00050@0000405@001@40
000100d000<100003@001@40000100`000<100003P001@40000100l000<100002`001@40000100/0
1P4<00050@00004040000`40000:00050@00004030001@40000100/000D100000@0>00030@0000`0
00D100000@0;00050@0000403@001@40000100l00P4A000060001@40000101D000D100000@0=0003
0@0000d000D100000@0=00030@0000d000D100000@0?00030@0000/000D100000@0;00050@000040
3@001@400001010000<100002P001@40000100`00P4200030@0000T000D100000@0>00030@0000`0
00D100000@0;00050@0000403@001@40000100d00P400`01000@000060001@40000101D000D10000
0@0=00030@0000d000D100000@0>00030@0000`000D100000@0=00813P001@40000100`000@10001
3@001@40000100`000<100000P4<00050@00004030020@8000<100002@001@40000100h000<10000
30001@40000100`00`4>00050@00004030000`4000020A40000H00050@0000405@001@40000100d0
00<100003@001@40000100l000<100002`001@40000100l000<100002`001@40000100`00P400`01
000<00050@00004030040@d000D100000@0<00040@010@`000D100000@0>00030@0000`000D10000
0@0;00050@0000403@001@40000100`000<100000P4A000060001@40000101D000D100000@0=0003
0@0000d000D100000@0;00050@0000403@001@40000100l000<100002`001@40000100d000<10040
3@001@40000100`000<100003P001@40000100`00P4>00050@0000403P000`40000<00050@000040
2`001@40000100d000D100000@0<00050@0000404@0001P00P400`010@0E008100<00@402`000`40
0@0?008100<00@402`000`4000020@d00P400`010@0;00050@0000403@020@03004100h00P4=0081
00<00@4030000`40000>008100<00@403@000`40000<008100<00@402`001@40000100d00P400`01
0@0;00050@0000403@020@03004100`000@100014P0001T00`4G00<13@020A000`4=00<13`030@d0
0`4?00<13`020@h00`4=00@13P030@l00`4<00<130050@h00`4=00<13`030@h00`4B0000o`1Y0000
2P3o0E@12`0009400:X1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/0Bd0002A00030B`/01l/00<1;2`07b`00`4/;00O;0030B`/01l/00<1;2`07R`00`40000[0000
T@00ZP6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TB`1;@000940
00<1;2`07b`00`4/;00O;0030B`/01l/00<1;2`07b`00`4/;00N;0030@0002/0002A002Z0I4/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;04]0000T@000`4/;00O;003
0B`/01l/00<1;2`07b`00`4/;00O;0030B`/01h/00<10000:`0009400:X1TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/0Bd0002A00030B`/01l/00<1;2`07b`00`4/
;00O;0030B`/01l/00<1;2`07R`00`40000[0000T@00ZP6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TB`1;@00094000<1;2`07b`00`4/;00O;0030B`/01l/00<1;2`0
7b`00`4/;00N;0030@0002/0000A00<11P000`40000400<13`030@H000<1000010030@l00`460003
0@0000@00`4?00<11P000`40000400<11`00ZP6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TB`11`030@H000<1000010030A<0000@008100<00@4030000`400@0>0081
00<00@4030000`400@0>008100<00@4030000`400@0>008100<00@4030000`400@0700030B`/01l/
00<1;2`07b`00`4/;00O;0030B`/01l/00<1;2`07R`00`400004008100<00@4030000`400@0C0000
40001@40000100`00`4>00050@00004030030@h000D100000@0<00<13P001@40000100`00`47002Z
0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;04600050@000040
30030A<0000@00050@0000402P000`40000@00050@0000402P000`40000@00050@0000402P000`40
000@00050@0000402P000`40000900030B`/01l/00<1;2`07b`00`4/;00O;0030B`/01l/00<1;2`0
7R`00`40000400050@0000402P000`40000E000040001@40000100/0104>00050@0000402`040@h0
00D100000@0;00@13P001@40000100/01047002Z0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;04600050@0000402`040A<0000@00050@00004030001040004=0005
0@00004030001040004=00050@00004030001040004=00050@00004030001040004602<18B`00`4/
;00O;0030B`/01l/8P4600050@00004030001040004B000040001@40000100/000<100403`001@40
000100/000<100403`001@40000100/000<100403`001@40000100/000<10040:P00I`6A;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@409`001@40000100/000<1004050000100
0P400`010@0;00030@0100l00P400`010@0;00030@0100l00P400`010@0;00030@0100l00P400`01
0@0;00030@0102X000<1;2`07b`00`4/;00O;0030B`/01l/00<100009@020@03004100/000<10040
500001400`4=00030@0000l00`4=00030@0000l00`4=00030@0000l00`4=00030@0002T006L1TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94102P00`4=00030@0001<0002c0003
0B`/01l/00<1;2`07b`00`4/;00O;0030@0004`0002c001W0I4/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA0@1>0000/`000`4/;00O;0030B`/01l/00<1;2`07b`00`40001<0000
/`00I`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40CP000;<000<1;2`0
7b`00`4/;00O;0030B`/01l/00<10000C0000;<006L1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94104h0002c00030B`/01l/00<1;2`07b`00`4/;00O;0030@0004`0002J00@1
10000`40000400<11`00I`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40
1`040@@000<1000010030CD0002L00030@0000X000<100401`000`4/;00O;0030B`/01l/00<1;2`0
7b`00`40000700030@0000X000<10040=@0009P01P4;00<11`00I`6A;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/T@401@060@/00`4e0000V0001@40000100X000<100002@000`4/
;00O;0030B`/01l/00<1;2`07b`00`40000300050@0000402P000`40000g0000V@001040004;00@1
1`00I`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@401P001040004;00@1
=@0009T00P400`01000;00040@000@H000<1;2`07b`00`4/;00O;0030B`/01l/00<1000010020@03
004000/000@10001=00009X000<100402`000`400@08001W0I4/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA0@0700030@0100/000<10040=P0009/00P4;00030@0100P000<1;2`0
7b`00`4/;00O;0030B`/01l/00<100001P020@/000<10040=P0009/00P4<00030@0000L006L1TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94100P00P4<00030@0003D0002c0003
0B`/01l/00<1;2`07b`00`4/;00O;0030@0004`0002c001W0I4/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA0@1>0000/`000`4/;00O;0030B`/01l/00<1;2`07b`00`40001<0000
/`00I`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40CP000;<000<1;2`0
7b`00`4/;00O;0030B`/01l/00<10000C0000;<006L1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94104h0002c00030B`/01l/00<1;2`07b`00`4/;00O;0030@0004`0002c001W
0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@1>0000/`000`4/;00O;003
0B`/01l/00<1;2`07b`00`40001<0000/`00I`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/T@40CP000;<000<1;2`07b`00`4/;00O;0030B`/01l/00<10000C0000;<006L1TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94104h0002c00030B`/01l/00<1;2`0
7b`00`4/;00O;0030@0004`0002c001W0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA0@1>0000/`000`4/;00O;0030B`/01l/00<1;2`07b`00`40001<0000/`00I`6A;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40CP000;<000<1;2`07b`00`4/;00O;003
0B`/01l/00<10000C0000;<006L1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94104h0002c00030B`/01l/00<1;2`07b`00`4/;00O;0030@0004`0002c001W0I4/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@1>0000/`000`4/;00O;0030B`/01l/00<1;2`0
7b`00`40001<0000/`00I`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40
CP000;<000<1;2`07b`00`4/;00O;0030B`/01l/00<10000C0000;<006L1TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94104h0002c00030B`/01l/00<1;2`07b`00`4/;00O;003
0@0004`0002c001W0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@1>0000
/`000`4/;00O;0030B`/01l/00<1;2`07b`00`40001<0000/`00I`6A;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/T@40CP000;<000<1;2`07b`00`4/;00O;0030B`/01l/00<10000
C0000;<006L1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94104h0002c0003
0B`/01l/00<1;2`07b`00`4/;00O;0030@0004`0002c001W0I4/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA0@1>0000/`000`4/;00O;0030B`/01l/00<1;2`07b`00`40001<0000
/`00I`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40CP000;<000<1;2`0
7b`00`4/;00O;0030B`/01l/00<10000C0000;<006L1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94104h0002c00030B`/01l/00<1;2`07b`00`4/;00O;0030@0004`0002c001W
0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@1>0000/`000`4/;00O;003
0B`/01l/00<1;2`07b`00`40001<0000/`00I`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/T@40CP000;<000<1;2`07b`00`4/;00O;0030B`/01l/00<10000C0000;<006L1TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94104h0002c00030B`/01l/00<1;2`0
7b`00`4/;00O;0030@0004`0002c001W0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA0@1>0000/`000`4/;00O;0030B`/01l/00<1;2`07b`00`40001<0000/`00I`6A;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40CP000;<000<1;2`07b`00`4/;00O;003
0B`/01l/00<10000C0000;<006L1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94104h0002c00030B`/01l/00<1;2`07b`00`4/;00O;0030@0004`0002c001W0I4/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@1>0000/`000`4/;00O;0030B`/01l/00<1;2`0
7b`00`40001<0000/`00I`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40
CP000;<000<1;2`07b`00`4/;00O;0030B`/01l/00<10000C0000;<08`40A94/TBbA;94/TBbA;94/
TBbA;94/TBbA;94/TBbA;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;941CP000=D0
00<1;2`07b`00`4/;00O;0030@0004`0003E00150I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/T@6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94104h0003E00030B`/01l/00<1;2`0
7b`00`40001<0000e@008P6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94S0Dh0003E0003
0B`/01l/00<10000KP000=D002<1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@1`0000
e@000`4/;00O;0030@0006h0003E000S0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40
L0000=D000<1;2`07b`00`40001^0000e@008`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;9410700002f00D10`050@D000<100000`030@@000<1;2`07b`00`40001^0000]P001@40000100<0
00D100000@0;00030@0100@002<1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@1`0000
]`000`40000500030@0000`00`4400030B`/01l/00<10000KP000;P000<100001@000`4000090003
0@0000H002<1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@1`0000^@000`4000050003
0@0000T0104400030B`/01l/00<100001@4400<11P000`40000300<1D`000;X000<100001@000`40
000900040@000@<002<1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@0200050@000040
0`000`4000020@/000<10040D`000;H000D100000@0300050@0000402P000`400@0500030B`/01l/
00D100000@0:00030@0000T00`5C0000]P000`4000020@<000<100000P4:00030@0100D002<1TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@0400030@0000L000<100001`000`40001E0000
]`030@D00`4<00030@0000@000<1;2`07b`00`40000300030@0000@00P4;00@1D`000=D002<1TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@0600030@0000D000<100002@001040005B0000
e@000`4/;00O;0040@000@<000<100001@000`40000800030@0105@0003E000S0I4/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/T@400P000`4000020@<000D100000@0:00030@0105@0003E0003
0B`/01l/00@100000`4500<130000`40001C0000e@008`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;9410700003E00030B`/01l/00<10000KP000=D002<1TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA0@1`0000e@000`4/;00O;0030@0006h0003E000S0I4/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/T@40L0000=D000<1;2`07b`00`40001^0000e@008`6A;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;9410700003E00030B`/01l/00<10000KP000=D002<1TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@1`0000e@000`4/;00O;0030@0006h0003E000S0I4/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40L0000=D000<1;2`07b`00`40001^0000e@008`6A
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410700003E00030B`/01l/00<10000KP000=D0
02<1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@1`0000e@000`4/;00O;0030@0006h0
003E000S0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40L0000=D000<1;2`07b`00`40
001^0000e@008`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410700003E00030B`/01l/
00<10000KP000=D002<1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@1`0000e@000`4/
;00O;0030@0006h0003E000S0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40L0000=D0
00<1;2`07b`00`40001^0000e@008`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410700
003E00030B`/01l/00<10000KP000=D002<1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
0@1`0000e@000`4/;00O;0030@0006h0003E000S0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/T@40L0000=D000<1;2`07b`00`40001^0000e@008`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;9410700003E00030B`/01l/00<10000KP000=D002<1TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA0@1`0000e@000`4/;00O;0030@0006h0003E000S0I4/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/T@40L0000=D000<1;2`07b`00`40001^0000e@008`6A;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;9410700003E00030B`/01l/00<10000KP000=D002<1TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@1`0000e@000`4/;00O;0030@0006h0003E000S0I4/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40L0000=D000<1;2`07b`00`40001^0000e@008`6A
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410700003E00030B`/01l/00<10000KP000=D0
02<1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@1`0000e@000`4/;00O;0030@0006h0
003E000S0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40L0000=D000<1;2`07b`00`40
001^0000e@008`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410700003E00030B`/01l/
00<10000KP000=D002<1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@1`0000e@000`4/
;00O;0030@0006h0003E000S0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40L0000=D0
00<1;2`07b`00`40001^0000e@008`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410700
003E00030B`/01l/00<10000KP000=D002<1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
0@1`0000e@000`4/;00O;0030@0006h0003E000S0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/T@40L0000=D000<1;2`07b`00`40001^0000e@008`6A;94/TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;9410700003E00030B`/01l/00<10000KP000=D002<1TBbA;94/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA0@1`0000e@000`4/;00O;0030@0006h0003E000S0I4/TBbA;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/T@40L0000=D000<1;2`07b`00`40001^0000e@008`6A;94/TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;9410700003E00030B`/01l/00<10000KP000=D002<1TBbA;94/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@1`0000e@000`4/;00O;0030@0006h0003E000S0I4/TBbA
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40L0000=D000<1;2`07b`00`40001^0000e@008`6A
;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;9410700003E00030B`/01l/00<10000KP000=D0
02<1TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA0@1`0000e@000`4/;00O;0030@0006h0
003E000S0I4/TBbA;94/TBbA;94/TBbA;94/TBbA;94/TBbA;94/T@40L0000=D08`5`0000o`1Y0000
o`1Y0000o`1Y0000o`1Y0000o`1Y0000o`1Y0000o`1Y0000o`1Y0000o`1Y0000o`1Y0000fP040@@0
00<100001@000`40000300<1M@000=`000<100000`000`40000;00030@0107D0003H00H11@000`40
000:00<1M@000=P000D100000@0600030@0000P000<10000M`000=T000@100011P000`40000900@1
M@000=T00P400`01000500030@0000X000@10001M0000001\
\>"],
  ImageRangeCache->{{{108, 395}, {315.438, 138.5}} -> {-0.429386, 3.90918, 
  0.00369537, 0.0309126}}]
}, Open  ]],

Cell["\<\
People may not be very good at explicitly manipulating \
probabilities and doing word problems that involve setting up contexts and \
formulating problems.  But this doesn't mean that they are no good at dealing \
with uncertainty and implicitly working with probabilities (or at least \
frequencies).  In fact, people appear to use probabilities in learning about \
language, e.g., word boundaries.  Animals also seem to use probabilities, but \
they also have a strong bias - a bias that is very useful in surviving.  This \
bias corresponds to having a strong expectation or prior probability.  Many \
of our problems in handling probabilities can be attributed to survival and \
our need to make decisions quickly, often on the basis of little evidence.  \
If you're a rat and you get poisoned eating in the trash bin of a particular \
restaurant, then you can't afford the \"luxury\" of eating out of that same \
bin enough times to make sure that you've seen a representative sample of the \
restaurant's cooking.  You're better off just avoiding that place.  Could \
this sort of behavior be justified in terms probability theory?  The answer \
is a definite yes.  And it has to do with bias, a notion that we'll return to \
frequently in this course.\.00g that place.  Could this sort of behavior be \
justified in terms probability theory?  The answer is a definite yes.  And it \
has to do with bias, a notion that we'll return to frequently in this course.
\
\>", "Text"]
}, Open  ]]
},
FrontEndVersion->"X 3.0",
ScreenRectangle->{{0, 1280}, {0, 1024}},
AutoGeneratedPackage->None,
WindowSize->{806, 655},
WindowMargins->{{Automatic, 188}, {83, Automatic}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"Magnification"->1},
Magnification->1.25,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1709, 49, 66, 0, 126, "Title"],

Cell[CellGroupData[{
Cell[1800, 53, 67, 0, 71, "Section"],
Cell[1870, 55, 266, 7, 57, "Text"],
Cell[2139, 64, 302, 7, 57, "Text"],
Cell[2444, 73, 161, 4, 56, "Text"],
Cell[2608, 79, 211, 6, 57, "Text"],
Cell[2822, 87, 765, 11, 144, "Text"],
Cell[3590, 100, 134, 3, 34, "Text"],
Cell[3727, 105, 148, 6, 35, "Text"],
Cell[3878, 113, 84, 3, 82, "Input"],
Cell[3965, 118, 298, 9, 57, "Text"],
Cell[4266, 129, 276, 7, 166, "Input"],

Cell[CellGroupData[{
Cell[4567, 140, 175, 5, 103, "Input"],
Cell[4745, 147, 39267, 1495, 233, 19678, 1249, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[44027, 1645, 686, 11, 156, "Text"],
Cell[44716, 1658, 96, 1, 61, "Input",
  InitializationCell->True],
Cell[44815, 1661, 40, 0, 61, "Input"],

Cell[CellGroupData[{
Cell[44880, 1665, 186, 6, 145, "Input"],
Cell[45069, 1673, 42731, 1032, 233, 15451, 691, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[87815, 2708, 172, 4, 56, "Text"],

Cell[CellGroupData[{
Cell[88012, 2716, 193, 6, 145, "Input"],
Cell[88208, 2724, 34754, 933, 233, 15501, 691, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[122977, 3660, 520, 8, 55, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[123534, 3673, 94, 3, 72, "Section"],
Cell[123631, 3678, 207, 4, 31, "Text"],
Cell[123841, 3684, 293, 5, 43, "Text"],
Cell[124137, 3691, 380, 6, 43, "Text"],
Cell[124520, 3699, 960, 14, 103, "Text"],

Cell[CellGroupData[{
Cell[125505, 3717, 194, 5, 53, "Input"],
Cell[125702, 3724, 35513, 1431, 185, 19512, 1229, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[161230, 5158, 227, 4, 31, "Text"],
Cell[161460, 5164, 106, 3, 20, "Text"],
Cell[161569, 5169, 95, 3, 20, "Text"],
Cell[161667, 5174, 237, 7, 31, "Text"],
Cell[161907, 5183, 106, 2, 20, "Text"],
Cell[162016, 5187, 106, 3, 20, "Text"],
Cell[162125, 5192, 122, 2, 20, "Text"],
Cell[162250, 5196, 387, 7, 43, "Text"],
Cell[162640, 5205, 168, 7, 20, "Text"],
Cell[162811, 5214, 113, 2, 20, "Text"],
Cell[162927, 5218, 110, 5, 20, "Text"],
Cell[163040, 5225, 322, 6, 43, "Text"],
Cell[163365, 5233, 250, 7, 31, "Text"],
Cell[163618, 5242, 62, 0, 20, "Text"],
Cell[163683, 5244, 195, 6, 20, "Text"],
Cell[163881, 5252, 69, 0, 40, "Input"],
Cell[163953, 5254, 35, 0, 20, "Text"],
Cell[163991, 5256, 62, 0, 40, "Input"],
Cell[164056, 5258, 51, 0, 20, "Text"],

Cell[CellGroupData[{
Cell[164132, 5262, 72, 0, 40, "Input"],
Cell[164207, 5264, 38, 1, 36, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[164282, 5270, 166, 5, 53, "Input"],
Cell[164451, 5277, 17735, 545, 185, 6539, 402, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[182201, 5825, 167, 4, 31, "Text"],
Cell[182371, 5831, 52, 0, 40, "Input"],
Cell[182426, 5833, 62, 0, 40, "Input"],

Cell[CellGroupData[{
Cell[182513, 5837, 178, 5, 62, "Input"],
Cell[182694, 5844, 24651, 914, 185, 12080, 754, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[207360, 6761, 171, 4, 31, "Text"],
Cell[207534, 6767, 171, 4, 31, "Text"],
Cell[207708, 6773, 138, 3, 20, "Text"],
Cell[207849, 6778, 115, 3, 20, "Text"],
Cell[207967, 6783, 161, 4, 31, "Text"],
Cell[208131, 6789, 708, 11, 79, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[208876, 6805, 59, 0, 72, "Section"],
Cell[208938, 6807, 290, 5, 43, "Text"],
Cell[209231, 6814, 507, 8, 55, "Text"],
Cell[209741, 6824, 288, 5, 43, "Text"],
Cell[210032, 6831, 161, 4, 31, "Text"],
Cell[210196, 6837, 605, 14, 55, "Text"],
Cell[210804, 6853, 129, 6, 20, "Text"],
Cell[210936, 6861, 106, 2, 20, "Text"],
Cell[211045, 6865, 316, 6, 43, "Text"],
Cell[211364, 6873, 381, 8, 43, "Text"],
Cell[211748, 6883, 122, 5, 20, "Text"],
Cell[211873, 6890, 116, 2, 20, "Text"],
Cell[211992, 6894, 734, 21, 55, "Text"],
Cell[212729, 6917, 145, 3, 20, "Text"],
Cell[212877, 6922, 398, 11, 114, "Input"],
Cell[213278, 6935, 36, 0, 40, "Input"],
Cell[213317, 6937, 193, 5, 62, "Input"],
Cell[213513, 6944, 196, 5, 62, "Input"],

Cell[CellGroupData[{
Cell[213734, 6953, 136, 3, 45, "Input"],
Cell[213873, 6958, 24379, 609, 185, 6052, 378, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],

Cell[CellGroupData[{
Cell[238289, 7572, 137, 3, 45, "Input"],
Cell[238429, 7577, 24717, 613, 185, 6067, 378, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],

Cell[CellGroupData[{
Cell[263183, 8195, 149, 3, 45, "Input"],
Cell[263335, 8200, 21216, 636, 185, 7404, 461, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],

Cell[CellGroupData[{
Cell[284588, 8841, 163, 5, 62, "Input"],
Cell[284754, 8848, 35210, 939, 185, 15473, 691, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[319979, 9790, 1492, 21, 127, "Text"]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

