(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 3.0,
MathReader 3.0, or any compatible application. The data for the notebook 
starts with the line of stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     86412,       2756]*)
(*NotebookOutlinePosition[     87245,       2785]*)
(*  CellTagsIndexPosition[     87201,       2781]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["\<\
Supplementary Notes on Representation, Logic and Problem \
Formulation\
\>", "Title"],

Cell[CellGroupData[{

Cell["Pigs, Porridge, Panthers, and Puzzles", "Section"],

Cell["\<\
In these notes, we explore some of the issues that surround \
representation, problem formulation, the use of logic, and approaches that \
exploit regularities (symmetries) in problems in order to expedite problem \
solving.  Analyzing puzzles is a good way to gain some into these issues.  \
Most puzzles you find in books and magazines aren't as messy or complicated \
as real-world problems but they illustrate basic principles that can be \
applied to solving any problem.\
\>", "Text"],

Cell[TextData[{
  "You've probably heard of some variant of the missionaries and cannibals \
problem.  In this problem, you're given the task of ferrying some misionaries \
and cannibals across a river using multiple trips in a limited-capacity boat. \
The tricky part is that you can't leave numbers of missionaries and \
canniabals on either bank of the river so that (depending on the version) the \
cannibals eat the missionaries or the missionaries convert the cannibals. The \
missionaries and cannibals problem is really a ",
  StyleBox["logistics",
    FontSlant->"Italic"],
  " problem; logistics is a subarea of operations research that deals with \
the procurement, distribution, maintenance, and replacement of materiel and \
personnel. Logistics problems are faced daily by local municipalities, the \
federal government, the military, and companies involved in shipping packages \
and transporting people. In the following, we consider a variant of the \
missionaries and cannibals problem that doesn't involve people eating other \
people, but is otherwise just as gruesome.  "
}], "Text"],

Cell["\<\
You're farmer and you have three items of produce that you want to \
sell at the market: a pig, a bushel of corn (it's pot of porridge in some \
variants), and a panther (a strange farm animal but that's the way the story \
goes in this variant).The market is across the river from your farm and you \
have to your transport your produce across the river in a small boat that can \
only carry you and one of your three items of produce at a time.  You can't \
leave the leave the pig and the corn alone on the same side of the river or \
the pig will eat the corn and you can't leave the panther and the pig alone \
on the same side of the river or the panther will eat the pig.  How do you \
get all of your produce to the market?\
\>", "Text"],

Cell[TextData[{
  "We're going to represent this problem using boolean logic (also called \
propositional logic).  The problem involves time passing so let's assume that \
there are discrete time points ",
  StyleBox["t = 0, 1, 2, ...,",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " up to ",
  StyleBox["T",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ". At any given point in time, each of our characters is on one side or the \
other (we'll ignore the time they are in transit on the river). We begin by \
introducing a set of boolean variables. Let ",
  StyleBox["pig[t] == True (False)",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["panther[t] == True (False)",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", and ",
  StyleBox["farmer[t] == True (False)",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", and ",
  StyleBox["corn[t] == True (False)",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", indicate that, respectively, the pig, panther, farmer, and the bushel of \
corn is on the market (farm) side of the river at time ",
  StyleBox["t",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ".  Using these variables we can write down various formulas that will \
allow us to formulate the ",
  StyleBox["pig and the panther problem",
    FontSlant->"Italic"],
  " (as we'll call it) as a logical problem.  For example, the following \
formula indicates that it can't be that the pig and the panther are on the \
farm side of the river at time ",
  StyleBox["0",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " if the farmer is on the market side of the river at time ",
  StyleBox["0",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[CellGroupData[{

Cell["Not[And[pig[0],panther[0],Not[farmer[0]]]] // FullForm", "Input"],

Cell[BoxData[
    TagBox[
      StyleBox[\(Not[And[pig[0], panther[0], Not[farmer[0]]]]\),
        ShowSpecialCharacters->False,
        ShowStringCharacters->True],
      FullForm]], "Output"]
}, Open  ]],

Cell["\<\
This is not the only way that we could have expressed the same \
constraint.  Here are two alternative but logically equivalent \
formulations.\
\>", "Text"],

Cell[CellGroupData[{

Cell["Or[Not[pig[0]],Not[panther[0]],farmer[0]] // FullForm", "Input"],

Cell[BoxData[
    TagBox[
      StyleBox[\(Or[Not[pig[0]], Not[panther[0]], farmer[0]]\),
        ShowSpecialCharacters->False,
        ShowStringCharacters->True],
      FullForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Implies[And[pig[0],panther[0]],farmer[0]] // FullForm", "Input"],

Cell[BoxData[
    TagBox[
      StyleBox[\(Implies[And[pig[0], panther[0]], farmer[0]]\),
        ShowSpecialCharacters->False,
        ShowStringCharacters->True],
      FullForm]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Before we proceed much further, a word about notation is in order. I like \
to use ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " as an interactive sketchpad for illustrating and animating points in \
lectures.  In this course, you won't be learning how to use ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " to write programs, but you will be able to use a program called ",
  StyleBox["MathReader",
    FontSlant->"Italic"],
  " to display lectures formatted as ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " notebooks.  ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " provides a dizzying array of formats for entering and displaying \
mathematical and logical formulas. Learning how to translate between \
notations is something you'll find useful in converting between a formal \
description of a problem (",
  StyleBox["e.g., ",
    FontSlant->"Italic"],
  "a description that you find in a paper or book) and a description suitable \
for programming purposes. Here is a small sampler of different notations for \
displaying logical formulas in ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  "."
}], "Text"],

Cell[CellGroupData[{

Cell["\<\
Module[{f = Or[And[pig[0],Not[panther[0]]],farmer[0]]},
        Print[FullForm[f]] ;
        Print[StandardForm[f]] ;
        Print[TraditionalForm[f]]]\
\>", "Input",
  Evaluatable->False],

Cell[BoxData[
    TagBox[
      StyleBox[\(Or[And[pig[0], Not[panther[0]]], farmer[0]]\),
        ShowSpecialCharacters->False,
        ShowStringCharacters->True],
      FullForm]], "Print"],

Cell[BoxData[
    TagBox[
      FormBox[
        \(pig[0] && \(\[InvisibleSpace]! \((panther[0])\)\) || farmer[0]\),
        "StandardForm"],
      StandardForm,
      Editable->True]], "Print"],

Cell[BoxData[
    TagBox[
      FormBox[
        \(pig(0) \[And] \(\[InvisibleSpace]! \((panther(0))\)\) \[Or] 
          farmer(0)\),
        "TraditionalForm"],
      TraditionalForm,
      Editable->True]], "Print"]
}, Open  ]],

Cell[TextData[{
  "As useful as it might be for you to learn to move easily back and forth \
between different notational variants, in the remainder of this notebook, \
we'll stick to the notation used in the textbook, ",
  StyleBox["Artificial Intelligence: Theory and Practice",
    FontSlant->"Italic"],
  ". The invocations in the subsection at the end of this section (which you \
really needn't look at) tell ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " how to display logical expressions using the notational conventions in \
the textbook. By the way, once you've evaluated all of the cells in this \
notebook during a session with ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  ", reevaluating any particular cell will result in a formula being \
displayed using the textbook conventions; for this reason, I've set the above \
cell so that it won't be evaluated and thus it will continue to accurately \
display some of ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  "'s other notational variants. "
}], "Text"],

Cell["\<\
Our immediate goal is to write down a formula that accurately \
represents the pig and the panther puzzle, in the sense that the formula is \
true if and only if the puzzle has a solution.  Once we have such a \
representation, then we'll want to determine if the formula is true or not, \
and, if it is true (and hence the puzzle has a solution), then we'll want to \
find a solution corresponding to a sequence of steps such that in each step \
the farmer rows across the river either alone or carrying an item of produce.\
\
\>", "Text"],

Cell[TextData[{
  "Just as there are multiple equivalent formulas for representing the \
constraint involving the pig and the panther that we illustrated above, there \
are multiple formulas representing the whole puzzle.  All of these formulas \
are pretty hairy, but some have a nice uniform structure that makes them more \
amenable to solution by machine.  In the following section, we take a short \
detour to talk more about logic and investigate some alternative forms called \
",
  StyleBox["normal forms",
    FontSlant->"Italic"],
  ". Many of you already know about normal forms but the review will probably \
be worthwhile as normal forms play an important role in automated inference. \
In addition to talking about the representational advantages of normal forms, \
we'll investigate a particular method for solving problems such as the pig \
and the panther problem that can be represented as boolean formulas. "
}], "Text"],

Cell[CellGroupData[{

Cell[TextData[{
  "Modifications to ",
  StyleBox["StandardForm",
    FontFamily->"Courier",
    FontSize->16],
  " to Mimic the Notation in the Book ",
  StyleBox["(Ignore this!)",
    FontSlant->"Italic"]
}], "Subsection"],

Cell[TextData[{
  "The following invocations involve low-level formatting issues. We want the \
logical expressions output by ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " to appear in the same notation used in the book, and these routines make \
this possible. "
}], "Text"],

Cell["Off[General::spell1] ;", "Input"],

Cell["\<\
AtomicQ[f_] := 
  Not[MemberQ[{And,Or,Not,Implies},Head[f]]]\
\>", "Input"],

Cell["\<\
LiteralQ[f_] := 
  Or[AtomicQ[f],
     And[SameQ[Head[f],Not],AtomicQ[f[[1]]]]]\
\>", "Input"],

Cell["\<\
Unprotect[Not] ;

Not /: MakeBoxes[Not[x_],TraditionalForm] := 
         If[LiteralQ[x],
            RowBox[{\"\[Not]\",
                    MakeBoxes[x,TraditionalForm]} ],
            RowBox[{\"\[Not]\",
                    \"(\",
                    MakeBoxes[x,TraditionalForm],
                    \")\"} ]]
      
Not /: MakeBoxes[Not[x_], StandardForm] := 
         If[LiteralQ[x],
            RowBox[{\"\[Not]\",
                    MakeBoxes[x,StandardForm]}],
            RowBox[{\"\[Not]\",
                    \"(\",
                    MakeBoxes[x,StandardForm],
                    \")\"} ]]
      
Protect[Not] ;\
\>", "Input"],

Cell[TextData[
"Unprotect[Implies] ;\n\nImplies /: MakeBoxes[Implies[x_,y_],TraditionalForm] \
:= \n             RowBox[{MakeBoxes[x,TraditionalForm],\n                     \
\"\[Superset]\",\n                     MakeBoxes[y,TraditionalForm]} ]\n      \
\nImplies /: MakeBoxes[Implies[x_,y_],StandardForm] := \n             \
RowBox[{MakeBoxes[x,StandardForm],\n                     \"\[Superset]\",\n   \
                  MakeBoxes[y,StandardForm]} ]\n      \nProtect[Implies] ;"], 
  "Input"],

Cell["\<\
Unprotect[And] ;

And /: MakeBoxes[And[x__],StandardForm] := 
         RowBox[{\"(\",
                 MakeBoxes[And[x],TraditionalForm],
                 \")\"}]
      
Protect[And] ;\
\>", "Input"],

Cell["\<\
Unprotect[Or] ;

Or /: MakeBoxes[Or[x__],StandardForm] := 
        RowBox[{\"(\",
                MakeBoxes[Or[x],TraditionalForm],
                \")\"}]
      
Protect[Or] ;\
\>", "Input"],

Cell["\<\
So now expressions print out as they would appear in the notation \
used in the textbook.\
\>", "Text"],

Cell[CellGroupData[{

Cell["Or[Not[Or[pig[0],panther[0]]],farmer[0]]", "Input"],

Cell[BoxData[
    \(\((\[Not] (\(pig(0) \[Or] panther(0)\)) \[Or] farmer(0))\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Or[Not[pig[0]],Not[panther[0]],farmer[0]]", "Input"],

Cell[BoxData[
    \(\((\[Not] pig(0) \[Or] \[Not] panther(0) \[Or] farmer(0))\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Implies[And[pig[0],panther[0]],farmer[0]]", "Input"],

Cell[BoxData[
    \(\((pig(0) \[And] panther(0))\) \[Superset] farmer[0]\)], "Output"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Normal Forms and Solving Problems Represented as Boolean Formulas\
\
\>", "Section"],

Cell[TextData[{
  StyleBox[
  "Let's start by introducing three boolean variables which we notate as "],
  StyleBox["v[[1]], v[[2]],",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[" and "],
  StyleBox["v[[3]]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[
  ".  For the time being let's assume that these variables have no \
interpretation, that is to say we won't assign them any values but rather \
think of them symbolically.  Keep in mind however that these are boolean \
variables and so they can only be assigned (interpreted as) boolean "],
  StyleBox["True",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[" or boolean "],
  StyleBox["False",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox["."]
}], "Text"],

Cell[TextData[{
  "The invocation ",
  StyleBox["Off[Part::partd]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " tells ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " not to warn me when I try to refer to a part of an object that doesn't \
exist.  In Mathematica, ",
  StyleBox["v[[i]]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " refers to the ",
  StyleBox["i",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "th entry of the list denoted by ",
  StyleBox["v",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ".  ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " doesn't really care that ",
  StyleBox["v",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " hasn't been defined and ",
  StyleBox["v[[1]]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " doesn't exist, but it is a common bug in writing ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " programs to incorrectly type (in either sense of the word) a variable and \
then refer to the parts of a nonexistent object or an object that is not a \
list.  By warning the programmer, ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " helps him or her to avoid a lot of bugs.  We could declare ",
  StyleBox["v",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " in such a way to avoid these annoying messages without subverting ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  "'s efforts to warn us, but it is more tedious that it is worth.  I'm \
assuming that you don't want to know any more about ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " than you need to know in order to understand my examples and \
illustrations.  Now I can refer to",
  StyleBox[" v[[1]], v[[2]],",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and ",
  StyleBox["v[[3]]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " symbolically in formulas without getting annoying warnings."
}], "Text"],

Cell["Off[Part::partd] ;", "Input"],

Cell[TextData[{
  StyleBox["From these variables, "],
  StyleBox["And",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[", "],
  StyleBox["Or",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[", and "],
  StyleBox["Not",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[" we can construct formulas."]
}], "Text"],

Cell[CellGroupData[{

Cell["Or[Not[v[[1]]],And[v[[3]],v[[2]]]]", "Input"],

Cell[BoxData[
    \(\((\[Not] v\[LeftDoubleBracket]1\[RightDoubleBracket] \[Or] 
        v\[LeftDoubleBracket]3\[RightDoubleBracket] \[And] 
          v\[LeftDoubleBracket]2\[RightDoubleBracket])\)\)], "Output"]
}, Open  ]],

Cell[TextData[{
  StyleBox["Here's a boolean formula expressed in a particular so-called "],
  StyleBox["normal form",
    FontSlant->"Italic"],
  StyleBox["."]
}], "Text"],

Cell[CellGroupData[{

Cell["f = And[Or[v[[1]],Not[v[[2]]]],Or[v[[3]],Not[v[[1]]]]]", "Input"],

Cell[BoxData[
    \(\((\(v\[LeftDoubleBracket]1\[RightDoubleBracket] \[Or] 
          \[Not] v\[LeftDoubleBracket]2\[RightDoubleBracket]\) \[And] 
        \(v\[LeftDoubleBracket]3\[RightDoubleBracket] \[Or] 
          \[Not] v\[LeftDoubleBracket]1\[RightDoubleBracket]\))\)\)], "Output"]
}, Open  ]],

Cell["It has two subformulas or clauses.", "Text"],

Cell[CellGroupData[{

Cell["Length[f] ", "Input"],

Cell[BoxData[
    \(2\)], "Output"]
}, Open  ]],

Cell["And we can refer to each of these parts individually.", "Text"],

Cell[CellGroupData[{

Cell["{f[[1]],f[[2]]}", "Input"],

Cell[BoxData[
    \({\((
        v\[LeftDoubleBracket]1\[RightDoubleBracket] \[Or] 
          \[Not] v\[LeftDoubleBracket]2\[RightDoubleBracket])\), 
      \((v\[LeftDoubleBracket]3\[RightDoubleBracket] \[Or] 
          \[Not] v\[LeftDoubleBracket]1\[RightDoubleBracket])\)}\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "The above formula is in ",
  StyleBox["conjunctive normal form",
    FontSlant->"Italic"],
  " or CNF. In general, a formula is in CNF if it is a conjunction of \
formulas each of which is a disjunction consisting of disjuncts that are \
either variables or their negations.  The disjunctions are called ",
  StyleBox["clauses",
    FontSlant->"Italic"],
  " and the disjuncts are called ",
  StyleBox["literals",
    FontSlant->"Italic"],
  ".  CNF is completely general in the sense that any boolean formula has an \
equivalent formula in CNF. In the following, the first formula is in CNF and \
second is not."
}], "Text"],

Cell[CellGroupData[{

Cell["And[Or[Not[v[[1]]],v[[2]]],Or[v[[1]]]]", "Input"],

Cell[BoxData[
    \(\((\(\[Not] v\[LeftDoubleBracket]1\[RightDoubleBracket] \[Or] 
          v\[LeftDoubleBracket]2\[RightDoubleBracket]\) \[And] 
        v\[LeftDoubleBracket]1\[RightDoubleBracket])\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["And[Or[Not[v[[1]]],And[v[[2]],v[[3]]]],Or[v[[1]]]]", "Input"],

Cell[BoxData[
    \(\((\(\[Not] v\[LeftDoubleBracket]1\[RightDoubleBracket] \[Or] 
          v\[LeftDoubleBracket]2\[RightDoubleBracket] \[And] 
            v\[LeftDoubleBracket]3\[RightDoubleBracket]\) \[And] 
        v\[LeftDoubleBracket]1\[RightDoubleBracket])\)\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "Question: What do we mean when we say two formulas are ",
  StyleBox["equivalent",
    FontSlant->"Italic"],
  "?  Answer: They have the same value (true or false) for every assignment \
to their constituent variables.  Question: What if a formula is always true \
or always false?  Can this even happen?  Answer: A formula which is always \
true is called a ",
  StyleBox["tautology",
    FontSlant->"Italic"],
  " (",
  StyleBox["e.g.,",
    FontSlant->"Italic"],
  " ",
  StyleBox["Or[v[[1]],Not[v[[1]]]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is true no matter what truth value we assign to ",
  StyleBox["v[[1]]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ") and one that is always false is called a contradiction (",
  StyleBox["e.g.,",
    FontSlant->"Italic"],
  " ",
  StyleBox["And[v[[1]],Not[v[[1]]]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ").  Given the definition of logical equivalence provided here, all \
tautologies are equivalent to one another and all contradictions are \
equivalent to one another.  Let's think more concretely about what it means \
for a variable or a formula to be assigned a value."
}], "Text"],

Cell[TextData[{
  "We can assign each variable a value of either true of false.  By assigning \
all of the variables in a formula values, we also assign the formula and each \
of its parts values.  How can this statement be so?  Well, the statement is \
not precisely so, but we're assuming that we've also assigned ",
  StyleBox["meaning",
    FontSlant->"Italic"],
  " to each of the so-called truth functions, ",
  StyleBox["And",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["Or",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", and ",
  StyleBox["Not",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "; in this case, we've assigned them the meaning corresponding to the \
\"standard\" truth tables."
}], "Text"],

Cell[CellGroupData[{

Cell["{f[[1]],f[[2]]} /. v -> {True,False,True}", "Input"],

Cell[BoxData[
    \({True, True}\)], "Output"]
}, Open  ]],

Cell[TextData[{
  StyleBox["Since "],
  StyleBox["f",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[
  " is a conjunction of these two clauses the formula is also true."]
}], "Text"],

Cell[CellGroupData[{

Cell["f /. v -> {True,False,True}", "Input"],

Cell[BoxData[
    \(True\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "The following routine formats and displays the truth tables for boolean \
formulas such as  ",
  StyleBox["And",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["Or",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["Not",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["Xor",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", and ",
  StyleBox["Implies.",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " "
}], "Text"],

Cell["\<\
DisplayTruthTable[fun_,arity_] :=
  Print[MatrixForm[Module[{a,b,t}, 
          t = Table[b = IntegerDigits[i,2,arity] ;
                    a = Table[If[b[[j]] == 0, False, True],
                              {j,1,arity}] ;
                    Append[a,Apply[fun,a]],
                    {i, 0, 2^arity-1}] ;
          Prepend[t,Append[Table[v[[i]],{i,1,arity}],
                           Apply[fun,
                                 If[arity == 1,
                                    {v[[1]]},
                                    If[arity == 2,
                                       {v[[1]],v[[2]]},
                                       {v[[1]],\"...\",v[[arity]]}]]]]]]]]\
\>\
", "Input"],

Cell[CellGroupData[{

Cell["DisplayTruthTable[And,3]", "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", GridBox[{
            {\(v\[LeftDoubleBracket]1\[RightDoubleBracket]\), 
              \(v\[LeftDoubleBracket]2\[RightDoubleBracket]\), 
              \(v\[LeftDoubleBracket]3\[RightDoubleBracket]\), 
              \((v\[LeftDoubleBracket]1\[RightDoubleBracket] \[And] "..." 
                  \[And] v\[LeftDoubleBracket]3\[RightDoubleBracket])\)},
            {"False", "False", "False", "False"},
            {"False", "False", "True", "False"},
            {"False", "True", "False", "False"},
            {"False", "True", "True", "False"},
            {"True", "False", "False", "False"},
            {"True", "False", "True", "False"},
            {"True", "True", "False", "False"},
            {"True", "True", "True", "True"}
            }], ")"}],
      (MatrixForm[ #]&)]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell["DisplayTruthTable[Implies,2]", "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", GridBox[{
            {\(v\[LeftDoubleBracket]1\[RightDoubleBracket]\), 
              \(v\[LeftDoubleBracket]2\[RightDoubleBracket]\), 
              \(v\[LeftDoubleBracket]1\[RightDoubleBracket] \[Superset] 
                v\[LeftDoubleBracket]2\[RightDoubleBracket]\)},
            {"False", "False", "True"},
            {"False", "True", "True"},
            {"True", "False", "False"},
            {"True", "True", "True"}
            }], ")"}],
      (MatrixForm[ #]&)]], "Print"]
}, Open  ]],

Cell["\<\
Determining whether or not a particular assignment makes a formula \
true or false is easy computationally.  How easy?\
\>", "Text"],

Cell[TextData[{
  "We say that a formula is ",
  StyleBox["satisfiable",
    FontSlant->"Italic"],
  " if there exists an assignment that makes the formula true.  You should \
know from your other classes that determining whether or not a formula is \
satisfiable is computationally difficult in the sense that the general \
problem is NP complete.  One characteristic of an NP-complete problem is that \
it is easy to check if you have a solution but there are too many possible to \
efficiently enumerate and check each one. One way to find out if a formula is \
satisfiable is to try every possible assignment."
}], "Text"],

Cell[CellGroupData[{

Cell["\<\
Print[MatrixForm[Module[{a,b,t}, 
                    t = Table[b = IntegerDigits[i,2,3] ;
                              a = Table[If[b[[j]] == 0, False, True],
                                        {j,1,3}] ;
                              {a, f /. v -> a},
                              {i, 0, 2^3-1}] ;
                    Prepend[t,{{\"v[[1]]\",\"v[[2]]\",\"v[[3]]\"},\"f\"}]]]]\
\
\>", "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", GridBox[{
            {\({"v[[1]]", "v[[2]]", "v[[3]]"}\), "\<\"f\"\>"},
            {\({False, False, False}\), "True"},
            {\({False, False, True}\), "True"},
            {\({False, True, False}\), "False"},
            {\({False, True, True}\), "False"},
            {\({True, False, False}\), "False"},
            {\({True, False, True}\), "True"},
            {\({True, True, False}\), "False"},
            {\({True, True, True}\), "True"}
            }], ")"}],
      (MatrixForm[ #]&)]], "Print"]
}, Open  ]],

Cell["\<\
Another way to find out if a formula is satisfiable is to generate \
random assignments.  This is really doing a random search in the space of \
boolean assignments.  If the formula is satisfiable then this often works \
rather well, but what if the formula is not satisfiable?\
\>", "Text"],

Cell["RandomBoolean[] := If[Random[Integer,1]==1,True,False]", "Input"],

Cell["RandomAssignment[n_Integer] := Table[RandomBoolean[],{n}]", "Input"],

Cell[CellGroupData[{

Cell["Table[f /. v -> RandomAssignment[3],{8}]", "Input"],

Cell[BoxData[
    \({True, True, False, False, False, False, True, False}\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "Sometimes we can make a sequence of ",
  StyleBox["local",
    FontSlant->"Italic"],
  " changes to an assignment so as to turn it into a satisfying assignment.  \
The sort of local changes that are most often used in practice correspond to \
taking an assignment and flipping one bit, ",
  StyleBox["i.e.,",
    FontSlant->"Italic"],
  " taking one variable which was assigned ",
  StyleBox["True",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and assigning it ",
  StyleBox["False",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " or visa versa.  The question is which bit to flip and how do we tell if \
we are making progress toward our goal of finding a satisfying assignment."
}], "Text"],

Cell[TextData[{
  "For the case of formulas such as ",
  StyleBox["f",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " above that are in conjunctive normal form, there is a pretty effective \
heuristic (rule of thumb) for measuring progress: you flip the bit \
corresponding to the variable such that the resulting assignment ends up \
satisfying the most clauses (breaking ties randomly).  Here's a simple \
procedure (with lots of print statements so we can watch how it works) that \
implements this heuristic."
}], "Text"],

Cell["\<\
Twiddle[a_List] := 
    Module[{best = {a}, s, g, h},
        g = Sum[If[f[[i]] /. v -> a, 1, 0], {i,1,Length[f]}] ;
        Print[\"Initial assignment = \", a]; 
        Print[\"Heuristic value for initial assignment = \", g] ;
        Do[s = ReplacePart[a,Not[a[[i]]],i] ;
           h = Sum[If[f[[i]] /. v -> s, 1, 0], {i,1,Length[f]}] ;
           Print[\"Current assignment = \", s] ;
           Print[\"Heuristic value for current assignment = \", h] ;
           If[h == g, AppendTo[best,s],
              If[h > g, best = {s} ; g = h]] ; 
           Print[\"Current best heuristic value = \", g] ;
           Print[\"Current top assignments = \", best],
           {i,1,Length[a]}] ;
        Return[best[[Random[Integer,{1,Length[best]}]]]]]\
\>", "Input"],

Cell[TextData[{
  "In the terminology of ",
  StyleBox["heuristic search",
    FontSlant->"Italic"],
  ", we say that ",
  StyleBox["Twiddle",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " chooses a bit to flip ",
  StyleBox["greedily",
    FontSlant->"Italic"],
  " in the sense that of all the bits it might flip, it chooses to flip the \
one that maximizes our heuristic measure. Can you think of an example, ",
  StyleBox["i.e.,",
    FontSlant->"Italic"],
  " a boolean function and an initial assignment to its variables, in which \
it is better ",
  StyleBox["not",
    FontSlant->"Italic"],
  " to be greedy? To demonstrate how ",
  StyleBox["Twiddle",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " works, lets take an assignment that satisfies ",
  StyleBox["f",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", flip one of its bits so that it no longer satisfies ",
  StyleBox["f",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", and then give it to ",
  StyleBox["Twiddle",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " as an intial assignment. "
}], "Text"],

Cell[CellGroupData[{

Cell["sat = {False,False,True}", "Input"],

Cell[BoxData[
    \({False, False, True}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["f /. v -> sat", "Input"],

Cell[BoxData[
    \(True\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["nosat = ReplacePart[sat,Not[sat[[2]]],2] ", "Input"],

Cell[BoxData[
    \({False, True, True}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["f /. v -> nosat", "Input"],

Cell[BoxData[
    \(False\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["found = Twiddle[nosat]", "Input"],

Cell[BoxData[
    InterpretationBox[
      \("Initial assignment = "\[InvisibleSpace]{False, True, True}\),
      SequenceForm[ "Initial assignment = ", {False, True, True}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \("Heuristic value for initial assignment = "\[InvisibleSpace]1\),
      SequenceForm[ "Heuristic value for initial assignment = ", 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \("Current assignment = "\[InvisibleSpace]{True, True, True}\),
      SequenceForm[ "Current assignment = ", {True, True, True}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \("Heuristic value for current assignment = "\[InvisibleSpace]2\),
      SequenceForm[ "Heuristic value for current assignment = ", 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \("Current best heuristic value = "\[InvisibleSpace]2\),
      SequenceForm[ "Current best heuristic value = ", 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \("Current top assignments = "\[InvisibleSpace]{{True, True, True}}\),
      SequenceForm[ "Current top assignments = ", {{True, True, True}}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \("Current assignment = "\[InvisibleSpace]{False, False, True}\),
      SequenceForm[ "Current assignment = ", {False, False, True}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \("Heuristic value for current assignment = "\[InvisibleSpace]2\),
      SequenceForm[ "Heuristic value for current assignment = ", 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \("Current best heuristic value = "\[InvisibleSpace]2\),
      SequenceForm[ "Current best heuristic value = ", 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \("Current top assignments = "\[InvisibleSpace]{{True, True, True}, {
            False, False, True}}\),
      SequenceForm[ 
      "Current top assignments = ", {{True, True, True}, {False, False, 
        True}}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \("Current assignment = "\[InvisibleSpace]{False, True, False}\),
      SequenceForm[ "Current assignment = ", {False, True, False}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \("Heuristic value for current assignment = "\[InvisibleSpace]1\),
      SequenceForm[ "Heuristic value for current assignment = ", 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \("Current best heuristic value = "\[InvisibleSpace]2\),
      SequenceForm[ "Current best heuristic value = ", 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \("Current top assignments = "\[InvisibleSpace]{{True, True, True}, {
            False, False, True}}\),
      SequenceForm[ 
      "Current top assignments = ", {{True, True, True}, {False, False, 
        True}}],
      Editable->False]], "Print"],

Cell[BoxData[
    \({False, False, True}\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "As expected, ",
  StyleBox["Twiddle",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " found an assignment satisfying ",
  StyleBox["f",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[CellGroupData[{

Cell["f /. v -> found", "Input"],

Cell[BoxData[
    \(True\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "Question: How could we use ",
  StyleBox["Twiddle",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " to design a program that tries to find a satisfying assignment if one \
exists and if the formula is not satisfiable returns a message \"this formula \
is not satisfiable\" which is indeed correct \"most of the time.\"  "
}], "Text"],

Cell[TextData[{
  "One answer to this question is the ",
  StyleBox["greedy satisfiability algorithm",
    FontSlant->"Italic"],
  " or GSAT. The outer loop of GSAT works as follows: Choose a random \
assignment and then in an inner loop try to turn it into a satisfying \
assignment by making a sequence of greedy bit flips. If you find a satisfying \
assignment starting from this random initial assignment, then stop and return \
it, otherwise choose another random assignment and try to turn it into a \
satisfying assignment. If necessary, continue trying random assignments for ",
  
  StyleBox["h",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " iterations. If after ",
  StyleBox["h",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " iterations you've found no satisfying assignment, then exit and report \
that the formula has no satisfying assignment or at least that you're willing \
to bet that it has no satisfying assignment. Choosing ",
  StyleBox["h",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " so that you're likely to win this bet is usually a matter of experiment \
and experience with the domain, ",
  StyleBox["i.e., ",
    FontSlant->"Italic"],
  "the set of problem instances that you're translating into boolean \
formulas."
}], "Text"],

Cell[TextData[{
  "Here is quick and dirty implementation of GSAT just for demonstration \
purposes. To make up for my characteristically short variable names, I'll \
provide a key: ",
  StyleBox["f",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is the boolean function we are trying to find a satisfying assignment \
for, ",
  StyleBox["n",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is the number of boolean variables in ",
  StyleBox["f",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["h",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is the maximum number of random starts mentioned above, ",
  StyleBox["k",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is the number of random starts so far, ",
  StyleBox["a",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and ",
  StyleBox["b",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " are boolean assignments corresponding to the best we've seen so far and \
the current assignment we're considering, and, finally, ",
  StyleBox["u",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and ",
  StyleBox["w",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " are heuristic values associated ",
  StyleBox["a",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and ",
  StyleBox["b",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ". As before, we reserve ",
  StyleBox["v",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " to represent boolean variables of the form ",
  StyleBox["v[[i]]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ". Here is the code for GSAT."
}], "Text"],

Cell["\<\
GSAT[f_,n_Integer,h_Integer] := 
  Module[{a, b, u, w, k = 0, flag},
    a = RandomAssignment[n] ;
    While[ k < h ,
      k ++ ; flag = True ;
      While[flag,
        flag = False ;
        u = Sum[If[f[[i]] /. v -> a, 1, 0], {i,1,Length[f]}] ;
        Do[b = ReplacePart[a,Not[a[[i]]],i] ;
           w = Sum[If[f[[i]] /. v -> b, 1, 0], {i,1,Length[f]}] ;
           If[w > u, a = b ; u = w; flag = True],
           {i,1,n}]] ;
      If[ f /. v -> a, Return[a], a = RandomAssignment[n]]] ;
    Return[False]]\
\>", "Input"],

Cell["And one simple experiment.", "Text"],

Cell[CellGroupData[{

Cell["found = GSAT[f,3,2]", "Input"],

Cell[BoxData[
    \({False, False, False}\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "As expected, ",
  StyleBox["GSAT",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " found an assignment satisfying ",
  StyleBox["f",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[CellGroupData[{

Cell["f /. v -> found", "Input"],

Cell[BoxData[
    \(True\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "Another variant of greedy satisfiability is called RWSAT or ",
  StyleBox["random walk satisfiability. ",
    FontSlant->"Italic"],
  "RWSAT is not quite as greedy as GSAT. The outer loop of RWSAT is the same \
as GSAT. However, in the inner loop of RWSAT, instead of greedily choosing \
from among single bit flips it greedily chooses from among a set of randomly \
generated sequences of bit flips of a length ",
  StyleBox["n",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " or less (where ",
  StyleBox["n",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is determined by experiment as in the case of ",
  StyleBox["h",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "). Note that the number of sequencs of length ",
  StyleBox["n",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is exponential in ",
  StyleBox["n",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " so we can't exhaustively search all such sequences for any reasonably \
large ",
  StyleBox["n",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " as we could all bit flips; hence we sample from the set of such sequences \
and then greedily chose the one that maximizes the number of satisfied \
clauses. If we make sure to include in our set of sequences all sequences of \
length one, ",
  StyleBox["i.e.,",
    FontSlant->"Italic"],
  " all single bit flips, then RWSAT should do at least as well GSAT though \
it may take longer on each iteration of the inner loop."
}], "Text"],

Cell[TextData[{
  "With the information provided in this subsection and in the appendix at \
the end of this notebook, you should have everything you need to implement a \
system that solves logistics problems using GSAT or RWSAT. For solving large, \
complicated logistics problems, I don't recommend using ",
  StyleBox["Mathematica. ",
    FontSlant->"Italic"],
  "A hybrid approach might be to build the boolean formula using the routines \
provided in this notebook and then write the formula to a file in an \
appropriate format. Then you can implement GSAT or RWSAT in C, C++, Java or \
whatever language you prefer using much more efficient data structures than \
the simple list structures I've used in this notebook. "
}], "Text"],

Cell[TextData[{
  "For more information on greedy methods for solving satisfiability problems \
see the article by Koutsoupias and Papadimitriou entitled \"On the greedy \
algorithm for satisfiability\" in ",
  StyleBox["Information Processing Letters",
    FontSlant->"Italic"],
  ", Volume 43, Pages 53-55, 1992. GSAT is described along with experimental \
results in an article entitled \"A New Method for Solving Hard Satisfiability \
Problems\", by Selman, Levesque, and Mitchell that appeared in the ",
  StyleBox["Proceedings of AAAI-92",
    FontSlant->"Italic"],
  ", Pages 440-446, 1992. You might also be interested in an article by \
Kirkpatrick and Selman that appeared in ",
  StyleBox["Science",
    FontSlant->"Italic"],
  ", entitled \"Critical Behavior in the Satisfiability of Random Boolean \
Formulae\", Volume 264, Pages 1297-1301, 1994."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["More Puzzling Porcine Problems", "Section"],

Cell["\<\
Now let's return to the pig and the panther problem.  We are going \
to represent the problem as a boolean formula in conjunctive normal form and \
so we will want to write down a set of clauses that can be conjoined together \
to form a CNF formula.\
\>", "Text"],

Cell[TextData[{
  StyleBox["The following degenerate clauses (the formula "],
  StyleBox["Or[v]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[" is equivalent to just "],
  StyleBox["v",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[
  ") must be true to characterize the initial state corresponding to all of \
the items of produce being on the farm side of the river."]
}], "Text"],

Cell[CellGroupData[{

Cell["init1 = Not[panther[0]]", "Input"],

Cell[BoxData[
    \(\[Not] panther[0]\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["init2 = Not[pig[0]]", "Input"],

Cell[BoxData[
    \(\[Not] pig[0]\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["init3 = Not[corn[0]]", "Input"],

Cell[BoxData[
    \(\[Not] corn[0]\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["init3 = Not[farmer[0]]", "Input"],

Cell[BoxData[
    \(\[Not] farmer[0]\)], "Output"]
}, Open  ]],

Cell[TextData[{
  StyleBox[
  "The rest of constraints are trickier to handle, but we'll take them one at \
a time and slowly.  We'll focus our immediate attention on specifying the \
constraint for a single time point and then extend to all time points.  Let's \
start with the physical constraint that only one animal can move at a time.  \
An animal, "],
  StyleBox["e.g.,",
    FontSlant->"Italic"],
  StyleBox[" the pig, has moved just in case either "],
  StyleBox["And[pig[t],Not[pig[t+1]]]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[" or "],
  StyleBox["And[Not[pig[t]],pig[t+1]]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[" for some "],
  StyleBox["t",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[
  ".  Here's a formula for the case in which the pig moves from the market \
side to the farm side and the corn stays on the market side."]
}], "Text"],

Cell[CellGroupData[{

Cell["Implies[And[pig[t],Not[pig[t+1]],corn[t]],corn[t+1]]", "Input"],

Cell[BoxData[
    \(\((pig(t) \[And] \[Not] pig(t + 1) \[And] corn(t))\) \[Superset] 
      corn[1 + t]\)], "Output"]
}, Open  ]],

Cell["which is equivalent to the clause", "Text"],

Cell[CellGroupData[{

Cell["Or[Not[pig[t]],pig[t+1],Not[corn[t]],corn[t+1]]", "Input"],

Cell[BoxData[
    \(\((\[Not] pig(t) \[Or] pig(t + 1) \[Or] \[Not] corn(t) \[Or] 
        corn(t + 1))\)\)], "Output"]
}, Open  ]],

Cell["\<\
Now we could proceed to write down the clauses for all the right \
combinations.  But I'm going to cheat.  I'm going to write down a set of more \
general formulas, one for each animal.\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\
solo1 = Implies[Or[And[pig[t],Not[pig[t+1]]],
                   And[Not[pig[t]],pig[t+1]]],
                And[Or[And[corn[t],corn[t+1]],
                       And[Not[corn[t]],Not[corn[t+1]]]],
                    Or[And[panther[t],panther[t+1]],
                       And[Not[panther[t]],Not[panther[t+1]]]]]]\
\>", "Input"],

Cell[BoxData[
    \(\((pig(t) \[And] \[Not] pig(t + 1) \[Or] 
          \[Not] pig(t) \[And] pig(t + 1))\) \[Superset] 
      \((\(corn(t) \[And] corn(t + 1) \[Or] 
            \[Not] corn(t) \[And] \[Not] corn(t + 1)\) \[And] 
          \(panther(t) \[And] panther(t + 1) \[Or] 
            \[Not] panther(t) \[And] \[Not] panther(t + 1)\))\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
solo2 = Implies[Or[And[corn[t],Not[corn[t+1]]],
                   And[Not[corn[t]],corn[t+1]]],
                And[Or[And[panther[t],panther[t+1]],
                       And[Not[panther[t]],Not[panther[t+1]]]],
                    Or[And[pig[t],pig[t+1]],
                       And[Not[pig[t]],Not[pig[t+1]]]]]]\
\>", "Input"],

Cell[BoxData[
    \(\((corn(t) \[And] \[Not] corn(t + 1) \[Or] 
          \[Not] corn(t) \[And] corn(t + 1))\) \[Superset] 
      \((\(panther(t) \[And] panther(t + 1) \[Or] 
            \[Not] panther(t) \[And] \[Not] panther(t + 1)\) \[And] 
          \(pig(t) \[And] pig(t + 1) \[Or] 
            \[Not] pig(t) \[And] \[Not] pig(t + 1)\))\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
solo3 = Implies[Or[And[panther[t],Not[panther[t+1]]],
                   And[Not[panther[t]],panther[t+1]]],
                And[Or[And[corn[t],corn[t+1]],
                       And[Not[corn[t]],Not[corn[t+1]]]],
                    Or[And[pig[t],pig[t+1]],
                       And[Not[pig[t]],Not[pig[t+1]]]]]]\
\>", "Input"],

Cell[BoxData[
    \(\((panther(t) \[And] \[Not] panther(t + 1) \[Or] 
          \[Not] panther(t) \[And] panther(t + 1))\) \[Superset] 
      \((\(corn(t) \[And] corn(t + 1) \[Or] 
            \[Not] corn(t) \[And] \[Not] corn(t + 1)\) \[And] 
          \(pig(t) \[And] pig(t + 1) \[Or] 
            \[Not] pig(t) \[And] \[Not] pig(t + 1)\))\)\)], "Output"]
}, Open  ]],

Cell["\<\
These obviously aren't clauses so what can I do with them.  I \
mentioned that CNF is a completely general form.  Not surprisingly I can \
convert any formula to an equivalent CNF formula.  I'm not going to describe \
the details of the algorithm here except to day that it involves using \
distribution and de Morgan's laws to transform the original formula into an \
equivalent one.  Think of the CNF conversion algorithm as a compiler that \
does some of the tedious work of translating a program into a form that can \
be used by a particular computer where in this case the computer is machine \
that evaluates a formula (determines if the formula is satisfiable or not and \
provides a satisfying assignment if so). The detailed algorithms are in the \
appendix at the end of this notebook.\
\>", "Text"],

Cell["\<\
In this case, I would just conjoin together the above three \
formulas to obtain a single formula, convert it to conjunctive normal form \
and then uses the clauses of this new formula to add to the set of clauses \
that I'm building for the pig and the panther problem. Again the details are \
in the appendix which is the last section of this notebook.\
\>", "Text"],

Cell[CellGroupData[{

Cell["solo = And[solo1,solo2,solo3] ", "Input"],

Cell[BoxData[
    \(\((\(pig(t) \[And] \[Not] pig(t + 1) \[Or] 
            \[Not] pig(t) \[And] pig(t + 1)\) \[Superset] 
          \(\(corn(t) \[And] corn(t + 1) \[Or] 
              \[Not] corn(t) \[And] \[Not] corn(t + 1)\) \[And] 
            \(panther(t) \[And] panther(t + 1) \[Or] 
              \[Not] panther(t) \[And] \[Not] panther(t + 1)\)\) \[And] 
        \(corn(t) \[And] \[Not] corn(t + 1) \[Or] 
            \[Not] corn(t) \[And] corn(t + 1)\) \[Superset] 
          \(\(panther(t) \[And] panther(t + 1) \[Or] 
              \[Not] panther(t) \[And] \[Not] panther(t + 1)\) \[And] 
            \(pig(t) \[And] pig(t + 1) \[Or] 
              \[Not] pig(t) \[And] \[Not] pig(t + 1)\)\) \[And] 
        \(panther(t) \[And] \[Not] panther(t + 1) \[Or] 
            \[Not] panther(t) \[And] panther(t + 1)\) \[Superset] 
          \(\(corn(t) \[And] corn(t + 1) \[Or] 
              \[Not] corn(t) \[And] \[Not] corn(t + 1)\) \[And] 
            \(pig(t) \[And] pig(t + 1) \[Or] 
              \[Not] pig(t) \[And] \[Not] pig(t + 1)\)\))\)\)], "Output"]
}, Open  ]],

Cell["\<\
We assume that the farmer and doesn't want to lose any produce, so \
we'll want to enforce the following two constraints.  First, the pig and the \
corn can't be alone without the farmer.\
\>", "Text"],

Cell[CellGroupData[{

Cell["Implies[And[pig[t],corn[t]],farmer[t]]", "Input"],

Cell[BoxData[
    \(\((pig(t) \[And] corn(t))\) \[Superset] farmer[t]\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Implies[And[Not[pig[t]],Not[corn[t]]],Not[farmer[t]]]", "Input"],

Cell[BoxData[
    \(\((\[Not] pig(t) \[And] \[Not] corn(t))\) \[Superset] 
      \(\[Not] farmer[t]\)\)], "Output"]
}, Open  ]],

Cell["Second, the pig and the panther can't be alone together.", "Text"],

Cell[CellGroupData[{

Cell["Implies[And[pig[t],panther[t]],farmer[t]]", "Input"],

Cell[BoxData[
    \(\((pig(t) \[And] panther(t))\) \[Superset] farmer[t]\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Implies[And[Not[pig[t]],Not[panther[t]]],Not[farmer[t]]]", "Input"],

Cell[BoxData[
    \(\((\[Not] pig(t) \[And] \[Not] panther(t))\) \[Superset] 
      \(\[Not] farmer[t]\)\)], "Output"]
}, Open  ]],

Cell["We can easily convert these formulas to clauses.", "Text"],

Cell[CellGroupData[{

Cell["dont1 = Or[pig[t],corn[t],Not[farmer[t]]]", "Input"],

Cell[BoxData[
    \(\((pig(t) \[Or] corn(t) \[Or] \[Not] farmer(t))\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["dont2 = Or[Not[pig[t]],Not[corn[t]],farmer[t]]", "Input"],

Cell[BoxData[
    \(\((\[Not] pig(t) \[Or] \[Not] corn(t) \[Or] farmer(t))\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["dont3 = Or[pig[t],panther[t],Not[farmer[t]]]", "Input"],

Cell[BoxData[
    \(\((pig(t) \[Or] panther(t) \[Or] \[Not] farmer(t))\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["dont4 = Or[Not[pig[t]],Not[panther[t]],farmer[t]]", "Input"],

Cell[BoxData[
    \(\((\[Not] pig(t) \[Or] \[Not] panther(t) \[Or] farmer(t))\)\)], "Output"]
}, Open  ]],

Cell[TextData[{
  StyleBox["The farmer's goal is to achieve the following for some time "],
  StyleBox["t",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox["."]
}], "Text"],

Cell[CellGroupData[{

Cell["And[panther[t],pig[t],corn[t],farmer[t]]", "Input"],

Cell[BoxData[
    \(\((panther(t) \[And] pig(t) \[And] corn(t) \[And] farmer(t))\)\)], 
  "Output"]
}, Open  ]],

Cell[TextData[{
  StyleBox["But this doesn't really help since we don't know which "],
  StyleBox["t",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[
  " will work in advance.  Finding out if there exists such a t is part of \
figuring out if this puzzle has a solution.  Knowing "],
  StyleBox["t",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[
  " would of course help but we simply don't know it.  Instead we add a \
formula of the following form."]
}], "Text"],

Cell[CellGroupData[{

Cell["\<\
Or[And[panther[1],pit[1],corn[1],farmer[1]],
   And[panther[2],pit[2],corn[2],farmer[2]],
   \"...\"
   And[panther[T],pit[T],corn[T],farmer[T]]]\
\>", "Input"],

Cell[BoxData[
    \(\((panther(1) \[And] pit(1) \[And] corn(1) \[And] farmer(1) \[Or] 
        panther(2) \[And] pit(2) \[And] corn(2) \[And] farmer(2) \[Or] 
        "..."\ \((panther(T) \[And] pit(T) \[And] corn(T) \[And] farmer(T))\))
      \)\)], "Output"]
}, Open  ]],

Cell[TextData[{
  StyleBox[
  "Can you think of a relatively easy way (without converting a big hairy \
formula to CNF) that you might express this constraint as a set of clauses by \
adding a set of additional variables?  Here's a hint.  Can you write the \
following two formulas as a set of clauses?  (By the way, "],
  StyleBox["Implies[f,g]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[" and "],
  StyleBox["Implies[g,f]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[" indicates that f and g are logically equivalent.)"]
}], "Text"],

Cell[CellGroupData[{

Cell["Implies[And[panther[t],pit[t],corn[t],farmer[t]],goal[t]]", "Input"],

Cell[BoxData[
    \(\((panther(t) \[And] pit(t) \[And] corn(t) \[And] farmer(t))\) 
      \[Superset] goal[t]\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Implies[goal[t],And[panther[t],pit[t],corn[t],farmer[t]]]", "Input"],

Cell[BoxData[
    \(goal[t] \[Superset] 
      \((panther(t) \[And] pit(t) \[And] corn(t) \[And] farmer(t))\)\)], 
  "Output"]
}, Open  ]],

Cell["\<\
If we can do it for one time point, then we can do it for any \
number of time points.  Now we need some way of linking these clauses \
together.  The following will work and it is certainly a clause.  Do these \
clauses together do the job?\
\>", "Text"],

Cell[CellGroupData[{

Cell["Or[goal[1],goal[2],\"...\",goal[T]]", "Input"],

Cell[BoxData[
    \(\((goal(1) \[Or] goal(2) \[Or] "..." \[Or] goal(T))\)\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "Now given this representation do you think you can prove that all of these \
formulas are satisfiable (and hence the big CNF formula is satisfiable) if \
and only if the puzzle has a solution in less than ",
  StyleBox["T",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " steps.  A variant of our greedy satisfiability algorithm will solve the \
problem posed above in a matter of seconds (not in interpreted ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  ", of course, but in compiled and optimized C code.  The fact that the \
formulas we build were pretty large shouldn't bother you.  They pale in \
comparison to the size of the compiled code for something as simple as \
\"Hello world\" on a Sparc.  The important aspect of the problem is that the \
size of the formula is linear in the number of time points; there are other \
sources of complexity but time is not one of them."
}], "Text"],

Cell["\<\
Well, now we have one representation for the pig and panther \
problem.  It was a bit tedious but mainly it required that we wrote down the \
statement of the problem in a logical form.  We didn't have to be \
particularly clever.  In this particular case, we actually used the notation \
of the propositional logic, but we could still use propositional logic and \
not use this notation.  Notation is just syntax, the important thing is that \
we preserce meaning, the semantics of our representation.\
\>", "Text"],

Cell["\<\
We're going to turn to the slides to think about another approach \
to representing this problem; an approach that provides insight into the \
structure of the problem and so will help us to reformulate the problem.\
\>",
   "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Some Puzzling Points to Reflect On", "Section"],

Cell["\<\
Here are some points regarding representations, logic and problem \
formulations to reflect on.  You might want to return to this section at \
different times during this course. I'm assuming that you're looked to the \
lecture notes on the graph-based formulation of the pig and the panther \
puzzle.\
\>", "Text"],

Cell["\<\
1.  Logic isn't a luxury or an alternative to something else; \
logic, along with its other mathematical cousins, linear algebra, geometry, \
and the integral and differential calculi, is the way that scientists write \
down their problems and their theories so as to communicate in precise and \
unambiguous terms.  POINT: logic and mathematical formalisms are essential to \
any science whether it be a science of the artificial (Herbert Simon's term) \
or of the natural world.\
\>", "Text"],

Cell["\<\
2.  It's not surprising that scientists and technologists working \
in artificial intelligence have considered using logic as a representation \
language or interlingua for an intelligent agent.  But these two uses are \
separate and very different.  It is not at all essential that the laws of \
intelligent (or even simply effective) thought are governed by the laws of \
logic and mathematics.  POINT: Logic can be used as a representation language \
but representation is not the same as logic.\
\>", "Text"],

Cell["\<\
3.  There is no single way to represent a problem and some \
representations are much better than others in the sense that they make it \
easier to solve a given problem.  POINT: Find the right way to represent a \
problem and it becomes trivial.\
\>", "Text"],

Cell["\<\
4.  Problem classes are what are potentially hard, not problem \
instances.  In asymptotic terms, the class of problems corresponding to \
boolean satisfiability on formulas of size 100 is easy, ditto for the class \
of all traveling salesman problems on 500 cities.  POINT: Size matters.\
\>", 
  "Text"],

Cell["\<\
5.  Given a finite class of problems, you're still stuck with \
potentially considering a very large number of possibilities for problem \
involving planning, pattern recognition, or language interpretation.  \
Sometimes you can dramatically reduce the number of possibilities by finding \
and exploiting regularities (symmetries) in the space of possibilities.  \
POINT: Finding a good representation matters.\
\>", "Text"],

Cell[TextData[{
  "6.  Finding these symmetries can be difficult.  The reason why we (human \
brains) can solve problems that our programs currently cannot is that nature \
(using the mechanisms of evolution and natural selection) has had a long time \
to discover regularities in the physical world and develop representations (",
  
  StyleBox["e.g.,",
    FontSlant->"Italic"],
  " neural organizations) to simplify our solving the challenges that nature \
confronts us with.  A good deal of the work in developing algorithms for \
speech recogition, machine vision, robot navigation involves finding those \
sources of structure in the interaction between the hardware, sensors and the \
physical world. POINT: For some real-world problems, finding the symmetries \
that distinguishes these problems from their more general NP-hard \
counterparts may be the only way that we will be able to  solve these \
problems in practice."
}], "Text"],

Cell["\<\
Aside: What could it mean for a neural network to constitute a \
representation or class of representations?\
\>", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Appendix of Procedures for Manipulating Logical Formulas", "Section"],

Cell[TextData[{
  "The following procedures implement algorithms to convert boolean formulas \
to one of two standard normal forms.  ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  "'s ",
  StyleBox["LogicalExpand",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " already implements conversion to disjunctive normal form (DNF) so most of \
the functions in this file are concerned with testing for and conversion to \
conjunctive normal form (CNF). We begin with some predicates for testing \
formulas. An atomic formula is simply a boolean variable; a literal is either \
atomic or a negated atom; a clause is a literal or a disjunction of literals; \
and a formula is in conjunctive normal form if it is either a clause or a \
conjunction of clauses."
}], "Text"],

Cell[TextData[{
  "Off[General::spell1] ; (* Just to keep ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " quiet. *)"
}], "Input"],

Cell["\<\
AtomicQ[f_] := 
  Not[MemberQ[{And,Or,Not,Implies},Head[f]]]\
\>", "Input"],

Cell["\<\
LiteralQ[f_] := 
  Or[AtomicQ[f],
     And[SameQ[Head[f],Not],AtomicQ[f[[1]]]]]\
\>", "Input"],

Cell["\<\
ClauseQ[f_] :=
  Or[LiteralQ[f],
     And[SameQ[Head[f],Or],
         Apply[And,Map[LiteralQ,Level[f,{1}]]]]]\
\>", "Input"],

Cell["\<\
CNFQ[f_] :=
  Or[ClauseQ[f],
     And[SameQ[Head[f],And],
         Apply[And,Map[ClauseQ,Level[f,{1}]]]]]\
\>", "Input"],

Cell[TextData[{
  "Reverse ",
  StyleBox["And",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and ",
  StyleBox["Or",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " for purposes of pushing a negation into a disjunction or conjunction.  ",
  StyleBox["Not[And[p,q]]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is equivalent to ",
  StyleBox["Or[Not[p],Not[q]]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and similarly ",
  StyleBox["Not[Or[p,q]]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is equivalent to ",
  StyleBox["And[Not[p],Not[q]]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ". As a simple exercise, you should convince your self that this is true."
}], "Text"],

Cell[BoxData[
    \(ReverseOrAnd[f_]\  := \ \n\ \ If[SameQ[f, And], Or, And]\)], "Input"],

Cell[TextData[{
  "The following procedure pushes all negation all the way in, ",
  StyleBox["i.e.,",
    FontSlant->"Italic"],
  " so that the formula has no negations except at the atomic level.  \
Mathematica routinely performs some simplifications so that ",
  StyleBox["Not[Not[p]]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is just ",
  StyleBox["p,",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " ",
  StyleBox["And[p]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and ",
  StyleBox["Or[p]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " are both ",
  StyleBox["p,",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " ",
  StyleBox["And[]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is ",
  StyleBox["True,",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and ",
  StyleBox["Or[]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is ",
  StyleBox["False.",
    FontFamily->"Courier",
    FontWeight->"Bold"]
}], "Text"],

Cell["\<\
NegativeIn[f_] :=
  If[LiteralQ[f],f,
     If[SameQ[Head[f],Not],
        Apply[ReverseOrAnd[Head[First[f]]],
              Map[(NegativeIn[Not[#]])&,Level[f,{2}]]],
        Apply[Head[f],Map[NegativeIn,Level[f,{1}]]]]]\
\>", "Input"],

Cell[TextData[{
  "It will simplify matters if the formulas we have to deal with consist of \
only ",
  StyleBox["Or, And, Not,",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and atomic formulas. We use the fact that ",
  StyleBox["Implies[p,q]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is logically equivalent to ",
  StyleBox["Or[Not[p],q]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " to justify the following conversion procedure."
}], "Text"],

Cell["\<\
EliminateImplications[f_] := 
  If[AtomicQ[f],f,
     If[SameQ[Head[f],Implies],
        Or[Not[EliminateImplications[f[[1]]]],
           EliminateImplications[f[[2]]]],
        Apply[Head[f],Map[EliminateImplications,Level[f,{1}]]]]]\
\>", "Input"],

Cell[TextData[{
  "As mentioned ",
  StyleBox["LogicalExpand",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " converts formulas to disjunctive normal form."
}], "Text"],

Cell["\<\
DNF[f_] := 
  LogicalExpand[EliminateImplications[f]]\
\>", "Input"],

Cell[TextData[{
  StyleBox["In the following, we'll make use of "],
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  StyleBox[
  "'s extensive symbolic manipulation capabilities to simplify our task of \
implementing a procedure for converting formulas to conjunctive normal form.  \
In particular.  we'll use Simplify to simplify nested conjunctions and \
disjunctions, and we'll also use "],
  StyleBox["Distribute",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[" to apply the distributive law for "],
  StyleBox["And",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[" and "],
  StyleBox["Or.",
    FontFamily->"Courier",
    FontWeight->"Bold"]
}], "Text"],

Cell[TextData[{
  StyleBox[
  "You are probably familiar with applying the distribution law for "],
  StyleBox["Plus",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[" ("],
  StyleBox["+",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[") and "],
  StyleBox["Times",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[" ("],
  StyleBox["*",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[") in arithmetic formulas.  Given a formula like "],
  StyleBox["((a + b) * (c * d * e)",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[
  " which is a product of sums, we can convert it to a sum of products "],
  StyleBox["(a * c * d * e) + (b * c * d * e)",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[
  " using the distribution law. In Mathematica, this is done by invoking "],
  StyleBox["Distribute[(a + b) * (c * d * e), Plus, Times]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[
  ".  Similarly, we can convert a disjunction into a conjunction, For \
example, from "],
  StyleBox["Distribute[Or[And[a,b], And[c,d,e]], And, Or]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[" we get"],
  StyleBox[" And[Or[a,c], Or[a,d], Or[a,e], Or[b,c], Or[b,d], Or[b,e]]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[
  ". You'll notice that in applying the distribution law formulas can get \
larger; in fact, in some cases they can get much larger, but there is no \
general, tractable method for avoiding such expansions (combinatorial \
explosions)."]
}], "Text"],

Cell["\<\
CNF[f_] := 
  AuxCNF[Simplify[NegativeIn[EliminateImplications[f]]]]\
\>", "Input"],

Cell["\<\
AuxCNF[f_] :=
  If[CNFQ[f], f,
    If[MatchQ[Head[f],Or],
       AuxCNF[Distribute[f,And,Or]],
       CNF[Simplify[Apply[And,
                          Map[(Distribute[#,And,Or])&,
                              Level[f,{1}]]]]]]]\
\>", "Input"],

Cell[TextData[StyleBox[
"Here are some examples of converting boolean formulas to normal forms."]], 
  "Text"],

Cell[CellGroupData[{

Cell["\<\
Module[{f1 = Not[And[Not[1],Not[Or[2,3]]]],
        f2 = Or[Not[1],Not[And[Or[1,2],3]]],
        f3 = Implies[1,Not[And[Implies[Not[1],2],3]]],
        f4 = Not[Or[1,And[2,1],And[3,4]]],
        f5 = Or[And[1,Or[2,3]],And[4,5],Or[6,7]],
        f6 = Not[And[Or[1,And[2,3]],4,Or[5,And[6,7],8]]]},
  Scan[(Print[\"Original formula: \",FullForm[#]] ;
        Print[\"CNF formula: \",FullForm[CNF[#]]] ;
        Print[\"DNF formula: \",FullForm[DNF[#]]])&, 
       {f1,f2,f3,f4,f5,f6}]]\
\>", "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Original formula: \"\>", "\[InvisibleSpace]", 
        TagBox[
          StyleBox[\(Not[And[Not[1], Not[Or[2, 3]]]]\),
            ShowSpecialCharacters->False,
            ShowStringCharacters->True],
          FullForm]}],
      SequenceForm[ "Original formula: ", 
        FullForm[ 
          Not[ 
            And[ 
              Not[ 1], 
              Not[ 
                Or[ 2, 3]]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"CNF formula: \"\>", "\[InvisibleSpace]", 
        TagBox[
          StyleBox[\(Or[1, 2, 3]\),
            ShowSpecialCharacters->False,
            ShowStringCharacters->True],
          FullForm]}],
      SequenceForm[ "CNF formula: ", 
        FullForm[ 
          Or[ 1, 2, 3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"DNF formula: \"\>", "\[InvisibleSpace]", 
        TagBox[
          StyleBox[\(Or[1, 2, 3]\),
            ShowSpecialCharacters->False,
            ShowStringCharacters->True],
          FullForm]}],
      SequenceForm[ "DNF formula: ", 
        FullForm[ 
          Or[ 1, 2, 3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Original formula: \"\>", "\[InvisibleSpace]", 
        TagBox[
          StyleBox[\(Or[Not[1], Not[And[Or[1, 2], 3]]]\),
            ShowSpecialCharacters->False,
            ShowStringCharacters->True],
          FullForm]}],
      SequenceForm[ "Original formula: ", 
        FullForm[ 
          Or[ 
            Not[ 1], 
            Not[ 
              And[ 
                Or[ 1, 2], 3]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"CNF formula: \"\>", "\[InvisibleSpace]", 
        TagBox[
          
          StyleBox[
            \(And[Or[Not[1], Not[1], Not[3]], Or[Not[1], Not[2], Not[3]]]\),
            ShowSpecialCharacters->False,
            ShowStringCharacters->True],
          FullForm]}],
      SequenceForm[ "CNF formula: ", 
        FullForm[ 
          And[ 
            Or[ 
              Not[ 1], 
              Not[ 1], 
              Not[ 3]], 
            Or[ 
              Not[ 1], 
              Not[ 2], 
              Not[ 3]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"DNF formula: \"\>", "\[InvisibleSpace]", 
        TagBox[
          StyleBox[\(Or[Not[1], Not[3]]\),
            ShowSpecialCharacters->False,
            ShowStringCharacters->True],
          FullForm]}],
      SequenceForm[ "DNF formula: ", 
        FullForm[ 
          Or[ 
            Not[ 1], 
            Not[ 3]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Original formula: \"\>", "\[InvisibleSpace]", 
        TagBox[
          StyleBox[\(Implies[1, Not[And[Implies[Not[1], 2], 3]]]\),
            ShowSpecialCharacters->False,
            ShowStringCharacters->True],
          FullForm]}],
      SequenceForm[ "Original formula: ", 
        FullForm[ 
          Implies[ 1, 
            Not[ 
              And[ 
                Implies[ 
                  Not[ 1], 2], 3]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"CNF formula: \"\>", "\[InvisibleSpace]", 
        TagBox[
          
          StyleBox[
            \(And[Or[Not[1], Not[1], Not[3]], Or[Not[1], Not[2], Not[3]]]\),
            ShowSpecialCharacters->False,
            ShowStringCharacters->True],
          FullForm]}],
      SequenceForm[ "CNF formula: ", 
        FullForm[ 
          And[ 
            Or[ 
              Not[ 1], 
              Not[ 1], 
              Not[ 3]], 
            Or[ 
              Not[ 1], 
              Not[ 2], 
              Not[ 3]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"DNF formula: \"\>", "\[InvisibleSpace]", 
        TagBox[
          StyleBox[\(Or[Not[1], Not[3]]\),
            ShowSpecialCharacters->False,
            ShowStringCharacters->True],
          FullForm]}],
      SequenceForm[ "DNF formula: ", 
        FullForm[ 
          Or[ 
            Not[ 1], 
            Not[ 3]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Original formula: \"\>", "\[InvisibleSpace]", 
        TagBox[
          StyleBox[\(Not[Or[1, And[2, 1], And[3, 4]]]\),
            ShowSpecialCharacters->False,
            ShowStringCharacters->True],
          FullForm]}],
      SequenceForm[ "Original formula: ", 
        FullForm[ 
          Not[ 
            Or[ 1, 
              And[ 2, 1], 
              And[ 3, 4]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"CNF formula: \"\>", "\[InvisibleSpace]", 
        TagBox[
          StyleBox[\(And[Not[1], Or[Not[2], Not[1]], Or[Not[3], Not[4]]]\),
            ShowSpecialCharacters->False,
            ShowStringCharacters->True],
          FullForm]}],
      SequenceForm[ "CNF formula: ", 
        FullForm[ 
          And[ 
            Not[ 1], 
            Or[ 
              Not[ 2], 
              Not[ 1]], 
            Or[ 
              Not[ 3], 
              Not[ 4]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"DNF formula: \"\>", "\[InvisibleSpace]", 
        TagBox[
          StyleBox[\(Or[And[Not[1], Not[3]], And[Not[1], Not[4]]]\),
            ShowSpecialCharacters->False,
            ShowStringCharacters->True],
          FullForm]}],
      SequenceForm[ "DNF formula: ", 
        FullForm[ 
          Or[ 
            And[ 
              Not[ 1], 
              Not[ 3]], 
            And[ 
              Not[ 1], 
              Not[ 4]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Original formula: \"\>", "\[InvisibleSpace]", 
        TagBox[
          StyleBox[\(Or[And[1, Or[2, 3]], And[4, 5], 6, 7]\),
            ShowSpecialCharacters->False,
            ShowStringCharacters->True],
          FullForm]}],
      SequenceForm[ "Original formula: ", 
        FullForm[ 
          Or[ 
            And[ 1, 
              Or[ 2, 3]], 
            And[ 4, 5], 6, 7]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"CNF formula: \"\>", "\[InvisibleSpace]", 
        TagBox[
          
          StyleBox[
            \(And[Or[1, 4, 6, 7], Or[1, 5, 6, 7], Or[2, 3, 4, 6, 7], 
              Or[2, 3, 5, 6, 7]]\),
            ShowSpecialCharacters->False,
            ShowStringCharacters->True],
          FullForm]}],
      SequenceForm[ "CNF formula: ", 
        FullForm[ 
          And[ 
            Or[ 1, 4, 6, 7], 
            Or[ 1, 5, 6, 7], 
            Or[ 2, 3, 4, 6, 7], 
            Or[ 2, 3, 5, 6, 7]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"DNF formula: \"\>", "\[InvisibleSpace]", 
        TagBox[
          StyleBox[\(Or[6, 7, And[1, 2], And[1, 3], And[4, 5]]\),
            ShowSpecialCharacters->False,
            ShowStringCharacters->True],
          FullForm]}],
      SequenceForm[ "DNF formula: ", 
        FullForm[ 
          Or[ 6, 7, 
            And[ 1, 2], 
            And[ 1, 3], 
            And[ 4, 5]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Original formula: \"\>", "\[InvisibleSpace]", 
        TagBox[
          StyleBox[\(Not[And[Or[1, And[2, 3]], 4, Or[5, And[6, 7], 8]]]\),
            ShowSpecialCharacters->False,
            ShowStringCharacters->True],
          FullForm]}],
      SequenceForm[ "Original formula: ", 
        FullForm[ 
          Not[ 
            And[ 
              Or[ 1, 
                And[ 2, 3]], 4, 
              Or[ 5, 
                And[ 6, 7], 8]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"CNF formula: \"\>", "\[InvisibleSpace]", 
        TagBox[
          
          StyleBox[
            \(And[Or[Not[1], Not[4], Not[5]], 
              Or[Not[1], Not[4], Not[6], Not[7]], Or[Not[1], Not[4], Not[8]], 
              Or[Not[2], Not[3], Not[4], Not[5]], 
              Or[Not[2], Not[3], Not[4], Not[6], Not[7]], 
              Or[Not[2], Not[3], Not[4], Not[8]]]\),
            ShowSpecialCharacters->False,
            ShowStringCharacters->True],
          FullForm]}],
      SequenceForm[ "CNF formula: ", 
        FullForm[ 
          And[ 
            Or[ 
              Not[ 1], 
              Not[ 4], 
              Not[ 5]], 
            Or[ 
              Not[ 1], 
              Not[ 4], 
              Not[ 6], 
              Not[ 7]], 
            Or[ 
              Not[ 1], 
              Not[ 4], 
              Not[ 8]], 
            Or[ 
              Not[ 2], 
              Not[ 3], 
              Not[ 4], 
              Not[ 5]], 
            Or[ 
              Not[ 2], 
              Not[ 3], 
              Not[ 4], 
              Not[ 6], 
              Not[ 7]], 
            Or[ 
              Not[ 2], 
              Not[ 3], 
              Not[ 4], 
              Not[ 8]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"DNF formula: \"\>", "\[InvisibleSpace]", 
        TagBox[
          
          StyleBox[
            \(Or[And[Not[1], Not[2]], And[Not[1], Not[3]], 
              And[Not[5], Not[6], Not[8]], And[Not[5], Not[7], Not[8]], 
              Not[4]]\),
            ShowSpecialCharacters->False,
            ShowStringCharacters->True],
          FullForm]}],
      SequenceForm[ "DNF formula: ", 
        FullForm[ 
          Or[ 
            And[ 
              Not[ 1], 
              Not[ 2]], 
            And[ 
              Not[ 1], 
              Not[ 3]], 
            And[ 
              Not[ 5], 
              Not[ 6], 
              Not[ 8]], 
            And[ 
              Not[ 5], 
              Not[ 7], 
              Not[ 8]], 
            Not[ 4]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell["\<\
Specifically, here is how you might convert the formula governing \
solo river crossings into a single formula in conjunctive normal form.\
\>", 
  "Text"],

Cell[CellGroupData[{

Cell["\<\
CNF[Or[And[pig[1],corn[1],farmer[1]],
       And[pig[2],corn[2],farmer[2]]]]\
\>", "Input"],

Cell[BoxData[
    \(\((\(pig(1) \[Or] pig(2)\) \[And] \(pig(1) \[Or] corn(2)\) \[And] 
        \(pig(1) \[Or] farmer(2)\) \[And] \(corn(1) \[Or] pig(2)\) \[And] 
        \(corn(1) \[Or] corn(2)\) \[And] \(corn(1) \[Or] farmer(2)\) \[And] 
        \(farmer(1) \[Or] pig(2)\) \[And] \(farmer(1) \[Or] corn(2)\) \[And] 
        \(farmer(1) \[Or] farmer(2)\))\)\)], "Output"]
}, Open  ]],

Cell["\<\
Question: How hard is it to determine whether a clause in DNF is \
satisfiable?\
\>", "Text"],

Cell["\<\
Question: How hard is it to determine whether a clause in CNF is \
satisfiable?\
\>", "Text"],

Cell["\<\
Question: How hard is it to convert a formula from CNF to DNF?\
\>",
   "Text"],

Cell["\<\
Question: How does the size of a CNF formula relate to the size of \
an equivalent DNF formula?\
\>", "Text"],

Cell[TextData[{
  "In case you were wondering, here is how we would construct a formula for \
multiple (",
  StyleBox["n",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ") time points.  Specifically, here is how we could construct the clauses \
for the CNF formula for the following schematic expression."
}], "Text"],

Cell[CellGroupData[{

Cell["\<\
Or[And[panther[1],pig[1],corn[1],farmer[1]],
   And[panther[2],pig[2],corn[2],farmer[2]],
   \"...\"
   And[panther[T],pig[T],corn[T],farmer[T]]]\
\>", "Input"],

Cell[BoxData[
    \(\((panther(1) \[And] pig(1) \[And] corn(1) \[And] farmer(1) \[Or] 
        panther(2) \[And] pig(2) \[And] corn(2) \[And] farmer(2) \[Or] 
        "..."\ \((panther(T) \[And] pig(T) \[And] corn(T) \[And] farmer(T))\))
      \)\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "First we use ",
  StyleBox["Mathematica'",
    FontSlant->"Italic"],
  "s ",
  StyleBox["Replace",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " command to make the apropriate substitutions for ",
  StyleBox["t = 1",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " up to ",
  StyleBox["t = n.",
    FontFamily->"Courier",
    FontWeight->"Bold"]
}], "Text"],

Cell["n = 2 ;", "Input"],

Cell["\<\
f = Apply[Or,Table[And[panther[t],pig[t],
                       corn[t],farmer[t]] /. t -> i,
                   {i,1,n}]] ;\
\>", "Input"],

Cell[TextData[StyleBox[
"Then we convert the resulting formula to CNF."]], "Text"],

Cell[CellGroupData[{

Cell["f = CNF[f]", "Input"],

Cell[BoxData[
    \(\((\(panther(1) \[Or] panther(2)\) \[And] \(panther(1) \[Or] pig(2)\) 
        \[And] \(panther(1) \[Or] corn(2)\) \[And] 
        \(panther(1) \[Or] farmer(2)\) \[And] \(pig(1) \[Or] panther(2)\) 
        \[And] \(pig(1) \[Or] pig(2)\) \[And] \(pig(1) \[Or] corn(2)\) \[And] 
        \(pig(1) \[Or] farmer(2)\) \[And] \(corn(1) \[Or] panther(2)\) \[And] 
        \(corn(1) \[Or] pig(2)\) \[And] \(corn(1) \[Or] corn(2)\) \[And] 
        \(corn(1) \[Or] farmer(2)\) \[And] \(farmer(1) \[Or] panther(2)\) 
        \[And] \(farmer(1) \[Or] pig(2)\) \[And] \(farmer(1) \[Or] corn(2)\) 
        \[And] \(farmer(1) \[Or] farmer(2)\))\)\)], "Output"]
}, Open  ]],

Cell[TextData[StyleBox[
"Finally, we have to separate out the clauses into a list of clauses."]], 
  "Text"],

Cell[CellGroupData[{

Cell["\<\
c = Table[f[[i]],{i,1,Length[f]}]
\
\>", "Input"],

Cell[BoxData[
    \({\((panther(1) \[Or] panther(2))\), \((panther(1) \[Or] pig(2))\), 
      \((panther(1) \[Or] corn(2))\), \((panther(1) \[Or] farmer(2))\), 
      \((pig(1) \[Or] panther(2))\), \((pig(1) \[Or] pig(2))\), 
      \((pig(1) \[Or] corn(2))\), \((pig(1) \[Or] farmer(2))\), 
      \((corn(1) \[Or] panther(2))\), \((corn(1) \[Or] pig(2))\), 
      \((corn(1) \[Or] corn(2))\), \((corn(1) \[Or] farmer(2))\), 
      \((farmer(1) \[Or] panther(2))\), \((farmer(1) \[Or] pig(2))\), 
      \((farmer(1) \[Or] corn(2))\), \((farmer(1) \[Or] farmer(2))\)}\)], 
  "Output"]
}, Open  ]],

Cell["\<\
CNF isn't very smart about repeated terms, but we can fix this \
easily enough. \
\>", "Text"],

Cell[CellGroupData[{

Cell["CNF[Or[pig[1],pig[1]]]", "Input"],

Cell[BoxData[
    \(\((pig(1) \[Or] pig(1))\)\)], "Output"]
}, Open  ]],

Cell["\<\
RemoveRepeatedTerms[fun_[args__]] := Apply[fun,Union[{args}]]\
\>", 
  "Input"],

Cell[CellGroupData[{

Cell["RemoveRepeatedTerms[Or[pig[1],pig[1]]]", "Input"],

Cell[BoxData[
    \(pig[1]\)], "Output"]
}, Open  ]],

Cell["\<\
There are a few other utiltities that would come in handy if we \
were to go the whole way and represent the pig and the panther problem as a \
big boolean function.  For completeness, we'll list them here and demonstrate \
their use with the following boolean function.\
\>", "Text"],

Cell[CellGroupData[{

Cell["f = And[panther[1],Or[pig[1],Not[panther[1]]]]", "Input"],

Cell[BoxData[
    \(\((panther(1) \[And] \(pig(1) \[Or] \[Not] panther(1)\))\)\)], "Output"]
}, Open  ]],

Cell["\<\
Here is a procedure that extracts the variable names from a \
formula.\
\>", "Text"],

Cell["\<\
Vars[expr_] := 
    If[Not[MemberQ[{And,Or,Not},Head[expr]]],{expr},
       Apply[Union,Table[Vars[expr[[i]]],{i,1,Length[expr]}]]]\
\>", "Input"],

Cell[CellGroupData[{

Cell["Vars[f]", "Input"],

Cell[BoxData[
    \({panther[1], pig[1]}\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "Here is a procedure that given a list of variable names creates new names \
for the variables so that we can specify assignments as a list of the form ",
  StyleBox["{v[[1]],...,v[[n]]}",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell["\<\
NewNames[variableList_List,headSymbol_] :=
    Table[{variableList[[i]],
           headSymbol[[i]]},
          {i,1,Length[variableList]}]\
\>", "Input"],

Cell[CellGroupData[{

Cell["NewNames[Vars[f],v]", "Input"],

Cell[BoxData[
    \({{panther[1], v\[LeftDoubleBracket]1\[RightDoubleBracket]}, {pig[1], 
        v\[LeftDoubleBracket]2\[RightDoubleBracket]}}\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "Using the following procedure we can go back and forth between our \
mnemonic variable names, ",
  StyleBox["e.g., ",
    FontSlant->"Italic"],
  StyleBox["panther[1]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", and the cryptic names, ",
  StyleBox["e.g.,",
    FontSlant->"Italic"],
  " ",
  StyleBox["v[[1]]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", that we use internally to simplify inference."
}], "Text"],

Cell["\<\
ReplacementList[b_List,forwardFlag_] := 
    If[forwardFlag, 
       Table[b[[i]][[1]] -> b[[i]][[2]],{i,1,Length[b]}],
       Table[b[[i]][[2]] -> b[[i]][[1]],{i,1,Length[b]}]]\
\>", "Input"],

Cell[CellGroupData[{

Cell["f /. ReplacementList[NewNames[Vars[f],v],True]", "Input"],

Cell[BoxData[
    \(\((v\[LeftDoubleBracket]1\[RightDoubleBracket] \[And] 
        \(v\[LeftDoubleBracket]2\[RightDoubleBracket] \[Or] 
          \[Not] v\[LeftDoubleBracket]1\[RightDoubleBracket]\))\)\)], "Output"]
}, Open  ]],

Cell["\<\
Using GSAT as our inference procedure, let's see how we might go \
from a formula involving our mnemonic variable names to a solution in terms \
of these mnemonic names.\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\
Module[{a, b = NewNames[Vars[f],v]},
  a = GSAT[f /. ReplacementList[b,True],Length[b],8] ;
  ReplacementList[b,True] /. v -> a ]\
\>", "Input"],

Cell[BoxData[
    \({panther[1] \[Rule] True, pig[1] \[Rule] True}\)], "Output"]
}, Open  ]]
}, Closed]]
},
FrontEndVersion->"X 3.0",
ScreenRectangle->{{0, 1280}, {0, 1024}},
WindowSize->{1014, 716},
WindowMargins->{{85, Automatic}, {109, Automatic}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"Magnification"->1},
Magnification->1.5,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1709, 49, 95, 3, 144, "Title"],

Cell[CellGroupData[{
Cell[1829, 56, 56, 0, 81, "Section"],
Cell[1888, 58, 500, 8, 115, "Text"],
Cell[2391, 68, 1104, 17, 215, "Text"],
Cell[3498, 87, 755, 11, 165, "Text"],
Cell[4256, 100, 1753, 49, 115, "Text"],

Cell[CellGroupData[{
Cell[6034, 153, 71, 0, 40, "Input"],
Cell[6108, 155, 193, 5, 36, "Output"]
}, Open  ]],
Cell[6316, 163, 167, 4, 31, "Text"],

Cell[CellGroupData[{
Cell[6508, 171, 70, 0, 40, "Input"],
Cell[6581, 173, 192, 5, 36, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[6810, 183, 70, 0, 40, "Input"],
Cell[6883, 185, 192, 5, 36, "Output"]
}, Open  ]],
Cell[7090, 193, 1187, 30, 91, "Text"],

Cell[CellGroupData[{
Cell[8302, 227, 199, 6, 62, "Input",
  Evaluatable->False],
Cell[8504, 235, 191, 5, 20, "Print"],
Cell[8698, 242, 194, 6, 20, "Print"],
Cell[8895, 250, 218, 7, 20, "Print"]
}, Open  ]],
Cell[9128, 260, 1053, 22, 91, "Text"],
Cell[10184, 284, 550, 9, 67, "Text"],
Cell[10737, 295, 939, 15, 103, "Text"],

Cell[CellGroupData[{
Cell[11701, 314, 224, 8, 57, "Subsection"],
Cell[11928, 324, 288, 7, 71, "Text"],
Cell[12219, 333, 39, 0, 38, "Input"],
Cell[12261, 335, 85, 3, 58, "Input"],
Cell[12349, 340, 104, 4, 78, "Input"],
Cell[12456, 346, 652, 22, 438, "Input"],
Cell[13111, 370, 496, 7, 278, "Input"],
Cell[13610, 379, 209, 9, 178, "Input"],
Cell[13822, 390, 201, 9, 178, "Input"],
Cell[14026, 401, 113, 3, 46, "Text"],

Cell[CellGroupData[{
Cell[14164, 408, 57, 0, 38, "Input"],
Cell[14224, 410, 92, 1, 70, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[14353, 416, 58, 0, 38, "Input"],
Cell[14414, 418, 93, 1, 70, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[14544, 424, 58, 0, 38, "Input"],
Cell[14605, 426, 86, 1, 70, "Output"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[14752, 434, 94, 3, 49, "Section"],
Cell[14849, 439, 783, 23, 118, "Text"],
Cell[15635, 464, 1963, 61, 220, "Text"],
Cell[17601, 527, 35, 0, 68, "Input"],
Cell[17639, 529, 356, 14, 41, "Text"],

Cell[CellGroupData[{
Cell[18020, 547, 51, 0, 40, "Input"],
Cell[18074, 549, 211, 3, 36, "Output"]
}, Open  ]],
Cell[18300, 555, 172, 5, 20, "Text"],

Cell[CellGroupData[{
Cell[18497, 564, 71, 0, 40, "Input"],
Cell[18571, 566, 287, 4, 36, "Output"]
}, Open  ]],
Cell[18873, 573, 50, 0, 20, "Text"],

Cell[CellGroupData[{
Cell[18948, 577, 27, 0, 40, "Input"],
Cell[18978, 579, 35, 1, 36, "Output"]
}, Open  ]],
Cell[19028, 583, 69, 0, 20, "Text"],

Cell[CellGroupData[{
Cell[19122, 587, 32, 0, 40, "Input"],
Cell[19157, 589, 288, 5, 36, "Output"]
}, Open  ]],
Cell[19460, 597, 644, 15, 55, "Text"],

Cell[CellGroupData[{
Cell[20129, 616, 55, 0, 40, "Input"],
Cell[20187, 618, 215, 3, 36, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[20439, 626, 67, 0, 40, "Input"],
Cell[20509, 628, 279, 4, 36, "Output"]
}, Open  ]],
Cell[20803, 635, 1199, 32, 91, "Text"],
Cell[22005, 669, 758, 21, 55, "Text"],

Cell[CellGroupData[{
Cell[22788, 694, 58, 0, 40, "Input"],
Cell[22849, 696, 46, 1, 36, "Output"]
}, Open  ]],
Cell[22910, 700, 199, 7, 20, "Text"],

Cell[CellGroupData[{
Cell[23134, 711, 44, 0, 40, "Input"],
Cell[23181, 713, 38, 1, 36, "Output"]
}, Open  ]],
Cell[23234, 717, 521, 23, 31, "Text"],
Cell[23758, 742, 705, 16, 148, "Input"],

Cell[CellGroupData[{
Cell[24488, 762, 41, 0, 40, "Input"],
Cell[24532, 764, 839, 17, 117, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[25408, 786, 45, 0, 40, "Input"],
Cell[25456, 788, 533, 12, 68, "Print"]
}, Open  ]],
Cell[26004, 803, 142, 3, 20, "Text"],
Cell[26149, 808, 626, 11, 67, "Text"],

Cell[CellGroupData[{
Cell[26800, 823, 410, 9, 88, "Input"],
Cell[27213, 834, 559, 13, 125, "Print"]
}, Open  ]],
Cell[27787, 850, 301, 5, 43, "Text"],
Cell[28091, 857, 71, 0, 40, "Input"],
Cell[28165, 859, 74, 0, 40, "Input"],

Cell[CellGroupData[{
Cell[28264, 863, 57, 0, 40, "Input"],
Cell[28324, 865, 87, 1, 36, "Output"]
}, Open  ]],
Cell[28426, 869, 732, 19, 55, "Text"],
Cell[29161, 890, 534, 11, 55, "Text"],
Cell[29698, 903, 774, 16, 157, "Input"],
Cell[30475, 921, 1116, 36, 67, "Text"],

Cell[CellGroupData[{
Cell[31616, 961, 41, 0, 40, "Input"],
Cell[31660, 963, 54, 1, 36, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[31751, 969, 30, 0, 40, "Input"],
Cell[31784, 971, 38, 1, 36, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[31859, 977, 58, 0, 40, "Input"],
Cell[31920, 979, 53, 1, 36, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[32010, 985, 32, 0, 40, "Input"],
Cell[32045, 987, 39, 1, 36, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[32121, 993, 39, 0, 40, "Input"],
Cell[32163, 995, 208, 4, 20, "Print"],
Cell[32374, 1001, 212, 4, 20, "Print"],
Cell[32589, 1007, 206, 4, 20, "Print"],
Cell[32798, 1013, 212, 4, 20, "Print"],
Cell[33013, 1019, 192, 4, 20, "Print"],
Cell[33208, 1025, 220, 4, 20, "Print"],
Cell[33431, 1031, 210, 4, 20, "Print"],
Cell[33644, 1037, 212, 4, 20, "Print"],
Cell[33859, 1043, 192, 4, 20, "Print"],
Cell[34054, 1049, 293, 7, 20, "Print"],
Cell[34350, 1058, 210, 4, 20, "Print"],
Cell[34563, 1064, 212, 4, 20, "Print"],
Cell[34778, 1070, 192, 4, 20, "Print"],
Cell[34973, 1076, 293, 7, 20, "Print"],
Cell[35269, 1085, 54, 1, 36, "Output"]
}, Open  ]],
Cell[35338, 1089, 232, 10, 20, "Text"],

Cell[CellGroupData[{
Cell[35595, 1103, 32, 0, 40, "Input"],
Cell[35630, 1105, 38, 1, 36, "Output"]
}, Open  ]],
Cell[35683, 1109, 360, 8, 31, "Text"],
Cell[36046, 1119, 1289, 30, 103, "Text"],
Cell[37338, 1151, 1653, 58, 79, "Text"],
Cell[38994, 1211, 539, 15, 148, "Input"],
Cell[39536, 1228, 42, 0, 20, "Text"],

Cell[CellGroupData[{
Cell[39603, 1232, 36, 0, 40, "Input"],
Cell[39642, 1234, 55, 1, 36, "Output"]
}, Open  ]],
Cell[39712, 1238, 229, 10, 20, "Text"],

Cell[CellGroupData[{
Cell[39966, 1252, 32, 0, 40, "Input"],
Cell[40001, 1254, 38, 1, 36, "Output"]
}, Open  ]],
Cell[40054, 1258, 1518, 40, 103, "Text"],
Cell[41575, 1300, 740, 12, 79, "Text"],
Cell[42318, 1314, 872, 17, 79, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[43227, 1336, 49, 0, 49, "Section"],
Cell[43279, 1338, 274, 5, 65, "Text"],
Cell[43556, 1345, 418, 12, 66, "Text"],

Cell[CellGroupData[{
Cell[43999, 1361, 40, 0, 68, "Input"],
Cell[44042, 1363, 51, 1, 68, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[44130, 1369, 36, 0, 68, "Input"],
Cell[44169, 1371, 47, 1, 68, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[44253, 1377, 37, 0, 40, "Input"],
Cell[44293, 1379, 48, 1, 36, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[44378, 1385, 39, 0, 40, "Input"],
Cell[44420, 1387, 50, 1, 36, "Output"]
}, Open  ]],
Cell[44485, 1391, 921, 24, 67, "Text"],

Cell[CellGroupData[{
Cell[45431, 1419, 69, 0, 40, "Input"],
Cell[45503, 1421, 117, 2, 36, "Output"]
}, Open  ]],
Cell[45635, 1426, 49, 0, 20, "Text"],

Cell[CellGroupData[{
Cell[45709, 1430, 64, 0, 40, "Input"],
Cell[45776, 1432, 118, 2, 36, "Output"]
}, Open  ]],
Cell[45909, 1437, 209, 4, 31, "Text"],

Cell[CellGroupData[{
Cell[46143, 1445, 340, 7, 79, "Input"],
Cell[46486, 1454, 357, 6, 54, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[46880, 1465, 340, 7, 79, "Input"],
Cell[47223, 1474, 357, 6, 54, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[47617, 1485, 340, 7, 79, "Input"],
Cell[47960, 1494, 357, 6, 54, "Output"]
}, Open  ]],
Cell[48332, 1503, 820, 12, 91, "Text"],
Cell[49155, 1517, 378, 6, 43, "Text"],

Cell[CellGroupData[{
Cell[49558, 1527, 47, 0, 40, "Input"],
Cell[49608, 1529, 1067, 18, 115, "Output"]
}, Open  ]],
Cell[50690, 1550, 211, 4, 31, "Text"],

Cell[CellGroupData[{
Cell[50926, 1558, 55, 0, 40, "Input"],
Cell[50984, 1560, 83, 1, 36, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[51104, 1566, 70, 0, 40, "Input"],
Cell[51177, 1568, 115, 2, 36, "Output"]
}, Open  ]],
Cell[51307, 1573, 72, 0, 20, "Text"],

Cell[CellGroupData[{
Cell[51404, 1577, 58, 0, 40, "Input"],
Cell[51465, 1579, 86, 1, 36, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[51588, 1585, 73, 0, 40, "Input"],
Cell[51664, 1587, 118, 2, 36, "Output"]
}, Open  ]],
Cell[51797, 1592, 64, 0, 20, "Text"],

Cell[CellGroupData[{
Cell[51886, 1596, 58, 0, 40, "Input"],
Cell[51947, 1598, 83, 1, 36, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[52067, 1604, 63, 0, 40, "Input"],
Cell[52133, 1606, 90, 1, 36, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[52260, 1612, 61, 0, 40, "Input"],
Cell[52324, 1614, 86, 1, 36, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[52447, 1620, 66, 0, 40, "Input"],
Cell[52516, 1622, 93, 1, 36, "Output"]
}, Open  ]],
Cell[52624, 1626, 187, 6, 20, "Text"],

Cell[CellGroupData[{
Cell[52836, 1636, 57, 0, 40, "Input"],
Cell[52896, 1638, 99, 2, 36, "Output"]
}, Open  ]],
Cell[53010, 1643, 499, 14, 43, "Text"],

Cell[CellGroupData[{
Cell[53534, 1661, 170, 5, 62, "Input"],
Cell[53707, 1668, 260, 4, 55, "Output"]
}, Open  ]],
Cell[53982, 1675, 576, 14, 55, "Text"],

Cell[CellGroupData[{
Cell[54583, 1693, 74, 0, 40, "Input"],
Cell[54660, 1695, 123, 2, 36, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[54820, 1702, 74, 0, 40, "Input"],
Cell[54897, 1704, 126, 3, 36, "Output"]
}, Open  ]],
Cell[55038, 1710, 265, 5, 31, "Text"],

Cell[CellGroupData[{
Cell[55328, 1719, 52, 0, 40, "Input"],
Cell[55383, 1721, 87, 1, 36, "Output"]
}, Open  ]],
Cell[55485, 1725, 933, 17, 91, "Text"],
Cell[56421, 1744, 526, 8, 55, "Text"],
Cell[56950, 1754, 242, 5, 31, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[57229, 1764, 53, 0, 49, "Section"],
Cell[57285, 1766, 325, 6, 90, "Text"],
Cell[57613, 1774, 504, 8, 115, "Text"],
Cell[58120, 1784, 522, 8, 115, "Text"],
Cell[58645, 1794, 270, 5, 65, "Text"],
Cell[58918, 1801, 315, 6, 43, "Text"],
Cell[59236, 1809, 434, 7, 55, "Text"],
Cell[59673, 1818, 944, 16, 91, "Text"],
Cell[60620, 1836, 132, 3, 20, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[60789, 1844, 75, 0, 49, "Section"],
Cell[60867, 1846, 782, 16, 141, "Text"],
Cell[61652, 1864, 140, 5, 72, "Input"],
Cell[61795, 1871, 85, 3, 88, "Input"],
Cell[61883, 1876, 104, 4, 53, "Input"],
Cell[61990, 1882, 134, 5, 62, "Input"],
Cell[62127, 1889, 130, 5, 62, "Input"],
Cell[62260, 1896, 743, 26, 43, "Text"],
Cell[63006, 1924, 89, 1, 48, "Input"],
Cell[63098, 1927, 1009, 41, 43, "Text"],
Cell[64110, 1970, 244, 7, 79, "Input"],
Cell[64357, 1979, 486, 15, 43, "Text"],
Cell[64846, 1996, 260, 7, 79, "Input"],
Cell[65109, 2005, 177, 6, 20, "Text"],
Cell[65289, 2013, 78, 3, 45, "Input"],
Cell[65370, 2018, 698, 20, 54, "Text"],
Cell[66071, 2040, 1602, 47, 103, "Text"],
Cell[67676, 2089, 93, 3, 45, "Input"],
Cell[67772, 2094, 254, 8, 88, "Input"],
Cell[68029, 2104, 110, 2, 20, "Text"],

Cell[CellGroupData[{
Cell[68164, 2110, 507, 11, 114, "Input"],
Cell[68674, 2123, 486, 15, 20, "Print"],
Cell[69163, 2140, 372, 11, 20, "Print"],
Cell[69538, 2153, 372, 11, 20, "Print"],
Cell[69913, 2166, 488, 15, 20, "Print"],
Cell[70404, 2183, 616, 21, 20, "Print"],
Cell[71023, 2206, 414, 13, 20, "Print"],
Cell[71440, 2221, 514, 15, 20, "Print"],
Cell[71957, 2238, 616, 21, 20, "Print"],
Cell[72576, 2261, 414, 13, 20, "Print"],
Cell[72993, 2276, 470, 14, 20, "Print"],
Cell[73466, 2292, 558, 18, 20, "Print"],
Cell[74027, 2312, 530, 17, 20, "Print"],
Cell[74560, 2331, 478, 14, 20, "Print"],
Cell[75041, 2347, 591, 18, 27, "Print"],
Cell[75635, 2367, 474, 14, 20, "Print"],
Cell[76112, 2383, 542, 16, 20, "Print"],
Cell[76657, 2401, 1308, 46, 58, "Print"],
Cell[77968, 2449, 854, 30, 28, "Print"]
}, Open  ]],
Cell[78837, 2482, 165, 4, 31, "Text"],

Cell[CellGroupData[{
Cell[79027, 2490, 101, 3, 45, "Input"],
Cell[79131, 2495, 371, 5, 54, "Output"]
}, Open  ]],
Cell[79517, 2503, 103, 3, 20, "Text"],
Cell[79623, 2508, 103, 3, 20, "Text"],
Cell[79729, 2513, 89, 3, 20, "Text"],
Cell[79821, 2518, 119, 3, 20, "Text"],
Cell[79943, 2523, 325, 8, 31, "Text"],

Cell[CellGroupData[{
Cell[80293, 2535, 170, 5, 62, "Input"],
Cell[80466, 2542, 260, 4, 55, "Output"]
}, Open  ]],
Cell[80741, 2549, 394, 16, 20, "Text"],
Cell[81138, 2567, 24, 0, 40, "Input"],
Cell[81165, 2569, 150, 4, 53, "Input"],
Cell[81318, 2575, 82, 1, 20, "Text"],

Cell[CellGroupData[{
Cell[81425, 2580, 27, 0, 40, "Input"],
Cell[81455, 2582, 664, 9, 75, "Output"]
}, Open  ]],
Cell[82134, 2594, 108, 2, 20, "Text"],

Cell[CellGroupData[{
Cell[82267, 2600, 59, 3, 45, "Input"],
Cell[82329, 2605, 582, 9, 85, "Output"]
}, Open  ]],
Cell[82926, 2617, 104, 3, 20, "Text"],

Cell[CellGroupData[{
Cell[83055, 2624, 39, 0, 40, "Input"],
Cell[83097, 2626, 59, 1, 36, "Output"]
}, Open  ]],
Cell[83171, 2630, 89, 3, 40, "Input"],

Cell[CellGroupData[{
Cell[83285, 2637, 55, 0, 40, "Input"],
Cell[83343, 2639, 40, 1, 36, "Output"]
}, Open  ]],
Cell[83398, 2643, 294, 5, 43, "Text"],

Cell[CellGroupData[{
Cell[83717, 2652, 63, 0, 40, "Input"],
Cell[83783, 2654, 92, 1, 36, "Output"]
}, Open  ]],
Cell[83890, 2658, 94, 3, 20, "Text"],
Cell[83987, 2663, 156, 4, 53, "Input"],

Cell[CellGroupData[{
Cell[84168, 2671, 24, 0, 40, "Input"],
Cell[84195, 2673, 54, 1, 36, "Output"]
}, Open  ]],
Cell[84264, 2677, 276, 7, 31, "Text"],
Cell[84543, 2686, 164, 5, 62, "Input"],

Cell[CellGroupData[{
Cell[84732, 2695, 36, 0, 40, "Input"],
Cell[84771, 2697, 157, 2, 36, "Output"]
}, Open  ]],
Cell[84943, 2702, 460, 16, 31, "Text"],
Cell[85406, 2720, 202, 5, 62, "Input"],

Cell[CellGroupData[{
Cell[85633, 2729, 63, 0, 40, "Input"],
Cell[85699, 2731, 215, 3, 36, "Output"]
}, Open  ]],
Cell[85929, 2737, 193, 4, 31, "Text"],

Cell[CellGroupData[{
Cell[86147, 2745, 154, 4, 53, "Input"],
Cell[86304, 2751, 80, 1, 36, "Output"]
}, Open  ]]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

